!********************************************
      MODULE solve_linear_module
!********************************************
         integer, save :: sgelsy_lwork=0, sgelsy_lastrows=0, sgelsy_lastcolumns=0
	 real, allocatable, dimension(:), save :: sgelsy_work
	 integer, allocatable, dimension(:), save :: sgelsy_zeroarr
      contains
        SUBROUTINE SolveLinear(InM,Inb,x)
	  IMPLICIT NONE
	  real, dimension(:,:) :: InM
	  real, dimension(size(InM,1),size(InM,2)) :: M
	  real, dimension(:) :: Inb
	  real, dimension(size(Inb)) :: b, x
	  integer :: rows, cols, rank, info
	  IF (size(M,1)/=size(b)) then !We have a problem here...
	    x(1)=1.0E30
	    RETURN
	  endif
	  M=InM
	  b=Inb
	  rows=size(M,1)
	  cols=size(M,2)
          IF ((sgelsy_lastrows/=rows).and.(sgelsy_lastcolumns/=cols)) THEN
	    !We need to allocate new workspace. Make workspace querry
            !Parameters: rows of M, columns of M, number of right-hand sides (here 1), M, first dimension of M (~row of M), b, rows of M, Array of zeroes of length cols, precision, rank of M (output),
            ! workspace, dimension of workspace (if -1, workspace querry is assumed), info (output)  
            sgelsy_lwork=-1
	    if (allocated(sgelsy_zeroarr)) deallocate(sgelsy_zeroarr)
	    allocate(sgelsy_zeroarr(1:cols))
	    sgelsy_zeroarr=0
	    if (.not.allocated(sgelsy_work)) allocate(sgelsy_work(1))
            call sgelsy(rows,cols,1,M,rows,b,rows,sgelsy_ZeroArr,0.01,rank,sgelsy_work,sgelsy_lwork,info)
            sgelsy_lwork=nint(sgelsy_work(1))
            deallocate(sgelsy_work)
            allocate(sgelsy_work(1:sgelsy_lwork))
            sgelsy_lastrows=rows
            sgelsy_lastcolumns=cols
            M=InM
	    b=Inb
	  ENDIF
	  sgelsy_zeroarr=0
	  CALL sgelsy(rows,cols,1,M,rows,b,rows,sgelsy_ZeroArr,0.01,rank,sgelsy_work,sgelsy_lwork,info)
	  IF (Info/=0) THEN
	    x(1)=1.0E30
	  ELSE
	    x=b
	  ENDIF
	END SUBROUTINE SolveLinear

      END MODULE solve_linear_module

!**************************
       module matrices
!**************************

!********************************
!Interfaces for generic functions
!********************************
       INTERFACE multmat
         module procedure mult_rmrm, mult_imim, mult_imrm, mult_rmim, &
                          mult_rvrm, mult_ivim, mult_ivrm, mult_rvim, &
                          mult_rmrv, mult_imiv, mult_imrv, mult_rmiv
       end interface

       INTERFACE printmat
         module procedure rprintmat, iprintmat, rprint, iprint
       end interface

       INTERFACE invert
         module procedure rinvert, iinvert
       end interface

       INTERFACE from0to1
         module procedure from0to1vec, from0to1num
       end interface

       INTERFACE determinant
         module procedure rdeterminant, ideterminant
       end interface

       INTERFACE isunitmatrix
         module procedure isunitmatrixr, isunitmatrixi
       end interface

       contains

!Returns the rank of a matrix
!#######################################
      SUBROUTINE GetRank(InMatrix,m,n,OutRank)
!#######################################
      IMPLICIT NONE
      REAL, PARAMETER :: eps=0.000001
      INTEGER :: m,n,row,i,ii,iii
      INTEGER :: OutRank, pivotnr
      REAL :: mxpivot, dummy
      REAL, DIMENSION(1:m,1:n) :: InMatrix,Matrix
      Matrix=InMatrix
      OutRank=0
      row=1
      DO i=1,n
      !Find the largest pivot
        mxpivot=abs(Matrix(row,i))
        pivotnr=row
        IF (row<m) THEN
          DO ii=row+1,m
            IF (abs(Matrix(ii,i))>mxpivot) THEN
              mxpivot=abs(Matrix(ii,i))
              pivotnr=ii
            ENDIF
          ENDDO
        ENDIF
        IF (abs(mxpivot)<eps) THEN 
          CYCLE
        ENDIF
  	! If necessary, switch the first line with the line with maximum pivot
  	IF (pivotnr/=row) THEN
  	  DO ii=i,n
  	    dummy=Matrix(row,ii)
  	    Matrix(row,ii)=Matrix(pivotnr,ii)
  	    Matrix(pivotnr,ii)=dummy
  	  ENDDO
        ENDIF
  	IF (row<m) THEN
          !Eliminate all numbers in the i-th column by Gaussian elimination
  	  DO ii=row+1,m
  	    DO iii=n,i,-1
              Matrix(ii,iii)=Matrix(ii,iii)-Matrix(ii,i)*Matrix(row,iii)/Matrix(row,i)
  	    ENDDO
  	  ENDDO
  	ENDIF
  	row=row+1
      ENDDO
      Outrank=row-1   
      END SUBROUTINE GetRank

! Bring a vector to interval 0-1
!*****************************************
      FUNCTION From0to1vec(vec,eps) result(modvec)
!*****************************************
        REAL, DIMENSION(:) :: vec
        REAL, DIMENSION(1:size(vec)) :: modvec
        REAL, OPTIONAL, INTENT(IN) :: eps
        REAL :: cureps=1.0E-5
        IF (present(eps)) cureps=eps
        modvec=vec-floor(vec)
	WHERE (modvec>1.0E0-cureps) modvec=modvec-1.0E0
	WHERE (abs(modvec)<cureps) modvec=0.0E0
      END FUNCTION From0to1vec

! Bring a number to interval 0-1
!*****************************************
      FUNCTION From0to1num(num,eps) result(modnum)
!*****************************************
        REAL :: num
        REAL :: modnum
        REAL, OPTIONAL, INTENT(IN) :: eps
        REAL :: cureps=1.0E-5
        IF (present(eps)) cureps=eps
        modnum=num-floor(num)
	IF (modnum>1.0E0-cureps) modnum=modnum-1.0E0
	IF (abs(modnum)<cureps) modnum=0.0E0
      END FUNCTION From0to1num


!**************************************************
      LOGICAL FUNCTION EqualVecI(V1,V2,dim)
!**************************************************
      IMPLICIT NONE
      integer :: i,dim
      integer, dimension(1:dim) :: V1,V2
      logical :: equal
      
      equal=.true.
      DO i=1,dim
        IF (abs(V1(i)-V2(i))/=0) THEN
	  equal=.false.
	  EXIT
        ENDIF
      ENDDO
      EqualVecI=equal
      END FUNCTION EqualVecI

! Returns unit matrix
!*****************************************
      FUNCTION UnitMat(Dim)
!*****************************************
        INTEGER :: Dim,k
	INTEGER, DIMENSION(1:Dim,1:Dim) :: UnitMat
        UnitMat=0
	DO k=1,Dim
	  UnitMat(k,k)=1
	ENDDO
      END FUNCTION UnitMat

!******************************************
      LOGICAL FUNCTION IsUnitMatrixR(Matrix,eps)
!******************************************
      IMPLICIT NONE
      integer :: dim
      real, dimension(:,:) :: Matrix
      integer :: iumi,iumii
      REAL, OPTIONAL, INTENT(IN) :: eps
      REAL :: cureps=1.0E-5
      IF (present(eps)) cureps=eps
      dim=size(Matrix,1)
      IsUnitMatrixR=.true.
      DO iumi=1,Dim
        DO iumii=1,Dim
          IF (((iumi==iumii).and.((abs(Matrix(iumi,iumii)-1.0E0))>cureps)).or. &
            ((iumi/=iumii).and.(abs(Matrix(iumi,iumii))>cureps))) THEN
            IsUnitMatrixR=.false.
          ENDIF
        ENDDO
        IF (.not.IsUnitMatrixR) EXIT
      ENDDO
      END FUNCTION IsUnitMatrixR

!******************************************
      LOGICAL FUNCTION IsUnitMatrixI(Matrix)
!******************************************
      IMPLICIT NONE
      integer :: dim
      integer, dimension(:,:) :: Matrix
      integer :: iumi,iumii
      dim=size(Matrix,1)
      IsUnitMatrixI=.true.
      DO iumi=1,Dim
        DO iumii=1,Dim
          IF (((iumi==iumii).and.(Matrix(iumi,iumii)/=1)).or. &
            ((iumi/=iumii).and.(Matrix(iumi,iumii)/=0))) THEN
            IsUnitMatrixI=.false.
          ENDIF
        ENDDO
        IF (.not.IsUnitMatrixI) EXIT
      ENDDO

      END FUNCTION IsUnitMatrixI

!*****************************
! Specific functions for multiplication of matrices and vectors*matrices for various combinations of real nad integer
!*****************************
! Matrix multiplication of REAL matrices
      function mult_rmrm(a,b) result(c)
      real :: a(:,:),b(:,:),c(1:size(a,1),1:size(b,2))
      integer :: i,k
      do i=1,size(a,1)
        do k=1,size(b,2)
          c(i,k)=sum(a(i,:)*b(:,k))
        enddo
      enddo
      end function mult_rmrm

! Matrix multiplication of INTEGER matrices
      function mult_imim(a,b) result(c)
      integer :: a(:,:),b(:,:),c(1:size(a,1),1:size(b,2))
      integer :: i,k
      do i=1,size(a,1)
        do k=1,size(b,2)
          c(i,k)=sum(a(i,:)*b(:,k))
        enddo
      enddo
      end function mult_imim

! Matrix multiplication of REAL and INTEGER matrix
      function mult_rmim(a,b) result(c)
      integer :: b(:,:)
      real :: a(:,:),c(1:size(a,1),1:size(b,2))
      integer :: i,k
      do i=1,size(a,1)
        do k=1,size(b,2)
          c(i,k)=sum(a(i,:)*b(:,k))
        enddo
      enddo
      end function mult_rmim

! Matrix multiplication of INTEGER and REAL matrix
      function mult_imrm(a,b) result(c)
      integer :: a(:,:)
      real :: b(:,:),c(1:size(a,1),1:size(b,2))
      integer :: i,k
      do i=1,size(a,1)
        do k=1,size(b,2)
          c(i,k)=sum(a(i,:)*b(:,k))
        enddo
      enddo
      end function mult_imrm

! REAL row vector times REAL matrix multiplication
      function mult_rvrm(a,b) result(c)
      real :: a(:)
      real :: b(:,:),c(1:size(b,2))
      integer :: i
      do i=1,size(b,2)
        c(i)=sum(a(:)*b(:,i))
      enddo
      end function mult_rvrm

! REAL row vector times INTEGER matrix multiplication
      function mult_rvim(a,b) result(c)
      integer :: b(:,:)
      real :: a(:),c(1:size(b,2))
      integer :: i
      do i=1,size(b,2)
        c(i)=sum(a(:)*b(:,i))
      enddo
      end function mult_rvim

! INTEGER row vector times REAL matrix multiplication
      function mult_ivrm(a,b) result(c)
      integer :: a(:)
      real :: b(:,:),c(1:size(b,2))
      integer :: i
      do i=1,size(b,2)
        c(i)=sum(a(:)*b(:,i))
      enddo
      end function mult_ivrm

! INTEGER row vector times INTEGER matrix multiplication
      function mult_ivim(a,b) result(c)
      integer :: a(:),b(:,:),c(1:size(b,2))
      integer :: i
      do i=1,size(b,2)
        c(i)=sum(a(:)*b(:,i))
      enddo
      end function mult_ivim

! REAL matrix times REAL column vector multiplication
      function mult_rmrv(a,b) result(c)
      real :: a(:,:)
      real :: b(:),c(1:size(a,1))
      integer :: i
      do i=1,size(a,1)
        c(i)=sum(a(i,:)*b(:))
      enddo
      end function mult_rmrv

! REAL matrix times INTEGER column vector multiplication
      function mult_rmiv(a,b) result(c)
      integer :: b(:)
      real :: a(:,:),c(1:size(a,1))
      integer :: i
      do i=1,size(a,1)
        c(i)=sum(a(i,:)*b(:))
      enddo
      end function mult_rmiv

! INTEGER matrix times REAL column vector multiplication
      function mult_imrv(a,b) result(c)
      integer :: a(:,:)
      real :: b(:),c(1:size(a,1))
      integer :: i
      do i=1,size(a,1)
        c(i)=sum(a(i,:)*b(:))
      enddo
      end function mult_imrv

! INTEGER matrix times INTEGER column vector multiplication
      function mult_imiv(a,b) result(c)
      integer :: a(:,:),b(:),c(1:size(a,1))
      integer :: i
      do i=1,size(a,1)
        c(i)=sum(a(i,:)*b(:))
      enddo
      end function mult_imiv

!*********************************
! Matrix inversion and determinant
!*********************************
!*******************************************************
!*    LU decomposition routines used by test_lu.f90    *
!*                                                     *
!*                 F90 version by J-P Moreau, Paris    *
!* --------------------------------------------------- *
!* Reference:                                          *
!*                                                     *
!* "Numerical Recipes by W.H. Press, B. P. Flannery,   *
!*  S.A. Teukolsky and W.T. Vetterling, Cambridge      *
!*  University Press, 1986".                           *
!*                                                     * 
!*******************************************************

!  ***************************************************************
!  * Given an N x N matrix A, this routine replaces it by the LU *
!  * decomposition of a rowwise permutation of itself. A and N   *
!  * are input. INDX is an output vector which records the row   *
!  * permutation effected by the partial pivoting; D is output   *
!  * as -1 or 1, depending on whether the number of row inter-   *
!  * changes was even or odd, respectively. This routine is used *
!  * in combination with LUBKSB to solve linear equations or to  *
!  * invert a matrix. Return code is 1, if matrix is singular.   *
!  ***************************************************************
       Subroutine LUDCMP(A,N,INDX,D,CODE)
       PARAMETER(TINY=1.5E-5)
       REAL ::  AMAX,DUM, SUM, A(N,N)
       REAL, ALLOCATABLE, DIMENSION(:) :: VV
       INTEGER :: CODE, D, INDX(N)
      
       ALLOCATE(VV(1:N))
       
       D=1
       CODE=0
      
       DO I=1,N
         AMAX=0.E0
         DO J=1,N
           IF (ABS(A(I,J)).GT.AMAX) AMAX=ABS(A(I,J))
         END DO ! j loop
         IF(AMAX.LT.TINY) THEN
           CODE = 1
           RETURN
         END IF
         VV(I) = 1.E0 / AMAX
       END DO ! i loop
      
       DO J=1,N
         DO I=1,J-1
           SUM = A(I,J)
           DO K=1,I-1
             SUM = SUM - A(I,K)*A(K,J) 
           END DO ! k loop
           A(I,J) = SUM
         END DO ! i loop
         AMAX = 0.E0
         DO I=J,N
           SUM = A(I,J)
           DO K=1,J-1
             SUM = SUM - A(I,K)*A(K,J) 
           END DO ! k loop
           A(I,J) = SUM
           DUM = VV(I)*ABS(SUM)
           IF(DUM.GE.AMAX) THEN
             IMAX = I
             AMAX = DUM
           END IF
         END DO ! i loop  
         
         IF(J.NE.IMAX) THEN
           DO K=1,N
             DUM = A(IMAX,K)
             A(IMAX,K) = A(J,K)
             A(J,K) = DUM
           END DO ! k loop
           D = -D
           VV(IMAX) = VV(J)
         END IF
      
         INDX(J) = IMAX
	 IF(ABS(A(J,J)) <TINY) THEN
           CODE = 1
           RETURN
         END IF
	 
         IF(J.NE.N) THEN
           DUM = 1.E0 / A(J,J)
           DO I=J+1,N
             A(I,J) = A(I,J)*DUM
           END DO ! i loop
         END IF 

       END DO ! j loop
       DEALLOCATE(VV)
      
       RETURN
       END subroutine LUDCMP


!  ******************************************************************
!  * Solves the set of N linear equations A . X = B.  Here A is     *
!  * input, not as the matrix A but rather as its LU decomposition, *
!  * determined by the routine LUDCMP. INDX is input as the permuta-*
!  * tion vector returned by LUDCMP. B is input as the right-hand   *
!  * side vector B, and returns with the solution vector X. A, N and*
!  * INDX are not modified by this routine and can be used for suc- *
!  * cessive calls with different right-hand sides. This routine is *
!  * also efficient for plain matrix inversion.                     *
!  ******************************************************************
       Subroutine LUBKSB(A,N,INDX,B)
       REAL  SUM, A(N,N),B(N)
       INTEGER INDX(N)
       II = 0      
       DO I=1,N
         LL = INDX(I)
         SUM = B(LL)
         B(LL) = B(I)
         IF(II.NE.0) THEN
           DO J=II,I-1
             SUM = SUM - A(I,J)*B(J)
           END DO ! j loop
         ELSE IF(SUM.NE.0.E0) THEN
           II = I
         END IF
         B(I) = SUM
       END DO ! i loop
      
       DO I=N,1,-1
         SUM = B(I)
         IF(I < N) THEN
           DO J=I+1,N
             SUM = SUM - A(I,J)*B(J)
           END DO ! j loop
         END IF
         B(I) = SUM / A(I,I)
       END DO ! i loop      
       RETURN
       END subroutine LUBKSB

!***********************************************************
!*  Inversion of a real square matrix by LU decomposition  *
!*  with dynamic allocations                               *
!*                                                         *
!*                       F90 version by J-P Moreau, Paris  *
!* ------------------------------------------------------- *
!***********************************************************
      FUNCTION RINVERT(A) result (Ainv) !Real input matrix

      integer :: n,d, rc
      real ::  A(:,:)   !input real matrix (n x n)
      real ::  Ainv(1:size(A,1),1:size(A,2))   !resulting inverted matrix (n x n)
      real, dimension(1:size(A,1),1:size(A,2)) ::  Atmp  !temporary matrix, not to destroy the original
      integer, dimension(1:size(A,1)) ::  INDX  !integer vector (n)
      n=size(A,1)    
      Atmp=A !Atmp is used so that the original matrix remains unaltered. Allows also an in-place inversion, where Ainv=A in memory
      Ainv=0.0E0
      do i=1, n 
        Ainv(i,i)=1.0E0
      end do
!call LU decomposition routine (only once)
      call LUDCMP(Atmp,n,INDX,D,rc)
!call solver if previous return code is ok
!to obtain inverse of A one column at a time
      if (rc.eq.0) then
        do j=1, n
          call LUBKSB(Atmp,n,INDX,Ainv(1,J))
        end do
      else
	Ainv(1,1)=1.0E30
      endif
      END FUNCTION RINVERT

      FUNCTION IINVERT(A) result (Ainv) !Integer input matrix
      integer :: n,d, rc
      integer ::  A(:,:)   !input integer matrix (n x n)
      real ::  Ainv(1:size(A,1),1:size(A,2))   !resulting inverted matrix (n x n)
      real, dimension(1:size(A,1),1:size(A,2)) ::  Atmp  !temporary matrix, not to destroy the original
      integer, dimension(1:size(A,1)) ::  INDX  !integer vector (n)
      n=size(A,1)    
      Atmp=float(A) !Atmp is used so that the original matrix remains unaltered. Allows also an in-place inversion, where Ainv=A in memory
      Ainv=0.0E0
      do i=1, n 
        Ainv(i,i)=1.0E0
      end do
!call LU decomposition routine (only once)
      call LUDCMP(Atmp,n,INDX,D,rc)
!call solver if previous return code is ok
!to obtain inverse of A one column at a time
      if (rc.eq.0) then
        do j=1, n
          call LUBKSB(Atmp,n,INDX,Ainv(1,J))
        end do
      else
	Ainv(1,1)=1.0E30
      endif
      END FUNCTION IINVERT

!*****************************************************
!* Calculate the determinant of a real square matrix *
!* A(n,n) by Gauss method with full pivoting.        *
!* ------------------------------------------------- *
!* Ref.: "Algebre - Algorithmes et programmes en     *
!*        Pascal by Jean-Louis Jardrin, Dunod -      *
!*        Bordas Paris, 1988 p. 76-79".              *
!* ------------------------------------------------- *
!The subroutine TSRGT applies to input real square matrix A(n,n) the upper
!triangularization algorithm of Gauss method with full pivoting and keeps
!trace of successive transfoRions done in integer vectors KP and LP.
!-----------------------------------------------------------------------------
!  Input parameters:
!  eps        precision (real*8)
!  n          size of A matrix (integer)
!  A          pointer to input real square matrix (real*8)
!  Output parameters:
!  it         flag=1 if A matrix ok, =0 if A matrix is singular (integer)
!  C          pointer to table storing main diagonal elements and supra-
!             diagonal elements of upper triangular matrix and the multi-
!             plying coefficients used during triangularization process
!  KP         table storing infoRions concerning the column exchanges
!             during process (integer)
!  LP         table storing infoRions concerning the line exchanges
!             during process (integer)
!-----------------------------------------------------------------------------
!The table C is first initialized to A matrix, then receives at each step k
!of the triangularization process, usefull elements of A matrix at step k for
!k=1,2,...n.
!The variables po(real*8), lo and ko(integer) store respectively pivot at step k,
!its line number and its column number.
!------------------------------------------------------------------------------
      Subroutine TSRGT(eps, n, A, it, C, Kp, Lp)
        real eps
        integer n,it
        real A(n,n), C(n,n)
        integer Kp(n),Lp(n) 
        real  po,t0
        C=A; it=1; k=1
        do while (it==1.and.k<n)
          po=C(k,k); lo=k; ko=k
          do i=k, n
            do j=k, n
              if (abs(C(i,j))>abs(po)) then
                po=C(i,j); lo=i; ko=j
              end if
            end do
          end do
          Lp(k)=lo; Kp(k)=ko
          if (abs(po)<eps) then
            it=0
          else
            if (lo.ne.k) then
              do j=k, n
                t0=C(k,j); C(k,j)=C(lo,j); C(lo,j)=t0
              end do
            end if
            if (ko.ne.k) then
              do i=1, n
                t0=C(i,k); C(i,k)=C(i,ko); C(i,ko)=t0
              end do
            end if 
            do i=k+1, n
              C(i,k)=C(i,k)/po
              do j=k+1, n
                C(i,j)=C(i,j)-C(i,k)*C(k,j)
              end do 
            end do
            k=k+1
          end if
        end do
        if (it==1.and.abs(C(n,n))<eps)  it=0
        return
      END SUBROUTINE TSRGT

!The function DMGT returns the determinant of a real square matrix
!A(n,n) by Gauss method with full pivoting.
!----------------------------------------------------------------------------
!  Input parameters:
!  eps        precision (real*8)
!  n          size of A matrix (integer)
!  A          pointer to input real square matrix
!  Output parameters:
!  None
!-----------------------------------------------------------------------------
!The procedure TSRGT is used to reduce A matrix to an upper triangular matrix.
!Output variables are it(integer), C(n,n), Kp(n) and Lp(n).
!If it=0, matrix A is singular, if it=1, matrix A is regular. Table C contains
!at location i,j (j>=i) the corresponding element of the upper triangular matrix.
!Tables Lp and Kp contain infoRions relative to exchanges of line or column
!that occured during the process. For instance, the element number k of Lp is
!an integer <> k if an exchange of line has been made at step k (k=1,2,...,n).
!The number of exchanges of lines and columns is stored in integer L. the
!determinant of A matrix is stored in d0 (real*8).
!-----------------------------------------------------------------------------
      Function RDeterminant(A, ineps) result(d0)
      real, dimension(:,:) ::  A
      real, optional :: ineps
      real :: d0, eps=1.0E-6
      integer :: n, it
      
      real :: C(1:size(A,1),1:size(A,2))
      integer :: Kp(1:size(A,1)), Lp(1:size(A,1))
      
        IF (present(ineps)) eps=ineps
        n=size(A,1)
        call TSRGT(eps,n,A,it,C,Kp,Lp)  !call triangularization subroutine
        if (it==0) then
          d0=0.E0  !matrix singular, det=0
        else       !matrix regular, det<>0
          d0=1.E0
          do k=1, n
      	  d0=d0*C(k,k)
          end do
          l=0
          do k=1, n-1
            if (Lp(k).ne.k)  l=l+1
            if (Kp(k).ne.k)  l=l+1
          end do
          if (MOD(l,2).ne.0) d0=-d0  !l is odd
        end if
        return
      End FUNCTION RDeterminant

!-----------------------------------------------------------------------------
      Function IDeterminant(A, ineps) result(i0)
      integer, dimension(:,:) ::  A
      real, optional :: ineps
      real :: d0, eps=1.0E-6
      integer :: n, i0, it
      
      real :: C(1:size(A,1),1:size(A,2))
      integer :: Kp(1:size(A,1)), Lp(1:size(A,1))
      
        IF (present(ineps)) eps=ineps
        n=size(A,1)
        call TSRGT(eps,n,float(A),it,C,Kp,Lp)  !call triangularization subroutine
        if (it==0) then
          d0=0.E0  !matrix singular, det=0
        else       !matrix regular, det<>0
          d0=1.E0
          do k=1, n
      	  d0=d0*C(k,k)
          end do
          l=0
          do k=1, n-1
            if (Lp(k).ne.k)  l=l+1
            if (Kp(k).ne.k)  l=l+1
          end do
          if (MOD(l,2).ne.0) d0=-d0  !l is odd
        end if
	i0=nint(d0)
        return
      End FUNCTION IDeterminant


!**************************
! Specific functions for printing matrices and vectors
!**************************
       subroutine rprintmat(matrix,form)
       integer :: i,ii
       REAL, DIMENSION(:,:) :: matrix !Assumed shape array
       CHARACTER(len = *), optional :: form       
       CHARACTER(len = 12) :: prnform       
       write(prnform,'(i2)') size(matrix,2)
       IF (present(form)) THEN
	 prnform='('//trim(prnform)//trim(form)//')'
         DO i=1,size(matrix,1)
           write(*,prnform) (matrix(i,ii),ii=1,size(matrix,2))
         ENDDO
       ELSE
         DO i=1,size(matrix,1)
           write(*,*) (matrix(i,ii),ii=1,size(matrix,2))
         ENDDO
       ENDIF
       write(*,*) 
       end subroutine rprintmat

       subroutine iprintmat(matrix, form)
       integer :: i,ii
       INTEGER, DIMENSION(:,:) :: matrix !Assumed shape array
       CHARACTER(len = *), optional :: form       
       CHARACTER(len = 12) :: prnform       
       write(prnform,'(i2)') size(matrix,2)
       IF (present(form)) THEN
	 prnform='('//trim(prnform)//trim(form)//')'
         DO i=1,size(matrix,1)
           write(*,prnform) (matrix(i,ii),ii=1,size(matrix,2))
         ENDDO
       ELSE
         DO i=1,size(matrix,1)
           write(*,*) (matrix(i,ii),ii=1,size(matrix,2))
         ENDDO
       ENDIF
       write(*,*) 
       end subroutine iprintmat

       subroutine rprint(vector, form)
       integer :: i
       REAL, DIMENSION(:) :: vector !Assumed shape array
       CHARACTER(len = *), optional :: form       
       CHARACTER(len = 12) :: prnform       
       write(prnform,'(i2)') size(vector)
       IF (present(form)) THEN
	 prnform='('//trim(prnform)//trim(form)//')'
         write(*,prnform) (vector(i),i=1,size(vector))
       ELSE
         write(*,*) (vector(i),i=1,size(vector))
       ENDIF
       write(*,*) 
       end subroutine rprint

       subroutine iprint(vector, form)
       integer :: i
       INTEGER, DIMENSION(:) :: vector !Assumed shape array
       CHARACTER(len = *), optional :: form       
       CHARACTER(len = 12) :: prnform       
       write(prnform,'(i2)') size(vector)
       IF (present(form)) THEN
	 prnform='('//trim(prnform)//trim(form)//')'
         write(*,prnform) (vector(i),i=1,size(vector))
       ELSE
         write(*,*) (vector(i),i=1,size(vector))
       ENDIF
       write(*,*) 
       end subroutine iprint

       end module matrices
