!     Last change:  LP   16 Mar 2005    7:46 am
!****
!****   SUBROUTINES used for the AbInitio module

! *************************************************
      SUBROUTINE GetWeakRefl
! *************************************************

      USE SF_Module
      
      IMPLICIT NONE
      
      INTEGER, PARAMETER :: maxsatmax=15
      real, DIMENSION(0:MTotalF) :: darr
      INTEGER :: i,ii,satmax,sloop,curmax
      INTEGER :: satnr
      REAL :: curweaklimit
      REAL :: select !function returning nth element of an array
      darr(0)=0.0E0
! Find out the highest satellite index      
      satmax=0
      IF (SDim>RDim) THEN
        DO i=1,MTotalF
          DO ii=RDim+1,SDim
            satmax=max(satmax,abs(MEMRefl(i)%hklfg(ii)))
          ENDDO
        ENDDO
      ELSE
        satmax=0
      ENDIF
      IF (satmax>maxsatmax) CALL StopProgram('Problem in GetWeakRefl - too many orders of satellites.')
      IF (.not.allocated(WeakRefl)) THEN
        ALLOCATE(WeakRefl(0:satmax))
        ALLOCATE(SaveWeakRefl(0:satmax))
      ENDIF
      SaveWeakrefl=0.0
      IF (WeakRatio<eps) RETURN
      DO sloop=0,satmax
        satnr=0
        DO i=1,MTotalF
          IF (SDim>RDim) THEN
            curmax=0
            DO ii=RDim+1,SDim
              curmax=max(curmax,abs(MEMRefl(i)%hklfg(ii)))
            ENDDO
            IF (curmax==sloop) then
              satnr=satnr+1 !counts the number of satellites in the class
                  darr(satnr)=MEMRefl(i)%F
            ENDIF
          ELSE !No satellites
            satnr=satnr+1 !counts the number of reflections
                darr(satnr)=MEMRefl(i)%F
              ENDIF
            ENDDO
            IF (weakreflmode==wrcode_absolute) THEN
              SaveWeakRefl(sloop)=WeakRatio
              IF (expandedlog) THEN
                ii=0
                DO i=1,satnr
                  IF (MEMRefl(i)%F<SaveWeakRefl(sloop)) ii=ii+1
                ENDDO
                write(lgu,'(a,f10.4)') 'Fraction of reflections treated as weak:',float(ii)/float(satnr)
              ENDIF
            ELSE
          SaveWeakRefl(sloop)=select(nint(satnr*WeakRatio),satnr,darr)
            ENDIF
      END DO
      WeakRefl=SaveWeakRefl
      END SUBROUTINE  GetWeakRefl

! *************************************************
      SUBROUTINE SetWeakCodes
! *************************************************
        USE SF_Module
        IMPLICIT NONE
        integer :: i
        real :: curweaklimit
        
        curweaklimit=WeakRefl(0)
        DO i=MStartF+1,MEndF
          IF (SDim>RDim) curweaklimit=WeakRefl(abs(MEMRefl(i)%hklFG(RDim+1)))
          IF ((MEMRefl(i)%code==code_default).and.(MEMRefl(i)%F<=curweaklimit)) THEN
            MEMRefl(i)%code=code_weak
          ELSEIF ((MEMRefl(i)%code==code_weak).and.(MEMRefl(i)%F>curweaklimit)) THEN
            MEMRefl(i)%code=code_default
          ENDIF
        ENDDO
      END SUBROUTINE SetWeakCodes

! ***** the main subroutine ***********************
      SUBROUTINE ChargeFlipping
!**************************************************
      USE SF_Module
      USE SF_Interfaces
      USE SF_Convergence
      USE SF_rhoiointerfaces
      USE FFT_Interfaces

      IMPLICIT NONE
      real, PARAMETER :: StartDeltaCoef=0.005, IncDeltaCoef=1.07, DecDeltaCoef=0.9,IncRelDeltaCoef=1.01,DecRelDeltaCoef=0.99
      real, PARAMETER :: AutoDeltaChargeSmall=0.8, AutoDeltaChargeBig=1.0
      INTEGER, PARAMETER :: RFcdim=11, CheckCyc=10
      INTEGER, PARAMETER :: normalcycle=0, guessdelta=1,polishing=2
      INTEGER :: cycletype
      INTEGER :: hmstartc,hmstepc, smstartc,smstepc
      integer :: i,ii,iii,suma
      INTEGER :: cyc,idummy
      real :: CC,SumFObs,SumFCalc,SumDeltaAbsF
      real :: FobsFcalc,FobsSq,FcalcSq
      real :: FOM=0.0 !Total figure of merit of the convergence test
      logical :: ShouldContinue
      real :: AUcnt
      integer :: GetSeedFromTime !Function to get number from miliseconds
      real :: lowlimit, highlimit, dummy,diff
      real :: reldeltavalue
      real :: posave,estdelta
      real :: phase
      INTEGER, DIMENSION(0:11) :: bins
      logical :: newline, autodelta, goto50, searchsymbck, TimeForAction
      character*(linelength) :: perform,LastRecord
      character*(filelength) :: rhofiletmp
      character*(linelength) :: rhotypetmp
      
      !Resample the arrays if NVox is different from NVoxIter - can happen for repeated runs
      IF (any(NVox/=NVoxIter)) CALL ChangeGrid(NVoxIter)
      IF (modelrhotype=='random') CALL GenerateRandomModel
      !Set the parameters of the algorithm to the user-defined values
      IF (InAutoDelta) THEN
        FlipDelta=0.0
      ELSE
        FlipDelta=CfInDelta
      ENDIF
      IF (DeltaType==reldelta) THEN
        IF (InAutoDelta) THEN
          RelDeltaValue=0.83
        ELSE
          RelDeltaValue=CfInDelta
        ENDIF
      ELSE
        RelDeltaValue=0.0
      ENDIF
      Autodelta=InAutoDelta
      perform=InPerform
      maxcycles=InMaxCycles
      hmstartc=InHmStartC
      hmstepc=InHmStepC
      smstartc=InSmStartC
      smstepc=InSmStepC
      convergencemode = InConvergenceMode
      WeakRefl=SaveWeakRefl
      goodconvcyc=MaxCycles+stabicyc+1
      gdmpar=InGdmPar
      call FramedLabel(1,'Iteration')
      estdelta=0.0
      DO i=MStartF+1,MEndF
!        IF (MEMrefl(i)%code==code_skip) write(*,*) MEMrefl(i)%hklfg
        IF (MEMrefl(i)%code==code_skip) CYCLE
        IF (IndepSFPos(MEMRefl(i)%hklFG,MEMRefl(i)%SFloc)) THEN!An independent reflection
          estdelta=estdelta+MEMRefl(i)%F**2            
        ENDIF
      ENDDO
      estdelta=sqrt(2*estdelta)/Cell(NCellPar+1)
      IF (DeltaType==KedDelta) FlipDelta=FlipDelta*estdelta

      IF (AutoRS) THEN
        RandomSeed=GetSeedFromTime()
      ELSE
        RandomSeed=RandomSeed+repeats-1
      ENDIF
      CALL MY_SRAND(randomseed)
      IF (UsePhases=='code') THEN            
        READ(cwfu,'(a)') codeword
        if (mod(repeats,NCodeWords)==0) REWIND cwfu
      ENDIF  

      write(lgu,'(a,f8.4)') 'Estimated delta:',estdelta
      write(lgu,'(a,i10)')  'Random seed:    ',RandomSeed
      IF (UsePhases=='code') write(lgu,'(a)') 'Codeword: '//trim(CodeWord)
      write(lgu,'('''')')
      IF (Terminal) THEN
        write(*,*)
        write(*,'(a)') 'Starting iteration:'
        write(*,*)
        write(*,'(a,f8.4)') 'Estimated delta:',estdelta
        write(*,'(a,i10)') 'Random seed:',RandomSeed
        IF (UsePhases=='code') write(*,'(a)') 'Codeword: '//trim(CodeWord)
        write(*,*)
      ENDIF
! Method specific settings
      IF ((perform=='lde').and.(AutoDelta)) THEN
        autodelta=.false.
            FlipDelta=estdelta
      ENDIF
      if (autodelta) then
       cycletype=guessdelta
       IF (.not.neutrons) THEN !For neutrons there is no iterative delta search
          write(lgu,'(a)') 'Searching for a proper delta:'
          IF (Terminal) write(*,'(a)') 'Searching for a proper delta:'
        ENDIF
      else
        cycletype=normalcycle
      endif
50    continue
!Initialize the algorithm
        CALL CFInit
        CALL ResetConv(1)
        cyc=0
!***************
!Start iteration
!***************
100     CONTINUE
        cyc=cyc+1
        IF (SetWeakCycle==cyc) THEN
          CALL SetWeakCodes
          if ((expandedlog).and.(cyc>1)) write(lgu,'(a)') '   Using weak reflection perturbation from now on'
          if (terminal.and.(cyc>1)) write(*,'(a)') 'Using weak reflection perturbation from now on                                '
        ENDIF
        if ((abs(gdmpar%beta1+gdmpar%beta2-1.)>eps).or.(abs(gdmpar%beta1)<eps)) rhoold=rho
        IF (abs(gdmpar%beta1)>eps) THEN
          IF (abs(gdmpar%gamma1D+1.)>eps) THEN
            CALL ModifyRho(gdmpar%gamma1D)
          ENDIF
          IF (abs(gdmpar%gamma1R+1.)>eps) then
            CALL RestoreAmplitudes(gdmpar%gamma1R)
          endif
!         Attention - small inconsistency here: MEMRefl%ABEM and MEMRefl%BMEM already contain updated values (F(n+1)),
!         but MEMRefl%FMEM still contain values of G(n). Therefore the statistics can be done later. FMEM is not needed except for the statistics
!         SF is not up-to date
        ENDIF
        IF (abs(gdmpar%beta2)>eps) THEN !second half of the difference map algorithm
          rho1=rho
          rho=rhoold
          IF (abs(gdmpar%gamma2R+1.)>eps) THEN
            CALL RestoreAmplitudes(gdmpar%gamma2R)
          ENDIF
          IF (abs(gdmpar%gamma2D+1.)>eps) THEN
            CALL ModifyRho(gdmpar%gamma2D)
          ENDIF
        ENDIF
!       Combine the different rhos if needed
        IF (abs(gdmpar%beta2)>eps) THEN
          if (abs(gdmpar%beta1+gdmpar%beta2-1.)>eps) THEN
            rho=(1-gdmpar%beta1-gdmpar%beta2)*rhoold+gdmpar%beta1*rho1+gdmpar%beta2*rho
          else
           rho=gdmpar%beta1*rho1+gdmpar%beta2*rho
          endif
        ELSE !Only first part of the formula is used:
          if (abs(gdmpar%beta1-1.)>eps) THEN
            rho=(1-gdmpar%beta1)*rhoold+gdmpar%beta1*rho
          endif
        ENDIF
!        CALL FFT(rho2sf)
!        CALL SF2MEMRefl
!        CALL MEMRefl2SF
!        CALL FFT(sf2rho)

!       Calculate the R-factor, correlation coefficient and peakiness
        CALL RhoAndSFStats
        IF ((cycletype/=polishing).and.(convergencemode/='never').and.(goodconvcyc>MaxCycles)) THEN
          IF ((convergencemode=='symmetry').and. &
              (((cyc<100).and.(mod(cyc,20)==0)).or.((cyc<1000).and.(mod(cyc,100)==0)).or.(mod(cyc,100)==0))) THEN
            searchsymbck=searchsym
            searchsym=.true.
            CALL LocateSymmetryPatt(SearchSymKey,DoNotTouch,0)
            CALL MEMRefl2SF
            CALL FFT(sf2rho)
            searchsym=searchsymbck
          ENDIF
          CALL TestConvergence(cyc,FOM)
        ENDIF
!        IF (mod(cyc,50)==0) THEN
!          IF (cyc>50) THEN
!            CALL MatchOldRho
!            write(*,*)
!            write(*,'(a,f5.2)') 'Match with old rho:                                                   ',TwoDenMatch
!          ENDIF
!          rho50=rho
!        ENDIF
        !Write the per-cycle output
        CALL PerCycleOutput

        ! If we search for the proper delta, let's test
        if (autodelta.and.(cyc==CheckCyc)) THEN
          CALL CheckDelta(goto50)
          if (goto50) goto 50
        ENDIF

        IF ((hmclasses>0).and.(TimeForAction(cyc,hmstartc,hmstepc))) THEN
          CALL HistogramMatch(newline)
          IF (hmrepart) THEN
            CALL RepartitionIntensities(.true.)
          ELSE
!           Replace the amplitudes with the experimental amplitudes
            DO i=MStartF+1,MEndF
              IF (MEMRefl(i)%FMEM>eps) THEN
                MEMRefl(i)%AMEM=MEMRefl(i)%AMEM*MEMRefl(i)%F/MEMRefl(i)%FMEM
                MEMRefl(i)%BMEM=MEMRefl(i)%BMEM*MEMRefl(i)%F/MEMRefl(i)%FMEM
              ELSE
                MEMRefl(i)%AMEM=0.0E0
                MEMRefl(i)%BMEM=0.0E0
              ENDIF
            ENDDO
            CALL MEMRefl2SF
            CALL FFT(sf2rho)
          ENDIF
          IF (hmwait) THEN
            open(hmwaitu,file=hmwaitoutfile)
            DO i=1,MTotalF
              write(hmwaitu,'(3i4,2f20.10)') MEMRefl(i)%hklFG(1:3),MEMRefl(i)%FMEM,phase(MEMRefl(i)%AMEM,MEMRefl(i)%BMEM)
            ENDDO
            close(hmwaitu)
            IF (terminal) write(*,'(a)') 'Waiting for the file sf_hmimport.dat to appear...'
            DO
              CALL SLEEP(1)
              INQUIRE(file=hmwaitinfile,EXIST=ShouldContinue)
              IF (ShouldContinue) EXIT
            ENDDO
            open(hmwaitu,file=hmwaitoutfile) !A simple system-independent way of deleting the file
            close(hmwaitu,STATUS='delete')
            open(hmwaitu,file=hmwaitinfile)
            DO i=1,MTotalF
              write(*,*) i,MTotalF
              read(hmwaitu,'(3i4,2f20.10)') ii,ii,ii,MEMRefl(i)%FMEM,dummy
              MEMRefl(i)%AMEM=MEMRefl(i)%FMEM*cos(dummy)
              MEMRefl(i)%BMEM=MEMRefl(i)%FMEM*sin(dummy)
              MEMRefl(i)%F=MEMRefl(i)%FMEM
            ENDDO
            close(hmwaitu,status='delete')
            CALL MEMRefl2SF
            CALL FFT(sf2rho)
          ENDIF
        ENDIF
        IF ((symmatch).and.(TimeForAction(cyc,smstartc,smstepc))) THEN
          IF (.not.newline) THEN
            if (terminal) write(*,*)
            write(lgu,*)
          ENDIF
          searchsymbck=searchsym
          searchsym=.true.
          CALL LocateSymmetryPatt(SearchSymKey,AverageSF,1)
          searchsym=searchsymbck
   !     Calculate the updated electron density
          IF (smrepart) THEN
            CALL RepartitionIntensities(.true.)
          ELSE
!           Replace the amplitudes with the experimental amplitudes
            DO i=MStartF+1,MEndF
              IF (MEMRefl(i)%FMEM>eps) THEN
                MEMRefl(i)%AMEM=MEMRefl(i)%AMEM*MEMRefl(i)%F/MEMRefl(i)%FMEM
                MEMRefl(i)%BMEM=MEMRefl(i)%BMEM*MEMRefl(i)%F/MEMRefl(i)%FMEM
              ELSE
                MEMRefl(i)%AMEM=0.0E0
                MEMRefl(i)%BMEM=0.0E0
              ENDIF
            ENDDO
            CALL MEMRefl2SF
            CALL FFT(sf2rho)
          ENDIF
        ENDIF
        IF ((envrhotype/='none').and.(TimeForAction(cyc,envstartc,envstepc))) CALL Envelope(newline)

        IF ((ViewProgress/=noviewer).and.(TimeForAction(cyc,viewprogressperiod,viewprogressperiod))) CALL WriteForViewer

!       Temporary code to write out every density
!        buffer='tempdensity_XXXXX.xplor'
!        write(buffer(13:17),'(i5.5)') cyc
!        call writexplor(rho,trim(buffer))
!       END temporary code

        CALL ModifyParameters
       
!       Now decide, whether or not the iteration should continue
        ShouldContinue=.true.
        IF (UserBreak) THEN
          ShouldContinue=.false.
        ELSEIF (cyc==MaxCycles) THEN
          ShouldContinue=.false.
        ELSEIF ((cyc-goodconvcyc)>=stabicyc) THEN
          ShouldContinue=.false.
        ENDIF
        IF (PresModeNr>0) THEN !Should I wait a bit before continuing to the next cycle?
          DO i=1,PresModeNr
            IF (cyc<=PresModeLimits(i)) THEN
              CALL Wait(PresModeTimes(i))
              EXIT
            ENDIF
          ENDDO
        ENDIF
        IF (ShouldContinue) goto 100
!**************************
!End of the iteration cycle
!**************************    
        CALL MEMRefl2SF
!        CALL SF2MEMRefl !This updates also FMEM
        CALL SF2Refl
        CurRhoProps%Rvalue=100.*RF
        CurRhoProps%RunNumber=repeats
        CurRhoProps%Peaks=min(skewrho/refskewrho, 999.99)
        IF (cycletype/=polishing) THEN
          write(lgu,'('''')')
          IF (UserBreak) THEN
            write(lgu,'(a,i6,a)') 'Calculation was interrupted by user after',cyc,' cycles.'
            if (terminal) write(*,'(a,i6,a)') 'Calculation was interrupted by user after',cyc,' cycles.'
          ELSEIF (cyc==MaxCycles) THEN
            write(lgu,'(a,i6,a)') 'Calculation is finished. No convergence was detected after',MaxCycles,' cycles.'
            if (terminal) write(*,'(a,i6,a)') 'Calculation is finished. No convergence was detected after',MaxCycles,' cycles.'
          ELSE
            write(lgu,'(a,i6,a)') 'Calculation successfully converged after',cyc,' cycles.'
            Converged=.true.
            NConverged=NConverged+1
            if (terminal) THEN
              write(*,*)
              write(*,'(a,i6,a)') 'Calculation successfully converged after',cyc,' cycles.'
              write(*,'('''')')
            ENDIF
          ENDIF
          TotalCycles=TotalCycles+cyc-stabicyc
          write(lgu,'(a)') 'Last iteration record:'
          write(lgu,'(a)') trim(LastRecord)
          write(lgu,'('''')')
          IF (polish) THEN
            write(lgu,'(i6,a)') polishcycles,' cycles of noise suppression follow:'
            if (terminal) write(*,'(i6,a)') polishcycles,' cycles of noise suppression follow:'
            cycletype=polishing
            IF (any(NVox/=NVoxFine)) THEN
              CALL ChangeGrid(NVoxFine)
              !Reset RefSkewRho so that Peaks are on the same scale as before
              refskewrho=refskewrho*(sum(rho**3)/NVoxTot)/skewrho
            ENDIF
            hmstartc=1000000000
            smstartc=1000000000
            perform='lde'
            call SetGdmPar('lde')
            cyc=0
            maxcycles=polishcycles
            convergencemode='never'
            CALL ResetConv(1)
            WeakRefl=0.0
            !Change the reflection codes
            CALL SetWeakCodes
            goto 100
          ENDIF
        ELSE
          if (terminal) write(*,*)
        ENDIF
        IF (analpeaks) CALL AnalysePeaks

!     And this is the END
        RETURN

      CONTAINS
      
! Covers all events that might lead to a modification of iteration parameters during iteration
!**********************************
        SUBROUTINE ModifyParameters
!**********************************
        logical :: UserCom
!!!! Check possible user commands
        IF (UserCom('delta','REA')) THEN
          if (DeltaType==RelDelta) then
            reldeltavalue=com%rl
          elseif(DeltaType==KedDelta) THEN
            flipdelta=com%rl*estdelta
          else
            flipdelta=com%rl
          endif
          write(lgu,'(a)')
          write(lgu,'(a,f8.4)') 'Changing delta to',com%rl
          IF (terminal) THEN
            write(*,*)
            write(*,'(a,f8.4)') 'Changing delta to',com%rl
          ENDIF
        ENDIF
        IF (UserCom('perform','CHR')) THEN
          IF (com%chr=='lde') THEN
            perform='lde'
            write(lgu,'(a)')
            write(lgu,'(a)') 'Changing method to LDE.'
            IF (terminal) THEN
              write(*,*)
              write(*,'(a)') 'Changing method to LDE.'
            ENDIF
          ELSEIF (com%chr=='cf') THEN
            perform='cf'
            write(lgu,'(a)')
            write(lgu,'(a)') 'Changing method to charge flipping.'
            IF (terminal) THEN
              write(*,*)
              write(*,'(a)') 'Changing method to charge flipping.'
            ENDIF
          ELSE
            write(lgu,'(a,f8.4)') trim(com%chr)//' is an invalid method. Ignoring the command.'
            IF (terminal) THEN
              write(*,*)
              write(*,'(a)') trim(com%chr)//' is an invalid method. Ignoring the command.'
            ENDIF
          ENDIF
        ENDIF
        IF (UserCom('weakratio','REA')) THEN
          weakratio=com%rl
          Call Getweakrefl
          CALL SetWeakCodes
          write(lgu,'(a)')
          write(lgu,'(a,f8.4)') 'Changing weakratio to',com%rl
          IF (terminal) THEN
            write(*,*)
            write(*,'(a,f8.4)') 'Changing weakratio to',com%rl
          ENDIF
        ENDIF
        IF (UserCom('density','CHR')) THEN !Write the current snapshot of the density
           rhofiletmp=com%chr
          write(lgu,'(a)')
          write(lgu,'(a)') 'Writing the current density to file '//trim(rhofiletmp)//'.'
          if (terminal) then
            write(*,*)
            write(*,'(a)') 'Writing the current density to file '//trim(rhofiletmp)//'.'
          endif
          rhotypetmp=OutRhotype(1)
          CALL WriteRho(rho,rhofiletmp,rhotypetmp)
        ENDIF
        IF (UserCom('fullreflections','CHR')) THEN
          rhofiletmp=com%chr
          CALL WriteFullRefl(rhofiletmp)
          write(lgu,'(a)')
          write(lgu,'(a)') 'Writing the full reflection list to file '//trim(rhofiletmp)//'.'
          if (terminal) then
            write(*,*)
            write(*,'(a)') 'Writing the full reflection list to file '//trim(rhofiletmp)//'.'
          endif
          CALL WriteFullRefl(rhofiletmp)
        ENDIF

       IF (cycletype/=polishing) UserBreak=UserCom('stop','NON') !Look at the command file for a command to stop the iteration

       if (cyc==StartNeutrons) THEN
         write(lgu,'(a)')
         write(lgu,'(a)') 'Switching to mode with negative densities allowed.'
         if (terminal) then
           write(*,*)
           write(*,*) 'Switching to mode with negative densities allowed.'
         endif
         neutrons=.true.
       ENDIF

       END SUBROUTINE ModifyParameters


!**********************************
        SUBROUTINE ModifyRho(gamma)
!**********************************
        IMPLICIT NONE
        REAL :: deltadummy, gamma

        deltadummy=FlipDelta*capvx

        IF (perform=='lde') THEN
          AUcnt=0.0
          posave=LDEThreshold*flipdelta*capvx
          IF (.not.neutrons) THEN
            DO i=0,NVoxTot-1
              IF (rho(i)<0.0E0) THEN
                rho(i)=0.0E0
              ELSE
                IF (posave>eps) rho(i)=rho(i)*(1.-exp(-0.5*(rho(i)/posave)**2))
                AUcnt=AUcnt+abs(rho(i))
              ENDIF
            END DO
          ELSE
            DO i=0,NVoxTot-1
              if (abs(rho(i))<posave) rho(i)=0.0
            ENDDO
          ENDIF
        ELSEIF (perform=='workcode') THEN
          CALL RestoreConnectivity(gamma)
        ELSE
          IF (DeltaType==RelDelta) THEN
            CALL GetRelDelta(RelDeltaValue)
            deltadummy=FlipDelta*capvx          
          ENDIF
          AUcnt=0.0
          IF (.not.neutrons) THEN
            DO i=0,NVoxTot-1
              IF (rho(i)<deltadummy) rho(i)=-gamma*rho(i)
            END DO
            IF (Autodelta.and.(cyc==CheckCyc)) THEN
              DO i=0,NVoxTot-1
                IF (rho(i)<deltadummy) AUcnt=AUcnt+abs(rho(i))
              END DO
            ENDIF
          ELSE
            DO i=0,NVoxTot-1
              IF (abs(rho(i))<deltadummy) rho(i)=-gamma*rho(i)
            ENDDO
            IF (Autodelta.and.(cyc==CheckCyc)) THEN
              DO i=0,NVoxTot-1
                IF (abs(rho(i))<deltadummy) AUcnt=AUcnt+abs(rho(i))
              END DO
            ENDIF
          ENDIF
        ENDIF
        END SUBROUTINE ModifyRho

!**********************************
        SUBROUTINE GetRelDelta(RelDeltaVal)
!**********************************
          real :: reldeltaVal

          lowlimit=10000000.
          highlimit=-10000000.
          DO i=0,NVoxTot-1
            IF (neutrons) THEN
              dummy=abs(rho(i))
            ELSE
              dummy=rho(i)
            ENDIF
            lowlimit=min(lowlimit,dummy)
            highlimit=max(highlimit,dummy)
          END DO
          ii=nint(ReldeltaVal*float(NVoxTot))
1000          DO i=0,11
            bins(i)=0
          ENDDO
          diff=highlimit-lowlimit
          DO i=0,NVoxTot-1
            IF (neutrons) THEN
              dummy=abs(rho(i))
            ELSE
              dummy=rho(i)
            ENDIF
            IF (dummy<lowlimit) THEN
              bins(0)=bins(0)+1
            ELSEIF (dummy>highlimit) THEN
              bins(11)=bins(11)+1
            ELSE
              idummy=int(10.0E0*(dummy-lowlimit)/diff)+1
              bins(idummy)=bins(idummy)+1
            END IF
          END DO
          suma=0
          i=-1
          DO WHILE (suma<ii)
            i=i+1
            suma=suma+bins(i)
          END DO
          lowlimit=lowlimit+0.1*float(i-1)*diff
          highlimit=lowlimit+0.1*diff
          IF (bins(i)>10) GOTO 1000
          Flipdelta=((lowlimit+highlimit)/2.)/capvx
        END SUBROUTINE GetRelDelta

!******************************************************
        SUBROUTINE RestoreAmplitudes(gamma)
!******************************************************
        IMPLICIT NONE
        real :: gamma
        real :: amem, bmem, fmem, fobs, frat, amemw, bmemw
        real :: WilsonBound, Fwb, WBrat
            LOGICAL :: UsePhasesHere, fdeltafhere

       CALL FFT(rho2sf)
       CALL SF2MEMRefl
       IF (fdeltaf) THEN
         SF=0.0
       ELSE
         SF=-gamma*SF
       ENDIF
       ! Copy the F(000)
       IF (fixf0) THEN
         MEMRefl(MStartF)%AMEM=MEMRefl(MStartF)%A
         MEMRefl(MStartF)%BMEM=0.0
       ELSEIF ((perform=='lde').or.(cycletype==polishing)) THEN
         MEMRefl(MStartF)%AMEM=0.0
         MEMRefl(MStartF)%BMEM=0.0
       ENDIF
       IF (cycletype==polishing) THEN
         fdeltafhere=.false.
       ELSE
         IF (abs(gamma)>0.) THEN
           fdeltafhere=.true.
             ELSE
           fdeltafhere=.false.
         ENDIF
       ENDIF
       SF(0)=MEMRefl(1)%AMEM
       UsePhasesHere=((UsePhases=='always').or.((UsePhases=='number').and.(setphasecycles>=cyc)))
       IF ((MissingMode==mm_boundsum).and.(MissingNr>0)) THEN
         DO i=2, MEndF
               IF (MEMRefl(i)%code==code_missing) WBrat=WBrat+MEMRefl(i)%FMEM
         ENDDO
         WBrat=WBRat/sumWB
         IF (WBrat<1.) WBrat=1.
       ENDIF
       DO i=2,MEndF
         select case(MEMRefl(i)%code)
         CASE (code_default)
           IF(MEMRefl(i)%FMEM>eps) THEN
             frat=MEMRefl(i)%F/MEMRefl(i)%FMEM
           ELSE
             frat=0.0
           ENDIF
           IF (fdeltafhere) THEN
!             frat=max((1.+gamma)*frat-gamma,0.0)!Fobs'=(1+gamma)*fobs-gamma*fcalc
             frat=(1.+gamma)*frat-gamma!Fobs'=(1+gamma)*fobs-gamma*fcalc
           ENDIF
           amem=MEMRefl(i)%AMEM*frat
           bmem=MEMRefl(i)%BMEM*frat
         CASE (code_weak)
           amemw=MEMRefl(i)%AMEM
           bmemw=MEMRefl(i)%BMEM
           IF (IndepSFPos(MEMRefl(i)%hklFG,MEMRefl(i)%SFloc)) THEN !An independent reflection
             amem=bmemw
             bmem=-amemw
           ELSE !A reflection for which a Friedel pair exists closer to the origin
             amem=-bmemw
             bmem=amemw
           ENDIF
         CASE (code_knownph)
           IF(UsePhasesHere) THEN
             amem=MEMRefl(i)%A
             bmem=MEMRefl(i)%B
               ELSE
                 IF(MEMRefl(i)%FMEM>eps) THEN
               frat=MEMRefl(i)%F/MEMRefl(i)%FMEM
             ELSE
                   frat=0.0
                   ENDIF
             IF (fdeltafhere) frat=(1.+gamma)*frat-gamma!Fobs'=(1+gamma)*fobs-gamma*fcalc
             amem=MEMRefl(i)%AMEM*frat
             bmem=MEMRefl(i)%BMEM*frat
           ENDIF
         CASE (code_missing)
           SELECT CASE(missingmode)
             CASE (mm_zero)
               amem=0.
               bmem=0.
             CASE (mm_float)
               amem=MEMRefl(i)%AMEM
               bmem=MEMRefl(i)%BMEM
             CASE (mm_bound)
               Fwb=WilsonBound(MEMRefl(i)%hklFG)
               IF (MEMRefl(i)%FMEM>MissingBndFactor*Fwb) THEN
                  frat=MissingBndFactor*Fwb/MEMRefl(i)%FMEM
                 amem=MEMRefl(i)%AMEM*frat
                 bmem=MEMRefl(i)%BMEM*frat
                   ELSE
                 amem=MEMRefl(i)%AMEM
                 bmem=MEMRefl(i)%BMEM
               ENDIF
               !And now that we modified the reflections, we may make the reflection, if necessary
               IF (fdeltafhere) THEN
                 IF(MEMRefl(i)%FMEM>eps) THEN
                   frat=MEMRefl(i)%F/MEMRefl(i)%FMEM
                 ELSE
                   frat=0.0
                 ENDIF
                 IF (fdeltafhere) THEN
!                  frat=max((1.+gamma)*frat-gamma,0.0)!Fobs'=(1+gamma)*fobs-gamma*fcalc
                   frat=(1.+gamma)*frat-gamma!Fobs'=(1+gamma)*fobs-gamma*fcalc
                 ENDIF
                 amem=MEMRefl(i)%AMEM*frat
                 bmem=MEMRefl(i)%BMEM*frat
               ENDIF

             CASE (mm_boundsum)
               amem=MEMRefl(i)%AMEM/WBrat
               bmem=MEMRefl(i)%BMEM/WBrat
               END SELECT
         CASE (code_skip)
            amem=0.
            bmem=0.
         END SELECT
         MEMRefl(i)%AMEM=amem
         MEMRefl(i)%BMEM=bmem
         SF(MEMRefl(i)%SFloc)=amem
         SF(MEMRefl(i)%SFloc+1)=bmem
       ENDDO

        IF (detailedphases>0) THEN
          IF (mod(cyc,detailedphases)==0) THEN
            write(detphu,'(a,i4,a,f6.2)') '# cycle ',cyc,' skewness ',skewrho/refskewrho
            DO i=MStartF+1, MEndF
              IF (IndepSFPos(MEMRefl(i)%hklFG,MEMRefl(i)%SFLoc)) THEN !An independent reflection
                WRITE (detphu,'(3i4,2f14.7)') &
                   (MEMRefl(i)%hklFG(ii),ii=1,SDim), MEMRefl(i)%AMEM, -1.0*MEMRefl(i)%BMEM
              ENDIF
            END DO
          ENDIF
        ENDIF
        CALL FFT(sf2rho)

      END SUBROUTINE RestoreAmplitudes

!******************************************************
        SUBROUTINE RhoAndSFStats
!******************************************************
        real :: averho,obsbymtot,calcbymtot

        averho=capvx*MEMRefl(1)%FMEM/Cell(NCellPar+1)
        skewrho=sum((rho-averho)**3)/NVoxTot
        sigmarho=sqrt(sum((rho-averho)**2)/NVoxTot)
        IF (DeltaType==KedDelta) FlipDelta=CFInDelta*sigmarho/capvx
        if ((cyc==10).and.(cycletype/=polishing)) refskewrho=skewrho
!        if (allocated(rhoold)) then
!          epsilon=sum(abs(rho-rhoold))/NVoxTot!sum(abs(rho))
!        else
!          epsilon=-1.0E30
!             endif
        SumDeltaAbsF=0.0E0
        SumFCalc=0.0E0
        DO Cnt1=2, MEndF
          IF (MEMRefl(Cnt1)%code/=code_skip) SumDeltaAbsF=SumDeltaAbsF+ABS(MEMRefl(Cnt1)%F-MEMRefl(Cnt1)%Fmem) !For R-value
        END DO
        RF=SumDeltaAbsF/SumFobs
        IF (ConvergenceMode=='normal') THEN !Calculation of the correlation coefficient is relatively costly - use only if needed
          SumFCalc=0.0E0
          DO Cnt1=2, MEndF 
            SumFcalc=SumFcalc+MEMRefl(Cnt1)%FMEM !For CC
          END DO
          FobsFcalc=0.0E0
          FcalcSq=0.0E0
          ObsByMtot=SumFObs/MTotalF
          CalcByMtot=SumFCalc/MTotalF
          DO  Cnt1=2, MEndF
            FobsFcalc=FobsFcalc+(MEMRefl(Cnt1)%F-ObsByMTot)*(MEMRefl(Cnt1)%Fmem-CalcByMTot) !For CC
            FcalcSq=FcalcSq+(MEMRefl(Cnt1)%Fmem-CalcByMTot)**2 !For CC
          ENDDO
          CC=100.*FobsFcalc/sqrt(FobsSq*FcalcSq)
        ELSE
          CC=0.
        ENDIF
        !Fill the array CycMem with the current values
        Cnt1=mod(cyc-TestConvRefCycle-1,NMemGlobal)+1
        CycMem(1,Cnt1)=RF
        CycMem(2,Cnt1)=MEMRefl(MStartF)%FMEM
        CycMem(3,Cnt1)=skewrho/refskewrho
        CycMem(4,Cnt1)=CC
        END SUBROUTINE RhoAndSFStats

!******************************************************
        SUBROUTINE CheckDelta(resume)
!******************************************************
          logical :: resume

          IF (MEMRefl(MStartF)%AMEM>AutoDeltaChargeBig*AUCnt) THEN
            IF (DeltaType==RelDelta) THEN
              RelDeltaValue=RelDeltaValue*IncRelDeltaCoef
            ELSE
              FlipDelta=Flipdelta*IncDeltaCoef
            ENDIF
            write(lgu,'(a,f8.3,a)') 'Total/flipped ratio = ',MEMRefl(MStartF)%AMEM/AUCnt,'. Increasing delta.'
            write(lgu,'('''')')
            if (terminal) THEN
              write(*,'('''')')
              write(*,'(a,f8.3,a)') 'Total/flipped ratio = ',MEMRefl(MStartF)%AMEM/AUCnt,'. Increasing delta.'
              write(*,'('''')')
            ENDIF
            resume=.true.
          ELSEIF (MEMRefl(MStartF)%AMEM<AutoDeltaChargeSmall*AUCnt) THEN
            IF (DeltaType==RelDelta) THEN
              RelDeltaValue=RelDeltaValue*DecRelDeltaCoef
            ELSE
              FlipDelta=Flipdelta*DecDeltaCoef
            ENDIF
            write(lgu,'(a,f8.3,a)') 'Total/flipped ratio = ',MEMRefl(MStartF)%AMEM/AUCnt,'. Decreasing delta.'
            write(lgu,'('''')')
            if (terminal) THEN
              write(*,'('''')')
              write(*,'(a,f8.3,a)') 'Total/flipped ratio = ',MEMRefl(MStartF)%AMEM/AUCnt,'. Decreasing delta.'
              write(*,'('''')')
            ENDIF
            resume=.true.
          ELSE
            write(lgu,'(a,f8.3,a)') 'Total/flipped ratio = ',MEMRefl(MStartF)%AMEM/AUCnt,'.'
            write(lgu,'(a)') 'Criterion for delta fulfilled, continuing iteration.'
!            write(lgu,'(i5,a,f7.3,a,f8.2,a,f8.2,a,f6.2)') cyc,'  R:', 100.*RF, '  Charge: ',MEMRefl(MStartF)%AMEM, '(',AUCnt,')', &
!                                                    skewrho/refskewrho
            if (terminal) THEN
              write(*,'('''')')
              write(*,'(a,f8.3,a)') 'Total/flipped ratio = ',MEMRefl(MStartF)%AMEM/AUCnt,'.'
              write(*,'(a)') 'Criterion for delta fulfilled, continuing iteration.'
!             write(*,'(i5,a,f7.3,a,f10.2,a,f10.2,a,f6.2)') cyc,'  R:', 100.*RF, '  Charge: ',MEMRefl(MStartF)%AMEM, &
!                                                  '(',AUCnt,')', skewrho/refskewrho
            ENDIF
            autodelta=.false.
            resume=.false.
           ENDIF
        END SUBROUTINE CheckDelta


!******************************************************
        SUBROUTINE PerCycleOutput
!******************************************************
        logical :: SpecialCycle
        character*16 :: FOMbuffer,SymBuffer
        SpecialCycle=.false.
        IF (((cyc<100).and.(mod(cyc,10)==0)).or. &
            ((cyc<1000).and.(mod(cyc,100)==0)).or. &
            (mod(cyc,1000)==0)) SpecialCycle=.true.

        IF (CycleType/=polishing) THEN
          IF (FOM>eps) THEN
            write(FOMbuffer,'(''   Score:'',f7.2)') FOM
          ELSEIF (.not.AutoDelta) THEN
            FOMbuffer='   Score:    ---'
          ELSE
            FOMbuffer=''
          ENDIF
        ELSE
          FOMbuffer=''
        ENDIF
        IF (convergencemode=='symmetry') THEN
          write(SymBuffer,'(a,f6.2)') '  Sym:',CurRhoProps%agrfac
    ELSE
      SymBuffer=''
    ENDIF
        write(LastRecord,'(i5,a,f7.3,a,f10.2,a,f6.2,a,a)') &
              cyc,'  R:', 100.*RF, '  Charge: ',MEMRefl(MStartF)%AMEM, &
              '   Peaks:',min(skewrho/refskewrho, 999.99), trim(SymBuffer),trim(FOMbuffer)
        IF (terminal) THEN
          write(*,'(a,a1)', ADVANCE='NO') trim(LastRecord),char(13)
          IF (SpecialCycle.and..not.Autodelta) THEN
            write(*,'('''')')
            newline=.true.
          ELSE
            newline=.false.
          ENDIF
        ENDIF

        IF ((SpecialCycle.and..not.Autodelta).or.(ExpandedLog).or.((cycleType==polishing).and.(cyc==MaxCycles))) THEN
          write(lgu,'(a)') trim(LastRecord)
        ENDIF
        call plivc
        END SUBROUTINE PerCycleOutput


!**********************************
      SUBROUTINE CFInit
!**********************************
      USE SF_Module
      IMPLICIT NONE
      INTEGER :: loc,locp1,i
      real :: phase !function returning the phase from the real and imaginary components
      real :: dummy,diff,maxf, phdiff, MY_RAND
      real :: normsample,GaussDev
      logical, save :: viewerrunning=.false.

      Converged=.false.
      CurRhoProps%agrfac=100.
      refskewrho=1.
      SF=0.0E0
      SumFObs=0.0
      FObsSq=0.0
      DO i=1,MTotalF
        IF (MEMRefl(i)%code/=code_skip) SumFObs=SumFObs+MEMRefl(i)%F
      ENDDO
      DO i=1,MTotalF
        FObsSq=FObsSq+((MEMRefl(i)%F-SumFobs/MTotalF)**2)
      ENDDO
      IF (modelrhotype=='none') THEN
        IF (.not.fixf0) THEN
!        IF (.true.) THEN
          MEMRefl(MStartF)%A=0.0E0
          MEMRefl(MStartF)%B=0.0E0
          SF(0)=0.0E0
          SF(1)=0.0E0
          MEMRefl(MStartF)%AMEM=0.0E0
          MEMRefl(MStartF)%BMEM=0.0E0
        ELSE
          SF(0)=MEMRefl(MStartF)%A
          SF(1)=0.0E0
          MEMRefl(MStartF)%AMEM=SF(0)
          MEMRefl(MStartF)%BMEM=SF(1)
        ENDIF
        IF (UsePhases=='code') THEN
          CALL PhasesFromCode(codeword)
        ELSE
              DO i=MStartF+1, MEndF
            Loc=MEMRefl(i)%SFLoc
            Locp1=Loc+1
            IF ((UsePhases/='no').and.(MEMRefl(i)%code==code_knownph)) THEN
              SF(Loc)=MEMRefl(i)%A
              SF(Locp1)=MEMRefl(i)%B
            ELSEIF (MEMRefl(i)%code==code_skip) THEN
              MEMRefl(i)%A=0.
              MEMRefl(i)%B=0.
              SF(Loc)=MEMRefl(i)%A
              SF(Locp1)=MEMRefl(i)%B
            ELSE
              dummy=MY_RAND()*2*pi
              MEMRefl(i)%A=MEMRefl(i)%F*cos(dummy)
              MEMRefl(i)%B=MEMRefl(i)%F*sin(dummy)
              SF(Loc)=MEMRefl(i)%A
              SF(Locp1)=MEMRefl(i)%B
            END IF
          ENDDO
        ENDIF
          ELSE
        IF (any(NVoxModel/=NVox)) THEN
          CALL ResampleDensity(NVoxModel,NVox,modelrho)
          NVoxModel=NVox
        ENDIF
            rho=modelrho
            CALL FFT(rho2sf)
            call SF2MEMRefl
            !Repartition the input intensities according to the model, if desired
            IF (modelrepart) CALL RepartitionIntensities(.true.)
        ! Randomize the phases of MEMRefl
        SF=0.0;phdiff=0.0; diff=0.0; maxf=1.
        DO i=MStartF,MEndF
          Loc=MEMRefl(i)%SFLoc
          Locp1=Loc+1
          dummy=Phase(MEMRefl(i)%AMEM,MEMRefl(i)%BMEM)
          IF (randmodelfactor>=-eps) THEN !add a uniform random number between -randmodelfactor and +randmodelfactor
            dummy=dummy+2.*pi*(2.*MY_RAND()-1.)*randmodelfactor
          ELSEIF (randmodelfactor<eps) THEN ! add a sample from a normal distribution N(0,randmodelfactor)
            IF (MEMRefl(i)%FMEM>eps) THEN
              normsample=abs(randmodelfactor)*GaussDev()
              dummy=dummy+2*pi*normsample/MEMRefl(i)%FMEM
              dummy=dummy+2*pi*abs(randmodelfactor)/MEMRefl(i)%FMEM
              phdiff=phdiff+min(normsample**2,(MEMRefl(i)%FMEM-normsample)**2)
              diff=diff+MEMRefl(i)%FMEM**2
              if (MEMRefl(i)%FMEM>4.13*abs(randmodelfactor)) THEN
                maxf=maxf*4.13*abs(randmodelfactor)/MEMRefl(i)%FMEM
              ENDIF
            ELSE
              dummy=2.*pi*(2.*MY_RAND()-1.)
            ENDIF
          ENDIF
          MEMRefl(i)%AMEM=MEMRefl(i)%F*cos(dummy)
          MEMRefl(i)%BMEM=MEMRefl(i)%F*sin(dummy)
          SF(Loc)=MEMRefl(i)%AMEM
          SF(Locp1)=MEMRefl(i)%BMEM
        ENDDO
        IF (randmodelfactor<0.0E0) THEN
          write(*,*) 'Distance from the model:', sqrt(phdiff),sqrt(diff)
          write(*,*) 'Volume ratio:', 1./maxf
        ENDIF
      ENDIF
      CALL FFT(sf2rho)

      !The backtransform is not out-of-place -> SF must be restored (? - Check after final reorganization)
      CALL MEMRefl2SF

      IF (inperform=='workcode') THEN
        ALLOCATE(skelet(0:NVoxTot-1))
        call CalculateConnectivity(phdiff,0.0,0.9)
        write(*,*) 'Connectivity of the starting map:',phdiff
        read(*,*)
      ENDIF

      if ((autodelta).and.(FlipDelta<eps)) FlipDelta=estdelta !The moment of setting the starting delta value
      IF (DeltaType==RelDelta) THEN
        CALL GetRelDelta(RelDeltaValue)
        write(lgu,'(a,f6.3,a1)') 'Current delta = ',RelDeltavalue*100.,'%'
        if (terminal) write(*,'(a,f6.3,a1)') 'Current delta = ',RelDeltaValue*100.,'%'
      ELSEIF(DeltaType==KedDelta) THEN
        write(lgu,'(a,f10.5,a,f10.5,a)') 'Current k_ed = ',FlipDelta/estdelta, '  (absolute delta =',FlipDelta,')'
        if (terminal) write(*,'(a,f10.5,a,f10.5,a)') 'Current k_ed = ',FlipDelta/estdelta, '  (absolute delta =',FlipDelta,')'
      ELSE
        write(lgu,'(a,f10.5,a,f10.5,a)') 'Current delta = ',FlipDelta, '  (k_ed =',FlipDelta/estdelta,')'
        if (terminal) write(*,'(a,f10.5,a,f10.5,a)') 'Current delta = ',FlipDelta, '  (k_ed =',FlipDelta/estdelta,')'
      ENDIF
      IF (neutrons) autodelta=.false. !No iterative search for delta in the case of neutrons
      IF (ViewProgress/=noviewer) THEN
        CALL WriteForViewer
        IF (.not.viewerrunning) THEN
          CALL RunViewer
          viewerrunning=.true.
        ENDIF
      ENDIF
!        IF (.not.allocated(rho50)) ALLOCATE(rho50(1:NVoxTot))

      END SUBROUTINE CFInit

    END SUBROUTINE ChargeFlipping



!******************************************************
      SUBROUTINE PhasesFromCode(cw)
!******************************************************
        USE SF_Module
        IMPLICIT NONE
        character*(linelength) :: cw
        integer :: i,ii, Loc, Locp1
        real :: dummy,phorig,phexpand, MY_RAND
        integer, dimension(SDim) :: trhkl
        real, dimension(SDim) :: dhkl,dtrhkl
        real, dimension(1:1) :: arone

        DO i=MStartF+1,MEndF
          MEMRefl(i)%AMEM=1.0E20 !Flag that phase hasn't been assigned
        ENDDO
        DO ii=1,len_trim(cw)
          if (cw(ii:ii)=='0') THEN
            phorig=pi
          ELSE
            phorig=0.
          ENDIF
          dhkl=float(Refl(ii)%hklfg)
          DO cnt2=1,NSymPrim !Only the "primitive" operations must be used, centering is not interesting here
            dtrhkl=multmat(dhkl,SmOp(cnt2)%R)
            trhkl=nint(dtrhkl)
            call multm(dtrhkl(1),SmOp(cnt2)%t,arone(1),1,SDim,1)
            phexpand=phorig-2.*pi*arone(1) !This contains the phase of the expanded reflection
            DO i=MStartF,MEndF !Look for the corresponding reflection in MEMRefl array
              IF (EqualVecI(trhkl(1:SDim),MEMRefl(i)%hklfg(1:SDim),SDim)) THEN
                MEMRefl(i)%AMEM=MEMRefl(i)%F*cos(phexpand)
                MEMRefl(i)%BMEM=MEMRefl(i)%F*sin(phexpand)                
                EXIT
              ENDIF
              trhkl=-trhkl
              IF (EqualVecI(trhkl(1:SDim),MEMRefl(i)%hklfg(1:SDim),SDim)) THEN
                MEMRefl(i)%AMEM=MEMRefl(i)%F*cos(-phexpand)
                MEMRefl(i)%BMEM=MEMRefl(i)%F*sin(-phexpand)                
                EXIT
              ENDIF
            ENDDO
          ENDDO
        ENDDO
!        DO i=MStartF+1,MEndF
!          write(*,*) i,(MEMRefl(i)%hklfg(ii),ii=1,SDim),MEMRefl(i)%AMEM,phase(MEMRefl(i)%AMEM,MEMRefl(i)%BMEM)
!        ENDDO
        DO i=MStartF,MEndF
          Loc=MEMRefl(i)%SFLoc
          Locp1=Loc+1
          IF (MEMRefl(i)%AMEM>1.0E19) THEN !This reflection has not been assigned a phase from code - make it random
            dummy=MY_RAND()*2*pi
            MEMRefl(i)%AMEM=MEMRefl(i)%F*cos(dummy)
            MEMRefl(i)%BMEM=MEMRefl(i)%F*sin(dummy)
          ENDIF
          SF(Loc)=MEMRefl(i)%AMEM
          SF(Locp1)=MEMRefl(i)%BMEM
        ENDDO
      END SUBROUTINE PhasesFromCode

!******************************************************
      SUBROUTINE TestConvergence(cyc, FOM)
!******************************************************
      USE SF_Module
      USE SF_Convergence
      IMPLICIT NONE
      LOGICAL :: TestConverged
      INTEGER :: i, cyc
      REAL :: FOM

      IF ((ConvergenceMode=='never').or.(goodconvcyc<MaxCycles).or.(cyc<skipstartcyc)) THEN
        RETURN
      ELSEIF (cyc==skipstartcyc) THEN
        CALL ResetConv(cyc)
        RETURN
      ELSEIF(ConvergenceMode=='peakiness') THEN
        IF (skewrho/refskewrho>goodskew) goodconvcyc=cyc
      ELSEIF(ConvergenceMode=='rvalue') THEN
        IF (RF<goodrval) goodconvcyc=cyc
      ELSEIF(ConvergenceMode=='charge') THEN
        IF (MEMRefl(MStartF)%AMEM<goodcharge) goodconvcyc=cyc
      ELSEIF(ConvergenceMode=='symmetry') THEN
        IF (currhoprops%agrfac<goodsym) goodconvcyc=cyc
      ELSEIF (ConvergenceMode=='normal') THEN
        DO i=1,NSpeed
          CALL ConvergenceFromCurves(cyc,convpar(i),FOM,TestConverged,.false.)
          IF (TestConverged) THEN
            goodconvcyc=cyc !Remember the cycle when the convergence was detected
            EXIT
          ENDIF
         ENDDO
      ENDIF
      END SUBROUTINE TestConvergence


!******************************************************
      SUBROUTINE ResetConv(cycles)
!******************************************************
        USE SF_Convergence

        integer :: i,cycles
        !filling the convpar array

        DO i=1,NSpeed
           convpar(i)%NMem=NMem(i)
           convpar(i)%delay=delay(i)
           convpar(i)%counted=counted(i)
           convpar(i)%CycMax=CycMax(i)
           convpar(i)%SlopeMax=SlopeMax(i)
           convpar(i)%CurvMax=CurvMax(i)
           convpar(i)%MinCycPeakPoints=MinCycPeakPoints(i)
           convpar(i)%MinSlopePeakPoints=MinSlopePeakPoints(i)
           convpar(i)%MinCurvPeakPoints=MinCurvPeakPoints(i)
           convpar(i)%OldPeakCrit=OldPeakCrit(i)
           convpar(i)%CycCount=0
           convpar(i)%SlopeCount=0
           convpar(i)%CurvCount1=0
           convpar(i)%CurvCount2=0
           convpar(i)%CycPeakPoints=0
           convpar(i)%CycPeakEndCycle=0
           convpar(i)%SlopePeakPoints=0
           convpar(i)%SlopePeakTopCycle=0
           convpar(i)%CurvPeak1Points=0
           convpar(i)%CurvPeak1EndCycle=0
           convpar(i)%CurvPeak1TopCycle=0
           convpar(i)%CurvPeak2Points=0
           convpar(i)%CycPeakTopRel=0.
           convpar(i)%SlopePeakTopRel=0.
           convpar(i)%SlopePeakTopAbs=0.
           convpar(i)%CurvPeak1TopRel=0.
           convpar(i)%CurvPeak1TopAbs=0.
           convpar(i)%CurvPeak2TopRel=0.
           convpar(i)%CurvPeak2TopAbs=0.
           convpar(i)%CycPeak=.false.
           convpar(i)%CycConverg=.false.
           convpar(i)%SlopePeak=.false.
           convpar(i)%SlopeConverg=.false.
           convpar(i)%CurvPeak1=.false.
           convpar(i)%CurvPeak2=.false.
           convpar(i)%CurvConverg=.false.
           convpar(i)%CycWait=.false.
           convpar(i)%SlopeWait=.false.
           convpar(i)%CurvWait=.false.
           convpar(i)%CycWaitSign=0
           convpar(i)%SlopeWaitSign=0
           convpar(i)%CurvWaitSign=0
        ENDDO
        TestConvRefCycle=cycles-1

        CycMem=0.
        SlopeTemp=0.
        SlopeMem=0.
        CurvMem=0.
        FOM=0.
        END SUBROUTINE ResetConv

!******************************************************
      SUBROUTINE ArraysFilling(cycles)
!******************************************************
      USE SF_Module
      USE SF_Convergence
      REAL, DIMENSION(1:NFOM), SAVE :: Sumx, SumSqx, Sumy, Sumxy
      INTEGER :: i,ii,cycles

      Sumx= NDeriv*(NDeriv+1)/2 ! For the Slope and Curvature calculations, we consider x = 1,2,...,NDeriv
      SumSqx=NDeriv*(NDeriv+1)*(2*NDeriv+1)/6
      Sumy=0.0
      Sumxy=0.0
!####### VALUE #######
 
 !CycMem is filled in SUBROUTINE Rfactor

!####### SLOPE #######

! Slope of NDeriv previous points for R, TotCh, score, CC
      IF (cycles-TestConvRefCycle>=NDeriv) THEN
         DO ii=1, 4   
            do i=1, NDeriv
               Sumy(ii)=Sumy(ii)+CycMem(ii,mod(cycles-TestConvRefCycle-NDeriv+i-1,NMemGlobal)+1)
               Sumxy(ii)=Sumxy(ii)+CycMem(ii,mod(cycles-TestConvRefCycle-NDeriv+i-1,NMemGlobal)+1)*i
            enddo
            !Filling SlopeTemp with slope
            SlopeTemp(ii,mod(cycles-TestConvRefCycle-1,NSlopeTemp)+1)=(Sumxy(ii)-Sumx(ii)*Sumy(ii)/NDeriv)  &
                                                                      /(SumSqx(ii)-(Sumx(ii)**2)/NDeriv)
         ENDDO
      ENDIF
! Smoothing the slope, smoothed slope stored in SlopeMem
      IF (cycles-TestConvRefCycle<NDeriv+NSlopeTemp-1) THEN
         DO ii=1,4
            SlopeMem(ii,mod(cycles-TestConvRefCycle-1,NMemGlobal)+1)=SlopeTemp(ii,mod(cycles-TestConvRefCycle-1,NSlopeTemp)+1)
         ENDDO
      ELSE
         DO ii=1, 4
            SlopeMem(ii,mod(cycles-TestConvRefCycle-1,NMemGlobal)+1)=0.0 ! Reset the current SlopeMem value to 0 before calculation
            do i=1, NSlopeTemp
               SlopeMem(ii,mod(cycles-TestConvRefCycle-1,NMemGlobal)+1)=SlopeMem(ii,mod(cycles-TestConvRefCycle-1,NMemGlobal)+1)  &
                                                                        +SlopeTemp(ii,i)
            enddo
            SlopeMem(ii,mod(cycles-TestConvRefCycle-1,NMemGlobal)+1)=SlopeMem(ii,mod(cycles-TestConvRefCycle-1,NMemGlobal)+1)  &
                                                                     /NSlopeTemp
         ENDDO
      ENDIF

!####### CURVATURE #######

! Second derivate (=slope of NDeriv previous slope points)
      Sumy=0.0      ! Reset the Sums to 0
      Sumxy=0.0
      IF (cycles-TestConvRefCycle>=NDeriv+NDeriv-1) THEN
         DO ii=1, 4   
            do i=1, NDeriv
               Sumy(ii)=Sumy(ii)+SlopeMem(ii,mod(cycles-TestConvRefCycle-NDeriv+i-1,NMemGlobal)+1)
               Sumxy(ii)=Sumxy(ii)+SlopeMem(ii,mod(cycles-TestConvRefCycle-NDeriv+i-1,NMemGlobal)+1)*i
            enddo
            !Filling CurvMem with second derivative
            CurvMem(ii,mod(cycles-TestConvRefCycle-1,NMemGlobal)+1)=(Sumxy(ii)-Sumx(ii)*Sumy(ii)/NDeriv)  &
                                                                    /(SumSqx(ii)-(Sumx(ii)**2)/NDeriv)
         ENDDO
      ENDIF

      END SUBROUTINE ArraysFilling

!******************************************************
      SUBROUTINE ConvergenceFromCurves(cycles,s,TotPeakscore,TestConverged,Debug)
!******************************************************
      USE SF_Module
      USE SF_Convergence
      IMPLICIT NONE
      INTEGER :: i,ii,cycles
      INTEGER, DIMENSION(1:NFOM) :: CycCriterion, SlopeCriterion, CurvCriterion
      REAL, DIMENSION(1:NFOM) :: CycAver, CycDev, SlopeAver, SlopeDev, CurvAver, CurvDev
      logical, dimension(1:NFOM) :: CycStop, SlopeStop, CurvStop
      REAL :: PeakTopCycleAver, CycPeakScore, SlopePeakScore, CurvPeakScore, TotPeakScore
      INTEGER :: PeakTopNbr !Number of detected peaks
      Type(convpar_type) :: s
      
      
      LOGICAL :: TestConverged,Debug
      
      CycCriterion=0
      SlopeCriterion=0
      CurvCriterion=0
      CycAver=0.0
      CycDev=0.0
      SlopeAver=0.0
      SlopeDev=0.0
      CurvAver=0.0
      CurvDev=0.0
      TestConverged=.false.
      
      CALL ArraysFilling(cycles)
      
!####### VALUE #######

! Average and standard deviation of the k-(s%NMem-s%delay) to k values (where k is the current cycle number)
! for R, Total Charge, score, Correlation Coefficient
      IF (cycles-TestConvRefCycle<(s%NMem-s%delay)) THEN
         DO ii=1,NFOM
            do i=1, cycles-TestConvRefCycle
              CycAver(ii)=CycAver(ii)+CycMem(ii,i)
            enddo
                  CycAver(ii)=CycAver(ii)/cycles-TestConvRefCycle
         ENDDO
      ELSE
         DO ii=1,NFOM !Counting, for the average, the "s%counted" first and "s%counted" last points of s%NMem, excluding the s%delay earliest points
            do i=1,s%counted !Counting the points k-(s%NMem-s%delay) to k-(s%NMem-s%delay)+s%counted
               CycAver(ii)=CycAver(ii)+CycMem(ii,mod(cycles-TestConvRefCycle-(s%NMem-s%delay)+i-1,NMemGlobal)+1)
            enddo
            do i=1,s%counted !Counting the points k-s%counted to k
               CycAver(ii)=CycAver(ii)+CycMem(ii,mod(cycles-TestConvRefCycle-(s%counted)+i-1,NMemGlobal)+1)
            enddo
            CycAver(ii)=CycAver(ii)/(2*s%counted)
         ENDDO
      ENDIF
      
      IF (cycles-TestConvRefCycle<(s%NMem-s%delay)) THEN
         DO ii=1,NFOM
           do i=1, cycles-TestConvRefCycle
             CycDev(ii)=CycDev(ii)+(CycMem(ii,i)-CycAver(ii))**2
           enddo
           IF (cycles-TestConvRefCycle>1) THEN
             CycDev(ii)=(CycDev(ii)/(cycles-TestConvRefCycle-1))**0.5
           ELSE
             CycDev(ii)=0.0
           ENDIF
         ENDDO
      ELSE
         DO ii=1,NFOM !Counting the s%counted first and s%counted last points of s%NMem, excluding the s%delay earliest points
            do i=1,s%counted !Counting the points k-(s%NMem-s%delay) to k-(s%NMem-s%delay)+s%counted
               CycDev(ii)=CycDev(ii)+(CycMem(ii,mod(cycles-TestConvRefCycle-(s%NMem-s%delay)+i-1,NMemGlobal)+1)-CycAver(ii))**2
            enddo
            do i=1,s%counted !Counting the points k-s%counted to k
               CycDev(ii)=CycDev(ii)+(CycMem(ii,mod(cycles-TestConvRefCycle-(s%counted)+i-1,NMemGlobal)+1)-CycAver(ii))**2
            enddo
            CycDev(ii)=(CycDev(ii)/((2*s%counted)-1))**0.5
         ENDDO
      ENDIF
!####### SLOPE #######
! Average and standard deviation of the k-(s%NMem-s%delay) to k values (where k is the current cycle number)
! for R, Total Charge, score, Correlation Coefficient
      IF (cycles-TestConvRefCycle<NDeriv) THEN
         SlopeAver=0.0
      ELSEIF (cycles-TestConvRefCycle<NDeriv+(s%NMem-s%delay)-1) THEN
         DO ii=1, NFOM
            do i=NDeriv, cycles-TestConvRefCycle
              SlopeAver(ii)=SlopeAver(ii)+SlopeMem(ii,mod(i-1,NMemGlobal)+1)
            enddo
                  SlopeAver(ii)=SlopeAver(ii)/(cycles-TestConvRefCycle-NDeriv+1)
         ENDDO
      ELSE
         DO ii=1, NFOM
            do i=1, (s%NMem-s%delay)
              SlopeAver(ii)=SlopeAver(ii)+SlopeMem(ii,mod(cycles-TestConvRefCycle-(s%NMem-s%delay)+i-1,NMemGlobal)+1)
            enddo
                  SlopeAver(ii)=SlopeAver(ii)/(s%NMem-s%delay)
         ENDDO
      ENDIF
      
      IF (cycles-TestConvRefCycle<NDeriv) THEN
         SlopeDev=0.0
      ELSEIF (cycles-TestConvRefCycle< NDeriv+(s%NMem-s%delay)-1) THEN
         DO ii=1, NFOM
           do i=NDeriv, cycles-TestConvRefCycle
             SlopeDev(ii)=SlopeDev(ii)+(SlopeMem(ii,mod(i-1,NMemGlobal)+1)-SlopeAver(ii))**2
           enddo
           SlopeDev(ii)=(SlopeDev(ii)/((cycles-TestConvRefCycle-NDeriv+1)-1))**0.5
         ENDDO
      ELSE
         DO ii=1, NFOM
            do i=1, (s%NMem-s%delay)
              SlopeDev(ii)=SlopeDev(ii)  &
                           +(SlopeMem(ii,mod(cycles-TestConvRefCycle-(s%NMem-s%delay)+i-1,NMemGlobal)+1)-SlopeAver(ii))**2
            enddo
            SlopeDev(ii)=(SlopeDev(ii)/((s%NMem-s%delay)-1))**0.5
         ENDDO
      ENDIF
!####### CURVATURE #######
! Average and standard deviation of the k-(s%NMem-s%delay) to k values (where k is the current cycle number)
! for R, Total Charge, score, Correlation Coefficient
      IF (cycles-TestConvRefCycle<2*NDeriv-1) THEN
         CurvAver=0.0
      ELSEIF (cycles-TestConvRefCycle<2*NDeriv-1+(s%NMem-s%delay)-1) THEN
         DO ii=1, NFOM
            do i=2*NDeriv-1, cycles-TestConvRefCycle
              CurvAver(ii)=CurvAver(ii)+CurvMem(ii,mod(i-1,NMemGlobal)+1)
            enddo
                  CurvAver(ii)=CurvAver(ii)/(cycles-TestConvRefCycle-2*NDeriv+2)
         ENDDO
      ELSE
         DO ii=1, NFOM
            do i=1, (s%NMem-s%delay)
              CurvAver(ii)=CurvAver(ii)+CurvMem(ii,mod(cycles-TestConvRefCycle-(s%NMem-s%delay)+i-1,NMemGlobal)+1)
            enddo
                  CurvAver(ii)=CurvAver(ii)/(s%NMem-s%delay)
         ENDDO
      ENDIF
      
      IF (cycles-TestConvRefCycle<2*NDeriv-1) THEN
         CurvDev=0.0
      ELSEIF (cycles-TestConvRefCycle<2*NDeriv+(s%NMem-s%delay)-2) THEN
         DO ii=1, NFOM
           do i=2*NDeriv-1, cycles-TestConvRefCycle
             CurvDev(ii)=CurvDev(ii)+(CurvMem(ii,mod(i-1,NMemGlobal)+1)-CurvAver(ii))**2
           enddo
           CurvDev(ii)=(CurvDev(ii)/((cycles-TestConvRefCycle-2*NDeriv+2)-1))**0.5
         ENDDO
      ELSE
         DO ii=1, NFOM
            do i=1, (s%NMem-s%delay)
              CurvDev(ii)=CurvDev(ii)+(CurvMem(ii,mod(cycles-TestConvRefCycle-(s%NMem-s%delay)+i-1,NMemGlobal)+1)-CurvAver(ii))**2
            enddo
            CurvDev(ii)=(CurvDev(ii)/((s%NMem-20)-1))**0.5
         ENDDO
      ENDIF
!####### Detection of the value increase/decrease ("half-peak") #######
        !Checks the WaitCheckCycle-th value; if it's too large (+ or -) then s%CycWait=.true. and nothing happens until the values are "small" again
        IF (cycles-TestConvRefCycle-s%NMem+1==WaitCheckCycle) THEN
           DO ii=1,NFOM
              IF (abs(CycMem(ii,WaitCheckCycle)-CycAver(ii))>s%CycMax*CycDev(ii)) THEN
                 s%CycWait(ii)=.true.
                 s%CycWaitSign(ii)=nint((CycMem(ii,WaitCheckCycle)-CycAver(ii))/abs(CycMem(ii,WaitCheckCycle)-CycAver(ii)))
              ELSE
                 s%CycWait(ii)=.false.
              ENDIF
           ENDDO
        ENDIF

        !Detection of the increase/decrease
        IF (cycles-TestConvRefCycle-s%NMem+1>=WaitCheckCycle) THEN

           DO ii=1,NFOM
              IF (s%CycWait(ii).eqv..true.) THEN
                 if ( s%CycWaitSign(ii)*(CycMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CycAver(ii))  &
                      <s%CycMax*CycDev(ii) ) then
                    s%CycWait(ii)=.false.
                 endif
              ENDIF
           ENDDO
           DO ii=1,NFOM
              IF (s%CycWait(ii).eqv..false.) THEN
                 IF (cycles-TestConvRefCycle>=s%NMem)THEN !When the s%NMem-th value is stored, then the search for a peak begins with the FIRST stored value
                    IF ( -1*ExpPeakSign(ii)*(CycMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CycAver(ii))  &
                         >s%CycMax*CycDev(ii) ) THEN ! The point belongs to the expected increase/decrease
                       s%CycCount(ii)=s%CycCount(ii)+1
                       ! Remembering the top (larger) value of the "half-peak" and the corresponding cycle
                       if ( (abs(CycMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CycAver(ii))/CycDev(ii)  &
                             >s%CycPeakTopRel(ii)) )then
                         s%CycPeakTopRel(ii)=abs(CycMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CycAver(ii))/CycDev(ii)
                       else
                          s%CycPeakTopRel(ii)=s%CycPeakTopRel(ii)
                       endif
                    ELSE ! The point is not part of the "half-peak"
                       IF (s%CycCount(ii)>=s%MinCycPeakPoints) THEN
                          s%CycPeak(ii)=.true. !There were already enough large points to consider them as a "half-peak"
                       ELSEIF (s%CycPeak(ii).eqv..false.) THEN
                          s%CycPeakTopRel(ii)=0.0
                       ENDIF
                       s%CycCount(ii)=0
                    ENDIF
                    IF (s%CycPeak(ii)) THEN !A "half-peak" has been detected
                       IF ( abs(CycMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CycAver(ii))  &
                            <=(s%CycMax/2.)*CycDev(ii) ) THEN !The current value is small enough to set CycConverg=.true.
!                          write (*,*)
!                          write(*,'(a,i2, a, i4)') 'CONVERGENCE DETECTED via Fig. of Merit n ', ii, ' at cycle n ', cycles-TestConvRefCycle-s%NMem+1
!                             read(*,*)
                          s%CycConverg(ii)=.true.
                       ENDIF
                    ENDIF
                 ENDIF
              ENDIF
           ENDDO
           !### Detects if a new higher peak comes ###
           DO ii=1,NFOM
              IF (s%CycPeak(ii).eqv..true.) THEN
                 IF ( (abs(CycMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CycAver(ii))/CycDev(ii)  &
                      >=s%CycPeakTopRel(ii)) ) THEN
                    s%CycPeakTopRel(ii)=abs(CycMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CycAver(ii))/CycDev(ii)
                    s%CycPeak(ii)=.false.
                    s%CycConverg(ii)=.false.
                    IF (debug) THEN
                      write(*,*)
                      write(*,'(a,i2)') 'NEW HIGHER PEAK for fig. of merit n ', ii
                      read(*,*)
                    ENDIF
                 ENDIF
              ENDIF
           ENDDO

        ENDIF
        !####### Counting the points in the "half-peak"; retaining the cycle n of the end of this peak #######
        DO ii=1, NFOM
           IF (s%CycPeak(ii).eqv..false.) THEN
              s%CycPeakPoints(ii)=s%CycCount(ii)
              s%CycPeakEndCycle(ii)=cycles-TestConvRefCycle
           ELSE
              s%CycPeakPoints(ii)=s%CycPeakPoints(ii)
              s%CycPeakEndCycle(ii)=s%CycPeakEndCycle(ii)
           ENDIF
        ENDDO
!####### Detection of the slope peak #######
        !Checks the WaitCheckCycle-th value; if it's too large (+ or -) then s%SlopeWait=.true. and nothing happens until the Slope value is "small" again
        IF (cycles-TestConvRefCycle-s%NMem+1==WaitCheckCycle) THEN
           DO ii=1,NFOM
              IF (abs(SlopeMem(ii,WaitCheckCycle)-SlopeAver(ii))>s%SlopeMax*SlopeDev(ii)) THEN
                 s%SlopeWait(ii)=.true.
                s%SlopeWaitSign(ii)=nint((SlopeMem(ii,WaitCheckCycle)-SlopeAver(ii))/abs(SlopeMem(ii,WaitCheckCycle)-SlopeAver(ii)))
              ELSE
                 s%SlopeWait(ii)=.false.
              ENDIF
           ENDDO
        ENDIF

        !Detection of the peak
        IF (cycles-TestConvRefCycle-s%NMem+1>=WaitCheckCycle) THEN

           DO ii=1,NFOM
              IF (s%SlopeWait(ii).eqv..true.) THEN
                 if ( s%SlopeWaitSign(ii)*(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-SlopeAver(ii))  &
                      <s%SlopeMax*SlopeDev(ii) ) then
                    s%SlopeWait(ii)=.false.
                 endif
              ENDIF
           ENDDO
           DO ii=1,NFOM
              IF (s%SlopeWait(ii).eqv..false.) THEN
                 IF (cycles-TestConvRefCycle>=NDeriv-1+s%NMem)THEN !When the s%NMem-th value is stored, then the search for a peak begins with the FIRST stored value
                    IF ( ExpPeakSign(ii)*(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-SlopeAver(ii))  &
                         >s%SlopeMax*SlopeDev(ii) ) THEN ! The point belongs to the expected peak
                       s%SlopeCount(ii)=s%SlopeCount(ii)+1
                       ! Remembering the top (larger) value of the peak and the corresponding cycle
                       if ( (abs(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-SlopeAver(ii))/SlopeDev(ii)  &
                            >s%SlopePeakTopRel(ii)).AND.(abs(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))  &
                            >s%SlopePeakTopAbs(ii)) ) then
                          s%SlopePeakTopRel(ii)=abs(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-SlopeAver(ii))  &
                                                /SlopeDev(ii)
                          s%SlopePeakTopAbs(ii)=abs(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))
                          s%SlopePeakTopCycle(ii)=cycles-TestConvRefCycle-s%NMem+1
                       else
                          s%SlopePeakTopRel(ii)=s%SlopePeakTopRel(ii)
                          s%SlopePeakTopAbs(ii)=s%SlopePeakTopAbs(ii)
                          s%SlopePeakTopCycle(ii)=s%SlopePeakTopCycle(ii)
                       endif
                    ELSE ! The point is not part of the peak
                       IF (s%SlopeCount(ii)>=s%MinSlopePeakPoints) THEN
                          s%SlopePeak(ii)=.true. !There were already enough large points to consider them as a peak
                       ELSEIF (s%SlopePeak(ii).eqv..false.) THEN
                          s%SlopePeakTopRel(ii)=0.0
                          s%SlopePeakTopAbs(ii)=0.0
                          s%SlopePeakTopCycle(ii)=0
                       ENDIF
                       s%SlopeCount(ii)=0
                    ENDIF
                    IF (s%SlopePeak(ii)) THEN !A peak has been detected
                       IF ( abs(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))  &
                            <=(s%SlopeMax/2.)*SlopeDev(ii) ) THEN !The current value is small enough to set s%SlopeConverg=.true.
!                          write (*,*)
!                          write(*,'(a,i2, a, i4)') 'CONVERGENCE DETECTED via slope n ', ii, ' at cycle n ', cycles-TestConvRefCycle-s%NMem+1
!                             read(*,*)
                          s%SlopeConverg(ii)=.true.
                       ENDIF
                    ENDIF
                 ENDIF
              ENDIF
           ENDDO
           !### Detects if a new higher peak comes ###
           DO ii=1,NFOM
              IF (s%SlopePeak(ii).eqv..true.) THEN
                 IF ( (abs(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-SlopeAver(ii))/SlopeDev(ii)  &
                      >=s%SlopePeakTopRel(ii)).AND.(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)  &
                      >=s%SlopePeakTopAbs(ii)) ) THEN
                    s%SlopePeakTopRel(ii)=abs(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-SlopeAver(ii))  &
                                          /SlopeDev(ii)
                    s%SlopePeakTopAbs(ii)=abs(SlopeMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))
                    s%SlopePeakTopCycle(ii)=cycles-TestConvRefCycle-s%NMem+1
                    s%SlopePeak(ii)=.false.
                    s%SlopeConverg(ii)=.false.
                    IF (debug) THEN
                      write(*,*)
                      write(*,'(a,i2)') 'NEW HIGHER PEAK for slope n ', ii
                        read(*,*)
                    ENDIF
                 ENDIF
              ENDIF
           ENDDO

        ENDIF
        !####### Counting the points in the slope peak; retaining the cycle n of the end of this peak #######
        DO ii=1, NFOM
           IF (s%SlopePeak(ii).eqv..false.) THEN
              s%SlopePeakPoints(ii)=s%SlopeCount(ii)
           ELSE
              s%SlopePeakPoints(ii)=s%SlopePeakPoints(ii)
           ENDIF
        ENDDO
!####### Detection of the double curvature peaks #######
        !Checks the WaitCheckCycle-th value; if it's too large (+ or -) then s%CurvWait=.true. and nothing happens until the Curvature value is "small" again
        IF (cycles-TestConvRefCycle-s%NMem+1==WaitCheckCycle) THEN
           DO ii=1,NFOM
              IF (abs(CurvMem(ii,WaitCheckCycle)-CurvAver(ii))>s%CurvMax*CurvDev(ii)) THEN
                 s%CurvWait(ii)=.true.
                 s%CurvWaitSign(ii)=nint((CurvMem(ii,WaitCheckCycle)-CurvAver(ii))/abs(CurvMem(ii,WaitCheckCycle)-CurvAver(ii)))
              ELSE
                 s%CurvWait(ii)=.false.
              ENDIF
           ENDDO
        ENDIF

        !Detection of the peaks
        IF (cycles-TestConvRefCycle-s%NMem+1>=10) THEN
           DO ii=1,NFOM
              IF (s%CurvWait(ii).eqv..true.) THEN
                 if ( s%CurvWaitSign(ii)*(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))  &
                      <s%CurvMax*CurvDev(ii) ) then
                    s%CurvWait(ii)=.false.
                 endif
              ENDIF
           ENDDO
           DO ii=1,NFOM
              IF (s%CurvWait(ii).eqv..false.) THEN

                 IF (cycles-TestConvRefCycle>=2*NDeriv-2+s%NMem)THEN !When the s%NMem-th value is stored, then the search for a peak begins with the FIRST stored value

                    IF (s%CurvPeak1(ii).eqv..false.) THEN !no peak is detected yet
                       IF ( ExpPeakSign(ii)*(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))  &
                            >s%CurvMax*CurvDev(ii) ) THEN ! The point belongs to the expected peak
                          s%CurvCount1(ii)=s%CurvCount1(ii)+1
                          ! Remembering the top (larger) value of the peak and the corresponding cycle
                          if ( (abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))/CurvDev(ii)  &
                               >s%CurvPeak1TopRel(ii)).AND.(abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))  &
                               >s%CurvPeak1TopAbs(ii)) ) then
                             s%CurvPeak1TopRel(ii)=abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))  &
                                                   /CurvDev(ii)
                             s%CurvPeak1TopAbs(ii)=abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))
                             s%CurvPeak1TopCycle(ii)=cycles-TestConvRefCycle-s%NMem+1
                          else
                             s%CurvPeak1TopRel(ii)=s%CurvPeak1TopRel(ii)
                             s%CurvPeak1TopAbs(ii)=s%CurvPeak1TopAbs(ii)
                             s%CurvPeak1TopCycle(ii)=s%CurvPeak1TopCycle(ii)
                          endif
                       ELSEIF ( ExpPeakSign(ii)*(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))  &
                                <-s%CurvMax*CurvDev(ii) ) THEN ! The point could belong to the second peak
                          if (s%CurvCount1(ii)>=s%MinCurvPeakPoints) then ! There were enough points for the first peak; the second peak begins
                                s%CurvPeak1(ii)=.true.
                             s%CurvCount1(ii)=0
                             s%CurvPeak2TopRel(ii)=abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii)) &
                               /CurvDev(ii)
                             s%CurvPeak2TopAbs(ii)=abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))
                             s%CurvCount2(ii)=1
                          else ! There were not enough points for the first peak; all is reset to 0
                             s%CurvCount1(ii)=0
                             s%CurvPeak1TopRel(ii)=0.0
                             s%CurvPeak1TopAbs(ii)=0.0
                             s%CurvPeak1TopCycle(ii)=0
                          endif
                       ELSE ! The point is too low to belong to any of the two peaks
                          if (s%CurvCount1(ii)>=s%MinCurvPeakPoints) then ! There were enough points for the 1st peak
                                s%CurvPeak1(ii)=.true.
                          else ! There were not enough points for the first peak; all is reset to 0
                             s%CurvPeak1TopRel(ii)=0.0
                             s%CurvPeak1TopAbs(ii)=0.0
                             s%CurvPeak1TopCycle(ii)=0
                          endif
                          s%CurvCount1(ii)=0
                       ENDIF

                    ELSEIF (s%CurvPeak2(ii).eqv..false.) THEN !the first peak only is detected
                          if ( -1*ExpPeakSign(ii)*(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))   &
                               >s%CurvMax*CurvDev(ii) ) then ! The point belongs to the second peak
                             s%CurvCount1(ii)=0
                             s%CurvCount2(ii)= s%CurvCount2(ii)+1
                             if ( (abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))/CurvDev(ii)  &
                                  >s%CurvPeak2TopRel(ii)).AND.(abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))  &
                                  >s%CurvPeak2TopAbs(ii)) ) then
                                s%CurvPeak2TopRel(ii)=abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1) &
                                                      -CurvAver(ii))/CurvDev(ii)
                                s%CurvPeak2TopAbs(ii)=abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))
                             else
                                s%CurvPeak2TopRel(ii)=s%CurvPeak2TopRel(ii)
                                s%CurvPeak2TopAbs(ii)=s%CurvPeak2TopAbs(ii)
                             endif
                          elseif( ExpPeakSign(ii)*(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))  &
                                  >s%CurvMax*CurvDev(ii) )then ! The point could belong to a new (higher?) first peak
                             IF (s%CurvCount2(ii)>=s%MinCurvPeakPoints) THEN ! There were enough points for the second peak
                                s%CurvPeak2(ii)=.true.
                                s%CurvCount1(ii)=s%CurvCount1(ii)+1
                                s%CurvCount2(ii)=0
                             ELSEIF (cycles-TestConvRefCycle-s%CurvPeak1EndCycle(ii)>s%CurvPeak1Points(ii)) THEN ! The second peak doesn't come soon enough after the first peak; the first peak is cancelled
                                s%CurvPeak1(ii)=.false.
                             s%CurvPeak1TopRel(ii)=abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))  &
                                                      /CurvDev(ii)
                                s%CurvPeak1TopAbs(ii)=abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))
                                s%CurvPeak1TopCycle(ii)=cycles-TestConvRefCycle-s%NMem+1
                                s%CurvPeak2TopRel(ii)=0.0
                                s%CurvPeak2TopAbs(ii)=0.0
                                s%CurvCount1(ii)=s%CurvCount1(ii)+1
                                s%CurvCount2(ii)=0
!                                write (*,*)
!                                write(*,'(a, i2, a)') 'False first peak n ', ii, ' detected'
                             ELSE
                                s%CurvCount1(ii)=s%CurvCount1(ii)+1
                                s%CurvCount2(ii)=0
                             ENDIF
                          else ! The point is too low to belong to any of the two peaks
                             IF (s%CurvCount2(ii)>=s%MinCurvPeakPoints) THEN ! There were enough points for the second peak
                                s%CurvPeak2(ii)=.true.
                                s%CurvCount1(ii)=0
                                s%CurvCount2(ii)=0
                             ELSEIF (cycles-TestConvRefCycle-s%CurvPeak1EndCycle(ii)>s%CurvPeak1Points(ii)) THEN ! The second peak doesn't come soon enough after the first peak; the first peak is cancelled
                                s%CurvPeak1(ii)=.false.
                                s%CurvPeak1TopRel(ii)=0.0
                                s%CurvPeak1TopAbs(ii)=0.0
                                s%CurvPeak1TopCycle(ii)=0
                                s%CurvPeak2TopRel(ii)=0.0
                                s%CurvPeak2TopAbs(ii)=0.0
                                s%CurvCount1(ii)=0
                                s%CurvCount2(ii)=0
!                                write (*,*)
!                                write(*,'(a, i2, a)') 'False first peak n ', ii, ' detected'
                             ENDIF
                          endif
                    ELSEIF  (s%CurvPeak2(ii).eqv..true.) THEN ! The second peak is detected; counting the points of an eventual new higher first peak
                       IF ( ExpPeakSign(ii)*(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))  &
                            >s%CurvMax*CurvDev(ii) ) THEN
                          s%CurvCount1(ii)=s%CurvCount1(ii)+1
                       ELSE
                          s%CurvCount1(ii)=0
                       ENDIF
                    ENDIF

                 ENDIF

              ENDIF
           ENDDO
        ENDIF
        DO ii=1, NFOM
           IF (s%CurvPeak2(ii)) THEN
              IF ( abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))<=(s%CurvMax/2.)*CurvDev(ii) ) THEN ! The current value is small enough to set CurvConverg=.true.
                 s%CurvConverg(ii)=.true.
!                 write (*,*)
!                 write(*,'(a,i2, a, i4)') 'CONVERGENCE DETECTED via curve n ', ii, ' at cycle n ', cycles-TestConvRefCycle-s%NMem+1
              ENDIF
           ENDIF
        ENDDO
        !### Detects if a new higher first peak comes ###
        DO ii=1,NFOM

           IF (s%CurvPeak1(ii).eqv..true.) THEN
              IF ( (ExpPeakSign(ii)*(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii))/CurvDev(ii)  &
                   >=s%CurvPeak1TopRel(ii)).AND.(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)  &
                   >=s%CurvPeak1TopAbs(ii)) ) THEN
                 s%CurvPeak1TopRel(ii)=ExpPeakSign(ii)*(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1)-CurvAver(ii)) &
                                       /CurvDev(ii)
                 s%CurvPeak1TopAbs(ii)=abs(CurvMem(ii,mod(cycles-TestConvRefCycle-s%NMem,NMemGlobal)+1))
                 s%CurvPeak2TopRel(ii)=0.0
                 s%CurvPeak2TopAbs(ii)=0.0
                 s%CurvPeak1TopCycle(ii)=cycles-TestConvRefCycle-s%NMem+1
                 s%CurvPeak1(ii)=.false.
                 s%CurvPeak2(ii)=.false.
                 s%CurvConverg(ii)=.false.
                 IF (debug) THEN
                   write(*,*)
                   write(*,'(a,i2)') 'NEW HIGHER FIRST PEAK for curve n ', ii
                   read(*,*)
                 ENDIF
              ENDIF
           ENDIF
        ENDDO

        !####### Counting the points in the curv. peaks; retaining the cycle n of the end of the 1st peak #######
        DO ii=1, NFOM
           IF (s%CurvPeak1(ii).eqv..false.) THEN
              s%CurvPeak1Points(ii)=s%CurvCount1(ii)
              s%CurvPeak1EndCycle(ii)=cycles-TestConvRefCycle
           ELSE
              s%CurvPeak1Points(ii)=s%CurvPeak1Points(ii)
              s%CurvPeak1EndCycle(ii)=s%CurvPeak1EndCycle(ii)
           ENDIF
           IF (s%CurvPeak2(ii).eqv..false.) THEN
              s%CurvPeak2Points(ii)=s%CurvCount2(ii)
           ELSE
              s%CurvPeak2Points(ii)=s%CurvPeak2Points(ii)
           ENDIF
        ENDDO
!####### Reseting too old peaks #######
        ! Calculating the average cycle of the tops of the peaks (first peak for the curvature)
        PeakTopNbr=0
        PeakTopCycleAver=0.0
        DO ii=1,NFOM
           IF (s%CycPeak(ii)) THEN
              PeakTopNbr=PeakTopNbr+1
              PeakTopCycleAver=PeakTopCycleAver+(s%CycPeakEndCycle(ii)-s%NMem+1)
           ENDIF
           IF (s%SlopePeak(ii)) THEN
              PeakTopNbr=PeakTopNbr+1
              PeakTopCycleAver=PeakTopCycleAver+s%SlopePeakTopCycle(ii)
           ENDIF
           IF (s%CurvPeak2(ii)) THEN
              PeakTopNbr=PeakTopNbr+1
              PeakTopCycleAver=PeakTopCycleAver+s%CurvPeak1TopCycle(ii)
           ENDIF
        ENDDO
        PeakTopCycleAver=PeakTopCycleAver/PeakTopNbr


        ! If the top of a peak was earlier than the PeakTopCycleAver - s%OldPeakCrit cycles, the peak is considered too old and is cancelled
        DO ii=1,NFOM
           IF (s%CycPeak(ii)) THEN
              IF (NINT(PeakTopCycleAver)-(s%CycPeakEndCycle(ii)-s%NMem+1)>s%OldPeakCrit) THEN
                 s%CycPeak(ii)=.false.
                 s%CycPeakTopRel(ii)=0.0
                 s%CycConverg(ii)=.false.
                 s%CycPeakPoints(ii)=0
                 s%CycPeakEndCycle(ii)=cycles-TestConvRefCycle
                 IF (debug) THEN
                   write(*,*)
                   write(*,'(a, i2, a)') 'Too old value "half-peak" n ', ii, ' detected'
                   read(*,*)
                 ENDIF
              ENDIF
           ENDIF
           IF (s%SlopePeak(ii)) THEN
              IF (NINT(PeakTopCycleAver)-s%SlopePeakTopCycle(ii)>s%OldPeakCrit) THEN
                 s%SlopePeak(ii)=.false.
                 s%SlopePeakTopRel(ii)=0.0
                 s%SlopePeakTopAbs(ii)=0.0
                 s%SlopePeakTopCycle(ii)=0
                 s%SlopeConverg(ii)=.false.
                 s%SlopePeakPoints(ii)=0
                 IF (debug) THEN
                   write(*,*)
                   write(*,'(a, i2, a)') 'Too old slope peak n ', ii, ' detected'
                   read(*,*)
                 ENDIF
              ENDIF
           ENDIF
           IF (s%CurvPeak2(ii)) THEN
              IF (NINT(PeakTopCycleAver)-s%CurvPeak1TopCycle(ii)>s%OldPeakCrit) THEN
                 s%CurvPeak1(ii)=.false.
                 s%CurvPeak1Points(ii)=0
                 s%CurvPeak1TopRel(ii)=0.0
                 s%CurvPeak1TopAbs(ii)=0.0
                 s%CurvPeak1TopCycle(ii)=0
                 s%CurvPeak1EndCycle(ii)=cycles-TestConvRefCycle
                 s%CurvPeak2(ii)=.false.
                 s%CurvPeak2Points(ii)=0
                 s%CurvPeak2TopRel(ii)=0.0
                 s%CurvPeak2TopAbs(ii)=0.0
                 s%CurvCount1(ii)=0
                 s%CurvCount2(ii)=0
                 s%CurvConverg(ii)=.false.
                 IF (debug) THEN
                   write (*,*)
                   write(*,'(a, i2, a)') 'Too old first peak n ', ii, ' detected'
                   read(*,*)
                 ENDIF
              ENDIF
           ENDIF
        ENDDO

!####### Counting the total score of the peaks #######

        CycPeakScore=0.0
        SlopePeakScore=0.0
        CurvPeakScore=0.0
        TotPeakScore=0.0

        DO ii=1,NFOM ! Counting the number of figures of merit detected
           IF (s%CycPeak(ii)) THEN
              CycCriterion(ii)=1
           ELSE
              CycCriterion(ii)=0
           ENDIF
           IF (s%SlopePeak(ii)) THEN
              SlopeCriterion(ii)=1
           ELSE
              SlopeCriterion(ii)=0
           ENDIF
           IF (s%CurvPeak2(ii)) THEN
              CurvCriterion(ii)=1
           ELSE
              CurvCriterion(ii)=0
           ENDIF
        ENDDO

        DO ii=1,NFOM
           IF (s%CycPeak(ii)) THEN
              CycPeakScore=CycPeakScore+s%CycPeakPoints(ii)*s%CycPeakTopRel(ii)
           ENDIF
           IF (s%SlopePeak(ii)) THEN
              SlopePeakScore=SlopePeakScore+s%SlopePeakPoints(ii)*s%SlopePeakTopRel(ii)
           ENDIF
           IF (s%CurvPeak2(ii)) THEN
             CurvPeakScore=CurvPeakScore+(s%CurvPeak1Points(ii)+s%CurvPeak2Points(ii))*(s%CurvPeak1TopRel(ii)+s%CurvPeak2TopRel(ii))
           ENDIF
        ENDDO
        TotPeakScore=CycPeakScore*SUM(CycCriterion)+SlopePeakScore*SUM(SlopeCriterion)+CurvPeakScore*SUM(CurvCriterion)

!####### Writing detailed results ####### (Only if debug .eqv..true.)
      IF (debug) THEN
         write(*,*)
         write(*,*) 'Using parameters with NMEM = ', s%NMem
         write(*,*) 

         write (*,'(a,i5,a10,i4)') 'checked cycle: ', cycles-TestConvRefCycle-s%NMem+1, 'line n ',   &
                                   mod(cycles-TestConvRefCycle-s%NMem, NMemGlobal)+1
         write(*,*)
         write(*,'(a,4l4)') 's%CycWait:    ', (s%CycWait(ii),ii=1,NFOM)

!         DO i=1,NMemGlobal
!            write(*,'(i5,f14.5,3f14.5)') i,(CycMem(ii,i),ii=1,NFOM)
!         ENDDO
!         write(*,*)
         write(*,'(4i4)') (s%CycCount(ii),ii=1,NFOM)
         write(*,'(a,4i4)') 'Peak pts:     ',(s%CycPeakPoints(ii),ii=1,NFOM)
!         write(*,'(4i4)') (s%CycPeakEndCycle(ii),ii=1,NFOM)
         write(*,'(a,4l4)') 'Peak:         ',(s%CycPeak(ii),ii=1,NFOM)
         write(*,'(a,4l4)') 'Stab.:        ',(s%CycConverg(ii),ii=1,NFOM)
         write(*,'(a,4i5)') 'Peak end cyc.:',(s%CycPeakEndCycle(ii)-s%NMem+1,ii=1,NFOM)
         write(*,'(a,4f8.2)') 'Top rel value:   ',(s%CycPeakTopRel(ii),ii=1,NFOM)
         write(*,*)
         write(*,'(a,4f14.5)') 'Aver', (CycAver(ii),ii=1,NFOM)
         write(*,'(a,4f14.5)') 'Dev', (CycDev(ii),ii=1,NFOM)
      
         write(*,*) '- - - - - - - - - - - - - - - - - - - - - - - - - - - -'

         write(*,'(a,4l4)') 's%SlopeWait:    ', (s%SlopeWait(ii),ii=1,NFOM)
!         write (*,*)
!         DO i=1,NSlopeTemp
!            write(*,'(i5,f14.5,3f10.5)') i,(SlopeTemp(ii,i),ii=1,NFOM)
!         ENDDO
!         write (*,*)    
!         DO i=1,s%NMem
!            write(*,'(i5,f14.5,3f10.5)') i,(SlopeMem(ii,i),ii=1,NFOM)
!         ENDDO
         write(*,'(4i4)') (s%SlopeCount(ii),ii=1,NFOM)
         write(*,'(a,4i4)') 'Peak pts:     ',(s%SlopePeakPoints(ii),ii=1,NFOM)
         write(*,'(a,4l4)') 'Peak:         ',(s%SlopePeak(ii),ii=1,NFOM)
         write(*,'(a,4l4)') 'Stab.:        ',(s%SlopeConverg(ii),ii=1,NFOM)
         write(*,'(a,4i5)') 'Peak top cyc.:',(s%SlopePeakTopCycle(ii),ii=1,NFOM)
         write(*,'(a,4f8.2)') 'Top rel value:   ',(s%SlopePeakTopRel(ii),ii=1,NFOM)
         write(*,'(a,4f8.2)') 'Top abs value:   ',(s%SlopePeakTopAbs(ii),ii=1,NFOM)
         write(*,*)
         write(*,'(a,4f9.5)') 'Aver ',(SlopeAver(ii),ii=1,NFOM)
         write(*,'(a,4f9.5)') 'Dev  ',(SlopeDev(ii),ii=1,NFOM)

         write(*,*) '- - - - - - - - - - - - - - - - - - - - - - - - - - - -'

!         write(*,*)

         write(*,'(a,4l4)') 's%CurvWait:       ', (s%CurvWait(ii),ii=1,NFOM)
!!         Write (*,*) (CurvSign(ii),ii=1,NFOM)
!         DO i=1,s%NMem
!            write(*,'(i5,f14.5,3f10.5)') i,(CurvMem(ii,i),ii=1,NFOM)
!         ENDDO
!         write(*,*)
         write(*,'(4i4)') (s%CurvCount1(ii),ii=1,NFOM)
         write(*,'(4i4)') (s%CurvCount2(ii),ii=1,NFOM)
         write(*,'(a,4i4)') 'Peak 1 pts:     ',(s%CurvPeak1Points(ii),ii=1,NFOM)
         write(*,'(a,4i4)') 'Peak 2 pts:     ',(s%CurvPeak2Points(ii),ii=1,NFOM)
         write(*,'(a,4l4)') 'Peak 1:         ',(s%CurvPeak1(ii),ii=1,NFOM)
         write(*,'(a,4l4)') 'Peak 2:         ',(s%CurvPeak2(ii),ii=1,NFOM)
         write(*,'(a,4l4)') 'Stab.:          ',(s%CurvConverg(ii),ii=1,NFOM)
         write(*,'(a,4i5)') 'Peak 1 top cyc.:',(s%CurvPeak1TopCycle(ii),ii=1,NFOM)
         write(*,'(a,4f8.2)') 'Top 1 rel value:   ',(s%CurvPeak1TopRel(ii),ii=1,NFOM)
         write(*,'(a,4f8.2)') 'Top 2 rel value:   ',(s%CurvPeak2TopRel(ii),ii=1,NFOM)
         write(*,'(a,4f8.2)') 'Top 1 abs value:   ',(s%CurvPeak1TopAbs(ii),ii=1,NFOM)
         write(*,'(a,4f8.2)') 'Top 2 abs value:   ',(s%CurvPeak2TopAbs(ii),ii=1,NFOM)
!         write(*,*) (s%CurvPeak1EndCycle(ii),ii=1,NFOM)
         write(*,*)
         write(*,'(a,4f9.5)') 'Aver ',(CurvAver(ii),ii=1,NFOM)
         write(*,'(a,4f9.5)') 'Dev  ',(CurvDev(ii),ii=1,NFOM)
         
         write(*,*) '- - - - - - - - - - - - - - - - - - - - - - - - - - - -'
         write (*,'(a,f7.0)') 'Top cycles average:',PeakTopCycleAver
         write(*,'(a,f10.2)') 'Cyc score:        ',CycPeakScore
         write(*,'(a,f10.2)') 'Slope score:      ',SlopePeakScore
         write(*,'(a,f10.2)') 'Curv score:       ',CurvPeakScore
         write(*,'(a,f10.2)') 'Total score:      ',TotPeakScore
      ENDIF  
      
!####### Detecting the end of the convergence #######      
          
       IF (TotPeakScore>=LimitFOM*1000.) THEN
!          write(*,*)
!           write(*,'(a,i2)') 'ALLELUIA!!! Mackie just detected a convergence of level ',Floor(TotPeakScore/1000)
!           read(*,*)
           DO ii=1,NFOM
              if (s%CycPeak(ii).eqv.s%CycConverg(ii)) then
                 CycStop(ii)=.true.
              else
                 CycStop(ii)=.false.
              endif
              if (s%SlopePeak(ii).eqv.s%SlopeConverg(ii)) then
                 SlopeStop(ii)=.true.
              else
                 SlopeStop(ii)=.false.
              endif
              if (s%CurvPeak1(ii).eqv.s%CurvConverg(ii)) then
                 CurvStop(ii)=.true.
              else
                 CurvStop(ii)=.false.
              endif
           ENDDO

           TestConverged=.true. !TestConverged is kept = true only if the following contitions are verified for all FOM
           DO ii=1,NFOM
              if (CycStop(ii).AND.SlopeStop(ii).AND.CurvStop(ii)   &
                  .AND.(s%CycCount(ii)==0).AND.(s%SlopeCount(ii)==0).AND.(s%CurvCount1(ii)==0)) then
                 TestConverged=TestConverged
              else
                 TestConverged=.false.
              endif
           ENDDO

           if (TestConverged) then
              IF (debug) THEN
                 write(*,*)
                 write(*,*) '*********************************************************'
                 write(*,'(a,f7.2)') 'Superflip converged at level ', TotPeakScore/1000
                 write(*,*)
                 write(*,'(a)') 'Following peaks have been detected:'
                 DO ii=1,NFOM
                    if (s%CycPeak(ii)) then
                       write (*,'(a,i1)') '   Cyc n ', ii
                    endif
                 ENDDO
                 DO ii=1,NFOM
                    if (s%SlopePeak(ii)) then
                       write (*,'(a,i1)') '   slope n ', ii
                    endif
                 ENDDO
                 DO ii=1,NFOM
                    if (s%CurvPeak2(ii)) then
                       write (*,'(a,i1)') '   curvature n ', ii
                    endif
                 ENDDO
                 write(*,*)
              ENDIF
!              write(*,'(a)') 'Following criterions are stabilized:'
!              DO ii=1,NFOM
!                 if (s%CycConverg(ii)) then
!                    write (*,'(a,i1)') '   Cyc n ', ii
!                 endif
!              ENDDO
!              DO ii=1,NFOM
!                 if (s%SlopeConverg(ii)) then
!                    write (*,'(a,i1)') '   slope n ', ii
!                 endif
!              ENDDO
!              DO ii=1,NFOM
!                 if (s%CurvConverg(ii)) then
!                    write (*,'(a,i1)') '   curvature n ', ii
!                 endif
!              ENDDO
           endif
       ENDIF
      
!       read(*,*)
      
       IF (debug) THEN
         write(*,*) '*********************************************************'
       ENDIF       
       TotPeakScore=TotPeakScore/1000.
       END SUBROUTINE ConvergenceFromCurves

! Modifies the density to match a predefined histogram
!******************************************************
      SUBROUTINE HistogramMatch(newline)
!******************************************************
      USE SF_Module
      USE FFT_Interfaces
      IMPLICIT NONE
      INTEGER :: i,ii
      LOGICAL :: newline
      REAL :: a(1:hmclasses),b(1:hmclasses)
      if (expandedlog) write(lgu,'(a)') 'Performing histogram matching.'
      if (terminal) then
        if (.not.newline) THEN
          write(*,*)
        ELSE
          write(*,*)
        ENDIF
        write(*,'(a)') 'Performing histogram matching.'
      endif

      CALL GetCurrentHistogram
      ! Fill the constants for each class. See Zhang and Main, Acta Cryst. A46, 41-46
      DO i=1,hmclasses
        a(i)=(hmgramm(i)%orig-hmgramm(i-1)%orig)/(hmgramm(i)%cur-hmgramm(i-1)%cur)
        b(i)=(hmgramm(i-1)%orig*hmgramm(i)%cur-hmgramm(i)%orig*hmgramm(i-1)%cur)/(hmgramm(i)%cur-hmgramm(i-1)%cur)
      ENDDO
      ! Now modify the original density according to the histogram:
      DO i=0,NVoxTot-1
        !Find the class which current voxel is in:
        DO ii=hmclasses-1,0,-1
          IF (rho(i)>hmgramm(ii)%cur) THEN
            rho(i)=a(ii+1)*rho(i)+b(ii+1)
            EXIT
          ENDIF
        ENDDO
      ENDDO
!     Calculate the modified structure factors
      CALL FFT(rho2sf)
      SF(0)=MEMRefl(1)%FMEM !Preserve the total charge, otherwise the iteration goes crazy
      CALL SF2MEMRefl
      END SUBROUTINE HistogramMatch

!******************************************************
      SUBROUTINE RepartitionIntensities(ForceLaue)  !Input is a MEMRefl array with structure factors resulting from some density modification, output are modified amplitudes in MEMRefl(i)%F, and updated phases
!******************************************************
      USE SF_Module
      USE FFT_Interfaces
      IMPLICIT NONE
      INTEGER :: i,ii,iii,iv,arrloc,ICoo(SDim),friedel
      REAL :: dma,dmb, tmpamp
      LOGICAL :: ForceLaue
      
!     Average the modified structure factors over symmetry. Use SF to speed up the process
      IF (ForceLaue) THEN
        CALL MEMRefl2SF
        DO i=MStartF+1,MEndF !Start from the second one, the first is F000
          dma=0.0E0;dmb=0.0E0
          DO ii=1,NSym
            ICoo=0
            DO iii=1,SDim
              DO iv=1,SDim
                ICoo(iii)=ICoo(iii)+MEMRefl(i)%hklFG(iv)*SmOp(ii)%R(iv,iii)
              END DO
            END DO
            CALL ArrayLocationSF(ICoo,ArrLoc,friedel)
            dma=dma+(SF(Arrloc)**2+SF(Arrloc+1)**2)
            dmb=dmb+1.
          END DO
          MEMRefl(i)%FMEM=sqrt(dma/dmb)
        END DO
      ENDIF
      
      IF (NGroups>0) THEN !Calculate the current intensities
        DO i=1,NGroups
          GroupInt(i)%cur=0.0E0
        ENDDO  
        DO i=MStartF+1,MEndF
          ii=MEMRefl(i)%group
          IF (ii>0) THEN
            IF (MEMRefl(i)%hklFG(1)==0) THEN
              GroupInt(ii)%cur=GroupInt(ii)%cur+0.5*(MEMRefl(i)%FMEM**2)
            ELSE
              GroupInt(ii)%cur=GroupInt(ii)%cur+MEMRefl(i)%FMEM**2
            ENDIF
          ENDIF
        ENDDO
      ENDIF
      DO ii=1,NGroups
        IF (GroupInt(ii)%cur<eps) GroupInt(ii)%cur=1. !To avoid division by zero. Value is unimportant, all reflections in the group are zero anyway
      ENDDO
!     Replace the structure factors by the repartitioned ones
      DO i=MStartF+1,MEndF
        dma=MEMRefl(i)%AMEM
        dmb=MEMRefl(i)%BMEM
        IF ((dma*dma+dmb*dmb)>eps) THEN
          tmpamp=sqrt(dma*dma+dmb*dmb)
        ELSE
          tmpamp=1.0
        ENDIF
        IF ((NGroups==0).or.(MEMRefl(i)%group==0)) THEN !Only phase is taken, amplitude remains
          MEMRefl(i)%AMEM=MEMRefl(i)%F*dma/tmpamp
          MEMRefl(i)%BMEM=MEMRefl(i)%F*dmb/tmpamp
          MEMRefl(i)%FMEM=MEMRefl(i)%F
        ELSE !Change the amplitude AND phase
          ii=MEMRefl(i)%group
          MEMRefl(i)%AMEM=MEMRefl(i)%FMEM*(dma/tmpamp)*sqrt(GroupInt(ii)%orig/GroupInt(ii)%cur)
          MEMRefl(i)%BMEM=MEMRefl(i)%FMEM*(dmb/tmpamp)*sqrt(GroupInt(ii)%orig/GroupInt(ii)%cur)
          MEMRefl(i)%FMEM=SQRT(MEMRefl(i)%AMEM**2+MEMRefl(i)%BMEM**2)
          MEMRefl(i)%F=MEMRefl(i)%FMEM
        ENDIF
      ENDDO

!     Fill the SF array by the new structure factors. This is the input to the next iteration cycle
      CALL MEMRefl2SF
!     Calculate the updated electron density
      CALL FFT(sf2rho)
!********* Listing of the reflections by groups
!      OPEN(99,file='groups.lst')
!      DO i=1,NGroups
!        write(99,'(a,i5)') 'Group nr.',i
!        DO ii=MStartF,MEndF
!          if (MEMRefl(ii)%group==i) THEN
!            write(99,*) (MEMRefl(ii)%hklfg(iii),iii=1,SDim),MEMRefl(ii)%F
!          ENDIF
!        ENDDO
!      ENDDO
!      CLOSE(99)
!      write(*,*) 'zapsano'
!      read(*,*)
!********* End of the listing of the reflections by groups
      END SUBROUTINE RepartitionIntensities
      
! Align the current density with the envelope, and set to zero the regions outside the envelope (in fact the transition will be smooth)
!******************************************
      SUBROUTINE Envelope(newline)
!******************************************
      USE SF_Module
      USE FFT_Interfaces
      IMPLICIT NONE
      INTEGER :: i
      REAL :: minrho,averho,correction
      LOGICAL :: newline

      if (expandedlog) write(lgu,'(a)') 'Matching the density to the envelope.'
      if (terminal) then
        if (.not.newline) write(*,'('''')') 
        write(*,'(a)') 'Matching the density to the envelope.'
      endif
! shift the phases of MEMRefl to best match the envelope. NOte: If envelope is sampled i=on a grid different from rho, it is resampled inside MatchReference
      CALL MatchReference(2,0)
! Get the density from the changed reflections
      CALL FFT(sf2rho)
! Modify the density to suppress the density outside the envelope
! First find the minimum and the envelope
      minrho=1.0E30
      DO i=0,NVoxTot-1
        minrho=min(envrho(i),minrho)
      ENDDO
      averho=EnvZeroLevel*cApVx
      DO i=0,NVoxTot-1
        correction=1.+min(0.,max((envrho(i)-averho)/abs(envcutmin*minrho-averho),-1.0))
        rho(i)=rho(i)*correction        
!        rho(i)=correction        
      ENDDO
!      CALL WriteRho('modifiedrho.xplor')
!      STOP
! Now we have the modified rho, so let's fill the reflections
      CALL FFT(rho2sf)
      CALL SF2MEMRefl    
      END SUBROUTINE Envelope


!******************************************
      SUBROUTINE AnalysePeaks
!******************************************
      USE SF_Module
      IMPLICIT NONE
      INTEGER, PARAMETER :: TotPeakCnt=10000
      TYPE PeakType
        INTEGER, DIMENSION(3) :: coo
        REAL :: rho
      END Type PeakType
      TYPE (PeakType), DIMENSION(1:TotPeakCnt) :: Peak
      INTEGER :: i,ii,ArrLoc,PeakNr,MinPeakPos
      INTEGER, DIMENSION(0:10) :: totalpeaks
      INTEGER, DIMENSION(0:40) :: DistHist
      REAL :: dummy, sigrho, dist, multsc
      REAL, DIMENSION(1:SDim) :: FracCoo
      INTEGER, DIMENSION(1:SDim) :: Dxyz,Dxyz2,StartC,EndC,AxOrd
      LOGICAL :: PeakYes
      
      peaknr=0
      totalpeaks=0
      DO i=1,SDim
        AxOrd(i)=i
      ENDDO
      
      sigrho=sqrt(sum(rho**2)/NVoxTot)
      DO ArrLoc=0,NVoxTot-1
        IF (rho(ArrLoc)<2.*sigrho) CYCLE !ignore too low (potential) maxima
!       Cycle over the cube 3x3x3... to find a maximum
        CALL Coordinates(ArrLoc,Dxyz)
        DO ii=1,SDim
          StartC(ii)=Dxyz(ii)-1
          EndC(ii)=Dxyz(ii)+1
          Dxyz2(ii)=StartC(ii)
        END DO
        PeakYes=.true.
        DO
          CALL ArrayLocation(Dxyz2,Cnt1)
          IF (rho(ArrLoc)<rho(Cnt1)) THEN
            PeakYes=.false.
            EXIT
          END IF
          ! move one point further            
          IF (.not.NextCoo(SDim,StartC,EndC,Dxyz2,AxOrd)) EXIT
        END DO
        IF (PeakYes) THEN !we have found a maximum, let's check, if it is large enough it
          i=int(rho(Arrloc)/sigrho)
          if(i>10) i=10
          totalpeaks(i)=totalpeaks(i)+1
          IF (PeakNr<TotPeakCnt) THEN
            PeakNr=PeakNr+1
            Peak(PeakNr)%coo=Dxyz
            Peak(PeakNr)%rho=rho(ArrLoc)
          ENDIF
        ENDIF
      ENDDO
      ! Bubblesort the peaks by z coordinate
        DO i=PeakNr-1,1,-1
          DO ii=1,i
            IF (Peak(ii)%coo(3)>Peak(ii+1)%coo(3)) THEN
              dummy=Peak(ii)%rho
              Dxyz=Peak(ii)%coo
              Peak(ii)=Peak(ii+1)
              Peak(ii+1)%rho=dummy
              Peak(ii+1)%coo=Dxyz
            END IF
          END DO
        END DO
        write(lgu,*) 'Total number of peaks in the map:',PeakNr
        write(lgu,*) 'Histogram of peaks:',totalpeaks(2:10)
        write(lgu,*)
        !Calculate the histogram of the peak distances
!        DistHist=0
!        DO i=1,PeakNr-1
!          DO ii=i+1,PeakNr
!            FracCoo=float(Peak(i)%coo-Peak(ii)%coo)/NVox
!            dist=sqrt(multsc(FracCoo,FracCoo,SDim,.false.))
!            StartC=-1
!            EndC=1
!            Dxyz2=StartC
!            DO
!              dist=min(dist,sqrt(multsc(FracCoo+float(Dxyz2),FracCoo+float(Dxyz2),SDim,.false.)))
!              IF (.not.NextCoo(SDim,StartC,EndC,Dxyz2,AxOrd)) EXIT
!            ENDDO
!            IF (dist<4.0) DistHist(nint(dist*10.))=DistHist(nint(dist*10.))+1
!          ENDDO
!        ENDDO
!        write(*,*) 'Histogram of distances:',DistHist
!        write(lgu,*) 'Histogram of distances:',DistHist
      END SUBROUTINE AnalysePeaks
