!******************************************
      SUBROUTINE INPUT
!******************************************
      USE Input_module
      USE sf_module
      USE sf_composition
      USE sf_convergence
      USE sf_interfaces
      
      INTEGER :: i,NPass, ErrCode
      REAL :: ResLimTmp1=-100., ResLimTmp2=-100.
      CHARACTER*(linelength) :: DummyString
      LOGICAL :: FlagZeroC=.false., FmtFlag, UserDefinedFillMissing=.false.
      INTEGER, ALLOCATABLE, DIMENSION(:) :: PresModeTmp 


      CALL InitInput
      IF (Terminal) write(*,'(a,a,a)') 'Reading the input file ',trim(infile),'.'

!********* Define the groups of keywords here
      GrpName='must'; GrpStatus=.true. ;CALL NewKwdGroup
      GrpName='opt';  GrpStatus=.false.;CALL NewKwdGroup
      GrpName='4D';   GrpStatus=.false.;CALL NewKwdGroup
      GrpName='refl'; GrpStatus=.false.;CALL NewKwdGroup
      GrpName='model';GrpStatus=.false.;CALL NewKwdGroup

!********* Define the keywords here
! keywords concerning the file handling
      KN='perform';         KEndN='';           Grp='opt' ; CALL NewKwd(kid_perform)
      KN='title';           KEndN='';           Grp='opt' ; CALL NewKwd(kid_title)
      KN='filebase';        KEndN='';           Grp='opt' ; CALL NewKwd(kid_base)
      KN='expandedlog';     KEndN='';           Grp='opt' ; CALL NewKwd(kid_explog)
      KN='terminal';        KEndN='';           Grp='opt' ; CALL NewKwd(kid_terminal)
      KN='fullreflections'; KEndN='';           Grp='opt' ; CALL NewKwd(kid_expreffile)
      KN='coverage';        KEndN='';           Grp='opt' ; CALL NewKwd(kid_coverage)
      KN='outputfile';      KEndN='';           Grp='must'; CALL NewKwd(kid_outfile)
      KN='outputformat';    KEndN='';           Grp='opt';  CALL NewKwd(kid_outformat)
      KN='referencefile';   KEndN='';           Grp='opt';  CALL NewKwd(kid_reffile)
      KN='referenceformat'; KEndN='';           Grp='opt';  CALL NewKwd(kid_refformat)
      KN='modelfile';       KEndN='';           Grp='model';CALL NewKwd(kid_modelfile)
      KN='modelformat';     KEndN='';           Grp='opt';  CALL NewKwd(kid_modelformat)
      KN='envelopefile';    KEndN='';           Grp='opt';  CALL NewKwd(kid_envfile)
      KN='envelopeformat';  KEndN='';           Grp='opt';  CALL NewKwd(kid_envformat)
      KN='commandfile';     KEndN='';           Grp='opt';  CALL NewKwd(kid_comfile)
      KN='rewriteoutput';   KEndN='';           Grp='opt';  CALL NewKwd(kid_rewrite)
! Basic crystallographic information
      KN='dimension';       KEndN='';           Grp='opt' ; CALL NewKwd(kid_sdim)
      KN='realdimension';   KEndN='';           Grp='opt' ; CALL NewKwd(kid_rdim)
      KN='cell';            KEndN='';           Grp='must'; CALL NewKwd(kid_cell)
      KN='qvectors';        KEndN='endqvectors';Grp='4D';   CALL NewKwd(kid_qvec)
      KN='symmetry';        KEndN='endsymmetry';Grp='must'; CALL NewKwd(kid_symmetry)
      KN='testsymmetry';    KEndN='endtestsymmetry';Grp='opt'; CALL NewKwd(kid_testsym)
      KN='twin';            KEndN='endtwin';    Grp='opt';  CALL NewKwd(kid_twin)
      KN='centers';         KEndN='endcenters' ;Grp='opt';  CALL NewKwd(kid_centers)
! Reflection list
      KN='dataformat';      KEndN='';           Grp='opt';  CALL NewKwd(kid_dataformat)
      KN='dataitemwidths';  KEndN='';           Grp='opt';  CALL NewKwd(kid_dataiw)
      KN='reflstartline';   KEndN='';           Grp='opt';  CALL NewKwd(kid_startline)
      KN='reflendline';     KEndN='';           Grp='opt';  CALL NewKwd(kid_endline)
      KN='reflectionlist';  KEndN='';           Grp='opt';  CALL NewKwd(kid_refllist)
      KN='fbegin';          KEndN='endf';       Grp='refl'; CALL NewKwd(kid_fbegin)
      KN='reslimit';        KEndN='';           Grp='opt';  CALL NewKwd(kid_reslimit)
      KN='resunits';        KEndN='';           Grp='opt';  CALL NewKwd(kid_resunits)
      KN='missing';         KEndN='';           Grp='opt';  CALL NewKwd(kid_missing)
! Setings of the algorithm
      KN='voxel';           KEndN='';           Grp='opt';  CALL NewKwd(kid_voxel)
      KN='finevoxel';       KEndN='';           Grp='opt';  CALL NewKwd(kid_finevoxel)
      KN='delta';           KEndN='';           Grp='opt';  CALL NewKwd(kid_delta)
      KN='neutrons';        KEndN='';           Grp='opt';  CALL NewKwd(kid_neutrons)
      KN='randomseed';      KEndN='';           Grp='opt';  CALL NewKwd(kid_rs)
      KN='fastfft';         KEndN='';           Grp='opt';  CALL NewKwd(kid_fastfft)
      KN='maxcycles';       KEndN='';           Grp='opt';  CALL NewKwd(kid_maxcycles)
      KN='addcycles';       KEndN='';           Grp='opt';  CALL NewKwd(kid_addcyc)
      KN='skipstartcycles'; KEndN='';           Grp='opt';  CALL NewKwd(kid_skipstartcyc)
      KN='usephases';       KEndN='';           Grp='opt';  CALL NewKwd(kid_usephases)
      KN='biso';            KEndN='';           Grp='opt';  CALL NewKwd(kid_biso)
      KN='normalize';       KEndN='';           Grp='opt';  CALL NewKwd(kid_normalize)
      KN='nresshells';      KEndN='';           Grp='opt';  CALL NewKwd(kid_nresshells)
      KN='weakratio';       KEndN='';           Grp='opt';  CALL NewKwd(kid_weakratio)
      KN='f+deltaf';        KEndN='';           Grp='opt';  CALL NewKwd(kid_fdeltaf)
      KN='searchsymmetry';  KEndN='';           Grp='opt';  CALL NewKwd(kid_searchsym)
      KN='derivesymmetry';  KEndN='';           Grp='opt';  CALL NewKwd(kid_derivesym)
      KN='phisymtype';      KEndN='';           Grp='opt';  CALL NewKwd(kid_coinc)
      KN='histogram';      KEndN='endhistogram';Grp='opt';  CALL NewKwd(kid_histogram)
      KN='hmparameters';    KEndN='';           Grp='opt';  CALL NewKwd(kid_hmpar)
      KN='hmwait';          KEndN='';           Grp='opt';  CALL NewKwd(kid_hmwait)
      KN='modelrepartition';KEndN='';           Grp='opt';  CALL NewKwd(kid_modelrepart)
      KN='smparameters';    KEndN='';           Grp='opt';  CALL NewKwd(kid_smpar)
      KN='envelopeparameters';KEndN='';           Grp='opt';  CALL NewKwd(kid_envpar)
      KN='bestdensities';   KEndN='';           Grp='opt';  CALL NewKwd(kid_bestden)
      KN='repeatmode';      KEndN='';           Grp='opt';  CALL NewKwd(kid_repeatmode)
      KN='convergencemode'; KEndN='';           Grp='opt';  CALL NewKwd(kid_convmode)
      KN='composition';     KEndN='';           Grp='opt';  CALL NewKwd(kid_composition)
      KN='fips';            KEndN='';           Grp='opt';  CALL NewKwd(kid_fips)
      KN='fwhmseparation';  KEndN='';           Grp='opt';  CALL NewKwd(kid_fwhmlim)
      KN='lambda';          KEndN='';           Grp='opt';  CALL NewKwd(kid_lambda)
      KN='polish';          KEndN='';           Grp='opt';  CALL NewKwd(kid_polish)
      KN='fixf0';           KEndN='';           Grp='opt';  CALL NewKwd(kid_fixf0)
      KN='detailedphases';  KEndN='';           Grp='opt';  CALL NewKwd(kid_detph)
      KN='analyzepeaks';    KEndN='';           Grp='opt';  CALL NewKwd(kid_analpeaks)
      KN='clusteranalysis'; KEndN='';           Grp='opt';  CALL NewKwd(kid_clusan)
      KN='viewprogress';    KEndN='';           Grp='opt';  CALL NewKwd(kid_progress)
      KN='presentationmode';KEndN='';           Grp='opt';  CALL NewKwd(kid_presmode)
      KN='ampformatch';     KEndN='';           Grp='opt';  CALL NewKwd(kid_ampformatch)
      KN='mem';             KEndN='';           Grp='opt';  CALL NewKwd(kid_mem)


      DO NPass=1,PassNr
        ErrNr=0
!       read the input line by line
100     ErrCode=0
        CALL NextLine(InBufferOrig,ErrCode) !Get next non-empty line of the input, strip out the comments
        IF (ErrCode>0) GOTO 900 !Some reading error
        IF (ErrCode<0) GOTO 200 !Finished reading the file
        CurKwd=FirstWord(InBufferOrig,.true.) !.true. means function FirstWord returns Buffer shortened - without the first word
        InBuffer=ToLow(InBufferOrig)
    CurKwd=ToLow(CurKwd) !InBufferOrig contins the original value of the keyword, InBuffer its lower-case version
        CurKwdId=0
        DO i=1,KwdNr
      IF (CurKwd==Kwd(i)%name) THEN
        CurKwdId=i                       !The first word of the line matches the ith keyword
            Kwd(i)%occ=Kwd(i)%occ+1          !Add one to the number of occurences of the keyword i
      END IF
    END DO

    IF (CurKwdId==0) GOTO 100 !No valid keyword was found, proceed to next line
        ErrCode=0
!*** Perform keyword-specific operations ***
        IF (.true..eqv..false.) THEN !Dummy line to allow for adding the keywords anywhere. Otherwise IF/EDNIF must be checked
! ##### Information about the files
        ELSEIF (CurKwdId==kid_title) THEN
          Title=trim(InBufferOrig)

        ELSEIF (CurKwdId==kid_perform) THEN
          InPerform=FirstWord(InBuffer,.true.)
      IF ((InPerform/='cf').and.(InPerform/='lde').and.(InPerform/='dm').and.(InPerform/='aar').and.(InPerform/='fourier')&
                           .and.(InPerform/='symmetry').and.(InPerform/='workcode').and.(InPerform/='general')&
                   .and.(InPerform/='clusters')) THEN
        CALL ErrorMsg('The value of the keyword perform must be CF, AAR, LDE, dm, general, fourier or symmetry.')
        ErrCode=100
      ENDIF
          IF ((InPerform=='workcode')) THEN
            IF (len_trim(InBuffer)/=0) THEN
          Buffer=FirstWord(InBUffer,.true.)
              CALL OneReal(Buffer,randfourierfactor,ErrCode)
            ENDIF
            IF (len_trim(InBuffer)/=0) THEN
          Buffer=FirstWord(InBuffer,.true.)
              CALL OneInt(Buffer,LargeGraph,ErrCode)
            ENDIF
            IF (len_trim(InBuffer)/=0) THEN
          Buffer=FirstWord(InBuffer,.true.)
              CALL OneReal(Buffer,connstdevfac,ErrCode)
            ENDIF
        IF (ErrCode/=0) CALL ErrorMsg('Error reading randfourierfactor.')
      ELSE
            randfourierfactor=0.0
      ENDIF
      IF (InPerform=='general') InPerform=trim(InPerform)//' '//trim(InBuffer)
      IF (InPerform=='clusters') ClusAn=.true.

        ELSEIF (CurKwdId==kid_base) THEN
          FileBase=trim(InBufferOrig)
           
        ELSEIF (CurKwdId==kid_outfile) THEN
          IF (NPass==1) CALL ReadRhoFiles

        ELSEIF (CurKwdId==kid_expreffile) THEN
          ExpRefFile=trim(InBufferOrig)

        ELSEIF (CurKwdId==kid_reffile) THEN
          RefRhoFile=trim(InBufferOrig)

        ELSEIF (CurKwdId==kid_modelfile) THEN
          ModelRhoFile=FirstWord(InBufferOrig,.true.)
          IF (ToLow(ModelRhoFile)=='patterson') THEN
        ModelRhoFile='patterson'
            ModelRhoType='patterson'
          ELSEIF (ToLow(ModelRhoFile)=='superposition') THEN
        ModelRhoFile='supminfunction'
            ModelRhoType='supminfunction'
          ELSEIF (ToLow(ModelRhoFile)=='random') THEN
        ModelRhoFile='random'
            ModelRhoType='random'
      ENDIF
      IF (len_trim(InBufferOrig)>0) CALL OneReal(InBufferOrig,randmodelfactor,ErrCode)

        ELSEIF (CurKwdId==kid_envfile) THEN
          EnvRhoFile=trim(InBufferOrig)

        ELSEIF (CurKwdId==kid_comfile) THEN
          ComFile=trim(InBufferOrig)

        ELSEIF (CurKwdId==kid_outformat) THEN
      IF (NPass==2) THEN
            DO i=1,OutrhoNr
          OutRhoType(i)=FirstWord(InBuffer,.true.)
          IF (len_trim(outrhotype(i))==0) THEN
!                CALL ErrorMsg('Error, less formats in keyword outputformat than filenames in keyword outputfile.')
!        ErrCode=100
!        EXIT
                OutRhoType(i)='none'
          ENDIF
              IF (OutRhoType(i)/='none'.and.OutRhoType(i)/='jana'.and.OutRhoType(i)/='xplor' &
         .and.OutRhoType(i)/='m80'.and.OutRhoType(i)/='shelxsym'.and.OutRhoType(i)/='ccp4') THEN
                CALL ErrorMsg('Sorry, '//trim(OutRhoType(i))//' is not a known density format.')
            ErrCode=100
          END IF
        ENDDO
      ENDIF

    ELSEIF ((CurKwdId==kid_refformat).or.(CurKwdId==kid_envformat).or.(CurKwdId==kid_modelformat)) THEN
          DummyString=InBuffer            
      IF (CurKwdId==kid_refformat) RefRhoType=InBuffer
      IF (CurKwdId==kid_modelformat) ModelRhoType=InBuffer
      IF (CurKwdId==kid_envformat) EnvRhoType=InBuffer
      IF (DummyString/='jana'.and.DummyString/='xplor'.and.DummyString/='m80'.and.DummyString/='cif' &
                             .and.DummyString/='ccp4') THEN
            CALL ErrorMsg('Sorry, '//trim(DummyString)//' is not a known format.')
        ErrCode=100
      END IF
           
        ELSEIF (CurKwdId==kid_explog) THEN
          CALL YesNo(InBuffer,ExpandedLog,ErrCode)
 
        ELSEIF (CurKwdId==kid_terminal) THEN
      Buffer=FirstWord(InBuffer,.true.)
          CALL YesNo(Buffer,Terminal,ErrCode)
      Buffer=FirstWord(InBuffer,.true.)
      IF (ErrCode==0) THEN
        IF (Buffer=='keep') then
              KeepWindow=.true.
        ELSEIF (len_trim(Buffer)/=0) THEN
          CALL ErrorMsg('Error, the second value of terminal must be "keep" (or blank)!')
          ErrCode=100
        ENDIF
      ENDIF
 
        ELSEIF (CurKwdId==kid_coverage) THEN
          CALL YesNo(InBuffer,CalcCoverage,ErrCode)
 
        ELSEIF (CurKwdId==kid_rewrite) THEN
          CALL YesNo(InBuffer,RewriteOutput,ErrCode)
 
! ##### Crystallographic information.

!       Dimension of the problem
        ELSEIF (CurKwdId==kid_sdim) THEN
          CALL OneInt(InBuffer,SDim,ErrCode)
          IF (SDim<1) THEN
        CALL ErrorMsg('Error, dimension must be a positive number!')
            ErrCode=100
      END IF

!       Dimension of the physical space
        ELSEIF (CurKwdId==kid_RDim) THEN
          CALL OneInt(InBuffer,RDim,ErrCode)
          IF (RDim<1) THEN
        CALL ErrorMsg('Error, realdimension must be a positive number!')
            ErrCode=100
      END IF

        ELSEIF (CurKwdId==kid_cell) THEN
          IF (NPass==2) THEN
        CALL ArrayReal(InBuffer,Cell,NCellPar,ErrCode)
        DO cnt1=1,RDim
          if (Cell(cnt1)<eps) THEN
            Write(Dummystring,'(a,i2,a)') 'Error, the cell parameter number',cnt1,' is not positive.'
            Call Errormsg(DummyString)
        ErrCode=100
          ENDIF
        END DO
        DO cnt1=RDim+1,NCellPar
          if (Cell(cnt1)>180..or.Cell(cnt1)<0.) THEN
            Write(Dummystring,'(a,i2,a)') 'Error, the angle number',cnt1-RDim,' is not in interval <0,180>.'
            Call Errormsg(DummyString)
        ErrCode=100
          ENDIF
        END DO
      ENDIF

        ELSEIF (CurKwdId==kid_qvec) THEN
          IF (NPass==2) CALL Qvectors(kid_qvec,ErrCode)

        ELSEIF (CurKwdId==kid_symmetry) THEN
          CALL ReadSymmetry(kid_symmetry,1,ErrCode)

        ELSEIF (CurKwdId==kid_testsym) THEN
          CALL ReadSymmetry(kid_testsym,2,ErrCode)

        ELSEIF (CurKwdId==kid_twin) THEN
          CALL ReadSymmetry(kid_twin,3,ErrCode)

        ELSEIF (CurKwdId==kid_centers) THEN
          CALL ReadCenters(kid_centers,ErrCode)

! ######### Reflections
        ELSEIF (CurKwdId==kid_dataiw) THEN
          DataItemWidths=trim(InBuffer)

        ELSEIF (CurKwdId==kid_startline) THEN
          ReflStartLine=trim(InBufferOrig)
          IF (ReflStartLine(len_trim(ReflStartLine):len_trim(ReflStartLine))=='"') &
          ReflStartLine(len_trim(ReflStartLine):len_trim(ReflStartLine))=' '
          IF (ReflStartLine(1:1)=='"') ReflStartLine=ReflStartLine(2:)
        ELSEIF (CurKwdId==kid_endline) THEN
          ReflEndLine=trim(InBufferOrig)
          IF (ReflEndLine(len_trim(ReflEndLine):len_trim(ReflEndLine))=='"') &
          ReflEndLine(len_trim(ReflEndLine):len_trim(ReflEndLine))=' '
          IF (ReflEndLine(1:1)=='"') ReflEndLine=ReflEndLine(2:)

        ELSEIF ((CurKwdId==kid_dataformat)) THEN
        CALL SetDataFormatPresets !Replaces dataformat shortcuts with the appropriate combination of variables
        IF (.not.hklxplor) THEN
          Buffer=InBuffer
          NDataItems=0
          DO
            NDataItems=NDataItems+1
            IF (NDataItems>MaxDataItems) THEN
              CALL ErrorMsg('Too many items in the keyword '//trim(CurKwd)//'.')
              ErrCode=100
              EXIT
            ENDIF
            DataItem(NDataItems)%string=FirstWord(Buffer,.true.)
            IF (len_trim(DataItem(NDataItems)%string)==0) THEN
              NDataItems=NDataItems-1
              EXIT
            ENDIF
          ENDDO
          IF (NDataItems==0) THEN
            CALL ErrorMsg('No value present after the keyword '//trim(CurKwd)//'!')
            ErrCode=100
          ENDIF
          IF (ErrCode==0) THEN !Check if every item appears only once and no forbidden item appears
            DO cnt1=1,NFormatFlags
              FormatFlag(cnt1)%count=0
            ENDDO
            DO cnt1=1,NDataItems
              FmtFlag=.false.
              DO cnt2=1,NFormatFlags
                IF (FormatFlag(cnt2)%string==DataItem(cnt1)%string) THEN
                  DataItem(cnt1)%flag=cnt2
                  FormatFlag(cnt2)%count=FormatFlag(cnt2)%count+1
                  FormatFlag(cnt2)%pos=cnt1
                  FmtFlag=.true.
                ENDIF
              ENDDO
              IF (.not.FmtFlag) THEN
                CALL ErrorMsg('String '//trim(DataItem(cnt1)%string)// &
                  ' is not a valid item in the keyword'//trim(CurKwd)//'.')
                ErrCode=100
              ENDIF
            ENDDO
            DO cnt1=1,NFormatFlags
              IF ((FormatFlag(cnt1)%count>1).and.(cnt1/=c_dummy)) THEN
                CALL ErrorMsg('Dataformat: Item '//trim(FormatFlag(cnt1)%string)//' appears more than once.')
                ErrCode=100
              ENDIF
            ENDDO
            IF (ErrCode==0) THEN !And now check the consistency.
              IF ((FormatFlag(c_a)%count>0).and.(FormatFlag(c_b)%count<0)) THEN !A nd B must be present simultaneously
                CALL ErrorMsg('Dataformat: item '//trim(FormatFlag(c_a)%string)//' is present but ' &
                                           //trim(FormatFlag(c_b)%string)//' is missing.')
                ErrCode=100
              ELSEIF ((FormatFlag(c_a)%count==0).and.(FormatFlag(c_int)%count==0).and.&
                (FormatFlag(c_amp)%count==0).and.(FormatFlag(c_ampdiff)%count==0)) THEN !No information about the intensity!
                CALL ErrorMsg('Dataformat: at least one of the items '//trim(FormatFlag(c_a)%string)// &
                            ', '//trim(FormatFlag(c_int)%string)//', ' &
                            //trim(FormatFlag(c_amp)%string)//' must be present in the list.')
                ErrCode=100
              ELSEIF ((FormatFlag(c_a)%count==0).and.(FormatFlag(c_b)%count>0)) THEN !A nd B must be present simultaneously
                CALL ErrorMsg('Dataformat: item '//trim(FormatFlag(c_b)%string)// &
                            ' is present but '//trim(FormatFlag(c_a)%string)//' is missing.')
                ErrCode=100
              ELSEIF ((FormatFlag(c_a)%count>0).and.(FormatFlag(c_int)%count>0)) THEN !IF AB is given, Intensity must not be and vice versa
                CALL ErrorMsg('Dataformat: items '//trim(FormatFlag(c_a)%string)// &
                            ' and '//trim(FormatFlag(c_int)%string)//' are not compatible.')
                ErrCode=100
              ELSEIF ((FormatFlag(c_a)%count>0).and.(FormatFlag(c_amp)%count>0)) THEN !IF AB is given, amplitude must not be and vice versa
                CALL ErrorMsg('Dataformat: items '//trim(FormatFlag(c_a)%string)// &
                            ' and '//trim(FormatFlag(c_amp)%string)//' are not compatible.')
                ErrCode=100
              ELSEIF ((FormatFlag(c_a)%count>0).and.(FormatFlag(c_phase)%count>0)) THEN !IF AB is given, phase must not be and vice versa
                  CALL ErrorMsg('Dataformat: items '//trim(FormatFlag(c_a)%string)// &
                            ' and '//trim(FormatFlag(c_phase)%string)//' are not compatible.')
                  ErrCode=100
              ELSEIF ((FormatFlag(c_int)%count>0).and.(FormatFlag(c_amp)%count>0)) THEN !IF intensity is given, amplitude must not be and vice versa
                  CALL ErrorMsg('Dataformat: items '//trim(FormatFlag(c_int)%string)// &
                            ' and '//trim(FormatFlag(c_amp)%string)//' are not compatible.')
                  ErrCode=100
              ELSEIF ((FormatFlag(c_ampdiff)%count>0).and.((FormatFlag(c_amp)%count>0).or.&
                (FormatFlag(c_int)%count>0).or.(FormatFlag(c_a)%count>0))) THEN !IF intensity is given, amplitude must not be and vice versa
                  CALL ErrorMsg('Dataformat: item '//trim(FormatFlag(c_ampdiff)%string)// &
                            ' is not compatible with '//trim(FormatFlag(c_amp)%string)// &
                ', '//trim(FormatFlag(c_int)%string)//'and '//trim(FormatFlag(c_a)%string))
                  ErrCode=100
              ENDIF
            ENDIF
            IF (FormatFlag(c_fwhm)%count==1) HaveFWHM=.true.
          ENDIF!(ErrCode==0)
        ENDIF!(hklxplor)
    ELSEIF (CurKwdId==kid_resunits) THEN
      IF (InBuffer=='d') THEN
        ResUnits=res_d
      ELSEIF (InBuffer=='sthl') THEN
        ResUnits=res_sthl
      ELSE
        CALL ErrorMSg('Error: the value of keyword resunits can be only d or sthl.')
        ErrCode=100
          ENDIF

    ELSEIF (CurKwdId==kid_reslimit) THEN
      IF (NPass==1) THEN
            Buffer=FirstWord(InBuffer, .true.)          
        IF ((Buffer=='d').or.(Buffer=='sthl')) THEN
          CALL ErrorMSg('Error: obsolete syntax for reslimit. Use keyword resunits to define d/sthl.')
          ErrCode=100
        ELSE
          call OneReal(Buffer,reslimtmp1,ErrCode)
              IF ((ErrCode==0).and.(len_trim(InBuffer)>0)) THEN
            call OneReal(InBuffer,reslimtmp2,ErrCode)
          ENDIF
        ENDIF
      ENDIF

    ELSEIF (CurKwdId==kid_ampformatch) THEN
      IF (NPass==1) THEN
            call OneReal(InBuffer,intformatchlow,ErrCode)
            IF ((ErrCode==0).and.(len_trim(InBuffer)>0)) THEN
          call OneReal(InBuffer,intformatchhigh,ErrCode)
        ENDIF
      ENDIF

        ELSEIF (CurKwdId==kid_missing) THEN
      IF (NPass==1) THEN
            Buffer=FirstWord(InBuffer, .true.)
            SELECT CASE (Buffer)
            CASE ('zero')
              MissingMode=mm_zero
            CASE ('float')
              MissingMode=mm_float
            CASE ('bound')
              MissingMode=mm_bound
            CASE ('boundsum')
              MissingMode=mm_boundsum
            CASE DEFAULT
              CALL ErrorMsg('Error: The first value of the keyword missing must be zero, float, bound, boundsum or extrapolate!')
              ErrCode=100
            END SELECT
            IF ((ErrCode==0).and.(len_trim(InBuffer)>0)) THEN
          call OneReal(InBuffer,FillMissingSthl,ErrCode)
              UserDefinedFillMissing=.true.
        ENDIF
        IF ((ErrCode==0).and.(len_trim(InBuffer)>0)) call OneReal(InBuffer,MissingBndFactor,ErrCode)
          ENDIF

        ELSEIF (CurKwdId==kid_refllist) THEN
          ReflectionList=FirstWord(InBuffer, .true.)
          IF ((ReflectionList/='general').and.(reflectionlist/='unique')) THEN
              CALL ErrorMsg('Error: The aprameter of the keyword reflectionlist must be unique or general!')
              ErrCode=100
          ENDIF

        ELSEIF (CurKwdId==kid_fbegin) THEN
          !Check, if there is a file name attached to the keyword
      IF (len_trim(InBuffer)>0) THEN
        Buffer=FirstWord(InBuffer,.true.)
        IF (Buffer=='randomstructure') THEN
          GenerateRandom=.true.
          IF(len_trim(InBuffer)/=0) CALL OneInt(InBuffer,RandStructureSeed,ErrCode)
        ELSE
          externrefl=.true.
          ExternReflFile=InBufferOrig
          OPEN (UNIT=erfu,FILE=trim(ExternReflFile),IoStat=ErrCode, STATUS='OLD')
          IF (ErrCode/=0) THEN
            CALL ErrorMsg('Error opening file '//trim(ExternReflFile)//&
                   '. Make sure the file exists and check the access rights!')
          END IF
          IF (hklxplor) CLOSE(erfu) !The file is read as an ordinary xplor file - it is opened in the routine
        ENDIF
      ELSE
        externrefl=.false.
      ENDIF
      IF ((ErrCode==0).and.(NPass==2)) CALL AllFromCell(ErrCode)
      IF ((ErrCode==0).and.(.not.GenerateRandom)) CALL ReadRefl(kid_fbegin,ErrCode)
      IF (GenerateRandom) TotalF=0
          IF (externrefl) CLOSE(erfu)

! ######### Settings of the algorithm
        ELSEIF (CurKwdId==kid_voxel) THEN
          IF (NPass==2) THEN
        Buffer=FirstWord(InBuffer,.false.)
        IF (Buffer=='auto') then
          autovoxels=.true.
        elseif (index(inbuffer,'angstrom')/=0) THEN !Voxel dividison is given by maximum voxel size
          autovoxels=.true.
          call OneReal(Buffer,IterGridResolution,ErrCode)
        elseif (index(inbuffer,'shannon')/=0) THEN !Voxel dividison is given by maximum voxel size
          autovoxels=.true.
          call OneReal(Buffer,IterShannonFactor,ErrCode)
        else
          autovoxels=.false.
              CALL ArrayInt(InBuffer,NVoxIter,SDim,ErrCode)
              IF (ErrCode==0) THEN
            IF (any(NVoxIter<=0)) THEN
              Call ErrorMsg('Error, the numbers of voxels must be a positive integer!')
                  ErrCode=100
                END IF
          END IF
        ENDIF
      END IF

        ELSEIF (CurKwdId==kid_finevoxel) THEN
          IF (NPass==2) THEN
        Buffer=FirstWord(InBuffer,.false.)
        IF (Buffer=='auto') then
          autofinevoxels=.true.
        elseif (index(inbuffer,'angstrom')/=0) THEN !Voxel dividison is given by maximum voxel size
          autofinevoxels=.true.
          call OneReal(Buffer,FineGridResolution,ErrCode)
        elseif (index(inbuffer,'shannon')/=0) THEN !Voxel dividison is given by the Shannon rate
          autofinevoxels=.true.
          FineGridResolution=0.0
          call OneReal(Buffer,FineShannonFactor,ErrCode)
        elseif (Buffer=='no') THEN
          autofinevoxels=.false.
          NVoxFine=0
        else
          autofinevoxels=.false.
              CALL ArrayInt(InBuffer,NVoxFine,SDim,ErrCode)
              IF (ErrCode==0) THEN
            IF (any(NVoxIter<=0)) THEN
              Call ErrorMsg('Error, the numbers of finevoxels must be a positive integer!')
                  ErrCode=100
                END IF
          END IF
        ENDIF
      END IF

        ELSEIF (CurKwdId==kid_delta) THEN
          Buffer=FirstWord(InBuffer,.true.)
          IF (Buffer=='auto') THEN
        InAutoDelta=.true.
!            reldelta=.false.
        CFInDelta=0.0
      ELSE
        InAutoDelta=.false.
            CALL OneReal(Buffer,CFInDelta,ErrCode)
          ENDIF
          IF ((InBuffer=='fraction').or.(InBuffer=='dynamic')) THEN
        DeltaType=RelDelta
      ELSEIF ((InBuffer=='absolute').or.(InBuffer=='static')) THEN
        DeltaType=StatDelta
      ELSEIF (InBuffer=='sigma') THEN
        DeltaType=KedDelta
      ELSEIF (InBuffer/=' ') THEN
            CALL ErrorMsg('Error: The second value of the keyword delta must be static (absolute) or dynamic (fraction)!')
        ErrCode=100
      ENDIF

        ELSEIF (CurKwdId==kid_neutrons) THEN
          Buffer=Firstword(InBuffer,.true.)
      IF ((Buffer/='no').and.(Buffer/='yes'))  THEN
        ! Try to read the number of cycles
        CALL OneInt(Buffer,StartNeutrons,ErrCode)
        IF (ErrCode/=0) THEN
          CALL ErrorMsg('Error: Wrong value of keyword neutrons. The allowed values are "no", "yes"'// &
                        ' or an integer number of cycles!')
            ELSE
          neutrons=.false.
        ENDIF
      ELSEIF (Buffer=='yes') THEN
        neutrons=.true.
        StartNeutrons=0
      ELSE
        neutrons=.false.
        StartNeutrons=0
      ENDIF


        ELSEIF (CurKwdId==kid_rs) THEN
          IF (InBuffer=='auto') THEN
        autoRS=.true.
      ELSE
        autoRS=.false.
            CALL OneInt(InBuffer,RandomSeed,ErrCode)
      END IF

        ELSEIF (CurKwdId==kid_fastfft) THEN
          CALL YesNo(InBuffer,FastFFT,ErrCode)
 
        ELSEIF (CurKwdId==kid_maxcycles) THEN
          IF (InMaxCycles<0) THEN 
        CALL OneInt(InBuffer,InMaxCycles,ErrCode)
            IF ((ErrCode==0).and.(InMaxCycles<0)) &
          CALL ErrorMsg('Error: The maximum number of cycles must be a non-negative integer.')
          ENDIF

        ELSEIF (CurKwdId==kid_addcyc) THEN
          CALL OneInt(InBuffer,stabicyc,ErrCode)
            IF ((ErrCode==0).and.(stabicyc<0)) THEN
          CALL ErrorMsg('Error: The number of additional cycles must be a non-negative integer.')
              ErrCode=100
        ENDIF
        ELSEIF (CurKwdId==kid_skipstartcyc) THEN
          CALL OneInt(InBuffer,skipstartcyc,ErrCode)
            IF ((ErrCode==0).and.(skipstartcyc<0)) THEN
          CALL ErrorMsg('Error: The number of skipped cycles must be a non-negative integer.')
              ErrCode=100
        ENDIF
        ELSEIF (CurKwdId==kid_usephases) THEN
          IF (NPass==1) THEN !Read the masic settings
            UsePhases=Firstword(InBuffer,.true.)
        IF ((UsePhases/='no').and.(UsePhases/='always').and.(UsePhases/='code').and. &
            (UsePhases/='firstcycle').and.(UsePhases/='yes'))  THEN
          ! Try to read the number of cycles
          CALL OneInt(UsePhases,setphasecycles,ErrCode)
          IF (ErrCode/=0) THEN
            CALL ErrorMsg('Error: Wrong value of keyword usephases. The allowed values are "no", "always", "firstcycle", '// &
        '"code" or an integer number of cycles!')
              ELSE
            UsePhases='number'
          ENDIF
              IF (UsePhases=='yes') UsePhases='firstcycle' !Obsolete synonym
        ENDIF
            IF (UsePhases=='code') THEN !Read the name of the code file
          buffer=InBufferOrig
          CodeWordFile=FirstWord(Buffer,.true.) !Skips the word "code"
          CodeWordFile=FirstWord(Buffer,.true.)
          IF (len_trim(CodeWordFile)==0) THEN
            CALL ErrorMsg('Error: Missing filename of the codeword file!')
            ErrCode=100
              ELSE
                OPEN (UNIT=cwfu,FILE=trim(CodeWordFile),IoStat=ErrCode, STATUS='OLD')
                IF (ErrCode/=0) THEN
                  CALL ErrorMsg('Error opening file '//trim(CodeWordFile)//&
                            '. Make sure the file exists and check the access rights!')
                END IF
          ENDIF
            ENDIF
          ELSEIF(NPass==2) THEN !Check the corretness of the codeword file and count the codewords
        NCodeWords=0
        DO
          read(cwfu,'(a)',END=1000) buffer
          IF (len_trim(Buffer)>0) THEN
                NCodeWords=NCodeWords+1
                CALL CheckAllowedChars(trim(buffer),6,ErrCode)
                IF (ErrCode/=0) THEN
          write(buffer,'(i3)') NCodeWords
                  CALL ErrorMsg('Error: the codeword number '//trim(buffer)//' contains characters other than 0 and 1.')
          EXIT
        ENDIF
        IF (len_trim(Buffer)>TotalF) THEN
          write(buffer,'(a,i3,a,i4,a)') 'Error: the codeword number ',NCodeWords, &
                ' contains more digits than the number reflections (',TotalF,')'
                  CALL ErrorMsg(buffer)
                  ErrCode=100
          EXIT
        ENDIF
              ENDIF
        ENDDO
1000        CONTINUE
            REWIND(cwfu)
      ENDIF
    ELSEIF (CurKwdId==kid_normalize) THEN
      IF (NPass==2) THEN
        IF ((InBuffer=='wilson').or.(InBuffer=='yes')) THEN
          NormalizeRefl=NormWilson
          IF (ElNr==0) THEN
            CALL ErrorMsg('Error: Unknown chemical composition, normalization by Wilson plot cannot be performed.'// &
                        'Use the keyword "composition" to define the type and number of atoms in the unit cell.')
                ErrCode=100
          ENDIF
        ELSEIF(InBuffer=='local') THEN
          NormalizeRefl=NormLocal
        ELSEIF(InBuffer=='atoms') THEN
          NormalizeRefl=NormAtoms
        ELSEIF(InBuffer=='no') THEN
          NormalizeRefl=NormNone
        ELSE
              CALL ErrorMsg('Sorry, '//trim(InBuffer)//' is not an allowed value for the keyword ' &
                        //trim(CurKwd)//'.')
              CALL ErrorMsg('Allowed values are "wilson", "atoms" and "local".')
          ErrCode=100
        ENDIF
      ENDIF

    ELSEIF (CurKwdId==kid_nresshells) THEN
          CALL OneInt(InBuffer,NResShells,ErrCode)

        ELSEIF (CurKwdId==kid_biso) THEN
          CALL OneReal(InBuffer,Biso,ErrCode)
      IF (len_trim(InBuffer)>0) THEN
        Buffer=firstword(InBuffer,.true.)
        if (Buffer=='fix') Bisofix=.true.
      ENDIF

        ELSEIF (CurKwdId==kid_fips) THEN
      Buffer=FirstWord(InBuffer,.true.)
          IF ((Buffer=='auto').or.(Buffer=='no')) THEN
        FIPS=Buffer
      ELSEIF(Buffer=='table') THEN !Request for a table of FIPS cycles
        FIPS='table'
            CALL ArrayReal(InBuffer,FIPSparams,4,ErrCode)
        IF ((ErrCode/=0).or.(FIPSParams(4)-float(nint(FIPSParams(4)))>eps)) THEN
          CALL ErrorMSg('Error. FIPS table must be followed by three real numbers')
          CALL ErrorMsg('(alpha start, alpha end, alpha step) and one integer (number of FIPS cycles).')
        ENDIF
      ELSE
        !Try reading the FIPS parameters
        CALL OneReal(Buffer,FIPSParams(1),ErrCode)
        IF (ErrCode==0) THEN
          CALL OneInt(InBuffer,cnt1,ErrCode)
          IF (ErrCode==0) FIPSParams(2)=float(cnt1)
        ENDIF
        IF (ErrCode==0) THEN
          FIPS='given'
        ELSE
          CALL ErrorMsg('Error, keyword FIPS must be followed by ''no'', ''auto'', ''table'',')
          CALL ErrorMsg('or one real and one integer representing the boostfactor and number of cycles.')
        ENDIF
      ENDIF

        ELSEIF (CurKwdId==kid_fwhmlim) THEN
          CALL OneReal(InBuffer,FWHMlim,ErrCode)

        ELSEIF (CurKwdId==kid_lambda) THEN
          CALL OneReal(InBuffer,lambda,ErrCode)

        ELSEIF (CurKwdId==kid_composition) THEN
          IF ((NPass==1).and.(ElNr==0)) THEN !The second condition is there to protect against doubled allocation of compos in case of a double occurence of composition in the input file
            IF (len_trim(InBufferORig)==0) THEN
              ElNr=0
        ELSE
          ElNr=1
        ENDIF
            DO i=1,len_trim(InBuffer)-1
              if (InBuffer(i:i).eq.' '.and.InBuffer(i+1:i+1).ne.' ') ElNr=ElNr+1
            ENDDO
            IF (ElNr>0) THEN
              ALLOCATE(Compos(1:ElNr,1:2))
              DO i=1,ElNr
                Compos(i,1)=0
                Compos(i,2)=1
                Buffer=FirstWord(InBufferOrig,.true.)
                !Find the first number in the string
            ii=1;iii=0
            DO WHILE ((ii/=0).and.(iii<=len_trim(Buffer)))
              iii=iii+1
                  CALL CheckAllowedChars(Buffer(iii:iii),5,ii)
            ENDDO
        IF ((iii==2).and.(Buffer(1:1)=='D')) THEN
          Compos(i,1)=1
                ELSE
          DO j=1,TotElNr
                    IF (Buffer(1:iii-1)==ElSmb(j)) Compos(i,1)=j
              END DO
                ENDIF
        IF (Compos(i,1)==0) THEN
              Buffer='Error, symbol '//Buffer(1:iii-1)//' is not a valid symbol of an element.'
              CALL ErrorMsg(Buffer)
              ErrCode=100
                  EXIT
                END IF
                IF (iii<=len_trim(Buffer)) THEN
                  DummyString=Buffer
                  CALL OneInt(Buffer(iii:),Compos(i,2),ErrCode)
          IF (ErrCode/=0) THEN
                 Buffer='Error, string '//trim(DummyString)//' does not have the format "element symbol and number".'
                    CALL ErrorMsg(Buffer)
                  ENDIF
            ENDIF
          END DO
            ELSE
          CALL ErrorMsg('Error, no elements found after the keyword composition.')
          ErrCode=100
        ENDIF
      ENDIF

        ELSEIF (CurKwdId==kid_weakratio) THEN
          CALL OneReal(InBuffer,WeakRatio,ErrCode)
          IF (weakratio<0.0E0) THEN
            CALL ErrorMsg('Error - weakratio must be a positive number!')
            ErrCode=100
          ENDIF
          IF (len_trim(InBuffer)/=0) THEN !Starting cycle for weakrefl
            CALL OneInt(InBuffer,SetWeakCycle,ErrCode)
            IF (setweakcycle<1) THEN
              CALL ErrorMsg('Error - the second argument of weakratio must be a positive integer!')
              ErrCode=100
            ENDIF
          ENDIF
          IF (len_trim(InBuffer)/=0) THEN !Mode of handling the weakreflections
            IF (InBuffer=='relative') THEN
              weakreflmode=wrcode_relative
            ELSEIF (InBuffer=='absolute') THEN
              weakreflmode=wrcode_absolute
            ELSE
              CALL ErrorMsg('Error, the third parameter of '//trim(CurKwd)//' must be relative or absolute!')
              ErrCode=100
            ENDIF
          ENDIF

        ELSEIF (CurKwdId==kid_fdeltaf) THEN
          fdeltaf=.true.
      IF (len_trim(Inbuffer)/=0) THEN
        CALL OneReal(InBuffer,fdeltafw,ErrCode)
            IF (fdeltafw<0.0E0) THEN
          CALL ErrorMsg('Error - the argument of f+deltaf must be a positive number!')
          ErrCode=100
            ENDIF
          ENDIF

        ELSEIF (CurKwdId==kid_derivesym) THEN
          derivesym=Firstword(InBuffer,.true.)
          IF ((derivesym/='no').and.(derivesym/='yes').and.(derivesym/='use')) THEN
            CALL ErrorMsg('Error: Wrong value of keyword derivesymmetry. The allowed values are "no", "yes", or "use"!')
        ErrCode=100
      ENDIF
      IF (ErrCode==0) THEN
            Buffer=FirstWord(InBuffer,.true.) !Get whatever follows the word 'yes', 'use', or 'no'
        IF ((derivesym/='no').and.(len_trim(Buffer)>0)) THEN
              CALL OneReal(Buffer,limitagrfac,ErrCode)
          IF (ErrCode/=0) CALL ErrorMsg('Error, ''derivesymmetry yes|use'' can be followed only by a real number.')
        ENDIF
          ENDIF

        ELSEIF (CurKwdId==kid_searchsym) THEN
          Buffer=Firstword(InBuffer,.true.)
      IF (Buffer=='no') THEN
            searchsym=.false.
        averagesym=NoAveraging
          ELSEIF (Buffer=='shift') THEN
            searchsym=.true.
        averagesym=NoAveraging
          ELSEIF (Buffer=='average') THEN
            searchsym=.true.
        averagesym=AveragePhases
          ELSEIF (Buffer=='fixaverage') THEN
            searchsym=.false.
        averagesym=AveragePhases
          ELSEIF (Buffer/=' ') THEN
            CALL ErrorMsg('Error: Wrong value of keyword searchsymmetry. The allowed values are "no", "shift", or "average"!')
        ErrCode=100
      ENDIF
      Buffer=FirstWord(Inbuffer,.true.)
      IF (len_trim(Buffer)/=0) THEN
        symroutine=trim(Buffer)
        IF ((symroutine/='old').and.(symroutine/='new')) THEN
              CALL ErrorMsg('Error: Wrong value of keyword searchsymmetry. The second string can be "old" or "new"!')
          ErrCode=100
            ENDIF
      ENDIF

        ELSEIF (CurKwdId==kid_coinc) THEN
          CALL OneInt(InBuffer,CoincCode,ErrCode)
      IF ((errcode==0).and.(CoincCode/=1).and.(CoinCCode/=2)) THEN
            CALL ErrorMsg('Error, the value of '//trim(CurKwd)//' must be 1 or 2!')
            ErrCode=100
      END IF

        ELSEIF (CurKwdId==kid_histogram) THEN
!      IF (NPass==2) THEN
          IF (firstword(InBuffer,.false.)=='composition') THEN !Calculate histogram from the composition by making a random structure
            IF ((RDim/=3).or.(SDim/=3)) THEN
              CALL ErrorMsg('Error: Calculation of histogram from composition works only for ordinary 3D structures.'//&
                    ' If you want to use the histogram-matching option, please supply an external histogram')
              ErrCode=100
            ELSEIF (ElNr==0) THEN
              CALL ErrorMsg('Error: Unknown chemical composition, the histogram cannot be calculated from composition.'// &
                    'Use the keyword "composition" to define the type and number of atoms in the unit cell.')
              ErrCode=100
            ELSE
              Buffer=FirstWord(InBuffer,.true.) !Remove the tag "composition" from InBuffer
              IF ((len_trim(InBuffer)==0).or.(InBuffer=='wilson')) THEN
                BisoHM=-100. !Flag that Biso is to be taken from Wilson plot
              ELSE
                CALL OneReal(InBuffer,BisoHM,ErrCode)
                IF (ErrCode/=0) THEN
              CALL ErrorMsg('Error - the second value after "histogram composition" must be either'// &
                   ' ''wilson'' or a real number!')
              ENDIF
            ENDIF
            HMClasses=-10 !Flag for histogram from random structure
          ENDIF
          ELSEIF (firstword(InBuffer,.false.)=='structure') THEN
              IF ((RDim/=3).or.(SDim/=3)) THEN
                CALL ErrorMsg('Error: Calculation of histogram from composition works only for ordinary 3D structures.'//&
                      ' If you want to use the histogram-matching option, please supply an external histogram')
                ErrCode=100
              ELSE
                Buffer=FirstWord(InBufferOrig,.true.) !Remove the tag "structure" from InBuffer
                ModelStrFile=FirstWord(InBufferOrig,.true.) 
            IF (len_trim(ModelStrFile)==0) THEN
                  CALL ErrorMsg('Error - the keyword "histogram structure" must be followed by the name of the structure file!')
              ErrCode=100
            ENDIF
            HMClasses=-20
              ENDIF
        ELSEIF (len_trim(InBuffer)/=0) THEN
              CALL ErrorMsg('Error - the allowed first value of the keyword histogram is composition or structure.')
          ErrCode=100
        ELSE
              CALL ReadHistogram(kid_histogram,ErrCode)
            ENDIF
!          ENDIF
!       HM parameters
        ELSEIF (CurKwdId==kid_hmpar) THEN
          Buffer=FirstWord(Inbuffer,.true.)
      IF (len_trim(buffer)==0) THEN
        ErrCode=100
        CALL ErrorMsg('Error, two values expected after the keyword '//trim(CurKwd)//'!')
          ELSE
        CALL OneInt(Buffer,Inhmstartc,ErrCode)
          ENDIF
      IF (ErrCode==0) THEN
            Buffer=FirstWord(Inbuffer,.true.)
        IF (len_trim(buffer)==0) THEN
          ErrCode=100
          CALL ErrorMsg('Error, two values expected after the keyword '//trim(CurKwd)//'!')
            ELSE
          CALL OneInt(Buffer,Inhmstepc,ErrCode)
            ENDIF
          ENDIF
      IF ((errcode==0).and.((inhmstartc<0).or.((inhmstartc>0).and.(inhmstepc<=0)))) THEN
            CALL ErrorMsg('Error, both values of '//trim(CurKwd)//' must be positive!')
            ErrCode=100
      END IF
          IF ((ErrCode==0).and.(len_trim(InBuffer)>0)) THEN
            Buffer=FirstWord(Inbuffer,.true.)
        CALL YesNo(Buffer,hmrepart,ErrCode)
        IF (ErrCode/=0) THEN
              CALL ErrorMsg('Error, the third value of '//trim(CurKwd)//' must be yes or no!')
        ENDIF
      ENDIF
        ELSEIF (CurKwdId==kid_hmwait) THEN
          Buffer=FirstWord(Inbuffer,.true.)
          CALL YesNo(Buffer,hmwait,ErrCode)

!       SM parameters - for symmetry matching
        ELSEIF (CurKwdId==kid_smpar) THEN
      symmatch=.true.
          Buffer=FirstWord(Inbuffer,.true.)
      IF (len_trim(buffer)==0) THEN
        ErrCode=100
        CALL ErrorMsg('Error, two values expected after the keyword '//trim(CurKwd)//'!')
          ELSE
        CALL OneInt(Buffer,Insmstartc,ErrCode)
          ENDIF
      IF (ErrCode==0) THEN
            Buffer=FirstWord(Inbuffer,.true.)
        IF (len_trim(buffer)==0) THEN
          ErrCode=100
          CALL ErrorMsg('Error, two values expected after the keyword '//trim(CurKwd)//'!')
            ELSE
          CALL OneInt(Buffer,Insmstepc,ErrCode)
            ENDIF
          ENDIF
      IF ((errcode==0).and.((insmstartc<0).or.((insmstartc>0).and.(insmstepc<=0)))) THEN
            CALL ErrorMsg('Error, both values of '//trim(CurKwd)//' must be positive!')
            ErrCode=100
      END IF
          IF ((ErrCode==0).and.(len_trim(InBuffer)>0)) THEN
            Buffer=FirstWord(Inbuffer,.true.)
        CALL YesNo(Buffer,smrepart,ErrCode)
        IF (ErrCode/=0) THEN
              CALL ErrorMsg('Error, the third value of '//trim(CurKwd)//' must be yes or no!')
        ENDIF
      ENDIF

! Parameters for matching envelope
        ELSEIF (CurKwdId==kid_envpar) THEN
          Buffer=FirstWord(Inbuffer,.true.)
      IF (len_trim(buffer)==0) THEN
        ErrCode=100
        CALL ErrorMsg('Error, at least two values expected after the keyword '//trim(CurKwd)//'!')
          ELSE
        CALL OneInt(Buffer,envstartc,ErrCode)
          ENDIF
      IF (ErrCode==0) THEN
            Buffer=FirstWord(Inbuffer,.true.)
        IF (len_trim(buffer)==0) THEN
          ErrCode=100
          CALL ErrorMsg('Error, at least two values expected after the keyword '//trim(CurKwd)//'!')
            ELSE
          CALL OneInt(Buffer,envstepc,ErrCode)
            ENDIF
          ENDIF
      IF ((errcode==0).and.((envstartc<=0).or.(envstepc<=0))) THEN
            CALL ErrorMsg('Error, both values of '//trim(CurKwd)//' must be positive integers!')
            ErrCode=100
      END IF
          IF (len_trim(InBuffer)/=0) THEN !Zero level for the envelope is given
            Buffer=FirstWord(Inbuffer,.true.)
          CALL OneReal(Buffer,EnvZeroLevel,ErrCode)
        IF (ErrCode/=0) THEN
          CALL ErrorMsg('Error, cannot read the third parameter of the keyword '//trim(CurKwd)//'!')
            ENDIF          
      ENDIF
          IF (len_trim(InBuffer)/=0) THEN !Cutmin is given
            Buffer=FirstWord(Inbuffer,.true.)
          CALL OneReal(Buffer,EnvCutMin,ErrCode)
        IF (ErrCode/=0) THEN
          CALL ErrorMsg('Error, cannot read the fourth parameter of the keyword '//trim(CurKwd)//'!')
            ENDIF          
      ENDIF

        ELSEIF (CurKwdId==kid_modelrepart) THEN
          CALL YesNo(InBuffer,modelrepart,ErrCode)
 
        ELSEIF (CurKwdId==kid_bestden) THEN
          CALL OneInt(InBuffer,Nbestrho,ErrCode)
      IF ((errcode==0).and.(NbestRho<=0)) THEN
            CALL ErrorMsg('Error, the value of '//trim(CurKwd)//' must be positive!')
            ErrCode=100
      END IF
      IF ((ErrCode==0).and.(len_trim(InBuffer)/=0)) THEN
        SELECT CASE (InBuffer)
          CASE('rvalue')
            SaveRhoFOM=FOMRvalue
          CASE('peakiness')
            SaveRhoFOM=FOMpeakiness
          CASE('reference')
            SaveRhoFOM=FOMReference
          CASE('symmetry')
            SaveRhoFOM=FOMSymAgr
              CASE DEFAULT
                CALL ErrorMsg('Error, the second value of '//trim(CurKwd)//' must be rvalue, reference or symmetry!')
                ErrCode=100
        END SELECT
      ENDIF

        ELSEIF (CurKwdId==kid_repeatmode) THEN
          RepeatMode=Firstword(InBuffer,.true.)
      IF ((RepeatMode/='never').and.(RepeatMode/='nosuccess').and.(RepeatMode/='always'))  THEN
        ! Try to read the number of repetitions
        CALL OneInt(RepeatMode,repeatattempts,ErrCode)
        IF (ErrCode/=0) THEN
          CALL ErrorMsg('Error: Wrong value of keyword repeatmode. The allowed values are "never", "nosuccess", "always"'// &
                        ' or an integer number of repetitions!')
            ELSE
          RepeatMode='number'
        ENDIF
        !Do we want to add up densities?
        IF (len_trim(InBuffer)/=0) THEN
          IF (trim(InBuffer)=='sumgood') THEN
            WantSumRho=sumgood
          ELSEIF (trim(InBuffer)=='sumall') THEN
            WantSumRho=sumall
              ELSE
            CALL ErrorMsg('Error: Wrong second value of keyword repeatmode. The allowed values are sumall or sumgood!')
          ENDIF
        ENDIF
      ENDIF

        ELSEIF (CurKwdId==kid_convmode) THEN
          InConvergenceMode=Firstword(InBuffer,.true.)
      IF ((InConvergenceMode/='never').and.(InConvergenceMode/='normal').and. &
          (InConvergenceMode/='rvalue').and.(InConvergenceMode/='charge').and.&
          (InConvergenceMode/='peakiness').and.(InConvergenceMode/='symmetry'))  THEN
            CALL ErrorMsg('Error: Wrong value of keyword convergencemode. The allowed values are "never", "normal", "rvalue",'// &
                      ' "charge", symmetry or "peakiness".')
            ErrCode=100
      ELSEIF((InConvergenceMode=='normal').and.(len_trim(InBuffer)>0)) THEN
        CALL OneReal(InBuffer,LimitFOM,ErrCode)
        IF ((ErrCode/=0).or.(LimitFOM<0.0)) THEN
          CALL ErrorMsg('Error: The second parameter of "convergencemode normal" must be a positive number.')
          ErrCode=100
        ENDIF
      ELSEIF((InConvergenceMode=='peakiness').and.(len_trim(InBuffer)>0)) THEN
        CALL OneReal(InBuffer,goodskew,ErrCode)
        IF ((ErrCode/=0).or.(goodskew<1.0)) THEN
          CALL ErrorMsg('Error: The second parameter of "convergencemode peakiness" must be a number larger than 1.')
          ErrCode=100
        ENDIF
      ELSEIF((InConvergenceMode=='symmetry').and.(len_trim(InBuffer)>0)) THEN
        CALL OneReal(InBuffer,goodsym,ErrCode)
        IF ((ErrCode/=0).or.(goodsym>100.)) THEN
          CALL ErrorMsg('Error: The second parameter of "convergencemode symmetry" must be a number smaller than 100.')
          ErrCode=100
        ENDIF
      ELSEIF((InConvergenceMode=='rvalue').and.(len_trim(InBuffer)>0)) THEN
        CALL OneReal(InBuffer,goodrval,ErrCode)
        IF ((ErrCode/=0).or.(goodrval<0.0)) THEN
          CALL ErrorMsg('Error: The second parameter of "convergencemode rvalue" must be a positive real number.')
          ErrCode=100
        ELSE
          goodrval=goodrval/100.
        ENDIF
      ELSEIF((InConvergenceMode=='charge').and.(len_trim(InBuffer)>0)) THEN
        CALL OneReal(InBuffer,goodcharge,ErrCode)
        IF ((ErrCode/=0).or.(goodcharge<0.0)) THEN
          CALL ErrorMsg('Error: The second parameter of "convergencemode charge" must be a positive real number.')
          ErrCode=100
        ENDIF
      ENDIF

        ELSEIF (CurKwdId==kid_polish) THEN
          CALL YesNo(InBuffer,Polish,ErrCode)
      IF (ErrCode==0) THEN
            Buffer=FirstWord(InBuffer,.true.) !Get rid of the yes in InBuffer
        IF ((polish).and.(len_trim(inbuffer)>0)) THEN !Try to read the number of polishcycles
              CALL OneInt(InBuffer,polishcycles,ErrCode)
          IF (ErrCode/=0) CALL ErrorMsg('Error, ''polish yes'' can be followed only by an integer number of cycles.')
        ENDIF
          ENDIF

        ELSEIF (CurKwdId==kid_mem) THEN
          CALL YesNo(InBuffer,mem,ErrCode)
          IF (ErrCode==0) THEN
            Buffer=FirstWord(InBuffer,.true.) !Get rid of the yes in InBuffer
            Buffer=FirstWord(InBuffer,.true.)
            IF (len_trim(buffer)>0) THEN !Try to read the number of memcycles
              CALL OneInt(Buffer,MEMcycles,ErrCode)
            ENDIF
            Buffer=FirstWord(InBuffer,.true.)
            IF (len_trim(buffer)>0) THEN !Try to read the number of polishcycles
              CALL OneReal(Buffer,MEMlimit,ErrCode)
            ENDIF
          ENDIF

        ELSEIF (CurKwdId==kid_fixf0) THEN
          CALL YesNo(InBuffer,fixf0,ErrCode)

!       detailed information about reflections in every cycle
        ELSEIF (CurKwdId==kid_detph) THEN
          CALL OneInt(InBuffer,detailedphases,ErrCode)

        ELSEIF (CurKwdId==kid_analpeaks) THEN
          CALL YesNo(InBuffer,AnalPeaks,ErrCode)
 
        ELSEIF (CurKwdId==kid_clusan) THEN
          CALL YesNo(InBuffer,ClusAn,ErrCode)
 
        ELSEIF (CurKwdId==kid_progress) THEN
          Buffer=FirstWord(InBuffer,.true.)
      IF (Buffer=='no') THEN
            ViewProgress=noviewer
      ELSEIF (Buffer=='chimera') THEN
            ViewProgress=chimera
      ELSEIF (Buffer=='vesta') THEN
            ViewProgress=vesta
          ELSE
            CALL ErrorMsg('Error: The only allowed value of viewprogress is currently "chimera" and "vesta".')
            ErrCode=100
      ENDIF
          IF ((ErrCode==0).and.(len_trim(InBuffer)/=0)) THEN
        CALL OneInt(Inbuffer,ViewProgressPeriod,ErrCode)
        IF ((ErrCode/=0).or.(ViewProgressPeriod<1)) &
          CALL ErrorMsg('Error, the second value of viewprogress must be a positive integer')
      ENDIF

        ELSEIF (CurKwdId==kid_presmode) THEN
          IF (NPass==1) THEN
            i=0
        DO
              Buffer=FirstWord(InBuffer,.true.)
          IF (len_trim(buffer)>0) THEN
            i=i+1
          ELSE
             EXIT
          ENDIF
        ENDDO
        IF (mod(i,2)/=0) THEN
              CALL ErrorMsg('Error: Even number of parameters expected after presentationmode')
              ErrCode=100
            ELSE
          allocate(PresModeTmp(1:i))
              allocate(PresModeLimits(1:i/2))
              allocate(PresModeTimes(1:i/2))
              PresModeNr=i/2
              Buffer=InBufferOrig
          CALL ArrayInt(Buffer,PresModeTmp,i,ErrCode)
          IF (ErrCode==0) THEN
            PresModeLimits=PresModeTmp((/(ii,ii=1,i-1,2)/))
            PresModeTimes=PresModeTmp((/(ii,ii=2,i,2)/))
          ENDIF
          DEALLOCATE(PresModeTmp)
        ENDIF
      ENDIF
    END IF !End of the ifblock containing handling of all keywords
        IF (ErrCode/=0) THEN
      CALL ErrorMsg('Error reading keyword '//trim(Kwd(CurKwdId)%name)//'.')
      CALL ErrorMsg(' ')
          ErrNr=ErrNr+1
    END IF


!*** End of keyword specific operations ***                 
        GOTO 100 !Go to reading next line
200     CONTINUE        
        ! Take care of the only case when two keywords are mutually exclusive
    IF ((Kwd(kid_centers)%occ==1).and.(sgcode/=sgcode_explicit)) THEN
          CALL ErrorMsg('Error, keyword centers must not be present, if symmetry is given by a symbol or space-group number!')
          ErrNr=ErrNr+1
        END IF

        CALL ErrorCheck !Check for errors during reading and take corresponding actions
        CALL AfterPass  !Actions to be taken after the file is read up to the end
        REWIND(infunit)
      END DO
!     Successfully finished reading
      IF (len_trim(DataItemWidths)/=0) DEALLOCATE(DataItemPos)
      IF (Terminal) write(*,'(a)') 'The input file '//trim(infile)//' was succesfully read.'
      GOTO 9999

!     error messages            
!     Only if no other error has been reported
900   IF (ErrNr==0) CALL StopProgram('Error reading the input file '//trim(infile)//'.')
      
9999  CLOSE(infunit)
      CONTINUE      

      CONTAINS

! This routine typically just reads the name of the input file
!****************************************
        SUBROUTINE InitInput
!****************************************
!          USE F90_UNIX_ENV, ONLY : GETARG,IARGC
      infile=' '
      IF (IARGC()>0) THEN
            CALL GETARG(1,inFile)
      ENDIF
      IF ((inFile=='-V').or.(inFile=='--version')) THEN
        write(*,'(a)') 'Superflip version: '//trim(versionstring)
        STOP
      ENDIF
100       continue
      IF (len_trim(infile)>0) THEN
            OPEN (UNIT=infunit,FILE=trim(Infile),IOSTAT=ErrCode,STATUS='OLD')
          ELSE
        WRITE(*,'(a)', ADVANCE='NO') 'Please, give the name of the input file: '
            READ(*,'(a)') inFile
        GOTO 100
      END IF
          IF (ErrCode/=0) THEN
            CALL ErrorMsg('Error, can''t open file "'//trim(Infile)//'".')
            inFile=''
        goto 100
        ErrCode=0
      END IF
!           Get number of cycles to compute 
          Buffer = ' '
      IF (IARGC()>1) THEN
            CALL GETARG(2,Buffer)
          END IF
          IF (len_trim(buffer)>0) THEN
            CALL OneInt(Buffer,InMaxCycles,ErrCode)
            IF ((ErrCode/=0).or.(InMaxCycles<0)) THEN
              CALL ErrorMsg('Error: The second program argument (maximum number of cycles) could not be read properly.')
          CALL StopProgram('Error: The maximum number of cycles must be a non-negative integer.')
        ENDIF
      ENDIF
      RETURN
    END SUBROUTINE InitInput


! Handles all actions taken after a pass is finished - typically allocation of dynamically allocated arrays
! and definitions of the group status
!****************************************
        SUBROUTINE AfterPass
!****************************************
          integer :: ErrCode2
      real :: multsc

          ErrCode=0
          IF (NPass==1) THEN
! Define the file names
            IF (len_trim(FileBase)==0) THEN !No filebase was given in the input file -> derive from the name of the input file
              cnt2=0
              DO cnt1=len_trim(infile),1,-1
            IF (infile(cnt1:cnt1)=='.') THEN
          cnt2=cnt1
          EXIT
        ENDIF
          END DO
              IF (cnt2==0) THEN !No dot found in the name of the input file
            Filebase=trim(Infile)
          ELSE ! Define Filebase as everything before the last dot
                Filebase=InFile(:cnt2-1)
          ENDIF
        ENDIF
            cnt1=len_trim(Filebase)
            LogFile=Filebase(:cnt1)//trim(lgext)
            IF (rewriteoutput) THEN
              OPEN (UNIT=lgu,FILE=trim(LogFile),IoStat=ErrCode2, POSITION='REWIND')
            ELSE
              OPEN (UNIT=lgu,FILE=trim(LogFile),IoStat=ErrCode2, POSITION='APPEND')
        ENDIF
        IF (ErrCode2/=0) THEN
          CALL ErrorMsg('Error opening log file '//trim(LogFile)//'. Check the access rights!')
              ErrCode=ErrCode2
        ELSE
          CALL VersionStamp(.true.)
        END IF
            if (len_trim(comfile)==0) ComFile=FileBase(:cnt1)//trim(cmext)
            TmpFile=FileBase(:cnt1)//trim(tmpext)


            IF (detailedphases>0) THEN
              DetPhFile=Filebase(:cnt1)//'.sfdtp'
              OPEN (UNIT=detphu,FILE=trim(DetPhFile),IoStat=ErrCode2, POSITION='REWIND')
              IF (ErrCode2/=0) THEN
            CALL ErrorMsg('Error opening file '//trim(DetPhFile)//'. Check the access rights!')
                ErrCode=ErrCode2
          END IF
            ENDIF
  
        IF (SDim>RDim) CALL FlagKwdGroup('4D',.true.)
            IF (SDim<RDim) THEN
          ErrCode=100
          CALL ErrorMsg('Error, total dimension is smaller than the dimension of the physical space.')
        ENDIF

            IF (InPerform/='symmetry') THEN
          CALL FlagKwdGroup('refl',.true.)
        ELSE
          CALL FlagKwdGroup('model',.true.)
            ENDIF
        IF (ResUnits==res_d) THEN
              IF (UserDefinedFillMissing) FillMissingSthl=1./(2.*max(FillMissingSthl,eps))
          IF (ResLimTmp2>-1.) THEN !both limits are given
                refldmin=min(reslimtmp1, reslimtmp2)
                refldmax=max(reslimtmp1, reslimtmp2)
                reflsthlmax=1./(2.*max(refldmin,eps))
            reflsthlmin=1./(2.*max(refldmax,eps))
          ELSEIF (ResLimTmp1>-1.) THEN !only the lower d value is given
            refldmin=reslimtmp1
                reflsthlmax=1./(2.*max(refldmin,eps))
          ENDIF
        ELSEIF (ResUnits==res_sthl) THEN
          IF (ResLimTmp2>-1.) THEN !both limits are given
                reflsthlmin=min(reslimtmp1, reslimtmp2)
                reflsthlmax=max(reslimtmp1, reslimtmp2)
                refldmin=1./(2.*max(reflsthlmax,eps))
            refldmax=1./(2.*max(reflsthlmin,eps))
          ELSEIF (ResLimTmp1>-1.) THEN !only the upper limit on sin(th)/lambda value is given
            reflsthlmax=reslimtmp1
                refldmin=1./(2.*max(reflsthlmax,eps))
          ENDIF
            ENDIF


        NCellPar=RDim*(RDim+1)/2
        ALLOCATE(Cell(1:NCellPar+1),STAT=ErrCode2)
            IF (ErrCode2/=0) CALL StopProgram('Error allocating array Cell! Too little memory?')

        ! Allocate the array holding the number of voxels
            ALLOCATE(NVoxIter(1:SDim),STAT=ErrCode2)
            IF (ErrCode2/=0) CALL StopProgram('Error allocating array NVox! Too little memory?')
            ALLOCATE(NVoxFine(1:SDim),STAT=ErrCode2)
            IF (ErrCode2/=0) CALL StopProgram('Error allocating array NVox! Too little memory?')
            ALLOCATE(NVox(1:SDim),STAT=ErrCode2)
            IF (ErrCode2/=0) CALL StopProgram('Error allocating array NVox! Too little memory?')

        ! Allocate the array holding the number of elements in the SF array
            ALLOCATE(NVoxSF(1:SDim),STAT=ErrCode2)
            IF (ErrCode2/=0) CALL StopProgram('Error allocating array NVoxSF! Too little memory?')

            ALLOCATE(DimStep(1:SDim),STAT=ErrCode2)
            IF (ErrCode2/=0) CALL StopProgram('Error allocating array DimStep! Too little memory?')

            ALLOCATE(DimStepSF(1:SDim),STAT=ErrCode2)
            IF (ErrCode2/=0) CALL StopProgram('Error allocating array DimStepSF! Too little memory?')

        ! The qvectors, if necessary
        IF (SDim>RDim) THEN
              ALLOCATE(QVec(1:RDim,1:SDim-RDim),STAT=ErrCode2)
              IF (ErrCode2/=0) CALL StopProgram('Error allocating array QVec! Too little memory?')
            END IF

        ! Symmetry
            IF (sgcode==sgcode_hall) THEN
              CALL SGfromHall(insgsymbol,NCenter,Center, NSym,SmOp)
              IF (NSym==0) THEN
                Buffer='Error, '//trim(insgsymbol)//' is not a valid Hall number.'
                CALL StopProgram(trim(Buffer))
              ENDIF
              CALL ReallocateSG(SmOp,SDim,NSym,NSym*NCenter+NTestSym+1)
            ELSEIF (sgcode==sgcode_number) THEN              
          Buffer=insgsymbol
              CALL OneInt(buffer,cnt1,ErrCode)
              IF (ErrCode/=0) THEN
                Buffer='Error, '//trim(insgsymbol)//' is not a valid space-group number.'
                CALL StopProgram(trim(Buffer))
                ErrCode=100
              ELSE
                ii=0
                DO i=1,SGlistLength
              IF ((cnt1==SGlist(i)%sgnr).and.(SGlist(i)%std=='S')) THEN
                    ii=i
                    EXIT
                  ENDIF
                ENDDO
                IF (ii/=0) THEN
                  CALL SGfromHall(SGlist(ii)%hall,NCenter,Center, NSym,SmOp)
                  CALL ReallocateSG(SmOp,SDim,NSym,NSym*NCenter+NTestSym+1)
                ELSE
                  Buffer='Error, '//trim(insgsymbol)//' is not a valid space-group number.'
                  CALL StopProgram(trim(Buffer))
                  ErrCode=100                
                ENDIF
              ENDIF
            ELSEIF (sgcode==sgcode_hm) THEN
          ii=0
              DO i=1,SGlistLength
            IF (trim(insgsymbol)==SGlist(i)%HM) THEN
                  ii=i
                  EXIT
                ENDIF
              ENDDO
              IF (ii/=0) THEN
                CALL SGfromHall(SGlist(ii)%hall,NCenter,Center, NSym,SmOp)
                CALL ReallocateSG(SmOp,SDim,NSym,NSym*Ncenter+NTestSym+1)
              ELSE
                Buffer='Error, the HM symbol '//trim(insgsymbol)//' was not found in the database.'
                CALL StopProgram(trim(Buffer))
                ErrCode=100                
              ENDIF
            ELSEIF (sgcode==sgcode_fp) THEN
              CALL StopProgram('Sorry, input of symmetry via fingerprint is not yet implemented.') 
              ErrCode=100
            ELSE
              IF (NCenter==0) NCenter=1 !For the case no centering vectors were given - the zero centering will be added automatically
              CALL AllocateSG(SmOp,SDim,NSym*NCenter+NTestSym+1)
              ALLOCATE(Center(1:SDim,1:NCenter),STAT=ErrCode2)
              IF (ErrCode2/=0) CALL StopProgram('Error allocating array Center! Too little memory?')
        ENDIF
            IF (derivesym=='use') THEN !We have to make a backup of the original symmetry operations
              CALL AllocateSG(SmOpOrig,SDim,size(SmOp))
              ALLOCATE(CenterOrig(1:SDim,1:NCenter),STAT=ErrCode2)
              IF (ErrCode2/=0) CALL StopProgram('Error allocating array CenterOrig! Too little memory?')
            ENDIF
            IF (NTwin>0) THEN
              ALLOCATE(TwinOp(NTwin),STAT=ErrCode2)
              DO cnt1=1,NTwin
            ALLOCATE(TwinOp(cnt1)%R(SDim,SDim),STAT=ErrCode2)
              ENDDO
              IF (ErrCode2/=0) CALL StopProgram('Error allocating twin operations! Too little memory?')
        ENDIF

! Reflections
            IF (TotalF>0) THEN
          ALLOCATE(Refl(1:TotalF),STAT=ErrCode2)
          DO i=1,TotalF
                ALLOCATE(Refl(i)%hklFG(1:SDim),STAT=ErrCode2)
          ENDDO
              IF (ErrCode2/=0) CALL StopProgram('Error allocating array Refl! Too little memory?')
            ENDIF
! Histogram 
            IF (hmclasses>0) THEN
              ALLOCATE(HMGramm(0:hmclasses),STAT=ErrCode2)
              IF (ErrCode2/=0) CALL StopProgram('Error allocating array HMGramm! Too little memory?')
        ENDIF
! The best R-values:
            ALLOCATE(BestRhoProps(1:Nbestrho),STAT=ErrCode2)
            IF (ErrCode2/=0) CALL StopProgram('Error allocating array BestRhoProps! Too little memory?')


! Reflection list has fixed format - record positions of the break points in the line
          IF (len_trim(DataItemWidths)>0) THEN
            InBuffer=DataItemWidths
        ALLOCATE(DataItemPos(1:NDataItems+SDim+1))
        DataItemPos(1)=1
            Buffer=FirstWord(InBuffer,.true.)
            CALL OneInt(Buffer,ii,ErrCode)
        IF (ErrCode/=0) RETURN
        DO i=1,SDim
          DataItemPos(i+1)=DataItemPos(i)+ii
        ENDDO
        DO i=SDim+1,NDataItems+SDim
          Buffer=FirstWord(InBuffer,.true.)
          IF (len_trim(buffer)==0) THEN
            Call ErrorMsg('Error - less items in keyword '//trim(Kwd(kid_dataiw)%name)// &
                          ' than in '//trim(Kwd(kid_dataformat)%name))
                EXIT
          ENDIF
          CALL OneInt(Buffer,ii,ErrCode)
          IF (ErrCode/=0) EXIT
          DataItemPos(i+1)=DataItemPos(i)+ii
        ENDDO
      ENDIF
      IF (InPerform=='symmetry') RepeatMode='never'

          ELSEIF (NPass==2) THEN
          IF (.not.UserDefinedFillMissing) FillMissingSthl=min(FillMissingSthl,MaxInputSthl)
          IF (.not.allocated(G)) CALL AllFromCell(ErrCode)
          IF ((TotalF==0).and.(IterGridResolution<eps)) IterGridResolution=0.7
        IF (autovoxels) CALL GetAutoVoxels(NVoxIter,IterGridResolution, IterShannonFactor)
        IF (autofinevoxels) THEN
          CALL GetAutoVoxels(NVoxFine,FineGridResolution,FineShannonFactor)
        ELSEIF(all(NVoxFine==0)) THEN
          NVoxFine=NVoxIter
        ENDIF
        NVox=NVoxIter
      ! Assign the dimensions of the SF array, calculate related quantities
            CALL NVoxRelatedArrays(NVox, NVoxSF, DimStep, DimStepSF, NVoxTot, NVoxTotSF)
            cApVx=Cell(NCellPar+1)/NVoxTot

            CALL SetGdmPar(InPerform)
        InGdmPar=GdmPar
! If InPerform='symmetry', or random structure generated, it can happen that there are no reflections in the input file. In that case they will be calculated from the modelrho,
! but the array must be allocated and the indices filled
        IF (TotalF==0) THEN
              Allocate(Refl(1:NVoxTotSF),STAT=ErrCode2)
              DO cnt1=1,NVoxTotSF
                ALLOCATE(Refl(cnt1)%hklFG(1:SDim),STAT=ErrCode2)
              ENDDO
          TotalAllocatedF=NVoxTotSF
              IF (ErrCode2/=0) CALL StopProgram('Error allocating array Refl! Too little memory?')
          cnt2=0
          DO cnt1=0,NVoxTotSF-1
            CALL CoordinatesSF(cnt1*2,Refl(cnt2+1)%hklFG)
                DO i=1,SDim
                  Refl(cnt2+1)%hklFG(i)=IMapToRange(Refl(cnt2+1)%hklFG(i),-NVox(i)/2,(NVox(i)+1)/2)
                ENDDO
!                dummy=sqrt(multsc(float(MEMRefl(cnt2+1)%hklFG),float(MEMRefl(cnt2+1)%hklFG),SDim,.true.))/2
                dummy=sinthl(float(Refl(cnt2+1)%hklFG))
        IF ((dummy<ResForSym).and.(dummy>=reflsthlmin).and.(dummy<=reflsthlmax)) THEN
!                IF (MEMRefl(cnt2+1)%hklFG(1)<NVox(1)/2) THEN !exclude the reflections in the middle of the odd division
                  IF (all(abs(Refl(cnt2+1)%hklFG)<NVox/2)) THEN !exclude the reflections in the middle of the odd division
                cnt2=cnt2+1
                Refl(cnt2)%F=0.0
                Refl(cnt2)%FMEM=0.0
                Refl(cnt2)%AMEM=0.0
                Refl(cnt2)%BMEM=0.0
            Refl(cnt2)%SFLoc=cnt1*2
          ENDIF
        ENDIF
          ENDDO
          StartF=1; EndF=cnt2;TotalF=cnt2
        ENDIF
! Try to guess the types of the density files
            DO cnt1=1,OutRhoNr
              IF (OutRhoType(cnt1)=='none') THEN
            CALL GuessType(OutRhoFile(cnt1),OutRhoType(cnt1))
            IF (OutRhoType(cnt1)=='unknown'.or.OutRhoType(cnt1)=='cif') THEN
              CALL ErrorMsg('Error, keyword outputformat is missing and I cannot guess the type from the file names!')
              CALL ErrorMsg('Please use the keyword outputformat or standard file extensions.')
                  ErrCode=100
            ENDIF
              ENDIF
        ENDDO
        IF ((RefRhoFile/='').and.(RefRhoType=='none')) THEN
          CALL GuessType(RefRhoFile,RefRhoType)
          IF (RefRhoType=='unknown') THEN
            CALL ErrorMsg('Error, keyword referenceformat is missing and I cannot guess the type from the file name!')
            CALL ErrorMsg('Please use the keyword referenceformat or one of the standard file extensions.')
          ENDIF
            ENDIF
        IF ((ModelRhoFile/='').and.(ModelRhoType=='none')) THEN
          CALL GuessType(ModelRhoFile,ModelRhoType)
              IF (ModelRhoType=='unknown') THEN
                CALL ErrorMsg('Error, keyword modelformat is missing and I cannot guess the type from the file name!') 
            CALL ErrorMsg('Please use the keyword modelformat or one of the standard file extensions.')
          ENDIF
            ENDIF
        IF ((EnvRhoFile/='').and.(EnvRhoType=='none')) THEN
          CALL GuessType(EnvRhoFile,EnvRhoType)
          IF (EnvRhoType=='unknown') THEN
            CALL ErrorMsg('Error, keyword envelopeformat is missing and I cannot guess the type from the file name!')
            CALL ErrorMsg('Please use the keyword envelopeformat or one of the standard file extensions.')
          ENDIF
            ENDIF
        !Allocate structure factors
            ALLOCATE(sf(0:2*NVoxTotSF-1),STAT=ErrCode2)
            IF (ErrCode2/=0) CALL StopProgram('Error allocating array sf! Too little memory?')
            !Make the expansion and check of the symmetry elements and a general check of consistency
        CALL CompleteSymmetry(ErrCode)
        IF (ErrCode/=0) CALL StopProgram('Error processing the symmetry information.')
            IF (derivesym=='use') THEN !Make a backup copy of the symmetry operations
          NSymOrig=NSym
          NSymPrimOrig=NSymPrim
          NCenterOrig=NCenter
          DO cnt1=1,size(SmOp)
                CALL CopySymOp(SmOp(cnt1),SmOpOrig(cnt1))
          ENDDO
          CenterOrig=Center
        ENDIF
        !If FWHM have been given, get the groups out of FWHM
        IF ((HaveFWHM).and.(NGroups==0)) CALL FWHM2Groups
            ! Allocate the array for group intensities
        IF (NGroups>0) THEN
              ALLOCATE(GroupInt(1:NGroups),STAT=ErrCode2)
              IF (ErrCode2/=0) CALL StopProgram('Error allocating array GroupInt! Too little memory?')
        ENDIF
            IF (stabicyc<0) THEN !Stabicyc was not set by user
          IF (InConvergenceMode=='normal') THEN
            stabicyc=0
          ELSE
            stabicyc=abs(stabicyc)
          ENDIF
        ENDIF
            IF (MissingMode==mm_undefined) THEN !Set defaults for missing mode
          IF(NormalizeRefl==NormNone) THEN
            MissingMode=mm_float
          ELSE
            MissingMode=mm_bound
          ENDIF
            ENDIF
      END IF
      IF (ErrCode/=0) CALL StopProgram('')
          InMaxCycles=abs(InMaxCycles)
    END SUBROUTINE AfterPass

! Translate datafrmat shortcuts
!****************************************
        SUBROUTINE SetDataFormatPresets
!****************************************
        IMPLICIT NONE

    IF (InBuffer=='m91') THEN
          InBuffer='intensity'
      ReflStartLine=''
      ReflEndLine=' 999'
      dataitemwidths='4 9'
    ELSEIF (InBuffer=='m90') THEN
          InBuffer='intensity'
      ReflStartLine='Data Block'
      ReflEndLine='Data Block'
      dataitemwidths='4 9'
    ELSEIF (InBuffer=='shelx') THEN
          InBuffer='intensity'
      ReflStartLine=''
      ReflEndLine='   0   0   0'
      dataitemwidths='4 8'
    ELSEIF (InBuffer=='xplor') THEN
      hklxplor=.true.
    ENDIF
    END SUBROUTINE SetDataFormatPresets

! Get all possible quantities from the cell, like reciprocal cell, metric tensors, volumes
!****************************************
    SUBROUTINE AllFromCell(ErrCd)
!****************************************
      IMPLICIT NONE
      INTEGER :: ErrCd, poscnt
      real, allocatable, dimension(:,:) :: latmat !Matrix holding the coordinates of the lattice - for calculation of the superspace metric tensor
      integer, allocatable, dimension(:,:) :: sumgammai,sumgammar
      double precision :: csa !cosinus of the angle between two lattice vectors
      real :: multsc !function returning a scalar product
      ALLOCATE(G(1:RDim,1:RDim),STAT=ErrCd)
      IF (ErrCd/=0) CALL StopProgram('Error allocating metric tensor G! Too little memory?')
      ALLOCATE(GS(1:SDim,1:SDim),STAT=ErrCd)
      IF (ErrCd/=0) CALL StopProgram('Error allocating superspace metric tensor GS! Too little memory?')
      ALLOCATE(RCell(1:NCellPar+1),STAT=ErrCd)
      IF (ErrCd/=0)  CALL StopProgram('Error allocating array for reciprocal cell RCell! Too little memory?')
      ALLOCATE(Gi(1:RDim,1:RDim),STAT=ErrCd)
      IF (ErrCd/=0)  CALL StopProgram('Error allocating reciprocal metric tensor Gi! Too little memory?')
      ALLOCATE(GSi(1:SDim,1:SDim),STAT=ErrCd)
      IF (ErrCd/=0) CALL StopProgram('Error allocating reciprocal metric tensor GSi! Too little memory?')
      ALLOCATE(latmat(1:RDim,1:SDim),STAT=ErrCd)
      IF (ErrCd/=0) CALL StopProgram('Error allocating lattice matrix LatMat! Too little memory?')
      ALLOCATE(sumgammar(1:RDim,1:RDim),STAT=ErrCd)
      IF (ErrCd/=0) CALL StopProgram('Error allocating lattice matrix LatMat! Too little memory?')
      ALLOCATE(sumgammai(1:SDim-RDim,1:SDim-RDim),STAT=ErrCd)
      IF (ErrCd/=0) CALL StopProgram('Error allocating lattice matrix LatMat! Too little memory?')

!     Fill Metric tensor
      PosCnt=0
      DO cnt1=RDim,1,-1
        DO cnt2=RDim,cnt1,-1
          IF (cnt1==cnt2) THEN
            csa=1.0D0
          ELSE
            PosCnt=PosCnt+1
            IF (PosCnt+RDim>NCellPar) THEN
              CALL ErrorMsg('Zle maticko, zle je, chci vic uhlu nez mam.')
              write(buffer,*) RDim, poscnt, cnt1, cnt2
              CALL ErrorMsg(Buffer)
              CALL STopProgram('Internal inconsistency of the program. Please send this message to the author.')
            END IF
            csa=cos(torad*Cell(RDim+PosCnt))
          ENDIF
          G(cnt2,cnt1)=Cell(cnt1)*Cell(cnt2)*csa
        END DO
      END DO
!     Only the upper triangle is filled, the lower is symmetrical, so let's fill it too
      DO cnt1=2,RDim
        DO cnt2=1,cnt1-1
          G(cnt2,cnt1)=G(cnt1,cnt2)
        END DO
      END DO
!     Get the reciprocal metric tensor and volumes
      Gi=invert(G)
      IF (Gi(1,1)>1.0E29) THEN !Singular matrix
        Call ErrorMsg('Error, the cell parameters define a cell with zero volume. Please correct the parameters.')
        RETURN
      END IF
! If superspace is used, fill also the superspace metric tensor, which will be used for checking of the symmetry operations
      IF (SDim>RDim) THEN
        latmat=0.0
        DO cnt1=1,RDim
          latmat(cnt1,cnt1)=1.E0
        ENDDO
        DO cnt1=1,SDim-RDim
          DO cnt2=1,RDim
            latmat(cnt2,cnt1+RDim)=qvec(cnt2,cnt1)
          ENDDO
        ENDDO
        DO cnt1=1,SDim
          DO cnt2=1,cnt1
            GSi(cnt1,cnt2)=multsc(latmat(1:RDim,cnt1),latmat(1:RDim,cnt2),RDim,.true.)
            IF ((cnt1>RDim).and.(cnt2>RDim)) GSi(cnt1,cnt2)=2.*GSi(cnt1,cnt2) !The factor 2 here is very tricky. Comes from the properties of the superspace, and from the requirement that internal base vectors e1, e2... transform the same way as the corresponding q-vectors
            GSi(cnt2,cnt1)=GSi(cnt1,cnt2)
          ENDDO
        ENDDO
!        sumgammai=0.0
!        sumgammar=0.0
!        DO cnt1=1,NSym
!          call printsymop(SmOp(cnt1))
!          sumgammar=sumgammar+SmOp(cnt1)%R(1:RDim,1:RDim)
!          sumgammai=sumgammai+SmOp(cnt1)%R(RDim+1:SDim,RDim+1:SDim)
!        ENDDO
!        call printmat(sumgammaR)
!        write(*,*)
!        call printmat(sumgammaI)
!        read(*,*)
!      DO cnt1=RDim+1,SDim
!        GSi(cnt1,cnt1)=GSi(cnt1,cnt1)+1.0E0 !this is the correction for the scalar product of the vectors d in the internal space.
!      ENDDO
        GS=INVERT(GSi)
        IF (GS(1,1)>1.0E29) THEN !Singular matrix
          Call ErrorMsg('Error, the q-vectors define a superspace cell with zero volume. Please check the input file.')
          RETURN
        END IF
      ELSE !for the normal, non-modulated case, GS=G, GSi=Gi
        GS=G
        GSi=Gi
      ENDIF
      Cell(NCellPar+1)=sqrt(Determinant(G,eps))
      RCell(NCellPar+1)=1./Cell(NCellPar+1)

!     and now the reciprocal cell
      DO cnt1=1,RDim
        RCell(cnt1)=SQRT(Gi(cnt1,cnt1))
      ENDDO
      PosCnt=0
      DO cnt1=RDim,1,-1
        DO cnt2=cnt1-1,1,-1
          PosCnt=PosCnt+1
          IF (PosCnt+RDim>NCellPar) THEN
            CALL ErrorMsg('Zle maticko, zle je, chci vic prvku Gi nez mam.')
            write(buffer,*) RDim, poscnt, cnt1, cnt2
            CALL ErrorMsg(Buffer)
            CALL STopProgram('Internal inconsistency of the program. Please send this message to the author.')
          END IF
          RCell(PosCnt+RDim)=Gi(cnt2,cnt1)/(RCell(cnt1)*RCell(cnt2))
          RCell(PosCnt+RDim)=acos(RCell(PosCnt+RDim))/torad
        END DO
      END DO
      DEALLOCATE(latmat)
      deallocate(sumgammar,sumgammai)
    END SUBROUTINE AllFromCell


! Sort the reflections by ascending sin(theta)/lambda order and assign the group number according to the FWHM and FWHMlim
!****************************************
        SUBROUTINE FWHM2Groups
!****************************************
        USE SF_Module
    IMPLICIT NONE
    logical :: GoodOrder=.true., separnext,separprev
    real :: multsc
    real, dimension(SDim) :: dhkl
    type(RefTyp) :: DummyRefl
    !First check, if the reflections are sorted in ascending sinthl order
    dhkl=float(Refl(1)%hklFG)
    Refl(1)%BMEM=0.5*sqrt(multsc(dhkl,dhkl,SDim,.true.))
    REFL(1)%BMEM=2.*asin(Refl(1)%BMEM*lambda)/torad
    DO cnt1=StartF+1,EndF
      dhkl=float(Refl(cnt1)%hklFG)
      Refl(cnt1)%BMEM=0.5*sqrt(multsc(dhkl,dhkl,SDim,.true.))
      REFL(cnt1)%BMEM=2.*asin(Refl(cnt1)%BMEM*lambda)/torad
      IF (Refl(cnt1)%BMEM<Refl(cnt1-1)%BMEM) GoodOrder=.false.
    ENDDO
    IF (.not.GoodOrder) THEN
      !Make a bubblesort of reflections. I should use a more sophisticated sort, I know...
      DO cnt1=EndF,2,-1
        DO cnt2=1,cnt1-1
          IF (Refl(cnt2)%BMEM>Refl(cnt2+1)%BMEM) THEN
            DummyRefl=Refl(cnt2)
            Refl(cnt2)=Refl(cnt2+1)
            Refl(cnt2+1)=DummyRefl
          ENDIF
        ENDDO
      ENDDO
    ENDIF
   !Now go through the reflections and assign the group numbers
    separprev=(FWHMlim*(Refl(1)%AMEM+Refl(2)%AMEM)/2.<(Refl(2)%BMEM-Refl(1)%BMEM))
    IF (separprev) THEN
      Refl(1)%group=0 !First reflection separated
    ELSE
      NGroups=NGroups+1
      Refl(1)%group=NGroups
    ENDIF
    DO cnt1=StartF+1,EndF-1
      !Calculate the distance to the next reflection
      separnext=(FWHMlim*(Refl(cnt1)%AMEM+Refl(cnt1+1)%AMEM)/2.<(Refl(cnt1+1)%BMEM-Refl(cnt1)%BMEM))
!      write(*,*) cnt1,FWHMlim*(Refl(cnt1)%AMEM+Refl(cnt1+1)%AMEM)/2.,Refl(cnt1)%BMEM,Refl(cnt1+1)%BMEM
      IF (separprev.and.separnext) THEN
        Refl(cnt1)%group=0 !Separated reflection
      ELSEIF(separprev) THEN !Start of a new group
        NGroups=NGroups+1
        Refl(cnt1)%group=NGroups
      ELSE !Continuation of the group
        Refl(cnt1)%group=NGroups
      ENDIF
      separprev=separnext
    ENDDO
    IF (separprev) THEN
      Refl(EndF)%group=0 !Last reflection separated
    ELSE
      Refl(EndF)%group=NGroups
    ENDIF
        END SUBROUTINE FWHM2Groups

! Read the q-vectors
!****************************************
        SUBROUTINE QVectors(kid,ErrCd)
!****************************************
      INTEGER :: i,ErrCd,kid
      CHARACTER*50 :: ErrText
          DO i=1,Sdim-RDim
            CALL NextLine(InBuffer,ErrCd)
            
        IF (ErrCd/=0) RETURN !Some error occured
            IF (FirstWord(InBuffer,.false.)==Kwd(kid)%endname) THEN !Premature end of the block
              CALL ErrorMsg('Error - '//trim(Kwd(kid)%endname)//' detected prematurely.')
              WRITE(ErrText,'(i1,a)') Sdim-RDim,' qvectors expected.'
          CALL ErrorMsg(ErrText)
          ErrCd=100
              RETURN
        END IF
            CALL ArrayReal(InBuffer,QVec(1,i),RDim,ErrCd)
        IF (ErrCd/=0) RETURN !Some error occured
      END DO
          CALL NextLine(InBuffer,ErrCd)
          IF (FirstWord(InBuffer,.false.)/=Kwd(kid)%endname) THEN !The end-keyword is not here!
            CALL ErrorMsg('Too many lines before '//trim(Kwd(kid)%endname))
        ErrCd=100
      END IF
    END SUBROUTINE QVectors

! Handle the symmetry operations
!****************************************
        SUBROUTINE ReadSymmetry(kid,key,ErrCd)
!****************************************
      INTEGER :: ErrCd,kid,ErrCd2,NSymLocal,NSymStart
      INTEGER :: Key !Key=1  - normal symmetry, Key=2 - testing symmetry operations, Key=3 - twin
      REAL, DIMENSION(SDim) :: DVec !Dummy vector

          IF (NPass==1) THEN !Count symmetry operations
        i=index(InBuffer,':')
            IF (i/=0) THEN !symmetry given by symbol or SG number
              Buffer=InBuffer(1:i-1)
              SELECT CASE(trim(Buffer))
                CASE('ccp4','number')
                  sgcode=sgcode_number
                CASE('hm')
                  sgcode=sgcode_hm
                CASE('hall')
                  sgcode=sgcode_hall
                CASE('fp')
                  sgcode=sgcode_fp
              END SELECT
              insgsymbol=InBufferOrig(i+1:)
        ELSE !symmetry given explicitly
              NSymLocal=0
100           CALL NextLine(InBuffer,ErrCd)            
          IF (ErrCd<0) THEN !Preliminary end of file
                CALL ErrorMsg('Preliminary end of file. No keyword '//trim(Kwd(kid)%endname)//' found.')
            RETURN !Some error occured
              ELSEIF (ErrCd/=0) THEN
                IF ((Key==1).or.(Key==2)) THEN
                  WRITE(Buffer,'(a,i6)') 'Error reading symmetry operation number',NSym
            ELSEIF(Key==3) THEN
                  WRITE(Buffer,'(a,i6)') 'Error reading twin matrix number',NSym
            ENDIF
            CALL ErrorMsg(Buffer)
            RETURN
          ENDIF
              NSymLocal=NSymLocal+1
              IF (FirstWord(InBuffer,.false.)/=Kwd(kid)%endname) GOTO 100
              NSymLocal=NSymLocal-1
          IF (Key==1) THEN
            NSym=NSymLocal
          ELSEIF (Key==2) THEN
            NTestSym=NSymLocal
          ELSEIF (Key==3) THEN
            NTwin=NSymLocal
          ENDIF
            ENDIF
          ELSEIF (NPass==2) THEN !Read the symmetry operations
            IF (Key==1) THEN
              IF (sgcode/=sgcode_explicit) RETURN !In this case the symmetry has already been read
          NSymLocal=NSym
          NSymStart=1
        ELSEIF (Key==2) THEN
          NSymLocal=NTestSym+NSym*NCenter
          NSymStart=NSym*NCenter+1
        ELSEIF (Key==3) THEN
          NSymLocal=NTwin
          NSymStart=1
        ENDIF
            ErrCd2=0
        DO cnt1=NSymStart,NSymLocal
              CALL NextLine(InBuffer,ErrCd)            
          IF (ErrCd/=0) RETURN !Some error occured
          InBuffer=tolow(InBuffer)
          IF ((key==1).or.(key==2)) THEN
                CALL OneSym(InBuffer,SmOp(cnt1)%R,SmOp(cnt1)%t,cnt1,ErrCd)
          ELSEIF (key==3) THEN
                CALL OneSym(InBuffer,TwinOp(cnt1)%R,DVec,cnt1,ErrCd)
              ENDIF
          IF (ErrCd/=0) ErrCd2=100
        END DO
        ErrCd=ErrCd2
          ENDIF
    END SUBROUTINE ReadSymmetry

! Decode one symmetry operation into a matrix and a vector
!****************************************
        SUBROUTINE OneSym(SymString,InRMat,LocTVec,SymOpPos,ErrCd)
!****************************************
      INTEGER :: ErrCd,i,ii,col,curbeg,curend,jex,jelom,jebod,lgth
      INTEGER :: indx,indlom,nom,denom,TruePos,SymOpPos
          INTEGER, DIMENSION(SDim,SDim) :: InRmat
          REAL, DIMENSION(SDim,SDim) :: LocRmat
          REAL, DIMENSION(SDim) :: LocTvec
      CHARACTER*(*) :: SymString
      CHARACTER*(linelength) :: Row
      CHARACTER*(15) :: rowpart
      CHARACTER*6 :: CooLetter='xyztuv'
          IF (SymOpPos>NSym) THEN
        TruePos=SymOpPos-NSym
      ELSE
         TruePos=SymOpPos
      ENDIF
      !First check if the format is x1, x2, x3... or x, y, z
      IF (index(SymString,'x1')==0) THEN !x y z format
            DO i=1,SDim
          DO
            ii=index(SymString,CooLetter(i:i))
            IF (ii==0) EXIT
            write(rowpart,'(a1,i1)')'#',i
            SymString=SymString(:ii-1)//trim(rowpart)//symString(ii+1:)
          ENDDO
        ENDDO
            DO i=1,len_trim(SymString)
          IF (SymString(i:i)=='#') SymString(i:i)='x'
        ENDDO
      ENDIF
      !Check if comma or space delimited
      IF (index(Symstring,',')/=0) THEN !Comma delimited
        CALL RemoveSpaces(SymString)
        ii=len_trim(SymString)
        DO i=1,ii
          if (SymString(i:i)==',') SymString(i:i)=' '
        ENDDO
      ENDIF

          DO i=1,SDim
        LocTVec(i)=0.0E0
        DO ii=1,SDim
          LocRMat(i,ii)=0.0E0
        ENDDO
        row=FirstWord(SymString,.true.)
        lgth=len_trim(row)
        IF (lgth==0) THEN
          write(row,'(a,i3)') 'Error, not enough components of the symmetry operation number',TruePos
              call ErrorMsg(row)
          ErrCd=100
          return
        ENDIF
            if (lgth==1) goto 9100
        CALL CheckAllowedChars(row(:lgth),1,ErrCd)
        IF (ErrCd/=0) goto 9100
        curbeg=1;jex=0;jelom=0;jebod=0;indx=0;indlom=0
        DO ii=1,lgth
          IF (row(ii:ii)=='x') THEN
            jex=jex+1
            indx=ii
          ELSEIF (row(ii:ii)=='/') THEN
            jelom=jelom+1
            indlom=ii
          ELSEIF (row(ii:ii)=='.') THEN
            jebod=jebod+1
          ENDIF
          IF ((((row(ii:ii)=='+').or.(row(ii:ii)=='-')).and.(ii>1)).or.(ii==len_trim(row))) THEN
                IF (ii==lgth) then
          curend=ii
                else
          curend=ii-1
        endif
        IF ((jelom+jebod+jex>1).or.(indlom==curbeg).or.(indlom==curend) &
            .or.(indx==curend).or.(indx-curbeg>1)) goto 9000
                !Row(curbeg,curend) contains exactly one entry, it can be a +-xi,fraction or a decimal number
        IF (jex>0) then !Part containing entry of the rotation matrix
          read(row(indx+1:curend),'(i2)',IOSTAT=ErrCd) col
          if (errcd/=0) goto 9000
          if (LocRMat(i,col)>eps) goto 9300
          if (row(curbeg:curbeg)=='-') THEN
            LocRMat(i,col)=-1.0E0
          elseif ((row(curbeg:curbeg)=='+').or.(row(curbeg:curbeg)=='x')) THEN
            LocRMat(i,col)=1.0E0
          else
            goto 9000
          endif
        elseif(jebod>0) then !decimal number
          if (LocTVec(i)>eps) goto 9200
          rowpart=row(curbeg:curend)
          call onereal(rowpart,LocTVec(i),ErrCd)
          if (ErrCd/=0) goto 9000
        elseif(jelom>0) then !fraction
          if (LocTVec(i)>eps) goto 9200
          rowpart=row(curbeg:indlom-1)
          call oneint(rowpart,nom,ErrCd)
          if (ErrCd/=0) goto 9000
          rowpart=row(indlom+1:curend)
          call oneint(rowpart,denom,ErrCd)
          if (ErrCd/=0) goto 9000
          LocTVec(i)=float(nom)/float(denom)
                else !no x, no point, no slash -> integer translational part! Pretty bizzare, but possible...
          if (LocTVec(i)>eps) goto 9200
          rowpart=row(curbeg:curend)
          call oneint(rowpart,nom,ErrCd)
          if (ErrCd/=0) goto 9000
          LocTVec(i)=float(nom)
            endif
        jex=0;jelom=0;jebod=0;curbeg=curend+1;indlom=0;indx=0
          ENDIF
        ENDDO
      END DO
          InRMat=nint(LocRMat)
      return
9000      write(row,'(a,i3,a,i2)') 'Syntax error in the symmetry operation number',TruePos,', component number ',i
          call ErrorMsg(row)
      ErrCd=100
      return
9100      write(row,'(a,i3)') 'Syntax error in the symmetry operation number',TruePos
          call ErrorMsg(row)
          ErrCd=100
          return
9200      write(row,'(a,i3,a,i2)') 'Error, more than one translational part in the symmetry operation number',TruePos, &
                              ', component number ',i
          call ErrorMsg(row)
          ErrCd=100
          return
9300      write(row,'(a,2i2)') 'Error, more than one definition of the matrix element',i,col
          call ErrorMsg(row)
      write(row,'(a,i3,a)') 'in symmetry operation number',TruePos,'.'
          call ErrorMsg(row)
          ErrCd=100
          return
    END SUBROUTINE OneSym

! Handle the symmetry operations
!****************************************
        SUBROUTINE ReadCenters(kid,ErrCd)
!****************************************
      INTEGER :: i,ErrCd,kid,ErrCd2
      REAL :: dummyreal
      LOGICAL :: LocalCFlag
      CHARACTER*(linelength) :: Buffer2
          IF (NPass==1) THEN !Count centering vectors
            NCenter=0
            FlagZeroC=.false.
100         CALL NextLine(InBuffer,ErrCd)            
        IF (ErrCd<0) THEN !Preliminary end of file
              CALL ErrorMsg('Preliminary end of file. No keyword '//trim(Kwd(kid)%endname)//' found.')
          RETURN
            ELSEIF (ErrCd/=0) THEN
              WRITE(Buffer,'(a,i6)') 'Error reading centering vector number',NCenter
          CALL ErrorMsg(Buffer)
          RETURN
        ENDIF
            NCenter=NCenter+1
!           Test for the presence of a zero centering vector
        IF (.not.FlagZeroC) THEN
              Buffer=InBuffer
          LocalCFlag=.true.
          DO
            If (len_trim(Buffer)==0) EXIT !We do not know SDim yet, therefore just checking all numbers up to the end of the string
                Buffer2=FirstWord(Buffer,.true.)
        CALL OneReal(Buffer2,dummyreal,ErrCd2)
            IF (ErrCd2==0) THEN
                  IF (abs(dummyreal)>eps) LocalCFlag=.false.
            ELSE
          LocalCFlag=.false.
              EXIT !If Error, then either the endcenters keyword has been reached, or there is some other error that will be caught later
                ENDIF
          ENDDO
          IF (LocalCFlag) FlagZeroC=.true.
            ENDIF
        IF (FirstWord(InBuffer,.false.)/=Kwd(kid)%endname) GOTO 100
            NCenter=NCenter-1
        IF (.not.FlagZeroC) NCenter=NCenter+1
          ELSEIF (NPass==2) THEN !Read the centering vectors            
        ErrCd2=0
            IF (.not.FlagZeroC) THEN !We have to add the first vector ourselves
          DO i=1,SDim
            Center(i,1)=0.0E0
          ENDDO
          cnt2=2
        ELSE
          cnt2=1
            ENDIF
        DO cnt1=cnt2,NCenter
              CALL NextLine(InBuffer,ErrCd)            
          IF (ErrCd/=0) RETURN !Some error occured
          CALL OneCen(InBuffer,cnt1,ErrCd)
          IF (ErrCd/=0) ErrCd2=100
        END DO
        ErrCd=ErrCd2
          ENDIF
    END SUBROUTINE ReadCenters

! Decode one centering vector
!****************************************
        SUBROUTINE OneCen(CenString,Pos,ErrCd)
!****************************************
      INTEGER :: ErrCd,Pos,i,ii,jelom,lgth
      INTEGER :: indlom,nom,denom
      CHARACTER*(*) :: CenString
      CHARACTER*(linelength) :: Row
      CHARACTER*(15) :: rowpart

      DO i=1,SDim
        Center(i,Pos)=0.0E0
        row=FirstWord(CenString,.true.)
        lgth=len_trim(row)
        IF (lgth==0) THEN
          write(row,'(a,i3)') 'Error, not enough components of the centering vector number',Pos
              call ErrorMsg(row)
          ErrCd=100
          return
        ENDIF

        CALL CheckAllowedChars(row(:lgth),2,ErrCd)
        IF (ErrCd/=0) goto 9100
            jelom=0
        DO ii=1,lgth
          IF (row(ii:ii)=='/') THEN
            jelom=jelom+1
            indlom=ii
          ENDIF
            ENDDO
        if (jelom>1) then
          goto 9000
            elseif(jelom>0) then !fraction
          rowpart=row(1:indlom-1)
          call oneint(rowpart,nom,ErrCd)
          if (ErrCd/=0) goto 9000
          rowpart=row(indlom+1:lgth)
          call oneint(rowpart,denom,ErrCd)
          if (ErrCd/=0) goto 9000
          center(i,pos)=float(nom)/float(denom)
        else !It can be only a plain number
          rowpart=row(1:lgth)
          call onereal(rowpart,Center(i,pos),ErrCd)
          if (ErrCd/=0) goto 9000
        endif
      END DO
      return
9000      write(row,'(a,i3,a,i2)') 'Syntax error in the centering vector number',Pos,', component number ',i
          call ErrorMsg(row)
      ErrCd=100
      return
9100      write(row,'(a,i3)') 'Syntax error in the centering vector number',Pos
          call ErrorMsg(row)
          ErrCd=100
          return
        END SUBROUTINE OneCen

! Handle the reflections
!****************************************
        SUBROUTINE ReadRefl(kid,ErrCd)
!****************************************
      INTEGER :: ErrCd,kid,TmpTotalF
      LOGICAL :: succ
      IF (NPass==1) THEN !Count the reflections
        TotalF=0
        IF (hklxplor) THEN
          CALL ReadHKLfromXplor(TotalF) !If called with TotalF=0, then only number of reflections is counted and returned in TotalF
        ELSE
          IF (ExternRefl) THEN
            IF (len_trim(ReflStartLine)>0) THEN  !Rewind to the beginning of the reflection records - skip possible header
              DO
                READ(erfu,'(a)',IOSTAT=ErrCd) InBuffer
                IF (ErrCd<0) THEN !End of file
                  CALL ErrorMsg('Error - the starting line "'//trim(ReflStartLine)// &
                          '" not found in the reflection file!')
                  RETURN
                ENDIF
                IF (InBuffer(1:len_trim(ReflStartLine))==ReflStartLine) EXIT
              ENDDO
            ENDIF
          ENDIF
100       CONTINUE
          IF (ExternRefl) THEN
            CALL NextLineExtern(InBuffer,ErrCd,erfu)
            IF (ErrCd<0) THEN !End of file - in case of external file this is no problem
              ErrCd=0
              RETURN
            ELSEIF ((len_trim(ReflEndLine)>0).and.(InBuffer(:len_trim(ReflEndLine))==ReflEndLine)) THEN !Final line detected, reading is finished
              RETURN
            ENDIF
          ELSE
            CALL NextLine(InBuffer,ErrCd)
            IF (ErrCd<0) THEN !Preliminary end of file
              CALL ErrorMsg('Preliminary end of file. No keyword '//trim(Kwd(kid)%endname)//' found.')
              RETURN !Some error occured
            ELSEIF(FirstWord(InBuffer,.false.)==Kwd(kid)%endname) THEN
              RETURN
            ELSEIF (ErrCd/=0) THEN
              WRITE(Buffer,'(a,i6)') 'Error reading reflection number',TotalF
              CALL ErrorMsg(Buffer)
              call ErrorMsg('Lately reading: '//trim(InBuffer))
              RETURN
            ENDIF
          ENDIF
          IF (ErrCd/=0) THEN
            WRITE(Buffer,'(a,i6)') 'Error reading reflection number',TotalF
            CALL ErrorMsg(Buffer)
            call ErrorMsg('Lately reading: '//trim(InBuffer))
            RETURN
          ENDIF
          TotalF=TotalF+1
          GOTO 100
        ENDIF !(hklxplor)

      ELSEIF (NPass==2) THEN !Read the reflections
        IF (terminal) THEN
          IF (externrefl) THEN
            write(*,'(a,a,a)') '  Reading reflections from file ',trim(ExternReflFile),'...'
          ELSE
            write(*,'(a)') '  Reading reflections...'
          ENDIF
        ENDIF
        IF (hklxplor) THEN
          CALL ReadhklfromXplor(TotalF)
          StartF=1;EndF=TotalF
          TotalAllocatedF=TotalF
        ELSE
          IF (ExternRefl) THEN
            IF (len_trim(ReflStartLine)>0) THEN !Rewind to the beginning of the reflection records - skip possible header
              DO
                READ(erfu,'(a)') InBuffer
                IF (InBuffer(1:len_trim(ReflStartLine))==ReflStartLine) EXIT
              ENDDO
            ENDIF
          ENDIF
          TmpTotalF=0
          DO cnt1=1,TotalF
            IF (ExternRefl) THEN
              CALL NextLineExtern(InBuffer,ErrCd,erfu)
            ELSE
              CALL NextLine(InBuffer,ErrCd)
            ENDIF
            IF (len_trim(DataItemWidths)>0) THEN !introduce spaces between individual data items to allow a free-format reading of fixed-format lines
              Buffer=InBuffer
              InBUffer=''
              DO i=1,SDim
                InBuffer=trim(InBuffer)//' '//Buffer(DataItemPos(i):DataItemPos(i+1)-1)
              ENDDO
              DO i=SDim+1,NDataItems+SDim
                InBuffer=trim(InBuffer)//' '//Buffer(DataItemPos(i):DataItemPos(i+1)-1)
              ENDDO
            ENDIF
            IF (ErrCd/=0) RETURN !Some error occured
            CALL OneRefl(InBuffer,TmpTotalF+1,ErrCd, succ)
            if (succ) TmpTotalF=TmpTotalF+1 !Count successful reflections
            IF (ErrCd/=0) RETURN
          END DO
          TotalAllocatedF=TotalF
          TotalF=TmpTotalF
          StartF=1;EndF=TotalF
        ENDIF
        if (terminal) write(*,'(a)') '  ...successfully finished.'
      ENDIF
    END SUBROUTINE ReadRefl

! Read one reflection
!************************************************
        SUBROUTINE OneRefl(ReflString,Pos,ErrCd,succ)
!************************************************
      INTEGER :: ErrCd,Pos,i
      CHARACTER*(*) :: ReflString
      CHARACTER*(linelength) :: Component,ErrSubStr
      CHARACTER*(linelength) :: ReflStringC
      LOGICAL :: unknownphase,succ
      REAL :: sthl, multsc
          ReflStringC=ReflString !Store the original to allow printing it in the error messages. The copy is progresively destroyed by FirstWord.
          ! Read the indices
          CALL CheckAllowedChars(trim(ReflStringC),4,ErrCd)
      IF (ErrCd/=0) THEN
            ErrSubstr=''
        GOTO 9000
      ENDIF
      DO i=1,SDim
        Component=FirstWord(ReflStringC,.true.)
        CALL OneInt(Component,Refl(Pos)%hklFG(i),ErrCd)
        IF (ErrCd/=0) THEN
              ErrSubstr='indices of the'
          GOTO 9000
        ENDIF
      ENDDO
      !Check, if the reflection falls in the resolution limits
      succ=.true.
      sthl=0.5*sqrt(multsc(float(Refl(Pos)%hklFG),float(Refl(Pos)%hklFG),SDim,.true.))
      maxinputsthl=max(maxinputsthl,sthl)
      IF ((sthl<reflsthlmin).or.(sthl>reflsthlmax)) THEN
        succ=.false.
        return
      ENDIF
          Refl(pos)%code=code_default
          !Now read the items according to the array dataitem
      unknownphase=.false.
      DO i=1,NDataItems
        Component=FirstWord(ReflStringC,.true.)
            IF ((trim(component)=='?').and.(DataItem(i)%flag==c_phase)) THEN
          unknownphase=.true.
        ELSE IF (len_trim(Component)==0) THEN !The behavior depends on what should be read
          IF ((DataItem(i)%flag==c_int).or.(DataItem(i)%flag==c_amp).or.(DataItem(i)%flag==c_a)) THEN
                ErrCd=100
        ErrSubstr='the component '//trim(DataItem(i)%string)//' of the'
            GOTO 9000
              ELSEIF((DataItem(i)%flag==c_group).or.(DataItem(i)%flag==c_b)) THEN
            DataItem(i)%value=0.0E0
          ELSEIF(DataItem(i)%flag==c_phase) THEN
            unknownphase=.true.
          ENDIF
        ELSE
          CALL OneReal(Component,DataItem(i)%value,ErrCode)
              IF (ErrCd/=0) THEN
        ErrSubstr='the component '//trim(DataItem(i)%string)//' of the'
            GOTO 9000
              ENDIF
        ENDIF
          ENDDO
          IF (FormatFlag(c_a)%count>0) THEN
        HaveAB=.true.
      ELSE
        HaveAB=.false.
      ENDIF
          !And now fill the Refl array according to the information from DataItem
          IF (FormatFlag(c_group)%count>0) THEN !Assign the group number
        Refl(pos)%group=nint(DataItem(formatflag(c_group)%pos)%value)
        NGroups=max(NGroups,Refl(pos)%group)
      ELSE
        Refl(pos)%group=0
      ENDIF
          IF (FormatFlag(c_fwhm)%count>0) THEN !Assign the fwhm - only temporarily needed -> store in AMEM component
        Refl(pos)%AMEM=DataItem(formatflag(c_fwhm)%pos)%value
          ENDIF
      IF (HaveAB) THEN
        Refl(pos)%A=DataItem(formatflag(c_a)%pos)%value
!!!!!!! Important: The convention in FFTW is OPOSITE to that in crystallography.
!!!!!!! Therefore internally the B component or Refl must be inverted on input and inverted again in poutput
!!!!!!! Unpleasant, but necessary
        Refl(pos)%B=-1*DataItem(formatflag(c_b)%pos)%value
        Refl(pos)%F=SQRT(Refl(pos)%A**2+Refl(pos)%B**2)
            if (Refl(pos)%B==-nophaseflag) Refl(pos)%B=nophaseflag+eps
      ELSEIF (FormatFlag(c_phase)%count>0) THEN !We have phases, we can caluclate A and B
        HaveAB=.true.
            IF(FormatFlag(c_int)%count>0) THEN
              IF (DataItem(formatflag(c_int)%pos)%value<eps) THEN
            Refl(pos)%F=0.0E0
          ELSE
            Refl(pos)%F=SQRT(DataItem(formatflag(c_int)%pos)%value)
          ENDIF
        ELSE
              Refl(pos)%F=DataItem(formatflag(c_amp)%pos)%value
        ENDIF
        IF (unknownphase) THEN
              Refl(pos)%A=Refl(pos)%F
          Refl(pos)%B=0.
        ELSE
              Refl(pos)%A=Refl(pos)%F*cos(2.*pi*DataItem(formatflag(c_phase)%pos)%value)
              Refl(pos)%B=-Refl(pos)%F*sin(2.*pi*DataItem(formatflag(c_phase)%pos)%value)
          Refl(pos)%code=code_knownph
        ENDIF
      ELSEIF (FormatFlag(c_int)%count>0) THEN !Only intensity is given
            IF (DataItem(formatflag(c_int)%pos)%value<eps) THEN
          Refl(pos)%F=0.0E0
        ELSE
          Refl(pos)%F=SQRT(DataItem(formatflag(c_int)%pos)%value)
            ENDIF
            Refl(pos)%A=Refl(pos)%F
            Refl(pos)%B=0.0E0
      ELSEIF (FormatFlag(c_amp)%count>0) THEN !Only amplitude is given
            Refl(pos)%F=DataItem(formatflag(c_amp)%pos)%value
            Refl(pos)%A=Refl(pos)%F
            Refl(pos)%B=0.0E0
      ELSEIF (FormatFlag(c_ampdiff)%count>0) THEN !Only amplitude is given
            Refl(pos)%F=abs(DataItem(formatflag(c_ampdiff)%pos)%value)
            Refl(pos)%A=Refl(pos)%F
            Refl(pos)%B=0.0E0
      ELSE
        Call ErrorMsg('Zle maticko, zle je, nemam z ceho pocitat reflexe.')
        CALL StopProgram('Internal error of the program. Please contact the author.')
      ENDIF
          RETURN
9000      WRITE(Buffer,'(a,i6)') 'Error reading '//trim(ErrSubStr)//' reflection number',Pos
          CALL ErrorMsg(Buffer)
          call ErrorMsg('Lately reading: '//ReflString)
          ErrCd=100
          RETURN
    END SUBROUTINE OneRefl

! Handle the symmetry operations
!****************************************
        SUBROUTINE ReadHistogram(kid,ErrCd)
!****************************************
      INTEGER :: ErrCd,kid
      LOGICAL :: ExplicitClasses
          IF (NPass==1) THEN !Count histogram classes
            hmclasses=-1
100         CALL NextLine(InBuffer,ErrCd)         
        IF (ErrCd<0) THEN !Preliminary end of file
              CALL ErrorMsg('Preliminary end of file. No keyword '//trim(Kwd(kid)%endname)//' found.')
          RETURN
            ELSEIF (ErrCd/=0) THEN
              WRITE(Buffer,'(a,i6)') 'Error reading histogramm class number',HMClasses
          CALL ErrorMsg(Buffer)
          RETURN
        ENDIF
            HMClasses=HMClasses+1
        IF (FirstWord(InBuffer,.false.)/=Kwd(kid)%endname) GOTO 100
            HMClasses=HMClasses-1
          ELSEIF (NPass==2) THEN !Read the classes
        DO cnt1=0,HMClasses
              CALL NextLine(InBuffer,ErrCd)
          IF (cnt1==0) THEN !Check if class borders are given or not
            Buffer=InBuffer
        DummyString=FirstWord(Buffer,.true.) !Cut the first number
        IF (len_trim(Buffer)>0) THEN !Classes are given explicitly
          ExplicitClasses=.true.
        ELSE
          ExplicitClasses=.false.
        ENDIF
          ENDIF
              IF (ErrCd/=0) RETURN !Some error occured
          IF (ExplicitClasses) THEN
                CALL Onereal(InBuffer,HMGramm(cnt1)%pos,ErrCd)
                CALL Onereal(InBuffer,HMGramm(cnt1)%orig,ErrCd)
                IF (ErrCd/=0) RETURN !Some error occured
              ELSE
                CALL Onereal(InBuffer,HMGramm(cnt1)%orig,ErrCd)
                IF (ErrCd/=0) RETURN !Some error occured
                HMgramm(cnt1)%pos=float(cnt1)/float(HMclasses)
          ENDIF
        END DO
      ENDIF
    END SUBROUTINE ReadHistogram

        SUBROUTINE CheckAllowedChars(String,key,ErrCd)
    integer, parameter :: CharNr=17
    integer :: ErrCd,loop,key
        character*(*) :: String
    character*1, dimension(1:CharNr) :: alchr,alcenchr,alsymchr,justnumchr,reflchr,digits,binary
    logical :: ok
    data alsymchr  /' ',' ','x','1','2','3','4','5','6','7','8','9','0','.','/','+','-'/
    data alcenchr  /' ',' ','1','2','3','4','5','6','7','8','9','0','.','/','+','-',' '/
        data justnumchr/' ',' ','1','2','3','4','5','6','7','8','9','0','.',' ','+','-',' '/
        data reflchr   /' ',' ','1','2','3','4','5','6','7','8','9','0','.',' ','+','-','?'/
        data digits    /'1','2','3','4','5','6','7','8','9','0','0','0','0','0','0','0','0'/
        data binary    /'0','1','1','1','1','1','1','1','1','1','1','1','1','1','1','1','1'/
        alsymchr(2)=char(9) !char(9) cannot be in the data statement
        alcenchr(2)=char(9)
        justnumchr(2)=char(9)
    if (key==1) then
      alchr=alsymchr
    elseif (key==2) then
      alchr=alcenchr
    elseif (key==3) then
      alchr=justnumchr
    elseif (key==4) then
      alchr=reflchr
    elseif (key==5) then
      alchr=digits
    elseif (key==6) then
      alchr=binary
    endif
    DO loop=1,len_trim(string)
      ok=.false.
      DO loop2=1,CharNr
        IF (string(loop:loop)==alchr(loop2)) THEN
          ok=.true.
          exit
        END IF
      END DO
          ErrCd=0
      if (.not.ok) then
        ErrCd=100
        RETURN
      endif
        END DO
        END SUBROUTINE CheckAllowedChars

        SUBROUTINE GuessType(FileName,FileType)
    IMPLICIT NONE
        CHARACTER*(*) :: FileName, FileType
    INTEGER :: i
    ! Search for the period
    DO i=len_trim(FileName),1,-1
      IF (FileName(i:i)=='.') THEN
        SELECT CASE (Filename(i+1:len_trim(Filename)))
          CASE ('m80')
                FileType='m80'
          CASE ('m81')
                FileType='jana'
          CASE ('xplor')
                FileType='xplor'
          CASE ('cif')
                FileType='cif'
          CASE ('ins')
                FileType='shelxsym'
          CASE ('ccp4')
                FileType='ccp4'
              CASE DEFAULT
            FileType='unknown'
        END SELECT
        EXIT
      ENDIF
    ENDDO
        END SUBROUTINE GuessType

        SUBROUTINE ReadRhoFiles
    IMPLICIT NONE
        OutRhoNr=0
    DO
      IF (len_trim(InBufferOrig)/=0) THEN
        OutRhoNr=OutRhoNr+1
        IF (OutRhoNr>MaxFileNames) THEN
              write(Buffer,'(i3)') MaxFileNames
          Buffer='Warning, maximum number of allowed density files exceeded. Only the first'// &
                      trim(Buffer)//' will be considered.'
          CALL ErrorMsg(Buffer)
          OutRhoNr=MaxFileNames
          EXIT
        ENDIF
          OutRhoFile(OutRhoNr)=FirstWord(InBufferOrig,.true.)
      ELSE
        EXIT
      ENDIF
    ENDDO
        END SUBROUTINE ReadRhoFiles



      END SUBROUTINE INPUT

!****************************************
        SUBROUTINE GetAutoVoxels(InNVox, GridResolution, Shannon)
!****************************************
        USE SF_Module
    USE SF_Interfaces
    IMPLICIT NONE
    INTEGER, DIMENSION(1:SDim) :: InNVox
    REAL :: GridResolution, Shannon
        INTEGER, DIMENSION(1:SDim) :: symfac, mxsymfac,maxhkl
    LOGICAL, DIMENSION(1:SDim,1:SDim) :: equivdim
    INTEGER :: os,i, cv,n2, n3, n5, n2max,n3max, n5max, ii,mindif
    REAL :: sh,dummy
    real, dimension(SDim) :: dhkl,dhkl2

        
        !Find the factors that the voxel spacing must contain due to the symmetry:
        mxsymfac=0
        do cv=1,NCenter
          do os=1,NSym
            do i=1,SDim
              sh=SmOp(os)%t(i)+Center(i,cv)
              symfac(i)=0
800           symfac(i)=symfac(i)+1
              dummy=sh*float(symfac(i))
              if (abs(dummy-nint(dummy)).ge.bigeps) goto 800
              if (symfac(i).gt.mxsymfac(i)) mxsymfac(i)=symfac(i)
        enddo
      enddo
        enddo

        !Find the maximum indices in the expanded reflection list
        maxhkl=0
    do cv=StartF,EndF
       dhkl=float(Refl(cv)%hklFG)
       DO os=1,NSym
         dhkl2=multmat(dhkl,SmOp(os)%R)
         DO i=1,SDim
           if (maxhkl(i)<abs(nint(dhkl2(i)))) maxhkl(i)=abs(nint(dhkl2(i)))
         ENDDO
       ENDDO
    enddo
    !If minimum resolution of the grid is required, then make sure maxhkl is not smaller that the minimum resolution
        IF (GridResolution>eps) THEN
          DO i=1,RDim
            maxhkl(i)=max(maxhkl(i),int(Cell(i)/(2*GridResolution)-1))
      ENDDO
    ENDIF
        !Make sure the grid is fine enough to accomodate high-angle missing reflections
        DO i=1,RDim
          maxhkl(i)=max(maxhkl(i),int(2.*FillMissingSthl/RCell(i))+1)
        ENDDO

    !Find directions that should have equal spacing, because they are mixed by symmetry
    equivdim=.false.
    do os=1,NSym
      do i=1,SDim
        do ii=1,SDim
              if (SmOp(os)%R(i,ii)/=0) THEN
            equivdim(i,ii)=.true.
            equivdim(ii,i)=.true.
          endif
        enddo
      enddo
    enddo

    !Make sure the equivalent directions have the same parameters
        do i=1,SDim
          do ii=1,SDim
        if (i==ii) cycle
        if (equivdim(i,ii)) then
          cv=max(maxhkl(i),maxhkl(ii))
          maxhkl(i)=cv; maxhkl(ii)=cv
          cv=lcm((/mxsymfac(i),mxsymfac(ii)/))
          mxsymfac(i)=cv; mxsymfac(ii)=cv
        endif
      enddo
    enddo
    symfac=maxhkl*shannon+2
    do i=1,SDim
          n2max=int(log(float(symfac(i)))/log(2.))+1
          n3max=int(log(float(symfac(i)))/log(3.))+1
          n5max=int(log(float(symfac(i)))/log(5.))+1
      mindif=1000000
      do n2=0,n2max
        do n3=0,n3max
          do n5=0,n5max
                ii=2**n2*3**n3*5**n5*mxsymfac(i)
                if (((ii>symfac(i)).and.((ii-symfac(i)).lt.mindif)).and.(ii>=8)) then
                  mindif=ii-symfac(i)
                  InNVox(i)=ii
                endif
          enddo
        enddo
      enddo
    enddo
        END SUBROUTINE GetAutoVoxels

      SUBROUTINE ShrinkHMSymbol(InHMSym,OutHMSym)
        character*(*) :: InHMSym, OutHMSym
    integer :: i, ii
    OutHMSym=''
    IF (len_trim(InHMSym)==0) RETURN
    i=0
    ii=0
    DO
      i=i+1
      IF ((InHMSym(i:i)/=' ').and.(InHMSym(i:i)/=':')) THEN
        ii=ii+1
        OutHMSym(ii:ii)=InHMSym(i:i)
      ENDIF
      IF ((InHMSym(i:i)==':').or.(i==len_trim(InHMSym))) EXIT
    ENDDO
      END SUBROUTINE ShrinkHMSymbol
