!********************************************
      MODULE sf_module
!********************************************
      USE symmetry_tools
! #### Constants      
      integer, parameter :: linelength=256 ! Maximum length of a line in an output file
      integer, parameter :: filelength=256 ! Maximum length of a file name
      integer, parameter :: extlength=10         ! Maximum length of a file extension (part after the last dot)
      integer, parameter :: MaxSymOps=150 !Maximum  number of symmetry operators
      integer, parameter :: MaxFileNames=10
      integer, parameter :: lgu=30 !Unit number of the log file
      integer, parameter :: rhu=90 !Unit number of the output rho file
      integer, parameter :: detphu=91 !Unit number of the file with detailed phases
      integer, parameter :: cwfu=92 !Unit number of the file with codewords for external phase assignment
      integer, parameter :: hmwaitu=93 !Unit number of the file for exporting and importing of the amplitudes repartitioned by an external program
      integer, parameter :: tmpu=99 !Unit number of the temporary file
      character*(extlength), parameter :: lgext='.sflog' !Extension of the log file
      character*(extlength), parameter :: cmext='.sfcom' !Extension of the command file
      character*(extlength), parameter :: tmpext='.sftmp' !Extension of the temporary file
      character*(132), parameter :: versionstring='02/13/14  8:48'
      integer, parameter :: refrhotar=1, envrhotar=2, modelrhotar=3, rhotar=4
      integer, parameter :: SearchSymKey=1, TestSymKey=2, range=7
      integer, parameter :: RelDelta=1, StatDelta=2, KedDelta=3
      integer, parameter :: NormNone=0, NormWilson=1, NormLocal=2, NormAtoms=3
      integer, parameter :: sumall=1,sumgood=2
      integer, parameter :: code_default    =0, & !reflection codes
                            code_weak       =1, &
                            code_knownph    =2, &
                            code_missing    =3, &
                            code_highres    =4, &
                            code_skip       =5
      
      integer, parameter :: mm_zero       =0, & !codes for handling missing reflections
                            mm_float      =1, &
                            mm_bound      =2, &
                            mm_boundsum   =3, &
                            mm_undefined  =-1
      
      integer, parameter :: FOMRValue     =0, & !codes for choosing the FOM for saving best densities
                            FOMReference  =1, &
                            FOMSymAgr     =2, &
                            FOMpeakiness  =3
      
      integer, parameter :: NoAveraging   =0, & !codes for handling the symmetry
                            AveragePhases =1, &
                            AverageSF     =2, &
                            DoNotTouch    =3
      
      integer, parameter :: noviewer      =0, & !codes for choosing the external program for viewing the progress
                            chimera       =1, &
                            vesta         =2
      
      integer, parameter :: UnknownSystem =0, & !codes for different systems
                            Unix          =1, &
                            Windows       =2
      
      integer, parameter :: r2cft         =1, & !codes for forward and backward transform
                            c2rft        =-1
      
      integer, parameter :: res_sthl      =0, & !codes for resolution types
                            res_d         =1
      
      integer, parameter :: pattersonmodel=0, & !codes for internal starting models
                            supminfunction=1
      
      integer, parameter :: sgcode_explicit=0, & !codes for space-group input
                            sgcode_hall=1, &
                            sgcode_hm=2, &
                            sgcode_number=3, &
                            sgcode_fp=4
      
      integer, parameter :: wrcode_relative=0, & !codes for handling weak reflections
                            wrcode_absolute=1

      real, parameter :: pi=3.14159265358979323846264335327941971
      real, parameter :: torad=pi/180.0E0
      double precision, parameter :: dpi=3.14159265358979323846264335327941971
      double precision, parameter :: dtorad=dpi/180.0D0
      real, parameter :: nophaseflag=0.123E-6
      real, parameter :: LDEThreshold=1.

! #### General variables holding crystallographic information      
      integer, save :: SDim=3 ! Dimension of the density/structure factors
      integer, save :: RDim=3 ! Dimension of the real space
      real, allocatable, dimension(:), save :: Cell, RCell !Direct and reciprocal cell
      integer, save :: NCellPar !Number of cell parameters, depends on RDim
      real, SAVE :: cApVx=1.0D0     ! 'cubic' Angstroem per voxel
      real, allocatable, dimension(:,:), save :: G, Gi ! Direct and reciprocal metric tensors
      real, allocatable, dimension(:,:), save :: GS, GSi ! Direct and reciprocal metric tensors in superspace
      real, allocatable, dimension(:,:), save :: QVec   ! q-vectors
      integer, save :: NSym, NSymPrim !Number of symmetry operators, number of the "primitive" ones, without combination with centering vectors
      integer, save :: sgcode=sgcode_explicit !Encodes the way symmetry is input
      character*132 :: insgsymbol=''
      integer, save :: NCenter=0
      integer, save :: NGen, NGCenter !Number of generators of the primitive SG, number of centering generators
      integer, save :: NSymOrig, NSymPrimOrig, NCenterOrig !Backup variables for recovering the symmetry if needed
      logical  :: Centro
      INTEGER, SAVE :: NTestSym=0     !number of symmetry operators to be tested
      INTEGER, SAVE :: NTwin=0        !number of twin matrices

      type(SymOp_type), allocatable, dimension(:) :: SmOp
      type(SymOp_type), allocatable, dimension(:) :: SmOpOrig
!      real, allocatable,dimension(:,:,:), save :: RMat,RMatOrig
!      real, allocatable,dimension(:,:), save :: TVec,TVecOrig
      real, allocatable,dimension(:,:), save :: Center,CenterOrig
      type(SymOp_type), allocatable, dimension(:) :: TwinOp
!      real, allocatable,dimension(:,:,:), save :: TwinMat
      logical, allocatable,dimension(:), save :: SurviveTwin !Flags the symmetry operations that surwive upon twinningMat
      
      character*48, save :: SGSymbol='---'
      integer, allocatable,dimension(:), save :: generators,gencenters
      
      type SGPropsType
        logical :: centro, enant
	integer :: nsym, nsymprim, ncen
      end type SGPropsType
      type (SGPropsType) :: SGProps
! Reflections:
      character*30 :: DataType='amplitude'
      logical, save :: haveAB                   ! Defines if the input file contained two-componented reflection input
      logical, save :: haveFWHM
      INTEGER, SAVE :: TotalF=0, TotalAllocatedF=0
      INTEGER, SAVE :: StartF=0, EndF=-1        ! Start and end element of the reflection array. For compatibility with BayMEM, where not only ""F": reflections are possible
      INTEGER, SAVE :: MTotalF=0                ! Total expanded reflections
      INTEGER, SAVE :: MStartF=0, MEndF=-1      ! Start and end element of expanded F-constraint in array. For compatibility with BayMEM, where not only ""F": reflections are possible
      TYPE RefTyp
        INTEGER :: SFLoc                        ! Corresponding location in SF-array of the reflection
        INTEGER, POINTER, DIMENSION(:) :: hklFG    ! hkl... of reflection
        INTEGER :: group, code
        REAL :: A, B, F             ! Structure factor of reflection, A=real, B=imaginary part and Sigma = e.s.d.
        REAL :: Amem, Bmem, Fmem    ! Calculated structure factor
!        REAL, DIMENSION(2) :: phold !Storing the old phases to calculate the phase difference
      END TYPE RefTyp  
      TYPE (RefTyp), DIMENSION (:), ALLOCATABLE, SAVE :: Refl   ! The reflections read from the input file
      TYPE (RefTyp), DIMENSION (:), ALLOCATABLE, SAVE :: MEMRefl! The reflection expanded to full sphere, sorted and used in the MEM    
      INTEGER, ALLOCATABLE, DIMENSION(:), SAVE :: MEMRefl2Refl !Array hodling the correspondance between the EMMRefl and Refl lists. Needed ofnly for histogramm matching.
      LOGICAL, ALLOCATABLE, DIMENSION(:), SAVE :: KnownPhase !Array hodling the information if the phase of the reflection is known or not.
      REAL :: reflsthlmax=10000., reflsthlmin=0.0, refldmin=0., refldmax=100000., MaxInputSthl=0.0
      REAL :: ResUnits=res_sthl
      CHARACTER(LEN=30) :: ReflectionList='general'

! #### Variables holding the settings of the algorithm
      TYPE gdmpar_type
        real :: beta1, beta2, gamma1D, gamma1R, gamma2D, gamma2R
	logical :: needrhoold,needrho1
      END TYPE gdmpar_type
      type(gdmpar_type) :: GdmPar, InGdmPar
      character*(linelength), save    :: InPerform='cf'
      LOGICAL, SAVE :: GenerateRandom=.false.
      INTEGER, SAVE :: InMaxCycles=-10000          ! Input of maximal cycles to compute. Negative sign is a flag for MaxCycles not being redefined by user.
      integer, allocatable, dimension(:), save :: NVox, NVoxIter, NVoxFine, NVoxref, NVoxModel, NVoxEnv !Number of voxels along each dimension
      real, save :: IterGridResolution=0.0 !User has the possibility to set the grid size
      real, save :: FineGridResolution=0.2 !User has the possibility to set the grid size
      real, save :: IterShannonFactor=2
      real, save :: FineShannonFactor=2
      logical, save :: autovoxels=.true. !Flag for calculating voxel dvision automatically
      logical, save :: autofinevoxels=.true. !Flag for calculating finevoxel dvision automatically
      integer, allocatable, dimension(:), save :: NVoxSF !Nuumber of voxels in the SF array
      integer, allocatable, dimension(:), save :: DimStep !Steps used in a fast calculaton of the position of the reflection/pixel in the main array
      integer, allocatable, dimension(:), save :: DimStepSF !Steps used in a fast calculaton of the position of the reflection/pixel in the SF array
      integer, save :: NVoxTot, NVoxTotSF    !The total number of voxels in rho and SF
      logical, save :: AutoRS=.true.,InAutoDelta=.true. !Flags for automatic definition of RandomSeed and Delta
      integer, save :: DeltaType=2   !Type of the input delta
      logical, save :: neutrons=.false. !Indicates that negative density is expected -> flip between -delta and +delta
      integer, save :: StartNeutrons !Gives the cycle number from which the iteration will be neutron-like
      real, save    :: FlipDelta,CFInDelta  !The basic parameter of the whole stuff      
      integer, save :: RandomSeed !Initialization of the random number generator
      integer, save :: RandStructureSeed=0 !Initialization of the random number generator for generating random structure
      LOGICAL, SAVE :: FastFFT=.false. !If true, FFTW_MEASURE is used insteda of FFTW_ESTIMATE - faster, but not reproducible
      character*10, save :: UsePhases='no' !Should the phases from the input file be used as starting phases?
      integer, save :: setphasecycles=0 !Number of cycles during which the phases should be set to the known values. Only with usephass='number'
      real, save    :: Biso=0.0E0 !Isotropic Debye-Waller factor
      logical, save :: Bisofix=.false. !Isotropic Debye-Waller factor
      real, save    :: BisoHM=0.0E0 !Isotropic Debye-Waller factor used for calculation of histogram
      integer, save :: NormalizeRefl=NormNone !Use Es instead of Fs?
      integer, save :: NResShells=0 !Number of resolution shells in local normalization
      logical, save :: karleexpand=.true.
      real, save    :: FWHMlim=0.2 !the fraction of FWHM under which reflections are considered to overlap totally
      real, save    :: lambda=0.710 !wavelength of the radiation. Used only for calculation of overlaps in powder!
      character*5, save :: FIPS='no' !Perform FIPS on the overlapping powder data?
      real, dimension(4), save    :: FIPSparams
      real, save    :: WeakRatio=0.0E0 !Proportion of the reflections to be treated as weak
      integer, save :: WeakReflMode=wrcode_relative !Mode of handling weakratio - absolute or relative
      real, save    :: fdeltafw=100000.0E0 !Range radius of the annulus around F(h) for the F(obs)+delta(F) method 
      logical, save :: fdeltaf=.false. !If true, the F(obs)+delta(F) will be used
      logical, save :: searchsym=.true.
      integer, save :: averagesym=AveragePhases
      character*3, save  :: derivesym='no' 
      real, save    :: LimitAgrFac=25.0  !Highest agreement factor still considered a match
      logical, save :: fixf0=.false. !keep F(0) constant?
      character*3   :: symroutine='new'
      real, SAVE :: MaxRvalue=-100.0D0         !If R/value falls below this, calculation finishes
      real, ALLOCATABLE, DIMENSION(:) :: WeakRefl, SaveWeakRefl !The threshold on the weak reflections
      integer, save :: SetWeakCycle=1 !From which cycle should the weak reflections be treated as weak
      integer, allocatable, dimension(:) :: zoneaxis
      real, save :: randmodelfactor=0.0
      real, save :: randfourierfactor=0.0
      REAL, save :: connstdevfac=1.4
      REAL, save :: FillMissingSthl=0.4
      REAL, save :: MissingBndFactor=2.5
      integer, save :: MissingMode=mm_undefined
      integer, save :: MissingNr
      REAL, save :: ResForSym=5.7 !Resolution in sin(th)/lambda for calculating the reflections for perform symmetry
      REAL, save :: sumWB !sum of wilson-plot based estimates of the missing reflections
      INTEGER, save :: CoincCode=1 !Determines the type of the coincidence coefficient used
      INTEGER, save :: LargeGraph=200
      REAL, save :: intformatchlow=0.0, intformatchhigh=1.0E30 !Limits for intensities taken into account for matchong with reference file

      logical, save :: polish=.true.
      integer, save :: polishcycles=5
      character*(9), save :: inconvergencemode='normal' !String influencing the convergence criterion
      logical, save :: mem=.false.
      logical, save :: twochannel=.false.
      real, save :: memlimit=10.
      integer, save :: memcycles=100

! Variables related to the repetition of the calculation
      character*(9), save :: repeatmode='never' !String determining, if and when the calculation should be repeated
      integer, save :: repeatattempts=0 !How many repetitions should I make
      integer, save :: repeats !Actually performed repetitions
      integer, save :: NCodeWords
      character*(linelength) :: CodeWord

      TYPE RhoPropsType
        integer :: RunNumber
        real :: Rvalue=0., peaks=0.,agrfac=100.0,refmatch=100.0
	character*60 :: DerSGSym='', OrigShift=''
      END Type RhoPropsType
      TYPE(RhoPropsType), SAVE :: CurRhoProps
      character*(4), save :: rhoprefix='best'
      integer, save :: Nbestrho=1 !Number of the best densities to be saved      
      integer, save :: Nsavedrho=0 !Number of already saved densities. Always smaller or equal to NBestRho
      type(RhoPropsType), allocatable, dimension(:),save :: BestRhoProps !Holds the properties of the best saved densities
      integer, save :: SaveRhoFOM=FOMRValue !Which FOM to use to decide if the current density is "good
      real, save  :: TwoDenMatch !A variable to export the density match from MatchReference
      real, allocatable, dimension(:), save  :: TwoDenShift !A variable to export the density match and shift from MatchReference
      logical, save :: Converged !Flag to indicate, if the current calculation converged or not
      integer, save :: WantSumRho=0, NSummed=0 !flag if the sumrho should be produced, number of densities summed up 
      integer, save :: NConverged=0 !Number of converged runs
      integer, save :: TotalCycles=0
      logical, save :: ClusAn=.false. !Should I perform cluster analysis of multiple densities?
      integer :: ViewProgress=noviewer, ViewProgressPeriod=5
      integer, allocatable, dimension(:) :: PresModeLimits, PresModeTimes !Slowing down the program for presentation purposes
      integer, save :: PresModeNr=0

!     Issues concerning histogram and symmetry matching
      integer, save :: Inhmstartc=0, Inhmstepc=0
      integer, save :: Insmstartc=0, Insmstepc=0
      LOGICAL, save :: symmatch=.false.
      LOGICAL, save :: hmrepart=.true.
      LOGICAL, save :: smrepart=.false.
      LOGICAL, save :: modelrepart=.false.
      LOGICAL, save :: hmwait=.false.
      integer, save :: hmclasses=0 !The number of classes the histogram should be divided to. Deduced from the input
      type hmgramm_type
        real :: pos, cur, orig
      end type hmgramm_type
      type(hmgramm_type), ALLOCATABLE, DIMENSION(:),save :: hmgramm
      integer :: NGroups=0 !Number of groups of overlapping reflections
      type(hmgramm_type),allocatable, dimension(:),save :: GroupInt !Array holding the Group intensity for each group of reflections      
      integer, save :: envstartc=200, envstepc=100
      real, save :: envzerolevel=0.0, envcutmin=0.75
      
   
! #### Variables related to the storage of the density, structure factors and FFT
      real, allocatable, dimension(:), target, save :: rho !This is what it says it is
      real, allocatable, dimension(:), target, save :: rhoold
      real, allocatable, dimension(:), target, save :: rho1
      INTEGER, ALLOCATABLE, DIMENSION(:) :: skelet
      real, allocatable, dimension(:), target, save :: refrho !Array holding the reference density
      real, allocatable, dimension(:), target, save :: modelrho !Array holding possible starting density
      real, allocatable, dimension(:), target, save :: envrho !Array holding the envelope density
      real, allocatable, dimension(:), target, save :: sumrho !Array holding the sum of all successful densities during repeated runs
      real, allocatable, dimension(:), target, save :: rho50  !Array holding the 50-cycles old density
      real, allocatable, dimension(:), target, save :: hklrho !A temporary array used to store hkl amplitudes in xplor format
      real, allocatable, dimension(:), save :: srho !Part of the density to be used in the spline interpolation
      real, allocatable, dimension(:), save :: SF  !Array holding the structure factors for the transformation by FFT
      TYPE FFTInstr
        INTEGER(KIND=8) :: plan
	INTEGER :: dir
	REAL :: Norm
      END TYPE FFTInstr
      TYPE(FFTInstr) :: rho2sf, sf2rho !Variables holding plans for FFT using FFTW library

! #### Variables specific for the density-analysis part

! #### Filing, i.e. names of the input and output files and related stuff
       integer, save :: SystemFlag !Flag holding the type of the operating system the program is running on
       character*(linelength), save    :: Title='title not defined' !The title of the job - just a flag written to the output file
       character*(filelength), save    :: FileBase=' ' !Base of the files superflip is using
       character*(filelength), save    :: InFile !The name of the input file
       character*(filelength), save    :: LogFile, OutFile, TmpFile!Names of the file with logging information, output information and temporary storage
       CHARACTER*(FileLength), SAVE    :: ComFile='' ! Name of the communication file
       character*(filelength), save    :: DetPhFile!File with detailed information about the phases in each cycle
       character*(filelength), save    :: CodeWordFile!File with codewrds encoding the phase information
       character*(filelength), dimension(1:MaxFileNames), save :: OutRhoFile='', CurRhoFile=''
       character*(filelength), save :: RefRhoFile='', ModelRhoFile='', EnvRhoFile='', ModelStrFile=''
       character*(linelength), dimension(1:MaxFileNames),save  :: OutRhoType='none'
       character*(linelength), save :: RefRhoType='none', ModelRhoType='none', EnvRhoType='none'
       integer, save :: OutRhoNr=0
       character*(filelength), save    :: ExpRefFile=' ' !File that will be used to write the complete list of reflections
       character*(filelength), save    :: hmwaitoutfile='sf_hmexport.dat' !The name of the file with reflections for external repartitioning
       character*(filelength), save    :: hmwaitinfile='sf_hmimport.dat' !The name of the file with reflections after external repartitioning
       logical, save :: ExpandedLog=.false. ! If yes, extended information about the progress is written in the LogFile
       logical, save :: Terminal=.true. ! If yes, extended information about the progress is written in the LogFile
       logical, save :: KeepWindow=.false. ! If yes, the program waits for any key at the end of the execution
       logical, save :: CalcCoverage=.true. ! If yes, coverage calculation is performed on the input data
       logical, save :: RewriteOutput=.true. ! If yes, the output density will rewrite possible existing older densities

! #### Other variables, counters etc.
       integer :: cnt1, cnt2, cnt3, cnt4 ! counters, loop variables
       character*(linelength) :: Buffer
       integer :: AllocError
       integer :: RecLFac=1 !Factor in the number of byted per record in the direct access to the files
       REAL, SAVE :: RF,skewrho,sigmarho,refskewrho, epsilon
       integer :: maxcycles
       integer :: detailedphases=0 !If positive, detailed phase information is written every nth cycle
       logical :: analpeaks=.false.

!      User communication with program
      TYPE COM_TYPE
        INTEGER :: int
	REAL :: rl
	CHARACTER*132 :: chr
      END TYPE COM_TYPE
      TYPE(COM_TYPE),SAVE :: Com
      LOGICAL, SAVE :: UserBreak=.false.

      end module sf_module

!********************************************
      MODULE Input_module
!********************************************
        USE sf_module
	
        integer, parameter :: MaxGroups=10 !Maximum number of groups
        integer, parameter :: MaxKwds=100 !Maximum number of keywords
        integer, parameter :: MaxKwdLen=30 !Maximal length of a keyword and groupname
        integer, parameter :: infunit=10 !Unit number the infile will be connected to
        integer, parameter :: erfu=50 !Unit number an external reflection file
        character*2, parameter :: comments='#!'
        INTEGER :: PassNr=2
        CHARACTER*(linelength) :: InBuffer,InBufferOrig,ExternReflFile

        TYPE GrpType
          character*(MaxKwdLen) ::name
	  logical :: status !If true, the keywords of this group ale compulsory, otherwise they are optional
	END TYPE GrpType
        TYPE (GrpType), DIMENSION(MaxGroups) :: KwdGroup
        Integer :: KwdGroupNr=0
	CHARACTER*(MaxKwdLen) :: GrpName !Temporary variable storing the name to be passed to the subroutine KwdGroup
        LOGICAL :: GrpStatus 

        Type KwdType
          character*(MaxKwdLen) ::name, endname
          integer :: group
	  integer :: occ
	END TYPE KwdType
        TYPE (KwdType), DIMENSION(1:MaxKwds) :: Kwd
        Integer :: KwdNr=0
        CHARACTER*(MaxKwdLen) :: KN,KEndN,Grp ! Temporary variables to be passed to the subroutine NewKwd
        CHARACTER*(MaxKwdLen) :: CurKwd
        INTEGER :: CurKwdId !Identification number of the current keyword during rading

!       Datatype handling:
        INTEGER, PARAMETER :: c_int=1,c_amp=2,c_a=3,c_b=4,c_phase=5,c_group=6,c_dummy=7,c_fwhm=8, c_ampdiff=9
        INTEGER, PARAMETER :: NFormatFlags=9, MaxDataItems=10
	INTEGER :: NDataItems=2
	TYPE DataItemType
	  character*20 :: string
	  integer :: flag
	  real :: value
	END TYPE DataItemType
        TYPE(DataItemType), DIMENSION(1:MaxDataItems) :: DataItem
	
	TYPE FormatFlagType
	  character*20 :: string
	  integer :: count,pos
	END TYPE FormatFlagType
        TYPE(FormatFlagType), DIMENSION(1:NFormatFlags) :: FormatFlag
	DATA FormatFlag(1)%String,FormatFlag(2)%String,FormatFlag(3)%String,FormatFlag(4)%String,FormatFlag(5)%String, &
	     FormatFlag(6)%String,FormatFlag(7)%String,FormatFlag(8)%String,FormatFlag(9)%String &
	     /'intensity','amplitude','a','b','phase','group','dummy','fwhm','amplitude_difference'/
!       Initialize the values to the default setting "a b"
	DATA FormatFlag(1)%count,FormatFlag(2)%count,FormatFlag(3)%count,FormatFlag(4)%count,FormatFlag(5)%count, &
	     FormatFlag(6)%count,FormatFlag(7)%count,FormatFlag(8)%count /0,0,1,1,0,0,0,0/
	DATA FormatFlag(1)%pos,FormatFlag(2)%pos,FormatFlag(3)%pos,FormatFlag(4)%pos,FormatFlag(5)%pos, &
	     FormatFlag(6)%pos,FormatFlag(7)%pos,FormatFlag(8)%pos /0,0,1,2,0,0,0,0/
        DATA DataItem(1)%string,DataItem(2)%string /'a','b'/
        DATA DataItem(1)%flag,DataItem(2)%flag /c_a,c_b/
	
        CHARACTER*(linelength) :: ReflStartLine='',ReflEndLine='',DataItemWidths=''
	LOGICAL :: externrefl, hklxplor=.false.
	INTEGER, ALLOCATABLE, DIMENSION(:) :: DataItemPos


        INTEGER :: ErrNr=0 !Number of errors occured during reading

        CONTAINS

	SUBROUTINE NewKwdGroup
           KwdGroupNr=KwdGroupNr+1
	   KwdGroup(KwdGroupNr)%name=GrpName
	   KwdGroup(KwdGroupNr)%status=GrpStatus
	END SUBROUTINE NewKwdGroup

	SUBROUTINE FlagKwdGroup(name,stat)
	   character*(*) :: name
	   logical :: stat
	   integer :: i
           DO i=1,KwdGroupNr
	     IF (KwdGroup(i)%name==name) KwdGroup(i)%status=stat
	   END DO
	END SUBROUTINE FlagKwdGroup

	SUBROUTINE NewKwd(kid)
           integer :: i,kid
	   KwdNr=KwdNr+1
	   kid=KwdNr !keyword identification, the keyword number is massed tot he variable, which can be then referred instead of the number itself
	   Kwd(KwdNr)%name=KN !Name of the (first part of the) keyword
	   Kwd(KwdNr)%endname=KEndN !Optional closing part of the keyword
	   Kwd(KwdNr)%occ=0 !Number of occurences in the iput file
           DO i=1,KwdGroupNr
	     if (Grp==KwdGroup(i)%name) Kwd(KwdNr)%group=i
	   END DO
	END SUBROUTINE NewKwd

        SUBROUTINE ErrorCheck
          IMPLICIT NONE
          integer :: i
          character*132 :: ErrMessage
	  
	  DO i=1,KwdNr

	    IF (Kwd(i)%occ>1) THEN
	      ErrNr=ErrNr+1
	      write(ErrMessage,'(3a,i2,a)') 'Error, keyword ',trim(Kwd(i)%name),' occured ', &
                                             Kwd(i)%occ,' times in the input file.'
	      CALL ErrorMsg(ErrMessage)
	      CALL ErrorMsg('')
            END IF

            IF ((KwdGroup(Kwd(i)%Group)%status.eqv..true.).and.(Kwd(i)%occ==0)) THEN
	      ErrNr=ErrNr+1
	      WRITE(*,'(3a,i2,a)') 'Error, the compulsory keyword ',trim(Kwd(i)%name),' is missing in the input file.'
	    ENDIF

            Kwd(i)%occ=0 !Make it zero to be "fresh" for next pass
	  END DO 

	  IF (ErrNr>0) THEN
	    WRITE(*,'(a,i3,a)') 'There were',ErrNr,' errors in the input file.'
	    WRITE(*,'(a)') 'Please check the input file and try again.'
	    STOP
	  END IF

	END SUBROUTINE ErrorCheck

        SUBROUTINE OneInt(Text,oInt,ErrCode)
          CHARACTER*(*) :: Text
          INTEGER :: oInt,ErrCode,dm
	  CHARACTER*10 :: form='(I..)'
	  CHARACTER*(linelength) :: word
          word=FirstWord(text,.true.)
	  dm=len_trim(word)
	  IF (dm==0) THEN
	    ErrCode=100
	  ELSE
	    write(form(3:4),'(i2)') max(15,dm)
            READ(word,form,IOSTAT=ErrCode) oInt
	  ENDIF
        END SUBROUTINE OneInt

        SUBROUTINE ArrayInt(Text,oArrInt,iCnt,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode,iCnt,i,dm
	  INTEGER, DIMENSION(iCnt) :: oArrInt
	  CHARACTER*(linelength) :: word
	  CHARACTER*10 :: form='(I..)'
          DO i=1,iCnt
            word=FirstWord(text,.true.)
            dm=len_trim(word)
	    IF (dm==0) THEN
	      CALL ErrorMsg('Error - not enough numbers in the list of integers.')
              write(word,'(''Expected'',i2,'', found '',i2)') iCnt,i-1
	      CALL ErrorMsg(word)
	      ErrCode=100
	      EXIT
	    ELSE
	      write(form(3:4),'(i2)') max(15,dm)
              READ(word,form,IOSTAT=ErrCode) oArrInt(i)
            ENDIF
	  END DO
        END SUBROUTINE ArrayInt

        SUBROUTINE OneReal(text,oReal,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode,dm
	  CHARACTER*10 :: form='(f...0)'
	  CHARACTER*(linelength) :: word
	  REAL :: oReal
          word=FirstWord(text,.true.)
          dm=len_trim(word)
	  IF (dm==0) THEN
	    ErrCode=100
	  ELSE
	    write(form(3:4),'(i2)') max(25,dm)
            READ(word,form,IOSTAT=ErrCode) oReal
	  ENDIF
        END SUBROUTINE OneReal

        SUBROUTINE ArrayReal(text,oArrReal,iCnt,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode,iCnt,i,dm
	  REAL, DIMENSION(iCnt) :: oArrReal
	  CHARACTER*(linelength) :: word
	  CHARACTER*10 :: form='(f...0)'
          DO i=1,iCnt
            word=FirstWord(text,.true.)
            dm=len_trim(word)
	    IF (dm==0) THEN
	      CALL ErrorMsg('Error - not enough numbers in the list of reals.')
              write(word,'(''Expected'',i2,'', found '',i2)') iCnt,i-1
	      CALL ErrorMsg(word)
	      ErrCode=100
	      EXIT
	    ELSE
	      write(form(3:4),'(i2)') max(25,dm)
              READ(word,form,IOSTAT=ErrCode) oArrReal(i)
            ENDIF
	  END DO
        END SUBROUTINE ArrayReal

        SUBROUTINE YesNo(text,oLogic,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode
	  LOGICAL :: oLogic
	  ErrCode=0
	  IF (FirstWord(text,.false.)=='yes') THEN
	    oLogic=.true.
	  ELSEIF (FirstWord(text,.false.)=='no') THEN
	    oLogic=.false.
          ELSE
	    ErrCode=100
	  END IF
	END SUBROUTINE YesNo

        FUNCTION FirstWord(ioText,cutword)
          IMPLICIT NONE
          CHARACTER*(*) :: ioText
          CHARACTER*132 :: FirstWord
          INTEGER :: wbeg,i,wend
          LOGICAL CutWord
          CHARACTER*1 :: Tab,SearchedChar1,SearchedChar2
          CHARACTER*1, PARAMETER :: Space=' ',DQuota='"'
	  
          FirstWord=' '
          Tab=CHAR(9)
          SearchedChar1=Space
          SearchedChar2=Tab
          wbeg=0
          DO wbeg=1,LEN(ioText)!Skip any leading spaces
            IF ((ioText(wbeg:wbeg)/=Space).AND.(ioText(wbeg:wbeg)/=Tab)) EXIT
          END DO
          if(wbeg>LEN(ioText)) THEN
            ioText=' '
            RETURN
          ENDIF
          IF (ioText(wbeg:wbeg)==DQuota) THEN
            SearchedChar1=DQuota
            SearchedChar2=DQuota
            wbeg=wbeg+1
          ENDIF
          DO i=wbeg,LEN(ioText)
            IF ((ioText(i:i)==SearchedChar1).OR.(ioText(i:i)==SearchedChar2)) THEN
              wend=i-1
              FirstWord=ioText(wbeg:wend)
              IF (CutWord) ioText=ioText(wend+2:LEN(ioText))
              EXIT
            END IF
          END DO
          DO wbeg=1,LEN(ioText)!Skip any leading spaces of the new ioText
	        IF ((ioText(wbeg:wbeg)/=Space).AND.(ioText(wbeg:wbeg)/=Tab)) EXIT
          END DO
          if(wbeg<LEN(ioText)) THEN
	        ioText(1:)=ioText(wbeg:)
          ELSE
	        ioText=' '
	      ENDIF
        END FUNCTION FirstWord

        FUNCTION ToLow(iText)
          IMPLICIT NONE
          CHARACTER*(*) :: iText
          CHARACTER*(len(iText)) :: ToLow
          INTEGER :: i
	  ToLow=''
          DO i=1,len_trim(iText)
	    SELECT CASE (iText(i:i))
	      CASE('A')
	        ToLow(i:i)='a'
	      CASE('B')
	        ToLow(i:i)='b'
	      CASE('C')
	        ToLow(i:i)='c'
	      CASE('D')
	        ToLow(i:i)='d'
	      CASE('E')
	        ToLow(i:i)='e'
	      CASE('F')
	        ToLow(i:i)='f'
	      CASE('G')
	        ToLow(i:i)='g'
	      CASE('H')
	        ToLow(i:i)='h'
	      CASE('I')
	        ToLow(i:i)='i'
	      CASE('J')
	        ToLow(i:i)='j'
	      CASE('K')
	        ToLow(i:i)='k'
	      CASE('L')
	        ToLow(i:i)='l'
	      CASE('M')
	        ToLow(i:i)='m'
	      CASE('N')
	        ToLow(i:i)='n'
	      CASE('O')
	        ToLow(i:i)='o'
	      CASE('P')
	        ToLow(i:i)='p'
	      CASE('Q')
	        ToLow(i:i)='q'
	      CASE('R')
	        ToLow(i:i)='r'
	      CASE('S')
	        ToLow(i:i)='s'
	      CASE('T')
	        ToLow(i:i)='t'
	      CASE('U')
	        ToLow(i:i)='u'
	      CASE('V')
	        ToLow(i:i)='v'
	      CASE('W')
	        ToLow(i:i)='w'
	      CASE('X')
	        ToLow(i:i)='x'
	      CASE('Y')
	        ToLow(i:i)='y'
	      CASE('Z')
	        ToLow(i:i)='z'
	      CASE DEFAULT
	        ToLow(i:i)=iText(i:i)
	    END SELECT
	  ENDDO
	END FUNCTION ToLow

        SUBROUTINE DelComments(ioText)
          INTEGER :: i,ii
	  logical :: CommChar
	  CHARACTER*(*) :: ioText
	  DO i=1,len_trim(ioText)
	    CommChar=.false.
	    DO ii=1,len_trim(Comments)
	      IF (ioText(i:i)==Comments(ii:ii)) CommChar=.true.
	    END DO
            IF (CommChar) EXIT
	  END DO
          IF (i>0) THEN
	    ioText=ioText(1:i-1)
	  ELSE
	    ioText=''
	  END IF
	END SUBROUTINE DelComments

        SUBROUTINE NextLine(oText,ErrCd)
	  CHARACTER*(*) :: oText
	  INTEGER :: ErrCd,i
100       READ(infunit,'(a)', IOSTAT=ErrCd) oText
	  DO i=1,len_trim(oText)
	    if (oText(i:i)==char(9)) oText(i:i)=' '
          ENDDO
	  CALL DelComments(oText)
	  IF ((len_trim(oText)==0).and.(ErrCd==0)) GOTO 100 !This line is empty or contains only comments
	END SUBROUTINE NextLine

        SUBROUTINE NextLineExtern(oText,ErrCd,fileunit)
	  CHARACTER*(*) :: oText
	  INTEGER :: ErrCd,fileunit
100       READ(fileunit,'(a)', IOSTAT=ErrCd) oText
	  IF ((len_trim(oText)==0).and.(ErrCd==0)) GOTO 100 !This line is empty
	END SUBROUTINE NextLineExtern

!        SUBROUTINE ErrorMsg(text)
!          CHARACTER*(*) :: text
!          write(*,'(3a)') trim(text)
!	END SUBROUTINE ErrorMsg

      END MODULE Input_module

!********************************************
      MODULE sf_composition
!********************************************
      INTEGER, PARAMETER :: TotElNr=98                          !Number of elements in the ElSmb array
      INTEGER, PARAMETER :: AppMemb=5
      CHARACTER*2, DIMENSION(1:TotElNr), PARAMETER :: ElSmb= &       ! Elemental symbols sorted by atomic number
        (/'H ','He','Li','Be','B ','C ','N ','O ','F ','Ne','Na','Mg','Al','Si','P ','S ','Cl','Ar','K ','Ca','Sc','Ti','V ','Cr',&
          'Mn','Fe','Co','Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr','Rb','Sr','Y ','Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd',&
          'In','Sn','Sb','Te','I ','Xe','Cs','Ba','La','Ce','Pr','Nd','Pm','Sm','Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb','Lu','Hf',&
          'Ta','W ','Re','Os','Ir','Pt','Au','Hg','Tl','Pb','Bi','Po','At','Rn','Fr','Ra','Ac','Th','Pa','U ','Np','Pu','Am','Cm',&
	  'Bk','Cf'/)
      INTEGER, DIMENSION(:,:),ALLOCATABLE, SAVE :: Compos ! Array holding the atomic numbers of atoms expected to be in the structure
      INTEGER, SAVE :: ElNr !Number of different elements
            REAL, SAVE :: ScaleWP, BisoWP, SumfWP, LnkWP,M2bisoWP !Wilson-plot derived quantities
      INTEGER :: sfci, sfcj
      REAL, DIMENSION(1:2*AppMemb,1:TotElNr) :: FF !& !Analytical approximation to atomic form factor
                                                   !FF(1,i)-FF(AppMEmb,i) contains the a coefficients of the ith element,  
						   !The b coefficient is in FF(AppMemb+1,2*AppMemb)
      DATA  ((FF(sfci,sfcj),sfci=1,2*AppMemb),sfcj=1,30) &
          /  0.49300,   0.32291,   0.14019,   0.04081,   0.00304,  10.51090,  26.12570,   3.14236,  57.79970,   0.00000,  &
             0.87340,   0.63090,   0.31120,   0.17800,   0.00640,   9.10370,   3.35680,  22.92760,   0.98210,   0.00000,  &
             1.12820,   0.75080,   0.61750,   0.46530,   0.03770,   3.95460,   1.05240,  85.39050, 168.26100,   0.00000,  &
             1.59190,   1.12780,   0.53910,   0.70290,   0.03850,  43.64270,   1.86230, 103.48300,   0.54200,   0.00000,  &
             2.05450,   1.33260,   1.09790,   0.70680,  -0.19320,  23.21850,   1.02100,  60.34980,   0.14030,   0.00000,  &
             2.31000,   1.02000,   1.58860,   0.86500,   0.21560,  20.84390,  10.20750,   0.56870,  51.65120,   0.00000,  &
            12.21260,   3.13220,   2.01250,   1.16630, -11.52900,   0.00570,   9.89330,  28.99750,   0.58260,   0.00000,  &
             3.04850,   2.28680,   1.54630,   0.86700,   0.25080,  13.27710,   5.70110,   0.32390,  32.90890,   0.00000,  &
             3.53920,   2.64120,   1.51700,   1.02430,   0.27760,  10.28250,   4.29440,   0.26150,  26.14760,   0.00000,  &
             3.95530,   3.11250,   1.45460,   1.12510,   0.35150,   8.40420,   3.42620,   0.23060,  21.71840,   0.00000,  &
             4.76260,   3.17360,   1.26740,   1.11280,   0.67600,   3.28500,   8.84220,   0.31360, 129.42400,   0.00000,  &
             5.42040,   2.17350,   1.22690,   2.30730,   0.85840,   2.82750,  79.26110,   0.38080,   7.19370,   0.00000,  &
             6.42020,   1.90020,   1.59360,   1.96460,   1.11510,   3.03870,   0.74260,  31.54720,  85.08860,   0.00000,  &
             6.29150,   3.03530,   1.98910,   1.54100,   1.14070,   2.43860,  32.33370,   0.67850,  81.69370,   0.00000,  &
             6.43450,   4.17910,   1.78000,   1.49080,   1.11490,   1.90670,  27.15700,   0.52600,  68.16450,   0.00000,  &
             6.90530,   5.20340,   1.43790,   1.58630,   0.86690,   1.46790,  22.21510,   0.25360,  56.17200,   0.00000,  &
            11.46040,   7.19640,   6.25560,   1.64550,  -9.55740,   0.01040,   1.16620,  18.51940,  47.77840,   0.00000,  &
             7.48450,   6.77230,   0.65390,   1.64420,   1.44450,   0.90720,  14.84070,  43.89830,  33.39290,   0.00000,  &
             8.21860,   7.43980,   1.05190,   0.86590,   1.42280,  12.79490,   0.77480, 213.18700,  41.68410,   0.00000,  &
             8.62660,   7.38730,   1.58990,   1.02110,   1.37510,  10.44210,   0.65990,  85.74840, 178.43700,   0.00000,  &
             9.18900,   7.36790,   1.64090,   1.46800,   1.33290,   9.02130,   0.57290, 136.10800,  51.35310,   0.00000,  &
             9.75950,   7.35580,   1.69910,   1.90210,   1.28070,   7.85080,   0.50000,  35.63380, 116.10500,   0.00000,  &
            10.29710,   7.35110,   2.07030,   2.05710,   1.21990,   6.86570,   0.43850,  26.89380, 102.47800,   0.00000,  &
            10.64060,   7.35370,   3.32400,   1.49220,   1.18320,   6.10380,   0.39200,  20.26260,  98.73990,   0.00000,  &
            11.28190,   7.35730,   3.01930,   2.24410,   1.08960,   5.34090,   0.34320,  17.86740,  83.75430,   0.00000,  &
            11.76950,   7.35730,   3.52220,   2.30450,   1.03690,   4.76110,   0.30720,  15.35350,  76.88050,   0.00000,  &
            12.28410,   7.34090,   4.00340,   2.34880,   1.01180,   4.27910,   0.27840,  13.53590,  71.16920,   0.00000,  &
            12.83760,   7.29200,   4.44380,   2.38000,   1.03410,   3.87850,   0.25650,  12.17630,  66.34210,   0.00000,  &
            13.33800,   7.16760,   5.61580,   1.67350,   1.19100,   3.58280,   0.24700,  11.39660,  64.81260,   0.00000,  &
            14.07430,   7.03180,   5.16520,   2.41000,   1.30410,   3.26550,   0.23330,  10.31630,  58.70970,   0.00000 /

      DATA  ((FF(sfci,sfcj),sfci=1,2*AppMemb),sfcj=31,60) &
          / 15.23540,   6.70060,   4.35910,   2.96230,   1.71890,   3.06690,   0.24120,  10.78050,  61.41350,   0.00000,  &
            16.08160,   6.37470,   3.70680,   3.68300,   2.13130,   2.85090,   0.25160,  11.44680,  54.76250,   0.00000,  &
            16.67230,   6.07010,   3.43130,   4.27790,   2.53100,   2.63450,   0.26470,  12.94790,  47.79720,   0.00000,  &
            17.00060,   5.81960,   3.97310,   4.35430,   2.84090,   2.40980,   0.27260,  15.23720,  43.81630,   0.00000,  &
            17.17890,   5.23580,   5.63770,   3.98510,   2.95570,   2.17230,  16.57960,   0.26090,  41.43280,   0.00000,  &
            17.35550,   6.72860,   5.54930,   3.53750,   2.82500,   1.93840,  16.56230,   0.22610,  39.39720,   0.00000,  &
            17.17840,   9.64350,   5.13990,   1.52920,   3.48730,   1.78880,  17.31510,   0.27480, 164.93400,   0.00000,  &
            17.56630,   9.81840,   5.42200,   2.66940,   2.50640,   1.55640,  14.09880,   0.16640, 132.37600,   0.00000,  &
            17.77600,  10.29460,   5.72629,   3.26588,   1.91213,   1.40290,  12.80060,   0.12560, 104.35400,   0.00000,  &
            17.87650,  10.94800,   5.41732,   3.65721,   2.06929,   1.27618,  11.91600,   0.11762,  87.66270,   0.00000,  &
            17.61420,  12.01440,   4.04183,   3.53346,   3.75591,   1.18865,  11.76600,   0.20479,  69.79570,   0.00000,  &
             3.70250,  17.23560,  12.88760,   3.74290,   4.38750,   0.27720,   1.09580,  11.00400,  61.65840,   0.00000,  &
            19.13010,  11.09480,   4.64901,   2.71263,   5.40428,   0.86413,   8.14487,  21.57070,  86.84720,   0.00000,  &
            19.26740,  12.91820,   4.86337,   1.56756,   5.37874,   0.80852,   8.43467,  24.79970,  94.29280,   0.00000,  &
            19.29570,  14.35010,   4.73425,   1.28918,   5.32800,   0.75154,   8.21758,  25.87490,  98.60620,   0.00000,  &
            19.33190,  15.50170,   5.29537,   0.60584,   5.26593,   0.69866,   7.98929,  25.20520,  76.89860,   0.00000,  &
            19.28080,  16.68850,   4.80450,   1.04630,   5.17900,   0.64460,   7.47260,  24.66050,  99.81560,   0.00000,  &
            19.22140,  17.64440,   4.46100,   1.60290,   5.06940,   0.59460,   6.90890,  24.70080,  87.48250,   0.00000,  &
            19.16240,  18.55960,   4.29480,   2.03960,   4.93910,   0.54760,   6.37760,  25.84990,  92.80290,   0.00000,  &
            19.18890,  19.10050,   4.45850,   2.46630,   4.78210,   5.83030,   0.50310,  26.89090,  83.95710,   0.00000,  &
            19.64180,  19.04550,   5.03710,   2.68270,   4.59090,   5.30340,   0.46070,  27.90740,  75.28250,   0.00000,  &
            19.96440,  19.01380,   6.14487,   2.52390,   4.35200,   4.81742,   0.42089,  28.52840,  70.84030,   0.00000,  &
            20.14720,  18.99490,   7.51380,   2.27350,   4.07120,   4.34700,   0.38140,  27.76600,  66.87760,   0.00000,  &
            20.29330,  19.02980,   8.97670,   1.99000,   3.71180,   3.92820,   0.34400,  26.46590,  64.26580,   0.00000,  &
            20.38920,  19.10620,  10.66200,   1.49530,   3.33520,   3.56900,   0.31070,  24.38790, 213.90400,   0.00000,  &
            20.33610,  19.29700,  10.88800,   2.69590,   2.77310,   3.21600,   0.27560,  20.20730, 167.20200,   0.00000,  &
            20.57800,  19.59900,  11.37270,   3.28719,   2.14678,   2.94817,   0.24448,  18.77260, 133.12400,   0.00000,  &
            21.16710,  19.76950,  11.85130,   3.33049,   1.86264,   2.81219,   0.22684,  17.60830, 127.11300,   0.00000,  &
            22.04400,  19.66970,  12.38560,   2.82428,   2.05830,   2.77393,   0.22209,  16.76690, 143.64400,   0.00000,  &
            22.68450,  19.68470,  12.77400,   2.85137,   1.98486,   2.66248,   0.21063,  15.88500, 137.90300,   0.00000 /

      DATA  ((FF(sfci,sfcj),sfci=1,2*AppMemb),sfcj=61,90) &
         /  23.34050,  19.60950,  13.12350,   2.87516,   2.02876,   2.56270,   0.20209,  15.10090, 132.72100,   0.00000,  &
            24.00420,  19.42580,  13.43960,   2.89604,   2.20963,   2.47274,   0.19645,  14.39960, 128.00700,   0.00000,  &
            24.62740,  19.08860,  13.76030,   2.92270,   2.57450,   2.38790,   0.19420,  13.75460, 123.17400,   0.00000,  &
            25.07090,  19.07980,  13.85180,   3.54545,   2.41960,   2.25341,   0.18195,  12.93310, 101.39800,   0.00000,  &
            25.89760,  18.21850,  14.31670,   2.95354,   3.58324,   2.24256,   0.19614,  12.66480, 115.36200,   0.00000,  &
            26.50700,  17.63830,  14.55960,   2.96577,   4.29728,   2.18020,   0.20217,  12.18990, 111.87400,   0.00000,  &
            26.90490,  17.29400,  14.55830,   3.63837,   4.56796,   2.07051,   0.19794,  11.44070,  92.65660,   0.00000,  &
            27.65630,  16.42850,  14.97790,   2.98233,   5.92046,   2.07356,   0.22355,  11.36040, 105.70300,   0.00000,  &
            28.18190,  15.88510,  15.15420,   2.98706,   6.75621,   2.02859,   0.23885,  10.99750, 102.96100,   0.00000,  &
            28.66410,  15.43450,  15.30870,   2.98963,   7.56672,   1.98890,   0.25712,  10.66470, 100.41700,   0.00000,  &
            28.94760,  15.22080,  15.10000,   3.71601,   7.97628,   1.90182,   9.98519,   0.26103,  84.32980,   0.00000,  &
            29.14400,  15.17260,  14.75860,   4.30013,   8.58154,   1.83262,   9.59990,   0.27512,  72.02900,   0.00000,  &
            29.20240,  15.22930,  14.51350,   4.76492,   9.24354,   1.77333,   9.37046,   0.29598,  63.36440,   0.00000,  &
            29.08180,  15.43000,  14.43270,   5.11982,   9.88750,   1.72029,   9.22590,   0.32170,  57.05600,   0.00000,  &
            28.76210,  15.71890,  14.55640,   5.44174,  10.47200,   1.67191,   9.09227,   0.35050,  52.08610,   0.00000,  &
            28.18940,  16.15500,  14.93050,   5.67589,  11.00050,   1.62903,   8.97948,   0.38266,  48.16470,   0.00000,  &
            27.30490,  16.72960,  15.61150,   5.83377,  11.47220,   1.59279,   8.86553,   0.41792,  45.00110,   0.00000,  &
            27.00590,  17.76390,  15.71310,   5.78370,  11.68830,   1.51293,   8.81174,   0.42459,  38.61030,   0.00000,  &
            16.88190,  18.59130,  25.55820,   5.86000,  12.06580,   0.46110,   8.62160,   1.48260,  36.39560,   0.00000,  &
            20.68090,  19.04170,  21.65750,   5.96760,  12.60890,   0.54500,   8.44840,   1.57290,  38.32460,   0.00000,  &
            27.54460,  19.15840,  15.53800,   5.52593,  13.17460,   0.65515,   8.70751,   1.96347,  45.81490,   0.00000,  &
            31.06170,  13.06370,  18.44200,   5.96960,  13.41180,   0.69020,   2.35760,   8.61800,  47.25790,   0.00000,  &
            33.36890,  12.95100,  16.58770,   6.46920,  13.57820,   0.70400,   2.92380,   8.79370,  48.00930,   0.00000,  &
            34.67260,  15.47330,  13.11380,   7.02588,  13.67700,   0.70100,   3.55078,   9.55642,  47.00450,   0.00000,  &
            35.31630,  19.02110,   9.49887,   7.42518,  13.71080,   0.68587,   3.97458,  11.38240,  45.47150,   0.00000,  &
            35.56310,  21.28160,   8.00370,   7.44330,  13.69050,   0.66310,   4.06910,  14.04220,  44.24730,   0.00000,  &
            35.92990,  23.05470,  12.14390,   2.11253,  13.72470,   0.64645,   4.17619,  23.10520, 150.64500,   0.00000,  &
            35.76300,  22.90640,  12.47390,   3.21097,  13.62110,   0.61634,   3.87135,  19.98870, 142.32500,   0.00000,  &
            35.65970,  23.10320,  12.59770,   4.08655,  13.52660,   0.58909,   3.65155,  18.59900, 117.02000,   0.00000,  &
            35.56450,  23.42190,  12.74730,   4.80703,  13.43140,   0.56336,   3.46204,  17.83090,  99.17220,   0.00000 /

      DATA  ((FF(sfci,sfcj),sfci=1,2*AppMemb),sfcj=91,TotElNr) &
         /  35.88470,  23.29480,  14.18910,   4.17287,  13.42870,   0.54775,   3.41519,  16.92350, 105.25100,   0.00000,  &
            36.02280,  23.41280,  14.94910,   4.18800,  13.39660,   0.52930,   3.32530,  16.09270, 100.61300,   0.00000,  &
            36.18740,  23.59640,  15.64020,   4.18550,  13.35730,   0.51193,   3.25396,  15.36220,  97.49080,   0.00000,  &
            36.52540,  23.80830,  16.77070,   3.47947,  13.38120,   0.49938,   3.26371,  14.94550, 105.98000,   0.00000,  &
            36.67060,  24.09920,  17.34150,   3.49331,  13.35920,   0.48363,   3.20647,  14.31360, 102.27300,   0.00000,  &
            36.64880,  24.40960,  17.39900,   4.21665,  13.28870,   0.46515,   3.08997,  13.43460,  88.48340,   0.00000,  &
            36.78810,  24.77360,  17.89190,   4.23284,  13.27540,   0.45102,   3.04619,  12.89460,  86.00300,   0.00000,  &
            36.91850,  25.19950,  18.33170,   4.24391,  13.26740,   0.43753,   3.00775,  12.40440,  83.78810,   0.00000  /

      !Variables for the structure definition
        TYPE Atom_type
	  integer :: nr,Z !number of atomic type, atomic number
	  character*8 :: name
	  real, dimension(1:3) :: xyz !Coordinates
	  real, dimension(1:6) :: U !Displacement
	  real :: occ !occupancy
	END TYPE Atom_type
	TYPE (Atom_type), ALLOCATABLE, DIMENSION(:) :: atoms
	INTEGER :: TotAtoms
	LOGICAL :: StrSymmetry=.false. !True if the atomic list contains only symmetry independent part
      end module sf_composition

!********************************************
      MODULE sf_symsearch
!********************************************
      USE SF_Module
      INTEGER, ALLOCATABLE, DIMENSION(:) :: ZeroArr !Array necessary for the call to the Sgelsy routine
      REAL, ALLOCATABLE, DIMENSION (:) :: work !work array for sgelsy
      INTEGER :: lwork !the length of the array lwork
      REAL, ALLOCATABLE, DIMENSION(:) :: stepd
      INTEGER, ALLOCATABLE, DIMENSION(:) ::mind,maxd
      INTEGER :: NGuessCen
      REAL, ALLOCATABLE, DIMENSION(:,:) :: GuessCen
      INTEGER, ALLOCATABLE, DIMENSION(:,:) :: GuessCenMult
      INTEGER, ALLOCATABLE, DIMENSION (:,:) :: P, HP, IR
      REAL, ALLOCATABLE, DIMENSION(:,:) :: RIR
      end module sf_symsearch

!********************************************
      MODULE sf_convergence
!********************************************
      INTEGER, PARAMETER :: NFOM=4 !Number of analysed figures of merit
      INTEGER, PARAMETER :: NSpeed=3 !Number of different analyse speeds.
      INTEGER, dimension(1:NFOM), parameter :: ExpPeakSign=(/-1 ,-1, 1, 1 /) !Expected "drop direction" (+ or -) for the values of the Figures of Merit, and expected Slope and first Curvature peak Sign for R-factor, Tot. Charge, Score, Correlation Coefficient
      integer, DIMENSION(1:NSpeed),parameter :: NMem=(/40,70,120/) !Nbr of points retained in CycMem, SlopeMem and CurvMem; used in TestConvergence; MUST BE >= NDeriv and >= delay+2*counted
      INTEGER, PARAMETER :: NMemGlobal=120 !MUST BE the largest value of NMem
      integer, DIMENSION(1:NSpeed), parameter :: delay=(/15,20,50/) !Nbr of points between the checked point and the points used to calculate the average and standard deviation
      integer, DIMENSION(1:NSpeed), parameter :: counted=(/10,10,10/) !The "counted" first and "counted" last values of the last (NMem-delay) values are used to calculate the VALUE average and standard deviation.
      integer, parameter :: WaitCheckCycle=10 !Cycle number at which the "Wait test" is done
      integer, parameter :: NSlopeTemp=3 !Nbre of points used for slope smoothing
      integer, parameter :: NDeriv=5 !Nbr of points used for slope and 2nd derivative calculations; must be <= NMem, NMem
      real, DIMENSION(1:NSpeed), parameter :: CycMax=(/25,10,7/) !Increase/decrease of the figures of merit are detected if the absolute(checked value) exceeds CycAver+CycMax*CycDev
      real, DIMENSION(1:NSpeed), parameter :: SlopeMax=(/15.,3.5,3.5/) !Slope points are detected if they exceed SlopeAver+SlopeMax*SlopeDev
      real, DIMENSION(1:NSpeed), parameter :: CurvMax=(/15.,3.5,3.5/) !Curvature points are detected if they exceed CurvAver+CurvMax*CurvDev
      integer, DIMENSION(1:NSpeed), parameter :: MinCycPeakPoints=(/7,7,7/)!A "half-peak" of a Figure of Merit is detected if there are "MinCycPeakPoints" points exceeding CycAverCycMax*CycDev
      integer, DIMENSION(1:NSpeed), parameter :: MinSlopePeakPoints=(/7,7,7/)!A peak of slope is detected if there are "MinSlopePeakPoints" points exceeding SlopeAverSlopeMax*SlopeDev
      integer, DIMENSION(1:NSpeed), parameter :: MinCurvPeakPoints=(/5,5,5/)!Idem for curvature
      integer, DIMENSION(1:NSpeed), parameter :: OldPeakCrit=(/20,20,20/)! A peak is declared too old if it happened OldPeakCrit cycles before the average cycle of all detected peaks (PeakTopCycleAver).

      REAL, DIMENSION(1:NFOM,1:NMemGlobal), SAVE :: CycMem
      REAL, DIMENSION(1:NFOM,1:NSlopeTemp), SAVE :: SlopeTemp
      REAL, DIMENSION(1:NFOM,1:NMemGlobal), SAVE :: SlopeMem
      REAL, DIMENSION(1:NFOM,1:NMemGlobal), SAVE :: CurvMem
      INTEGER, SAVE :: TestConvRefCycle !Reference cycle: the cycle number of a call to Reset Subroutine

      TYPE convpar_type
	integer :: NMem, delay, counted
	real :: CycMax, SlopeMax, CurvMax
	integer :: MinCycPeakPoints, MinSlopePeakPoints, MinCurvPeakPoints
	integer :: OldPeakCrit
	INTEGER, DIMENSION(1:NFOM) :: CycCount, SlopeCount, CurvCount1, CurvCount2
        INTEGER, DIMENSION(1:NFOM) :: CycPeakPoints, CycPeakEndCycle
        INTEGER, DIMENSION(1:NFOM) :: SlopePeakPoints, SlopePeakTopCycle
        INTEGER, DIMENSION(1:NFOM) ::CurvPeak1Points, CurvPeak1EndCycle, CurvPeak1TopCycle, CurvPeak2Points
        REAL, DIMENSION(1:NFOM) :: CycPeakTopRel, SlopePeakTopRel, SlopePeakTopAbs, &
                                         CurvPeak1TopRel, CurvPeak1TopAbs, CurvPeak2TopRel, CurvPeak2TopAbs
        logical, dimension(1:NFOM) :: CycPeak, CycConverg, SlopePeak, SlopeConverg, CurvPeak1, CurvPeak2, CurvConverg
        logical, dimension(1:NFOM) :: CycWait, SlopeWait, CurvWait
        integer, dimension(1:NFOM) :: CycWaitSign, SlopeWaitSign, CurvWaitSign
      END TYPE convpar_type
      TYPE (convpar_type), DIMENSION(1:3), save :: convpar
      real, save :: LimitFOM=0.8, goodskew=3.0, goodrval=0.30, goodcharge=0., goodsym=30.
      integer, save :: stabicyc=-30 !Number of cycles for stabilization of the calculation
      integer, save :: skipstartcyc=0 !Number of initial cycles not to be considered for convergence
      integer :: goodconvcyc
      character*(20) :: convergencemode
      END MODULE SF_Convergence

!********************************************
      MODULE sf_rhoiointerfaces
!********************************************
      INTERFACE

        Subroutine WriteJana(inrho,rhofile)
          USE SF_Module
          real, target, dimension(0:) :: inrho
          CHARACTER*(*) :: rhofile
        end subroutine WriteJana

        Subroutine WriteXplor(inrho,rhofile)
          USE SF_Module
          real, target, dimension(0:) :: inrho
          CHARACTER*(*) :: rhofile
        end subroutine WriteXplor

        Subroutine WriteCCP4(inrho,rhofile)
          USE SF_Module
          real, target, dimension(0:) :: inrho
          CHARACTER*(*) :: rhofile
        end subroutine WriteCCP4

        SUBROUTINE ReadXPlor(inrho,RhoFile,gettinghkl)
          USE SF_Module
          USE Input_module
          IMPLICIT NONE
          real, target, dimension(0:) :: inrho      
          CHARACTER*(*) :: rhofile  ! Filename
          LOGICAL, OPTIONAL :: gettinghkl
        END SUBROUTINE ReadXPlor
	
        SUBROUTINE ReadJana(inrho,RhoFile)
          USE SF_Module
          IMPLICIT NONE
          real, target, dimension(0:) :: inrho      
          CHARACTER*(*) :: rhofile  ! Filename
        END SUBROUTINE ReadJana

        SUBROUTINE ReadCCP4(inrho,RhoFile)
          USE SF_Module
          IMPLICIT NONE
          real, target, dimension(0:) :: inrho      
          CHARACTER*(*) :: rhofile  ! Filename
        END SUBROUTINE ReadCCP4

        SUBROUTINE RhoFromCIF(inrho,RhoFile)
          USE SF_Module 
          USE SF_Composition
          IMPLICIT NONE
          real, target, dimension(0:) :: inrho      
          CHARACTER :: rhofile*(FileLength)  ! Filename
        END SUBROUTINE RhoFromCIF
      END INTERFACE

      END MODULE sf_RhoIOInterfaces

!********************************************
      MODULE FFT_Interfaces
!********************************************
! Explicit intefaces
      INTERFACE

      SUBROUTINE MakeFFTPlan(InNVox,inrho,insf,dir,infastfft,instr)
        USE SF_Module
        IMPLICIT NONE
        INTEGER, PARAMETER :: FFTW_FORWARD=+1
        INTEGER, PARAMETER :: FFTW_BACKWARD=-1
        INTEGER, PARAMETER :: FFTW_MEASURE=0
        INTEGER, PARAMETER :: FFTW_PATIENT=32
        INTEGER, PARAMETER :: FFTW_ESTIMATE=64
        INTEGER, DIMENSION(:) :: InNVox
        REAL, TARGET, DIMENSION(:) :: inrho, insf
        INTEGER :: Method, dir
        LOGICAL :: InFastFFT
        TYPE(FFTInstr) :: instr
      END SUBROUTINE MakeFFTPlan

      SUBROUTINE FFT(instr, inrho)
        USE SF_Module
        IMPLICIT NONE
        TYPE(FFTInstr) :: instr
        REAL, OPTIONAL, TARGET, DIMENSION(:) :: inrho
      END SUBROUTINE FFT

      END INTERFACE
     
      END MODULE FFT_Interfaces

!********************************************
      MODULE sf_interfaces
!********************************************
! Explicit intefaces
      INTERFACE

      INTEGER FUNCTION LCM(inarr)
        IMPLICIT NONE
	INTEGER, DIMENSION(:) :: inarr
	INTEGER, DIMENSION(1:size(inarr)) :: cumarr
      END FUNCTION LCM

      SUBROUTINE ResampleDensity(NVox1,NVox2,inrho)
        USE SF_Module
        USE FFT_Interfaces
        IMPLICIT NONE
        INTEGER, DIMENSION(SDim) :: NVox1,NVox2
        real, allocatable, dimension(:) :: inrho
      END SUBROUTINE ResampleDensity
      
      Subroutine WriteRho(inrho,rhofile,rhotype)
        USE SF_Module
        USE SF_RhoIOInterfaces
        real, target, dimension(0:) :: inrho
        character*(filelength) :: rhofile
        character*(linelength) :: rhotype
      END SUBROUTINE WriteRho
      
      SUBROUTINE ReadRho(inrho,rhofile,rhotype)
      USE SF_Module
      USE SF_RhoIOInterfaces
      real, target, dimension(0:) :: inrho
      character*(filelength) :: rhofile
      character*(linelength) :: rhotype
      end subroutine ReadRho

      LOGICAL function IndepSFPos(hkl,SFpos)
        use sf_module
	integer, dimension(:) :: hkl
	integer :: SFpos
      END FUNCTION IndepSFPos
            
      SUBROUTINE TimeTick(level, text, unit)
        integer :: level
        character*(*) :: text
        integer, optional :: unit
      END SUBROUTINE TimeTick

      SUBROUTINE GetDandS(SymOpNr,dvector,svector,key,NeedNewMap,InIntrVec)
      USE SF_Module
      USE SF_SymSearch
      IMPLICIT NONE
      INTEGER :: SymOpNr,key
!      TYPE(SymOp_type) :: InSymOp
      REAL, DIMENSION(1:SDim) :: dvector,svector
      REAL, DIMENSION(:),OPTIONAL :: InIntrVec
      LOGICAL :: NeedNewMap
      END SUBROUTINE GetDandS

      END INTERFACE

      END MODULE sf_interfaces
