      subroutine InputLog
      USE sf_module
      USE sf_composition
      USE sf_convergence
      USE Input_module
      IMPLICIT NONE
      integer :: SmtCnt1, SmtCnt2
      character*19 :: CellForm='(a,..f12.4,a,f15.4)'
      character*14 ::  VoxForm='(a,..i6,a,i10)'
      character*14 ::  RDimRealForm='(..f9.4)'
      character*14 ::  SDimRealForm='(..f9.4)'
      character*11 :: SymForm='(..i3,f9.4)'
      character*11 :: TwinSymForm='(..i3,5X)'
      character*55 :: SymLabelForm='(i.,'' symmetry operations found, their list follows:'')'
      character*68 :: TestLabelForm='(i.,'' symmetry operations found for testing, their list follows:'')'
      character*68 :: TwinLabelForm='(i.,'' twin matrices found, their list follows:'')'
      character*7  :: CompForm= '(a,i3)'

      write(CellForm(4:5),'(i2)') NCellPar
      write(VoxForm(4:5),'(i2)') SDim
      write(RDimRealForm(2:3),'(i2)') RDim
      write(SDimRealForm(2:3),'(i2)') SDim
      write(SymForm(2:3),'(i2)') SDim
      write(TwinSymForm(2:3),'(i2)') SDim
      IF (NSym>0) write(SymLabelForm(3:3),'(i1)') int(log10(float(NSym))+eps)+1
      IF (NTestSym>0) write(TestLabelForm(3:3),'(i1)') int(log10(float(NTestSym))+eps)+1
      IF (NTwin>0) write(TwinLabelForm(3:3),'(i1)') int(log10(float(NTwin))+eps)+1
      call FramedLabel(1,'Following data were read from the input file or set as default:')
      write(lgu,'(2a)') 'Job title: ',trim(Title)
      call FramedLabel(2,'Information about files:')
      write(lgu,'(2a)') 'Name of the input file: ',trim(InFile)
      IF (OutRhoNr==1) THEN
         write(lgu,'(4a)')   'Density will be writen in ',trim(OutRhoType(1)),' format to file   ', &
	                    trim(OutRhoFile(1))
      ELSE
         write(lgu,'(4a)')   'Density will be writen in following files:'
         DO cnt1=1,OutRhoNr
          write(lgu,'(4a)') '   ',trim(OutRhoFile(cnt1)),', with type ',trim(OutRhoType(cnt1))
         ENDDO
      ENDIF
!      IF ((outrhotype=='xplor').and.(SDim/=3)) THEN
!        write(lgu,'(a)') 'Warning: the Xplor format in dimension different from 3 is superflip''s extension '// &
!                       'to the standard format.' 
!      ENDIF
      IF (RefRhoType/='none') THEN
        write(lgu,'(4a)') 'Reference density will be read in ',trim(RefRhoType),' format from file ',trim(RefRhoFile)
      END IF
      IF (EnvRhoType/='none') THEN
        write(lgu,'(4a)') 'Envelope will be read in ',trim(EnvRhoType),' format from file ',trim(EnvRhoFile)
      END IF
      IF (ModelRhoType/='none') THEN
        SELECT CASE (ModelRhoType)
        CASE('patterson')
          write(lgu,'(a)') 'Patterson function with origin peak removed will be used as a starting density.' 
        CASE('supminfunction')
          write(lgu,'(a)') 'Superposition minimum function will be used as a starting density.' 
        CASE('random')
          write(lgu,'(a)') 'A random structure generated from the composition will be used as a starting density.'
        CASE DEFAULT
          write(lgu,'(4a)') 'The starting density will be read in ',trim(ModelRhoType),' format from file ', trim(ModelRhoFile)
        END SELECT
      END IF
      IF (RewriteOutput) THEN
        write(lgu,'(2a)') 'Warning: If the outputfile exists, it will be overwritten by the new density.'
      ELSE
        write(lgu,'(2a)') 'Warning: If the outputfile exists, the density will be written to another file.'
        write(lgu,'(2a)') '         (See the end of the log-file for the name of the output file).'
      ENDIF

      IF (ExpandedLog) THEN
        write(lgu,'(2a)') 'Logfile will contain extended information about the calculation.'
      ELSE
        write(lgu,'(2a)') 'Logfile will contain only basic information about the calculation.'
      ENDIF
      IF (CalcCoverage) THEN
        write(lgu,'(2a)') 'Logfile will contain information about the data coverage.'
      ENDIF
      call FramedLabel(2,'Crystallographic information:')
      write(lgu,'(a,i3)') 'Superspace dimension:           ',SDim	
      write(lgu,'(a,i3)') 'Dimension of the physical space:',RDim	
      write(lgu,cellform) 'Direct cell parameters:       ',(Cell(cnt1),cnt1=1,NCellPar),'  Volume:',Cell(NCellPar+1)
      write(lgu,cellform) 'Reciprocal cell parameters:   ',(RCell(cnt1),cnt1=1,NCellPar),'  Volume:',RCell(NCellPar+1)
      IF (Sdim-RDim>0) THEN
        write(lgu,'(''Q-vectors:'')')	        
        do cnt1=1,SDim-RDim
          write(lgu,rdimrealform) (QVec(cnt2,cnt1),cnt2=1,RDim)
        enddo
      ENDIF
      IF (ElNr>0) THEN
        write(lgu,'(a)',ADVANCE='NO') 'Wilson plot will be based on this composition: '
	DO cnt1=1,ElNr
	  write(compform(5:5),'(i1)') int(log10(float(max(Compos(cnt1,2),1))))+1
	  write(lgu,compform,ADVANCE='NO') trim(ElSmb(Compos(cnt1,1))),Compos(cnt1,2)
	ENDDO
	write(lgu,*)
      ENDIF
      if (NTwin>0) THEN
        write(lgu,'('''')')
        write(lgu,TwinLabelForm) NTwin
        write(lgu,'('''')')
        SmtCnt1=-2
        SmtCnt2=0
        do
          SmtCnt1=min(SmtCnt1+3,NTwin)
	  SmtCnt2=min(SmtCnt2+3,NTwin)
	  do cnt2=1,SDim
            do cnt1=SmtCnt1,SmtCnt2
              IF (cnt2==1) THEN
	        write(lgu,'(i4,'':'')',ADVANCE='NO') cnt1
	      ELSE
	        write(lgu,'(''     '')',ADVANCE='NO')
              ENDIF
	      write(lgu,TwinSymForm,ADVANCE='NO') (TwinOp(cnt1)%R(cnt2,cnt3),cnt3=1,SDim)
	    enddo
            write(lgu,'('''')')
          enddo
          write(lgu,'('''')')
          if (SmtCnt2==NTwin) EXIT
        enddo
      endif

      write(lgu,SymLabelForm) NSymPrim
      write(lgu,'('''')')
      SmtCnt1=-2
      SmtCnt2=0
      do
        SmtCnt1=min(SmtCnt1+3,NSymPrim)
	SmtCnt2=min(SmtCnt2+3,NSymPrim)
	do cnt2=1,SDim
          do cnt1=SmtCnt1,SmtCnt2
            IF (cnt2==1) THEN
              IF (SurviveTwin(cnt1)) THEN
                write(lgu,'(i4,'': '')',ADVANCE='NO') cnt1
	      ELSE
                write(lgu,'(i4,''*:'')',ADVANCE='NO') cnt1
              ENDIF
	    ELSE
	      write(lgu,'(''      '')',ADVANCE='NO')
            ENDIF
	    write(lgu,SymForm,ADVANCE='NO') (SmOp(cnt1)%R(cnt2,cnt3),cnt3=1,SDim),SmOp(cnt1)%t(cnt2)
	  enddo
          write(lgu,'('''')')
        enddo
        write(lgu,'('''')')
	  
        if (SmtCnt2==NSymPrim) EXIT
      enddo

      IF (NTwin>0) THEN
        DO cnt1=1,NSymPrim
          IF (.not.SurviveTwin(cnt1)) THEN
            write(lgu,'(''The starred symmetry operations will not be used in the expansion of the reflections'')')
            write(lgu,'(''because they are not preserved by the twin matrices listed above.'')')
            write(lgu,'('''')')
            EXIT
	  ENDIF
	ENDDO
      ENDIF
      
      IF (Centro) THEN
        write(lgu,'(''The structure is centrosymmetric.'')')      
      ELSE
        write(lgu,'(''The structure is non-centrosymmetric.'')')      
      ENDIF
      
      IF (NCenter>1) THEN
        write(lgu,'(''The symmetry operations are to be combined with the following centering vectors:'')')
        do cnt1=1,NCenter
          write(lgu,sdimrealform) (Center(cnt2,cnt1),cnt2=1,SDim)
        enddo
	write(lgu,'('''')')
      ENDIF

      if (NTestSym>0) THEN
        write(lgu,TestLabelForm) NTestSym
        write(lgu,'('''')')
        SmtCnt1=-2+NSym
        SmtCnt2=NSym
        do
          SmtCnt1=min(SmtCnt1+3,NSym+NTestSym)
	  SmtCnt2=min(SmtCnt2+3,NSym+NTestSym)
	  do cnt2=1,SDim
            do cnt1=SmtCnt1,SmtCnt2
              IF (cnt2==1) THEN
	        write(lgu,'(i4,'':'')',ADVANCE='NO') cnt1-NSym
	      ELSE
	        write(lgu,'(''     '')',ADVANCE='NO')
              ENDIF
	      write(lgu,SymForm,ADVANCE='NO') (SmOp(cnt1)%R(cnt2,cnt3),cnt3=1,SDim),SmOp(cnt1)%t(cnt2)
	    enddo
            write(lgu,'('''')')
          enddo
          write(lgu,'('''')')
          if (SmtCnt2==NSym+NTestSym) EXIT
        enddo
      endif

      call FramedLabel(2,'Settings of the algorithm:')
      write(lgu,voxform)  'Number of voxels:               ',(NVoxIter(cnt1),cnt1=1,SDim),'   Total:',product(NVoxIter)
      IF ((any(NVoxIter/=NVoxFine)).or.(polish)) THEN
        write(lgu,voxform)  'Number of voxels for polishing: ',(NVoxFine(cnt1),cnt1=1,SDim),'   Total:',product(NVoxFine)
      ENDIF
      IF (inperform=='cf') THEN
        write(lgu,'(a)')  'The iteration scheme will be standard charge flipping (CF).'
      ELSEIF (inperform=='aar') THEN
        write(lgu,'(a)')  'The iteration scheme will be averaged alternating reflections (AAR).'
      ELSEIF (inperform=='lde') THEN
        write(lgu,'(a)')  'The iteration scheme will be low-density elimination (LDE).'
      ELSEIF (inperform=='dm') THEN
        write(lgu,'(a)')  'The iteration scheme will be difference map (DM).'
      ELSEIF (inperform(1:7)=='general') THEN
        write(lgu,'(a,6f6.2)')  'The iteration scheme will be "general" with the following parameters:',&
              gdmpar%beta1,gdmpar%gamma1R,gdmpar%gamma1D, gdmpar%beta2,gdmpar%gamma2D,gdmpar%gamma2R
      ELSEIF (inperform=='fourier') THEN
        write(lgu,'(a)')  'The Fourier transform of the input data will be performed.'
      ELSEIF (inperform=='symmetry') THEN
        write(lgu,'(a)')  'The symmetry of the model file will be analyzed.'
      ENDIF
      IF (fdeltaf) THEN
        IF (fdeltafw>1000) THEN
          write(lgu,'(a)')  'The F+deltaF reciprocal-space modification will be used with no limit on the amplitude change.'
        ELSE
          write(lgu,'(a,f5.2)')  'The F+deltaF reciprocal-space modification will be used with the amplitude change limited to F*',&
          fdeltafw
        ENDIF
      ENDIF
      IF ((inperform/='fourier').and.(inperform/='symmetry')) THEN
        IF (InAutoDelta) THEN
          write(lgu,'(a)')  'Delta will be determined automatically.'
        ELSE
          write(lgu,'(a,f8.3)')  'Delta: ',CfInDelta
          IF (DeltaType==RelDelta) THEN
            write(lgu,'(a,f8.3,a)') '  Dynamic delta will be used:',CFInDelta, &
	              ' of all pixels with the smallest density will be flipped in each cycle.'
          ELSEIF (DeltaType==StatDelta) THEN
            write(lgu,'(a,f8.3,a)') '  Static delta will be used: pixels with density below', &
	              CFInDelta,' will be flipped in each cycle.'
          ELSEIF (DeltaType==KedDelta) THEN
            write(lgu,'(a,f8.3,a)') '  Static delta will be used: pixels with density below', &
	              CFInDelta,'*sigma(map) will be flipped in each cycle.'
	  END IF
        ENDIF
      ENDIF
      IF (neutrons) THEN
        IF (StartNeutrons==0) THEN
          write(lgu,'(a)') 'Both positive and negative densities will be allowed to evolve - be aware of Babinet solutions.'
        ELSE
          write(lgu,'(a,i5,a)') 'Negative densities will be allowed to evolve starting from cycle number ',StartNeutrons,'.'
        ENDIF
      ENDIF

      IF (inconvergencemode=='never') THEN
        write(lgu,'(a,i5,a)') 'Convergencemode set to never - the iteration will be stopped after ',InMaxCycles,' cycles.'
      ELSE
        write(lgu,'(a,i9,a)') 'The iteration will be stopped when the convergence is detected or after ',InMaxCycles,' cycles.'
      ENDIF
      IF (stabicyc/=0) THEN
        write(lgu,'(a,i5,a)') 'After the detection of convergence',stabicyc,' cycles of iteration will be added.'
      ENDIF
      IF (polish) THEN
        write(lgu,'(a,i4,a)') 'After the regular iteration',polishcycles,' cycles of "density polishing" will be applied.'
      ENDIF
      IF (AutoRS) THEN
        write(lgu,'(a)')  'The random number generator will be initialized automatically.'
      ELSE
        write(lgu,'(a,i10)')  'Seed for the random number generator:',RandomSeed
      ENDIF

      IF (repeatmode/='never') THEN
        IF (repeatmode=='nosuccess') THEN
          write(lgu,'(a)') 'The whole calculation will be repeated until convergence is detected.'
        ELSEIF (repeatmode=='always') THEN
          write(lgu,'(a)') 'The whole calculation will be repeated indefinitely.'
        ELSEIF (repeatmode=='number') THEN
           write(lgu,'(a,i3,a)') 'The whole calculation will be repeated',repeatattempts,' times.'
        ENDIF
        IF (NBestRho>1) THEN
	  write(lgu,'(i3,a)') NBestrho,' densities with the best R-value will be saved with the names bestXX_YYYY,'// &
	                     ' where XX is the serial number and YYYY is the name of the output density given by user.'
	ENDIF
      ENDIF


      write(lgu,*)
      IF (ModelRhoType/='none') THEN
        write(lgu,'(a)')  'The initial phases of the structure factors will be taken from the starting density file.'      
      ELSEIF (UsePhases=='firstcycle') THEN
        write(lgu,'(a)') 'Known phases from the input file will be used as initial phases,'// &
	                 ' unknown phases will be initiated randomly.'
      ELSEIF(UsePhases=='always') THEN
        write(lgu,'(a)') 'Known phases from the input file will be used in each iteration cycle,'// &
	                 ' unknown phases will be initiated randomly.'
      ELSEIF(UsePhases=='number') THEN
        write(lgu,'(a,i6,a)') 'Known phases from the input file will be used in first',setphasecycles,&
	                      ' iteration cycles, unknown phases will be initiated randomly..'
      ELSE
        write(lgu,'(a)')  'Random initial phases will be assigned to the structure factors.'
      ENDIF
      write(lgu,'(a,f8.3)') 'Isotropic Debye-Waller factor:', biso
      IF (weakreflmode==wrcode_relative) THEN
        write(lgu,'(a,f8.3)') 'Proportion of reflections to be treated as weak:', weakratio
      ELSEIF(weakreflmode==wrcode_absolute) THEN
        write(lgu,'(a,f8.3)') 'Reflection amplitude for treating the reflection as weak:', weakratio
      ENDIF
      IF (externrefl) THEN
        WRITE(lgu,'(a,a)') 'Reflections read from external reflection file:',trim(externreflfile)
      ELSE
        WRITE(lgu,'(a,a)') 'Reflections read from internal reflection list.'
      ENDIF
      IF (reflsthlmax<10000.) THEN !Resolution limits imposed by user
        IF (resunits==res_d) THEN
	  IF (refldmax<1000.) THEN
	    write(lgu,'(a,f8.2,a,f8.2,a)') 'Only reflection with d > ',refldmin,' and d < ',refldmax,' will be used.'
	  ELSE
	    write(lgu,'(a,f8.2,a)') 'Only reflection with d > ',refldmin,' will be used.'
	  ENDIF
	ELSEIF (resunits==res_sthl) THEN
	  write(lgu,'(a,f5.3,a,f5.3,a)') 'Only reflection with sin(th)/lambda > ',reflsthlmin,' and sin(th)/lambda < ', &
	       reflsthlmax,' will be used.'
	ENDIF
      ENDIF
      IF ((FillMissingSthl>0.).and.(MissingMode/=mm_zero)) THEN !Special treatment of missing reflections
        IF (resunits==res_d) THEN
	  write(lgu,'(a,f8.2,a)') 'All reflections with d > ',1/(2*max(FillMissingSthl,eps)), &
	    ' not present in the input data set will be treated a missing.'
	ELSEIF (resunits==res_sthl) THEN
	  write(lgu,'(a,f5.3,a)') 'All reflections with sin(th)/lambda < ',FillMissingSthl, &
	    ' not present in the input data set will be treated a missing.'
	ENDIF
        SELECT CASE(MissingMode)
	  CASE(mm_float)
          write(lgu,'(a)') 'The amplitudes of the missing reflections will be left unconstrained.'
	  CASE(mm_bound)
            IF (NormalizeRefl==NormNone) THEN
              write(lgu,'(a,f5.1,a)') 'The amplitudes of the missing reflections will be constrained to less than',&
                MissingBndFactor,' times the Wilson-plot-based intensity.'
            ELSE
              write(lgu,'(a,f5.1,a)') 'The normalized amplitudes of the missing reflections will be'// &
	        ' constrained to less than',MissingBndFactor,'.'
	    ENDIF
	  CASE(mm_boundsum)
            IF (NormalizeRefl==NormNone) THEN
              write(lgu,'(a)') 'The sum of amplitudes of the missing reflections will be constrained to the '//&
	         'Wilson-plot-based value.'
            ELSE
              write(lgu,'(a)') 'The average normalized amplitude of the missing reflections will be constrained to 1.'
	  ENDIF
	END SELECT
      ELSE
        write(lgu,'(a)') 'Reflections missing in the input - if any - will be reset to 0 at each cycle.'
      ENDIF

      IF (hmclasses/=0) THEN !Histogram matching
        write(lgu,'(a,i6,a,i6,a)') 'The density will be matched to the histogram every',inhmstepc, &
	                           ' cycles starting from the',inhmstartc,'th cycle.'
	    IF (NGroups>0) THEN
          write(lgu,'(a)') '  The reflection amplitudes within each reflection group will be'// &
	                 ' repartitioned after each histogram-matching procedure.'
        ENDIF
      ENDIF
      IF (envrhotype/='none') THEN !Envelope matching
        write(lgu,'(a,i6,a,i6,a)') 'The density will be matched to the envelope every',envstepc, &
	                           ' cycles starting from the',envstartc,'th cycle.'
        write(lgu,'(a,f6.3,a,f6.3,a)') 'Zero level of the envelope is',EnvZeroLevel,', the minimum cutoff is',envcutmin,'.'
      ENDIF
      write(lgu,*)
      IF (averagesym/=NoAveraging) THEN
        write(lgu,'(a)') 'The resulting density will be shifted and averaged according to the symmetry operations given above.'
      ELSEIF(SearchSym) THEN
        write(lgu,'(a)') 'The resulting density will be shifted so that it best matches the symmetry operations given above.'
	write(lgu,'(a)') 'No averaging of the density pixels will be performed.'
      ELSE
	write(lgu,'(a)') 'No search for the position of the symmetry elements will be performed.'
      ENDIF
      IF (searchsym.and.(symroutine=='new')) THEN
	write(lgu,'(a)', ADVANCE='NO') 'Following symmetry operations will be used to locate the origin of symmetry:' 
	DO cnt1=1,NGen
          write(lgu,'(i4)',ADVANCE='NO') Generators(cnt1)
	ENDDO      
        write(lgu,*)
      ENDIF
      IF (refrhotype/='none') THEN
        write(lgu,'(a)') 'The resulting density will be shifted so that is matches the reference density.'        
      ENDIF

      end subroutine InputLog

      subroutine FramedLabel(key,label)
        USE sf_module
        character*(*) :: label
        character*(linelength) :: frame
	integer :: key !key=1 means full frame with #, key=2 is second-order frame with ---

	if (len_trim(label)>linelength) THEN
	  CALL ErrorMsg('Zle maticko, zle, co to tam pises do logfajlu?')
	  CALL ErrorMsg(trim(label))
	  CALL StopProgram('Internal inconsistency of the program. Please send this message to the author.')
	endif
	IF (Key==1) THEN
          do i=1,len_trim(label)+4
            frame(i:i)='#'
          enddo
          write(lgu,'(a)') ''
          write(lgu, '(a)') frame(:len_trim(label)+4)
          write(lgu,'(3a)') '# ',trim(label),' #'
          write(lgu, '(a)') frame(:len_trim(label)+4)
          write(lgu,'(a)') ''
        ELSEIF (Key==2) THEN
          do i=1,len_trim(label)
            frame(i:i)='-'
          enddo
          write(lgu,'(a)') ''
          write(lgu, '(a)') frame(:len_trim(label))
          write(lgu,'(a)') trim(label)
          write(lgu, '(a)') frame(:len_trim(label))
        END IF
      end subroutine FramedLabel

      subroutine WriteRefl
      USE sf_module
      USE Input_module
      IMPLICIT NONE
      CHARACTER*53 ::  CFForm='(   I4,2F14.7,3X,2F14.7)'
      CHARACTER*56 ::  CFGForm='(   I4,2F14.7,3X,2F14.7,i8)'
      CHARACTER*43 ::  FRForm='(   I4,2F14.7)'
      INTEGER :: i,ii

      WRITE (CFForm(2:4),'(I3)') SDim
      WRITE (CFGForm(2:4),'(I3)') SDim
      WRITE (FRForm(2:4),'(I3)') SDim
!!!!!!! Important: The convention in FFTW is OPOSITE to that in crystallography.
!!!!!!! Therefore internally the B component or Refl must be inverted on input and inverted again in output
!!!!!!! Unpleasant, but necessary
      IF (expandedlog) THEN
        call FramedLabel(1,'Structure factors as phased by charge flipping')
        IF (NGroups==0) THEN
          WRITE (lgu,'(A)') &
                 '     #     h   k   l...     A              B             A(new)         B(new)'
          DO i=StartF, EndF
            WRITE (lgu,'(I6,A)',ADVANCE='NO') i, ': '
            WRITE (lgu,CFForm) &
            (Refl(i)%hklFG(ii),ii=1,SDim), Refl(i)%A, -Refl(i)%B, Refl(i)%Amem, -Refl(i)%Bmem
          END DO
        ELSE
          WRITE (lgu,'(A)') &
               '     #     h   k   l...     A              B             A(new)         B(new)      Group'
          DO i=StartF, EndF
            WRITE (lgu,'(I6,A)',ADVANCE='NO') i, ': '
            WRITE (lgu,CFGForm) &
            (Refl(i)%hklFG(ii),ii=1,SDim), Refl(i)%A, -Refl(i)%B, Refl(i)%Amem, -Refl(i)%Bmem,Refl(i)%group
          END DO
        ENDIF      
        write(lgu,'('''')')
      ELSE
        write(lgu,'(a)') 'You can obtain more information about the reconstructed reflection phases by using ''expandedlog yes''.'
        write(lgu,'('''')')
      ENDIF
      end subroutine WriteRefl

      subroutine WriteFullRefl(ReflFile)
      USE sf_module
      USE Input_module
      IMPLICIT NONE
      CHARACTER*(*) :: ReflFile
      CHARACTER*43 ::  FRForm='(   I4,2F14.7,F14.4,F14.1)'
      INTEGER :: i,ii, ErrCode,dcoo(SDim)
      LOGICAL :: Succ
      REAL       Fmodul, PhiCF

        WRITE (FRForm(2:4),'(I3)') SDim
!!!!!!! Important: The convention in FFTW is OPOSITE to that in crystallography.
!!!!!!! Therefore internally the B component or Refl must be inverted on input and inverted again in output
!!!!!!! Unpleasant, but necessary
        OPEN (UNIT=99,FILE=trim(ReflFile),IOSTAT=ErrCode)
        IF (ErrCode/=0) THEN
          CALL ErrorMsg('Error, can''t open file "'//trim(ReflFile)//'".')
          CALL ErrorMsg('The complete list of reflection has not been saved.')
          RETURN
        END IF  
	Fmodul=SQRT(MEMRefl(MStartF)%AMEM**2+MEMRefl(MStartF)%BMEM**2)
!	IF ((abs(MEMRefl(MStartF)%AMEM)>eps).AND.(abs(MEMRefl(MStartF)%BMEM)>eps)) then
	   PhiCF=-atan2(MEMRefl(MStartF)%BMEM,MEMRefl(MStartF)%AMEM)/torad  	
!	ELSE 
!	   PhiCF=0.0 	
!	endif 	
        WRITE (99,FRForm) &
         (MEMRefl(MStartF)%hklFG(ii),ii=1,SDim), MEMRefl(MStartF)%AMEM, -MEMRefl(MStartF)%BMEM,Fmodul,PhiCF
        DO i=MStartF+1, MEndF          
	  Succ=.true.
	  IF (MEMRefl(i)%hklFG(1)==0) THEN
            call CoordinatesSF(MEMRefl(i)%SFloc,DCoo)	  
            DO ii=1,SDim !Filter out the duplicitly present reflections if the h=0 layer
	      IF (DCoo(ii)==0) CYCLE !Skip zero indices
	      IF (DCoo(ii)>NVox(ii)/2) THEN !The first non-zero index must be smaller than NVox(ii)/2
	        Succ=.false.              
              ENDIF
              EXIT
            ENDDO
          ENDIF
	  IF (Succ) THEN
	    Fmodul=SQRT(MEMRefl(i)%AMEM**2+MEMRefl(i)%BMEM**2)
!	    IF ((abs(MEMRefl(i)%AMEM)>eps).AND.(abs(MEMRefl(i)%BMEM)>eps)) then
	      PhiCF=-atan2(MEMRefl(i)%BMEM,MEMRefl(i)%AMEM)/torad  	
!            ELSE 
!	      PhiCF=0.0 	
!            endif 	
	    WRITE (99,FRForm) &
              (MEMRefl(i)%hklFG(ii),ii=1,SDim), MEMRefl(i)%AMEM, -1.0*MEMRefl(i)%BMEM, Fmodul, PhiCF
            WRITE (99,FRForm) &
              (-MEMRefl(i)%hklFG(ii),ii=1,SDim), MEMRefl(i)%AMEM, -1.0*(-MEMRefl(i)%BMEM), Fmodul,-1.0*PhiCF
          ENDIF
	END DO
        CLOSE(99)
      end subroutine WriteFullRefl

! Let's save the density in various formats
!*****************************************
      Subroutine WriteRho(inrho,rhofile,rhotype)
!*****************************************
      USE SF_Module
      USE SF_RhoIOInterfaces
      real, target, dimension(0:) :: inrho
      character*(filelength) :: rhofile
      character*(linelength) :: rhotype
      character*(extlength) :: ext
      character*(filelength) :: tmprhofilebody='sfrho..', tmprhofile

      logical :: exflag
      integer :: i
      
      SELECT CASE(RhoType)
        CASE ('jana')
	  ext='.m81'	  
        CASE ('xplor')
	  ext='.xplor'
        CASE ('m80')
	  ext='.m80'
        CASE ('shelxsym')
	  ext='.ins'
        CASE ('ccp4')
	  ext='.ccp4'
      END SELECT
      ! IF existing files should not be rewritten, let's take care of it
      INQUIRE (FILE=rhofile,EXIST=exflag) !test, if the file exists, if not, nothing happens
      IF (exflag.and..not.rewriteoutput) then
        tmprhofile=trim(tmprhofilebody)//ext
	DO i=0,99
	  write(tmprhofile(6:7),'(i2.2)') i
          INQUIRE (FILE=TmpRhoFile,EXIST=exflag) !test, if the file exists, if not, nothing happens
	  IF (.not.exflag) EXIT
	  IF (i==99) THEN !No free slot has been found
	    write(lgu,'(a)') 'Error, the requested output file '//trim(rhofile)//' exists.'
	    write(lgu,'(a)') 'All files sfrho00'//trim(ext)//' through sfrho99'//trim(ext)//' exist too.'
	    write(lgu,'(a)') 'Cannot write the output density. Please change the output file, delete some of the files,'
	    write(lgu,'(a)') 'or use setting rewriteoutput yes to rewrite existing output files.'
            CALL StopProgram('Error during writing of the density. Check the log file.')
	  ENDIF
	ENDDO
	write(lgu,'(a)') 'Warning, the requested output file '//trim(rhofile)//' exists.'
        write(lgu,'(a)') 'Rewriting of the output files is disabled.'
        write(lgu,'(a)') 'The density will be written to the file '//trim(tmprhofile)//'.'  
        rhofile=tmprhofile
      ENDIF
      SELECT CASE(RhoType)
        CASE ('jana')
	  call WriteJana(inrho,rhofile)
        CASE ('xplor')
	  call WriteXplor(inrho,rhofile)
        CASE ('ccp4')
	  call WriteCCP4(inrho,rhofile)
        CASE ('m80')
	  call WriteM80(rhofile)
        CASE ('shelxsym')
	  IF (SDim==3) THEN
            call WriteShelxSym(rhofile)
          ELSE
  	    write(lgu,'(a)') 'Sorry, shelx-like symmetry file is available only for 3D problems.'
	    IF (terminal) write(*,'(a)') 'Sorry, shelx-like symmerty file is available only for 3D problems.'
	  ENDIF
      END SELECT
     
      end subroutine WriteRho
      
! Let's save the density in various formats
!*****************************************
      Subroutine ReadRho(inrho,rhofile,rhotype)
!*****************************************
      USE SF_Module
      USE SF_RhoIOInterfaces
      real, target, dimension(0:) :: inrho
      character*(filelength) :: rhofile
      character*(linelength) :: rhotype
      SELECT CASE(RhoType)
        CASE ('jana')
	  call ReadJana(inrho,rhofile)
        CASE ('xplor')
	  call ReadXplor(inrho,rhofile,.false.)
        CASE ('ccp4')
	  call ReadCCP4(inrho,rhofile)
   	 CASE('cif')
          call RhoFromCIF(inrho,rhofile)
      END SELECT
     
      end subroutine ReadRho

!Get the NVox information from the density file
!****************************************
      Subroutine NVoxFromFile(rhofile,rhotype,InNVox)
!****************************************
      USE SF_module
      IMPLICIT NONE
      character*(filelength) :: rhofile
      character*(linelength) :: rhotype
      integer, dimension(1:SDim) :: InNVox
      SELECT CASE(RhoType)
        CASE ('jana')
	  call NVoxFromJana(rhofile, InNVox)
        CASE ('xplor')
	  call NVoxFromXplor(rhofile, InNVox)
        CASE ('ccp4')
	  call NVoxFromCCP4(rhofile, InNVox)
   	 CASE('cif')
          InNVox=NVox
      END SELECT

      END SUBROUTINE NVoxFromFile

!Write the density in the JANA format
!****************************************
      Subroutine WriteJana(inrho,rhofile)
!****************************************
      USE SF_Module
      real, target, dimension(0:) :: inrho
      CHARACTER*(*) :: rhofile
      INTEGER :: nxny, nmap, mapa, nsubs
      REAL :: rmax, rmin
      REAL, ALLOCATABLE :: table(:)
      REAL :: xfmn(1:6), xfmx(1:6), xdf(1:6)
      INTEGER :: iorien(1:6),NVoxtmp(1:6)
      INTEGER :: GetRecLFac !Function returning the length of a unit record
      LOGICAL :: Satelitybyly      

      DO i=1,SDim
	iorien(i)=i
        xfmn(i)=0.E0
        xfmx(i)=float(NVox(i)-1)/float(NVox(i))
        xdf(i)=1./float(NVox(i))
	NVoxTmp(i)=NVox(i)
      END DO
      DO i=SDim+1,6
        iorien(i)=i
        xfmn(i)=0.E0
        xfmx(i)=0.E0
        xdf(i)=1.E0
	NVoxTmp(i)=1
      END DO
      nxny=NVox(1)*NVox(2)
      nmap=NVoxTot/nxny
      mapa=4 !Indicates an Fobs map
      rmin=100000.
      rmax=-100000.
      nsubs=1
      IF (SDim>RDim) THEN
        SatelityByly=.true.
      ELSE
        SatelityByly=.false.
      ENDIF      
      ALLOCATE (table(nxny))

!     Write the header of the output map
      RecLFac=GetReclFac()
      open(rhu,file=rhofile,form='unformatted', access='direct',recl=(max(nxny,35)*ReclFac),ERR=900)
      WRITE (rhu,rec=1,ERR=910) (NVoxtmp(i),i=1,6), nxny, nmap, (xfmn(i),xfmx(i),i=1,6), (xdf(i),i=1,6), iorien, mapa, nsubs, &
                                Satelitybyly
!     write the density
      cnt1=-1
      DO i=1,nmap
        DO ii=1,nxny
          cnt1=cnt1+1
          table(ii)=inrho(cnt1)/cApVx
	  rmax=max(table(ii),rmax)
	  rmin=min(table(ii),rmin)
	END DO
        WRITE (rhu,rec=i+1,ERR=910) table
      END DO
      WRITE (rhu,rec=nmap+2,ERR=910) rmax, rmin
      CLOSE(rhu,ERR=920)
      WRITE(lgu,'(A)') 'Electron density written to file '//trim(rhofile)//'.'
      write(lgu,'('''')')
      GOTO 999

!       Error handling      
900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine WriteJana:'
      WRITE (lgu,'(A)') 'Error! Cannot open density file for writing!'
      WRITE (lgu,'(A)') 'Output file: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopPRogram('Error during writing of the density. Check the log file.')
      
910   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine WriteJana:'
      WRITE (lgu,'(A)') 'Error! Cannot write into density file.'
      WRITE (lgu,'(A)') 'Output file: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Something may have been saved. Stopping.'
      WRITE (lgu,'(A)') ''
      CLOSE(rhu)
      CALL StopProgram('Error during writing of the density. Check the log file.')

920   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine WriteJana:'
      WRITE (lgu,'(A)') 'Error! Cannot close density file.'
      WRITE (lgu,'(A)') 'Output file: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Density might have been saved. Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during writing of the density. Check the log file.')


999   CONTINUE
      DEALLOCATE(table)
      end subroutine WriteJana
      
!Read the reference density in JANA format
!****************************************
      Subroutine ReadJana(inrho,rhofile)
!****************************************
      USE SF_Module
      real, target, dimension(0:) :: inrho      
      CHARACTER*(*) :: rhofile
      INTEGER :: nxny, nmap, mapa, nsubs,target
      REAL, ALLOCATABLE :: table(:)
      REAL :: xfmn(1:6), xfmx(1:6), xdf(1:6), CurCapVx
      INTEGER :: iorien(1:6),CurNVox(1:6)
      INTEGER :: GetRecLFac !Function returning the length of a unit record
      LOGICAL :: Satelitybyly      

      RecLFac=GetReclFac()
      open(rhu,file=rhofile,form='unformatted', access='direct',status='OLD',recl=35*ReclFac,ERR=900)

!     Read the information from the header
!      write(*,*) 'Prvni rtadek'
      READ (rhu,rec=1,ERR=910) (CurNVox(i),i=1,6), nxny, nmap, (xfmn(i),xfmx(i),i=1,6), (xdf(i),i=1,6), iorien, mapa, nsubs, &
            Satelitybyly
      ALLOCATE (table(nxny))
!     Make some basic check of consistency
!      DO cnt1=1,SDim
!        IF (NVoxtmp(cnt1)/=NVox(cnt1)) THEN
!	  write(lgu,'(a,i2,a)') 'Error, the voxel division in the reference density along dimension',cnt1,&
!	                   ' is inconsistent with the input file.'
!          CALL StopProgram('Error during reading of the reference density. Check the log file.')
!        ENDIF
!      ENDDO
      IF (size(inrho)/=product(CurNVox(1:SDim))) THEN
        CALL ErrorMsg('Error, incorrectly allocated density array.')
        CALL StopProgram('Internal error of the program. Please inform the autor.')
      ENDIF
      CurCapvx=Cell(NCellPar+1)/PRODUCT(CurNVox)
      DO cnt1=1,SDim
        IF (abs(xfmn(cnt1))>eps.or.abs(xfmx(cnt1)-float(CurNVox(cnt1)-1)/float(curNVox(cnt1)))>10.*eps) THEN
	  write(lgu,'(a,i2,a)') 'Warning, the start or end of the grid in the density '//trim(rhofile)//' along dimension',cnt1,&
	                   'does not lie on the borders of the unit cell.'
	ENDIF
      ENDDO
      DO cnt1=1,SDim
        IF (iorien(cnt1)/=cnt1) THEN
	  write(lgu,'(a)') 'Warning, the order of axes in the density'//trim(rhofile)//' is not 123...'
          exit
	ENDIF
      ENDDO

      close(rhu)
      open(rhu,file=rhofile,form='unformatted', access='direct',recl=(max(nxny,35)*ReclFac),ERR=900)

!     read the density
      cnt1=-1
      DO i=1,nmap
        read(rhu,rec=i+1,Err=910) table
	DO ii=1,nxny
          cnt1=cnt1+1
          inrho(cnt1)=table(ii)*CurCApVx
	END DO
      END DO
      CLOSE(rhu,ERR=920)
      GOTO 999

!       Error handling      
900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadJana:'
      WRITE (lgu,'(A)') 'Error! Cannot open density file for reading!'
      WRITE (lgu,'(A)') 'Input file: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during reading of the reference density. Check the log file.')
      
910   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadJana:'
      WRITE (lgu,'(A)') 'Error! Cannot read from the density file.'
      WRITE (lgu,'(A)') 'Input file: '//trim(rhofile)
      WRITE (lgu,'(A)') ''
      CLOSE(rhu)
      CALL StopProgram('Error during reading of the reference density. Check the log file.')

920   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadJana:'
      WRITE (lgu,'(A)') 'Error! Cannot close the density file.'
      WRITE (lgu,'(A)') 'Input file: '//trim(rhofile)
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during reading of the reference density. Check the log file.')


999   CONTINUE
      DEALLOCATE(table)
      end subroutine ReadJana

!Read the reference density in JANA format
!****************************************
      Subroutine NVoxFromJana(rhofile, InNVox)
!****************************************
      USE SF_Module
      IMPLICIT NONE
      CHARACTER*(*) :: rhofile
      INTEGER, DIMENSION(1:SDim) :: InNVox
      INTEGER :: NVoxtmp(1:6), i
      INTEGER :: GetRecLFac !Function returning the length of a unit record

      RecLFac=GetReclFac()
      open(rhu,file=rhofile,form='unformatted', access='direct',status='OLD',recl=35*ReclFac,ERR=900)

!     Read the information from the header
!      write(*,*) 'Prvni rtadek'
      READ (rhu,rec=1,ERR=900) (NVoxtmp(i),i=1,6)
      InNVox=NVoxTmp(1:SDim)
      CLOSE(rhu)
      RETURN

900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadJana:'
      WRITE (lgu,'(A)') 'Error! Cannot open or read density file!'
      WRITE (lgu,'(A)') 'Input file: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during reading of the reference density. Check the log file.')
            
      END SUBROUTINE NVoxFromJana


!        Write electron density in XPlor format
!******************************************************************
      SUBROUTINE WriteXPlor (inrho,rhofile)
!******************************************************************
      
      USE SF_Module
      IMPLICIT NONE
      real, target, dimension(0:) :: inrho
      CHARACTER*(*) :: rhofile  ! Filename
      REAL :: average
      INTEGER :: Loc,nmap,ntable

      OPEN (rhu, ERR=900, FILE=rhofile)
!       Header of file
!       Empty line
      WRITE (rhu, '(a)', ERR=910) ' '
!       number of title lines
      WRITE (rhu, '(i7)', ERR=910) 1
!       title
      WRITE (rhu, '(a)', ERR=910) trim(title)
! Special handling of 2D cases, to make them accessible by normal 3D viewers
      IF ((RDim==2).and.(RDIM==SDim)) THEN
!       Voxel spacing
        WRITE (rhu, '(9I8)', ERR=910) NVox(1), 0, NVox(1)-1, NVox(2), 0, NVox(2)-1, 1, 0, 0
!       Cellparameters
        WRITE (rhu, '(6F12.4)', ERR=910) Cell(1),Cell(2),10.,Cell(3), 90.0, 90.0
      ELSE
!       Voxel spacing
        DO Cnt1=1, SDim
          WRITE (rhu, '(3I8)', ADVANCE='NO', ERR=910) NVox(Cnt1), 0, NVox(Cnt1)-1
        END DO
        WRITE (rhu, '(a)', ERR=910) ''
!       Cellparameters
        DO Cnt1=1, NCellPar
          WRITE (rhu, '(F12.4)', ADVANCE='NO', ERR=910) Cell(Cnt1)
        END DO
        WRITE (rhu, '(a)', ERR=910) ''
       ENDIF
!      String ZYX defining the order of axes
      WRITE (rhu, '(a)', ERR=910) 'ZYX'
!      body of the file
      loc=0
      average=0.0E0
      IF (SDim>1) THEN
        ntable=NVox(1)*NVox(2)
      ELSE
        ntable=NVox(1)
      ENDIF
      nmap=NVoxtot/ntable
      DO cnt1=0,nmap-1
        WRITE (rhu, '(i7)', ERR=910) cnt1
        WRITE (rhu, '(6e12.5)', ERR=910) inrho(loc:loc+ntable-1)/CapVx
        loc=loc+ntable
      END DO
      average=sum(inrho)/CapVx
!       footer, number -9999
      WRITE (rhu, '(a)', ERR=910) '-9999'
!       average and sigma(average)
      average=sum(inrho)/float(NVoxTot)      
      WRITE (rhu, '(2e12.5)', ERR=910) average/CApVx, sqrt(sum((inrho-average)**2)/NVoxTot)/CApVx
                            
      CLOSE (rhu, ERR=920)
      WRITE(lgu,'(A)') 'Electron density written to file '//trim(rhofile)
      write(lgu,'('''')')
      GOTO 999

!       Error handling      
900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine WriteXPlor:'
      WRITE (lgu,'(A)') 'Error! Cannot open density file for writing!'
      WRITE (lgu,'(A)') 'Output file: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during writing of the density. Check the log file.')
      
910   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine WriteXPlor:'
      WRITE (lgu,'(A)') 'Error! Cannot write into density file.'
      WRITE (lgu,'(A)') 'Output file: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Something may have been saved. Stopping.'
      WRITE (lgu,'(A)') ''
      CLOSE(rhu)
      CALL stopProgram('Error during writing of the density. Check the log file.')

920   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine WriteXPlor:'
      WRITE (lgu,'(A)') 'Error! Cannot close density file.'
      WRITE (lgu,'(A)') 'Output file: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Density might have been saved. Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during writing of the density. Check the log file.')

999   CONTINUE
      END SUBROUTINE WriteXPlor

!        Read reflections from xplpor-like format
!******************************************************************
      SUBROUTINE ReadhklfromXPlor(nrefl)
!******************************************************************
      USE SF_Module
      USE Input_module
      USE sf_rhoiointerfaces
      IMPLICIT NONE
      integer :: nrefl, i,cnt,Loc,ErrCode
      integer, dimension(SDim) :: NVoxhkl,NVoxstart,StartC,EndC,CurC, AxOrd
      integer, dimension(3*SDim) :: NVoxtmp
      character(LEN=linelength) :: rhotype
      logical :: havezero
       AxOrd=(/(i,i=1,SDim)/)
      !If nrefl=0, read the file, count and return the number of reflections, otherwise use previously read hklrho to extract the hkl list
      IF (nrefl==0) THEN
        NVoxhkl=1
        rhotype='xplor'
        CALL NVoxFromFile(externreflfile,rhotype,NVoxhkl)
        ALLOCATE(hklrho(0:product(NVoxhkl)-1),STAT=ErrCode)
        IF (ErrCode/=0) CALL StopProgram('Error allocating array hklrho! Too little memory?')
        CALL ReadXplor(hklrho,externreflfile,.true.)
        DO i=0,product(NVoxhkl)-1
          IF (hklrho(i)>=0.0E0) nrefl=nrefl+1
        ENDDO
      ELSE
        OPEN (rhu, FILE=externreflfile,STATUS='OLD')
        DO i=1,2
          READ (rhu, '(a)') Buffer
        ENDDO
        CALL OneInt(Buffer,Loc,ErrCode)
        DO cnt1=1,Loc
          READ(rhu,'(a)')
        ENDDO
!         Voxel spacing
        READ(rhu,'(a)') Buffer
        CALL ArrayInt(Buffer,NVoxTmp,3*SDim,ErrCode)
        DO i=1,SDim
          NVoxStart(i)=NVoxTmp(3*i-1)
          NVoxHKL(i)=NVoxTmp(3*i-2)
        ENDDO
        CLOSE(rhu)
	    StartC=NVoxStart
        EndC=NVoxStart+NVoxHKL-1
        CurC=StartC
        i=-1
        cnt=0
        havezero=.false.
        DO
          i=i+1
          IF (all(CurC==0)) THEN !make sure the zero reflection is the first
            loc=1
            havezero=.true.
          ELSEIF (hklrho(i)>=0.0) THEN
            cnt=cnt+1
            loc=cnt
          ENDIF
          IF ((all(CurC==0)).or.(hklrho(i)>=0.0)) THEN
            Refl(loc)%F=hklrho(i)
            Refl(loc)%A=Refl(loc)%F
            Refl(loc)%B=0.
            Refl(loc)%hklFG=CurC
            Refl(loc)%code=code_default
            Refl(loc)%group=0
          ENDIF
          IF (.not.NextCoo(SDim,StartC,EndC,CurC,AxOrd)) EXIT
	    ENDDO
        IF (.not.havezero) THEN
          Refl(1)%F=0.
          Refl(1)%A=0.
          Refl(1)%B=0.
          Refl(1)%hklFG=0
          Refl(1)%code=code_default
          Refl(1)%group=0
        ENDIF
        DEALLOCATE(hklrho)
      ENDIF
      END SUBROUTINE ReadHKLFromXplor


!        Read electron density in XPlor format
!******************************************************************
      SUBROUTINE ReadXPlor(inrho,RhoFile,gettinghkl)
!******************************************************************
      USE SF_Module
      USE Input_module
      IMPLICIT NONE
      real, target, dimension(0:) :: inrho      
      CHARACTER*(*) :: rhofile  ! Filename
      character*10 :: zyxstr
      character*8 :: efmt='( e  .0)', ffmt='( f  .0)', curfmt 
      INTEGER, DIMENSION(3*SDim) :: NVoxTmp
      INTEGER, DIMENSION(SDim) :: CurNVox
      REAL, DIMENSION(NCellPar) :: Celltmp
      INTEGER :: Loc,nmap,ntable,linecnt,ErrCode
      REAL :: CurCapvx
      LOGICAL, OPTIONAL :: gettinghkl
      linecnt=0
      OPEN (rhu, ERR=900, FILE=rhofile,STATUS='OLD')
! First of all try to determine the format
      READ (rhu, '(a)', ERR=950)
      READ (rhu, '(a)', ERR=950) Buffer
      CALL OneInt(Buffer,Loc,ErrCode)
      IF (ErrCode/=0) GOTO 950
      DO cnt1=1,Loc+6
        linecnt=linecnt+1
        READ(rhu,'(a)',ERR=950)
      ENDDO
      READ (rhu, '(a)', ERR=950) Buffer
      !Buffer contains the first line of pixels
      IF ((index(Buffer,'E')/=0).or.(index(Buffer,'r')/=0)) THEN
        curfmt=efmt
      ELSE
        curfmt=ffmt      
      ENDIF
      loc=0
      DO cnt1=1,len_trim(Buffer)
        if (Buffer(cnt1:cnt1)=='.') loc=loc+1
      ENDDO
      write(curfmt(2:2),'(i1)') loc
      write(curfmt(4:5),'(i2)') len_trim(buffer)/loc
      REWIND(rhu)
!       Header of file
!       Empty line
      linecnt=linecnt+1
      ErrCode=0
      READ (rhu, '(a)', ERR=950)
!       number of title lines, skip them
      linecnt=linecnt+1
      READ (rhu, '(a)', ERR=950,END=950) Buffer
      CALL OneInt(Buffer,Loc,ErrCode)
      IF (ErrCode/=0) GOTO 950
      DO cnt1=1,Loc
        linecnt=linecnt+1
        READ(rhu,'(a)',ERR=950)
      ENDDO
!       Voxel spacing
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=950) Buffer
      CALL ArrayInt(Buffer,NVoxTmp,3*SDim,ErrCode)
      DO cnt1=1,SDim
        CurNVox(cnt1)=NVoxTmp(3*cnt1-2)
	IF((NVoxtmp(3*cnt1-1)/=0).or.(NVoxtmp(3*cnt1)/=CurNVox(cnt1)-1)) THEN
	  write(lgu,'(a,i2,a)') 'Warning, the start or end of the grid in the density along dimension',cnt1,&
	                   'does not lie on the borders of the unit cell.'
	ENDIF
      ENDDO
      IF (size(inrho)/=PRODUCT(CurNVox)) THEN
        CALL ErrorMsg('Error, incorrectly allocated density array.')
        CALL StopProgram('Internal error of the program. Please inform the autor.')
      ENDIF
      IF (ErrCode/=0) GOTO 950
!       Cellparameters
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=950) Buffer
      IF(gettinghkl) THEN
        CurCapvx=1.
      ELSE
        CurCapvx=Cell(NCellPar+1)/PRODUCT(CurNVox)
        CALL ArrayReal(Buffer,Celltmp,NCellPar,ErrCode)
        IF (ErrCode/=0) GOTO 950
        DO cnt1=1,SDim
          IF (abs(Celltmp(cnt1)-Cell(cnt1))>10.*eps) THEN
	        write(lgu,'(a,i2,a)') 'Warning, cell parameter number',cnt1,' in the density '//&
	                   ' is different from what is given in the input file.'
	      ENDIF
        ENDDO
      ENDIF
!      String ZYX defining the order of axes
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=950) zyxstr
      if (zyxstr/='ZYX') THEN
	write(lgu,'(a,i2,a)') 'Warning, string ZYX expected on line',linecnt,' of the density. Instead found '//zyxstr
      ENDIF
!      body of the file
      IF (SDim>1) THEN
        ntable=CurNVox(1)*CurNVox(2)
      ELSE
        ntable=CurNVox(1)
      ENDIF
      nmap=product(curnvox)/ntable
      loc=0
      DO cnt1=0,nmap-1
        READ (rhu, '(a)', ERR=950) Buffer
        CALL OneInt(Buffer,cnt3,ErrCode)
        IF ((Cnt3/=cnt1).and.(.not.gettinghkl)) THEN
          WRITE(lgu,'(a,i5,a)') 'Error, inconsistency in the density on line',linecnt,'.'
          WRITE(lgu,'(a,i4,a,i4,a)') 'Label of the layer nr.',cnt1,' is',cnt3,'.'
          CALL StopProgram('Error during reading of the reference density. Check the log file.')
        ENDIF
        IF (ErrCode/=0) GOTO 950
        READ (rhu,curfmt, ERR=950) inrho(loc:loc+ntable-1)
        inrho(loc:loc+ntable-1)=inrho(loc:loc+ntable-1)*CurCApVx
        loc=loc+ntable
      END DO
      CLOSE (rhu, ERR=920)
      GOTO 999

!       Error handling      
900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadXPlor:'
      WRITE (lgu,'(A)') 'Error! Cannot open density file for reading!'
      WRITE (lgu,'(A)') 'Input file name: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during reading of the reference density. Check the log file.')
      

920   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadXPlor:'
      WRITE (lgu,'(A)') 'Error! Cannot close density file.'
      WRITE (lgu,'(A)') 'Input file: '//trim(rhofile)
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during reading of the reference density. Check the log file.')

950   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadXPlor:'
      WRITE(lgu,'(a,i5)') 'Error reading reference density at layer',cnt1+1
      CALL StopProgram('Error during reading of the reference density. Check the log file.')

999   CONTINUE
      END SUBROUTINE ReadXPlor

!Get the number of voxels for the Xplor file
!****************************************
      Subroutine NVoxFromXplor(rhofile, InNVox)
!****************************************
      USE SF_Module
      USE Input_module
      IMPLICIT NONE
      CHARACTER*(*) :: rhofile
      INTEGER, DIMENSION(1:SDim) :: InNVox
      INTEGER, DIMENSION(1:3*SDim) :: NVoxtmp
      INTEGER :: linecnt, loc, ErrCode

      OPEN (rhu, ERR=900, FILE=rhofile,STATUS='OLD')
      ErrCode=0
      linecnt=1
!       Header of file
!       Empty line
      READ (rhu, '(a)', ERR=950)
!       number of title lines, skip them
      linecnt=linecnt+1
      READ (rhu, '(a)', ERR=950,END=950) Buffer
      CALL OneInt(Buffer,Loc,ErrCode)
      IF (ErrCode/=0) GOTO 950
      DO cnt1=1,Loc
        linecnt=linecnt+1
        READ(rhu,'(a)',ERR=950)
      ENDDO
!       Voxel spacing
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=950) Buffer
      CALL ArrayInt(Buffer,NVoxTmp,3*SDim,ErrCode)
      IF (ErrCode/=0) GOTO 950
      DO cnt1=1,SDim
        InNVox(cnt1)=NVoxTmp(3*cnt1-2)
      ENDDO
      CLOSE(rhu)
      RETURN

900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadXPlor:'
      WRITE (lgu,'(A)') 'Error! Cannot open density file for reading!'
      WRITE (lgu,'(A)') 'Input file name: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during reading of the reference density. Check the log file.')

950   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadXPlor:'
      WRITE(lgu,'(a,i5)') 'Unexpected format of the file '//trim(rhofile)//'at line',linecnt
      CALL StopProgram('Error during reading of the reference density. Check the log file.')

      END SUBROUTINE NVoxFromXplor
      
!        Write electron density in CCP4 format
!******************************************************************
      SUBROUTINE WriteCCP4 (inrho,rhofile)
!******************************************************************
      
      USE SF_Module
      IMPLICIT NONE
      real, target, dimension(0:) :: inrho
      CHARACTER*(*) :: rhofile  ! Filename
      CHARACTER*80 :: symstring='X,Y,Z   ', label='CCP4 map generated by Superflip'
      REAL(KIND=4) :: average, sigma, zero
      integer :: pos,NVoxTot256, symlen, reclen
      INTEGER(KIND=4), dimension(0:100) :: nrarr
      INTEGER :: i,k, getreclfac, istamp

      istamp=X'4444'
      nrarr(0:100)=(/(k,k=0,100)/)
      average=(sum(rho)/NVoxTot)/CapVx
      sigma=sqrt(sum((rho-average)**2)/NVoxTot)/CapVx
      RecLFac=GetReclFac()
      symlen=8
      reclen=256+symlen/4
      OPEN (rhu, ERR=900, FILE=rhofile, ACCESS='DIRECT', FORM='UNFORMATTED',RECL=reclen*reclfac, STATUS='REPLACE')

!first write the header
      write(rhu,rec=1) (NVox(k),k=1,3),nrarr(2),(nrarr(0),k=1,3),(NVox(k),k=1,3),(Cell(k),k=1,6),(k,k=1,3), &
                       minval(rho)/CapVx,maxval(rho)/CapVx,average,nrarr(1),symlen,(nrarr(0),k=1,28),'MAP ',istamp,sigma,nrarr(1), &
!                       label(1:80),('    ',k=1,180),symstring(1:symlen*4)
                       label(1:80),('    ',k=1,180),'X,Y,Z   '
!and now dump the whole array as it is      
      pos=1
      DO i=0,NVoxTot-1,reclen
  	    pos=pos+1
        write(rhu,rec=pos) inrho(i:min(i+reclen-1,NVoxTot-1))/CApVx
	  ENDDO
      CLOSE(rhu)
      GOTO 999

!       Error handling      
900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine WriteCCP4:'
      WRITE (lgu,'(A)') 'Error! Cannot open density file for writing!'
      WRITE (lgu,'(A)') 'File name: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during writing of the reference density. Check the log file.')

999   CONTINUE      
      END SUBROUTINE WriteCCP4

!        Read electron density in CCP4 format
!******************************************************************
      SUBROUTINE ReadCCP4 (inrho,rhofile)
!******************************************************************
      
      USE SF_Module
      IMPLICIT NONE
      real, target, dimension(0:) :: inrho      
      CHARACTER*(*) :: rhofile  ! Filename
      REAL(KIND=4) :: rdummy, CurCapvx
      integer :: target, idummy, NSymBytes
      integer, dimension(SDim) :: CurNVox, IntBuffer
      integer, dimension(NCellPar) :: CurCellPar
      INTEGER :: i,getreclfac


      RecLFac=GetReclFac()
      OPEN (rhu, ERR=900, FILE=rhofile, ACCESS='DIRECT', FORM='UNFORMATTED',RECL=reclfac, STATUS='OLD')

!first read the number of voxels
      DO i=1,SDim
        read(rhu,rec=i) CurNVox(i)
      ENDDO
      IF (size(inrho)/=PRODUCT(CurNVox)) THEN
        CALL ErrorMsg('Error, incorrectly allocated density array.')
        CALL StopProgram('Internal error of the program. Please inform the autor.')
      ENDIF
      CurCapvx=Cell(NCellPar+1)/PRODUCT(CurNVox)
      READ(rhu,rec=4) idummy
      IF (IDummy/=2) THEN
         write(lgu,'(a,i2,a)') 'Error, the mode of the CCP4 map '//trim(rhofile)//' should be 2, and it is',cnt1,'.'
         CALL StopProgram('Error during reading of the density. Check the log file.')
      ENDIF
!check if the map is in the whole box
      DO i=1,SDim
        read(rhu,rec=i+4) IntBuffer(i)
      ENDDO
      IF (any(IntBuffer/=0)) THEN
         write(lgu,'(a)') 'Error, the CCP4 map '//trim(rhofile)// &
	      ' does not contain the density in the whole unit cell, or the density is translated.'
         CALL StopProgram('Error during reading of the density. Check the log file.')
      ENDIF     
      DO i=1,SDim
        read(rhu,rec=i+7) IntBuffer(i)
      ENDDO
      IF (any(IntBuffer/=CurNVox)) THEN
         write(lgu,'(a)') 'Error, the CCP4 map '//trim(rhofile)// &
	       ' does not contain the density in the whole unit cell, or the density is translated.'
         CALL StopProgram('Error during reading of the density. Check the log file.')
      ENDIF     

      !Finally Check the cell parameters
      DO i=1,NCellPAr
        read(rhu,rec=i+10) CurCellPar(i)
      ENDDO
      IF (any(abs(CurCellPar-Cell)>10.*eps)) THEN
	  write(lgu,'(a)') 'Warning, cell parameters in the density '//trim(rhofile)// &
	       ' are different from what is given in the input file.'
      ENDIF
      READ(rhu,rec=24) NSymBytes !How many bytes used to store the symmetry operations; usually 0
      !And now read the map
      DO i=0,PRODUCT(CurNVox)-1
        read(rhu,rec=256+nsymbytes/4+1+i) rdummy
        inrho(i)=rdummy*CurCapVx
      ENDDO

      CLOSE(rhu)
      GOTO 999

!       Error handling      
900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadCCP4:'
      WRITE (lgu,'(A)') 'Error! Cannot open density file for writing!'
      WRITE (lgu,'(A)') 'Input file name: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during reading of the reference density. Check the log file.')

999   CONTINUE   
      END SUBROUTINE ReadCCP4

!Get the number of voxels from the CCP4 file
!****************************************
      Subroutine NVoxFromCCP4(rhofile, InNVox)
!****************************************
      USE SF_Module
      IMPLICIT NONE
      CHARACTER*(*) :: rhofile
      INTEGER, DIMENSION(1:SDim) :: InNVox
      INTEGER :: GetRecLFac !Function returning the length of a unit record
      INTEGER :: i

      RecLFac=GetReclFac()
      OPEN (rhu, ERR=900, FILE=rhofile, ACCESS='DIRECT', FORM='UNFORMATTED',RECL=reclfac, STATUS='OLD')

!read the number of voxels
      DO i=1,SDim
        read(rhu,rec=i,ERR=900) InNVox(i)
      ENDDO
      CLOSE(rhu)
      RETURN

900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine ReadCCP4:'
      WRITE (lgu,'(A)') 'Error! Cannot open or read density file!'
      WRITE (lgu,'(A)') 'Input file: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during reading of the reference density. Check the log file.')
            
      END SUBROUTINE NVoxFromCCP4
      

!        Make an electron density from a CIF file
!******************************************************************
      SUBROUTINE RhoFromCIF(inrho,RhoFile)
!******************************************************************
      USE Input_Module      
      USE SF_Module 
      USE SF_Composition
      USE FFT_Interfaces
      IMPLICIT NONE
      real, target, dimension(0:) :: inrho      
      CHARACTER :: rhofile*(FileLength)  ! Filename
	CALL ReadCif(RhoFile)
	StrSymmetry=.true. !Means the structure contains only symmetry-independent part
	CALL SFFromStructure
	CALL FFT(sf2rho,rho)
        inrho=rho
        IF (TotAtoms>0) DEALLOCATE(Atoms)
      END SUBROUTINE RhoFromCIF
      

!        Write the list of reflections in m80 format
!******************************************************************
      SUBROUTINE WriteM80(rhofile)
!******************************************************************
      
      USE SF_Module
      IMPLICIT NONE
      CHARACTER :: rhofile*(FileLength)  ! Filename
      CHARACTER*(14) :: m80fmt='(..i4,13e12.5)'
      REAL :: a,b,f
      INTEGER ::ii 			!Function returning the length of the string
!!!!!!! Important: The convention in FFTW is OPOSITE to that in crystallography.
!!!!!!! Therefore internally the B component or Refl must be inverted on input and inverted again in poutput
!!!!!!! Unpleasant, but necessary	    
      write(m80fmt(2:3),'(i2)') SDim+1
      OPEN (rhu, ERR=900, FILE=rhofile)
      IF (averagesym==NoAveraging) THEN !Reflections not averaged - we have to write out all reflections
        DO cnt1=MStartF,MEndF
          a=MEMRefl(cnt1)%AMEM
          b=-MEMRefl(cnt1)%BMEM
	  f=MEMRefl(cnt1)%FMEM
	  write(rhu,m80fmt) (MEMRefl(cnt1)%hklFG(ii),ii=1,SDim),1,f,f,f,a,b,(0.,ii=1,8)
	ENDDO
      ELSE !Only symmetry-indepenent reflections have to be written out
        f=MEMRefl(MStartF)%FMEM !Write the F(0) from the MEM list, it is not guaranteed to be in the Refl lList
        write(rhu,m80fmt) (MEMRefl(mstartf)%hklFG(ii),ii=1,SDim),1,f,f,f,f,0.,(0.,ii=1,8)
        DO cnt1=StartF,EndF
          IF (Refl(cnt1)%sfloc==0) CYCLE !Skip the F(0) reflection
          a=Refl(cnt1)%AMEM
          b=-Refl(cnt1)%BMEM
          f=Refl(cnt1)%FMEM
          write(rhu,m80fmt) (Refl(cnt1)%hklFG(ii),ii=1,SDim),1,f,f,f,a,b,(0.,ii=1,8)
	ENDDO
      ENDIF
      CLOSE(rhu)

      RETURN

900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine Writem80:'
      WRITE (lgu,'(A)') 'Error! Cannot open density file for writing!'
      WRITE (lgu,'(A)') 'File name: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during writing of the density. Check the log file.')
      
      END SUBROUTINE WriteM80

!        Write the list of reflections in m80 format
!******************************************************************
      SUBROUTINE WriteShelxSym(rhofile)
!******************************************************************
      USE SF_Module
      IMPLICIT NONE
      CHARACTER :: rhofile*(FileLength)  ! Filename
      CHARACTER*(LineLength) :: ModSymStr !Symmetry string modified for CIF or INS purposes
      LOGICAL :: succ
      REAL :: suma
      INTEGER, DIMENSION(SDim,SDim) :: iDummyMat
      REAL, DIMENSION(SDim) :: DummyVec,DummyVec2
      INTEGER :: LattNr,isum,iabssum,InvOp,i,ii,iii

      OPEN (rhu, ERR=900, FILE=rhofile)
      write(rhu,'(a)') 'TITL '//trim(title)//' - space group proposed by Superflip: '//trim(sgsymbol)
      write(rhu,'(a,f7.5,6f9.4)') 'CELL ',0.71073,(Cell(ii),ii=1,6)
      !Find the proper lattice centering:
      !First make a checksum and control that only 1/2 and 1/3 or 2/3 are present in the centings
      suma=0.0;succ=.true.
      DO i=1,NCenter
 	DO ii=1,3
 	  suma=suma+float(2**(ii-1))*Center(ii,i) 	  
 	  if ((abs(center(ii,i)-0.0E0)>eps).and.(abs(center(ii,i)-0.5E0)>eps).and. &
              (abs(center(ii,i)-0.3333333333E0)>eps).and.(abs(center(ii,i)-0.666666666E0)>eps)) succ=.false. 
 	ENDDO
      ENDDO
      IF (.not.succ) suma=-100.
      IF (NCenter==1) THEN
 	LattNr=-1
      ELSEIF (Ncenter==2) THEN !The only allowed (and possible) is A, B, C and I
 	IF (abs(suma-3.5)<eps) THEN
          LattNr=-2 !I-Centering
 	ELSEIF (abs(suma-3.0)<eps) THEN
          LattNr=-5 !A-centering
 	ELSEIF (abs(suma-2.5)<eps) THEN
          LattNr=-6 !B-centering
 	ELSEIF (abs(suma-1.5)<eps) THEN
          LattNr=-7 !C-centering
 	ELSE
 	  succ=.false.
 	ENDIF
      ELSEIF (Ncenter==3) THEN !The only allowed is R obverse
 	IF (abs(Center(2,2)-Center(3,2))<eps) THEN
          LattNr=-3 !R-centering obverse
 	ELSE
 	  succ=.false.
 	ENDIF
      ELSEIF (Ncenter==4) THEN ! The only allowed is F
 	IF (abs(suma-7)<eps) THEN
          LattNr=-4 !F-centering
 	ELSE
 	  succ=.false.
 	ENDIF
      ENDIF
      IF(.not.succ) THEN
        write(*,'(a)') 'Sorry, SHELX format allows only the standard centerings P, I, F, A, B, C or  R obverse.' 
        write(*,'(a)') 'Your centering is none of them. The lattice will be set to P.'
 	LattNr=-1
      ENDIF
      !Check if we have an inversion center
      DO i=1,NSymPrim
 	isum=0;iabssum=0;InvOp=0
 	DO ii=1,3
 	  DO iii=1,3
 	    isum=isum+SmOp(i)%R(ii,iii)
 	  ENDDO
          iabssum=iabssum+abs(SmOp(i)%R(ii,ii))
        ENDDO
        IF ((isum==-3.).and.(iabssum)==3) THEN !Centrosymmetric
 	  !SHELX accepts inversion only at the origin
 	  suma=sum(abs(SmOp(i)%t))
          IF (suma>eps) THEN !It is centrosymmetric, but the inversion center does not lie in the origin -> like if it were not centro...
!            write(*,'(a)') 'Sorry, SHELX requires the inversion center to lie at the origin.' 
!            write(*,'(a)') 'Your inversion center does not.' 
            InvOp=0
          ELSE
 	    InvOp=i
            LattNr=-LattNr	  
 	  ENDIF
 	  EXIT
 	ENDIF
      ENDDO
      write(rhu,'(a,i2)') 'LATT ',LattNr
      !Write out the symmetry operations. Without inversion-center related!
      DO i=2,NSymPrim
        succ=.true.
 	IF (InvOp>0) THEN !Check, if this one is not related to some of the previous
          DO ii=1,i-1
 	    CALL ProdOp(SmOp(ii)%R,SmOp(ii)%t,SmOp(InvOp)%R,SmOp(InvOp)%t,iDummyMat,DummyVec,SDim)		    
 	    DO iii=1,ncenter
	      dummyvec2=dummyvec+Center(1:SDim,iii)
	      IF (EqualOp(SmOp(i)%R,SmOp(i)%t,iDummyMat,DummyVec2,SDim)) succ=.false.
            ENDDO
 	  ENDDO
        ENDIF
 	IF (succ) THEN
 	  ModSymStr=SymOpIntTab(SmOp(i),'xyz')
          CALL RemoveSpaces(ModSymStr)
	  write(rhu,'(a)') 'SYMM '//trim(ModSymStr)
 	ENDIF
      ENDDO
      write(rhu,'(a)') 'END'
      RETURN
      
900   WRITE (lgu,'(A)') ''
      WRITE (lgu,'(A)') 'Routine WriteShelxSym:'
      WRITE (lgu,'(A)') 'Error! Cannot open file for reading!'
      WRITE (lgu,'(A)') 'Output file name: '//trim(rhofile)
      WRITE (lgu,'(A)') 'Stopping.'
      WRITE (lgu,'(A)') ''
      CALL StopProgram('Error during writing the Shelx file. Check the log file.')
      END SUBROUTINE WriteShelxSym


      subroutine TimeStamp(InText)
      USE SF_Module
      IMPLICIT NONE
      character*(*) :: InText
      character*60 :: DatTimForm
      integer :: MonthIndex
      CHARACTER, SAVE :: Date*8, Time*10
      CHARACTER*3,DIMENSION(1:12) :: Months      
!      CHARACTER, SAVE :: Year*4, Month*2, Day*2, Hour*2, Minutes*2, Seconds*2
      DATA Months /'JAN','FEB','MAR','APR','MAY','JUN','JUL','AUG','SEP','OCT','NOV','DEC'/
      CALL DATE_AND_TIME (Date, Time)
      READ(Date(5:6),'(i2)') MonthIndex
!      Year=Date(1:4)
!      Month=Date(5:6)
!      Day=Date(7:8)
!      Hour=Time(1:2)
!      Minutes=Time(3:4)
!      Seconds=Time(5:6)
      DatTimform='(a,'' '',a2,''.'',a3,'' '',a4,'', '',a2,'':'',a2,'':'',a2)'
      Write(Buffer,DatTimForm) trim(InText),Date(7:8),Months(MonthIndex),Date(1:4),Time(1:2),Time(3:4),Time(5:6)
      CALL FramedLabel(2,Buffer)
      if (terminal) write(*,'(a)') trim(Buffer)
      end subroutine TimeStamp

      subroutine VersionStamp(bigstamp)
      USE SF_Module
      IMPLICIT NONE
      LOGICAL :: BigStamp
        IF (BigStamp) THEN
          write(lgu,'(a)') '**************************************************************'
          write(lgu,'(a)') '                           Superflip'
          write(lgu,'(a)') '                    Version: '//trim(versionstring)
	  write(lgu,'(a)') 'Palatinus, L. & Chapuis, G.(2007): J. Appl. Cryst. 40, 786-790'
	  write(lgu,'(a)') '              http://superspace.epfl.ch/superflip'
          write(lgu,'(a)') '**************************************************************'
	  write(lgu,*)
	ELSE
          write(lgu,'(a)') 'Superflip version: '//trim(versionstring)
	  IF (terminal) THEN
	    write(*,'(a)') 'Superflip version: '//trim(versionstring)
	    write(*,*)
          ENDIF
	ENDIF
      end subroutine VersionStamp
