!Library contains subroutines, which serve for spline interpolation of the electron density
!5.7.2002+ Lukas Palatinus

      SUBROUTINE GetPeak(Centr, ExPeak,outrho) !Gets coordinates of the pixel, returns accurate coordinates and density of the maximum
      USE SF_Module
      IMPLICIT NONE
      INTEGER :: Centr(SDim) ! Central voxel of the maximum, dummy coords
      INTEGER :: AxOrd(SDim),DerOrd(SDim),Orig(SDim)
      real :: ExPeak (SDim) ! The exact(er) peak position
      real :: outrho !peak electron density
      real, DIMENSION(range) :: y,y2 !arrays for subroutine splines
      INTEGER :: ii ! Counters
      LOGICAL :: Succ !flag of success

      ExPeak=float(Centr)
      call CPRefine(ExPeak,Succ) !returns the coordinates of the maximum, succ is false, if no maximum could be found
      IF (Succ) THEN
        DerOrd=0
        DO ii=1,SDim
          AxOrd(ii)=ii
          Orig(ii)=int(ExPeak(ii)-float(range)/2.+range)+1-range
        END DO
        call FillRhoCut(AxOrd,Orig) !fill the part of the electron density around the maximum
        call RhoAtPoint(SDim,range,DerOrd,AxOrd,ExPeak,orig,srho,outrho,y,y2) !returns the peak density
      ELSE
        ExPeak=-1.0E30
	outrho=-1.0E30
      END IF
      END SUBROUTINE GetPeak

      SUBROUTINE CPRefine(pt,succ) ! find stationary point (derivatives=0)
      USE SF_Module
        
	IMPLICIT NONE
        INTEGER, PARAMETER :: ptchangelim=20
        real, PARAMETER :: MaxGradShift=0.5
        real, PARAMETER :: MaxPtShift=1.
	real :: pt(SDim),ptor(SDim) !point, where the density is to be determined
			! Old pt to calculate discrepancy, pt to define origin of the hypercube
			!PtOr was added for the cases pt is close to the pixel border. In that case, changing origin can cause oscillations.
        INTEGER, DIMENSION(SDim) :: AxOrd !order of the axes in the output rcut
        INTEGER, DIMENSION(SDim) :: DerOrd,Orig
        real :: rout,disc, Grad, GradOld, GradOlder, red, dump !output rho, 1st and 2nd solution of the qadr. equation, discrepancy
        real, DIMENSION(range) :: y,y2
        real, ALLOCATABLE :: H(:,:),Hinv(:,:),Gr(:), Sh(:) !Hessian, its inversion and gradient
	LOGICAL :: succ !success
        INTEGER :: i,ii,iii, iv, ptchange !counters

        ALLOCATE(H(SDim,SDim),STAT=AllocError)
        ALLOCATE(Hinv(SDim,SDim),STAT=AllocError)
        ALLOCATE(Gr(SDim),STAT=AllocError)
        ALLOCATE(Sh(SDim),STAT=AllocError)
        IF (AllocError.GT.0) STOP 'Error allocating in H, Hinv and Gr in CpRefine! Too little memory?'
        do i=1,SDim
          Orig(i)=int(pt(i)-float(range)/2.+range)+1-range !set the origin of the region +- range/2 around pt
          DerOrd(i)=0
        end do
	ptchange=0 !counts the number of cycles
        dump=1.
        DO ii=1,SDim
          AxOrd(ii)=ii
        END DO
        ptor=pt
	gradold=0.0; gradolder=0.0
        DO
          ptchange=ptchange+1
	  succ=.true.
	  IF (ptchange>ptchangelim) THEN !trying too long time
	    Succ=.false.
	    EXIT
	  END IF
          call FillRhoCut(AxOrd,Orig) !fill the area of rho with properly permuted axes
!********* Gradient and Hessian
          DO iii=1,SDim
	    DerOrd=0
	    DerOrd(iii)=1
            call RhoAtPoint(SDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
            Gr(iii)=rout
          END DO
	  DO iii=1,SDim
	    DO iv=1,SDim
              DerOrd=0
	      DerOrd(iii)=DerOrd(iii)+1
	      DerOrd(iv)=DerOrd(iv)+1
              call RhoAtPoint(SDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
              H(iii,iv)=rout	      
	    END DO
	  END DO
          Hinv=invert(H)
          IF (Hinv(1,1)>1.0E29) THEN
            succ=.false.
	    EXIT
	  END IF
	  Sh=0.0D0
          Grad=0.0D0
	  DO iii=1,SDim
	    DO iv=1,SDim
	      Sh(iii)=Sh(iii)+Gr(iv)*Hinv(iv,iii)
	    END DO
            Grad=Grad+Sh(iii)**2
	  END DO
	  Grad=sqrt(Grad)
          IF (Grad>MaxGradShift) THEN
	    Red=MaxGradShift/Grad
	  ELSE
	    Red=1.0
	  END IF
          disc=0.0D0
	  DO iii=1,SDim
	    pt(iii)=pt(iii)-sh(iii)*red*dump
            disc=disc+(pt(iii)-ptor(iii))**2
	  END DO
	  disc=sqrt(disc)
	  IF (disc>MaxPtShift) Succ=.false.
	  IF (ptchange>2) THEN
	    if ((gradold-grad)>(gradolder-gradold)) dump=dump*0.5
	  END IF
          gradolder=gradold
	  gradold=grad
	  IF (.not.Succ.or.grad<eps) EXIT !loop until .not.succ or disc is sufficiently small
	END DO
      END SUBROUTINE CPRefine
      

      SUBROUTINE FillRhoCut(AxOrd,Org) !fills part of the electron density from -range/2 to +range/2
        USE SF_Module
        IMPLICIT NONE
	INTEGER :: ArrLoc
        INTEGER :: Dxyz(SDim),EndC(SDim)     ! output from subrt. Coordinates, input to ArrayLocation
        INTEGER, DIMENSION(*) :: AxOrd,Org !order of the axes in the output rcut, origin in the rho
	
	DO Cnt1=1,SDim !set start and end of the cube
          Dxyz(Cnt1)=Org(Cnt1)
          EndC(AxOrd(Cnt1))=Org(AxOrd(Cnt1))+range-1
	END DO
        Cnt1=0
        DO
	  Call ArrayLocation(Dxyz,ArrLoc) !get the location of this pixel in rho
          srho(Cnt1)=rho(ArrLoc)
          cnt1=cnt1+1
          IF (.not.NextCoo(SDim,Org,EndC,Dxyz,AxOrd)) EXIT
        END DO
      END SUBROUTINE FillRhoCut


      SUBROUTINE RhoAtPoint(Dim,range,DerOrd,AxOrd,pt,orig,inrho,rout,y,y2) !returns electron density at pt and parameters y,y2 od the spline interpolation
        IMPLICIT NONE
	INTEGER :: i,Pcnt,SCnt ! counters
        INTEGER :: Dim,ADim !Dimension of the rhocut, actual dimension during the spline interpolation
        INTEGER :: range !length of the vertices of the "hypercube"
        real,DIMENSION(*) :: pt !point, where the density is to be determined
        INTEGER, DIMENSION(*) :: Orig !origin of the rhocut
        real :: inrho(0:range**Dim-1) !Input array containing the cut of the density around pt
        real :: rcut(0:range**Dim-1) !Working array
        real,DIMENSION(range) :: x,y,y2 !arrays for the routines spline and splint
        INTEGER, DIMENSION(*) :: AxOrd !order of the axes in the output rcut
        INTEGER, DIMENSION(*) :: DerOrd  ! Array defining the derivative of the function returned by RhoAtPoint
        real, DIMENSION(0:2) :: rval ! Density and its derivatives at the point pt, output
        real :: rout !output

        rcut=inrho
        DO Adim=Dim,1,-1 !Going from Dim to 1, at each cycle interpolating Adim-1 dimensional section of rcut
          Pcnt=1 !counter of the old rho
	  Scnt=0 !counter of the new, Adim-1 dimensional rho
	  DO i=0,range**ADim-1 ! repeat over all pixels of ald rcut
            x(Pcnt)=float(Orig(AxOrd(Dim-Adim+1))+Pcnt-1)
	    y(Pcnt)=rcut(i) !fill one row
            PCnt=PCnt+1
	    IF (PCnt==range+1) THEN !if the row is full, interpolate the value at pt
	      PCnt=1
	      CALL spline(x,y,range,y2)
	      CALL splint(x,y,y2,range,pt(AxOrd(Dim-Adim+1)),rval) !rval(0) contains the density, rval(1) the first derivative and rval(2) the second derivative
              rcut(Scnt)=rval(DerOrd(AxOrd(Dim-Adim+1))) !assign the result to rho(Scnt).
	      						!It is perfectly safe to save to the same rho, that point will not be needed in this cycle anymore.
	      SCnt=SCnt+1
	    END IF
!            continue to the next row
          END DO
          ! the ADim-1 dimensional "first slice" of rcut now contains the interpolated section and we can continu with next step
        END DO
	rout=rcut(0)
      END SUBROUTINE RhoAtPoint
    

      SUBROUTINE Spline(x,y,n,y2) !copied from Numerical Recipies
      INTEGER :: i,k,n
      real :: x(n),y(n),y2(n)
      real :: p,sig,u(n)
      u(1)=0.
      y2(1)=0.
      do i=2,n-1
        sig=(x(i)-x(i-1))/(x(i+1)-x(i-1))
        p=sig*y2(i-1)+2.
        y2(i)=(sig-1.)/p
        u(i)=(6.*((y(i+1)-y(i))/(x(i+1)-x(i))-(y(i)-y(i-1))/(x(i)-x(i-1)))/(x(i+1)-x(i-1))-sig*u(i-1))/p
      end do
      y2(n)=0.
      do k=n-1,1,-1
        y2(k)=y2(k)*y2(k+1)+u(k)
      end do
      return
      END SUBROUTINE Spline

      SUBROUTINE splint(xa,ya,y2a,n,x,y) !copied from Num.Rec., strongly simplified and modified
      INTEGER :: n,khi,klo
      real :: x,xa(n),y2a(n),ya(n)
      real, DIMENSION(0:2) :: y ! value and its derivatives
      real :: a,b
      
!     x lies in our case always exactly in the middle of the range ->
!     no need to bisect, we can calculate klo and khi directly
      klo=int(x-xa(1)+float(n))-n+1
      khi=klo+1
!     moreover, the spacing between the x-values is always 1 -> we can forget h; h=1
      a=xa(khi)-x
      b=x-xa(klo)
!     value
      y(0)=a*ya(klo)+b*ya(khi)+((a**3-a)*y2a(klo)+(b**3-b)*y2a(khi))/6.
!     the first derivative
      y(1)=(ya(khi)-ya(klo))+((1-3*a**2)*y2a(klo)+(3*b**2-1)*y2a(khi))/6.
!     the second derivative
      y(2)=a*y2a(klo)+b*y2a(khi)
      return
      END

      SUBROUTINE splint_full(xa,ya,y2a,n,x,y)
        INTEGER n
        REAL x,y,xa(n),y2a(n),ya(n)
        INTEGER k,khi,klo
        REAL a,b,h
        klo=1 !We will find the right place in the table by means of bisection.
              !This is optimal if sequential calls to this routine are at random
              !values of x. If sequential calls are in order, and closely
              !spaced, one would do better to store previous values of
              !klo and khi and test if they remain appropriate on the
              !next call.
        khi=n
     1  if (khi-klo.gt.1) then
          k=(khi+klo)/2
          if(xa(k).gt.x)then
            khi=k
          else
            klo=k
          endif
          goto 1
        endif !klo and khi now bracket the input value of x.
        h=xa(khi)-xa(klo)
        if (h.eq.0.) STOP 'Problem in spline interpolation!' !bad xa input in splint The xas must be distinct.
        a=(xa(khi)-x)/h !Cubic spline polynomial is now evaluated.
        b=(x-xa(klo))/h
        y=a*ya(klo)+b*ya(khi)+((a**3-a)*y2a(klo)+(b**3-b)*y2a(khi))*(h**2)/6.
        return
      END
