! Central subroutine that governs any operation on the density after the iteration stopped
! Basic assumption is that the rho, MEMRefl and Refl arrays are all consistent
!*******************************************    
      SUBROUTINE HandleSymmetry
!*******************************************    
      USE SF_Module
      USE FFT_Interfaces
      IMPLICIT NONE
        character*14 ::  VoxForm='(a,..i6,a,i10)'
        write(VoxForm(4:5),'(i2)') SDim

        IF (RefRhoType/='none') THEN
          CALL MatchReference(1,2) !Shift the phases of the MEMRefl to match the referencefile as closely as possible
	ELSEIF ((WantSumRho/=0).and.(NSummed>0)) THEN
          CALL MatchReference(3,2) !Shift the phases of the MEMRefl to match the referencefile as closely as possible
	ENDIF
	IF (derivesym/='no') THEN
          CALL FramedLabel(1,'Checking the density for symmetry')
          if (terminal) THEN
	    write(*,*)
	    write(*,'(a)') 'Checking the density for symmetry:'            
	    write(*,*)
	  ENDIF
	  call DeriveSymmetry
	  IF ((derivesym=='use').and.(autofinevoxels)) THEN !Check the NVox spacing to be consistent with the new symmetry
            CALL GetAutoVoxels(NVoxFine,FineGridResolution, FineShannonFactor)
            IF (any(NVox/=NVoxFine)) THEN
	      CALL ChangeGrid(NVoxFine)
	      write(lgu,'(a)') 'Current voxel spacing is incompatible with the derived symmetry.'
              write(lgu,voxform)  'Resampling to new voxel spacing:     ',(NVoxFine(cnt1),cnt1=1,SDim),'   Total:',product(NVoxFine)
	    ENDIF
	  ENDIF
	ENDIF
	IF (searchSym.or.(averagesym/=NoAveraging)) then
	  IF (searchsym) THEN
            CALL FramedLabel(1,'Search for the origin of the space group')
            if (terminal) write(*,'(a)') 'Searching for the origin of the space group:'
          ELSE
            CALL FramedLabel(1,'Averaging the density over symmetry')
            if (terminal) write(*,'(a)') 'Averaging the density over symmetry:'
	  ENDIF
          CALL LocateSymmetryPatt(SearchSymKey,AverageSym,2)
	endif
	IF (NTestSym>0) THEN
          CALL FramedLabel(1,'Testing for presence of the test symmetry operations')
          if (terminal) write(*,'(a)') 'Testing for presence of the test symetry operations:'
          CALL LocateSymmetryPatt(TestSymKey,NoAveraging,2)
	ENDIF
! All operations on the density were performed in the reflection space -> calculate rho form SF
        CALL FFT(sf2rho)
	IF (ViewProgress/=noviewer) CALL WriteForViewer
      END SUBROUTINE HandleSymmetry
      
      ! Find the position of origin by evaluating the autocorrelation function ("Patterson")
! If Key=SearchSymKey the origin i searched, for key=TestSymKey only the test symmetry operations are evaluated
!************************************************
      SUBROUTINE LocateSymmetryPatt(key, aveflag, verbose)
!************************************************
      USE SF_Module 
      USE Input_module
      USE SF_SymSearch
      USE SF_Interfaces
      
      IMPLICIT NONE
      REAL, PARAMETER :: difflimit=0.7E0
      INTEGER :: verbose !0=absolutely no output, 1 - only overall sym agr fac, 2 - everything
      INTEGER :: key !If key=SearchSym the origin is searched, for key=TestSym only the presence of a symmetry operation is evaluated
      INTEGER :: aveflag
!     Variables for sgelsy
      REAL, ALLOCATABLE, DIMENSION (:,:) :: DummyMat,DummyMat2,AugMAtrix
      REAL, ALLOCATABLE, DIMENSION (:) :: augrhs,daugrhs,daugrhs2
      type(SymOpPropsType) :: TmpSymProps
      INTEGER :: loopnr, enant=1
!     Variables holding the multiple possible solutions
      integer, allocatable, dimension(:) :: genpos, gencnt !Position of the first element belonging to the generator in the rhslist and the number of rhs for that generator
      integer :: totalrhs
      real, allocatable, dimension(:,:) :: rhslist, drhslist !list of right-hand sieds and a dummy list ussed for reallocatin the right list
  
      INTEGER, ALLOCATABLE, DIMENSION(:) :: StartC, EndC, AxOrd,IndexOrder

      INTEGER :: i,ii,iv,arrloc,twotosdim,n,cursymop,symoploop
      REAL, DIMENSION(SDim) :: dhkl,dhkl2,patpeak,rhs,Cooshift,CooShiftEnant,minCooShift
      REAL :: rdummy, rdummy2,mintotshift,coinc,coincenant,coincidencers,mindiff
      LOGICAL :: GoodTrans, CheckLattTrans

      IF ((NSymPrim==1).and.(key==SearchSymKey)) THEN !only trivial symmetry, no averaging possible
        IF (verbose>1) CALL ErrorMsg('Only trivial symmery is present, no symmetry search can be performed.')
        RETURN
      ENDIF
      twotosdim=2**SDim
      ALLOCATE(P(1:SDim,1:twotosdim),STAT=ArrLoc)
      ALLOCATE(HP(1:SDim,1:twotosdim),STAT=ArrLoc)
      ALLOCATE(IR(1:SDim,1:SDim),STAT=ArrLoc)
      ALLOCATE(DummyMat(1:SDim,1:SDim),STAT=ArrLoc)
      ALLOCATE(RIR(1:SDim,1:SDim),STAT=ArrLoc)
      ALLOCATE(ZeroArr(1:SDim),STAT=ArrLoc)
      ALLOCATE(genpos(1:NGen),STAT=ArrLoc)
      ALLOCATE(gencnt(1:NGen),STAT=ArrLoc)
      n=max(NGen,SDim)
      ALLOCATE(AxOrd(1:n),STAT=ArrLoc)
      DO i=1,n
        AxOrd(i)=i
      ENDDO
      ALLOCATE(IndexOrder(1:n),STAT=ArrLoc)
      ALLOCATE(StartC(1:n),STAT=ArrLoc)
      ALLOCATE(EndC(1:n),STAT=ArrLoc)
      ALLOCATE(srho(0:range**SDim-1),STAT=ArrLoc)
      IF (ArrLoc/=0) THEN
        STOP 'Error allocating temporary arrays for the symmetry search. Too little memory?'
      ENDIF
      
      CALL PrepareForSym(1)
        IF (.not.searchsym.and.(aveflag/=NoAveraging).and.(aveflag/=DoNotTouch)) THEN
          mincooshift=0.0
	  cooshift=0.0
	  GoodTrans=.true.
	  goto 1000
        ENDIF
	Enant=1
!Calculation returns to this point if enantiomorphic SG is tested
100     Totalrhs=0
	genpos=0
	gencnt=0      
      IF (Verbose>1) THEN
        IF (Enant==-1) THEN
	      write(lgu,'(a)') 'Enantiomorphic space group - trying the inverse image:'
	      write(lgu,*)
	      IF (terminal) THEN
            write(*,'(a)') 'Enantiomorphic space group - trying the inverse image:'
	        write(*,*)
	      ENDIF
	    ENDIF
      ENDIF
        IF (ExpandedLog.and.(key==SearchSymKey).and.(verbose>1)) THEN
          write(lgu,'(a)') 'Positions of the peaks in the autocorrelation function:'
        ELSEIF(key==TestSymKey) THEN
          if (terminal) write(*,'(a)') 'Number   agreement         coordinates'
	  write(lgu,'(a)') 'Number   agreement         coordinates'
        ENDIF
	IF (key==SearchSymKey) THEN
          LoopNr=NGen
	ELSEIF (key==TestSymKey) THEN
	  LoopNr=NTestSym
	ENDIF
        DO symoploop=1,LoopNr
          IF (key==SearchSymKey) THEN
            cursymop=generators(symoploop)
          ELSEIF (key==TestSymKey) THEN
            cursymop=NSym+symoploop
          ENDIF
     CALL Getdands(cursymop,patpeak,mincooshift,1,.true.)
            CALL MEMRefl2SF
            rdummy=CoincidenceRS(cursymop,MinCooShift)
	  IF (key==TestSymKey) THEN !Evaluate the quality of the best solution, write out and go to the next symop
            CALL MEMRefl2SF
            rdummy=CoincidenceRS(cursymop,MinCooShift)
! temporary trick!
!            write(lgu,*) 'Shift 0.00 0.00 0.00:',rdummy
!            dhkl=mincooshift
!            dhkl(1)=dhkl(1)+0.25
!            rdummy=CoincidenceRS(cursymop,dhkl)
!            write(lgu,*) 'Shift 0.25 0.00 0.00:',rdummy
!            dhkl=mincooshift
!            dhkl(3)=dhkl(3)+0.1666666
!            rdummy=CoincidenceRS(cursymop,dhkl)
!            write(lgu,*) 'Shift 0.00 0.00 0.16:',rdummy
!            dhkl=mincooshift
!            dhkl(1)=dhkl(1)+0.25
!            dhkl(3)=dhkl(3)+0.1666666
!            rdummy=CoincidenceRS(cursymop,dhkl)
!            write(lgu,*) 'Shift 0.25 0.00 0.16:',rdummy
!            write(lgu,*)
! end of temporary trick
	    if (terminal) then
	      write(*,'(i4,4X,f8.4,4X)',ADVANCE='NO') cursymop-NSym, rdummy
	      DO i=1,SDim
	        write(*,'(f8.4)',ADVANCE='NO') MinCooShift(i)	    
	      ENDDO
	      write(*,'('''')')
              if (cursymop==NTestSym+NSym) write(*,'('''')') !Separation line at the end of the list
	    endif
	    write(lgu,'(i4,4X,f8.4,4X)',ADVANCE='NO') cursymop-NSym, rdummy
	    DO i=1,SDim
	      write(lgu,'(f8.4)',ADVANCE='NO') MinCooShift(i)	    
	    ENDDO
	    write(lgu,'('''')')	    
	    CYCLE
	  ENDIF
          IF (ExpandedLog.and.(verbose>1)) THEN
	    write(lgu,'(a,i3,a)',ADVANCE='NO') 'Symop Nr.',cursymop, ' has peak at '
            DO i=1,SDim
	      write(lgu,'(f8.4)',ADVANCE='NO') patpeak(i)
	    ENDDO
	    write(lgu,'('''')')
	  ENDIF
	  
! Find the list of all possible alternative peak positions, that might lead to an origin position in the unit cell
! MinD and MaxD have been precalculted in the subroutine PrepareForSym
          DO i=1,SDim
	    StartC(i)=int((float(mind(i))-patpeak(i))/stepd(i))
	    EndC(i)=int((float(maxd(i))-patpeak(i))/stepd(i))
	  ENDDO
          IndexOrder=StartC
! Search each possible combination, check, if it represents a correct centering vector
          DO
            DO i=1,SDim
	      dhkl(i)=float(IndexOrder(i))*stepd(i)
	    ENDDO
            GoodTrans=CheckLattTrans(dhkl,1)
	    IF (GoodTrans) THEN
              do i=1,SDim
	        dhkl2(i)=dhkl(i)+patpeak(i)
	      enddo
	      rhs=dhkl2 !preserve the right-hand side for checking purposes
	      ! calling lapack
              DummyMat=RIR
	      ZeroArr=0
              call sgelsy(SDim,SDim,1,DummyMat,SDim,dhkl2,SDim,ZeroArr,0.01,n,work,lwork,arrloc)
	      call multm(RIR,dhkl2,dhkl,SDim,SDim,1)
	      DO i=1,SDim
	        IF ((abs(dhkl(i)-rhs(i))*float(NVox(i)))>1.) THEN
		  GoodTRans=.false.
		  EXIT
		ENDIF
	      ENDDO
	      IF (GoodTrans) THEN
                !We have to save the right-hand side. It's a bit tricky, since we do not know the number apriori.
	        !Therefore the list must be reallocated each time (2DO - linked list here!)
                IF (Totalrhs==0) THEN
	          ALLOCATE(rhslist(1:SDim,1:1))
	        ELSE
	          ALLOCATE(drhslist(1:SDim,1:TotalRhs))
		  drhslist=rhslist
		  deallocate(rhslist)
		  allocate(rhslist(1:SDim,1:Totalrhs+1))
		  DO i=1,TotalRhs
		    DO ii=1,SDim
		      rhslist(ii,i)=drhslist(ii,i)
		    ENDDO
		  ENDDO
		  DEALLOCATE(drhslist)
	        ENDIF
	        Totalrhs=Totalrhs+1
		gencnt(symoploop)=gencnt(symoploop)+1
		if (genpos(symoploop)==0) genpos(symoploop)=TotalRhs
		DO i=1,SDim
		  rhslist(i,totalrhs)=rhs(i)
		ENDDO
	      ENDIF
	    ENDIF
	    IF (.not.NextCoo(SDim,StartC,EndC,IndexOrder,AxOrd)) EXIT
	  ENDDO          
	ENDDO
        IF (key==TestSymKey) THEN
          GOTO 9999 !Go to the deallocation part, no need to search for the origin
        ENDIF
!	write(*,*) 'Konecny prehled potencialnich rhs:'
!	DO i=1,NGen
!	  write(*,*) 'Symop:',generators(i)
!	  DO ii=genpos(i),genpos(i)+gencnt(i)-1
!          write(*,*) (rhslist(iii,ii),iii=1,SDim)
!	  ENDDO
!	ENDDO
! Prepare everything for the final calculation of the solution
        CALL MEMRefl2SF !Prepared for calculation of the coincidence
        !First allocate and fill the augmented matrix
	allocate(DummyMat2(1:SDim*NGen,1:SDim))
	allocate(AugMatrix(1:SDim*NGen,1:SDim))
	allocate(augrhs(1:SDim*NGen))
	allocate(daugrhs(1:SDim*NGen))
	allocate(daugrhs2(1:SDim*NGen))
        do n=1,NGen
	  do i=1,SDim
	    do ii=1,SDim
              iv=generators(n)
	      AugMatrix(i+(n-1)*SDim,ii)=-SmOp(iv)%R(i,ii)
              if (i==ii) AugMatrix(i+(n-1)*SDim,ii)=1.+AugMatrix(i+(n-1)*SDim,ii)
	    enddo
	  enddo
	enddo
        DummyMat2=AugMatrix
        !Make a workspace querry
        lwork=-1
        call sgelsy(SDim*NGen,SDim,1,DummyMat2,SDim*NGen,dAugrhs,SDim*NGen,ZeroArr,0.01,n,work,lwork,arrloc)
        lwork=nint(work(1))
        deallocate(work)
        allocate(work(1:lwork))
	DO i=1,NGen
          StartC(i)=genpos(i)
          IndexOrder(i)=genpos(i)
	  EndC(i)=genpos(i)+gencnt(i)-1
	  AxOrd(i)=i
	ENDDO
	ZeroArr=0
	mindiff=1000000.
	mincooshift=0.0E0
	Cooshift=0.0E0
	mintotshift=10000000.
        DO
          !Fill the corresponding right-hand sides
	  DO i=1,NGen
	    DO ii=1,SDim
	      Augrhs(ii+(i-1)*SDim)=rhslist(ii,IndexOrder(i))
	    ENDDO
	  ENDDO
	  daugrhs=augrhs
	  Dummymat2=AugMatrix
	  zeroarr=0  
          call sgelsy(SDim*NGen,SDim,1,DummyMat2,SDim*NGen,dAugrhs,SDim*NGen,ZeroArr,0.01,n,work,lwork,arrloc)
	  call multm(AugMatrix,dAugrhs,dAugrhs2,SDim*NGen,SDim,1)
!	  write(*,*) 'Rhs:',(Augrhs(iii),iii=1,SDim*NGen)
!	  write(*,*) 'Solution:',(dAugrhs(iii),iii=1,SDim*NGen)
!	  write(*,*) 'Check:',(dAugrhs2(iii),iii=1,SDim*NGen)
!          write(*,*) 'Difference:',(abs(dAugrhs2(i)-Augrhs(i))*float(NVox(mod(i-1,SDim)+1)),i=1,SDim*NGen)

!         Calculate the length of the difference in pixels
          rdummy=0.0
	  DO i=1,SDim*NGen
	    rdummy=rdummy+((dAugrhs2(i)-Augrhs(i))*float(NVox(mod(i-1,SDim)+1)))**2
	  ENDDO
	  rdummy=sqrt(rdummy)/NGen	  
          if (rdummy<mindiff+eps) THEN !Store the best result
            IF ((RefRhoType/='none').or.(WantSumRho/=0.and.Nsummed>0)) THEN !If we aligned with reference file before, we search for the smallest possible shift! Do not exit as soon as any shift was found.
              rdummy2=0.0E0
	      DO i=1,SDim
	        rdummy2=rdummy2+(dAugrhs(i)-float(nint(dAugrhs(i))))**2
	      ENDDO
              IF (rdummy2<mintotshift) THEN
                DO i=1,SDim
                  minCooshift(i)=dAugrhs(i)
                ENDDO
                mindiff=rdummy
                mintotshift=rdummy2
	      ENDIF
            ELSE
              mindiff=rdummy
              DO i=1,SDim
	        minCooshift(i)=dAugrhs(i)
              ENDDO
            ENDIF
            if (rdummy<=difflimit) THEN !If rdummy is under difflimit value, there is no need to search further
	      GoodTrans=.true.
            ENDIF
	  ENDIF
	  IF ((.not.NextCoo(NGen,StartC,EndC,IndexOrder,AxOrd)).or.((GoodTrans).and.(RefRhoFile=='none'))) EXIT	
	ENDDO
1000    CONTINUE 
        Cooshift=mincooshift
        coinc=CoincidenceRS(0,CooShift)
	CooShift=From0to1(CooShift,0.0005)
        IF (verbose>1) THEN
	  IF (searchsym) THEN
	    write(lgu,'('''')')
	    write(lgu,'(a)') 'Position of the origin in the CF map: '
!            write(*,*) 'Tak jsem tu'
!            write(*,*) SDim
            DO i=1,SDim
	      write(lgu,'(f8.4)',ADVANCE='NO') CooShift(i)
	      write(CurRhoProps%OrigShift(8*i-7:),'(f8.4)') CooShift(i)
	    ENDDO
!	    write(*,*) 'Za smyckou'
            write(lgu,'('''')')	      
	  ENDIF
	  IF (terminal.and.(refrhotype/='none').and.searchsym) THEN
	    write(*,'(a)') 'Position of the origin in the CF map: '
            DO i=1,SDim
	      write(*,'(f8.4)',ADVANCE='NO') CooShift(i)
	    ENDDO
            write(*,'('''')')	      
	  ENDIF
          write(lgu,'(a)') 'Agreement factors of individual generators:'
          write(lgu,'(a)') 'Number smb agreement'   
          IF (terminal) THEN
            write(*,'(a)') 'Agreement factors of individual generators:'
            write(*,'(a)') 'Number smb agreement'   
          ENDIF
          ALLOCATE(TmpSymProps%dir(SDim),TmpSymProps%IntrVec(SDim))
	  DO i=1,NGen
            rdummy=CoincidenceRS(generators(i),CooShift)           
            CALL FillSymProps(SmOp(generators(i)),TmpSymProps,RDim)
	    buffer=SymOpSymbol(SmOp(generators(i)),TmpSymProps,RDim,1)
            write(lgu,'(i4,3X,a5,f8.2)') generators(i),adjustl(buffer),rdummy
            IF (terminal) THEN
              write(*,'(i4,3X,a5,f8.2)') generators(i),adjustl(buffer),rdummy
            ENDIF
	  ENDDO
          DEALLOCATE(TmpSymProps%dir,TmpSymProps%IntrVec)
	  IF (.not.GoodTrans) THEN !No solution within difflimit was found, let's write out a warning
            write(lgu,'('''')')
            write(lgu,'(a,f6.2,a)') 'Warning, a discrepancy in the least-squares solution is ',mindiff, &
	                            ', normal values are below 0.5.'
	    write(lgu,'(a)') 'The solution might be unreliable (no convergence or wrong symmetry...?)'
	    write(lgu,'('''')')
            IF (Terminal) THEN
              write(*,'(a,f6.2,a)') 'Warning, a discrepancy in the least-squares solution is ',mindiff, &
	                              ', normal values are below 0.5.'
	      write(*,'(a)') 'The solution might be unreliable (no convergence or wrong symmetry...?)'
	      write(*,'('''')')
            ENDIF
          ENDIF
	  write(lgu,'('''')')
          write(lgu,'(a,f7.2)') 'Overall agreement factor:',coinc
          write(lgu,'('''')')
          IF (Terminal) THEN
            write(*,'('''')')
            write(*,'(a,f7.2)') 'Overall agreement factor:',coinc
            write(*,'('''')')
          ENDIF
	ELSEIF(verbose==1) THEN
	  IF (terminal) write(*,'(a,f7.2)') 'Performing symmetry matching: Overall agreement',coinc    
	  write(lgu,'(a,f7.2)') 'Performing symmetry matching: Overall agreement',coinc    
	ENDIF
        !Deallocate the arrays specific of the SearchSym part
	IF (allocated(AugMatrix)) THEN
          deallocate(AugMatrix)
	  deallocate(Augrhs)
	  deallocate(dAugrhs)
	  deallocate(dAugrhs2)
	  deallocate(Dummymat2)
	  if (allocated(rhslist)) deallocate(rhslist)
        ENDIF
      
        IF ((enant==1).and.(derivesym/='use').and.(SGProps%enant).and..not.(.not.searchsym.and.(aveflag/=NoAveraging).and. &
            (aveflag/=DoNotTouch))) THEN !In this case we have to test the inverse image
          enant=-1
          DO i=MStartF, MEndF
	    MEMRefl(i)%BMEM=-MEMRefl(i)%BMEM
	  ENDDO
	  coincEnant=coinc
	  CooShiftEnant=CooShift
	  GOTO 100 !Try the same procedure on the inverse image
	ENDIF
    IF (enant==-1) THEN !Choose the correct image
      IF ((coinc>coincenant).or.(aveflag==DoNotTouch)) THEN !we just want to keep the original signs
        IF (coinc>coincenant) THEN !The first image was better
          CooShift=CooShiftEnant
          coinc=CoincEnant
        ENDIF
        DO i=MStartF, MEndF
          MEMRefl(i)%BMEM=-MEMRefl(i)%BMEM
        ENDDO
        CALL MEMRefl2SF
      ENDIF
    ENDIF
        CurRhoProps%agrfac=coinc
        IF (aveflag/=DoNotTouch) CALL ShiftOrAverage(CooShift,aveflag)

! And fill the SF array with the shifted density or recover the SF in case of a testsymmetry run
9999    CALL MEMRefl2SF
        CALL SF2MEMRefl !Updates also FMEM
        CALL SF2Refl !To have everything complete	
        ! Deallocate the remaining arrays   
	deallocate(P)
	deallocate(HP)
	deallocate(IR)
	deallocate(RIR)
	deallocate(DummyMat)
	deallocate(ZeroArr)
	deallocate(work)
        deallocate(srho)
        deallocate(stepd,mind,maxd)

      END SUBROUTINE LocateSymmetryPatt

!A function that is true if the argument is a valid lattice translation vector
!*********************************************
      LOGICAL FUNCTION CheckLattTrans(InVec,key)
!*********************************************
        USE SF_Module
	USE SF_SymSearch
	IMPLICIT NONE
        INTEGER :: i,key !key=1 is normal symmetry, key=2 means guessed centers
        REAL, DIMENSION(1:SDim) :: InVec,DVec
	DVec=InVec !Make the operation with DVEc, the input vector should remain unchanged

        CheckLattTrans=.false.
	IF (((key==1).and.(NCenter>1)).or.((key==2).and.(NGuessCen>1))) THEN !Only in that case it makes sense to check
        !Map the shift to the unit cell
          DVec=from0to1(DVec)
	  IF (Key==1) THEN
            DO i=1,NCenter
              IF (maxval(abs(dvec-Center(1:SDim,i)))<eps) THEN
	        checkLattTrans=.true.
	        EXIT
	      ENDIF
	    ENDDO
	  ELSEIF (Key==2) THEN
            DO i=1,NGuessCen
              IF (maxval(abs(dvec-GuessCen(1:SDim,i)))<eps) THEN
	        checkLattTrans=.true.
	        EXIT
	      ENDIF
	    ENDDO	  
	  ENDIF
	ELSE
	  CheckLattTrans=.true.
	ENDIF
      END FUNCTION CheckLattTrans

! Prepare the arrays and constants for the search of the position of a SymOp
!****************************************
      SUBROUTINE PrepareForSym(key)
!****************************************
      USE SF_Module
      USE SF_SymSearch
      IMPLICIT NONE
      integer :: key !key=1 -> normal centers, key=2 -> GuessCenters
      integer :: AllocErr,i,ii,n
      REAL, ALLOCATABLE, DIMENSION (:,:) :: DummyMat
      REAL, ALLOCATABLE, DIMENSION(:) :: rsarr
      INTEGER, ALLOCATABLE, DIMENSION(:) :: StartC, EndC, AxOrd,RunC
      REAL, DIMENSION(1) :: worktmp

      ALLOCATE(AxOrd(1:SDim),STAT=AllocErr)
      ALLOCATE(RunC(1:SDim),STAT=AllocErr)
      ALLOCATE(StartC(1:SDim),STAT=AllocErr)
      ALLOCATE(EndC(1:SDim),STAT=AllocErr)
      ALLOCATE (rsarr(1:SDim),STAT=AllocErr)
      ALLOCATE (DummyMat(1:SDim,1:SDim),STAT=AllocErr)
      ALLOCATE (stepd(SDim),mind(SDim),maxd(SDim),STAT=AllocErr)
      IF (AllocErr/=0) THEN
        STOP 'Error allocating temporary arrays for the symmetry search. Too little memory?'
      ENDIF
      zeroarr=0
!     Make a querry about the necessary size of the work space for sgelsy and allocate the work array accordingly
      lwork=-1
      rsarr=0.0E0
      DummyMat=0.0E0
      call sgelsy(SDim,SDim,1,DummyMat,SDim,rsarr,SDim,ZeroArr,0.01,n,worktmp,lwork,AllocErr)
      lwork=nint(worktmp(1))
      allocate(work(1:lwork))
! Find the the smallest steps along each dimension that have to be checked to cover all lattice translations
      stepd=1.0E0
      IF ((key==1).and.(NCenter>1)) THEN
        DO i=1,SDim
          DO ii=1,NCenter
            if ((Center(i,ii)>eps).and.(Center(i,ii)<stepd(i))) stepd(i)=Center(i,ii)
          ENDDO
        ENDDO
      ELSEIF ((key==2).and.(NGuessCen>1)) THEN
        DO i=1,SDim
          DO ii=1,NGuessCen
            if ((GuessCen(i,ii)>eps).and.(GuessCen(i,ii)<stepd(i))) stepd(i)=GuessCen(i,ii)
          ENDDO
        ENDDO
      ENDIF
! Prepare the matrix P that holds all possible combinations of 1 and 0
      StartC=0
      EndC=1
      DO i=1,SDim
        AxOrd(i)=i
      ENDDO
      RunC=StartC
      n=0
      DO
        n=n+1
        DO i=1,SDim
          P(i,n)=RunC(i)
        ENDDO
        IF (.not.NextCoo(SDim,StartC,EndC,RunC,AxOrd)) EXIT
      ENDDO	
      DEALLOCATE(rsarr)
      DEALLOCATE (DummyMat)
      DEALLOCATE(StartC)
      DEALLOCATE(EndC)
      DEALLOCATE(RunC)
      DEALLOCATE(AxOrd)
      END SUBROUTINE PrepareForSym

!Find the coodrinates of the d vector and one of the possible s vectors corresponding to it
!*************************************************
      SUBROUTINE GetDandS(SymOpNr,dvector,svector,key,NeedNewMap,InIntrVec)
!*************************************************
      USE SF_Module
      USE SF_SymSearch
      IMPLICIT NONE
      INTEGER :: i,n,bestii,SymOpNr,AllocErr,pixloop
      INTEGER :: key !If key==1, take normal centering vectors, if key==2, take GuessCen
      REAL :: rdummy, rdummy2
      REAL, DIMENSION(1:SDim) :: dvector,svector, patpeak,dhkl,dhkl2
      INTEGER, DIMENSION(1:SDim) :: coo
      INTEGER, DIMENSION(1:SDim) :: StartC, EndC, AxOrd,StartC2, EndC2, RunC2
      REAL, DIMENSION(:),OPTIONAL :: InIntrVec
      REAL, DIMENSION(1:SDim) :: IntrVec,BestCenVec
      REAL, DIMENSION (1:SDim,1:SDim) :: DummyMat
      REAL :: CoincidenceRS, coinc
      LOGICAL :: GoodTrans, CheckLattTrans
      LOGICAL :: NeedNewMap
      IF (present(InIntrVec)) THEN
        IntrVec=InIntrVec
      ELSE
        Intrvec=0.
      ENDIF
      AxOrd=(/(i,i=1,SDim)/)
! First do some preparatory work - find the parameters of the search for the possible solutions
! Find out the extreme possible values of d supposed s is in [0;1]^Sdim 
      !Prepare matrix IR=(I-R)
      IR=-SmOp(SymOpNr)%R+UnitMat(SDim)
      RIR=float(IR)
      HP=multmat(IR,P)
      mind=minval(HP,dim=2)
      maxd=maxval(HP,dim=2)
      ! Now make the autocorrelation map and find the best acceptable peak in the Patterson
      IF (NeedNewMap) THEN
	CALL MEMRefl2SF
	CALL GenPatterson(SymOpNr)
        CALL MEMRefl2SF !gets destroyed by FFT
        !Normalize the density so that the peak has an (arbitrarily chosen) value 100.
	rdummy=maxval(rho)
	rho=100.*rho/rdummy        
      ENDIF
      rdummy=-1.0E30
      rdummy2=100000.
      StartC=0
      ENdC=NVox-1
      coo=maxval(abs(IR),dim=2)
      DO i=1,SDim
        if (coo(i)==0) THEN
          StartC(i)=nint(intrvec(i)*NVox(i))
          EndC(i)=StartC(i)
        ENDIF
      ENDDO
      coo=StartC
      DO
        CALL ArrayLocation(coo,pixloop)
	IF (rho(pixloop)>rdummy) THEN
        ! Check, if the "peak" gives a viable solution
	  patpeak=float(coo)/NVox-IntrVec
          ! find the limits of the search:
          DO i=1,SDim
	    StartC2(i)=int((float(mind(i))-patpeak(i))/stepd(i))
	    EndC2(i)=int((float(maxd(i))-patpeak(i))/stepd(i))
	  ENDDO
          RunC2=StartC2
          ! Search each possible combination, check, if it represents a correct centering vector
	  DO
	    dhkl=RunC2*stepd
            GoodTrans=CheckLattTrans(dhkl,key)
	    IF (GoodTrans) THEN
              dhkl2=dhkl+patpeak
	      dvector=dhkl2 !Remember the right-hand side of (I-R)s=d
	      ! calling lapack
              DummyMat=RIR
              ZeroArr=0
              call sgelsy(SDim,SDim,1,DummyMat,SDim,dhkl2,SDim,ZeroArr,0.01,n,work,lwork,AllocErr)
              dhkl=multmat(RIR,dhkl2) !dhkl2~s, calculate d_new=(I-R)s
              IF (maxval(abs(dhkl-dvector)*NVox)>=1.) GoodTrans=.false.
	    ENDIF
	    IF (GoodTrans.or.(.not.NextCoo(SDim,StartC2,EndC2,RunC2,AxOrd))) EXIT
	  ENDDO          
          IF (GoodTrans) THEN !This peak passed the check and is the highest up to now -> remember it
!!!!! The following commented-out part is a trick to use coincidenceRS to select the best position. Much slower, and not better, but who knows, maybe a new CoincidenceRS will be better
!            dhkl=SmOp(SymOpNr)%t
!            IF (present(inintrvec)) THEN
!	      SmOp(SymOpNr)%t=intrvec
!	    ENDIF
!	    coinc=CoincidenceRS(SymOpNr,dhkl2)
!            SmOp(SymOpNr)%t=dhkl
!            write(*,*) rho(pixloop), rdummy, coinc, rdummy2, coo
!	    IF (coinc<rdummy2) THEN
!	      write(*,*) 'better'
              rdummy=rho(pixloop)
              rdummy2=coinc
	      bestii=pixloop
	      bestcenvec=RunC2*stepd
!	    ENDIF
	  ENDIF      
	ENDIF
!	IF  ((rdummy>(100.-eps)).or.(.not.NextCoo(SDim,StartC,EndC,coo,AxOrd))) EXIT
	IF  (.not.NextCoo(SDim,StartC,EndC,coo,AxOrd)) EXIT
      ENDDO
   
!     Get the precise position of the peak by spline interpolation
      call Coordinates(bestii,coo)
      CALL GetPeak(coo,dvector,rdummy)
      IF (rdummy<-1.0E29) dvector=coo !The interpolation was not succesfull, the original coordinates must be taken
      dvector=dvector/NVox
      !And finally make sure that the peak really lies on the allowed subspace, i.e. that (I-R)s=d is exactly valid for some s.
      dhkl2=dvector-IntrVec+bestcenvec
      DummyMat=RIR
      ZeroArr=0
      call sgelsy(SDim,SDim,1,DummyMat,SDim,dhkl2,SDim,ZeroArr,0.01,n,work,lwork,AllocErr)
      dvector=multmat(RIR,dhkl2)-bestcenvec+IntrVec !calculate d_new=(I-R)s, subtract the centering added before, and add the intrinsic vector
      svector=dhkl2 !Save the solution for the best peak - is useful fro TestSymmetry
      END SUBROUTINE GetDandS

! Routine taken from LocateSymmetryRS, but here all symops are used together to refine the position of the origin
!Attention: if symnr=-1, shift between maps in MEMRefl and SF is evaluated. For symnr=-2 inversion is applied and the shift is evaluated
!**************************************************
      SUBROUTINE RefineOrigPos(OrCoo,SymDim,RefDim,SymNr,succ)
!**************************************************
        USE SF_Module
        USE SF_Interfaces
	IMPLICIT NONE
	INTEGER, PARAMETER :: ptchangelim=40
	real, PARAMETER :: MaxGradShift=0.2
	real, PARAMETER :: SmallSh=0.05
	real, PARAMETER :: MaxPtShift=2.
	real :: disc, Grad, GradOld, GradOlder, red, dump
	real :: deriv1, deriv2,maxsh
	real, dimension(SDim) :: OrCoo
	real :: ptor(SDim) !point, where the density is to be determined
	real, ALLOCATABLE :: H(:,:),Hinv(:,:),Ho(:,:),Gr(:), Sh(:) !Hessian, its inversion and gradient
        INTEGER :: DerOrd(2), ptchange, SymNr, SymDim,i,iii,iv
	INTEGER, DIMENSION(SymDim) :: RefDim
        LOGICAL :: succ

	ALLOCATE(H(SymDim,SymDim),STAT=AllocError)
	ALLOCATE(Hinv(SymDim,SymDim),STAT=AllocError)
	ALLOCATE(Ho(SymDim,SymDim),STAT=AllocError)
	ALLOCATE(Gr(SymDim),STAT=AllocError)
	ALLOCATE(Sh(SymDim),STAT=AllocError)
	IF (AllocError.GT.0) STOP 'Error allocating in H, Hinv and Gr in CpRefine! Too little memory?'
	DO i=1,SymDim
          ptor(i)=OrCoo(RefDim(i))
        END DO
        ptchange=0
	dump=1.
        red=1.
	gradold=0.0
	gradolder=0.0
        DO
	  ptchange=ptchange+1
          succ=.true.
          IF (ptchange>ptchangelim) THEN !trying too long time
            Succ=.false.
            EXIT
          END IF
!******** Gradient and Hessian
	  DO iii=1,SymDim
	    gr(iii)=deriv1(SymNr,OrCoo,RefDim(iii))
	  END DO
	  DO iii=1,SymDim
            DO iv=iii,SymDim
	      DerOrd=0
	      DerOrd(1)=RefDim(iii)
	      DerOrd(2)=RefDim(iv)
	      H(iii,iv)=deriv2(SymNr,OrCoo,DerOrd)
              H(iv,iii)=H(iii,iv)
            END DO
          END DO
	  Hinv=invert(H)
          IF (Hinv(1,1)>1.0E29) THEN
            succ=.false.	       
            EXIT
          END IF
          Sh=0.0E0
	  Grad=0.0E0
          DO iii=1,SymDim
            DO iv=1,SymDim
              Sh(iii)=Sh(iii)+Gr(iv)*Hinv(iv,iii)
            END DO
            Grad=Grad+(Sh(iii)*NVox(RefDim(iii)))**2
          END DO
          Grad=sqrt(Grad)
	  IF (Grad>MaxGradShift) THEN
            Red=MaxGradShift/Grad
          ELSE
            Red=1.0
          END IF
	  disc=0.0E0
	  maxsh=0.
          DO iii=1,SymDim
	    maxsh=max(maxsh,abs(sh(iii)*red*dump*NVox(RefDim(iii))))
            OrCoo(RefDim(iii))=OrCoo(RefDim(iii))-sh(iii)*red*dump
	    disc=disc+((OrCoo(RefDim(iii))-ptor(iii))*NVox(RefDim(iii)))**2
          END DO
          disc=sqrt(disc)
          IF (disc>MaxPtShift) Succ=.false.
          IF (ptchange>2) THEN
            if ((gradold-grad)>(gradolder-gradold)) dump=dump*0.5
          END IF
	  gradolder=gradold
	  gradold=grad
!          formstr='( f8.3,2X,f7.4)'
!          write(formstr(2:2),'(i1)') SDim
!	  IF (ExpandedLog) THEN
!            dummy=CoincidenceRS(SymNr,OrCoo)
!	    write(lgu,formstr) (OrCoo(iii)*NVox(iii),iii=1,SDim),dummy
!          ENDIF
!          IF (Terminal) THEN
!	    dummy=CoincidenceRS(SymNr,OrCoo)
!      	    write(*,formstr) (OrCoo(iii)*NVox(iii),iii=1,SDim),dummy
!	  ENDIF  
          IF (.not.Succ.or.maxsh<smallsh) EXIT !loop until .not.succ or disc is sufficiently small
        END DO
!        maxsum=CoincidenceRS(SymNr,OrCoo)
        DEALLOCATE(H)
        DEALLOCATE(Hinv)
        DEALLOCATE(Ho)
        DEALLOCATE(Gr)
        DEALLOCATE(Sh)
      END SUBROUTINE RefineOrigPos
      
! Function returning the coincidence of two maps. IF Symop>0, a map is compared with its symmetry-transformed image.
! For Symop=0 cummulative coincidence of all syops is calculated. For SymOp<-1 only shift between map in MEMRefl and SF is calculated, with SymOp=-2 the inversion is also appplied
! SymOp=-3 indicates cummulative coincidence of all generators
!**********************************************
      real FUNCTION CoincidenceRS(symop,shift)
!**********************************************
      USE SF_Module

      IMPLICIT NONE
      
      real, parameter :: expph=0.0833333 !=(1/0.5)*int_0^0.5 (ph^2 d ph)=average phase squared
      real :: phdiff,phase
      real :: cosph, fsq, sumn, sumfsqphsq, sumfsqcosph, sumffourcosph, sumfsq, sumffour, sumffouronecosphsq
      real :: meanIo, meanIc, meanIoIc, meanIosq, meanIcsq, Crand
      real, dimension(SDim) :: Shift,svec
      integer, dimension(SDim) :: ICoo
      integer :: ii,symop,Arrloc,friedel,Lstart,Lend,CurSymOp,CurSymOpLoop,enant

      IF (SymOp==0) THEN
        LStart=2
	LEnd=NSymPrim
	Enant=1
      ELSEIF ((SymOp==-1).or.(SymOp==-2)) THEN
        LStart=1
	LEnd=1
	IF (SymOp==-1) THEN
	  Enant=1
	ELSE
	  Enant=-1
	ENDIF
      ELSEIF (SymOp==-3) THEN
        LStart=1
	LEnd=NGen
	Enant=1
      ELSE
        LStart=SymOp
	LEnd=SymOp
	Enant=1
      ENDIF
      sumfsqphsq=0.; sumfsqcosph=0.; sumffourcosph=0.; sumfsq=0.; sumffour=0.; sumffouronecosphsq=0.; sumn=0.
      DO CurSymOpLoop=LStart,LEnd
        IF (SymOp==-3) THEN !Assign the sym number form the list of generators
	  CurSymOp=generators(CurSymOpLoop)
	ELSE !Just copy the loop number, it is automatically the number of symmetry operation
	  CurSymOp=CurSymOpLoop
	ENDIF
	DO ii=MStartF+1,MEndF !Skip the first reflection, which is the F(0)
          IF ((MEMRefl(ii)%code==code_missing).or.(MEMRefl(ii)%code==code_skip)) cycle !Do not take the missing reflections into account
!         Applying the symop to each reflection
          IF (cursymop/=1) THEN
	    ICoo=multmat(MEMRefl(ii)%hklFG,SmOp(cursymop)%R)
            CALL ArraylocationSF(Icoo, Arrloc,friedel)
          ELSE
            CALL ArraylocationSF(MEMRefl(ii)%hklFG, Arrloc,friedel)
	  ENDIF
! calculate h*[(I-R)*Shift-Tvec] or h*shift for SymOp<0
	  IF (SymOp>=0) THEN
	    svec=multmat(SmOp(cursymop)%R,shift)
	    svec=shift-svec-SmOp(cursymop)%t
	  ELSE
	    svec=shift
	  ENDIF
          phdiff=sum(MEMRefl(ii)%hklFG*svec)
	  phdiff=phdiff-(Phase(MEMRefl(ii)%Amem,enant*MEMRefl(ii)%BMem)-   &
                  Phase(SF(Arrloc),float(friedel)*SF(Arrloc+1)))/(2.*pi)
          phdiff=phdiff-aint(phdiff)
          if (abs(phdiff)>0.5E0) THEN
	    if (phdiff<0.) THEN 
	      phdiff=phdiff+1.0
	    else
	      phdiff=phdiff-1.0
            end if
	  END IF
          cosph=cos(2*pi*phdiff)
	  fsq=MEMRefl(ii)%FMEM*MEMRefl(ii)%FMEM
          sumfsqphsq=sumfsqphsq+(fsq*phdiff**2)
!	  write(*,*) fsq,phdiff, sumfsqphsq
          sumfsqcosph=sumfsqcosph+(fsq*cosph)
          sumffourcosph=sumffourcosph+(fsq*fsq*cosph)
	  sumfsq=sumfsq+MEMRefl(ii)%FMEM**2
          sumffour=sumffour+fsq*fsq
          sumffouronecosphsq=sumffouronecosphsq+fsq*fsq*(1.+cosph)**2
	  sumn=sumn+1.
!	  sfamp=sqrt(SF(Arrloc)**2+SF(Arrloc+1)**2)
!          suma=suma+(MEMRefl(ii)%FMEM*sfamp*phdiff**2)
!	  sumden=sumden+MEMRefl(ii)%FMEM*sfamp
        END DO
      ENDDO
      SELECT CASE(CoincCode)
      CASE(1)
        CoincidenceRS=100.*(sumfsqphsq/sumfsq)/expph !Normalize to the expected average value of S
      CASE(2)
        meanIo=sumfsq/sumn
	meanIc=2*(sumfsq+sumfsqcosph)/sumn
	meanIoIc=2*(sumffour+sumffourcosph)/sumn
	meanIosq=sumffour/sumn
	meanIcsq=4*sumffouronecosphsq/sumn
	Crand=sqrt((meanIosq-meanIo**2)/(1.5*meanIoSq-meanIo**2)) !Calculate the expected correlation coefficient for random phases, which is NOT zero!
	CoincidenceRS=100.*(1.-(meanIoIc-meanIo*meanIc)/sqrt((meanIosq-meanIo**2)*(meanIcsq-meanIc**2)))/(1.-Crand)
      END SELECT
      END FUNCTION CoincidenceRS


! Function returning the coincidence of two maps. IF Symop>0, a map is compared with its symmetry-transformed image.
! For Symop=0 cummulative coincidence of all syops is calculated. For SymOp<-1 only shift between map in MEMRefl and SF is calculated, with SymOp=-2 the inversion is also appplied
! SymOp=-3 indicates cummulative coincidence of all generators
!**********************************************
      real FUNCTION CoincidenceRSBurla(symop,shift)
!**********************************************
      USE SF_Module

      IMPLICIT NONE
      
      real, parameter :: expph=0.0833333 !=(1/0.5)*int_0^0.5 (ph^2 d ph)=average phase squared
      real :: sumcos,sumsin,sumden,phdiff,phase
      real, dimension(SDim) :: Shift,svec
      integer, dimension(SDim) :: ICoo
      integer :: i,ii,iii,iv,symop,Arrloc,friedel,Lstart,Lend,CurSymOp,CurSymOpLoop,enant

      sumcos=0.0E0
      sumsin=0.0E0
      sumden=0.0E0
      IF (SymOp==0) THEN
        LStart=2
	LEnd=NSymPrim
	Enant=1
      ELSEIF ((SymOp==-1).or.(SymOp==-2)) THEN
        LStart=1
	LEnd=1
	IF (SymOp==-1) THEN
	  Enant=1
	ELSE
	  Enant=-1
	ENDIF
      ELSEIF (SymOp==-3) THEN
        LStart=1
	LEnd=NGen
	Enant=1
      ELSE
        LStart=SymOp
	LEnd=SymOp
	Enant=1
      ENDIF
      DO CurSymOpLoop=LStart,LEnd
        IF (SymOp==-3) THEN !Assign the sym number form the list of generators
	  CurSymOp=generators(CurSymOpLoop)
	ELSE !Just copy the loop number, it is automatically the nbumber of symmetry operation
	  CurSymOp=CurSymOpLoop
	ENDIF
	DO ii=MStartF+1,MEndF !Skip the first reflection, which is the F(0)
!         Applying the symop to each reflection
!         H by R
          DO iii=1, SDim
            Icoo(iii)=0
            DO iv=1, SDim
              Icoo(iii)=ICoo(iii)+MEMRefl(ii)%hklFG(iv)*SmOp(cursymop)%R(iv,iii)
            END DO
          END DO
          CALL ArraylocationSF(Icoo, Arrloc,friedel)
	  phdiff=0.0
! calculate h*[(I-R)*Shift-Tvec] of h*shift for SymOp<0
            svec=0.0E0
          DO i=1,SDim
            IF (SymOp>=0) THEN
              DO iii=1,SDim
	        svec(i)=svec(i)+SmOp(cursymop)%R(i,iii)*Shift(iii)
	      END DO
              svec(i)=-svec(i)+Shift(i)-SmOp(cursymop)%t(i)
            ELSE
	      svec(i)=Shift(i)
	    ENDIF
	    phdiff=phdiff+MEMRefl(ii)%hklFG(i)*svec(i)
	  END DO             
	  phdiff=phdiff-(Phase(MEMRefl(ii)%Amem,enant*MEMRefl(ii)%BMem)-   &
                  Phase(SF(Arrloc),float(friedel)*SF(Arrloc+1)))/(2.*pi)
!          phdiff=phdiff-aint(phdiff)
!          if (abs(phdiff)>0.5E0) THEN
!	    if (phdiff<0.) THEN 
!	      phdiff=phdiff+1.0
!	    else
!	      phdiff=phdiff-1.0
!            end if
!	  END IF
          sumcos=sumcos+MEMRefl(ii)%FMEM**2*cos(2*pi*phdiff)
          sumsin=sumsin+MEMRefl(ii)%FMEM**2*abs(sin(2*pi*phdiff))
	  sumden=sumden+MEMRefl(ii)%FMEM**2
        END DO
      ENDDO
      CoincidenceRSBurla=(sumcos-sumsin)/sumden
      END FUNCTION CoincidenceRSBurla


      FUNCTION deriv1(snr,x,coo) !Evaluates the derivative of the function CoincidenceRS along the coordinate coo
      USE SF_Module
      
      IMPLICIT NONE 

      INTEGER coo,snr 
      real deriv1,x(SDim),xx1(SDim),xx2(SDim)
      real h,CoincidenceRS

      h=0.01/NVox(coo)
   
      xx1=x; xx1(coo)=xx1(coo)-h
      xx2=x; xx2(coo)=xx2(coo)+h
      deriv1=(CoincidenceRS(snr,xx2)-CoincidenceRS(snr,xx1))/(2.0*h)
      END


      FUNCTION deriv2(snr,x,DerOrd) !Evaluates the derivative of the function CoincidenceRS along the coordinate coo
      USE SF_Module
      
        IMPLICIT NONE 

        INTEGER snr,DerOrd(2),i
        real deriv2,x(SDim),xx1(SDim),xx2(SDim),tmpder(2)
        real h1,h2,CoincidenceRS
      

        h1=0.01/NVox(DerOrd(1))
        h2=0.01/NVox(DerOrd(2))
   
! Get two derivatives along derord1 in points separated by 2h2 along derord2
        DO i=-1,1,2
          xx1=x; xx2=x
          xx1(DerOrd(1))=xx1(DerOrd(1))+i*h1
          xx2(DerOrd(1))=xx2(DerOrd(1))+i*h1

	  xx1(DerOrd(2))=xx1(DerOrd(2))-h2
          xx2(DerOrd(2))=xx2(DerOrd(2))+h2
          tmpder(1+(i+1)/2)=(CoincidenceRS(snr,xx2)-CoincidenceRS(snr,xx1))/(2.0*h2)
        END DO 
        deriv2=(tmpder(2)-tmpder(1))/(2.0*h1)
      END FUNCTION Deriv2

! Finds the shift between the current density and the reference density and shifts the phases accorddingly
! If verbose is true, the "verbose" mode is used, othervise silent mode without refinement 
!******************************************************
      SUBROUTINE MatchReference(denskey,verbose)
!******************************************************
      USE SF_Module
      USE SF_Interfaces
      USE FFT_Interfaces
      IMPLICIT NONE
      real :: peakvalue,rdummy, CoincidenceRS
      integer, dimension(SDim) :: PCoo, RefDim
      real, dimension(SDim) :: DCoo
      integer :: i,ii,enant,enantloop,lend, peakpos
      integer :: denskey !1=reference density, 2=envelope density, 3=summed densities
      logical :: Succ
      integer :: verbose
!      IF (expandedlog) write(lgu,'(a)') 'Matching the density with the reference density:'
!     We have to fill the structure factors of refrho in SF:
! Now construct the "Generalized Patterson" from MEMRefl and SF with symmetry operation 1 = do nothing with the SF      
      IF (Centro) THEN 
        lend=1
      ELSE
        lend=-1
      ENDIF
      peakvalue=0.0E0
      DO enantloop=1,lend,-2 !For centro only one loop will be performed, for non-centro two
        IF (DensKey==1) THEN
          IF (any(NVoxRef/=NVox)) THEN
            CALL ResampleDensity(NVoxRef,NVox,refrho)
            NVoxRef=NVox
          ENDIF
          rho=refrho
        ELSEIF (DensKey==2) THEN
          IF (any(NVoxEnv/=NVox)) THEN
            CALL ResampleDensity(NVoxEnv,NVox,envrho)
            NVoxEnv=NVox
          ENDIF
          rho=envrho
        ELSEIF (DensKey==3) THEN
          rho=sumrho
        ELSEIF (DensKey==4) THEN
          rho=rho50
        ENDIF
        CALL FFT(rho2sf)
        !Quick hack to allow matching only to some reflections
        DO i=1,MTotalF
          IF ((MEMRefl(i)%F<intformatchlow).or.(MEMRefl(i)%F>intformatchhigh)) THEN
            SF(MEMRefl(i)%SFloc)=0.0
            SF(MEMRefl(i)%SFloc+1)=0.0
          ENDIF
        ENDDO
        CALL GenPatterson(enantloop)
        !Find the highest peak. It defines the shift vector between the current density and the reference file Currho(r)=Refrho(r-PCoo)
        DO ii=0,NVoxTot-1
          IF (rho(ii)>peakvalue) THEN
            PeakValue=rho(ii)	    
            PeakPos=ii
            enant=enantloop
          ENDIF
        ENDDO
      ENDDO
      CALL Coordinates(Peakpos,PCoo)
! Fill the refrho SF, we'll need them for refinement
      IF (DensKey==1) THEN
        rho=refrho
      ELSEIF (DensKey==2) THEN
        rho=envrho
      ELSEIF (DensKey==3) THEN
        rho=sumrho
      ELSEIF (DensKey==4) THEN
        rho=rho50
      ENDIF
      CALL FFT(rho2sf)
!     Refine the shift
      DO i=1,SDim
        dcoo(i)=float(PCoo(i))/float(NVox(i))
      ENDDO
      IF (verbose>0) THEN
        DO i=1,SDim
          RefDim(i)=i
        ENDDO
!      IF (Expandedlog) THEN     
!        write(lgu,'(a)') 'I fould following shift:'
!        DO i=1,SDim
!          write(lgu,'(i5)',ADVANCE='NO') PCoo(i)
!        ENDDO
!	write(lgu,'('''')')
! 	write(lgu,'(a)') 'Refining the  shift:'
!      ENDIF
        succ=.true.
        IF (Enant==1) THEN
          IF ((verbose/=1).and.((intformatchlow<eps).and.(intformatchhigh>1.0E29))) CALL RefineOrigPos(DCoo,SDim,RefDim,-1,succ)
 	      rdummy=CoincidenceRS(-1,DCoo)
 	    ELSE
          IF ((verbose/=1).and.((intformatchlow<eps).and.(intformatchhigh>1.0E29))) CALL RefineOrigPos(DCoo,SDim,RefDim,-2,succ) !quick hack here to speed up the calculation. Verbose==1 is for creation of a correlation table
	      rdummy=CoincidenceRS(-2,DCoo)
        ENDIF
        IF ((.not.Succ).and.(verbose>1)) THEN
          write(lgu,'(a)') 'Warning, a problem occured in the refinement of the shift between the current and reference density.'
          write(lgu,'(a)') 'The refined position might be unrieliable.'
        ENDIF
      ENDIF
      IF (Enant==-1) THEN
        DO i=MStartF, MEndF
          MEMRefl(i)%BMem=-MEMRefl(i)%BMEM
        ENDDO
      ENDIF
      IF (verbose>1) THEN
        write(lgu,'(a,f7.2,a)') 'The density was aligned with the reference file, agreement ',rdummy,'%.'
        write(lgu,'(a)',ADVANCE='NO') 'Shift: '
        DO i=1,SDim
          write(lgu,'(f9.5)',ADVANCE='NO') DCoo(i)
        ENDDO
        write(lgu,'('''')')
        write(lgu,'('''')')
        IF (Terminal) THEN
          write(*,'(a,f7.2,a)') 'The density was aligned with the reference file, agreement ',rdummy,'%.'
          write(*,'(a)',ADVANCE='NO') 'Shift: '
          DO i=1,SDim
            write(*,'(f9.5)',ADVANCE='NO') DCoo(i)
          ENDDO
          write(*,'('''')')
          write(*,'('''')')
        ENDIF
      ELSEIF(verbose==1) THEN
        TwoDenMatch=rdummy
        IF (.not.allocated(TwoDenShift)) allocate(TwoDenShift(1:SDim))
        TwodenShift=DCoo
      ENDIF
      CurRhoProps%refmatch=rdummy
      CALL MEMRefl2SF
      CALL ShiftOrAverage(DCoo,NoAveraging)
      END SUBROUTINE MatchReference
      
! Finds the shift between the current density and the density recorded several cycles ago
!******************************************************
      SUBROUTINE MatchOldRho
!******************************************************
      USE SF_Module
      USE SF_Interfaces
      USE FFT_Interfaces
      IMPLICIT NONE
      real :: peakvalue,rdummy, CoincidenceRS
      integer, dimension(SDim) :: PCoo, RefDim
      real, dimension(SDim) :: DCoo
      integer :: i,ii,enant,enantloop,lend, peakpos
      logical :: Succ
      peakvalue=0.0E0
      rho=rho50
      CALL FFT(rho2sf)
      CALL GenPatterson(1)
      !Find the highest peak. It defines the shift vector between the current density and the reference file Currho(r)=Refrho(r-PCoo)
      DO ii=0,NVoxTot-1
        IF (rho(ii)>peakvalue) THEN
          PeakValue=rho(ii)	    
          PeakPos=ii
        ENDIF
      ENDDO
      CALL Coordinates(Peakpos,PCoo)
      rho=rho50
      CALL FFT(rho2sf)
 !     Refine the shift
      DO i=1,SDim
        dcoo(i)=float(PCoo(i))/float(NVox(i))
      ENDDO
      DO i=1,SDim
        RefDim(i)=i
      ENDDO
      CALL RefineOrigPos(DCoo,SDim,RefDim,-1,succ)
      rdummy=CoincidenceRS(-1,DCoo)	  
      twodenmatch=rdummy
      !Get back the original rho
      CALL MEMRefl2SF
      CALL FFT(sf2rho)
      END SUBROUTINE MatchOldRho

! Calculate a "Generalized Patterson, i.e. FT of product MEMRefl*SF, with SF possibly modified by some symop
! The highest peak in Patterson will correspond to vector d such that rho_MEMRefl(Rr+t+d)~rho_SF(r)
!******************************************************
      SUBROUTINE GenPatterson(SymOp)
!******************************************************
      USE SF_Module
      USE SF_Interfaces
      USE FFT_Interfaces

      IMPLICIT NONE

      REAL, DIMENSION(SDim,SDim) :: InvRMat
      integer, dimension(SDim) :: Icoo
      real :: tmpA, tmpB, phdiff, phase
      integer :: friedel,i,ii,iii,iv,arrloc,symop,cursymop,enant
      type ABtype
        real :: A,B
      end type ABtype
      type(ABtype), allocatable, dimension(:) :: MEMreflTmp
      
      ALLOCATE(MEMReflTmp(MStartF:MEndF))      
      IF (SymOp==-1) THEN !Flag for the inverse image of a non-centrosymmetric density
        InvRMat=0.0E0
	DO i=1,SDim
	  InvRMat(i,i)=1.
	ENDDO
	enant=-1
	cursymop=1 !To prevent generating a dummy vector TVec for SymOp=-1
      ELSE
        InvRMat=SmOp(SymOp)%R
        enant=1
	cursymop=symop
      ENDIF
      DO i=MStartF,MEndF
! If a density is rotated by {R|t}, F_rot(h)=exp(2pi*i*ht)*F_orig(hR)
! The convention is that the shift vector points from the current origin to the desired origin. For a comparison of
! the current density with the reference density this means vector from a point in the reference density to the equivalent point in the current density.
! The convolution is defined as rho1*rho2=integral(rho1(r).rho2(u-r)). Such a function will have a peak for u= vector from rho1 to rho2.
! Thus, rho1 must be the reference file, rho2 must be inverse to the current density.By convolution theorem we get F(Patt)=F_ref.F^*_cur
	DO iii=1,SDim
          Icoo(iii)=0
          DO iv=1, SDim
            Icoo(iii)=ICoo(iii)+nint(float(MEMRefl(i)%hklFG(iv))*InvRmat(iv,iii))
          END DO
        END DO
        CALL ArraylocationSF(Icoo, Arrloc,friedel)
	phdiff=0.0
! calculate phdiff=-h*Tvec
        DO ii=1,SDim
          phdiff=phdiff-2.*pi*float(MEMRefl(i)%hklFG(ii))*SmOp(cursymop)%t(ii)
        END DO
        phdiff=friedel*Phase(SF(Arrloc),SF(Arrloc+1))+phdiff
        tmpA=SQRT(SF(Arrloc)**2+SF(Arrloc+1)**2)*cos(phdiff)
        tmpB=SQRT(SF(Arrloc)**2+SF(Arrloc+1)**2)*sin(phdiff)
        !F^*(MEMRefl)*F(SF)
        MEMReflTmp(i)%A=MEMRefl(i)%AMEM*tmpA+float(enant)*MEMRefl(i)%BMEM*tmpB
        MEMReflTmp(i)%B=MEMRefl(i)%AMEM*tmpB-float(enant)*MEMRefl(i)%BMEM*tmpA
      ENDDO
      SF=0.0E0
      DO i=MStartF,MEndF
        SF(MEMRefl(i)%SFloc)=MEMReflTmp(i)%A
        SF(MEMRefl(i)%SFloc+1)=MEMReflTmp(i)%B
      ENDDO
      !Now calculate the "patterson", and that's the end of the story
      CALL FFT(sf2rho)
!      IF (InvRMat(1,1)>0.) THEN
!      buffer='patterson_   .xplor'
!      write(Buffer(11:13),'(i3.3)') cursymop
!      buffer2='xplor'
!      call WriteRho(trim(buffer),trim(buffer2))
      DEALLOCATE(MEMReflTmp)
      END SUBROUTINE GenPatterson

!Shift the reflections phases to shift the rho by CooShift, optionally average the phases over the symmetry
!Assumes than MEMRefl2SF has been preformed and SF contains the correct structure factors!
! The shift is performed according to the r(new)=r(old)-CooShift!! In other words, vector CooShift points from the current origin to the desired origin
! If a density should be shifted by -d, the structure factors change according to:
! F_shift=exp(-2*pi*i*h*d)*F_orig
! The relation between the symmetrically related reflections by symop {R|t} is:
!F(hR)=exp(-2*pi*i*h*t)*F(h) Thus, we have to correct the phase of F(hR) by +2*pi*h*t to get to the phase that corresponds to F(h)
!*********************************************
      SUBROUTINE ShiftOrAverage(CooShift,aveflag)
!*********************************************
      USE SF_Module
      IMPLICIT NONE

      DOUBLE PRECISION, PARAMETER :: deps=1.0D-5
      REAL, DIMENSION(SDim) :: CooShift
      REAL, DIMENSION(SDim,SDim) :: InvMat
      DOUBLE PRECISION dA,dB,dsum, dPhase
      REAL :: fsum, dummy,asum,bsum
      integer :: nsum
      INTEGER :: i,ii,iii,iv, friedel, arrloc
      INTEGER, DIMENSION(SDim) :: hkl
      INTEGER :: AveFlag
      
!...shift the phases so that the density is properly positioned
      DO i=MStartF+1,MEndF !Start from the second one, the first is F000
        dsum=0.0D0
        DO iii=1,SDim
           hkl(iii)=MEMRefl(i)%hklFG(iii)
           dsum=dsum-dble(hkl(iii))*dble(CooShift(iii))
        END DO
        CALL ArrayLocationSF(hkl,ArrLoc,friedel)
        dsum=DPhase(dble(SF(Arrloc)),dble(friedel)*dble(SF(Arrloc+1)))+dsum*2.*pi
        MEMRefl(i)%AMEM=MEMRefl(i)%F*dcos(dsum)
        MEMRefl(i)%BMEM=MEMRefl(i)%F*dsin(dsum)
      END DO
      CALL MEMRefl2SF
! If desired, average the phases of symmetrically equivalent reflections...
      IF (aveflag/=NoAveraging) THEN
        DO i=MStartF+1,MEndF !Start from the second one, the first is F000
          dA=0.0D0; dB=0.0D0
!          write(*,*) 'Reflexe:',(MEMRefl(i)%hklfg(ii),ii=1,3),MEMRefl(i)%AMEM,MEMRefl(i)%BMEM, &
!	        DPhase(dble(MEMRefl(i)%AMEM),dble(MEMRefl(i)%BMEM))          
          fsum=0.0; asum=0.0; bsum=0.0; nsum=0
	      !First check, if the reflection is not forbidden by centering
          DO ii=1,NCenter
            fsum=sum(Center(1:SDim,ii)*MEMRefl(i)%hklFG)
            fsum=fsum-nint(fsum)
            IF (abs(fsum)>bigeps) THEN
              MEMRefl(i)%AMEM=0.0
              MEMRefl(i)%BMEM=0.0
              MEMRefl(i)%FMEM=0.0
              EXIT
            ENDIF
          ENDDO
          IF (abs(fsum)>bigeps) CYCLE
          DO ii=1,NSymPrim
            hkl=0; dsum=0.0D0
            InvMat=invert(SmOp(ii)%R)
            DO iii=1,SDim
              DO iv=1,SDim
                hkl(iii)=hkl(iii)+MEMRefl(i)%hklFG(iv)*nint(InvMat(iv,iii))
              END DO
              dsum=dsum+dble(hkl(iii))*dble(SmOp(ii)%t(iii))          
            END DO
            CALL ArrayLocationSF(hkl,ArrLoc,friedel)
!            write(*,*) 'Obraz pod symop',ii,' je:',(hkl(iii),iii=1,3),SF(Arrloc),float(friedel)*SF(Arrloc+1), &
!	                DPhase(dble(SF(Arrloc)),dble(friedel)*dble(SF(Arrloc+1)))
!            write(*,*) 'friedel:',friedel, 'Arrloc:',Arrloc
!            write(*,*) 'Symmetry shift:',sum
            dsum=DPhase(dble(SF(Arrloc)),dble(friedel)*dble(SF(Arrloc+1)))+(dsum-dble(int(dsum)))*2.*pi
            dummy=sqrt(SF(Arrloc)**2+SF(ArrLoc+1)**2)
            IF (dummy>eps) THEN
              fsum=fsum+dummy
              nsum=nsum+1
              dA=dA+dcos(dsum)
              dB=dB+dsin(dsum)
              asum=asum+dummy*dcos(dsum)
              bsum=bsum+dummy*dsin(dsum)
            ENDIF
!	         write(*,*) 'sum, dA, dB = ',dsum,da,db,dcos(dsum),dsin(dsum)
!            if (hkl(3)>0) write(*,*) (hkl(iii),iii=1,SDim),sqrt(SF(Arrloc)**2+SF(ArrLoc+1)**2)
          END DO
!	  read(*,*)
          IF ((abs(dA)<deps).and.(abs(dB)<deps)) THEN !Sum of vectors is zero - either the reflection is zero, or it is an extinct reflection due to the symmetry!
            MEMRefl(i)%AMEM=0.0
            MEMRefl(i)%BMEM=0.0
            MEMRefl(i)%FMEM=0.0
         !write(*,*) 'Extinct:', (MEMRefl(i)%hklFG(iv),iv=1,SDim), MEMRefl(i)%F, dA, dB
          ELSE	  
            IF (AveFlag==AveragePhases) THEN
              ! Average of phases and amplitudes separately
              MEMRefl(i)%AMEM=(fsum/float(nsum))*dcos(DPhase(dA,dB))
              MEMRefl(i)%BMEM=(fsum/float(nsum))*dsin(DPhase(dA,dB))	               
            ELSEIF (AveFlag==AverageSF) THEN
              ! Average of complex structure factors
              MEMRefl(i)%AMEM=(asum/float(nsum))
              MEMRefl(i)%BMEM=(bsum/float(nsum))     
              MEMRefl(i)%FMEM=SQRT(MEMRefl(i)%AMEM**2+MEMRefl(i)%BMEM**2)
            ENDIF
          ENDIF
!	       write(*,*) 'Zaver:', DPhase(dA,dB),MEMRefl(i)%AMEM,MEMRefl(i)%BMEM
!	       read(*,*)
        END DO
        CALL MEMRefl2SF
      ENDIF
      CALL SF2Refl
    END SUBROUTINE ShiftorAverage

! Derive the space group of the reconstructed density
!**************************************
      SUBROUTINE DeriveSymmetry
!**************************************
      USE SF_Module
      USE SF_SymSearch
      USE SF_Interfaces
      IMPLICIT NONE
      REAL, PARAMETER :: CenterFactor=0.05
      integer :: i,ii,iii,iv,n,nmin,bestcen,twotosdim,AllocLoop,Rorder,AllocErr,NTmpSym,NDerSym, EstimNDerSym
      INTEGER :: prevs,curs
      LOGICAL :: AddCen
      INTEGER :: GoodSymNr,CrSystem, maxfrac, SGpos 
      INTEGER, DIMENSION(SDim) :: StartC,EndC,StartC2orig,StartC2,EndC2orig,EndC2,RunCoo,Rrow,AxOrd
      INTEGER, DIMENSION(1:6) :: OrderSum
      INTEGER, DIMENSION(1:3**SDim,1:SDim,1:SDim) :: PotenR !Potential rows of matrices
      TYPE ShiftScoreType
        INTEGER :: pos, score
      END TYPE ShiftScoreType
      TYPE(ShiftScoreType), ALLOCATABLE, DIMENSION(:) :: ShiftScore
      type(SymOp_type), allocatable, dimension(:) :: TmpSmOp, DerSmOp
      type(SymOp_type) :: DummySmOp
      real, allocatable,dimension(:) :: TmpAgrFac
      real, allocatable,dimension(:,:) :: OrList
      type(SymOpPropsType),allocatable, dimension(:) ::DerSymProps
      type(SymOpPropsType) :: TmpSymProps
      INTEGER, DIMENSION(SDim) ::NRrows !Number of potential rows for each row of R
      REAL, DIMENSION(SDim) :: dvec,dvector,svector
      REAL, DIMENSION(1:RDim) :: qr, qi
      REAL :: rdummy,rdummy2,rdummy3,coincidencers,coincidencersburla,from0to1_old
      REAL, DIMENSION(1:SDim,1:SDim) :: TestR,TestRT,dmat !Rotation matrix and its transpose, dummy matrices
      REAL, ALLOCATABLE, DIMENSION(:) :: RRatList !agreement factors for centering vectors
      INTEGER, DIMENSION(SDim,SDim) :: ITestR
      INTEGER, ALLOCATABLE, DIMENSION(:,:) :: GrpTable
      LOGICAL :: nonrot,nonmax,different,Rounded,FunnyNumber,derCentro
      CHARACTER*256 :: buffer2,FP
      CHARACTER*48 :: SGPart1,SGPart2
      CHARACTER*1  :: CenSymbol
      CHARACTER*17 :: TmpForm='(   i3,   f14.10)'
      CHARACTER*3 :: AxStr
      INTEGER, DIMENSION(1:4,1:6) :: & !Array holding the special directions, the fourth element is a unique number of that direction calculated as a+10*b+100*c
	SpDir = RESHAPE ( &
	 (/ 1, 0, 0, 1,   &
            0, 1, 0, 10,  & 
            0, 0, 1, 100, & 
            1, 1, 0, 11,  & 
            2, 1, 0, 12,  & 
            1, 1, 1, 111  & 
	  /),(/4,6/))
      AxOrd=(/(i,i=1,SDim)/)
      write(TmpForm(2:4),'(i3)') SDim**2
      write(TmpForm(8:10),'(i3)') SDim !Need to store one SDim*SDim matrix + one translation vector
      open(unit=tmpu,file=TmpFile)
      NTmpSym=0
      rdummy=0.0
      !First analyze the density for centering
      CALL GenPatterson(1)
      rdummy=maxval(rho)
      IF (expandedlog) THEN
        write(lgu,'(a)') 'Candidates for centering vectors:'
        buffer2=''
	buffer2(7*SDim/2-5:)='Coordinates'
        buffer2(7*SDim:)='   R-ratio  accepted?'
        write(lgu,'(a)') trim(buffer2)
      ENDIF
      DO AllocLoop=1,2
        NGuessCen=0
        DO i=0,NVoxTot-1
          IF (rho(i)>rdummy*(1.-5.*CenterFactor)) THEN
            call coordinates(i,Rrow)
            !is this a real maximum?
            StartC=-1; EndC=1; RunCoo=StartC
            nonmax=.false.
            DO
              call ArrayLocation(RunCoo+Rrow,ii)
	          IF ((rho(ii)>(rho(i))).and.(ii/=i)) THEN
                nonmax=.true.
	            EXIT
              ENDIF
              IF (.not.NextCoo(SDim,StartC,EndC,RunCoo,AxOrd)) EXIT
            ENDDO
            IF (.not.nonmax) THEN
              !Calculate R=SUM(F allowed)/SUM(F all)
              rdummy2=MEMRefl(1)%AMEM**2/sf2rho%Norm
              DO ii=MStartF+1,MEndF
                IF (.not.IndepSFPos(MEMRefl(ii)%hklFG, MEMRefl(ii)%SFloc)) CYCLE
                rdummy3=sum(MEMRefl(ii)%hklFG*float(Rrow)/float(NVox))
	            IF (abs(rdummy3-nint(rdummy3))<eps) rdummy2=rdummy2+2.*(MEMRefl(ii)%AMEM**2+MEMRefl(ii)%BMEM**2)/sf2rho%Norm
              ENDDO
              IF ((expandedlog).and.(AllocLoop==1)) THEN
                DO iii=1,SDim
                  write(lgu,'(f7.4)', ADVANCE='NO') float(Rrow(iii))/NVox(iii)
	            ENDDO
                write(lgu,'(f7.2)', ADVANCE='NO') rdummy2/rdummy
                IF (rdummy2/rdummy>(1.-CenterFactor)) THEN
                  write(lgu,'(''     yes'')', ADVANCE='NO')
                ELSE
                  write(lgu,'(''     no'')', ADVANCE='NO')
                ENDIF
                write(lgu,*)
              ENDIF
              IF (rdummy2/rdummy>(1.-CenterFactor)) THEN
                NGuessCen=NGuessCen+1
                IF (AllocLoop==2) THEN
                  GuessCen(1:SDim,NGuessCen)=float(Rrow)/float(NVox)
                  rratlist(NGuessCen)=rdummy2/rdummy
!                 write(*,*) rho(i)
!		          call printmat(GuessCen(1:SDim,NGuessCen))
                ENDIF
              ENDIF
            ENDIF
          ENDIF
        ENDDO
        IF (AllocLoop==1) THEN
          ALLOCATE(GuessCen(1:SDim,1:NGuessCen))
          ALLOCATE(RRatList(1:NGuessCen))
        ENDIF
      ENDDO
!     Sort the centering vectors by their R-ratio
      DO i=2,NGuessCen-1 !Start from 2, the first is always 000
        DO ii=i+1,NGuessCen
          IF (rratlist(i)<rratlist(ii)) THEN
            dvec=GuessCen(1:SDim,i)
            GuessCen(1:SDim,i)=GuessCen(1:SDim,ii)
            GuessCen(1:SDim,ii)=dvec
            rdummy2=rratlist(i)
            rratlist(i)=rratlist(ii)
            rratlist(ii)=rdummy2
          ENDIF
        ENDDO
      ENDDO
!     Now check that the guesscen array really forms a translation group
      DO
        DO i=1,NGuessCen
          DO ii=1,NGuessCen
            dvec=GuessCen(1:SDim,i)+GuessCen(1:SDim,ii)
            dvec=dvec-int(dvec)
            different=.true.
            DO iii=1,NGuessCen
              if (sum(abs(dvec-GuessCen(1:SDim,iii)))<eps) THEN
                different=.false.
                exit
              ENDIF
            ENDDO
            IF (different) EXIT
          ENDDO
          IF (different) EXIT
        ENDDO
        IF ((.not.different).or.(NGuessCen==1)) EXIT
        NGuessCen=NGuessCen-1 !Remove the worst centering and retry
      ENDDO

      !Calculate the multiplicity of the centering vectors
      ALLOCATE(GuessCenMult(1:SDim,1:NGuessCen))
      DO i=1,NGuessCen
        GuessCenMult(1:SDim,i)=1
        DO ii=1,SDim
          dvec=0.
          DO
            dvec=dvec+GuessCen(1:SDim,NGuessCen)
            if (abs(dvec(ii)-nint(dvec(ii)))>eps) THEN
              GuessCenMult(ii,i)=GuessCenMult(ii,i)+1
            ELSE
              EXIT
            ENDIF
          ENDDO
        ENDDO
      ENDDO

     IF (expandedlog) write(lgu,*)

      ALLOCATE (ZeroArr(1:SDim))
      twotosdim=2**SDim
      ALLOCATE(P(1:SDim,1:twotosdim),STAT=AllocErr)
      ALLOCATE(HP(1:SDim,1:twotosdim),STAT=AllocErr)
      ALLOCATE(IR(1:SDim,1:SDim),STAT=AllocErr)
      ALLOCATE(RIR(1:SDim,1:SDim),STAT=AllocErr)
      ALLOCATE(srho(0:range**SDim-1),STAT=AllocErr)
      ALLOCATE(TmpSymProps%dir(SDim),TmpSymProps%IntrVec(SDim))
      ALLOCATE(DummySmOp%R(1:SDim,1:SDim))
      ALLOCATE(DummySmOp%t(1:SDim))
      IF (AllocErr/=0) STOP 'Error allocating arrays P,HP, IR and RIR!'
      !Check each row of the potential rotation matrix
      NRrows=0
      PotenR=0
      DO i=1,SDim
        StartC=-1
	EndC=1
	Rrow=StartC
	DO
          ! For each row of the matrix rGir=Gi(i,i), because it is mapping certain linear combination of the reciprocal vectors onto the basis vector
	  dvec=multmat(Rrow,GSi)
          rdummy=sum(dvec*Rrow)
	  if (abs(rdummy-GSi(i,i))<eps) THEN
	    NRrows(i)=NRrows(i)+1
	    PotenR(NRrows(i),1:SDim,i)=Rrow
	  ENDIF
	  IF (.not.NextCoo(SDim,StartC,EndC,Rrow,AxOrd)) EXIT
        ENDDO
      ENDDO 
      CALL PrepareForSym(2) !prepare the variables for the symsearch
!      DO i=1,SDim
!        write(*,*) 'Potential rows number',i,':'
!        DO ii=1,NRrows(i)
!	  write(*,*) (PotenR(ii,iii,i),iii=1,SDim)
!	ENDDO
!      ENDDO
      StartC=1
      EndC=NRrows
      Rrow=StartC
      DO
        DO i=1,SDim
	  DO ii=1,SDim
            TestR(i,ii)=float(PotenR(Rrow(i),ii,i))
            TestRT(ii,i)=float(PotenR(Rrow(i),ii,i))	  
	  ENDDO
	ENDDO
        IF (IsUnitMatrix(TestR)) GOTO 1000 !Trivial symmetry operation
        ! Test, if this matrix fits
	dmat=multmat(TestRT,GS)
	dmat=multmat(dmat,TestR)
	dmat=multmat(dmat,GSi)
	IF (IsUnitMatrix(dmat)) THEN
          !This is a potential rotational part of the matrix. Find the corresponding translational parts
	  CALL ProjectionOperator(TestR,TestRT,ROrder,SDim) !TestRT is the projection operator P=sum_1^n R^i, where n is the order of R
	  !Now we have to search for the shift vector. Allowed shift vectors fulfill the condition Rorder*P*tau=N, where N is any lattice translation vector
	  StartC2Orig=-Rorder
	  EndC2Orig=Rorder
          !Chceck for zero lines of the projecton operator to save time.A zero line means the corresponding component of the vector must be zero	    
	  where (maxval(abs(TestRT),dim=2)<eps) EndC2=0
!          !Check for zero columns. A zero column means the corresponding component can be arbitrary - use 0
!	  where (maxval(abs(TestRT),dim=1)<eps) EndC2=0
	            
	  !Cycle over the centering vectors
	  DO AllocLoop=1,NGuessCen
            StartC2=StartC2Orig*GuessCenMult(1:SDim,AllocLoop)+1
            EndC2=EndC2Orig*GuessCenMult(1:SDim,AllocLoop)-1
	    RunCoo=StartC2
!            write(*,*) Allocloop,NGuessCen
!            read(*,*)
	    DO !loop over all possible combination of coordinates of intrinsic vector
	      dvector=float(RunCoo)/float(GuessCenMult(1:SDim,AllocLoop)*Rorder) !potential translation vector tau
!              write(*,*) dvector
              !Does ROrder*dvector arrive at the current centering?
	      dvec=ROrder*dvector-GuessCen(1:SDim,AllocLoop)
	      dvec=From0to1(dvec)
	      IF (maxval(dvec)<eps) THEN
	        !Is dvector a pure intrinsic vector?
		IF (maxval(abs(dvector-multmat(TestRT,dvector)))<eps) THEN
                  !This is an acceptable vector. However, we have a problem here, that too many equivalent vectors penetrate here - let's filter out at least some of them
                  nonrot=.true.
                  IF (maxval(abs(dvector-From0to1(dvector)))>eps) THEN !This vector contains negative components, let's check, if they are essential (Remark: they are NEVER essential, unless the symmetry operation is a diagonal one)
                    dvec=From0to1(dvector)
                    IF (maxval(abs(dvec-multmat(TestRT,dvec)))<eps) nonrot=.false. !If the vector mapped to 0-1 is still an intrinsic vector, forget about this dvec
                  ENDIF
                  IF (nonrot) THEN
                    !Save the operation to the temporary file
                    write(tmpu,TmpForm) ((nint(TestR(i,ii)),ii=1,SDim),i=1,SDim),(dvector(i),i=1,SDim)
                    NTmpSym=NTmpSym+1                    
                  ENDIF
		ENDIF 
              ENDIF
	      IF (.not.NextCoo(SDim,StartC2,EndC2,RunCoo,AxOrd)) EXIT
	    ENDDO
	  ENDDO
	ENDIF
1000	IF (.not.NextCoo(SDim,StartC,EndC,Rrow,AxOrd)) EXIT        
      ENDDO
      close(tmpu)   
      ALLOCATE(TmpSmOp(1:NTmpSym), STAT=AllocError)
      DO cnt1=1,NTmpSym
        ALLOCATE(TmpSmOp(cnt1)%R(SDim,SDim),TmpSmOp(cnt1)%t(SDim),STAT=AllocError)
      ENDDO
      IF (AllocError/=0) STOP 'Error allocating array TmpSmOp! Too little memory?'
      ALLOCATE(TmpAgrFac(1:NTmpSym),STAT=AllocError)
      IF (AllocError/=0) STOP 'Error allocating array TmpAgrFac! Too little memory?'

      TmpAgrFac=0.0
      open(unit=tmpu,file=TmpFile)
      DO AllocLoop=1,NTmpSym
        read(tmpu,TmpForm) ((TmpSmOp(AllocLoop)%R(i,ii),ii=1,SDim),i=1,SDim),(TmpSmOp(AllocLoop)%t(i),i=1,SDim)
      ENDDO
      close(tmpu,STATUS='DELETE')

      !Now select from each group just theTmpSmOp with the "nicest" vector
      DO i=1,NTmpSym-1
        IF (TmpAgrFac(i)>9000.) CYCLE
        !Find the last symop with the same rotation part
  	ii=i
	DO
	  IF (any(TmpSmOp(ii+1)%R/=TmpSmOp(i)%R)) EXIT
	  ii=ii+1
	  if (ii==NTmpSym) EXIT
	ENDDO
	IF (i==ii) CYCLE !Only one symop with the same rotation part found
	bestcen=i !bestcen holds the number of the representative symop of the group
	DO iii=i+1,ii
          IF (TmpAgrFac(iii)>9000.) CYCLE
	  IF (.not.SymOpsJustShifted(TmpSmOp(bestcen)%R,TmpSmOp(bestcen)%t,TmpSmOp(iii)%R,TmpSmOp(iii)%t,GuessCen(1:SDim,1:NGuessCen))) &
	      CYCLE !Different symops
          IF (count(abs(TmpSmOp(iii)%t)>eps)>count(abs(TmpSmOp(bestcen)%t)>eps)) THEN
	    TmpAgrFac(iii)=10000.
          ELSEIF (count(abs(TmpSmOp(iii)%t)>eps)<count(abs(TmpSmOp(bestcen)%t)>eps)) THEN
	    TmpAgrFac(bestcen)=10000.
	    bestcen=iii
	  ELSEIF (sum(abs(TmpSmOp(iii)%t))>sum(abs(TmpSmOp(bestcen)%t))) THEN
	    TmpAgrFac(iii)=10000.
	  ELSEIF (sum(abs(TmpSmOp(iii)%t))<sum(abs(TmpSmOp(bestcen)%t))) THEN
	    TmpAgrFac(bestcen)=10000.
	    bestcen=iii	    
	  ELSEIF (sum(TmpSmOp(iii)%t)<=sum(TmpSmOp(bestcen)%t)) THEN !This should select tau that are "more positive", like 0 0 -1/4 is worse than 0 0 1/4
	    TmpAgrFac(iii)=10000.
	  ELSE
	    TmpAgrFac(bestcen)=10000.
	    bestcen=iii
	  ENDIF
	ENDDO
      ENDDO

      !Time to evaluate the position and agreement factors of the symops
      ITestR=UnitMat(SDim)
!      open(67, file='ccvsphisym.dat',position='APPEND')
      DO AllocLoop=1,NTmpSym
	write(*,'(i4,a1,a1)',ADVANCE='NO') nint(100.*float(AllocLoop)/float(NTmpSym)),'%',char(13)
        IF (TmpAgrFac(AllocLoop)>9000.) cycle 
	!write the curent operation to the (NSym+NTestSym+1)th symop
	n=NSym+NTestSym+1
	CALL CopySymOp(TmpSmOp(AllocLoop),SmOp(n))
        SmOp(n)%t=0.
	IF (any(ITestR/=TmpSmOp(AllocLoop)%R)) THEN
	  different=.true.
	ELSE
	  different=.false.
	ENDIF
	!Find the position of the operation
	CALL GetDandS(n,dvector,svector,2,different,-TmpSmOp(AllocLoop)%t) !negative intrinsic part, because GenPatterson returns by mistake an inverted image. Uz je to v programu tak zapletene, ze se to neda jednoduse opravit, ale casem to budu muset udelat
	TmpSmOp(AllocLoop)%t=from0to1(-dvector) !Problem with the GenPatterson function - returns inverted image.
        call MEMRefl2SF
	dvec=0.0
	CALL CopySymOp(TmpSmOp(AllocLoop),SmOp(n))
        TmpAgrFac(AllocLoop)=CoincidenceRS(n,dvec)
!        CoincCode=1
!	rdummy=CoincidenceRS(n,dvec)
!	CoincCode=2
!        buffer=' '
!	CALL FillSymProps(TmpSmOp(AllocLoop),TmpSymProps,RDim)
!	buffer2=SymOpSymbol(TmpSmOp(AllocLoop),TmpSymProps,RDim,2)
!	write(67,'(a,''    '',a15,2f9.4)') trim(FileBase),trim(buffer2),rdummy,CoincidenceRS(n,dvec)
!	CoincCode=1
	ITestR=TmpSmOp(AllocLoop)%R !Remember last rotation part used to generate Patterson
      ENDDO
!      write(67,*)
!      write(67,*)
!      CLOSE (67)
! Sort the symops by the agreement factor
      DO i=NTmpSym,2,-1
        DO ii=1,i-1
	  IF (TmpAgrFac(ii)>TmpAgrFac(ii+1)) THEN
	    call CopySymOp(TmpSmOp(ii),DummySmOp)
	    call CopySymOp(TmpSmOp(ii+1),TmpSmOp(ii))
	    call CopySymOp(DummySmOp,TmpSmOp(ii+1))
            rdummy=TmpAgrFac(ii+1)
	    TmpAgrFac(ii+1)=TmpAgrFac(ii)
	    TmpAgrFac(ii)=rdummy
	  ENDIF
	ENDDO
      ENDDO
! Write out the info
      IF (NGuessCen>1) THEN
        if (terminal) write(*,'(a)') 'Centering vectors:'
        write(30,'(a)') 'Centering vectors:'      
        DO i=1,NGuessCen
          DO ii=1,SDim
            if (terminal) write(*,'(f7.3)',ADVANCE='NO') GuessCen(ii,i)
            write(30,'(f7.3)',ADVANCE='NO') GuessCen(ii,i)
	  ENDDO
          IF (terminal) write(*,*)
	  write(30,*)	
        ENDDO
        write(30,*)
        if (terminal) write(*,*)
      ENDIF
      if (terminal) THEN
        write(*,'(a)') 'Symmetry operations compatible with the lattice and centering:'
        write(*,'(a)') '                              Symmetry operation             agreement factor'
      ENDIF
      write(30,'(a)') 'Symmetry operations compatible with the lattice and centering:'
      write(30,'(a)') '                              Symmetry operation             agreement factor'
      !Commented out is the possibility to print out ALL the symops
!      DO AllocLoop=1,NTmpSym
!        buffer=' '
!	CALL FillSymProps(TmpSmOp(AllocLoop),TmpSymProps,RDim)
!	buffer2=SymOpSymbol(TmpRMat(1,1,AllocLoop),TmpTvec(1,AllocLoop),TmpSymProps,RDim,2)
!	write(buffer(16-len_trim(buffer2):),'(a)') trim(buffer2)
!        buffer=trim(buffer)//': '//SymOpIntTab(TmpRMat(1,1,AllocLoop),TmpTVec(1,AllocLoop),'x123')
!        write(buffer(len_trim(buffer)+1:),'(f9.3)') TmpAgrFac(AllocLoop)
!        ii=len_trim(buffer)+4
!        DO i=1,max(1,nint((100.-TmpAgrFac(AllocLoop))/2.))
!          write(buffer(ii+i:ii+i),'(a1)') 'X'
!        ENDDO
!        if (terminal) write(*,'(a)') trim(buffer)
!        write(30,'(a)') trim(buffer)
!      ENDDO

      GoodSymNr=0
      DO AllocLoop=1,NTmpSym
        IF (TmpAgrFac(AllocLoop)>9999.) CYCLE
        buffer=' '
	CALL FillSymProps(TmpSmOp(AllocLoop),TmpSymProps,RDim)
	buffer2=SymOpSymbol(TmpSmOp(AllocLoop),TmpSymProps,RDim,2)
	write(buffer(16-len_trim(buffer2)+1:),'(a)') trim(buffer2)
	dvec=TmpSmOp(AllocLoop)%t
	TmpSmOp(AllocLoop)%t=from0to1(TmpSymProps%IntrVec)
        buffer=trim(buffer)//': '//SymOpIntTab(TmpSmOp(AllocLoop),'x123')
	TmpSmOp(AllocLoop)%t=dvec
        write(buffer(len_trim(buffer)+1:),'(f9.3)') TmpAgrFac(AllocLoop)
        ii=len_trim(buffer)+4
        DO i=1,max(1,nint((100.-TmpAgrFac(AllocLoop))/2.))
          write(buffer(ii+i:ii+i),'(a1)') 'X'
        ENDDO
        if (terminal) write(*,'(a)') trim(buffer)
        write(lgu,'(a)') trim(buffer)
        IF (TmpAgrFac(AllocLoop)<LimitAgrFac) GoodSymNr=GoodSymNr+1
      ENDDO
      IF (GoodSymNr==0) THEN
        IF (terminal) write(*,*) 'No symmetry operations with agreement factor below',LimitAgrFac,'detected!'
        write(30,'(a,f7.2,a)') 'No symmetry operations with agreement factor below',LimitAgrFac,'detected!'
      ENDIF
      !Allocate derived symmetry operations. One problem still remains - how to estimate the total number of symmetry operators? Factor 3 used here is just a (hopefully) sufficient margin
      EstimNDerSym=(GoodSymNr+1)*NGuessCen*3
      ALLOCATE(DerSmOp(1:EstimNDerSym),STAT=AllocError)
      DO cnt1=1,EstimNDerSym
        ALLOCATE(DerSmOp(cnt1)%R(SDim,SDim),DerSmOp(cnt1)%t(SDim),STAT=AllocError)
      ENDDO
      IF (AllocError/=0) STOP 'Error allocating array DerSmOp! Too little memory?'
      !First Symop is always unit matrix - let's fill it in
      DerSmOp(1)%t=0.0
      DerSmOp(1)%R=UnitMat(SDim)
      NDerSym=1
      !Now loop over the symops and build the space group
      DO i=1,NTmpSym
        IF (TmpAgrFac(i)>LimitAgrFac) CYCLE
	!Test if this symop is new
	different=.true.
	DO ii=1,NDerSym
          DO iii=1,NGuessCen
	    dvec=TmpSmOp(i)%t+GuessCen(1:SDim,iii)
	    IF (EqualOp(TmpSmOp(i)%R,dvec,DerSmOp(ii)%R,DerSmOp(ii)%t,SDim)) THEN
	      different=.false.
	      exit
	    ENDIF
	  ENDDO
          if (.not.different) EXIT
	ENDDO
	IF (.not.different) CYCLE
        !If we are here, then curent TmpSymOp is not in the space group yet - let's add it to the list of symops and complete the space group
        n=NDerSym !Remember the current NDerSym
        NDerSym=NDerSym+1
        IF (NDerSym>EstimNDerSym) THEN
          CALL AllocateOneMoreDerSymOp
          EstimNDerSym=NDerSym
        ENDIF	
	CALL CopySymOp(TmpSmOp(i),DerSmOp(NDerSym))
	prevs=1
	DO
	  prevs=prevs+1
	  IF (prevs>NDerSym) EXIT
	  curs=n
	  DO
	    curs=curs+1
	    if (curs>NDerSym) EXIT
            CALL ProdOp(DerSmOp(curs)%R,DerSmOp(curs)%t,DerSmOp(prevs)%R,DerSmOp(prevs)%t,ITestR,dvec,SDim)
            !Check if we didn't get additional centering (appears often for modulated structures, if the modulation is weak!)
            IF (IsUnitMatrix(ITestR)) THEN
	      addcen=.true.
              DO iv=1,NGuessCen
	        dvector=dvec+GuessCen(1:SDim,iv)
		rdummy=sum(abs(dvector-nint(dvector)))
	        IF (rdummy<0.05) THEN
	          addcen=.false.
	          EXIT
	        ENDIF
	      ENDDO
	      IF (addcen) THEN
	        NDerSym=n
	        EXIT
	      ENDIF
            ELSE
	      AddCen=.false.
	    ENDIF
	    !Check if the new operation is really new:
	    different=.true.
	    DO iii=1,NDerSym
	      DO iv=1,NGuessCen
	        IF (EqualOp(DerSmOp(iii)%R,DerSmOp(iii)%t,ITestR,dvec+GuessCen(1:SDim,iv),SDim)) THEN
	          different=.false.
	          EXIT
	        ENDIF
	      ENDDO
              if (.not.different) EXIT
	    ENDDO
	    IF (different) THEN !New SymOp
	      NDerSym=NDerSym+1
	      IF (NDerSym>EstimNDerSym) THEN
	        CALL AllocateOneMoreDerSymOp
                EstimNDerSym=NDerSym
	      ENDIF
	      DerSmOp(NDerSym)%t=dvec
              DerSmOp(NDerSym)%R=ITestR
	    ENDIF
          ENDDO
          IF (AddCen) EXIT
        ENDDO
      ENDDO
      IF (terminal) write(*,*)

      !Create and fill the array of properties of the symops
      ALLOCATE(DerSymProps(1:NDerSym))
      maxfrac=maxval(GuessCenMult)
      DO i=1,NDerSym
        ALLOCATE(DerSymProps(i)%dir(SDim),DerSymProps(i)%IntrVec(SDim))
        CALL FillSymProps(DerSmOp(i),DerSymProps(i),RDim)
	maxfrac=max(maxfrac,DerSymProps(i)%introrder)
      ENDDO
      if (terminal) write(*,*)
!     Shift the origin so that the origin-dependent shifts are not completely arbitrary, but multiples of n/NDerSym
      dvector=0.0
      DO i=1,NDerSym
        dvector=dvector+DerSmOp(i)%t
!        call printsymop(DerSmOp(i))
      ENDDO
!      write(*,*) 'Sum of t:',dvector
      call ShiftSGOrigin(DerSmOp,-1.*dvector/NDerSym,.true.)
!      DO i=1,NDerSym
!        call printsymop(DerSmOp(i))
!      ENDDO      
!      read(*,*)
!***************** The space group is determined, but the origin is still almost arbitrary, and the t-values may have some discrepancy.
!***************** We have to make sure that the t-vectors are consistent
      !Fill the Group table
!      ALLOCATE(GrpTable(1:NDerSym,1:NDerSym))
!      GrpTable=0
!      DO i=1,NDerSym
!        DO ii=1,NDerSym
!          ItestR=multmat(DerSmOp(ii)%R,DerSmOp(i)%R)
!          DO iii=1,NDerSym
!            IF (all(ItestR==DerSmOp(iii)%R)) THEN
!              GrpTable(i,ii)=iii
!              EXIT
!            ENDIF
!          ENDDO          
!        ENDDO
!      ENDDO
!      !Make a space-group table with t-vectors
!      DO i=1,NSym
!        DO ii=1,NSym
!          dvec=multmat(DerSmOp(ii)%R,DerSmOp(i)%t)+DerSmOp(ii)%t
!          write(*,*) i,ii,GrpTable(i,ii)
!          DO iii=1,SDim
!            write(*,'(3f10.5)') from0to1(dvec(iii),bigeps),from0to1(DerSmOp(GrpTable(i,ii))%t(iii),bigeps), &
!              from0to1(dvec(iii),bigeps)-from0to1(DerSmOp(GrpTable(i,ii))%t(iii),bigeps)
!          ENDDO
!          write(*,*)
!        ENDDO
!      ENDDO
!      READ(*,*)
!      !Setup a matrix for solution of linear equations
!      DEALLOCATE(GrpTable)
!*****************
      RRow=0 !Holds the directions that have been already shifted
      !Sort the symops by the signifficancy of them for the determination of the origin shift
      ALLOCATE (ShiftScore(1:NDerSym))
      ShiftScore%pos=(/(i,i=1,NDerSym)/)
      ShiftScore%score=0
      ShiftScore(1)%score=100000
      DerCentro=.false.
      DO i=2,NDerSym
	IF ((DerSymProps(i)%trace==-SDim).and.(DersymProps(i)%rank==SDim)) THEN !Inversion center
	  ShiftScore(i)%score=10000 ! Use the inversion center to shift the SG
          DerCentro=.true.
	ELSE
	  n=0
	  DO ii=1,RDim
	    n=n+abs(DerSymProps(i)%dir(ii))
	  ENDDO
	  ShiftScore(i)%score=100*(RDim+1-n) !The simpler the direction, the better
	  ShiftScore(i)%score=ShiftScore(i)%score+10*DerSymProps(i)%order !The higher the order, the better
	  IF (DerSymProps(i)%order>3) THEN
            ShiftScore(i)%score=ShiftScore(i)%score+DerSymProps(i)%det !For axes I prefer simple axes before rotoinversions	  
	  ELSE
            ShiftScore(i)%score=ShiftScore(i)%score-DerSymProps(i)%det !Mirror planes are preferred over two-fold axes
	  ENDIF
	ENDIF
      ENDDO
      !Bubblesort ShiftScore by decreasing Score
      DO i=NDerSym,2,-1
        DO ii=1,i-1
	  IF (ShiftScore(ii)%score<ShiftScore(ii+1)%score) THEN
            n=ShiftScore(ii)%pos
	    ShiftScore(ii)%pos=ShiftScore(ii+1)%pos
	    ShiftScore(ii+1)%pos=n
            n=ShiftScore(ii)%score
            ShiftScore(ii)%score=ShiftScore(ii+1)%score
            ShiftScore(ii+1)%score=n
          ENDIF
        ENDDO	
      ENDDO
      
      DO AllocLoop=1,NDerSym
        i=ShiftScore(AllocLoop)%pos
	different=.false.
	DO ii=1,SDim
          dvector(ii)=DerSmOp(i)%t(ii)-DerSymProps(i)%IntrVec(ii)	  
	  IF ((RRow(ii)==0).and.(abs(dvector(ii))>bigeps)) THEN
	    different=.true.
	  ELSEIF (RRow(ii)==1) THEN !This direction has been already shifted, check, if the shift was enough
            nonrot=.true.
            DO iii=1,maxfrac
	      if ((iii==5).and.(SDim<4)) cycle
	      if (abs((DerSmOp(i)%t(ii)*float(iii)-float(nint(DerSmOp(i)%t(ii)*float(iii)))))<bigeps) THEN
	        nonrot=.false.
	        EXIT
	      ENDIF
	    ENDDO
	    IF (nonrot) different=.true.
	  ENDIF
        ENDDO
	IF (.not.different) THEN
!          write(*,*) i
!	  write(*,*) 'Already properly shifted'
	  CYCLE
	ENDIF	
        dmat=UnitMat(SDim)-DerSmOp(i)%R
        ZeroArr=0
        ! Find the position of the symmetry operation
 	call sgelsy(SDim,SDim,1,dmat,SDim,dvector,SDim,ZeroArr,0.01,n,work,lwork,AllocErr)
        svector=dvector
        DO ii=1,SDim
          if (RRow(ii)==1) svector(ii)=0.0 !protect agains reshifting of already shifted directions
	  if (abs(svector(ii))>1.5/NVox(ii)) RRow(ii)=1
	ENDDO
	!New t_new after an origin shift by s of an operation {R|t} can be calculated as t_new=(R-I)s+t
        FunnyNumber=.false. !Detects "non-conventional" translational parts. If false at the end of the loop, no need for further shifting
        DO ii=2,NDerSym
          dvector=multmat(DerSmOp(ii)%R,svector)
          DerSmOp(ii)%t=DerSmOp(ii)%t-svector+dvector
	  DerSmOp(ii)%t=DerSmOp(ii)%t-nint(DerSmOp(ii)%t)
	  DO iii=1,SDim
	    !Now round the TVec if it is close to a "typical" Tvec value
            Rounded=.false.
            if (abs(DerSmOp(ii)%t(iii))<1.5/NVox(iii)) THEN
	      DerSmOp(ii)%t(iii)=0.0
	      Rounded=.true.
	    ELSE
	      DO iv=2,maxfrac
	        IF (abs(DerSmOp(ii)%t(iii)-1./float(iv))<1.5/NVox(iii)) THEN
	          DerSmOp(ii)%t(iii)=1./float(iv)
		  Rounded=.true.
		  EXIT
	        ELSEIF (abs(DerSmOp(ii)%t(iii)+1./float(iv))<1.5/NVox(iii)) THEN
	          DerSmOp(ii)%t(iii)=-1./float(iv)
                  Rounded=.true.
		  EXIT
		ENDIF
              ENDDO
	    ENDIF
	    IF (.not.Rounded) FunnyNumber=.true.
	  ENDDO
        ENDDO
        DO ii=1,NDerSym
          CALL FillSymProps(DerSmOp(ii),DerSymProps(ii),RDim)
        ENDDO 
        different=.false.	
	DO ii=1,RDim
	  IF (RRow(ii)/=1) Different=.true.
	ENDDO
        IF ((.not.FunnyNumber).and.(.not.different)) THEN
!	  write(*,*) 'All vectors are OK'
!	  EXIT
	ENDIF
        IF ((AllocLoop==2).and.(DerCentro)) EXIT !If we have inversion center, we want to shift only to the inverion center
      ENDDO
      !Now change all the translational parts to positive translations
      DO i=2,NDerSym
        DerSmOp(i)%t=from0to1(DerSmOp(i)%t,bigeps)
      ENDDO

!      Write(*,*) 'Po posunuti na rozumny stred:'
!      DO ii=1,NDerSym
!!        i=ShiftScore(ii)%pos
!        i=ii
        !buffer=' '
!	buffer2=SymOpSymbol(DerSmOp(i),DerSymProps(i),RDim,2)
!	write(buffer(16-len_trim(buffer2)+1:),'(a)') trim(buffer2)
!	buffer=trim(buffer)//': '//SymOpIntTab(DerSmOp(i),'x123')
!        write(30,'(a)') trim(buffer)
!        if (terminal) write(*,'(a)') trim(buffer)
!      ENDDO

! At this moment we have shifted the SG to a reasonable origin (to an inversion center, if any, or to another nice position. Now let's test all potentially intersting alternative positions and select the point with highest site symmetry
      IF (DerCentro) THEN !Possible alternative inversion center are in the middle of all lattice vectors
        allocate(OrList(1:SDim,1:2**SDim+NGuessCen-1))
	OrList=0.
	OrList(1:SDim,1:2**SDim)=P/2.0
	n=twotosdim
        DO i=2,NGuessCen
	  different=.true.
	  DO ii=1,twotosdim
	    IF (maxval(abs(Orlist(1:SDim,ii)-GuessCen(1:SDim,i)))<eps) THEN
	      OrList(1:SDim,ii)=OrList(1:SDim,ii)/2. !Replace the centering vector by its half
	      different=.false.
	    ENDIF
	  ENDDO
	  IF (different) THEN
	    n=n+1
	    Orlist(1:SDim,n)=GuessCen(1:SDim,i)/2.
	  ENDIF
	ENDDO
        nmin=0 !holds the best site symmetry        
	DO i=1,n
	  ii=SiteSymmetry(OrList(1:SDim,i),DerSmOp,GuessCen)
          IF (ii>nmin) THEN
	    nmin=ii
	    svector=OrList(1:SDim,i) !holds the best origin
	  ENDIF
	ENDDO
      ELSE !For non-centrosymmetric SG do nothing special at the moment
        svector=0.0
      ENDIF      
      !And now shift to the best origin
      DO ii=2,NDerSym
        dvector=multmat(DerSmOp(ii)%R,svector)
        DerSmOp(ii)%t=DerSmOp(ii)%t-svector+dvector
        DerSmOp(ii)%t=DerSmOp(ii)%t-nint(DerSmOp(ii)%t)
      ENDDO
      !change all the translational parts to positive translations
      DO i=2,NDerSym
        DerSmOp(i)%t=from0to1(DerSmOp(i)%t,bigeps)
        CALL FillSymProps(DerSmOp(i),DerSymProps(i),RDim)
      ENDDO
      !And the last thing to do is to change the translation parts to a) eliminate superfluous centering vectors b) make the intrinsic part as simple as possible
      IF (NGuessCen>1) THEN
        DO i=2,NDerSym
	  DO ii=2,NGuessCen
            dvector=from0to1(DerSmOp(i)%t+GuessCen(1:SDim,ii))
	    !get the intrinsic vector from the new vector
            CALL ProjectionOperator(float(DerSmOp(i)%R),dmat,iii,SDim)
            svector=multmat(dmat,dvector)	  
	    !now compare the two
	    IF (maxval(abs(from0to1(svector-DerSymProps(i)%IntrVec)))<eps) THEN !intr vectors equal
	      IF (count(abs(DerSmOp(i)%t)>eps)>count(abs(dvector)>eps)) DerSmOp(i)%t=dvector
	    ELSE
              IF (count(abs(DerSymProps(i)%IntrVec)>eps)>count(abs(svector)>eps)) THEN
                DerSmOp(i)%t=dvector
              ELSEIF(count(abs(DerSymProps(i)%IntrVec)>eps)==count(abs(svector)>eps)) THEN
                IF (sum(abs(DerSymProps(i)%IntrVec))>sum(abs(svector))) THEN
                  DerSmOp(i)%t=dvector
	        ELSEIF (sum(abs(DerSymProps(i)%IntrVec))==sum(abs(svector))) THEN
	          IF (sum(svector)<sum(DerSymProps(i)%IntrVec)) THEN
		    DerSmOp(i)%t=dvector
	          ELSEIF (sum(svector)==sum(DerSymProps(i)%IntrVec)) THEN
		    DO iv=1,SDim
		      IF ((abs(svector(iv))>eps).and.(abs(DerSymProps(i)%IntrVec(iv))<eps)) THEN
                        DerSmOp(i)%t=dvector
			EXIT
		      ELSEIF ((abs(svector(iv))<eps).and.(abs(DerSymProps(i)%IntrVec(iv))>eps)) THEN
		        EXIT
		      ENDIF
		    ENDDO
		  ENDIF
		ENDIF
	      ENDIF
	    ENDIF
	  ENDDO
        ENDDO
      ENDIF
      call FramedLabel(2,'Space group derived from the symmetry operations:')
      IF (terminal) THEN
        write(*,'(a)') '*************************************************'
        write(*,'(a)') 'Space group derived from the symmetry operations:'
        write(*,'(a)') '*************************************************'
      ENDIF
      FP=''
!      sgfp_verbosity=3
      !If 3D, check the database
      IF ((SDim==3).and.(RDim==3)) THEN
        CALL PrivilegedOrigin(DerSmOp(1:NDerSym),GuessCen(1:SDim,1:NGuessCen),DVec)
	    CALL SGFP(DerSmOp(1:NDerSym),GuessCen(1:SDim,1:NGuessCen),DVec,FP)
	SGpos=0
        Buffer=FP(1:index(FP,'(')-2)
	DO i=1,SGlistLength
	  IF (Buffer==SGlist(i)%SGFP) THEN
	    IF (SGpos==0) SGpos=i
	    IF (SGlist(i)%std=='S') THEN
              SGpos=i
              EXIT
	    ENDIF
	  ENDIF
	ENDDO
      ELSE
        IF (SDim<5) THEN
          CALL PrivilegedOrigin(DerSmOp(1:NDerSym),GuessCen(1:SDim,1:NGuessCen),DVec)
	  CALL SGFP(DerSmOp(1:NDerSym),GuessCen(1:SDim,1:NGuessCen),DVec,FP)
        ENDIF
        SGpos=-1
      ENDIF
      IF (SGpos>0) THEN
        write(lgu,'(a)') '    HM symbol:     '//trim(SGlist(SGpos)%HM)
        write(lgu,'(a)') '    Hall symbol:   '//trim(SGlist(SGpos)%Hall)
        write(lgu,'(a)') '    Fingerprint:   '//trim(SGlist(SGpos)%SGFP)//' '//trim(SGlist(SGpos)%OrShift)
	SGsymbol=SGlist(SGpos)%HM
   	IF (terminal) THEN
   	  write(*,'(a)') '    HM symbol:     '//trim(SGlist(SGpos)%HM)
          write(*,'(a)') '    Hall symbol:   '//trim(SGlist(SGpos)%Hall)
          write(*,'(a)') '    Fingerprint:   '//trim(SGlist(SGpos)%SGFP)//' '//trim(SGlist(SGpos)%OrShift)
   	ENDIF
	CALL DeallocateSG(DerSmOp)
	DEALLOCATE(GuessCen)
	CALL SGFromHall(SGlist(SGpos)%Hall,NGuessCen,GuessCen,NDerSym,DerSmOp)
      ELSE
   	!Estimate the space group symbol
   	CALL TentativeHMSymbol(DerSmOp(1:NDerSym),GuessCen,QVec,RDim,SGsymbol)
        IF (SGPos==0) THEN !Non-standard 3D space group
   	  write(lgu,'(a)') 'The space group was not found in the database.'
        ENDIF
        IF (len_trim(SGsymbol)/=0) THEN
          write(lgu,'(a)') '    Tentative HM symbol: '//trim(SGsymbol)
   	  IF (terminal) THEN
   	    write(*,'(2a)')'    Tentative HM symbol: ',SGSymbol
   	  ENDIF
        ENDIF
   	IF (len_trim(FP)/=0) THEN
          write(lgu,'(a)') '    Fingerprint:         '//trim(FP)	  
   	  IF (terminal) THEN
            write(*,'(a)') '    Fingerprint:         '//trim(FP)	  
   	  ENDIF
   	ENDIF
      ENDIF   
      CurRhoProps%DerSGSym=trim(SGSymbol)
      !The order of the operatos might have changed -> update the symprops
      DO i=1,NDerSym
        CALL FillSymProps(DerSmOp(i),DerSymProps(i),RDim)
      ENDDO

      IF (NGuessCen>1) THEN
        write(30,'(a)') '    Centering vectors:'
        IF (terminal) write(*,'(a)') '    Centering vectors:'
        DO i=1,NGuessCen
          write(30,'(a)',ADVANCE='NO') '    '
          if (terminal) write(*,'(a)',ADVANCE='NO') '    '
	  DO ii=1,SDim
            write(30,'(f7.3)',ADVANCE='NO') GuessCen(ii,i)
            IF (terminal) write(*,'(f7.3)',ADVANCE='NO') GuessCen(ii,i)
	  ENDDO
	  write(30,*)
	  if (terminal) write(*,*)
        ENDDO
      ENDIF
      write(30,'(a)') '    Symmetry operations:'
      IF (terminal) THEN
        write(*,'(a)') '    Symmetry operations:'
      ENDIF
      
      DO ii=1,NDerSym
!        i=ShiftScore(ii)%pos
        i=ii
	buffer=' '
	buffer2=SymOpSymbol(DerSmOp(i),DerSymProps(i),RDim,2)
	write(buffer(16-len_trim(buffer2)+1:),'(a)') trim(buffer2)
	buffer=trim(buffer)//': '//SymOpIntTab(DerSmOp(i),'x123')
        write(30,'(a)') trim(buffer)
        if (terminal) write(*,'(a)') trim(buffer)
      ENDDO
!      write(lgu,'(a)') '-------------------------------------------------'
      write(lgu,*)
      IF (terminal) THEN
!       write(*,'(a)') '*************************************************'
        write(*,*)
      ENDIF


      IF (derivesym=='use') THEN !copy the DerSym operations to the RMat and TVec arrays, so that thay can be used in the symmetry averaging procedure
        DO cnt1=1,size(SmOp)
	  DEALLOCATE(SmOp(cnt1)%R)
	  DEALLOCATE(SmOp(cnt1)%t)
	ENDDO
        DEALLOCATE(SmOp,Center)
	NSymPrim=NDerSym
	NSym=NDerSym*NGuessCen
	NCenter=NGuessCen
        ALLOCATE(SmOp(NSym+1))
	DO cnt1=1,NSym+1
	  ALLOCATE(SmOp(cnt1)%R(SDim,SDim))
	  ALLOCATE(SmOp(cnt1)%t(SDim))
	ENDDO
        ALLOCATE(Center(1:SDim,1:NCenter))
        Center=GuessCen
	DO cnt1=1,NDerSym
          DO cnt2=1,NCenter
	    DO i=1,SDim
	      DO ii=1,SDim
	        SmOp(cnt1+NDerSym*(cnt2-1))%R(i,ii)=DerSmOp(ShiftScore(cnt1)%pos)%R(i,ii)
              ENDDO
              SmOp(cnt1+NDerSym*(cnt2-1))%t(i)=From0to1_old(DerSmOp(ShiftScore(cnt1)%pos)%t(i)+GuessCen(i,cnt2),eps)
	    ENDDO
	  ENDDO
	ENDDO
	IF (ALLOCATED(OrList)) DEALLOCATE(OrList)
	CALL FindGenerators
	CALL FillSGProps
      ENDIF


9999  DEALLOCATE(P)
      DEALLOCATE(HP)
      DEALLOCATE(IR)
      DEALLOCATE(RIR)
      DEALLOCATE(ZeroArr)
      DEALLOCATE(srho)
      DEALLOCATE(GuessCen)
      DEALLOCATE(GuessCenMult)
      DEALLOCATE(work)
      DO cnt1=1,size(DerSmOp)
        DEALLOCATE(DerSmOp(cnt1)%R)
	DEALLOCATE(DerSmOp(cnt1)%t)
      ENDDO
      DEALLOCATE(DerSmOp)
      DO cnt1=1,size(TmpSmOp)
        DEALLOCATE(TmpSmOp(cnt1)%R,TmpSmOp(cnt1)%t)
      ENDDO
      DEALLOCATE(TmpSmOp)
      DEALLOCATE(DummySmOp%R,DummySmOp%t)
      DEALLOCATE(TmpSymProps%dir,TmpSymProps%IntrVec)
      DO i=1,NDerSym
        DEALLOCATE(DerSymProps(i)%dir,DerSymProps(i)%IntrVec)
      ENDDO
      DEALLOCATE(DerSymProps)
      deallocate(stepd,mind,maxd)

      CONTAINS
	
        SUBROUTINE AllocateOneMoreDerSymOp
	  type(SymOp_Type), dimension(:), allocatable :: TmpDerSmOp
	  integer :: l1
	  allocate(tmpdersmop(1:NDerSym-1))
	  DO l1=1,NDerSym-1
            allocate(TmpDerSmOp(l1)%R(1:SDim,1:SDim))
            allocate(TmpDerSmOp(l1)%t(1:SDim))
	  ENDDO
	  DO l1=1,NDerSym-1
            CALL CopySymOp(DerSmOp(l1), TmpDerSmOp(l1))
	    DEALLOCATE(DerSmOp(l1)%R)
	    DEALLOCATE(DerSmOp(l1)%t)
	  ENDDO
	  deallocate(DerSmOp)
	  allocate(DerSmOp(1:NDerSym))
	  DO l1=1,NDerSym
            allocate(DerSmOp(l1)%R(1:SDim,1:SDim))
            allocate(DerSmOp(l1)%t(1:SDim))
 	  ENDDO
	  DO l1=1,NDerSym-1
            CALL CopySymOp(TmpDerSmOp(l1),DerSmOp(l1))
	    DEALLOCATE(TmpDerSmOp(l1)%R)
	    DEALLOCATE(TmpDerSmOp(l1)%t)
	  ENDDO
	  DEALLOCATE(TmpDerSmOp)
!	  write(*,*) 'Incerased NDersym from',NDerSym-1,' to',NDerSym
	END SUBROUTINE AllocateOneMoreDerSymOp
	
      END SUBROUTINE DeriveSymmetry




