      program superflip
      USE SF_Module
      USE SF_Interfaces
      USE SF_RhoIOInterfaces
      USE FFT_Interfaces
      IMPLICIT NONE
      logical :: RhoShouldBeWritten, RepeatCF !Two functions governing the cycling of the calculation      
      integer :: rholoop,reflloop
      real :: amem, bmem
      call TestSystem
      call Input
      call TimeStamp('Start of the calculation:')
      call InputLog
      IF (.not.KarleExpand) CALL NormalizeLocal
      CALL SetupRhoAndFFT
      IF (GenerateRandom) CALL GenerateRandomStructure
      IF (TotalF>0) call ProcessReflections !TotalF can be 0 only for InPerform='symmetry'
      call ReadAndFillOtherDensities !Fill the reference density, model density or enevelope, generate random structure if required
      IF (InPerform=='clusters') GOTO 1000
      IF ((FIPS/='no').and.(NGroups>0)) CALL MakeFIPS
      IF (HMClasses<0) CALL GenerateHistogram !Histogram will be calculated from a random structure or from external structure
      call GetWeakRefl !Calculate the thresholds for weak reflections      
      repeats=0      
      DO
        repeats=repeats+1
        IF ((InPerform/='fourier').and.(InPerform/='symmetry')) THEN
          call ChargeFlipping
          call HandleSymmetry
!          CALL CLEAN
! Write some information in the output file
          CALL WriteRefl
        ELSEIF (InPerform=='fourier') THEN
          CALL Fourier(-1) !Transform from structure factors to density
!          CALL CLEAN
        ELSEIF (InPerform=='symmetry') THEN
          IF (any(NVoxModel/=NVox)) THEN
            CALL ResampleDensity(NVoxModel,NVox,modelrho)
            NVoxModel=NVox
          ENDIF
          rho=modelrho
          CALL Fourier(1) !Transform from density to structure factors
	      CALL HandleSymmetry
	    ENDIF
        IF (RhoShouldBeWritten()) THEN        
          DO RhoLoop=1,OutRhoNr
            call WriteRho(rho,CurRhoFile(RhoLoop),OutRhoType(RhoLoop))
          ENDDO
        ENDIF
        IF (mem) CALL RunMEM
        IF (.not.RepeatCF()) EXIT
      ENDDO

      IF ((WantSumRho/=0).and.(NSummed>0)) THEN
        rho=sumrho/NSummed
        !Recover the experimental amplitudes
!        CALL FFT(rho2sf)
!        CALL SF2MEMRefl
!        DO reflloop=1,MTotalF
!          select case(MEMRefl(reflloop)%code)
!  CASE (code_missing,code_skip)
!            amem=0.; bmem=0.
!	  CASE DEFAULT
!    IF(MEMRefl(reflloop)%FMEM>eps) THEN
!             amem=MEMRefl(reflloop)%AMEM*MEMRefl(reflloop)%F/MEMRefl(reflloop)%FMEM
!             bmem=MEMRefl(reflloop)%BMEM*MEMRefl(reflloop)%F/MEMRefl(reflloop)%FMEM
!	    ELSE
!              amem=0.0; bmem=0.0
!	    ENDIF
!          END SELECT
!          MEMRefl(reflloop)%AMEM=amem
!          MEMRefl(reflloop)%BMEM=bmem
!          MEMRefl(reflloop)%FMEM=sqrt(amem**2+bmem**2)
!        ENDDO
!        CALL MEMRefl2SF
!        CALL FFT(sf2rho)
        DO RhoLoop=1,OutRhoNr
          call WriteRho(rho,OutRhoFile(RhoLoop),OutRhoType(RhoLoop))
        ENDDO
      ENDIF

      IF (len_trim(ExpRefFile)/=0) THEN
        CALL FFT(rho2sf,rho)
        CALL SF2MEMRefl
        CALL WriteFullRefl(ExpRefFile)
      ENDIF
1000  IF (ViewProgress/=noviewer) CALL RemoveForViewer
      IF (ClusAn) CALL DensityCorrelationTable
      call TimeStamp('End of the calculation:')
      call VersionStamp(.false.)
      call FinalCleanUp
      if (KeepWindow) THEN
        write(*,'(a)') 'Press Enter to exit'
	    read(*,*)
      ENDIF
      end program superflip

      SUBROUTINE Clean !Algorithm to remove Fourier artifacts
        USE SF_Module
        USE SF_Interfaces
        USE SF_RhoIOInterfaces
        USE FFT_Interfaces

        IMPLICIT NONE
        REAL, PARAMETER :: minfrac=0.2, peakfrac=0.2
        INTEGER :: i,ii,sfloc, sflocp1,maxpos, peakloop, NCorrPeaks
        REAL :: maxpeak, maxshapefunc, maxorigpeak,hlength,sigmasq, dummy, multsc
        INTEGER, DIMENSION(1:SDim) :: maxcoo, dcoo
        real, dimension(1:SDim) :: rdummy
        REAL, ALLOCATABLE, DIMENSION(:) :: onepeakrho, restrho,peakrho
        !Fill the SF array with ones
        allocate(onepeakrho(0:NVoxTot-1))
        allocate(restrho(0:NVoxTot-1))
        allocate(peakrho(0:NVoxTot-1))
        IF (.not.allocated(modelrho)) THEN
          allocate(modelrho(0:NVoxTot-1))
          modelrho=rho
        ENDIF
!       prepare values
        peakrho=0.0
        rho=modelrho
        restrho=modelrho
        maxorigpeak=maxval(rho)

!       Here we can clculate the shape function
!        SF=0.0E0
!        DO i=MStartF,MEndF
!          IF (MEMRefl(i)%code/=code_skip) SF(MEMRefl(i)%SFloc)=1
!        ENDDO
!        CALL FFT(sf2rho)
!        call writexplor(rho,'shapef.xplor')

        NCorrPeaks=0
        DO
          Ncorrpeaks=NCorrpeaks+1
          maxpeak=-10000.
          DO i=0,NVoxTot-1
            if (restrho(i)>maxpeak) THEN
              maxpeak=restrho(i)
              maxpos=i
            ENDIF
          ENDDO
          IF (maxpeak<minfrac*maxorigpeak) EXIT
          CALL Coordinates(maxpos,maxcoo)
          write(*,*) 'Peak, coordinates:', NCorrPeaks, maxpeak/capvx, maxcoo
!         find the domain of the peak:
          hlength=0.0
          DO i=1,SDim
            dcoo=maxcoo
            DO ii=1,NVox(i)/2
              dcoo(i)=ii+maxcoo(i)
              IF (dcoo(i)>NVox(i)-1) dcoo(i)=DCoo(i)-NVox(i)
              CALL ArrayLocation(dcoo,sfloc)
              IF (restrho(sfloc)<peakfrac*maxpeak) THEN
                hlength=max(hlength,Cell(i)*ii/NVox(i))
                EXIT
              ENDIF
            ENDDO
          ENDDO
          onepeakrho=0.0
          DO i=0,NVoxTot-1
            if (restrho(i)>peakfrac*maxpeak) THEN !Check the distance
              CALL Coordinates(i,dcoo)
              DO ii=1,SDim
                IF ((dcoo(ii)-maxcoo(ii))>NVox(ii)/2) THEN
                  dcoo(ii)=dcoo(ii)-NVox(ii)
                ELSEIF ((dcoo(ii)-maxcoo(ii))<-NVox(ii)/2) THEN
                  dcoo(ii)=dcoo(ii)+NVox(ii)
                ENDIF
              ENDDO
              rdummy=float(dcoo-maxcoo)/float(NVox)
              dummy=sqrt(multsc(rdummy,rdummy,SDim,.false.))
              IF (dummy<hlength) THEN
                onepeakrho(i)=restrho(i)
!TEMPORARY for treating Eu2Si2O7
!                dcoo(1)=dcoo(1)+NVox(1)/2
!                CALL Arraylocation(dcoo,ii)
!                onepeakrho(ii)=restrho(ii)
              ENDIF
          ENDIF
        ENDDO
!        call writexplor(onepeakrho,'onepeakrho.xplor')
        peakrho=peakrho+onepeakrho
!       Make a convolution of the sharpened peak with the shape function
        rho=onepeakrho
        CALL FFT(rho2sf)
        CALL SF2MEMRefl
        SF=0.0
        CALL MEMRefl2SF
        CALL FFT(sf2rho)
!       Rescale to fit the maximum
        dummy=maxval(rho)
        rho=rho*maxpeak/dummy
!       Subtract the result from the original density outside of the peak domain
        DO i=0,NVoxTot-1
          if (onepeakrho(i)<0.01) THEN
            restrho(i)=restrho(i)-rho(i)
          ELSE
            restrho(i)=0.0E0
          ENDIF
        ENDDO
      ENDDO
      rho=restrho+peakrho
      call writexplor(rho,'cleaned.xplor')
      DEALLOCATE(peakrho, restrho, onepeakrho)
      write(*,*) 'Hotovo'
      STOP
      END SUBROUTINE Clean


      SUBROUTINE RunMEM !Algorithm to perform a simplified MEM calculation on the density
        USE SF_Module
        USE SF_Interfaces
        USE SF_RhoIOInterfaces
        USE FFT_Interfaces

        IMPLICIT NONE

        real, parameter :: poscharge=0.05, lambdadecrement=0.7
        real :: lmbd,rnom,rden, rval,fcon,fconold, entropy, dstar, multsc, priorrho
        integer :: i,ii, loc, locp1
        real, allocatable, dimension(:) :: memrhop, memrhom

        !Make sure the MEMRefl array contains correct structure factors
!        rho=modelrho
        call FFT(rho2SF)
        call SF2MEMRefl

        IF (allocated(memrhop)) deallocate(memrhop)
        allocate(memrhop(0:NVoxTot-1))
        IF (allocated(memrhom)) deallocate(memrhom)
        allocate(memrhom(0:NVoxTot-1))
        !Define prior
        IF (.not.twochannel) THEN
          memrhop=capvx*MEMRefl(MStartF)%FMEM/Cell(NCellPar+1)
          memrhom=0.0
        ELSE
          memrhop=capvx*MEMRefl(MStartF)%FMEM/Cell(NCellPar+1)+poscharge*capvx
          memrhom=poscharge*capvx
        ENDIF
        priorrho=capvx*MEMRefl(MStartF)%FMEM/Cell(NCellPar+1)
        rho=memrhop-memrhom
        CALL FFT(rho2SF)
        !Guess the suitable lambda
        CALL InitLambda
        !Main MEM cycle
        rho=memrhop-memrhom
        CALL FFT(rho2SF)
        fconold=10000000.
        DO ii=1,memcycles
          CALL FillSFWithFconDer
          CALL FFT(sf2rho)
          DO i=0,NVoxTot-1
            memrhop(i)=memrhop(i)*exp(-lmbd*rho(i))
            IF (twochannel) memrhom(i)=memrhom(i)*exp(lmbd*rho(i))
          ENDDO
!          do i=0,10
!            write(*,*) memrhop(i),-lmbd*rho(i),exp(-lmbd*rho(i))
!          enddo
          !Fill new structure factors
          rho=memrhop-memrhom
          CALL FFT(rho2sf)
          !Get R-value
          rnom=0; rden=0;fcon=0.0
          DO i=MStartF+1, MEndF
            dstar=sqrt(multsc(float(MEMRefl(i)%hklFG),float(MEMRefl(i)%hklFG),SDim,.true.))
            IF (dstar>2*MEMlimit) CYCLE
            Loc=MEMRefl(i)%SFLoc
            Locp1=Loc+1
            rnom=rnom+abs(sqrt(SF(Loc)**2+SF(Locp1)**2)-MEMRefl(i)%FMEM)
            rden=rden+MEMRefl(i)%FMEM
            fcon=fcon+(SF(Loc)-MEMRefl(i)%AMEM)**2+(SF(Locp1)-MEMRefl(i)%BMEM)**2
          END DO
          entropy=-sum(memrhop*log(memrhop/priorrho))
          rval=rnom/rden
          fcon=fcon/MTotalF
          write(*,*) 'cycle:',ii, 'R-value:',100*rnom/rden,'Fcon:',fcon, 'Lambda:',lmbd, 'Entropy:',entropy
          if (fcon>fconold) lmbd=lmbd*lambdadecrement
          fconold=fcon
          if (rval<0.01) EXIT
        ENDDO
        write(lgu,'(a)') 'Entropy maximization was performed. Final numbers are:'
        write(lgu,'(a,f10.3,a,f10.3,a,f15.4)')'R-value:',100*rnom/rden,'  Fcon:',fcon, '  Entropy:',entropy
        memrhop=memrhop-memrhom
        CALL writexplor(memrhop,'memmed.xplor')
        DEALLOCATE(memrhop,memrhom)

      CONTAINS

      SUBROUTINE InitLambda
        IMPLICIT NONE

        REAL, PARAMETER :: maxexp=50
        REAL :: minrho(1:1)
        ! find the largest lambda, which does not cause over/underflow
        CALL FillSFWithFconDer
        CALL FFT(sf2rho)
        minrho(1)=minval(rho)
        lmbd=log(maxexp)/abs(minrho(1))
      END SUBROUTINE InitLambda

      END SUBROUTINE RunMEM


      SUBROUTINE FillSFWithFconDer
        USE SF_Module
        USE SF_Interfaces

        IMPLICIT NONE

        integer :: loc, locp1, i
        real :: dstar, multsc
        real, allocatable, dimension(:) :: dummyreal, dummyimag

        allocate(dummyreal(MStartF:MendF))
        allocate(dummyimag(MStartF:MendF))
        DO i=MStartF,MEndF
          Loc=MEMRefl(i)%SFLoc
          Locp1=Loc+1
          dummyreal(i)=SF(Loc)
          dummyimag(i)=SF(Locp1)
        ENDDO
        SF=0.0
        DO i=MStartF, MEndF
          dstar=sqrt(multsc(float(MEMRefl(i)%hklFG),float(MEMRefl(i)%hklFG),SDim,.true.))
          IF (dstar>2*MEMlimit) CYCLE
          Loc=MEMRefl(i)%SFLoc
          Locp1=Loc+1
          SF(Loc)=2*(DummyReal(i)-MEMRefl(i)%AMEM)
          SF(Locp1)=2*(DummyImag(i)-MEMRefl(i)%BMEM)
        END DO
        deallocate(dummyreal)
        deallocate(dummyimag)

      END SUBROUTINE FillSFWithFconDer
