! Expand the symmetry, check the consistency with NVox, combine with centering vectors and create the integer counterparts
!*********************************
      SUBROUTINE CompleteSymmetry(ErrCd)
!*********************************
      USE SF_Module
      USE Input_Module
      IMPLICIT NONE
      integer :: arrel,i,ErrCd
      logical :: zero,nonrot,nontrans
      real :: dummy,dummy2
      real, dimension(SDim) :: darray, darray2
      real, allocatable, dimension(:,:) :: dummymat,dummymat2
      integer, dimension(SDim,SDim) :: idummymat, idummymat2
      
      ! First shift the zero centering to the first place if necessary
      DO cnt1=1,NCenter
	zero=.true.
	DO cnt2=1,SDim
	  if (abs(center(cnt2,cnt1))>eps) zero=.false.
	ENDDO
	if (zero) THEN
	  IF (cnt1/=1) THEN
	    DO cnt2=1,SDim
	     dummy=Center(cnt2,cnt1)
	     Center(cnt2,cnt1)=Center(cnt2,1)
	     Center(cnt2,1)=dummy
            ENDDO
	  ENDIF
	  exit
	ENDIF
      ENDDO

      ! Map the centering vectors onto the unit cell
      DO cnt1=1,NCenter
        DO i=1,SDim
	  DO
	    IF (Center(i,cnt1)<0.E0) THEN
	      Center(i,cnt1)=Center(i,cnt1)+1.0E0
	    ELSEIF (Center(i,cnt1)>=1.E0) THEN
	      Center(i,cnt1)=Center(i,cnt1)-1.0E0
	    ELSE
	      EXIT
	    ENDIF
	  ENDDO
        ENDDO
      ENDDO

 !    Expand the symmetry operations over the centering vectors     
      NSymPrim=NSym     
      NSym=NSym*NCenter
      IF (NCenter>1) THEN
!       We need just NCenter identical copies of symops
        DO cnt1=NSymPrim+1,NSym
	  arrel=mod(cnt1-1,NSymPrim)+1
	  SmOp(cnt1)%R=SmOp(arrel)%R
        END DO
!       The translation vectors must be expanded and checked to lie between 0 and 1
        DO cnt1=1,NSymPrim
          DO cnt2=1,NCenter
            arrel=NSymPrim*(cnt2-1)+cnt1
	    DO i=1,SDim
              SmOp(arrel)%t(i)=SmOp(cnt1)%t(i)+Center(i,cnt2)
	      DO
	        IF (SmOp(arrel)%t(i)<0.E0) THEN
		  SmOp(arrel)%t(i)=SmOp(arrel)%t(i)+1.0E0
		ELSEIF (SmOp(arrel)%t(i)>=1.E0) THEN
		  SmOp(arrel)%t(i)=SmOp(arrel)%t(i)-1.0E0
		ELSE
		  EXIT
		ENDIF
	      ENDDO
	    ENDDO
	  ENDDO
	END DO
      ENDIF

! Check for the presence of the identity operation and switch it to the first place, if neccesary.
      DO cnt1=1,NSym
        dummy=0.0; dummy2=0.0
	DO cnt2=1,SDim
	  dummy=dummy+SmOp(cnt1)%R(cnt2,cnt2)
	  dummy2=dummy2+abs(SmOp(cnt1)%t(cnt2))
	ENDDO
	if ((abs(dummy-float(SDim))<eps).and.(dummy2<eps)) THEN !The identity operation has been located
	  IF (cnt1/=1) THEN
	    idummymat=SmOp(1)%R
            darray=SmOp(1)%t
            SmOp(1)=SmOp(cnt1)
            SmOp(cnt1)%R=idummymat
	    SmOp(cnt1)%t=darray
	  ENDIF
	  exit
	ENDIF
      ENDDO

! Check for the uniqueness of the centering vectors
      DO cnt1=1,NCenter
        zero=.false. !If zero is .true., a repetition of one centering vector was found
	DO cnt2=cnt1+1,NCenter
          IF (EqualOp(SmOp(1)%R,Center(1,cnt1),SmOp(1)%R,Center(1,cnt2),SDim)) THEN
            write(Buffer,'(a,i3,a,i3,a)') 'Error, the centering vectors',cnt1,' and',cnt2, &
                                    ' are equivalent.'
            CALL ErrorMsg(buffer)
            ErrCd=100
	    EXIT
	  ENDIF
        ENDDO
	IF (ErrCd/=0) EXIT
      ENDDO
      IF (ErrCd/=0) RETURN

! Check for the uniqueness of the symmetry operations
      DO cnt1=1,NSymPrim
        zero=.false. !If zero is .true., a repetition of one symmetry operation was found
	DO cnt2=cnt1+1,NSym
          IF (EqualOp(SmOp(cnt1)%R,SmOp(cnt1)%t,SmOp(cnt2)%R,SmOp(cnt2)%t,SDim)) THEN
            IF (cnt2>NSymPrim) THEN !Equivalece by a centering vector
              write(Buffer,'(a,i3,a,i3,a)') 'Error, the symmetry operations',cnt1,' and',mod(cnt2-1,NSymPrim)+1, &
	                                    ' are translationally equivalent.'
              CALL ErrorMsg(buffer)
              write(Buffer,'(a,i3,a,i3,a,i3,a)') 'SymOp',cnt1,' = SymOp', mod(cnt2-1,NSymPrim)+1,' + Center',1+cnt2/NSymPrim,'.'
            ELSE !Completely identical
              write(Buffer,'(a,i3,a,i3,a)') 'Error, the symmetry operations',cnt1,' and',cnt2,' are identical!'
            ENDIF
	    CALL ErrorMsg(buffer)
            ErrCd=100
	    EXIT
	  ENDIF
        ENDDO
	IF (ErrCd/=0) EXIT
      ENDDO
      IF (ErrCd/=0) RETURN

! Check the internal consistency of each operation - it must be a rotation matrix!
      allocate(dummymat(1:SDim,1:SDim))
      allocate(dummymat2(1:SDim,1:SDim))

      DO cnt1=1,NSymPrim !Only the first NSymPrim operations must be checked, the rest is identical
        dummymat=invert(SmOp(cnt1)%R)
        IF (dummymat(1,1)>1.0E29) THEN
          write(Buffer,'(a,i3,a)') 'Error, the rotation part of the symmetry operation number',cnt1, &
                                   ' is a singular matrix!'
	  CALL ErrorMsg(buffer)
	  ErrCd=100
          CYCLE
        ENDIF
!	write(*,*) 'Gsi'
!	call printmat(Gsi,'(f10.5)')
!	write(*,*) 'Gs'
!	call printmat(Gs,'(f10.5)')
	CALL multm(GS,dummymat,dummymat2,SDim,SDim,SDim)
!	write(*,*) 'R'
!	call printmat(SmOp(cnt1)%R,'(i5)')
!	write(*,*) 'R-1'
!	call printmat(dummymat,'(f10.5)')
!    write(*,*) 'GS*R-1'
!	call printmat(dummymat2,'(f10.5)')
	CALL multm(dummymat2,GSi,dummymat,SDim,SDim,SDim)
!   write(*,*) 'Gs*R-1*Gs'
!   call printmat(dummymat,'(f10.5)')

! Dummymat now contains G*R^-1*Gi, we are ready to check against R^T
        dummymat2=transpose(SmOp(cnt1)%R)
!    write(*,*) 'RT'
!	call printmat(dummymat2,'(f10.5)')
        IF (maxval(abs(dummymat2-dummymat))>eps) THEN
          write(Buffer,'(a,i3,a)') 'Error, the rotation part of the symmetry operation number',cnt1, &
                          ' is not properly defined or is incompatible with the cell parameters.(Error indicator: R^T*G/=G*R^-1)'
	  CALL ErrorMsg(buffer)
	  ErrCd=100
        ENDIF
      ENDDO
      IF (ErrCd/=0) THEN
        deallocate(dummymat)
        deallocate(dummymat2)
        RETURN
      ENDIF
! If we are here, we have passed the check of the individual operations vs. cell. What has to be checked now is the consistency with the voxel spacing.
      darray=1./NVox
      DO cnt1=1,NSymPrim !Centering vectors will be checked separately
         darray2=multmat(SmOp(cnt1)%R,darray)
         IF (any(abs(darray2*NVox-nint(darray2*NVox))>bigeps)) THEN
           write(Buffer,'(a,i3,a)') 'Error, the rotational part of the symmetry operation number',cnt1, &
	                            ' is not consistent with the voxel division.'
           CALL ErrorMsg(buffer)
           ErrCd=100
         ENDIF
         darray2=SmOp(cnt1)%t*NVox
         IF (any(abs(darray2-nint(darray2))>bigeps)) THEN
           write(Buffer,'(a,i3,a)') 'Error, the translational part of the symmetry operation number',cnt1, &
	                           ' is not consistent with the voxel division.'
           CALL ErrorMsg(buffer)
           ErrCd=100
         ENDIF
      ENDDO
      
      !Do the consistency check on the voxels also for the centering vectors. Separate check is done to avoid a large listing of errors on symmetry operations, if only the centering is wrong
      DO cnt1=1,NCenter
        nontrans=.false.
        DO cnt2=1,SDim
          dummy=Center(cnt2,cnt1)*float(NVox(cnt2))
          IF (abs(dummy-anint(dummy))>bigeps) nontrans=.true.
          IF (nontrans) EXIT
	ENDDO
        IF (nontrans) THEN
          write(Buffer,'(a,i3,a)') 'Error, the centering vector number',cnt1,' is not consistent with the voxel division.'
	  CALL ErrorMsg(buffer)
	  ErrCd=100
        ENDIF	
      ENDDO
      IF (ErrCd/=0) THEN
        deallocate(dummymat)
        deallocate(dummymat2)
        RETURN
      ENDIF
!     What comes next is the check of the completeness of the space group.Each combination of the operations must give another valid operation
! First do the check with the centering vectors
      DO cnt1=1,NCenter
        DO cnt2=1,NCenter	 
	  CALL ProdOp(SmOp(1)%R(1,1),Center(1,cnt1),SmOp(1)%R(1,1),Center(1,cnt2),idummymat,darray,SDim)
          zero=.true. !If zero is true, no matching centering vector has been found
	  DO cnt3=1,NCenter
            IF (EqualOp(idummymat,darray,idummymat,Center(1,cnt3),SDim)) THEN
	      zero=.false.
	      EXIT
	    ENDIF
	  ENDDO
          IF (zero) THEN
            write(Buffer,'(a,i3,a,i3,a)') 'Error, the list of centering vectors is incomplete! Sum of vectors',cnt1, &
	                                  ' and',cnt2,' was not found in the list.'
	    CALL ErrorMsg(buffer)
	    ErrCd=100
          ENDIF
	ENDDO
      ENDDO
      IF (ErrCd/=0) THEN
        deallocate(dummymat)
        deallocate(dummymat2)
        RETURN
      ENDIF
      DO cnt1=1,NSym
        DO cnt2=1,NSym
          CALL ProdOp(SmOp(cnt1)%R,SmOp(cnt1)%t,SmOp(cnt2)%R,SmOp(cnt2)%t,idummymat,darray,SDim)
          zero=.true. !If zero is true, no matching operation has been found
	  DO cnt3=1,NSym
            IF (EqualOp(idummymat,darray,SmOp(cnt3)%R,SmOp(cnt3)%t,SDim)) THEN
	      zero=.false.
	      EXIT           
	    ENDIF
	  ENDDO
          IF (zero) THEN
            write(Buffer,'(a,i3,a,i3,a)') 'Error, the list of operations is incomplete! Product of operations',cnt1, &
	                                  ' and',cnt2,' was not found in the list.'
	    CALL ErrorMsg(buffer)
	    ErrCd=100
          ENDIF
	ENDDO
      ENDDO
      IF (ErrCd/=0) RETURN
      !Find the generators
      CALL FindGenerators
      CALL FillSGProps

! If twin matricees are present, find the symops that survive the twinning
      ALLOCATE(SurviveTwin(1:NSymPrim),STAT=cnt1)
      IF (cnt1/=0) CALL StopProgram('Error allocating array SurviveTwin! Too little memory?')
      SurviveTwin=.true.
      IF (NTwin>0) THEN
        SurviveTwin=.true.
	darray=0.0
	DO cnt1=1,NSymPrim
          DO cnt2=1,NTwin
	    !Only if ST=TS, S survives
            idummymat=multmat(SmOp(cnt1)%R,TwinOp(cnt2)%R)
            idummymat2=multmat(TwinOp(cnt2)%R,SmOp(cnt1)%R)
	    IF (.not.EqualOp(idummymat,darray,idummymat2,darray,SDim)) THEN
	      SurviveTwin(cnt1)=.false.
	      EXIT
	    ENDIF
	  ENDDO	        
        ENDDO
      ENDIF
      
      deallocate(dummymat)
      deallocate(dummymat2)

      RETURN
      
      END SUBROUTINE CompleteSymmetry

!*****************************************************************************
      SUBROUTINE FillSGProps
!*****************************************************************************
      USE SF_Module
      IMPLICIT NONE
      LOGICAL :: IsEnant
      ! Check for the presence of the inversion center.
      Centro=.false.
      SGProps%Centro=.false.
      DO cnt1=1,NSym
        IF (all(SmOp(cnt1)%R==-UnitMat(SDim))) THEN
	  Centro=.true.
	  SGProps%centro=.true.
	  EXIT
	ENDIF
      ENDDO

      !Check, if the group has an enantiomorphic pair
      SGProps%enant=.false.
      IsEnant=.true.
      DO cnt1=1,NSymPrim
        IF (Determinant(SmOp(cnt1)%R)<0.) THEN
	  isenant=.false.
	  EXIT
        ENDIF
      ENDDO
      IF (IsEnant) THEN
	DO cnt1=2,NSymPrim
	  IsEnant=.true.
	  DO cnt2=cnt1,NSym
	    IF (SymOpsJustShifted(SmOp(cnt1)%R,-SmOp(cnt1)%t,SmOp(cnt2)%R,SmOp(cnt2)%t,Center)) THEN
	      IsEnant=.false.
	      EXIT
	    ENDIF
	  ENDDO
	  IF (IsEnant) THEN
	    SGProps%enant=.true.
	    EXIT
	  ENDIF
        ENDDO
      ENDIF

      END SUBROUTINE FillSGProps
      
!*****************************************************************************
      SUBROUTINE FindGenerators
!*****************************************************************************
      USE SF_Module

      IMPLICIT NONE
      INTEGER :: i,ii,cursymop
      type symopsorttype
        integer :: symop,rank,order,score
      end type symopsorttype
      type(symopsorttype), allocatable, dimension(:) :: symopsort
      integer, allocatable, dimension(:) :: genlist
      real, allocatable, dimension(:) :: darray
      real, allocatable, dimension(:,:) :: dummymat,dummymat2
      integer, dimension(SDim,SDim) :: idummymat
     

      allocate(dummymat(1:SDim,1:SDim))
      allocate(dummymat2(1:SDim,1:SDim))
      allocate(darray(1:SDim))
      allocate(symopsort(1:NSymPrim))
      allocate(genlist(1:NSym))
      
! Pick up the generators:
      !calculate the dimension of the invariant subspace of each symop
      DO i=1,NSymPrim
        DO cnt1=1,SDim
	  DO cnt2=1,SDim
	    dummymat(cnt1,cnt2)=SmOp(i)%R(cnt1,cnt2)
	    if (cnt1==cnt2) dummymat(cnt1,cnt2)=dummymat(cnt1,cnt2)-1.
	  ENDDO
	ENDDO
        CALL GetRank(dummymat,SDim,SDim,cnt1)
	symopsort(i)%rank=cnt1
	symopsort(i)%symop=i
      ENDDO
      ! Bubblesort symopsort by the rank
      DO i=NSymPrim,2,-1
        DO ii=1,i-1
	  IF (symopsort(ii+1)%rank<symopsort(ii)%rank) THEN
	    cnt1=symopsort(ii+1)%rank
	    cnt2=symopsort(ii+1)%symop
	    symopsort(ii+1)=symopsort(ii)
	    symopsort(ii)%rank=cnt1
	    symopsort(ii)%symop=cnt2
	  ENDIF
	ENDDO
      ENDDO
      genlist=0
      DO cnt1=2,NSymPrim
        cursymop=symopsort(cnt1)%symop
        if (genlist(cursymop)==0) then
	  genlist(cursymop)=2
	endif
        DO cnt2=2,NSym
	  if ((genlist(cursymop)>0).and.(genlist(cnt2)>0)) THEN
	    CALL ProdOp(SmOp(cursymop)%R,SmOp(cursymop)%t,SmOp(cnt2)%R,SmOp(cnt2)%t,idummymat,darray,SDim)
	    DO cnt3=1,NSym
              IF (EqualOp(idummymat,darray,SmOp(cnt3)%R,SmOp(cnt3)%t,SDim)) THEN
                if (genlist(cnt3)==0) genlist(cnt3)=1
	        EXIT
	      ENDIF
	    ENDDO
	    CALL ProdOp(SmOp(cnt2)%R,SmOp(cnt2)%t,SmOp(cursymop)%R,SmOp(cursymop)%t,idummymat,darray,SDim)
	    DO cnt3=1,NSym
              IF (EqualOp(idummymat,darray,SmOp(cnt3)%R,SmOp(cnt3)%t,SDim)) THEN
                if (genlist(cnt3)==0) genlist(cnt3)=1
	        EXIT
	      ENDIF
	    ENDDO
	  ENDIF
	ENDDO
      ENDDO
      NGen=0
      DO cnt1=1,NSymPrim
        if (genlist(symopsort(cnt1)%symop)==2) then
          NGen=NGen+1
	endif
      ENDDO
      IF (ALLOCATED(generators)) DEALLOCATE(generators)
      ALLOCATE(generators(1:NGen))
      cnt2=0
      DO cnt1=1,NSymPrim
        if (genlist(symopsort(cnt1)%symop)==2) then
          cnt2=cnt2+1
	  generators(cnt2)=symopsort(cnt1)%symop
	endif
      ENDDO
!Repeat the same procedure with the centering vectors:
      genlist=0
      IF (NCenter==1) THEN !Only the trivial centering present
        genlist(1)=2
      ELSE
        DO cnt1=2,NCenter
          if (genlist(cnt1)==0) then
	    genlist(cnt1)=2
	  endif
          DO cnt2=2,NCenter
	    if ((genlist(cnt1)>0).and.(genlist(cnt2)>0)) THEN
              CALL ProdOp(SmOp(1)%R,Center(1,cnt1),SmOp(1)%R,Center(1,cnt2),idummymat,darray,SDim)
	      DO cnt3=1,NSym
                IF (EqualOp(idummymat,darray,idummymat,Center(1,cnt3),SDim)) THEN
                  if (genlist(cnt3)==0) genlist(cnt3)=1
	          EXIT
	        ENDIF
	      ENDDO
	    ENDIF
	  ENDDO
        ENDDO
      ENDIF
      NGCenter=0
      DO cnt1=1,NCenter
        if (genlist(cnt1)==2) then
          NGCenter=NGCenter+1
	endif
      ENDDO
      IF (ALLOCATED(gencenters)) DEALLOCATE(gencenters)
      ALLOCATE(gencenters(1:NGCenter))
      cnt2=0
      DO cnt1=1,NCenter
        if (genlist(cnt1)==2) then
          cnt2=cnt2+1
	  gencenters(cnt2)=cnt1
	endif
      ENDDO

      deallocate(symopsort)
      deallocate(dummymat)
      deallocate(dummymat2)
      deallocate(darray)
      deallocate(genlist)

      END SUBROUTINE FindGenerators

! The input reflections are expanded, averaged and checked for doubles. MEMRefl array is filled
!************************************
      SUBROUTINE ProcessReflections
!************************************
      USE SF_Module
      USE SF_Composition
      USE Input_Module
      IMPLICIT NONE 
      integer, parameter :: hklcharnr=6,mxcoverclasses=100
      real, parameter :: cclsw=0.05 !width of coverage  classes
      integer :: arrloc,i,occup,ClassNr,ClCnt1,ClCnt2,covsum1,covsum2,friedel
      integer :: IMapToRange !Shift the number to lie between given boundaries
      integer :: ExtinctNr
      logical :: Extinct !Function, that is true if its argument is a reflection extinct due to the symmetry
      logical :: Skip
      integer, dimension(SDim) :: trhkl,maxhkl
      integer, allocatable, dimension(:,:) :: qstararray,qstar
      integer :: nqstar
      logical :: goodsat
      integer, dimension(1:mxcoverclasses,1:2) :: cover !Array holding the information about the coverage
      integer, dimension(0:NVoxTotSF-1) :: OrigReflNr
      real, dimension(SDim) :: dtrhkl
      real :: phdiff,ph1,Phase,sthlmax
      real :: multsc !Function returning a scalar product of two vectors
      real :: bfactor !Function returning the correction for the D-W factor
      real :: normalize !Function returning the normalized structure factor
      character*1, dimension(SDim) :: hklchar
      character*24 :: ReflForm='(i6,'':'',..i5,.f16.6,i14)'
      character*27 :: ReflGForm='(i6,'':'',..i5,.f16.6,i14,i8)'
      character*16 ::  LargereflForm='(a,f14.6,a,..i5)'
      
      hklchar(1)='h'
      if (SDim>=2) hklchar(2)='k'
      if (SDim>=3) hklchar(3)='l'
      if (SDim>=4) hklchar(4)='m'
      if (SDim>=5) hklchar(5)='n'
      if (SDim>=6) hklchar(6)='o'

      IF (Terminal) write(*,'(a)') 'Processing the reflections...'
      write(LargeReflForm(12:13),'(i2)') SDim
      write(ReflForm(9:10),'(i2)') SDim
      write(ReflGForm(9:10),'(i2)') SDim
      IF (haveAB) THEN
        ReflForm(14:14)='3'
        ReflGForm(14:14)='3'
      ELSE
        ReflForm(14:14)='1'
        ReflGForm(14:14)='1'
      ENDIF
! First some log-info about the input reflections      
      call FramedLabel(1,'Information about reflections:')
      write(lgu,'(a,i9)') 'Number of reflections in the input file: ',TotalF
      IF (ReflectionList/='unique') CALL AverageReflections
      IF (NGroups>0) THEN
        cnt1=0
	DO i=StartF,EndF
	  if (Refl(i)%group==0) cnt1=cnt1+1
	ENDDO
	write(lgu,'(a,i6)') 'Number of separated reflections        : ',cnt1
	write(lgu,'(a,i6)') 'Number of overlapping reflections      : ',TotalF-cnt1
	write(lgu,'(a,i6)') 'Number of overlap groups               : ',NGroups
	write(lgu,*)
	IF (FIPS=='table') THEN
          write(*,'(a,i6)') 'Number of reflections in the input file: ',TotalF
          write(*,'(a,i6)') 'Number of separated reflections        : ',cnt1
	  write(*,'(a,i6)') 'Number of overlapping reflections      : ',TotalF-cnt1
	  write(*,'(a,i6)') 'Number of overlap groups               : ',NGroups
	ENDIF
      ENDIF
!     Assign the location of the reflections in the array
      DO cnt1=1,TotalF
	Call ArrayLocationSF(Refl(cnt1)%hklfg,Refl(cnt1)%SFLoc,friedel)
	Refl(cnt1)%SFLoc=Refl(cnt1)%SFLoc*friedel
      ENDDO
      IF (ExpandedLog) THEN !Let's write a list, if the log should be detailed
        write(lgu,'(''Reflections read from the input file and averaged:'')')
        WRITE(Buffer,'(a)') 'Number:'
        DO cnt1=1,SDim
	  cnt2=7+5*cnt1
          WRITE(Buffer(cnt2:cnt2),'(a1)') hklchar(cnt1)
	ENDDO
        cnt2=len_trim(Buffer)+1
        IF (NGroups==0) THEN
          IF (haveAB) THEN
	    WRITE(Buffer(cnt2:),'(a)') '         F               A               B      ArrayLocation'
          ELSE
	    WRITE(Buffer(cnt2:),'(a)') '         F          ArrayLocation'
	  ENDIF
        ELSE
          IF (haveAB) THEN
	    WRITE(Buffer(cnt2:),'(a)') '         F               A               B      ArrayLocation  Group'
          ELSE
	    WRITE(Buffer(cnt2:),'(a)') '         F          ArrayLocation  Group'
	  ENDIF	
	ENDIF
	write(lgu,'(a)') trim(Buffer)
	DO cnt1=1,TotalF
          IF (NGroups==0) THEN
            IF (haveAB) THEN
	      write(lgu,reflform) cnt1,(Refl(cnt1)%hklFG(cnt2),cnt2=1,SDim),Refl(cnt1)%F,Refl(cnt1)%A,Refl(cnt1)%B,Refl(cnt1)%SFloc
             ELSE            
	      write(lgu,reflform) cnt1,(Refl(cnt1)%hklFG(cnt2),cnt2=1,SDim),Refl(cnt1)%F,Refl(cnt1)%SFloc
	    ENDIF
	  ELSE
            IF (haveAB) THEN
	      write(lgu,reflgform) cnt1,(Refl(cnt1)%hklFG(cnt2),cnt2=1,SDim),Refl(cnt1)%F,Refl(cnt1)%A,Refl(cnt1)%B, &
                                   Refl(cnt1)%SFloc,Refl(cnt1)%group
             ELSE            
	      write(lgu,reflgform) cnt1,(Refl(cnt1)%hklFG(cnt2),cnt2=1,SDim),Refl(cnt1)%F, &
	                           Refl(cnt1)%SFloc,Refl(cnt1)%group
	    ENDIF
          ENDIF
	ENDDO
        write(lgu,'('''')')
      ENDIF
! Check for presence of extinct reflections
      ExtinctNr=0
      DO cnt1=1,TotalF
        IF (extinct(Refl(cnt1)%hklfg,occup)) THEN !occup holds the number of the symop responsible for the extinction
          ExtinctNr=ExtinctNr+1
	  IF(expandedlog) CALL ExtinctRefl(Refl(cnt1)%hklFG,cnt1,occup,ExtinctNr,Refl(cnt1)%F) 
	ENDIF
      ENDDO
      IF (ExtinctNr>0) THEN
        write(lgu,'(a,i5,a)') 'Info:',ExtinctNr,' reflections found that are extinct according to the symmetry.'
        write(lgu,'(a,i5,a)') 'These reflections will be used in the calculation.'
      ENDIF

      maxhkl=0
      SF=0.0
      OrigReflNr=0
      DO cnt1=1,TotalF
!        IF (extinct(Refl(cnt1)%hklfg,occup)) CYCLE !Ignore the extinct reflections            
        DO cnt2=1,NSymPrim !Only the "primitive" operations must be used, centering is not interesting here
!          IF (.not.SurviveTwin(cnt2)) CYCLE !This symop cannot be used for expansion of the reflections, it is corrupted by the twinning
          Skip=.false.
          trhkl=multmat(refl(cnt1)%hklfg,SmOp(cnt2)%R)
	      maxhkl=max(maxhkl,abs(trhkl))
          IF (any(maxhkl>(NVox-1)/2)) Skip=.true.
	      IF (Skip) EXIT !This reflection has too large indices - error message will be issued, but Doubled check must be skipped
	      Call ArrayLocationSF(trhkl,arrloc,friedel)
	      phdiff=sum(Refl(cnt1)%hklFG*SmOp(cnt2)%t)
          ph1=Phase(Refl(cnt1)%A,Refl(cnt1)%B)+2.*pi*phdiff !This contains the phase of the transformed reflection. Attention: according to the crystallographic conventions phdiff should be subtracted, but fftw has opposite convention...
          SF(ArrLoc)=Refl(cnt1)%F*cos(ph1)
          SF(ArrLoc+1)=friedel*Refl(cnt1)%F*sin(ph1)
          OrigReflNr(Arrloc/2)=cnt1
	      IF (.not.centro) THEN
            Call ArrayLocationSF(-trhkl,arrloc,friedel)
            SF(ArrLoc)=Refl(cnt1)%F*cos(ph1)
            SF(ArrLoc+1)=-friedel*Refl(cnt1)%F*sin(ph1)
	        OrigReflNr(Arrloc/2)=cnt1
	      ENDIF
	    ENDDO
      ENDDO
      write(lgu,'(a)',ADVANCE='NO') 'Maximum indices in expanded reflection set:'
      DO i=1,SDim
        write(lgu,'(i4)',ADVANCE='NO') maxhkl(i)
      ENDDO
      write(lgu,'('''')')
      cnt1=0
      DO i=1,SDim
        If (maxhkl(i)>(NVox(i)-1)/2) THEN
          write(Buffer,'(a,i3)') 'Error, the maximum of the reflection index '//hklchar(i)//' is',maxhkl(i)
          Call ErrorMsg(Buffer)
          Call ErrorMsg('This reaches or exceeds half of the voxel division along that axis.')
          Call ErrorMsg('Please, increase the voxel division appropriately.')
          cnt1=100
        ENDIF
      ENDDO
      IF (cnt1/=0) CALL StopProgram('')
      write(lgu,'('''')')

! At this point the array SF contains all expanded reflections. We have to count them and write the necessary info into the MEMRefl array
      !For superspace determine the star of satellites. This is trivial for most cases, but complicated for 5D hexagonal
      IF (SDim>RDim) THEN
        allocate(qstar(1:SDim-RDim,1:200)) !200 is just a number, I hope never ever will this number be exceeded
        qstar(1:SDim-RDim,1)=0
        nqstar=1
        DO cnt1=1,TotalF
          goodsat=.true.
          DO cnt2=1,Nqstar
            IF (all(Refl(cnt1)%hklfg(RDim+1:SDim)==qstar(1:SDim-RDim,cnt2))) THEN
              goodsat=.false.
              EXIT
            ENDIF
          ENDDO
          IF (goodsat) THEN
            nqstar=nqstar+1
            qstar(1:SDim-RDim,nqstar)=refl(cnt1)%hklfg(RDim+1:SDim)
          ENDIF
        ENDDO
      ENDIF
      MTotalF=0
      MissingNr=0
      DO cnt1=1,NVoxTotSF-1 !skip the 000 reflection, we want it in any case
        IF (OrigReflNr(cnt1)/=0) THEN
	  MTotalF=MTotalF+1
        ELSEIF (FillMissingSthl>eps) THEN
          CALL CoordinatesSF(2*cnt1,trhkl)
          DO cnt3=1,SDim
	    trhkl(cnt3)=IMapToRange(trhkl(cnt3),-NVox(cnt3)/2,(NVox(cnt3)+1)/2)
	  ENDDO
	  !Exclude satellites with too high indices
	  IF (SDim>RDim) THEN
!            IF (any(abs(trhkl(RDim+1:SDim))>maxhkl(RDim+1:SDim))) CYCLE
	    goodsat=.false.
	    DO cnt3=1,Nqstar
	      IF (all(trhkl(RDim+1:SDim)==qstar(1:SDim-RDim,cnt3))) THEN
	        goodsat=.true.
	        EXIT
	      ENDIF
	    ENDDO
	    IF (.not.goodsat) CYCLE
	  ENDIF
          phdiff=multsc(float(trhkl),float(trhkl),SDim,.true.) !Gives the length of the diffraction vector
          if (phdiff>eps) phdiff=0.5*sqrt(phdiff)
	  IF ((phdiff<FillMissingSthl).and.(.not.extinct(trhkl,occup))) THEN
	    OrigReflNr(cnt1)=-1
            MissingNr=MissingNr+1
	    MTotalF=MTotalF+1
	  ENDIF
        ENDIF
      ENDDO
      write(lgu,'(a, i7)') 'Number of reflections treated as missing:',MissingNr
      MTotalF=MTotalF+1 !1 for the F(0) reflection
      MStartF=1;MEndF=MTotalF
      ALLOCATE(MEMRefl(1:MTotalF),STAT=cnt2)
      DO cnt1=1,MTotalF
        ALLOCATE(MEMRefl(cnt1)%hklFG(SDim),STAT=cnt2)
      ENDDO
      IF (cnt2/=0) CALL StopProgram('Error allocating array MEMRefl! Too little memory?')
      IF (((hmclasses/=0).and.(NGroups>0).and.(repeatmode/='never')).or.(FIPS/='no')) THEN !We will need to recover the F's after each attempt 
        ALLOCATE(MEMRefl2Refl(1:MTotalF),STAT=cnt1)
        MEMRefl2Refl=0
        IF (cnt1/=0) CALL StopPRogram('Error allocating array MEMRefl2Refl! Too little memory?')
      ENDIF
!Special handling of F(0)
      IF (OrigReflNr(0)==0) THEN !F(000) is not present in the input file -> initialization to 0
        MEMRefl(1)%F=0.0E0
        MEMRefl(1)%A=0.0E0
        MEMRefl(1)%B=0.0E0
      ELSE
        MEMRefl(1)%F=Refl(OrigReflNr(0))%F
        MEMRefl(1)%A=Refl(OrigReflNr(0))%A
        MEMRefl(1)%B=Refl(OrigReflNr(0))%B
      ENDIF
      MEMRefl(1)%hklfg=0
      MEMRefl(1)%code=0
      MEMRefl(1)%group=0
      MEMRefl(1)%SFloc=0
      cnt2=1; sthlmax=0.0;cover=0
      DO cnt1=1,NVoxTotSF-1
        classNr=OrigReflNr(cnt1)
        IF (ClassNr==0) cycle
        CALL CoordinatesSF(2*cnt1,trhkl)
        DO cnt3=1,SDim
          trhkl(cnt3)=IMapToRange(trhkl(cnt3),-NVox(cnt3)/2,(NVox(cnt3)+1)/2)
        ENDDO
        cnt2=cnt2+1
!       preparation of the coverage calculation.
        if ((calccoverage).and.(ClassNr>-1)) THEN
          phdiff=multsc(float(trhkl),float(trhkl),SDim,.true.) !Gives the length of the diffraction vector
          if (phdiff>eps) THEN
            phdiff=0.5*sqrt(phdiff)
          ELSE
	        phdiff=0.0E0
	      ENDIF
	      sthlmax=max(sthlmax,phdiff)
	      cnt3=int(phdiff/cclsw)+1
          IF (cnt3>mxcoverclasses) THEN
	        CALL ErrorMsg('Error - you have an unrealistic combination of cell parameters and reflection indices')
	        write(Buffer,LargeReflForm) 'Sin(theta)/lambda =',phdiff,' for reflection ',(trhkl(i),i=1,SDim)
            CALL ErrorMsg(Buffer)
	        CALL StopProgram('')
          ENDIF
          IF (.not.extinct(trhkl,occup)) cover(cnt3,1)=cover(cnt3,1)+1        
        ENDIF
        MEMRefl(cnt2)%SFloc=2*cnt1
	    MEMRefl(cnt2)%hklfg=trhkl
        IF (((hmclasses/=0).and.(NGroups>0).and.(repeatmode/='never')).or.(FIPS/='no')) THEN !We will need to recover the F's after each attempt 
          MemRefl2Refl(cnt2)=ClassNr
        ENDIF
        IF (classNr>0) THEN
          MEMRefl(cnt2)%F=sqrt(SF(2*cnt1)**2+SF(2*cnt1+1)**2)
          MEMRefl(cnt2)%A=SF(2*cnt1)
          MEMRefl(cnt2)%B=SF(2*cnt1+1)
          MEMRefl(cnt2)%code=Refl(ClassNr)%code
          MEMRefl(cnt2)%group=Refl(ClassNr)%group !Copy the group number
	    ELSE
          MEMRefl(cnt2)%F=0.0
          MEMRefl(cnt2)%A=0.0
          MEMRefl(cnt2)%B=0.0
          MEMRefl(cnt2)%code=code_missing
          MEMRefl(cnt2)%group=0
	    ENDIF
	    MEMRefl(cnt2)%AMEM=0.0E0
	    MEMRefl(cnt2)%BMEM=0.0E0
	    MEMRefl(cnt2)%FMEM=0.0E0
      ENDDO
      !The reflection array is filled. What remains is doing some statistics and, possibly, modification of the intensities
      IF (ElNr>0) CALL WilsonPlot
      IF (MissingMode==mm_boundsum) CALL WilsonBoundSum
!       IF we have powder or twin, calculate the experimental group amplitudes
      IF (NGroups>0) THEN
        CALL CalculateGroups(1) 
!        IF (FIPS/='no') CALL MakeFIPS
      ENDIF
! Normalize and/or correct for the overall B-factor
      IF ((NormalizeRefl==NormLocal).and.(karleexpand)) THEN
        CALL NormalizeLocal
      ELSE
        DO cnt1=MStartF,MEndF
          IF (NormalizeRefl==NormWilson) THEN
            MEMRefl(cnt1)%F=Normalize(MEMRefl(cnt1)%hklfg,MEMRefl(cnt1)%F,Biso)
            phdiff=phdiff+MEMRefl(cnt1)%F**2
          ELSEIF (NormalizeRefl==NormAtoms) THEN
            MEMRefl(cnt1)%F=Normalize(MEMRefl(cnt1)%hklfg,MEMRefl(cnt1)%F,0.0)
            phdiff=phdiff+MEMRefl(cnt1)%F**2
          ELSE !Only correct for B
            MEMRefl(cnt1)%F=MEMRefl(cnt1)%F*Bfactor(MEMRefl(cnt1)%hklFG,Biso)
          ENDIF
        END DO
      ENDIF
!If zone filter is set, mark out-of-zone reflections as skipped
!      allocate(zoneaxis(1:SDim))
!      zoneaxis=(/0, 1, 0/)
!      DO cnt1=MStartF,MEndF
!        IF (sum(MEMRefl(cnt1)%hklfg*zoneaxis)/=0) MEMrefl(cnt1)%code=code_skip
!      ENDDO

!     The last thing to do is to calculate the coverage statistics.
      IF (calccoverage) THEN
        cnt2=0
        DO cnt1=1,NVoxTotSF-1
          CALL CoordinatesSF(2*cnt1,trhkl)
	  DO cnt3=1,SDim
            trhkl(cnt3)=IMapToRange(trhkl(cnt3),-NVox(cnt3)/2,(NVox(cnt3)+1)/2)
            dtrhkl(cnt3)=float(trhkl(cnt3))
	  ENDDO
!       Exclude higher satellite orders than measured!
      IF (SDim>RDim) THEN
!            IF (any(abs(trhkl(RDim+1:SDim))>maxhkl(RDim+1:SDim))) CYCLE
        goodsat=.false.
        DO cnt3=1,Nqstar
          IF (all(trhkl(RDim+1:SDim)==qstar(1:SDim-RDim,cnt3))) THEN
            goodsat=.true.
            EXIT
          ENDIF
        ENDDO
        IF (.not.goodsat) CYCLE
      ENDIF
!          IF (SDim>RDim) THEN
!	    skip=.false.
!	    DO cnt3=RDim+1,SDim
!	      if(abs(trhkl(cnt3))>maxhkl(cnt3)) skip=.true.
!	    ENDDO
!	    IF (skip) cycle
!          ENDIF
          phdiff=multsc(dtrhkl,dtrhkl,SDim,.true.) !Gives the length of the diffraction vector
          if (phdiff>eps) THEN
	    phdiff=0.5*sqrt(phdiff)
	  ELSE
	    phdiff=0.0E0	  
	  ENDIF
	  if (phdiff>sthlmax) cycle
	  cnt2=cnt2+1
	  cnt3=int(phdiff/cclsw)+1
          IF (.not.extinct(trhkl,occup)) cover(cnt3,2)=cover(cnt3,2)+1        
        ENDDO
        !Write the record into the logfile
        call FramedLabel(2,'Coverage statistics of the expanded reflections by shells:')
        if (terminal) write(*,'(a)') 'Coverage statistics of the expanded reflections by shells:'
        ClCnt1=-7
        ClCnt2=0
        ClassNr=int(sthlmax/cclsw)+1
        do
          ClCnt1=min(ClCnt1+8,ClassNr)
	  ClCnt2=min(ClCnt2+8,ClassNr)
          write(lgu,'(''Resolution (sin(th)/l):'')',ADVANCE='NO')
          if (terminal) write(*,'(''Resolution (sin(th)/l):'')',ADVANCE='NO')
          do cnt1=ClCnt1,ClCnt2
            write(lgu,'(f9.3)',ADVANCE='NO') min(float(cnt1)*cclsw,sthlmax)
	    if (terminal) write(*,'(f9.3)',ADVANCE='NO') min(float(cnt1)*cclsw,sthlmax)
          enddo
          write(lgu,'('''')')
	  if (terminal) write(*,'('''')')
          write(lgu,'(''Resolution (d_min):    '')',ADVANCE='NO')
          if (terminal) write(*,'(''Resolution (d_min):    '')',ADVANCE='NO')
          do cnt1=ClCnt1,ClCnt2
            write(lgu,'(f9.3)',ADVANCE='NO') max(0.5/(float(cnt1)*cclsw),0.5/sthlmax)
            if (terminal) write(*,'(f9.3)',ADVANCE='NO') max(0.5/(float(cnt1)*cclsw),0.5/sthlmax)
          enddo
          write(lgu,'('''')')
          if (terminal) write(*,'('''')')
          write(lgu,'(''Obs. refl. in shell:   '')',ADVANCE='NO')
          if (terminal) write(*,'(''Obs. refl. in shell:   '')',ADVANCE='NO')
          do cnt1=ClCnt1,ClCnt2
            write(lgu,'(i9)',ADVANCE='NO') cover(cnt1,1)
            if (terminal) write(*,'(i9)',ADVANCE='NO') cover(cnt1,1)
          enddo
          write(lgu,'('''')')
          if (terminal) write(*,'('''')')
          write(lgu,'(''Total refl. in shell:  '')',ADVANCE='NO')
          if (terminal) write(*,'(''Total refl. in shell:  '')',ADVANCE='NO')
          do cnt1=ClCnt1,ClCnt2
            write(lgu,'(i9)',ADVANCE='NO') cover(cnt1,2)
            if (terminal) write(*,'(i9)',ADVANCE='NO') cover(cnt1,2)
          enddo
          write(lgu,'('''')')
          if (terminal) write(*,'('''')')
          write(lgu,'(''Coverage in shell:     '')',ADVANCE='NO')
          if (terminal) write(*,'(''Coverage in shell:     '')',ADVANCE='NO')
          do cnt1=ClCnt1,ClCnt2
            IF (cover(cnt1,2)/=0) THEN
              write(lgu,'(f8.1,''%'')',ADVANCE='NO') 100.*float(cover(cnt1,1))/float(cover(cnt1,2))
              if (terminal) write(*,'(f8.1,''%'')',ADVANCE='NO') 100.*float(cover(cnt1,1))/float(cover(cnt1,2))
            ELSE
              write(lgu,'(f8.1,''%'')',ADVANCE='NO') 0.
              if (terminal) write(*,'(f8.1,''%'')',ADVANCE='NO') 0.
	    ENDIF
	  enddo
          write(lgu,'('''')')
          if (terminal) write(*,'('''')')
          write(lgu,'(''Commulative coverage:  '')',ADVANCE='NO')
          if (terminal) write(*,'(''Commulative coverage:  '')',ADVANCE='NO')
          do cnt1=ClCnt1,ClCnt2
	    covsum1=0;covsum2=0;
	    DO cnt2=1,cnt1
               covsum1=covsum1+cover(cnt2,1)
               covsum2=covsum2+cover(cnt2,2)
	    ENDDO
            IF (CovSum2/=0) THEN
              write(lgu,'(f8.1,''%'')',ADVANCE='NO') 100.*float(covsum1)/float(covsum2)
              if (terminal) write(*,'(f8.1,''%'')',ADVANCE='NO') 100.*float(covsum1)/float(covsum2)
            ELSE
              write(lgu,'(f8.1,''%'')',ADVANCE='NO') 0.
              if (terminal) write(*,'(f8.1,''%'')',ADVANCE='NO') 0.
	    ENDIF
	  enddo
          write(lgu,'('''')')
          if (terminal) write(*,'('''')')
          write(lgu,'('''')')
          if (terminal) write(*,'('''')')
          if (ClCnt2==ClassNr) EXIT
        enddo
      ENDIF
      IF (expandedlog.and.((intformatchlow>eps).or.(intformatchhigh<1.0E29))) THEN !Filtering of reflections for fit - list them
        write(lgu,*) 'The limits on amplitudes for matching with reference file are: ', &
              intformatchlow,',',intformatchhigh
        write(lgu,'(a)') 'The following reflections will be used for matching with the reference file:'
        DO i=MStartF,MEndF
          IF ((MEMRefl(i)%F>intformatchlow).and.(MEMRefl(i)%F<intformatchhigh)) THEN
            DO cnt2=1,SDim            
              write(lgu,'(i4)',ADVANCE='NO') MEMRefl(i)%hklFG(cnt2)
            ENDDO
            write(lgu,'(f16.6)') MEMRefl(i)%F
          ENDIF
        ENDDO
      ENDIF
      IF (allocated(qstar)) DEALLOCATE(qstar)
      IF (Terminal) write(*,'(a)') '...successfully finished.'
      END SUBROUTINE ProcessReflections



! Average the input reflections and replace the Refl array with the averaged reflections
!********************************************************************
      SUBROUTINE AverageReflections
!********************************************************************
      USE SF_Module

      IMPLICIT NONE
      TYPE ListNode
        integer :: number
        type(ListNode), pointer :: next
      END TYPE ListNode
      type LinkedList
        Type(ListNode), pointer :: first
      END type LinkedList
      type(LinkedList) :: EquiRefChain
      type(ListNode), pointer :: previousnode, currentnode
      REAL, DIMENSION(1:TotalF) :: refldstar
      INTEGER, DIMENSION(1:TotalF) :: index
      REAL :: multsc, avea,aveb, totalint, dummy, rintnom, rintdenom
      REAL :: phase
      INTEGER :: i, ii, friedel, totaln, newtotalf, cursmop, refrefnr
      INTEGER, DIMENSION(1:SDim) :: hkltmp
      LOGICAL :: isequal
      
      ALLOCATE(EquiRefChain%first)
      NULLIFY(EquiRefChain%first%next)
      !We have to create a list of reflections sorted by sinthl
      DO i=1,TotalF
        refldstar(i)=multsc(float(Refl(i)%hklFG),float(Refl(i)%hklFG),SDim,.true.)
      ENDDO
      CALL indexx(TotalF,refldstar,index)
      !array index contains the position of the ith reflection sorted by dstar
      !now go through the list and average equivalent reflections
      newtotalf=0
      cnt1=0
      rintnom=0.0
      rintdenom=0.0
      open(unit=tmpu,file=TmpFile)
      DO
        cnt1=cnt1+1
        IF (index(cnt1)/=0) THEN !A new reflection
          totaln=0
          avea=0.
          aveb=0.
          totalint=0.
          refrefnr=index(cnt1)
          cnt2=cnt1-1
          currentnode=>EquiRefChain%first
          DO
            cnt2=cnt2+1
            if (cnt2>TotalF) EXIT
            if (index(cnt2)==0) CYCLE
            if (refldstar(index(cnt2))-eps>refldstar(refrefnr)) EXIT
            !Check, if the current reflection really is equivalent to the reference reflection
            isequal=.false.
            DO ii=1,NSymPrim
              hkltmp=multmat(Refl(index(cnt2))%hklfg,SmOp(ii)%R)
              IF (maxval(abs(hkltmp-Refl(refrefnr)%hklFG))==0) THEN
                isequal=.true.
                friedel=1
                cursmop=ii
                EXIT
              ENDIF
              IF (.not.centro) THEN
                IF (maxval(abs(-hkltmp-Refl(refrefnr)%hklFG))==0) THEN
                  isequal=.true.
                  friedel=-1
                  cursmop=ii
                  EXIT
                ENDIF
              ENDIF
            ENDDO
            IF (isequal) THEN
              IF ((refl(index(cnt2))%code/=refl(refrefnr)%code).or.(refl(index(cnt2))%group/=refl(refrefnr)%group)) THEN
                CALL ErrorMsg('Inconsistent input. Some equivalent reflections have different group number')
                CALL StopProgram('or "known-phase" flags. Please check the reflection list.')
              ENDIF
              IF (haveab.or.Refl(refrefnr)%code==code_knownph) THEN
                !transform the phase of the current reflection by the SymOp
                dummy=-2.*pi*sum(Refl(index(cnt2))%hklFG*SmOp(cursmop)%t)+friedel*Phase(Refl(index(cnt2))%A,Refl(index(cnt2))%B)
                totalint=totalint+Refl(index(cnt2))%F**2
                avea=avea+Refl(index(cnt2))%F*cos(dummy)
                aveb=aveb+Refl(index(cnt2))%F*sin(dummy)
              ELSE
                totalint=totalint+Refl(index(cnt2))%F**2
              ENDIF
              totaln=totaln+1
              !insert the number of reflection in the linked list
              currentnode%number=index(cnt2)
              if (.not.associated(currentnode%next)) THEN
                allocate(currentnode%next)
                nullify(currentnode%next%next)
              endif
              currentnode=>currentnode%next
              index(cnt2)=0
            ENDIF
          ENDDO
          !Record the new averaged reflection
          IF (haveab.or.Refl(refrefnr)%code==code_knownph) THEN
            totalint=totalint/totaln
            dummy=phase(avea,aveb)
            avea=sqrt(totalint)*cos(dummy)
            aveb=sqrt(totalint)*sin(dummy)
          ELSE
            totalint=totalint/totaln
            avea=sqrt(totalint)
            aveb=0.
          ENDIF
          write(tmpu,*) Refl(refrefnr)%hklFG,avea,aveb, Refl(refrefnr)%code,  Refl(refrefnr)%group
          newtotalf=newtotalf+1
          !calculate contribution to Rint
          currentnode=>EquiRefChain%first
          IF (totaln>1) THEN
            DO i=1,totaln
              rintnom=rintnom+abs(Refl(currentnode%number)%F**2-totalint)
              rintdenom=rintdenom+totalint
              currentnode=>currentnode%next
            ENDDO
          ENDIF
        ENDIF
        IF (cnt1==TotalF) EXIT
      ENDDO
      !Nullify the linked list
      currentnode=>EquiRefChain%first
      totaln=1
      DO
        if (associated(currentnode%next)) THEN
          previousnode=>currentnode
          currentnode=>previousnode%next
          deallocate(previousnode)
          totaln=totaln+1
        ELSE
          deallocate(currentnode)
          EXIT
        ENDIF
      ENDDO
      !Report on averaging
      call FramedLabel(2,'Averaging of reflections:')
      write(lgu,'(a,f6.3)') 'Redundancy   :',float(TotalF)/float(newtotalf)
      IF (rintdenom>eps) THEN
        write(lgu,'(a,f6.3)') 'Rint         :',100.*rintnom/rintdenom
      ELSE
        write(lgu,'(a)') 'Rint         : -----'
      ENDIF
      write(lgu,*)
      
      REWIND(tmpu)
      DO i=1,TotalAllocatedF
        deallocate(Refl(i)%hklFG)	
      ENDDO
      deallocate(Refl)
      TotalF=NewTotalF
      TotalAllocatedF=NewTotalF
      StartF=1
      EndF=TotalF
      ALLOCATE(Refl(TotalF))
      DO i=1,TotalF
        ALLOCATE(Refl(i)%hklFG(1:SDim))
      ENDDO
      DO i=1,TotalF
        read(tmpu,*) Refl(i)%hklFG,Refl(i)%A,Refl(i)%B, Refl(i)%code, Refl(i)%group
        Refl(i)%F=sqrt(Refl(i)%A**2+Refl(i)%B**2)
      ENDDO
      close(tmpu, STATUS='DELETE')
      END SUBROUTINE AverageReflections

!       Quicksort from Numerical Recipes, changed from F77 to F90
!*******************************************************************************
      SUBROUTINE indexx(n,arr,indx)
!*******************************************************************************
      IMPLICIT NONE
      
      INTEGER :: n,indx(n)
      REAL    :: arr(n), a
      INTEGER, PARAMETER :: M=7, NSTACK=50
      INTEGER :: i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)

      DO j=1,n
        indx(j)=j
      END DO
      jstack=0
      l=1
      ir=n
1     IF (ir-l.lt.M) THEN
        DO j=l+1,ir
          indxt=indx(j)
          a=arr(indxt)
          DO i=j-1,l,-1
            IF (arr(indx(i)).le.a) GOTO 2
            indx(i+1)=indx(i)
          END DO
          i=l-1
2         indx(i+1)=indxt
        END DO
        IF (jstack.eq.0) RETURN
        ir=istack(jstack)
        l=istack(jstack-1)
        jstack=jstack-2
      ELSE
        k=(l+ir)/2
        itemp=indx(k)
        indx(k)=indx(l+1)
        indx(l+1)=itemp
        IF (arr(indx(l)).gt.arr(indx(ir))) THEN
          itemp=indx(l)
          indx(l)=indx(ir)
          indx(ir)=itemp
        END IF
        IF (arr(indx(l+1)).gt.arr(indx(ir))) THEN
          itemp=indx(l+1)
          indx(l+1)=indx(ir)
          indx(ir)=itemp
        END IF
        IF (arr(indx(l)).gt.arr(indx(l+1))) THEN
          itemp=indx(l)
          indx(l)=indx(l+1)
          indx(l+1)=itemp
        END IF
        i=l+1
        j=ir
        indxt=indx(l+1)
        a=arr(indxt)
3       CONTINUE
          i=i+1
        IF (arr(indx(i)).lt.a) GOTO 3
4       CONTINUE
          j=j-1
        IF (arr(indx(j)).gt.a) GOTO 4
        IF (j.lt.i) GOTO 5
        itemp=indx(i)
        indx(i)=indx(j)
        indx(j)=itemp
        GOTO 3
5       indx(l+1)=indx(j)
        indx(j)=indxt
        jstack=jstack+2
        IF (jstack.gt.NSTACK) CALL StopProgram('NSTACK too small in indexx')
        IF (ir-i+1.ge.j-l) THEN
          istack(jstack)=ir
          istack(jstack-1)=i
          ir=j-1
        ELSE
          istack(jstack)=j-1
          istack(jstack-1)=l
          l=i
        END IF
      END IF
      GOTO 1
      END SUBROUTINE indexx

! Return the status of a reflection. If true, reflection is extinct
!********************************************************************
      LOGICAL FUNCTION Extinct(Refl1,symop)
!********************************************************************
      USE SF_Module
      IMPLICIT NONE
      integer, dimension(SDim) :: trhkl,Refl1
      real, dimension(SDim) :: dhkl,dtrhkl
      real :: phdiff
      real, dimension(1:1) :: arone
      integer :: i,ii,symop
      logical :: same
      DO i=1,SDim
        dhkl(i)=float(Refl1(i))
      ENDDO
      Extinct=.false.
      DO i=2,NSym !We have to take all the symops, centering is important, but the first symop is identity, not interesting
        dtrhkl=multmat(dhkl,SmOp(i)%R)
	same=.true.
	DO ii=1,SDim
          trhkl(ii)=nint(dtrhkl(ii))
          if (trhkl(ii)/=Refl1(ii)) THEN
	    same=.false.
	    EXIT
	  ENDIF
	ENDDO
	IF (.not.same) CYCLE
        ! The reflection is depicted onto itself -> check for the phase difference
	call multm(dhkl,SmOp(i)%t,arone,1,SDim,1)
        phdiff=arone(1)
	!Phase difference must be an integer, otherwise the reflection is forbidden
	IF (abs(phdiff-anint(phdiff))>eps) THEN
	  Extinct=.true.
	  symop=i
          EXIT
	ENDIF
      ENDDO
      END FUNCTION Extinct

! Write the information about the extinct reflections to the log-file
!********************************************************************
      SUBROUTINE FriedelRefl(pos1,pos2,FriedelNr)
!********************************************************************
      USE SF_Module
      IMPLICIT NONE
      integer :: pos1,pos2,i,FriedelNr
      character*44 :: drflform='(i5,'':'',..i4,f14.6,5X,i5,'':'',..i4,f14.6)'
      IF (FriedelNr==1) THEN !The first friedel pair, write the header to the logfile
        call FramedLabel(2,'Independently listed Friedel pairs found in the  input file:')
        write(lgu,'(a)') 'For each of the two reflections its position, indices and amplitude are listed.'
      ENDIF
      write(drflform(9:10),'(i2)') SDim
      write(drflform(30:31),'(i2)') SDim
      write(lgu,drflform) pos1,(Refl(pos1)%hklfg(i),i=1,SDim),Refl(pos1)%F,pos2,(Refl(pos2)%hklfg(i),i=1,SDim),Refl(pos2)%F
      END SUBROUTINE FriedelRefl

! Write the information about the extinct reflections to the log-file
!********************************************************************
      SUBROUTINE ExtinctRefl(refl1,pos,symop,ExtinctNr,intens)
!********************************************************************
      USE SF_Module
      IMPLICIT NONE
      integer, dimension(SDim) :: refl1
      integer :: pos,symop,i,ExtinctNr
      real :: intens
      character*60 :: drflform='(i5,'':'',..i4,5X,''SymOp:'',i3,''  Amplitude:'',f10.4)'
      IF (ExtinctNr==1) THEN !The first doubled reflection, write the header to the logfile
        call FramedLabel(2,'Extinct reflections found in the  input file:')        
        write(lgu,'(a)') 'The reflection (number, indices) is extinct due to the indicated symmetry operation.'
      ENDIF
      write(drflform(9:10),'(i2)') SDim
      write(lgu,drflform) pos,(Refl1(i),i=1,SDim),symop,intens
      END SUBROUTINE ExtinctRefl

! Write the information about the doubled reflections to the log-file
!********************************************************************
      SUBROUTINE DoubledRefl(refl1,pos1,intens1,refl2,pos2,intens2,DoubledNr)
!********************************************************************
      USE SF_Module
      IMPLICIT NONE
      integer, dimension(SDim) :: refl1,refl2
      integer :: pos1,pos2,DoubledNr,i
      real :: intens1,intens2
      character*60 :: drflform='(i5,'':'',..i4,f10.4,5X,i5,'':'',..i4,f10.4)'
      IF (DoubledNr==1) THEN !The first doubled reflection, write the header to the logfile
        call FramedLabel(2,'Doubled reflections found in the  input file:')        
        write(lgu,'(a)') 'The first reflection (number, indices,amplitude) is equivalent to the second reflection '// &
	                 ' (number, indices,amplitude) found previously.'
      ENDIF
      write(drflform(9:10),'(i2)') SDim
      write(drflform(30:31),'(i2)') SDim
      write(lgu,drflform) pos1,(Refl1(i),i=1,SDim),intens1,pos2,(Refl2(i),i=1,SDim),intens2
      END SUBROUTINE DoubledRefl      

!Subroutine returning a normalized value of a reflection given the indices, amplitude, and B-factor
!**********************************************
      REAL FUNCTION Normalize(InHKL,InF,InBiso)
!**********************************************
        USE SF_Module
	USE SF_composition
        IMPLICIT NONE
        INTEGER :: i,Inhkl(SDim)
	REAL :: InF,InBiso, sumf,sthl
	REAL :: dhkl(SDim), FormF, Multsc,BFactor
        IF (ElNr==0) CALL StopProgram('Error, attempting to normalize with zero elements.')
	dhkl=float(Inhkl)
	sumf=0
	sthl=multsc(dhkl,dhkl,SDim,.true.)
	sthl=sqrt(sthl)/2.
	DO i=1,ElNr
          sumf=sumf+float(Compos(i,2))*FormF(Compos(i,1),sthl)**2
	ENDDO
	sumf=sqrt(sumf)
	Normalize=Bfactor(Inhkl,InBiso)*InF/sumf
      END FUNCTION Normalize


!Function returning the atomic form factor
!**********************************************
      REAL FUNCTION FormF(el,sthl)
!**********************************************
	USE SF_composition
        IMPLICIT NONE
        INTEGER :: i,el
	REAL :: sthl
	FormF=0.0E0
	DO i=1,AppMemb
          FormF=FormF+FF(i,el)*exp(-FF(i+AppMemb,el)*sthl**2)
	ENDDO
      END FUNCTION FormF

!Subroutine returning a normalized value of a reflections given the indices, amplitude, and B-factor
!**********************************************
      SUBROUTINE WilsonPlot
!**********************************************
        USE SF_Module
	USE SF_Composition
        IMPLICIT NONE
	INTEGER, PARAMETER :: wprows=20
	INTEGER, PARAMETER :: mxwpclasses=200
	INTEGER :: ststart=2
	REAL, DIMENSION(0:mxwpclasses) :: logi
	REAL :: dhkl(SDim), ststep=0.01,stw=0.05
	INTEGER, DIMENSION(0:mxwpclasses) :: nrfl
	INTEGER :: i,ii,start,stp,mxel, wlstep
	REAL :: stolq, formf, multsc,minlogi,maxlogi
        REAL :: w,wx,wy,wxy,wx2, det !Sums for least squares fit
	character*(mxwpclasses+20) :: wpline
	LOGICAL :: weights=.false.
	mxel=0
	nrfl=0
	logi=0.0E0
	DO i=MStartF+1,MEndF
	  IF (MEMRefl(i)%code==code_missing) CYCLE
          dhkl=float(MEMRefl(i)%hklFG)
	  !Calculate sin(th)/l)**2
 	  stolq=multsc(dhkl,dhkl,SDim,.true.)/4.
          start=max(0,int((stolq-stw/2.)/ststep))
	  stp=max(0,int((stolq+stw/2.)/ststep)+1)
          if (stp>mxwpclasses) CYCLE
	  DO ii=start,stp
	    IF (abs(float(ii)*ststep-stolq)<stw/2.) THEN
              logi(ii)=logi(ii)+MEMRefl(i)%F**2
	      nrfl(ii)=nrfl(ii)+1
	      mxel=max(mxel,ii)
	    ENDIF            
	  ENDDO
	ENDDO
	minlogi=1000000.;maxlogi=-10000000000.
        DO i=ststart,mxel
	  if (nrfl(i)>0) THEN
            logi(i)=logi(i)/float(nrfl(i))
            !Calculate the sumf2 for given stolq
            stolq=float(i)*ststep
	    SumfWP=0.0E0
            DO ii=1,ElNr
              SumfWP=SumfWP+float(Compos(ii,2))*FormF(Compos(ii,1),sqrt(stolq))**2
	    ENDDO
            if (logi(i)>eps) THEN
	      logi(i)=log(logi(i)/SumfWP)
	    ELSEIF (i>ststart) THEN
	      logi(i)=logi(i-1)
	    ELSE
	      ststart=ststart+1
	    ENDIF
	  ELSEIF (i>ststart) THEN
	      logi(i)=logi(i-1)
          ELSE
	    ststart=ststart+1
	  endif
	  maxlogi=max(maxlogi,logi(i))
	  minlogi=min(minlogi,logi(i))
	ENDDO
	!Fit a weighted least-squares line
        w=0; wx=0.0; wy=0.0; wxy=0.0; wx2=0.0
	DO i=ststart,mxel
          IF (weights) THEN
            w=w+float(nrfl(i))
	    wx=wx+float(nrfl(i))*float(i)*ststep
	    wx2=wx2+float(nrfl(i))*(float(i)*ststep)**2
	    wy=wy+float(nrfl(i))*logi(i)
	    wxy=wxy+float(nrfl(i))*float(i)*ststep*logi(i)
	  ELSE
            w=w+1
	    wx=wx+float(i)*ststep
	    wx2=wx2+(float(i)*ststep)**2
	    wy=wy+logi(i)
	    wxy=wxy+float(i)*ststep*logi(i)
	  ENDIF
	ENDDO
	det=wx2*w-wx**2!Determinant of the 2x2 matrix
	M2bisoWP=(wxy*w-wy*wx)/det
	LnkWP=(wx2*wy-wx*wxy)/det
	BisoWP=-M2bisoWP/2
        IF ((NormalizeRefl==NormWilson).and.(.not.bisofix)) Biso=BisoWP !We need to correct for the Wilson-plot derived Biso in the normalization procedure
	ScaleWP=sqrt(exp(LnkWP))
	!Write the output on the screen and in the log-file
        call FramedLabel(2,'Wilson Plot:')
	write(lgu,*)
	write(lgu,'(a)') 'circles: <I>/SUM(f^2)'
	write(lgu,'(a,f9.4,a,f8.3,a)') 'dots   : linear fit y =',LnkWP,' + ',M2BisoWP,' * [sin(theta)/lambda]^2'
        write(lgu,*)
        IF (mxel<50) THEN
	  wlstep=2
	ELSE
	  wlstep=1
	ENDIF
	DO i=wprows,0,-1
	  wx=float(i)*(maxlogi-minlogi)/float(wprows)+minlogi !Temporary storage for current scale level
	  wpline=''
	  if (mod(i,wprows/4)==0) THEN
	    write(wpline,'(f10.5,''-|'')') exp(wx)
	  ELSEIF(i==wprows-2) THEN
	    write(wpline,'(''  Scale^2 -|'')')
	  ELSE
	    write(wpline,'(''          -|'')')
          ENDIF
	  wy=(wx-LnkWP)/M2BisoWP
	  ii=nint(wy/ststep)
          DO ii=ststart,mxel
            IF (abs(wx-(float(ii)*ststep*M2BisoWP+LnkWP))<0.5*(maxlogi-minlogi)/float(wprows)) &
	      write(wpline(ii*wlstep+12:ii*wlstep+12),'(''.'')')
	  ENDDO
          DO ii=ststart,mxel
	    IF (nint(float(wprows)*(logi(ii)-minlogi)/(maxlogi-minlogi))==i) write(wpline(ii*wlstep+12:ii*wlstep+12),'(''o'')')
	  ENDDO	  
          write(lgu,'(a)') trim(wpline)
	ENDDO
	wpline=''
	DO i=0,mxel*wlstep
	  write(wpline(12+i:12+i),'(''-'')')
	ENDDO
	ii=-10
	DO
	  ii=ii+10
          IF (ii>mxel) EXIT	  
	  wpline(12+ii*wlstep:12+ii*wlstep)='I'
	ENDDO
        write(lgu,'(a)') trim(wpline)
	ii=-10
	wpline=''
	DO
	  ii=ii+10
          IF (ii>mxel) EXIT	  
	  write(wpline(10+ii*wlstep:),'(f5.2)') ststep*float(ii)
	ENDDO
        write(lgu,'(a)') trim(wpline)
	wpline=''	
	write(wpline(11+wlstep*mxel/2-10:),'(a)') '[sin(theta)/lambda]^2'
        write(lgu,'(a)') trim(wpline)
	write(lgu,*)
	write(lgu,'(a,f10.3)') 'Scale (F_obs = K * F_abs):',ScaleWP
	write(lgu,'(a,f10.3)') 'Biso:                     ',BisoWP
        write(lgu,*)
        END SUBROUTINE WilsonPlot

!**********************************************
      REAL FUNCTION WilsonBound(hkl)
!**********************************************
      USE SF_module
      USE SF_composition
      INTEGER, DIMENSION(SDim) :: hkl
      REAL :: sthl, multsc
      IF (NormalizeRefl/=NormNone) THEN
        WilsonBound=1.0
      ELSE      
        sthl=sqrt(multsc(float(hkl),float(hkl),SDim,.true.))/2.
        WilsonBound=LnkWP+M2BisoWP*sthl**2
        WilsonBound=sqrt(exp(WilsonBound)*SumfWP)      
      ENDIF
      END FUNCTION WilsonBound
      

!**********************************************
      SUBROUTINE WilsonBoundSum
!**********************************************
      USE SF_Module
      IMPLICIT NONE
      INTEGER :: i
      REAL :: WilsonBound
      SumWB=0.
      DO i=MStartF, MEndF
        IF (MEMRefl(i)%code==code_missing) SumWB=SumWB+WilsonBound(MEMRefl(i)%hklFG)
      ENDDO
      END SUBROUTINE WilsonBoundSum
      
!**********************************************
      SUBROUTINE NormalizeLocal
!**********************************************
!avdl freely adapted from CCP4 program ECALC, subroutine Karle
!!avdl What should be done if nf(i)=0? (=number of reflections in a shell) for some i
!avdl This code is not very economic, since it recalculates mult, epsi and stolq several
!avdl times; this can of course be avoided by creating - allocatable - arrays at some point
        USE SF_Module
	USE SF_Interfaces
        IMPLICIT NONE
!former number of shells
	INTEGER, PARAMETER :: nqmax=500
	INTEGER, DIMENSION(nqmax) :: nf
	REAL, DIMENSION(4) :: cc
	REAL, DIMENSION(nqmax) :: sff,sfs
	INTEGER :: i, in, kq, m,jj, istart,iend, nresshellsorig
	REAL :: smin,smax, dq, qm, q, sq, ef, epsi, s, ff
	REAL :: stolq
	REAL, dimension(SDim) :: dhkl	
	REAL FSPLIN, multsc
!
        IF (nresshells==0) THEN !determine the optimal number of shells. Taken over from CCP4, origin of this formula unknown
				!but for typical reflection numbers this equals essentially to 200 reflections per shell on average
          M=200 !number of reflections in one shell
          nresshells = MIN(MAX((TotalF+M/2)/M,1),100)
          IF (nresshells<2) nresshells=2
	ENDIF
    nresshellsorig=nresshells
	smin=100.0
	smax=0.0
	IF (KarleExpand) THEN
	  istart=MStartF+1
	  iend=MEndF
	ELSE
	  istart=1
	  iend=TotalF
        ENDIF
	IF (abs(nresshells).gt.nqmax) CALL StopProgram('Number of shells too large, decrease nresshells <=500')
!avdl -- determine minimum and maximum resolution
        DO i=istart,iend
          IF (KarleExpand) THEN
	    IF (MEMRefl(i)%code==code_missing) CYCLE
	    dhkl=float(MEMRefl(i)%hklFG)
          ELSE
            dhkl=float(Refl(i)%hklFG)
          ENDIF
!Calculate 4*sin(th)/l)**2
 	  stolq=multsc(dhkl,dhkl,SDim,.true.)
          smin=min(smin,stolq)
	  smax=max(smax,stolq)
	ENDDO
!	write(*,'('' smin, smax'',2f14.3)')smin,smax
!avdl -- number of intervals: this is arbitrary!???
!        nqmax
100     CONTINUE !return point if nresshells is too large and zero reflections in szhell are encountered
        DO i=1,abs(nresshells)
          nf(i)=0
          sff(i)=0.0
        ENDDO
!avdl-- start of resolutioin interval, there is something like s**1.5
        qm = smin*sqrt(abs(smin))
!avdl-- step size of interval division
        dq = (smax*sqrt(abs(smax))-qm)/abs(nresshells)
!avdl-- determine <f**2> in each shell
        DO i=istart,iend
          IF (KarleExpand)THEN
            IF (.not.IndepSFpos(MEMRefl(i)%hklFG,MEMRefl(i)%SFloc)) CYCLE
            IF (MEMRefl(i)%code==code_missing) CYCLE
            dhkl=float(MEMRefl(i)%hklFG)
            call muleps(dhkl,m,epsi,SDim)
	        ff=MEMRefl(i)%F
            m=1
!           epsi=1.0
          ELSE
            dhkl=float(Refl(i)%hklFG)
            call muleps(dhkl,m,epsi,SDim)
            ff=Refl(i)%F
          ENDIF
          stolq=multsc(dhkl,dhkl,SDim,.true.)
          q=stolq*sqrt(stolq)
          in=min(int((q-qm)/dq)+1,abs(nresshells))
          nf(in)=nf(in)+m
          sff(in)=sff(in)+m*ff*ff/epsi
        ENDDO
        DO i=1,abs(nresshells)
   	      IF (nf(i).gt.0) THEN
	        sff(i)=sff(i)/nf(i)
          ELSE
	        IF (nresshells>1) THEN
	          nresshells=nresshells-1
	          GOTO 100
	        ELSE
	          CALL StopProgram(' The local normalization procedure failed - number of resolution shells decreased to 1. Stopping.')
	        ENDIF
	      ENDIF
	    ENDDO
!avdl--- some smoothing
	IF (nresshells>1) THEN
          sfs(1)=0.45*sff(1)+0.55*sff(2)
	  sfs(abs(nresshells))=0.45*sff(abs(nresshells))+0.55*sff(abs(nresshells)-1)
          DO i=2,abs(nresshells)-1
            sfs(i)=0.4*sff(i)+0.3*(sff(i-1)+sff(i+1))
	  ENDDO
        ENDIF
!avdl--- some statistics to see what is going on -- from ECALC
      IF (nresshellsorig/=nresshells) WRITE (lgu,'(A,i5,a,i5,a)') 'Warning, number of resolution shells decreased from ',&
                                                                   nresshellsorig,' to',nresshells,'.'
      call FramedLabel(2,'F**2 statistics:')
      WRITE (lgu,'(///,A,A,/)')&
      '    SHELL   NO OF REFLS   RESOLUTION    MEAN DSTAR^3',&
      '     MEAN F^2    SMOOTHED F^2 '
      DO i=1,abs(nresshells)
              s = (qm+i*dq)**(-1./3.)
              q = qm + (i-.5)*dq
          WRITE (lgu,'(1X,I8,6X,I8,F13.2,F16.6,F13.0,F16.0)')i,nf(i),s,q,sff(i),sfs(i)
      ENDDO
!avdl -- initialise spline calculation
        kq=0
	sq = qm + .5*dq
!avdl -- final pass-through to obtain pseudo-E's
	DO i=1,abs(nresshells)
	  nf(i)=0
	  sff(i)=0.0
	ENDDO
        DO i=istart,iend
          IF (KarleExpand)THEN
	    IF (MEMRefl(i)%code==code_missing) CYCLE
            dhkl=float(MEMRefl(i)%hklFG)
            call muleps(dhkl,m,epsi,SDim)
	    ff=MEMRefl(i)%F
	    m=1
!	    epsi=1.0
          ELSE
            dhkl=float(Refl(i)%hklFG)
	    ff=Refl(i)%F
            call muleps(dhkl,m,epsi,SDim)
          ENDIF
 	  stolq=multsc(dhkl,dhkl,SDim,.true.)
	  q=stolq*sqrt(stolq)
          ef = epsi * fsplin(q,sq,dq,sfs,abs(nresshells),kq,cc)
!          ef = epsi * rval(0)
          ef=ff/sqrt(abs(ef))
          IF (KarleExpand) THEN
	     MEMRefl(i)%F=ef
	  ELSE
	     Refl(i)%F=ef
          ENDIF
          IF (IndepSFpos(MEMRefl(i)%hklFG,MEMRefl(i)%SFloc)) THEN
            in=min(int((q-qm)/dq)+1,abs(nresshells))
	    nf(in)=nf(in)+m
	    sff(in)=sff(in)+m*ef*ef/epsi
	  ENDIF
	  IF (nresshells.lt.0)THEN
             IF (KarleExpand) THEN
 	  write(lgu,'(3i5,f9.4,f9.2,i3,2f9.4)')(int(dhkl(in)),in=1,3),ef,MEMRefl(i)%F,m,epsi,q
             ELSE
          write(lgu,'(3i5,f9.4,i3,2f9.4)')(int(dhkl(in)),in=1,3),Refl(i)%F,m,epsi,q
             ENDIF
	  ENDIF
	ENDDO
      call FramedLabel(2,'E**2 statistics from Karle normalization:')
	      WRITE (lgu,'(///,A,A,/)')&
      '    SHELL   NO OF REFLS   RESOLUTION    MEAN DSTAR^3',&
      '     MEAN E^2  '
      DO i=1,abs(nresshells)
              s = (qm+i*dq)**(-1./3.)
              q = qm + (i-.5)*dq
          WRITE (lgu,'(1X,I8,6X,I8,F13.2,F16.6,F13.3)')i,nf(i),s,q,sff(i)/nf(i)
      ENDDO
      ff=0.0
      jj=0
      do i=1,abs(nresshells)
        ff=ff+sff(i)
	jj=jj+nf(i)
      ENDDO
      write(lgu,'('' <E**2>_pseudo = '',f9.3)')ff/jj
      END SUBROUTINE NormalizeLocal


!     ======================================
      REAL FUNCTION FSPLIN(X,XO,XD,YA,N,J,C)
!     ======================================
!
!==== CUBIC SPLINE INTERPOLATION.
!==== X  = INTERPOLATION POINT IN X.
!==== XO = FIRST SAMPLE POINT IN X.
!==== XD = INCREMENT IN X BETWEEN SAMPLE POINTS.
!==== YA = ARRAY OF N SAMPLE POINTS IN Y.
!==== N  = NUMBER OF SAMPLE POINTS.
!==== J  = RETURNED ARRAY INDEX. MUST BE SET TO ZERO BEFORE FIRST CALL.
!==== C  = RETURNED ARRAY OF 4 CUBIC COEFFICIENTS. J AND C MUST NOT BE
!         MODIFIED BETWEEN CALLS.
!==== FSPLIN = INTERPOLATED VALUE IN Y.
!
      IMPLICIT NONE
!     .. Scalar Arguments ..
      INTEGER :: J,N
      REAL :: X,XD,XO
!     ..
!     .. Array Arguments ..
      REAL :: C(4),YA(N)
!     ..
!     .. Local Scalars ..
      INTEGER :: I,I1
      REAL :: G1,G2,S,U0,U1,U2,V0,V1,V2,X1,XD2,YD,YD1
!     ..
!     .. Intrinsic Functions ..
      INTRINSIC MIN
!     ..
      IF (N.LE.2) THEN
        J = 1
        IF (N.EQ.1) THEN
          FSPLIN = YA(1)
        ELSE
          FSPLIN = YA(1) + (YA(2)-YA(1))*(X-XO)/XD
        ENDIF
      ELSE
!==== COMPUTE WHICH INTERVAL TO USE.
        I = MIN(MAX(INT((X-XO)/XD),0)+1,N-1)
        X1 = XO + (I-1)*XD
!==== SAME AS LAST CALL ?
        IF (I.NE.J) THEN
!==== NO, RESET FOR NEXT CALL.
          J = I
!
!==== COMPUTE WEIGHTED DISPLACEMENT VECTOR (U1,V1) BETWEEN I AND I+1.
          XD2 = XD*XD
          I1 = I + 1
          YD1 = YA(I1) - YA(I)
          S = XD2 + YD1*YD1
          U1 = XD/S
          V1 = YD1/S
          C(1) = YA(I)
!
          IF (I.GT.1) THEN
!==== COMPUTE VECTOR (U0,V0) BETWEEN I-1 AND I, AND GRADIENT G1 AT I.
            YD = YA(I) - YA(I-1)
            S = XD2 + YD*YD
            U0 = XD/S
            V0 = YD/S
            G1 = (V0+V1)/ (U0+U1)
!
            IF (I1.EQ.N) GOTO 10
          ENDIF
!==== COMPUTE VECTOR (U2,V2) BETWEEN I+1 AND I+2, AND GRADIENT G2 AT I+1.
          YD = YA(I1+1) - YA(I1)
          S = XD2 + YD*YD 
          U2 = XD/S
          V2 = YD/S
          G2 = (V1+V2)/ (U1+U2)
!
          IF (I.EQ.1) THEN
!==== CASE OF FIRST INTERVAL, COMPUTE QUADRATIC COEFFIENTS.
            C(2) = YD1/XD
            C(3) = G2 - C(2)
            C(2) = C(2) - C(3)
            C(3) = C(3)/XD
            C(4) = 0.
          ELSE
!==== GENERAL CASE, COMPUTE CUBIC COEFFICENTS.
            C(2) = G1
            C(3) = YD1/XD
            C(4) = G1 + G2 - 2.*C(3)
            C(3) = (C(3)-C(4)-G1)/XD
            C(4) = C(4)/XD2
          ENDIF
          GOTO 20
!
!==== CASE OF LAST INTERVAL, COMPUTE QUADRATIC COEFFICIENTS.
10        C(2) = G1
          C(3) = (YD1/XD-G1)/XD
          C(4) = 0.
        ENDIF
!
!==== COMPUTE FUNCTION VALUE AND RETURN.
20      S = X - X1
        FSPLIN = C(1) + S*(C(2)+S*(C(3)+S*C(4)))
      ENDIF
      END FUNCTION FSPLIN


     SUBROUTINE MULEPS(dhkl,m,e,n1)
     USE SF_Module
! this is only needed if the reduced data set is normalized
!     =================
!
!==== CALC MULTIPLICITY AND EPSILON
!
!
      IMPLICIT NONE


      INTEGER :: m
      REAL :: e
!
      INTEGER :: i,j,k,n1,jj
!     ..
!     ..
!
      REAL, dimension(n1) :: dhkl,dhkl2	
      REAL, DIMENSION(1:6), PARAMETER :: MultCoef= &       ! Multiplicity coefficients for reflection comparaison
        (/1.,97.,945.,9987.,99976.,999933./)

      e = 1.
      m = 1
      IF (NSym.gt.1) THEN
        dhkl2=multmat(dhkl,SmOp(1)%R)
        i=0
        DO jj=1,n1
          i = i+ int(dhkl2(jj)*MultCoef(jj))
	ENDDO
        DO k = 2,NSymPrim
          dhkl2=multmat(dhkl,SmOp(k)%R)
          j=0
	  DO jj=1,n1
            j = j+ int(dhkl2(jj)*MultCoef(jj))
	  ENDDO
          IF (i.eq.j .OR. i.eq.-j) m = m + 1
          IF (i.eq.j) e = e + 1.
        ENDDO
        m = NSymPrim/m
      ENDIF
      END SUBROUTINE MULEPS



!Subroutine concentrating all the operations on the reflection array in case of powders
!**********************************************
      SUBROUTINE MakeFIPS
!**********************************************
        USE SF_Module
	USE SF_Composition
        USE FFT_Interfaces
        IMPLICIT NONE
        real :: alpha, Alpha0,StepAlpha,DeltaY,SumY1,SumY2,PrevDeltaY,LastCycDeltaY,MaxRho1,MaxRho2
	real :: MinDeltaY(1:3,1:2)
	integer :: i,ii,mlt,FIPSCycle,previous1,previous2, boostcycle
        integer :: MaxFIPSCycles

        !Write initial log-file and terminal information:
	IF (Terminal) THEN
	  write(*,'(a)') 'Repartitioning the intensites using FIPS:'
          SELECT CASE (FIPS)
	    CASE('auto')
              write(*,'(a)') '  I will try to determine the best boost factor automatically.'	    
	    CASE('table')
              write(*,'(a)') '  A table of FIPS results will be produced.'	    
	    CASE('given')
              write(*,'(a)') '  The parameters of FIPS were read from the input file:'	    
              write(*,'(a,f6.3)') '    Boostfactor:',FIPSParams(1)
              write(*,'(a,i3)')   '    Cycles     :',nint(FIPSParams(2))
          END SELECT
	ENDIF
        CALL FramedLabel(2,'Repartitioning the intenisites using FIPS:')
        SELECT CASE (FIPS)
	  CASE('auto')
            write(lgu,'(a)') '  I will try to determine the best boost factor automatically.'	    
	  CASE('table')
            write(lgu,'(a)') '  A table of FIPS results will be produced.'	    
	  CASE('given')
            write(lgu,'(a)') '  The parameters of FIPS were read from the input file:'	    
            write(lgu,'(a,f6.3)') '    Boostfactor:',FIPSParams(1)
            write(lgu,'(a,i3)')   '    Cycles     :',nint(FIPSParams(2))
        END SELECT
        IF ((expandedLog).or.(FIPS=='table')) THEN
	  write(lgu,*)
	  write(lgu,'(a)') 'What follows is the record of the FIPS iterations. Maximum 15 cycles are performed for each Alpha.'
	  write(lgu,'(a)') 'Each line represents one run with given alpha.'
	  write(lgu,'(a)') 'Each entry in line stands for one FIPS cycle and has the form DeltaY/R-value.'
	  write(lgu,'(a)') 'R-value is calculated with respect to the input amplitudes.'
          write(lgu,*)
	ENDIF        

        ! Reequipartition the data
	DO i=1,NGroups
	  cnt1=0
	  DO ii=1,MTotalF
	    IF (MEMRefl(ii)%group==i) cnt1=cnt1+1
	  ENDDO
	  DO ii=1,MTotalF
	    if (MEMRefl(ii)%group==i) THEN
              MEMRefl(ii)%F=sqrt(GroupInt(i)%orig/float(cnt1))
	    ENDIF
	  ENDDO	  
	ENDDO
        DO i=MStartF, MEndF
	  SF(MEMRefl(i)%SFloc)=MEMRefl(i)%F
	ENDDO
        DO i=StartF,EndF
	  Refl(i)%AMEM=Refl(i)%F !Store the original in AMEM
	  Refl(i)%F=SF(abs(Refl(i)%SFLoc))
	  Refl(i)%BMEM=Refl(i)%F !Store the equipartitioning in BMEM
        ENDDO
        IF (FIPS=='table') THEN
          !Calculate the R-value
          SumY1=0.0E0;SumY2=0.0E0
	  DO i=StartF,EndF
            IF (Refl(i)%group/=0) THEN
  	      SumY1=SumY1+(Refl(i)%AMEM-Refl(i)%F)**2
              SumY2=SumY2+Refl(i)%AMEM**2
            ENDIF
	  ENDDO
          write(*,'(a,f10.2)') 'Equipartitioned R-value:',100.*SumY1/SumY2
        ENDIF
!       Calculte F(0) from the composition and Wilson-plot derived scale
        MEMRefl(MStartF)%F=0.0
	IF (ElNr>0) THEN
          DO i=1,ElNr
            MEMRefl(MStartF)%F=MEMRefl(MStartF)%F+float(Compos(i,1)*Compos(i,2))
	  ENDDO
          MEMRefl(MstartF)%F=MEMRefl(MstartF)%F*ScaleWP
        ENDIF


        SELECT CASE (FIPS)
          CASE('auto') !Search for the optimal settings
            MaxFipsCycles=10
            Alpha=0.0; StepAlpha=1.0
            PrevDeltaY=1.0E30
            ! First localize roughly the correct alpha
	    DO
              Alpha=Alpha+StepAlpha
              CALL OneFIPS
	      IF (DeltaY>PrevDeltaY) EXIT
	      PrevDeltaY=DeltaY
            ENDDO
            !Iterate Around the alpha with the lowest DeltaY
            Alpha=Alpha-StepAlpha
            MinDeltaY(1,1)=1.0E30
            MinDeltaY(3,1)=1.0E30
	    MinDeltaY(2,1)=PrevDeltaY
	    MinDeltaY(2,2)=Alpha	
	    DO !Repeat until reasonably low DeltaY is achieved
	      StepAlpha=StepAlpha/4.
1000          CONTINUE
	      Alpha0=MinDeltaY(2,2)
              IF (MinDeltaY(1,1)>1.0E29) THEN
                Alpha=Alpha0-StepAlpha
                CALL OneFips
                MinDeltaY(1,1)=DeltaY
	        MinDeltaY(1,2)=Alpha
	      ENDIF
              IF (MinDeltaY(3,1)>1.0E29) THEN	  
                Alpha=Alpha0+StepAlpha
	        CALL OneFips
                MinDeltaY(3,1)=DeltaY
	        MinDeltaY(3,2)=Alpha
              ENDIF
	      IF (MinDeltaY(1,1)<MinDeltaY(2,1)) THEN
                MinDeltaY(3,1)=MindeltaY(2,1)
                MinDeltaY(3,2)=MindeltaY(2,2)
                MinDeltaY(2,1)=MindeltaY(1,1)
                MinDeltaY(2,2)=MindeltaY(1,2)
	        MindeltaY(1,1)=1.0E30
	        GOTO 1000
	      ELSEIF (MinDeltaY(3,1)<MinDeltaY(2,1)) THEN
                MinDeltaY(1,1)=MindeltaY(2,1)
                MinDeltaY(1,2)=MindeltaY(2,2)
                MinDeltaY(2,1)=MindeltaY(3,1)
                MinDeltaY(2,2)=MindeltaY(3,2)
	        MindeltaY(3,1)=1.0E30
	        GOTO 1000
	      ENDIF
	      IF (abs(MinDeltaY(1,1)-MinDeltaY(2,1))<0.01) EXIT
              MinDeltaY(1,1)=1.0E30
              MinDeltaY(3,1)=1.0E30
	    ENDDO
	    !Run one last FIPS iteration with the correct parameters
            IF (ExpandedLog) THEN
	      WRITE(lgu,*)
	      WRITE(lgu,'(a)') 'Final FIPS run:'
	    ENDIF
	    Alpha=MinDeltaY(2,2)
	    CALL OneFIPS
            IF (Terminal) THEN
	      write(*,'(a,f8.4)') 'Optimal boost factor:',Alpha
              write(*,'(a,f8.4)') 'Delta Y             :',DeltaY
            ENDIF
            write(lgu,'(a,f8.4)') 'Optimal boost factor:',Alpha
            write(lgu,'(a,f8.4)') 'Delta Y             :',DeltaY
          CASE ('given')
            Alpha=FIPSParams(1)
            MaxFipsCycles=nint(FIPSParams(2))
	    CALL OneFips
	  CASE ('table')
            MaxFipsCycles=nint(FIPSParams(4))
	    BoostCycle=nint((FIPSParams(2)-FIPSParams(1))/FIPSParams(3))
	    DO mlt=0,BoostCycle
	      Alpha=FIPSParams(1)+float(mlt)*FIPSParams(3)
	      CALL OneFIPS
	    ENDDO
            CALL StopProgram('A table of FIPS parameters has been produced. The program will stop now.')
        END SELECT

        !MEMRefl%F already contains the FIPSed amplitudes, now rewrite also the A and B components and update the Refl array        
	DO i=MStartF+1,MEndF
	  SumY1=MEMRefl(i)%A
	  SumY2=MEMRefl(i)%B
	  SumY1=sqrt(SumY1**2+SumY2**2)
	  IF (SumY1>eps) THEN
            MEMRefl(i)%A=MEMRefl(i)%A*MEMRefl(i)%F/SumY1
	    MEMRefl(i)%B=MEMRefl(i)%B*MEMRefl(i)%F/SumY1
          ELSE
            MEMRefl(i)%A=0.0E0
	    MEMRefl(i)%B=0.0E0
	  ENDIF
	  !Put the updated values into the SF array
	  SF(MEMRefl(i)%SFLoc)=MEMRefl(i)%A
	  SF(MEMRefl(i)%SFLoc+1)=MEMRefl(i)%B
	ENDDO
!        DO i=StartF,EndF
        DO i=StartF,EndF
	  Refl(i)%A=SF(abs(Refl(i)%SFLoc))
	  Refl(i)%B=SF(abs(Refl(i)%SFLoc)+1)
	  IF (Refl(i)%SFLoc<0) Refl(i)%B=-Refl(i)%B
	  Refl(i)%F=sqrt(Refl(i)%A**2+Refl(i)%B**2)
	ENDDO
      CONTAINS
	SUBROUTINE OneFIPS
	IMPLICIT NONE
          !Fill MEMRefl by the original (equipartitioned)
	  DO i=MStartF+1,MEndF
	    IF (MEMRefl(i)%code/=code_missing) THEN
              MEMRefl(i)%F=Refl(MEMRefl2Refl(i))%F
	    ELSE
	      MEMRefl(i)%F=0.0
	    ENDIF
	  ENDDO
	  FIPSCycle=0
	  IF ((ExpandedLog).or.(FIPS=='table')) write(lgu,'(a,f7.4,a)', ADVANCE='NO') 'Alpha=',alpha,':'
	  IF (FIPS=='table') write(*,'(a,f7.4,a)', ADVANCE='NO') 'Alpha=',alpha,':'
	  LastCycDeltaY=1000000.
	  DO
	     FIPSCycle=FipsCycle+1
	    !Fill SF with the intensities
	    SF=0.0E0
            DO i=MStartF,MEndF
              SF(MEMRefl(i)%SFloc)=MEMRefl(i)%F**2
	    ENDDO
	    CALL FFT(sf2rho)
	    maxrho1=0.0E0
	    DO i=0,NVoxTot-1
              maxrho1=max(maxrho1,rho(i))
	    ENDDO
	    !Square rho and get new structure factors
            rho=rho**2
	    maxrho2=0.0E0
	    DO i=0,NVoxTot-1
              maxrho2=max(maxrho2,rho(i))
	    ENDDO	    	    
            !Normalize rho to avoid too large numbers in subsequent calculations
	    rho=rho*(maxrho1/maxrho2)
	    CALL FFT(rho2sf)
            DO i=MStartF+1,MEndF
	      IF ((SF(MEMRefl(i)%SFloc)>=0.0E0).and.(MEMRefl(i)%code/=code_missing)) THEN 
	        MEMRefl(i)%FMEM=sqrt(max(SF(abs(Refl(MEMRefl2Refl(i))%SFloc)),0.0E0)) !To maintain the perfect symmetry, otherwise the exponentiation tends to destroy it
              ELSE
	        MEMRefl(i)%FMEM=MEMRefl(i)%F
	      ENDIF
	    ENDDO
	    CALL CalculateGroups(2) !Calculate the current group intensities
	    DO i=MStartF+1,MEndF
  	      IF (MEMRefl(i)%group==0) CYCLE !Do nothing with the separated reflections
	      ii=MEMRefl(i)%group
	      IF (GroupInt(ii)%cur>eps) THEN
	        MEMRefl(i)%FMEM=MEMRefl(i)%FMEM*sqrt(GroupInt(ii)%orig/GroupInt(ii)%cur)
	      ELSE
	        MEMRefl(i)%FMEM=0.0E0
	      ENDIF
	    ENDDO

	    !Repartition the overlaping intensities by the new squared Patterson, see eq.7 in Estermann and Gramlich, J.Appl.C. 26,396-404
            DO i=MStartF+1,MEndF
              IF (MEMRefl(i)%group==0) CYCLE !Do nothing with the separated reflections
              IF (Refl(MEMRefl2Refl(i))%F>eps) THEN
                IF (abs(alpha*(MEMRefl(i)%FMEM-Refl(MEMRefl2Refl(i))%F)/Refl(MEMRefl2Refl(i))%F)<20.) THEN
                  MEMRefl(i)%FMEM=sqrt(exp(alpha*((MEMRefl(i)%FMEM-Refl(MEMRefl2Refl(i))%F)/Refl(MEMRefl2Refl(i))%F)))
	         ELSE
		   MEMRefl(i)%FMEM=sqrt(exp(30.0E0))
	        ENDIF
	      ELSE
                MEMRefl(i)%FMEM=0.0
	      ENDIF
	    ENDDO
            !And renormalize them again to the right group intensities
            CALL CalculateGroups(2)
	    DO i=MStartF+1,MEndF
	      IF (MEMRefl(i)%group==0) THEN
	        IF (MEMRefl(i)%code/=code_missing) THEN
                  MEMRefl(i)%FMEM=Refl(MEMRefl2Refl(i))%F
		  MEMRefl(i)%F=MEMRefl(i)%FMEM
                ENDIF
                CYCLE !Do nothing with the separated reflections, just assign the observed amplitude
	      ENDIF
	      ii=MEMRefl(i)%group
	      IF (GroupInt(ii)%cur>eps) THEN
                MEMRefl(i)%F=MEMRefl(i)%FMEM*sqrt(GroupInt(ii)%orig/GroupInt(ii)%cur)
	      ELSE
	        MEMRefl(i)%F=0.0E0
	      ENDIF
            ENDDO
            !One cycle is finished, calculate the statistics. The statistics involves normalized intensities
	    !and reflections _consecutive_ in the diffraction pattern -> use array Refl to get position, but
            !MEMRefl (SF) to get the current amplitude
            DO i=MStartF+1,MEndF
              SF(MEMRefl(i)%SFloc)=MEMRefl(i)%F  !It's easier to find the amplitudes in SF than in MEMRefl from Refl
            ENDDO
            cnt1=0;SumY1=0.0;cnt2=0;SumY2=0.0;previous1=0;previous2=0
	    DO i=2,TotalF
	      IF (Refl(i)%group==0) THEN
	        if (previous1==0) THEN
		  previous1=i
		ELSE
		  cnt1=cnt1+1
                   IF (SF(abs(Refl(previous1)%SFLoc))>eps) THEN
                     SumY1=SumY1+(abs(SF(abs(Refl(i)%SFLoc))-SF(abs(Refl(previous1)%SFLoc)))&
		                 /SF(abs(Refl(previous1)%SFLoc)))
                   ENDIF
                  previous1=i
		ENDIF
	      ELSE
	        IF (previous2==0) THEN
		  previous2=i
		ELSEIF (Refl(previous2)%group/=Refl(i)%group) THEN
		  previous2=0
		ELSE
                  cnt2=cnt2+1
!                  SumY2=SumY2+abs(Normalize(Refl(i)%hklFG,SF(abs(Refl(i)%SFLoc)),BisoWP)-   &
!		                  Normalize(Refl(previous2)%hklFG,SF(abs(Refl(previous2)%SFLoc)),BisoWP))
!                  SumY2=SumY2+abs(Normalize(Refl(i)%hklFG,SF(abs(Refl(i)%SFLoc)),BisoWP)**2-   &
!		                  Normalize(Refl(previous2)%hklFG,SF(abs(Refl(previous2)%SFLoc)),BisoWP)**2)
                   IF (SF(abs(Refl(previous2)%SFLoc))>eps) THEN
!                     SumY2=SumY2+(abs(SF(abs(Refl(i)%SFLoc))**2-SF(abs(Refl(previous2)%SFLoc))**2)&
!		                 /SF(abs(Refl(previous2)%SFLoc))**2)
                     SumY2=SumY2+(abs(SF(abs(Refl(i)%SFLoc))-SF(abs(Refl(previous2)%SFLoc)))&
		                 /SF(abs(Refl(previous2)%SFLoc)))
                   ENDIF
                  previous2=i
                ENDIF
	      ENDIF
	    ENDDO
            SumY1=SumY1/float(cnt1)
            SumY2=SumY2/float(cnt2)
            DeltaY=((SumY1-SumY2)/SumY1)**2
! Write a file with the group R-values
!	    write(rhonm,'(a11,i3.3,a1,i2.2,a4)') 'zsm_myfips_',nint(10.*Alpha),'_',FIPSCycle,'.lst'
!	    open(88,file=rhonm)
!	    SumY1=0.0;SumY2=0.0
!	    IF (Refl(1)%group/=0) THEN
!              SumY1=SumY1+(Refl(1)%AMEM-SF(abs(refl(1)%SFLoc)))**2
!              SumY2=SumY2+Refl(1)%AMEM**2
!	    ENDIF
!	    DO i=StartF+1,EndF
!	      if ((Refl(i)%group==Refl(i-1)%group).and.(Refl(i)%group/=0)) THEN
!                SumY1=SumY1+(Refl(i)%AMEM-SF(abs(refl(i)%SFLoc)))**2
!	        SumY2=SumY2+Refl(i)%AMEM**2
!	      elseif (Refl(i-1)%group/=0) THEN
!	        write(88,'(i4,f10.5)') Refl(i-1)%group,100.*SumY1/SumY2
!		SumY1=0.0;SumY2=0.0
!	      ENDIF
!	      IF (Refl(i)%group>0) WRITE(88,'(3i4,3f10.5,i4)') (refl(i)%hklfg(ii),ii=1,3), &
!	                                 Refl(i)%AMEM,Refl(i)%BMEM,SF(refl(i)%SFLoc),refl(i)%group
!  	    ENDDO
!	    close(88)
            !Calculate the R-value
	    SumY1=0.0E0;SumY2=0.0E0
	    DO i=StartF,EndF
	      IF (Refl(i)%group/=0) THEN
                SumY1=SumY1+(Refl(i)%AMEM-SF(abs(refl(i)%SFLoc)))**2
	        SumY2=SumY2+Refl(i)%AMEM**2
!                SumY1=SumY1+(Refl(i)%BMEM-SF(abs(refl(i)%SFLoc)))**2
!	        SumY2=SumY2+Refl(i)%BMEM**2
	      ENDIF
	    ENDDO
            IF ((ExpandedLog).or.(FIPS=='table')) write(lgu,'(1X,f9.3,''/'',f6.2)',ADVANCE='NO') &
	         min(DeltaY,9999.999),100.*SumY1/SumY2
            IF (FIPS=='table') write(*,'(1X,f9.3,''/'',f6.2)',ADVANCE='NO') min(DeltaY,9999.999),100.*SumY1/SumY2
            IF (FIPS=='auto') THEN
              SumY1=abs(DeltaY/LastCycDeltaY-1.0)
	    ELSE
	      SumY1=1.0E30
	    ENDIF
	    IF ((SumY1<0.001).or.(FIPSCycle>=MaxFipsCycles)) EXIT
	    LastCycDeltaY=DeltaY
	  ENDDO	
	  IF ((ExpandedLog).or.(FIPS=='table')) write(lgu,*)
	  IF (FIPS=='table') write(*,*)
	END SUBROUTINE OneFIPS
      END SUBROUTINE MakeFIPS

!Calculate the Group intensity of overlapping reflections. key=1 - get original input group intensities; key=2 - fill current intensities
!**********************************************
      SUBROUTINE CalculateGroups(key)
!**********************************************
        USE SF_Module
        IMPLICIT NONE
        INTEGER :: i,ii,key

          DO i=1,NGroups
	    IF (key==1) GroupInt(i)%orig=0.0E0
	    IF (key==2) GroupInt(i)%cur=0.0E0
          ENDDO  
	  DO i=MStartF+1,MEndF
	    IF (MEMRefl(i)%group==0) CYCLE
	    ii=MEMRefl(i)%group
            IF (MEMRefl(i)%hklfg(1)==0) THEN !These reflections are twice in the list due to the special layout of the array SF
              IF (key==1) GroupInt(ii)%orig=GroupInt(ii)%orig+0.5*(MEMRefl(i)%F**2)
              IF (key==2) GroupInt(ii)%cur=GroupInt(ii)%cur+0.5*(MEMRefl(i)%FMEM**2)
            ELSE
              IF (key==1) GroupInt(ii)%orig=GroupInt(ii)%orig+MEMRefl(i)%F**2
              IF (key==2) GroupInt(ii)%cur=GroupInt(ii)%cur+MEMRefl(i)%FMEM**2
	    ENDIF
	  ENDDO

      END SUBROUTINE CalculateGroups


!Subroutine returning the position of an integer array in the pixel array
!**********************************************
      SUBROUTINE ArrayLocation(InCoo,OutLoc)
!**********************************************
        USE SF_Module
        integer, dimension(SDim) :: InCoo
        integer :: OutLoc,i,htmp
        OutLoc=0
        DO i=1,SDim
          htmp=InCoo(i)
          DO WHILE (htmp>=NVox(i))
            htmp=htmp-NVox(i)
          END DO
          DO WHILE (htmp<0)
            htmp=htmp+NVox(i)
          END DO
          OutLoc=OutLoc+DimStep(i)*htmp !*IMapToRange(InCoo(i),0,NVox(i))
        ENDDO
      END SUBROUTINE ArrayLocation

!Subroutine returning the position of an integer array in the pixel array
!**********************************************
      SUBROUTINE ArrayLocationSF(InCoo,OutLoc,friedel)
!**********************************************
        USE SF_Module
        integer, dimension(SDim) :: InCoo
	integer :: OutLoc,i,htmp,friedel
        OutLoc=0;friedel=1
	DO i=1,SDim
          htmp=friedel*InCoo(i)
          DO WHILE (htmp>=NVox(i))
            htmp=htmp-NVox(i)
          END DO
          DO WHILE (htmp<0)
            htmp=htmp+NVox(i)
          END DO
	  IF ((i==1).and.(htmp>(NVoxSF(1)-1))) THEN
	    friedel=-1
	    htmp=NVox(1)-htmp
	  ENDIF
	  OutLoc=OutLoc+DimStepSF(i)*htmp
	ENDDO
	OutLoc=2*OutLoc
      END SUBROUTINE ArrayLocationSF

!Subroutine returning the position of an integer array in the pixel array
!A copy of ArrayLocation, but with NVox and DimStep as arguments
!**********************************************
      SUBROUTINE ArrayLocationSFGen(InCoo,InNVox,InNVoxSF,InDimStepSF,OutLoc,friedel)
!**********************************************
        USE SF_Module
        integer, dimension(SDim) :: InCoo, InNVox, InNVoxSF,InDimStepSF
        integer :: OutLoc,i,htmp,friedel
        OutLoc=0;friedel=1
        DO i=1,SDim
          htmp=friedel*InCoo(i)
          DO WHILE (htmp>=InNVox(i))
            htmp=htmp-InNVox(i)
          END DO
          DO WHILE (htmp<0)
            htmp=htmp+InNVox(i)
          END DO
          IF ((i==1).and.(htmp>(InNVoxSF(1)-1))) THEN
            friedel=-1
            htmp=InNVox(1)-htmp
          ENDIF
          OutLoc=OutLoc+InDimStepSF(i)*htmp
        ENDDO
        OutLoc=2*OutLoc
      END SUBROUTINE ArrayLocationSFGen

!Subroutine returning the coordinates in the grid for given arraylocation
!**********************************************
      SUBROUTINE Coordinates(inLoc,OutCoo)
!**********************************************
        USE SF_Module
        integer, dimension(SDim) :: OutCoo
	integer :: i,InLoc,InLoc2,rest
        inLoc2=inLoc
        rest=inloc2
	DO i=SDim,1,-1
	  outCoo(i)=rest/Dimstep(i)
	  rest=mod(rest,DimStep(i))
        ENDDO
      END SUBROUTINE Coordinates

!Subroutine returning the coordinates in the grid for given arraylocation in SF array
!**********************************************
      SUBROUTINE CoordinatesSF(inLoc,OutCoo)
!**********************************************
        USE SF_Module
        integer, dimension(SDim) :: OutCoo
	integer :: i,InLoc,InLoc2,rest
        inLoc2=inLoc/2
        rest=inloc2
	DO i=SDim,1,-1
	  outCoo(i)=rest/DimstepSF(i)
	  rest=mod(rest,DimStepSF(i))
        ENDDO
      END SUBROUTINE CoordinatesSF

!Subroutine returning the coordinates in the grid for given arraylocation in SF array
!A copy of CoordinatesSF, but with NVoxSF and DimStepSF as arguments
!**********************************************
      SUBROUTINE CoordinatesSFGen(inLoc,InDimStepSF,OutCoo)
!**********************************************
        USE SF_Module
	integer, dimension(SDim) :: InNVoxSF, InDimStepSF
        integer, dimension(SDim) :: OutCoo
	integer :: i,InLoc,InLoc2,rest
        inLoc2=inLoc/2
        rest=inloc2
	DO i=SDim,1,-1
	  outCoo(i)=rest/InDimstepSF(i)
	  rest=mod(rest,InDimStepSF(i))
        ENDDO
      END SUBROUTINE CoordinatesSFGen

!Subroutine updating the MEMRefl array by the values from the SF array
!**********************************************
      SUBROUTINE SF2MEMRefl
!**********************************************
      USE SF_Module
      integer :: i,sfloc,sflocp1
      real :: a,b
      DO i=MStartF,MEndF
        sfloc=MEMRefl(i)%SFloc
	sflocp1=sfloc+1
        a=SF(sfloc); b=SF(sflocp1)
        MEMRefl(i)%AMEM=a
        MEMRefl(i)%BMEM=b
        MEMRefl(i)%FMEM=sqrt(a*a+b*b)
      ENDDO
      END SUBROUTINE SF2MEMRefl

!Subroutine assigning the MEMRefl values to the array SF
!**********************************************
      SUBROUTINE MEMRefl2SF
!**********************************************
      USE SF_Module
      integer :: i,sfloc,sflocp1
      SF=0.0E0
      DO i=MStartF,MEndF
        sfloc=MEMRefl(i)%SFloc
        sflocp1=sfloc+1
        IF (MEMRefl(i)%code/=code_skip) THEN
          SF(sfloc)=MEMRefl(i)%AMEM
          SF(sflocp1)=MEMRefl(i)%BMEM
        ENDIF
      ENDDO
      END SUBROUTINE MEMRefl2SF

!Subroutine updating the Refl array by the values from the SF array
!**********************************************
      SUBROUTINE SF2Refl
!**********************************************
      USE SF_Module
      integer :: i
      DO i=StartF,EndF
        Refl(i)%AMEM=SF(abs(Refl(i)%SFloc))
        Refl(i)%BMEM=SF(abs(Refl(i)%SFloc)+1)*sign(1,Refl(i)%SFloc)
        Refl(i)%FMEM=sqrt(Refl(i)%AMEM**2+Refl(i)%BMEM**2)
      ENDDO
      END SUBROUTINE SF2Refl

!Function that will shift the first argument so that LBound<=Number<UBound
!**************************************************************
      INTEGER FUNCTION IMapToRange(Number,LBound,UBound)
!**************************************************************
      integer :: Number,LBound,UBound
      integer :: step,shift
	  
      shift=Number
      step=UBound-LBound
      DO
        IF (shift<LBound) THEN
          shift=shift+step
        ELSEIF (shift>=UBound) THEN
          shift=shift-step
        ELSE
          EXIT
        ENDIF
      ENDDO
      IMapToRange=shift
      END FUNCTION IMapToRange
!*******************************

      LOGICAL FUNCTION TimeForAction(cyc,start,interval)
        IMPLICIT NONE
	integer :: cyc, start, interval
          IF (start==0) THEN
	    TimeForAction=.false.	    
	  ELSEIF ((cyc<start).or.(mod(cyc-start,interval)/=0)) THEN
	    TimeForAction=.false.	    
	  ELSE
	    TimeForAction=.true.
	  ENDIF
      END FUNCTION TimeForAction
      
! Matrix multiplication of REAL matrices
!***********************************************************
      subroutine multm(a,b,c,n1,n2,n3)
!***********************************************************
      real :: a(1:n1*n2),b(1:n2*n3),c(1:n1*n3)
      real :: p
      integer :: i,j,k,ik,jk,jkp
      do i=1,n1
        jkp=1
        ik=i
        do k=1,n3
          ij=i
          jk=jkp
          p=0.
          do j=1,n2
            p=p+a(ij)*b(jk)
            ij=ij+n1
            jk=jk+1
          enddo
          c(ik)=p
          ik=ik+n1
          jkp=jkp+n2
        enddo
      enddo
      end subroutine multm

! Matrix multiplication of INTEGER matrices
!***********************************************************
      subroutine multmI(a,b,c,n1,n2,n3)
!***********************************************************
      integer :: a(1:n1*n2),b(1:n2*n3),c(1:n1*n3)
      integer :: p
      integer :: i,j,k,ik,jk,jkp
      do i=1,n1
        jkp=1
        ik=i
        do k=1,n3
          ij=i
          jk=jkp
          p=0
          do j=1,n2
            p=p+a(ij)*b(jk)
            ij=ij+n1
            jk=jk+1
          enddo
          c(ik)=p
          ik=ik+n1
          jkp=jkp+n2
        enddo
      enddo
      end subroutine multmI
      
! returns the scalar product of two vectors
!******************************************************************
      REAL FUNCTION multsc(iVec1,iVec2,Dim,iRec)
!******************************************************************
      USE SF_Module
      
      IMPLICIT NONE
      INTEGER :: Dim !Dimension of the input vectors, allowed only RDim or SDim
      INTEGER :: l1,l2 !loop1, loop2
      REAL, DIMENSION(Dim) :: iVec1,iVec2 ! Input vectors
      REAL, DIMENSION(Dim) :: dvec1,dvec2,dvec3
      REAL, DIMENSION(1:1) :: mul
      LOGICAL :: iRec !Is the vector in reciprocal space?      
      multsc=0.0E0
      IF (iRec) THEN
        DO l1=1,Dim
	  dvec1(l1)=ivec1(l1)
	  dvec2(l1)=ivec2(l1)
	ENDDO
        IF (Dim>RDim) THEN !Reciprocal lattice vector, to determine the scalar product we have to add qvectors
!         add q-vectors
	  DO l1=1,Dim-RDim
	    DO l2=1,RDim
	      dvec1(l2)=dvec1(l2)+dvec1(l1+RDim)*qvec(l2,l1)
	      dvec2(l2)=dvec2(l2)+dvec2(l1+RDim)*qvec(l2,l1)
	    END DO
	  END DO
	ENDIF
	call multm(dvec1,Gi,dvec3,1,RDim,RDim)
        call multm(dvec3,dvec2,mul,1,RDim,1)
      ELSE
        call multm(ivec1,G,dvec1,1,RDim,RDim)
        call multm(dvec1,ivec2,mul,1,RDim,1)
      END IF
      multsc=mul(1)
      END FUNCTION multsc

! returns the sin(theta)/lambda value of a reciprocal vector
!******************************************************************
      REAL FUNCTION sinthl(iVec)
!******************************************************************
      USE SF_Module
      
      IMPLICIT NONE
      REAL, DIMENSION(SDim) :: iVec
      REAL :: multsc
      sinthl=sqrt(multsc(iVec,iVec,SDim,.true.))/2.
      END FUNCTION sinthl

!Subroutine that makes a product of two symmetry operations
!**************************************************
      SUBROUTINE ProdOp(InR1,T1,InR2,T2,IRout,Tout,n)
!**************************************************
      real, parameter :: eps=0.01
      integer :: n,i
      integer, dimension(1:n,1:n) :: InR1,InR2,IRout
      real, dimension(1:n,1:n) :: R1,R2,Rout
      real, dimension(1:n) :: T1,T2,Tout,temp

      R1=float(InR1)
      R2=float(InR2)
      call multm(R2,R1,Rout,n,n,n)
      call multm(R2,t1,temp,n,n,1)
      Tout=T2+temp
      DO i=1,n
        DO
	  IF (Tout(i)<-eps) THEN
	    Tout(i)=Tout(i)+1.0E0
	  ELSEIF (Tout(i)>(1.E0-eps)) THEN
	    Tout(i)=Tout(i)-1.0E0
	  ELSE
	    EXIT
          ENDIF
        ENDDO
      ENDDO
      IRout=nint(Rout)
      END SUBROUTINE ProdOp

! function returning phase of a complex number with components iA and iB
!******************************************************
       FUNCTION Phase(iA,iB)
!******************************************************

       IMPLICIT NONE
       real :: iA,iB, phase

       phase=atan2(iB,iA)

       END FUNCTION Phase

! double precision variant of the function above
!******************************************************
       FUNCTION DPhase(diA,diB)
!******************************************************

       IMPLICIT NONE
       
       double precision, parameter :: eps=0.000001
       double precision :: dPhase,diA,diB,amp
       
       amp=sqrt(diA*diA+diB*diB)
       IF (amp>eps) THEN
         DPhase=dacos(diA/amp)
       ELSE
         DPhase=0.0D0
       END IF
       if (diB<0.0) dPhase=-dPhase        
       END FUNCTION DPhase

! Change the grid resolution of rho, and all other quantities related to it
!******************************************************************
      SUBROUTINE ChangeGrid(NewNVox)
!******************************************************************
      USE SF_Module
      USE SF_Interfaces
      USE FFT_Interfaces
      IMPLICIT NONE
      INTEGER, DIMENSION(SDim) :: NewNVox
      INTEGER, DIMENSION(SDim) :: OldNVox
      INTEGER, DIMENSION(SDim) :: NewDimStep, NewDimStepSF, NewNVoxSF, Coo
      INTEGER :: NewNVoxTot, NewNVoxTotSF
      INTEGER :: i, ErrCode, friedel, loc

      OldNVox=NVox
      CALL ResampleDensity(OldNVox, NewNVox, rho)
!      if (allocated(sumrho)) CALL ResampleDensity(OldNVox, NewNVox, sumrho)
      CALL NVoxRelatedArrays(NewNVox,NewNVoxSF, NewDimStep, NewDimStepSF, NewNVoxTot, NewNVoxTotSF)
      DEALLOCATE(SF)
      ALLOCATE(sf(0:2*NewNVoxTotSF-1),STAT=ErrCode)
      IF (ErrCode/=0) CALL StopProgram('Error reallocating array sf! Too little memory?')
      NVox=NewNVox
      NVoxSF=NewNVoxSF
      DimStep=NewDimStep
      DimStepSF=NewDimStepSF
      NVoxTot=NewNVoxTot
      NVoxTotSF=NewNVoxTotSF
      cApVx=Cell(NCellPar+1)/NVoxTot

      DO i=1,TotalF
	IF (any(Refl(i)%hklfg>NVoxSF-1)) &
	  CALL StopProgram('Error, resampling to a grid that cannot hold all reflection indices! Stoping...')
	Call ArrayLocationSF(Refl(i)%hklfg,Loc,friedel)
	Refl(i)%SFLoc=Loc*friedel
      ENDDO
      DO i=MStartF, MEndF
	IF (any(MEMRefl(i)%hklfg>NVoxSF-1)) &
	  CALL StopProgram('Error, resampling to a grid that cannot hold all reflection indices! Stoping...')
	Call ArrayLocationSF(MEMRefl(i)%hklfg,MEMRefl(i)%SFLoc,friedel)
	IF (friedel<0) THEN 
	 write(*,*) MEMRefl(i)%hklfg,MEMRefl(i)%SFLoc,friedel, NVoxSF 
         CALL StopProgram('Friedel must not be negative in ChangeGrid. Internal error of the program, please contact the author.')
        ENDIF
      ENDDO      
      CALL MakeFFTPlan(NVox,rho,SF,r2cft,fastfft,rho2sf)
      CALL MakeFFTPlan(NVox,rho,SF,c2rft,fastfft,sf2rho)
      !And because MakeFFTPlan transform can destroy the contents of the arrays, we have to recalculate rho
      CALL MEMRefl2SF
      CALL FFT(sf2rho)
      END SUBROUTINE ChangeGrid


!******************************************************************
      SUBROUTINE NVoxRelatedArrays(InNVox,OutNVoxSF, OutDimStep, OutDimStepSF, OutNVoxTot, OutNVoxTotSF)      
!******************************************************************
      USE SF_Module
      IMPLICIT NONE
      INTEGER, DIMENSION(SDim) :: InNVox, OutDimStep, OutDimStepSF, OutNVoxSF
      INTEGER :: OutNVoxTot, OutNVoxTotSF
      INTEGER :: nraloop
      OutNVoxSF=InNVox; OutNVoxSF(1)=InNVox(1)/2+1
      OutDimStepSF(1)=1
      OutDimStep(1)=1
      DO nraloop=2,SDim
        OutDimStep(nraloop)=OutDimStep(nraloop-1)*InNVox(nraloop-1)
	OutDimStepSF(nraloop)=OutDimStepSF(nraloop-1)*OutNVoxSF(nraloop-1)
      ENDDO
      OutNVoxTot=product(InNVox)
      OutNVoxTotSF=product(OutNVoxSF)

      END SUBROUTINE NVoxRelatedArrays


!******************************************************************
      SUBROUTINE ResampleDensity(NVox1,NVox2,inrho)
!******************************************************************
      USE SF_Module
      USE FFT_Interfaces
      IMPLICIT NONE
      INTEGER, DIMENSION(SDim) :: NVox1,NVox2      
      real, allocatable, dimension(:) :: inrho

      INTEGER, DIMENSION(SDim) :: NVoxSF1, NVoxSF2, DimStep1, DimStep2, DimStepSF1, DimStepSF2
      INTEGER, DIMENSION(SDim) :: Coo, Coo2
      real, allocatable, dimension(:) :: SF1, SF2
      integer :: dim, NVoxTotSF1, NVoxTotSF2, Loc, Loc2, friedel, i
      integer :: ErrCode
      TYPE(FFTInstr) :: FFTInstr1,FFTInstr2
      CHARACTER*(FileLength) :: densityfile
      !Allocate the SF arrays
      NVoxSF1=NVox1; NVoxSF1(1)=NVox1(1)/2+1; NVoxTotSF1=product(NVoxSF1)
      NVoxSF2=NVox2; NVoxSF2(1)=NVox2(1)/2+1; NVoxTotSF2=product(NVoxSF2)
      DimStepSF1(1)=1
      DimStepSF2(1)=1
      DO cnt1=2,SDim
	DimStepSF1(cnt1)=DimStepSF1(cnt1-1)*NVoxSF1(cnt1-1)
        DimStepSF2(cnt1)=DimStepSF2(cnt1-1)*NVoxSF2(cnt1-1)
      ENDDO      
      ALLOCATE(sf1(0:2*NVoxTotSF1-1),STAT=ErrCode)
      IF (ErrCode/=0) CALL StopProgram('Error allocating arrays for resampling of the density. Too little memory?')

      !Create the FFT plan for the first transform
      CALL MakeFFTPlan(NVox1,inrho,sf1,r2cft,fastfft,fftinstr1)
      CALL FFT(fftinstr1,inrho)

      !deallocate the density - not needed anymore, and allocate the new SF array
      DEALLOCATE(inrho, STAT=ErrCode)
       
      ALLOCATE(sf2(0:2*NVoxTotSF2-1),STAT=ErrCode)
      IF (ErrCode/=0) CALL StopProgram('Error allocating arrays for resampling of the density. Too little memory?')
      allocate(inrho(0:product(NVox2)-1), STAT=ErrCode)
      CALL MakeFFTPlan(NVox2,inrho,sf2,c2rft,fastfft,FFTInstr2)

      !Transfer the structure factors from SF1 to SF2
      SF2=0.0
      DO cnt1=0,2*NVoxtotSF1-1,2
        CALL CoordinatesSFGen(cnt1,DimStepSF1,Coo)
	DO cnt2=2,SDim
	  IF (coo(cnt2)>1+(NVox1(cnt2)-1)/2) coo(cnt2)=coo(cnt2)-NVox1(cnt2)
        ENDDO
	IF (any(abs(coo)>(NVox2-1)/2)) CYCLE !This can happen if the resolution is reduced. 2DO - aliasing here!
	CALL ArrayLocationSFGen(Coo,NVox2,NVoxSF2,DimStepSF2,Loc,friedel)
	SF2(loc)=SF1(cnt1)
	SF2(loc+1)=SF1(cnt1+1)
      ENDDO      
      
      !Now dealocate SF1 and allocate again inrho
      deallocate(SF1)
      !create the FFT plan for the back transform to the new rho, and make the transform
      CALL FFT(FFTInstr2, inrho)
      deallocate(SF2)
      END SUBROUTINE ResampleDensity
      
!       Initalisation
!******************************************************************
      SUBROUTINE MakeFFTPlan(InNVox,inrho,insf,dir,infastfft,instr)
!******************************************************************
      USE SF_Module
      IMPLICIT NONE
      INTEGER, PARAMETER :: FFTW_FORWARD=+1
      INTEGER, PARAMETER :: FFTW_BACKWARD=-1
      INTEGER, PARAMETER :: FFTW_MEASURE=0
      INTEGER, PARAMETER :: FFTW_PATIENT=32
      INTEGER, PARAMETER :: FFTW_ESTIMATE=64
      INTEGER, DIMENSION(:) :: InNVox
      REAL, TARGET, DIMENSION(:) :: inrho, insf
      INTEGER :: Method, dir, i
      LOGICAL :: InFastFFT
      TYPE(FFTInstr) :: instr
      
      IF (InFastFFT) THEN
        Method=FFTW_MEASURE
      ELSE
        Method=FFTW_ESTIMATE        
      ENDIF
      instr%dir=dir
      IF (dir==r2cFT) THEN
        call sfftw_plan_dft_r2c(instr%plan,size(InNVox),InNVox,inrho,insf,method)
        instr%norm=1.
      ELSEIF (dir==c2rft) THEN
        call sfftw_plan_dft_c2r(instr%plan,size(InNVox),InNVox,insf,inrho,method)
        instr%norm=product(InNVox)
      ELSE
         CALL StopProgram('Error, Wrong FFT direction. Internal error of the program, please contact author.')
      ENDIF
      IF (instr%plan==0) &
         CALL StopProgram('Error, cannot initiate the FFT routine. The communication with FFTW3 does not work. Sorry.')
      END SUBROUTINE MakeFFTPlan

!     Interface to the FFTW subroutine
!******************************************************************
      SUBROUTINE FFT(instr, inrho)
!******************************************************************
!
      USE SF_Module
      IMPLICIT NONE
      TYPE(FFTInstr) :: instr
      REAL, OPTIONAL, TARGET, DIMENSION(:) :: inrho
   
      call sfftw_execute(instr%plan)
      IF (instr%dir==-1) THEN
        IF (present(inrho)) THEN
          inrho=inrho/instr%norm
        ELSE
          rho=rho/instr%norm
        ENDIF
      ENDIF

      END SUBROUTINE FFT

! Function that determines the value of the Recl factor / differs between DEC and other systems
!******************************************************
       INTEGER FUNCTION GetRecLFac()
!******************************************************
       USE SF_MODULE
       IMPLICIT NONE
       integer, parameter :: recordsize=4
       integer :: size,status, nameln,orecl
       integer(KIND=4) :: dummy
       orecl=recordsize
       nameln=len_trim(TmpFile)
       open (tmpu,access='DIRECT', file=TmpFile(:nameln), form='UNFORMATTED',STATUS='UNKNOWN',recl=orecl)
       dummy=132
       write(tmpu,rec=1) dummy
       close(tmpu)
       call statc(TmpFile(:nameln), nameln, size, status)
       if (size==recordsize) THEN 
         GetRecLFac=recordsize
       elseif (size==recordsize**2) THEN
         GetRecLFac=1
       else
         CALL StopProgram('Error detecting record size for direct file access. Please contact the author of the program.')
       endif
       open(tmpu,file=TmpFile(:nameln))
       close(tmpu,status='delete')
       end function GetRecLFac
       
       
! Subroutine that generates a Patterson or superposition minimum map
!******************************************************
       SUBROUTINE SupPatterson(key)
!******************************************************
       USE SF_Module
       USE SF_rhoiointerfaces
       USE FFT_Interfaces
       IMPLICIT NONE
       integer :: key, i,ii, sfloc,sflocp1, arrloc, maxi
       real :: maxrho, phdiff, phase, multsc, length
       integer, dimension(1:SDim) :: Dxyz, startC, EndC, Dxyz2, Dxyz3, AxOrd
       real, dimension(1:SDim) :: RDxyz
       logical :: ismax, IsHarker
       AxOrd=(/(i,i=1,SDim)/)

       SF=0.0E0
       DO i=MStartF,MEndF
         sfloc=MEMRefl(i)%SFloc
	 sflocp1=sfloc+1
         SF(sfloc)=MEMRefl(i)%F**2
         SF(sflocp1)=0.0E0
       ENDDO
       CALL FFT(sf2rho)
       IF (key==pattersonmodel) THEN !Try to remove the origin peak
         length=sum(rho)/NVoxTot !Calculate the mean value         
         DO i=1,NCenter
	   Dxyz2(1:RDim)=1+int(2.0/(Cell(1:RDim)/NVox(1:RDim))) !Define the box with side 2 angstrom
	   IF (SDim>RDim) Dxyz2(RDim+1:SDim)=NVox(RDim+1:SDim)/2
	   StartC=-Dxyz2+nint(Center(1:SDim,i)*NVox)
           EndC=Dxyz2+nint(Center(1:SDim,i)*NVox)
           Dxyz=StartC
           DO
             CALL Arraylocation(Dxyz,arrloc)
             rho(arrloc)=length
	     IF (.not.NextCoo(SDim,StartC,EndC,Dxyz,AxOrd)) EXIT
	   ENDDO	 
         ENDDO
	 modelrho=rho
       ENDIF
!       call writexplor(rho,'patterson_noorigpeak.xplor')
!       call stopprogram('patterson no orig')
       IF (key==supminfunction) THEN
         !Now find a highest peak that is not an origin peak
         maxrho=0.0
         DO i=1,NVoxTot-1 !Skipping i=0, that's the origin peak
           if ((rho(i)<maxrho).or.(rho(i)>0.2*rho(0))) CYCLE !Second condition to avoid pseudosymmetric peaks
	   CALL Coordinates(i,Dxyz)
           StartC=Dxyz-1
           EndC=Dxyz+1
           Dxyz2=StartC
	   ismax=.true.
	   DO
	     CALL Arraylocation(Dxyz2,arrloc)
	     if (rho(arrloc)>rho(i)) THEN
	       ismax=.false.
	       exit
	     ENDIF
	     IF (.not.NextCoo(SDim,StartC,EndC,Dxyz2,AxOrd)) EXIT
	   ENDDO
           IF ((ismax).and.(rho(i)>maxrho)) THEN
	     !Test, if it is a Harker vector and not too close to the origin	     
	     length=100.
	     DO ii=1,NCenter
               RDxyz=float(Dxyz)/NVox-Center(1:SDim,ii)
	       RDxyz=From0to1(RDxyz)
	       where(RDxyz>0.5) RDxyz=RDxyz-1.
               length=min(length,sqrt(multsc(RDxyz(1:RDim),RDxyz(1:RDim),RDim,.false.)))
	     ENDDO
	     IF ((.not.IsHarker(float(Dxyz)/float(NVox))).and.(length>2.5)) THEN
	       maxi=i
	       maxrho=rho(i)
	     ENDIF
	   ENDIF
         ENDDO
         call Coordinates(maxi,Dxyz)
!         write(*,*) 'Selected peak:',maxi, maxrho, float(Dxyz)/NVox, &
!              sqrt(multsc(float(Dxyz(1:3))/NVox(1:3),float(Dxyz(1:3))/NVox(1:3),RDim,.false.))
!        Now calculate a patterson shifted by -Dxyz
         RDxyz=float(Dxyz)/float(NVox)
         SF=0.0E0
         DO i=MStartF,MEndF
           sfloc=MEMRefl(i)%SFloc
           sflocp1=sfloc+1
           phdiff=-2.*pi*sum(MEMRefl(i)%hklfg*RDxyz)
           SF(sfloc)=MEMRefl(i)%F**2*cos(phdiff)
           SF(sflocp1)=MEMRefl(i)%F**2*sin(phdiff)
         ENDDO
         modelrho=rho
         CALL FFT(sf2rho)
         !Now calculate the minimum superposition function:
         modelrho=min(rho,modelrho)
         rho=modelrho
!         call writexplor(rho,'patterson_sup.xplor')
!         call stopprogram('The superposition map is saved under patterson_sup.xplor. Stopping.')
       ENDIF
! Replace the amplitudes by the experimental amplitudes. Not necessary actually
!       rho=modelrho
!       call FFT(rho2sf)
!       call sf2memrefl
!       sf=0.0
!       DO i=MStartF,MEndF
!         sfloc=MEMRefl(i)%SFloc
!	 sflocp1=sfloc+1
!	 phdiff=phase(MEMRefl(i)%AMEM,MEMRefl(i)%BMEM)
!         SF(sfloc)=MEMRefl(i)%F*cos(phdiff)
!         SF(sflocp1)=MEMRefl(i)%F*sin(phdiff)
!       ENDDO
!       call fft(sf2rho)
!       modelrho=rho
!       call writexplor(rho,'patterson_sup.xplor')
!       call stopprogram('patterson written')
       END SUBROUTINE SupPatterson
       
! This function is true if the input vector is a Harker vector of the structure
!******************************************************
       LOGICAL FUNCTION IsHarker(Vec)
!******************************************************
       USE SF_Module
       USE SF_SymSearch
       IMPLICIT NONE
       integer, PARAMETER :: NTry=1
       REAL :: MY_RAND
       REAL, DIMENSION(1:SDim) :: Vec, RandVec,SumRandH
       REAL, DIMENSION(1:SDim,1:SDim) :: ImR
       INTEGER :: i, ii, iii

      IsHarker=.false.
      DO i=1,NSym !All symops including the centering
!        write(*,*) 'Input vector:', Vec
!	write(*,*) 'SymOp:',i
        !A Harker vector h fulfills the equation h + t = (I-R)x, where x is some vector, and {R|t} is a symmetry operation.
	ImR=UnitMat(SDim)-SmOp(i)%R
        SumRandH=0.0
	DO ii=1,NTry
	  DO iii=1,SDim
	    RandVec(iii)=MY_RAND()
	  ENDDO
	  SumRandH=SumRandH+matmul(ImR,RandVec)-SmOp(i)%t
!	  write(*,*) 'SumRandH:',SumRandH
	ENDDO
        SumRandH=From0to1(SumRandH-NTry*Vec)
	IF (any(abs(SumRandH)<eps)) THEN
!	  write(*,*) 'Neprosel'
	  IsHarker=.true.
	  EXIT
	ELSE
!	  write(*,*) 'Prosel'
	ENDIF
      ENDDO
!      write(*,*) '***** Konecny vysledek:',IsHarker
!      write(*,*)
!      read(*,*)
      END FUNCTION IsHarker

! Subroutine preparing the SF array and calling fourier transform.
!******************************************************
       SUBROUTINE Fourier(direction)
!******************************************************
       USE SF_Module
       USE FFT_Interfaces
       IMPLICIT NONE
       integer :: i,direction,GetSeedFromTime
       real :: dummy, MY_RAND, phase,normsample,GaussDev
       IF (Direction==-1) THEN
         IF (AutoRS) RandomSeed=GetSeedFromTime()
         CALL MY_SRAND(randomseed)
         SF=0.0E0
         DO i=MStartF,MEndF
           dummy=Phase(MEMRefl(i)%A,MEMRefl(i)%B)
           normsample=(2.*MY_RAND()-1.)*2.*randfourierfactor
!           normsample=abs(randfourierfactor)*GaussDev()
           dummy=dummy+2*pi*normsample
           SF(MEMRefl(i)%SFloc)=MEMRefl(i)%F*cos(dummy)
           SF(MEMRefl(i)%SFloc+1)=MEMRefl(i)%F*sin(dummy)
!           write(*,*) i, MEMRefl(i)%hklfg, dummy,MEMRefl(i)%F*cos(dummy),MEMRefl(i)%F*sin(dummy), MEMRefl(i)%A, MEMRefl(i)%B
!	   write(*,*) SF(MEMRefl(i)%SFloc), SF(MEMRefl(i)%SFloc+1)
         ENDDO       
         CALL SF2MEMRefl !To have the "MEM" components updated
         CALL SF2Refl    !To have the "MEM" components updated
         CALL FFT(sf2rho)
       ELSEIF (Direction==1) THEN
         CALL FFT(rho2sf)
         CALL SF2MEMRefl
         CALL SF2Refl
         IF (InPerform=='symmetry') THEN !Assign also MEMRefl(i)%F
	   DO i=MStartF,MEndF
	     MEMRefl(i)%F=MEMRefl(i)%FMEM
           ENDDO
         ENDIF
       ENDIF
       END SUBROUTINE Fourier 

      subroutine MY_SRAND(rseed)
        INTEGER :: rseed, seed_size
	INTEGER, ALLOCATABLE, DIMENSION(:) :: arrseed
        !  Initialize the random number seed.
        call random_seed()
        !  Determine the size of the random number seed.
        call random_seed(size=seed_size)
        !  Allocate a seed of the right size.
        allocate (arrseed(seed_size))
        arrseed=rseed
        call random_seed(put=arrseed)
	deallocate(arrseed)
      end subroutine MY_SRAND

      real function MY_RAND()
        call random_number(my_rand)
      end function MY_RAND

! Shift coordinates in the hypercube between StartC and EndC
!*****************************************************************************
      FUNCTION NextCoo(Dim,StartC,EndC,Coo,AxOrd)
!*****************************************************************************
      IMPLICIT NONE

      LOGICAL :: NextCoo
      INTEGER :: Dim,i
      INTEGER, DIMENSION(Dim) :: StartC,EndC,Coo,AxOrd
        IF (Dim==0) THEN
	  NextCoo=.false.
	  goto 200
	ENDIF
        NextCoo=.true.
        i=1
100     Coo(AxOrd(i))=Coo(AxOrd(i))+1
        IF (Coo(AxOrd(i))>EndC(AxOrd(i))) THEN
          Coo(AxOrd(i))=StartC(AxOrd(i))
          i=i+1
          IF (i==Dim+1) THEN
	    NextCoo=.false.
	    goto 200
	  END IF
	  goto 100
        END IF      
200     RETURN
      END FUNCTION NextCoo

! Calculate the f of the average atom at given indices
!*****************************************************************************
      FUNCTION UFactor(hkl)
!*****************************************************************************
      USE SF_Module

      IMPLICIT NONE
      real :: a(6),b(6) !parameters of the fitting
      integer, DIMENSION(SDim) :: hkl
      real, DIMENSION(SDim) :: dhkl
      integer :: cnt
      real :: UFactor, sthl, multsc
      DATA a/2.09921E0,1.80832E0,1.26159E0,0.56775E0,0.26303E0,0.00010E0/
      DATA b/13.18997E0,30.37956E0,0.69255E0,0.16381E0,68.42774E0,0.44083E0/
      
      dhkl=float(hkl)
      sthl=0.5*sqrt(multsc(dhkl,dhkl,SDim,.true.))
      ufactor=0.0E0
      DO cnt=1,6
	ufactor=ufactor+a(cnt)*exp(-b(cnt)*sthl**2)
      END DO
      UFactor=6./UFactor
      END FUNCTION UFactor

! Calculate the B-factor of a iven reflection
!*****************************************************************************
      FUNCTION BFactor(hkl,B)
!*****************************************************************************
      USE SF_Module

      IMPLICIT NONE
      
      integer, DIMENSION(SDim) :: hkl
      real, DIMENSION(SDim) :: dhkl
      real :: BFactor, sthl, multsc,B
      
      dhkl=float(hkl)
      sthl=0.5*sqrt(multsc(dhkl,dhkl,SDim,.true.))
      BFactor=exp(B*sthl**2)
      END FUNCTION BFactor

!From Numerical recipes in Fortran
!Returns the kth smallest value in the array arr(1:n). The input array will be rearranged
!to have this value in location arr(k), with all smaller elements moved to arr(1:k-1) (in
!arbitrary order) and all larger elements in arr[k+1..n] (also in arbitrary order).
!**********************
      FUNCTION select(k,n,arr)
      INTEGER k,n
      REAL select,arr(n)
      INTEGER i,ir,j,l,mid
      REAL a,temp
      l=1
      ir=n
1     if(ir-l.le.1)then
        if(ir-l.eq.1)then
          if(arr(ir).lt.arr(l))then
            temp=arr(l)
            arr(l)=arr(ir)
            arr(ir)=temp
          endif
        endif
        select=arr(k)
        return
      else
        mid=(l+ir)/2
        temp=arr(mid)
        arr(mid)=arr(l+1)
        arr(l+1)=temp
        if(arr(l).gt.arr(ir))then
          temp=arr(l)
          arr(l)=arr(ir)
          arr(ir)=temp
        endif
        if(arr(l+1).gt.arr(ir))then
          temp=arr(l+1)
          arr(l+1)=arr(ir)
          arr(ir)=temp
        endif
        if(arr(l).gt.arr(l+1))then
          temp=arr(l)
          arr(l)=arr(l+1)
          arr(l+1)=temp
        endif
        i=l+1
        j=ir
        a=arr(l+1)
3       continue
          i=i+1
        if(arr(i).lt.a)goto 3
4       continue
          j=j-1
        if(arr(j).gt.a)goto 4
        if(j.lt.i)goto 5
        temp=arr(i)
        arr(i)=arr(j)
        arr(j)=temp
        goto 3
5       arr(l+1)=arr(j)
        arr(j)=a
        if(j.ge.k)ir=j-1
        if(j.le.k)l=i
      endif
      goto 1
      END

! Check, if the current Rho should be written out
!*****************************************************************************
      LOGICAL FUNCTION RhoShouldBeWritten()
!*****************************************************************************
      USE SF_Module
      IMPLICIT NONE
      integer :: Nbestorder,CurSavedRho, MaxRValPos
      real :: MaxRVal
      Character*(6) :: BRForm='(i...)'
      IF (RepeatMode=='never') THEN
        CurRhoFile=OutRhoFile
        RhoShouldBeWritten=.true.
      ELSEIF (NBestRho==1) THEN
        IF ((BestFOM(CurRhoProps)<BestFOM(BestRhoProps(1))).or.(NSavedrho==0)) THEN
          BestRhoProps(1)=CurRhoProps
	  CurRhoFile=OutRhoFile
          RhoShouldBeWritten=.true.
          write(lgu,'(a,i4,a)') 'Current FOM is the best until now. Saving the density.'
          IF (terminal) write(*,'(a)') 'Current FOM is the best until now. Saving the density.'
          NSavedRho=1
        ELSE  
	  RhoShouldBeWritten=.false.
          write(lgu,'(a,i4,a)') 'Current FOM is worse then the best. Density will not be saved.'
          IF (terminal) write(*,'(a)') 'Current FOM is worse then the best. Density will not be saved.'
	ENDIF
        write(lgu,'(a)') 'Properties of the saved density:'
        write(lgu,'(a)',ADVANCE='NO') '  Run  Rvalue   Peaks  Symm.'
        IF (len_trim(CurRhoProps%DerSGSym)>0) write(lgu,'(a)', ADVANCE='NO') '    Der.SG'
        IF (refrhotype/='none') write(lgu,'(a)', ADVANCE='NO')'  Ref.match      Origin shift'
        write(lgu,*)
        if (terminal) THEN
	  write(*,'(a)') 'Properties of the saved density:'
          write(*,'(a)',ADVANCE='NO') '  Run  Rvalue   Peaks  Symm.'
          IF (len_trim(CurRhoProps%DerSGSym)>0) write(*,'(a)', ADVANCE='NO') '    Der.SG'
	  IF (refrhotype/='none') write(*,'(a)', ADVANCE='NO')'  Ref.match      Origin shift'
          write(*,*)
        ENDIF
	CALL WriteBestRhoLine(BestRhoProps(1),0)
      ELSE
        IF (NSavedRho<NBestRho) THEN
	  NSavedRho=NSavedRho+1
          CurSavedRho=NSavedRho
	  BestRhoProps(NSavedRho)=CurRhoProps	  
	  RhoShouldBeWritten=.true.
	ELSE
	  MaxRVal=-1000.
	  MaxRValPos=0
	  DO cnt1=1,NBestRho !Find the highest RValue
	    IF (BestFOM(BestRhoProps(cnt1))>MaxRVal) THEN
	      MaxRVal=BestFOM(BestRhoProps(cnt1))
	      MaxRValPos=cnt1
	    ENDIF
	  ENDDO  
          IF (BestFOM(CurRhoProps)<MaxRVal) THEN
	    BestRhoProps(MaxRValPos)=CurRhoProps
	    CurSavedRho=MaxRValPos
            RhoShouldBeWritten=.true.
	    write(lgu,'(a,i4,a)') 'Current FOM ranks among the ',NBestRho,' best. Saving the density.'
            IF (terminal) write(*,'(a,i4,a)') 'Current FOM ranks among the',NBestRho,' best. Saving the density.'
	  ELSE
	    RhoShouldBeWritten=.false.
	    write(lgu,'(a,i4,a)') 'Current FOM does not rank among the ',NBestRho,' best. Density will not be saved.'
            IF (terminal) write(*,'(a,i4,a)') 'Current FOM does not rank among the ',NBestRho,&
	                                      ' best. Density will not be saved.'
	  ENDIF
	  !Write record in the log file
	ENDIF
        write(lgu,'(a)') 'Properties of the saved densities:'
        write(lgu,'(a)',ADVANCE='NO') '  Nr. Run  Rvalue   Peaks  Symm.'
        IF (len_trim(CurRhoProps%DerSGSym)>0) write(lgu,'(a)', ADVANCE='NO') '    Der.SG'
        IF (refrhotype/='none') write(lgu,'(a)', ADVANCE='NO')'  Ref.match      Origin shift'
        write(lgu,*)
        if (terminal) THEN
	  write(*,'(a)') 'Properties of the saved densities:'
          write(*,'(a)',ADVANCE='NO') '  Nr. Run  Rvalue   Peaks  Symm.'
          IF (len_trim(CurRhoProps%DerSGSym)>0) write(*,'(a)', ADVANCE='NO') '    Der.SG'
	  IF (refrhotype/='none') write(*,'(a)', ADVANCE='NO')'  Ref.match      Origin shift'
          write(*,*)
        ENDIF
	DO cnt1=1,NSavedRho
          CALL WriteBestRhoLine(BestRhoProps(cnt1),cnt1)
        ENDDO
	write(lgu,'('''')')
	IF (RhoShouldBeWritten) THEN
          NBestOrder=int(log10(float(NBestRho))+eps)+1
          WRITE(BRForm(3:5),'(i1,''.'',i1)') NBestOrder,NBestOrder
          CurRhofile=rhoprefix
	  DO cnt1=1,OutRhoNr
            WRITE(CurRhofile(cnt1)(len_trim(rhoprefix)+1:len_trim(rhoprefix)+NBestOrder),BRForm) CurSavedRho
            CurRhoFile(cnt1)=trim(CurRhoFile(cnt1))//'_'//OutRhoFile(cnt1)
          ENDDO
	ENDIF
      ENDIF
      CONTAINS
      SUBROUTINE WriteBestRhoLine(RhoProps, Index)
        IMPLICIT NONE	
	TYPE(RhoPropsType) :: RhoProps
	INTEGER :: index, loop, loopstart, loopend, loopstep
	IF (terminal) THEN
	  loopstart=6; loopend=lgu; loopstep=lgu-6
	ELSE
	  loopstart=lgu; loopend=lgu; loopstep=1
	ENDIF
	DO loop=loopstart,loopend, loopstep
  	  IF (Index>0) write(loop,'(i4)', ADVANCE='NO') index
          write(loop,'(i4,3f8.2)', ADVANCE='NO') RhoProps%RunNumber,RhoProps%Rvalue,RhoProps%Peaks,RhoProps%AgrFac
	  IF (len_trim(RhoProps%DerSGSym)>0) write(loop,'(a10)', ADVANCE='NO') adjustr(trim(RhoProps%DerSGSym))
	  IF (refrhotype/='none') write(loop,'(f9.2,1X,a)', ADVANCE='NO') RhoProps%refmatch,trim(RhoProps%OrigShift)
	  write(loop,*)
	ENDDO
      END SUBROUTINE WriteBestRhoLine
      
      REAL FUNCTION BestFOM(RhoProps)
        IMPLICIT NONE
	TYPE(RhoPropsType) :: RhoProps
        SELECT CASE (SaveRhoFom)
	CASE (FOMRvalue)
	  BestFOM=RhoProps%Rvalue
	CASE (FOMpeakiness)
	  BestFOM=-RhoProps%peaks
	CASE (FOMReference)
	  BestFOM=RhoProps%Refmatch
	CASE (FOMSymAgr)
	  BestFOM=RhoProps%agrfac
	END SELECT
      END FUNCTION BestFOM
      END FUNCTION RhoShouldBeWritten

! Check, if the CF calculation should be repeated
!*****************************************************************************
      LOGICAL FUNCTION RepeatCF()
!*****************************************************************************
      USE SF_Module
      IMPLICIT NONE
      real :: bfactor !Returns correction for the isotropic D_W. factor
      real :: normalize !Returns a normalized structure factor
      real :: lowconf, upconf
      logical :: WriteStats=.false.
      
      IF (RepeatMode=='never') THEN
        RepeatCF=.false.
      ELSEIF (UserBreak) THEN
        RepeatCF=.false.
        write(lgu,'(a)') 'Calculation interrupted by user.'
	write(lgu,'('''')')
	IF (terminal) THEN
	  write(*,'(a)') 'Calculation interrupted by user.'
          write(*,'('''')')
        ENDIF
	WriteStats=.true.      
      ELSEIF (RepeatMode=='always') THEN
	IF ((Converged.and.(WantSumRho==SumGood)).or.(WantSumRho==SumAll)) THEN
	  sumrho=sumrho+rho
	  NSummed=NSummed+1
          write(lgu,'(a,i4)') 'Current density added to the averaged density. Number of averaged densities:',NSummed
	  write(lgu,'('''')')
	  IF (terminal) THEN
            write(*,'(a,i4)') 'Current density added to the averaged density. Number of averaged densities:',NSummed
            write(*,'('''')')
          ENDIF
	ENDIF
        RepeatCF=.true.
        write(lgu,'(a)') 'RepeatMode always: Repeating the calculation.'
	write(lgu,'('''')')
	IF (terminal) THEN
	  write(*,'(a)') 'RepeatMode always: Repeating the calculation.'
          write(*,'('''')')
        ENDIF
      ELSEIF (RepeatMode=='nosuccess') THEN
	IF (Converged) THEN
	  RepeatCF=.false.
	ELSE
	  RepeatCF=.true.
          write(lgu,'(a)') 'RepeatMode nosuccess and no convergence detected: Repeating the calculation.'
	  write(lgu,'('''')')
          IF (terminal) THEN
	    write(*,'(a)') 'RepeatMode nosuccess and no convergence detected: Repeating the calculation.'
            write(*,'('''')')
          ENDIF
	ENDIF
      ELSEIF (RepeatMode=='number') THEN
	IF ((Converged.and.(WantSumRho==SumGood)).or.(WantSumRho==SumAll)) THEN
          sumrho=sumrho+rho
	  NSummed=NSummed+1
          write(lgu,'(a,i4)') 'Current density added to the averaged density. Number of averaged densities:',NSummed
	  write(lgu,'('''')')
	  IF (terminal) THEN
            write(*,'(a,i4)') 'Current density added to the averaged density. Number of averaged densities:',NSummed
            write(*,'('''')')
          ENDIF
	ENDIF
        IF (repeats<repeatattempts) THEN
	  RepeatCF=.true.
          write(lgu,'(a,i3,a,i3,a)') 'Run number',repeats,'. Still',repeatattempts-repeats,' to go.'
	  write(lgu,'('''')')
          IF (terminal) THEN
            write(*,'(a,i3,a,i3,a)') 'Run number',repeats,'. Still',repeatattempts-repeats,' to go.'
	    write(*,'('''')')
          ENDIF      
        ELSE
	  RepeatCF=.false.
	  WriteStats=.true.
          write(lgu,'(a,i3,a)') 'Last run from',repeatattempts,' completed.' 
	  write(lgu,'('''')')	  
          IF (terminal) THEN
            write(*,'(a,i3,a)') 'Last run from',repeatattempts,' completed.' 
	    write(*,'('''')')
          ENDIF      
	ENDIF
      ENDIF
      IF (WriteStats) THEN
          IF (NConverged>0) call betaconfidence(float(TotalCycles)/float(NConverged), NConverged,0.05,lowconf,upconf)
          write(lgu,'(a)') 'Summary of all runs:'
	  write(lgu,'(a,i7)')     'Number of attempts          :', repeats
	  write(lgu,'(a,i7)')     'Maximum cycles              :', InMaxCycles
	  write(lgu,'(a,i7)')     'Number of successes         :', NConverged
	  write(lgu,'(a,f7.1,a)') 'Success rate [%]            :', 100.*float(NConverged)/float(repeats)
	  if (NConverged>0) THEN
            write(lgu,'(a,i7)')   'Mean cycles per convergence(beta):', nint(float(TotalCycles)/float(NConverged))
            write(lgu,'(a,i7,a,i7)')   '95% confidence interval for beta :', nint(lowconf),' - ',nint(upconf)
            write(lgu,'(a,f5.2,a,f5.2)')   '95% confidence interval for SR   :', &
	                                   1.-exp(-InMaxCycles/upconf),' - ',1.-exp(-InMaxCycles/lowconf) 
	  ELSE
            write(lgu,'(a,i7,a)') 'Average cycles per solution :', InMaxcycles+1, ' or more.'
          ENDIF
          IF (terminal) THEN
            write(*,'(a,i3,a)') 'Last run from',repeats,' completed.' 
	    write(*,'('''')')
	    write(*,'(a)') 'Summary of all runs:'
	    write(*,'(a,i7)')     'Number of attempts               :', repeats
	    write(*,'(a,i7)')     'Maximum cycles                   :', InMaxCycles
	    write(*,'(a,i7)')     'Number of successes              :', NConverged
	    write(*,'(a,f7.1,a)') 'Success rate (SR) [%]            :', 100.*float(NConverged)/float(repeats)
	    if (NConverged>0) THEN
              write(*,'(a,i7)')   'Mean cycles per convergence(beta):', nint(float(TotalCycles)/float(NConverged))
	      write(*,'(a,i7,a,i7)')   '95% confidence interval for beta :', nint(lowconf),' - ',nint(upconf)
              write(*,'(a,f5.2,a,f5.2)')   '95% confidence interval for SR   :', &
	                                   1.-exp(-InMaxCycles/upconf),' - ',1.-exp(-InMaxCycles/lowconf) 
	    ELSE
              write(*,'(a,i7,a)') 'Average cycles per solution :', InMaxcycles+1, ' or more.'
            ENDIF
	    write(*,*)
          ENDIF      
      ENDIF
      IF ((RepeatCF).and.(NGroups>0).and.(hmclasses>0)) THEN !We have to recover the MEMRefl%F values, they have been modified by histogram matching and repartitioning
        DO cnt1=MStartF+1,MEndF
	  IF (MEMRefl(cnt1)%code/=code_missing) THEN
        IF (NormalizeRefl==NormWilson) THEN
          MEMRefl(cnt1)%F=Normalize(Refl(MEMRefl2Refl(cnt1))%hklfg,Refl(MEMRefl2Refl(cnt1))%F,Biso)
        ELSEIF (NormalizeRefl==NormAtoms) THEN
          MEMRefl(cnt1)%F=Normalize(Refl(MEMRefl2Refl(cnt1))%hklfg,Refl(MEMRefl2Refl(cnt1))%F,0.0)
	    ELSE
          MEMRefl(cnt1)%F=Refl(MEMRefl2Refl(cnt1))%F*Bfactor(MEMRefl(cnt1)%hklFG,Biso)
	    ENDIF
	  ELSE
        MEMRefl(cnt1)%F=0.0
	  ENDIF
	ENDDO
      ENDIF
      !We must also recover the symmetry operations, if derivesym use was used
      IF (derivesym=='use') THEN
        NSym=NSymOrig
        NSymPrim=NSymPrimOrig
        NCenter=NCenterOrig
        CALL DeallocateSG(SmOp)
        CALL AllocateSG(SmOp,SDim,NSymPrim*NCenter+NTestSym+1)
        CALL CopySG(SmOpOrig,SmOp)
	DEALLOCATE(Center)
        ALLOCATE(Center(1:SDim,1:NCenter))
	Center=CenterOrig
        CALL FindGenerators
      ENDIF
      !Finally we should restore the setting of weak reflections:
      IF (weakratio>0.0) CALL GetWeakRefl
      END FUNCTION RepeatCF

! Calculates the confidence interval of beta for the sample from exponential distribution with hnown \hat{beta} and known number of attempts
!Uses function chiinv that in turn uses chisqsum, gser, gcf, gammln
!********************************
      SUBROUTINE BetaConfidence(beta,nsamples,alpha,lowerlim,upperlim)
        IMPLICIT NONE
	INTEGER :: nsamples
	REAL :: beta, alpha, lowerlim, upperlim, chiinv
	lowerlim=beta*2.*nsamples/chiinv(2*nsamples,1.-alpha/2.)
	upperlim=beta*2.*nsamples/chiinv(2*nsamples,alpha/2.)
      END SUBROUTINE BetaConfidence
!********************************

! Calculates the value of inverse cummulative distribution function of Chi^2
!********************************
      FUNCTION chiinv(df,alpha)
!********************************
      IMPLICIT NONE
      REAL, PARAMETER :: eps=3.e-7
      INTEGER, PARAMETER :: maxiter=1000
      REAL :: alpha, chiinv, chisqcum
      REAL :: Ulimit, Llimit, test
      INTEGER :: df, niter
      !estimate the limits of the search
      IF (alpha<0.5) THEN
        Ulimit=float(df)
        Llimit=df/10.
      ELSE
        Llimit=float(df)
	Ulimit=df*5.
      ENDIF
      DO
        IF (chisqcum(df,Ulimit)<alpha) THEN
	  Ulimit=Ulimit*2
        ELSE
	  EXIT
	ENDIF
      ENDDO
      DO
        IF (chisqcum(df,Llimit)>alpha) THEN
	  Llimit=Llimit/2.
        ELSE
	  EXIT
	ENDIF
      ENDDO
      niter=0
      DO
        test=chisqcum(df,(Llimit+Ulimit)/2.)
	IF (test>alpha) THEN
	  Ulimit=(Llimit+Ulimit)/2.
	ELSE
	  Llimit=(Llimit+Ulimit)/2.
	ENDIF
	niter=niter+1
	IF ((abs(test-alpha)<eps).or.(niter>maxiter)) EXIT
      ENDDO
      chiinv=(Llimit+Ulimit)/2.
      END FUNCTION chiinv
      
!From numerical recipes in fortran, function gammp, just with shuffled arguments to calculate chiinv
!*************************
      FUNCTION chisqcum(k,inx)
!*************************
      REAL inx,x,chisqcum
      integer k
!     USES gcf,gser
!     Returns the incomplete gamma function P(a, x).
      REAL gammcf,gamser,gln,a
      a=float(k)/2.
      x=inx/2
      if(x.lt.a+1.)then !Use the series representation.
        call gser(gamser,a,x,gln)
        chisqcum=gamser
      else !Use the continued fraction representation
        call gcf(gammcf,a,x,gln)
        chisqcum=1.-gammcf !and take its complement.
      endif
      END

      SUBROUTINE gser(gamser,a,x,gln)
      INTEGER ITMAX
      REAL a,gamser,gln,x,EPS
      PARAMETER (ITMAX=1000,EPS=3.e-7)
      !USES gammln
      !Returns the incomplete gamma function P(a, x) evaluated by its series representation as
      !gamser. Also returns ln?(a) as gln.
      INTEGER n
      REAL ap,del,sum,gammln
      gln=gammln(a)
      if(x.le.0.)then
        if(x.lt.0.) CALL StopProgram('x < 0 in gser')
        gamser=0.
        return
      endif
      ap=a
      sum=1./a
      del=sum
      do n=1,ITMAX
        ap=ap+1.
        del=del*x/ap
        sum=sum+del
        if(abs(del).lt.abs(sum)*EPS)goto 1
      enddo
      CALL StopProgram('a too large, ITMAX too small in gser')
1     gamser=sum*exp(-x+a*log(x)-gln)
      return
      END

      SUBROUTINE gcf(gammcf,a,x,gln)
      INTEGER ITMAX
      REAL a,gammcf,gln,x,EPS,FPMIN
      PARAMETER (ITMAX=1000,EPS=3.e-7,FPMIN=1.e-30)
      ! USES gammln
      !Returns the incomplete gamma function Q(a, x) evaluated by its continued fraction representation
      !as gammcf. Also returns ln ?(a) as gln.
      !Parameters: ITMAX is the maximum allowed number of iterations; EPS is the relative accuracy;
      !FPMIN is a number near the smallest representable floating-point number.
      INTEGER i
      REAL an,b,c,d,del,h,gammln
      gln=gammln(a)
      b=x+1.-a !Set up for evaluating continued fraction by modified
      c=1./FPMIN !Lentz's method (5.2) with b0 = 0.
      d=1./b
      h=d
      do i=1,ITMAX !Iterate to convergence.
        an=-i*(i-a)
        b=b+2.
        d=an*d+b
        if(abs(d).lt.FPMIN)d=FPMIN
        c=b+an/c
        if(abs(c).lt.FPMIN)c=FPMIN
        d=1./d
        del=d*c
        h=h*del
        if(abs(del-1.).lt.EPS)goto 1
      enddo
      CALL StopProgram('a too large, ITMAX too small in gcf')
1     gammcf=exp(-x+a*log(x)-gln)*h !Put factors in front.
      return
      END

      FUNCTION gammln(xx)
      REAL gammln,xx
      !Returns the value ln[gamma(xx)] for xx > 0.
      INTEGER j
      DOUBLE PRECISION ser,stp,tmp,x,y,cof(6)
      !Internal arithmetic will be done in double precision, a nicety that you can omit if five-figure
      !accuracy is good enough.
      SAVE cof,stp
      DATA cof,stp/76.18009172947146d0,-86.50532032941677d0, &
      24.01409824083091d0,-1.231739572450155d0,.1208650973866179d-2, &
      -.5395239384953d-5,2.5066282746310005d0/
      x=xx
      y=x
      tmp=x+5.5d0
      tmp=(x+0.5d0)*log(tmp)-tmp
      ser=1.000000000190015d0
      do j=1,6
        y=y+1.d0
        ser=ser+cof(j)/y
      enddo
      gammln=tmp+log(stp*ser/x)
      return
      END

! Find a least common multiple 
! A beautifully simple algorithm found at http://www.cut-the-knot.org/Curriculum/Arithmetic/LCM.shtml
! original reference A. J. Goldman; N. J. Fine, El655, The American Mathematical Monthly, Vol. 71, No. 9. (Nov., 1964), pp. 1045-1046.
!*****************************************
      INTEGER FUNCTION LCM(inarr)
!*****************************************
        IMPLICIT NONE
	INTEGER, DIMENSION(:) :: inarr
	INTEGER, DIMENSION(1:size(inarr)) :: cumarr
        INTEGER :: minpos

        cumarr=inarr
100     lcm=minval(cumarr)
        IF (lcm==maxval(cumarr)) RETURN
	minpos=minloc(cumarr, dim=1)
	cumarr(minpos)=cumarr(minpos)+inarr(minpos)
	goto 100
	
      END FUNCTION LCM


! Generate histogram from a structure
!*****************************************
      SUBROUTINE GenerateHistogram
!*****************************************
      USE SF_Module
      USE Input_Module
      USE SF_Composition
      USE FFT_Interfaces
      
      IMPLICIT NONE
!      character*(filelength) :: rhofiletmp
!      character*(linelength) :: rhotypetmp
      
      IF (HMClasses==-10) THEN
        CALL RandomStructure(.false.)
	CALL SFFromStructure
	CALL FFT(sf2rho)
!        rhofiletmp='test.xplor'
!	rhotypetmp='xplor'
!	write(*,*) 'Writing density'
!	CALL WriteRho(rhofiletmp,rhotypetmp)
	CALL MakeHistogramFromRho
!       write histogram to the log file
        IF (expandedlog) THEN
          CALL FramedLabel(2,'Histogram generated from composition:')
          write(lgu,'(a)') ' Class nr   class position    histogram value'
	  DO cnt1=0,HMClasses
	    write(lgu,'(i6,2f17.5)') cnt1, HMGramm(cnt1)%pos, HMGramm(cnt1)%orig/cApVx
	  ENDDO
        ENDIF
      ELSEIF (HMClasses==-20) THEN
	CALL ReadCif(ModelStrFile)
	StrSymmetry=.true. !Means the structure contains only symmetry-independent part
	CALL SFFromStructure
	CALL FFT(sf2rho)
!        rhofiletmp='test.xplor'
!	rhotypetmp='xplor'
!	write(*,*) 'Writing density'
!	CALL WriteRho(rhofiletmp,rhotypetmp)
	CALL MakeHistogramFromRho
        IF (expandedlog) THEN
          CALL FramedLabel(2,'Histogram generated from used-supplied structure')
          write(lgu,'(a)') ' Structure file: '//trim(ModelStrFile)
          write(lgu,'(a)') ' Class nr   class position    histogram value'
	  DO cnt1=0,HMClasses
	    write(lgu,'(i6,2f17.5)') cnt1, HMGramm(cnt1)%pos, HMGramm(cnt1)%orig/cApVx
	  ENDDO
        ENDIF
      ELSE
        CALL StopProgram('Unforeseen error: HMClasses has unknown value. Please inform the author.')
      ENDIF
      IF (Totatoms>0) DEALLOCATE(Atoms)
      END SUBROUTINE GenerateHistogram

! Generate random structure
!*****************************************
      SUBROUTINE RandomStructure(SymFlag)
!*****************************************
      USE SF_Module
      USE SF_Composition
      
      IMPLICIT NONE
      REAL, PARAMETER :: GoodDist=1.5 !Mnimum distance between two atoms in the random structure
      LOGICAL :: SymFlag
      INTEGER :: i,ii,iii,iv,symloop,curpos,nrpos,GetSeedFromTime, curseed, Multip
      REAL :: MY_RAND,MinDist,diffvec(SDim),AtDist,multsc, curocc
      REAL, DIMENSION(SDim) :: RandVector
      IF (BisoHM==-100.) BisoHM=BisoWP !use Biso from the wilson plot

      IF (SymFlag) THEN
        Multip=NSym
      ELSE
        Multip=1
      ENDIF
      IF (ALLOCATED(Atoms)) DEALLOCATE(Atoms)
      TotAtoms=0
      DO i=1,ElNr
        TotAtoms=TotAtoms+(int((float(Compos(i,2))-eps)/multip)+1)*multip
      ENDDO
      ALLOCATE(Atoms(1:TotAtoms))
      !Check the volume per atom
      IF ((Cell(NCellPar+1)/Totatoms)/(4.*pi*(Gooddist/2.)**3/3.)<3.) THEN
        write(Buffer,'(a,f8.3,a)') 'Warning, the volume per atom according to the composition is ',Cell(NCellPar+1)/Totatoms,'.'
        call ErrorMsg(trim(Buffer))
        write(buffer,'(a,f6.3,a)') 'This is too small, I won''t be able to make a random structure with minimal'// &
	     ' atom-atom distance ',GoodDist,'.'
        call ErrorMsg(trim(Buffer))
	CALL StopProgram('Please check your composition and/or cell dimensions. Stopping.')
      ENDIF
      curpos=0
      IF (randstructureseed==0) THEN
        curseed=GetSeedFromTime()
      ELSE
        curseed=randstructureseed
      ENDIF
      write(lgu,'(a,i12)') 'Random seed for random structure:',curseed
      CALL MY_SRAND(curseed)
      DO i=1,ElNr
!        IF (mod(Compos(i,2),Multip)/=0) & 
!	  CALL StopPRogram('Error, the composition is not an integer multiple of the SG multiplicity!')
        nrpos=Compos(i,2)/Multip+1
        DO ii=1,nrpos          
          if (ii<nrpos) THEN
            curocc=1.
          ELSEIF (mod(Compos(i,2),Multip)/=0) THEN
            curocc=float(mod(Compos(i,2),Multip))/multip            
          ELSE
            EXIT !Thus is the case when compos(i,2)/multip is an integer, and thus no partially occupied atoms are needed
          ENDIF
          DO
	    cnt1=curpos
	    DO iii=1,3
	      RandVector(iii)=MY_RAND()
	    ENDDO
            DO symloop=1,Multip
	      cnt1=cnt1+1
              Atoms(cnt1)%U=BisoHM/(8.0*pi**2)
              Atoms(cnt1)%xyz=MatMul(SmOp(symloop)%R,RandVector)+SmOp(symloop)%t
              Atoms(cnt1)%occ=curocc
	      Atoms(cnt1)%nr=i
	      Atoms(cnt1)%Z=compos(i,1)
	    ENDDO
	    !Check the distance of the last added atom to all other atoms
	    IF (cnt1>1) THEN
	      MinDist=GoodDist+1.
	      DO iv=1,cnt1-1
	        DO cnt2=1,3
		  diffvec(cnt2)=Atoms(cnt1)%xyz(cnt2)-Atoms(iv)%xyz(cnt2)
		  if (abs(diffvec(cnt2)+1.)<abs(diffvec(cnt2))) THEN
		    diffvec(cnt2)=diffvec(cnt2)+1.
		  elseif (abs(diffvec(cnt2)-1.)<abs(diffvec(cnt2))) THEN
		    diffvec(cnt2)=diffvec(cnt2)-1.
	 	  endif
		ENDDO
                AtDist=sqrt(multsc(diffvec,diffvec,3,.false.))
                MinDist=min(MinDist,AtDist)
		IF (MinDist<GoodDist) EXIT
	      ENDDO
 	      IF (MinDist>GoodDist) EXIT
	    ELSE
	      EXIT
	    ENDIF
	  ENDDO
          curpos=curpos+Multip
	ENDDO
      ENDDO      
      StrSymmetry=SymFlag
      END SUBROUTINE RandomStructure

! Fill the structure factors to the array SF
!*******************************************
      SUBROUTINE MakeHistogramFromRho
!*******************************************
      USE SF_Module
      IMPLICIT NONE
      integer, parameter :: MaxClasses=200, minvox=100
      real, parameter :: minclassstep=0.02
      integer :: i,ii,currhopos,MainLoop
      real :: rhomax,rhomin,rnth,currho
      logical :: assigned
            
      CALL Heap(NVoxTot,rho)      
      rhomax=rho(NVoxTot-1)
      rhomin=rho(0)
      rnth=(rhomax-rhomin)/float(MaxClasses)
      i=0;HMClasses=0
      DO MainLoop=1,2
        IF (MainLoop==2) THEN
          ALLOCATE(HMGramm(0:HMClasses))
          hmgramm(HMClasses)%orig=rho(NVoxTot-1)
          hmgramm(HMClasses)%pos=1.
          i=HMClasses
        ENDIF
        currhopos=NVoxTot-1
        currho=rho(NVoxTot-1)
        DO
          i=i-1
	  Assigned=.false.
	  DO ii=currhopos-minvox,0,-1
!	    write(*,*) currhopos,ii,i,currho,rho(ii),abs(currho-rho(ii)),rnth
	    if ((abs(currho-rho(ii))>rnth).or.&
	        ((float(currhopos)-float(ii))>((NVoxTot-1)*minclassstep))) THEN
	      IF (MainLoop==1) THEN
	        HMClasses=HMClasses+1
	      ELSE
                hmgramm(i)%orig=rho(ii)
	        hmgramm(i)%pos=float(ii)/(NVoxTot-1)
	      ENDIF
	      currho=rho(ii)
	      currhopos=ii
	      Assigned=.true.
	      EXIT
	    ENDIF
	  ENDDO
          if (((currhopos-minvox)<0).or.(.not.assigned)) EXIT
        ENDDO
      ENDDO
      hmgramm(0)%orig=rho(0)
      hmgramm(0)%pos=0.
      END SUBROUTINE MakeHistogramFromRho

           
! Fill the current histogram
!*****************************************
      SUBROUTINE GetCurrentHistogram
!*****************************************
      USE SF_Module
      REAL :: select !Function selecting an nth element of an array
      INTEGER :: i

      hmgramm(0)%cur=rho(0)
      hmgramm(hmclasses)%cur=rho(0)
      DO i=0,NVoxTot-1
        SF(i)=rho(i) ! make a backup of rho in SF
        hmgramm(0)%cur=min(hmgramm(0)%cur,rho(i))
        hmgramm(hmclasses)%cur=max(hmgramm(hmclasses)%cur,rho(i))
      ENDDO
      DO i=1,hmclasses-1
        hmgramm(i)%cur=select(nint(hmgramm(i)%pos*float(NVoxTot-1))+1,NVoxTot,rho)        
      ENDDO
      !return rho back to its original
      DO i=0,NVoxTot-1
        rho(i)=SF(i)
      ENDDO
      END SUBROUTINE GetCurrentHistogram



      
! Fill the structure factors to the array SF
!*****************************************
      SUBROUTINE SFFromStructure
!*****************************************
      USE SF_Module
      USE SF_Composition
      
      IMPLICIT NONE
      INTEGER :: i, ii
      COMPLEX :: StrFac,StructureFactor
      logical :: satellite
      SF=0.0
      satellite=.false.
      DO i=MStartF,MEndF
        IF (SDim>RDim) THEN	   
          IF (sum(abs(MEMRefl(i)%hklfg(RDim+1:SDim)))>0) THEN
            satellite=.true.
          ELSE
            satellite=.false.
          ENDIF
	ENDIF
        IF (.not.satellite) THEN
	  StrFac=StructureFactor(MEMRefl(i)%hklFG)
	ELSE
	  StrFac=0.0
	ENDIF
!	write(*,*) strfac
!	read(*,*)
	SF(MEMRefl(i)%SFLoc)=REAL(StrFac)
	SF(MEMRefl(i)%SFloc+1)=-AIMAG(StrFac) !Minus, because of the oposite convention of FFTW3
!AIMAG(z) -- imaginary part of a complex number.
!When supplied with a COMPLEX argument, this function returns the imaginary part, for example, AIMAG(0.9,0.2) is 0.2.
!REAL or DBLE -- real part of a complex number.
!CONJG(z) -- conjugate of a complex number.
!ABS(z) -- absolute value of a complex number.
      ENDDO
      END SUBROUTINE SFFromStructure

      COMPLEX FUNCTION StructureFactor(hkl)
      USE SF_Module
      USE SF_Composition
      IMPLICIT NONE
      INTEGER, DIMENSION (1:6,1:2) :: six2nine = RESHAPE ( (/1,2,3,1,1,2,1,2,3,2,3,3 /),(/6,2/))
      INTEGER :: i,ii,iii,ic,jc,symcnt,symlooplimit
      real :: scprod,sthl,FormF,multsc,thermfac,curxyz(3),curU(6), curR(1:3,1:3), curt(1:3)
      INTEGER, DIMENSION(SDim) :: hkl      
      REAL, DIMENSION(SDim) :: dhkl
      REAL, DIMENSION(1:3,1:3) :: RmatTr,Umat,TmpMat
      
      dhkl=float(hkl)
      sthl=multsc(dhkl,dhkl,SDim,.true.)
      sthl=sqrt(sthl)/2.
      StructureFactor=0.
      if (StrSymmetry) THEN
        SymLoopLimit=NSym
      ELSE
        SymLoopLimit=1
      ENDIF
      DO i=1,TotAtoms
        DO symcnt=1,SymLoopLimit
	  CurR=SmOp(symcnt)%R(1:3,1:3)
	  Curt=SmOp(symcnt)%t(1:3)
          curxyz=multmat(CurR,Atoms(i)%xyz)
	  DO ii=1,3
	    curxyz(ii)=curxyz(ii)+curt(ii)
	  ENDDO
	  !Transform U according to RMat U(trans)=R.U.R^T
	  Umat(1,1)=Atoms(i)%U(1);Umat(2,2)=Atoms(i)%U(2);Umat(3,3)=Atoms(i)%U(3)
	  Umat(1,2)=Atoms(i)%U(4);Umat(2,1)=Atoms(i)%U(4)
	  Umat(1,3)=Atoms(i)%U(5);Umat(3,1)=Atoms(i)%U(5)
	  Umat(2,3)=Atoms(i)%U(6);Umat(3,2)=Atoms(i)%U(6)
          DO ii=1,3
	    DO iii=1,3
	      RmatTr(ii,iii)=CurR(iii,ii)
	    ENDDO
	  ENDDO
          TmpMat=multmat(CurR,Umat)
	  Umat=multmat(TmpMat,RMatTr)
	  curU(1)=Umat(1,1);curU(2)=Umat(2,2);curU(3)=Umat(3,3)
          curU(4)=Umat(1,2);curU(5)=Umat(1,3);curU(6)=Umat(2,3)
	  scprod=0.
	  DO ii=1,3
	    scprod=scprod+hkl(ii)*curxyz(ii)
	  ENDDO
          thermfac=0.0
	  DO ii=1,6
	    ic=six2nine(ii,1)
	    jc=six2nine(ii,2)
	    thermfac=thermfac+curU(ii)*RCell(ic)*RCell(jc)*hkl(ic)*hkl(jc)  
	    if (ii>3) thermfac=thermfac+curU(ii)*RCell(ic)*RCell(jc)*hkl(ic)*hkl(jc)
	  ENDDO
          thermfac=exp(-2*pi**2*thermfac)
	  StructureFactor=StructureFactor+Atoms(i)%occ*FormF(Atoms(i)%Z,sthl)*thermfac*exp(CMPLX(0.,2*pi*scprod))
	ENDDO
      ENDDO
      END FUNCTION StructureFactor


!Read a structure from a CIF file
!*****************************************
      SUBROUTINE ReadCIF(ciffile)
!*****************************************
      USE SF_Module
      USE SF_Composition
      USE Input_Module

      INTEGER, PARAMETER :: CifNamesNr=17
      CHARACTER*(filelength) :: ciffile,Buffer2
      LOGICAL :: OnAtoms
      INTEGER, DIMENSION(1:20) :: NamePos
      INTEGER :: i,ii
      real :: dummy
      CHARACTER*100, dimension(CifNamesNr) :: cifnames = (/ &
            '_atom_site_label                ', &
            '_atom_site_type_symbol          ', &
            '_atom_site_fract_x              ', &
            '_atom_site_fract_y              ', &
            '_atom_site_fract_z              ', &
            '_atom_site_adp_type             ', &
            '_atom_site_U_iso_or_equiv       ', &
            '_atom_site_symmetry_multiplicity', &
            '_atom_site_occupancy            ', &
            '_atom_site_aniso_label          ', &
            '_atom_site_aniso_type_symbol    ', &
            '_atom_site_aniso_U_11           ', &
            '_atom_site_aniso_U_22           ', &
            '_atom_site_aniso_U_33           ', &
            '_atom_site_aniso_U_12           ', &
            '_atom_site_aniso_U_13           ', &
            '_atom_site_aniso_U_23           '/)
         
      integer :: staterr
      
      
      TotAtoms=0
      open(50,file=ciffile,status='OLD',IOSTAT=staterr)
      IF (StatErr/=0) CALL StopProgram('Error opening cif file '//trim(ciffile)//'! Check the file please.')
100   READ(50,'(a)',END=200) Buffer      
      Buffer2=FirstWord(Buffer,.false.)
      IF (Buffer2=='loop_') THEN
        OnAtoms=.false.
        !Skip the data items and test for the atom-specific one
110	READ(50,'(a)',END=200) Buffer
        Buffer2=FirstWord(Buffer,.false.)
        IF (Buffer2==cifnames(1)) OnAtoms=.true.
        IF (Buffer2(1:1)=='_') GOTO 110
	IF ((OnAtoms).and.(len_trim(Buffer)/=0).and.(Buffer(1:1)/='#')) THEN
	  Totatoms=1
120       READ(50,'(a)',END=200) Buffer
          Buffer2=FirstWord(Buffer,.false.)
          IF ((len_trim(Buffer)/=0).and.(Buffer2(1:1)/='_').and.(Buffer(1:1)/='#')) THEN
	    TotAtoms=TotAtoms+1
            GOTO 120
	  ENDIF
	  goto 200
	ENDIF
      ENDIF      
      goto 100
200   continue
      IF (TotAtoms==0) CALL StopProgram('Error, no atoms found in the CIF file! Stopping.')
      REWIND(50)
      ALLOCATE(Atoms(1:TotAtoms))
      DO i=1,TotAtoms
        Atoms(i)%U=0.0
	Atoms(i)%occ=1.0
	Atoms(i)%nr=0
	Atoms(i)%Z=0
      ENDDO
      !Now read in the atoms
300   READ(50,'(a)',END=400) Buffer      
301   Buffer2=FirstWord(Buffer,.false.)
      cnt1=0;NamePos=0
      IF (Buffer2=='loop_') THEN        
        OnAtoms=.false.
        !Find the position of the interesting flags
310	READ(50,'(a)',END=400) Buffer
        cnt1=cnt1+1
        Buffer2=FirstWord(Buffer,.false.)
	DO i=1,CifNamesNr
          IF (Buffer2==cifnames(i)) THEN
            Namepos(cnt1)=i
          ENDIF
        ENDDO
        IF (Buffer2(1:1)=='_') GOTO 310
	IF (len_trim(Buffer)/=0) THEN
	  cnt2=0
320       continue
          IF ((len_trim(Buffer)/=0).and.(FirstWord(Buffer,.false.)/='_')) THEN
	    cnt2=cnt2+1
            cnt1=0
	    DO
	      cnt1=cnt1+1
	      Buffer2=FirstWord(Buffer,.true.)
	      IF (Buffer2=='?') CYCLE
	      IF (len_trim(Buffer2)==0) EXIT
	      SELECT CASE (Namepos(cnt1))
	        CASE (1)!Name
!		  write(*,*) 'Name:',Buffer2
	        CASE (2)!Type
                  DO i=1,ElNr
                    IF (Buffer2==ElSmb(compos(i,1))) Atoms(cnt2)%nr=i
	          END DO		  
                  DO i=1,TotElNr
                    IF (Buffer2==ElSmb(i)) Atoms(cnt2)%Z=i
	          END DO		  
                  IF (Atoms(cnt2)%Z==0) THEN !Unknown atomic type
                    Buffer='Error reading the CIF file, '//trim(ciffile)
	            write(*,*)
		    CALL ErrorMsg(Buffer)
                    Buffer=trim(Buffer2)//' is an unknown atomic type. Check the file please! Stopping.'
	            CALL StopProgram(Buffer)
                  ENDIF
	        CASE (3)!x
!		  write(*,*) 'x:',Buffer2
                  Atoms(cnt2)%xyz(1)=StripBrackets(Buffer2)
	        CASE (4)!y
!		  write(*,*) 'y:',Buffer2
                  Atoms(cnt2)%xyz(2)=StripBrackets(Buffer2)
	        CASE (5)!z
!		  write(*,*) 'z:',Buffer2
                  Atoms(cnt2)%xyz(3)=StripBrackets(Buffer2)
	        CASE (6)!adptype
!		  write(*,*) 'adptype:',Buffer2
	        CASE (7)!Uiso
!		  write(*,*) 'Uiso:',Buffer2
                  IF (Atoms(cnt2)%U(1)==0.0) Atoms(cnt2)%U(1)=StripBrackets(Buffer2) !Only if aniso hasn't yet been assigned
	        CASE (8)!symmult
!		  write(*,*) 'symmult:',Buffer2
		  read(Buffer2,*) ii
                  Atoms(cnt2)%occ=Atoms(cnt2)%occ*(float(ii)/float(NSym))
	        CASE (9)!occupancy
		  dummy=StripBrackets(Buffer2)
                  Atoms(cnt2)%occ=Atoms(cnt2)%occ*dummy
	        CASE (10)!NameUani
!		  write(*,*) 'NameUani:',Buffer2
	        CASE (11)!TypeUani
!		  write(*,*) 'TypeUani:',Buffer2
	        CASE (12)!U11
!		  write(*,*) 'U11',Buffer2
                  Atoms(cnt2)%U(1)=StripBrackets(Buffer2)
	        CASE (13)!U22
!		  write(*,*) 'U22:',Buffer2
                  Atoms(cnt2)%U(2)=StripBrackets(Buffer2)
	        CASE (14)!U33
!		  write(*,*) 'U33:',Buffer2
                  Atoms(cnt2)%U(3)=StripBrackets(Buffer2)
	        CASE (15)!U12
!		  write(*,*) 'U12:',Buffer2
                  Atoms(cnt2)%U(4)=StripBrackets(Buffer2)
	        CASE (16)!U13
!		  write(*,*) 'U13:',Buffer2
                  Atoms(cnt2)%U(5)=StripBrackets(Buffer2)
	        CASE (17)!U23
!		  write(*,*) 'U23:',Buffer2
                  Atoms(cnt2)%U(6)=StripBrackets(Buffer2)
	      END SELECT
	    ENDDO
	    READ(50,'(a)',END=400) Buffer !Read next atom
            IF (Buffer/='loop_') GOTO 320
	  ENDIF
	  goto 301
	ENDIF
      ENDIF      
      goto 300
400   continue
      DO i=1,TotAtoms
        IF ((Atoms(i)%U(2)==0.0).and.(Atoms(i)%U(3)==0.0)) THEN
	  Atoms(i)%U(2)=Atoms(i)%U(1)
	  Atoms(i)%U(3)=Atoms(i)%U(1)
	ENDIF
      ENDDO
      DO i=1,TotAtoms
!        write(*,*) i,Atoms(i)%nr,Atoms(i)%xyz,Atoms(i)%occ,Atoms(i)%U
      ENDDO
      CLOSE(50)
      
      CONTAINS
        REAL FUNCTION StripBrackets(Number)
	CHARACTER*(filelength) Number
        INTEGER :: loop
	DO loop=1,len_trim(Number)
	  IF (Number(loop:loop)=='(') THEN
	    Number=Number(1:loop-1)
	    EXIT
	  ENDIF
	ENDDO
	READ(Number,*,ERR=900) StripBrackets
	RETURN
900 STOP 'Syntax error while reading the information about atoms! Please check the CIF file.'
	END FUNCTION StripBrackets
      END SUBROUTINE ReadCIF
      

!*****************************************
      SUBROUTINE GenerateRandomStructure
!*****************************************
        USE SF_Module
        USE SF_Composition
        USE SF_Interfaces
        USE FFT_Interfaces
        IMPLICIT NONE
        integer :: i
        complex :: StrFac, StructureFactor

        CALL RandomStructure(.true.)
	call FramedLabel(2,'Atomic coordinates of the generated random structure:')
	DO i=1,TotAtoms
	  write(lgu,'(a,3f12.6)') ElSmb(Compos((Atoms(i)%nr),1)),Atoms(i)%xyz
	ENDDO
	write(lgu,*)
        StrSymmetry=.false.
        DO i=StartF,EndF
          StrFac=StructureFactor(Refl(i)%hklFG)
	  Refl(i)%F=abs(StrFac)
          IF (Refl(i)%SFLoc>0) THEN
	    SF(Refl(i)%SFloc)=REAL(StrFac)
	    SF(Refl(i)%SFloc+1)=-AIMAG(StrFac)
	  ELSE
	    SF(Refl(i)%SFloc)=REAL(StrFac)
	    SF(Refl(i)%SFloc+1)=AIMAG(StrFac)
	  ENDIF
        ENDDO
!	DO i=1,MTotalF
!	  MEMRefl(i)%F=MEMRefl(i)%FMEM
!	ENDDO
	CALL FFT(sf2rho)
	if (allocated(refrho)) deallocate(refrho)
	allocate(refrho(0:NVoxTot-1))
	if (.not.allocated(NVoxRef)) ALLOCATE(NVoxRef(1:SDim))
	NVoxRef=NVox
	refrho=rho
        refrhotype='internal'
      END SUBROUTINE GenerateRandomStructure

!*****************************************
      SUBROUTINE GenerateRandomModel
!*****************************************
        USE SF_Module
        USE SF_Composition
        USE SF_Interfaces
        USE FFT_Interfaces
        IMPLICIT NONE
        integer :: i
        complex :: StrFac, StructureFactor

        CALL RandomStructure(.false.)
!	call FramedLabel(2,'Atomic coordinates of the generated random structure:')
!	DO i=1,TotAtoms
!	  write(lgu,'(a,3f12.6)') ElSmb(Compos((Atoms(i)%nr),1)),Atoms(i)%xyz
!	ENDDO
!	write(lgu,*)
        StrSymmetry=.false.
	SF=0.0
        DO i=MStartF+1,MEndF
          StrFac=StructureFactor(MEMRefl(i)%hklFG)
          SF(MEMRefl(i)%SFloc)=REAL(StrFac)
          SF(MEMRefl(i)%SFloc+1)=-AIMAG(StrFac)
        ENDDO
	CALL FFT(sf2rho)
	modelrho=rho
      END SUBROUTINE GenerateRandomModel


!*****************************************
      SUBROUTINE SetupRhoAndFFT
!*****************************************
      USE SF_Module
      USE FFT_Interfaces
      IMPLICIT NONE
      integer :: ErrCode2, i
      ! Allocate the density
      ALLOCATE(rho(0:NVoxtot-1),STAT=ErrCode2)
      CALL MakeFFTPlan(NVox,rho,SF,r2cft,fastfft,rho2sf)
      CALL MakeFFTPlan(NVox,rho,SF,c2rft,fastfft,sf2rho)

      IF (gdmpar%needrhoold) ALLOCATE(rhoold(0:NVoxtot-1),STAT=ErrCode2)
      IF (gdmpar%needrho1) ALLOCATE(rho1(0:NVoxtot-1),STAT=ErrCode2)
      IF (ErrCode2/=0) CALL StopProgram('Error allocating arrays rho, rhoold or rho1! Too little memory?')
      IF (WantSumRho/=0) THEN
        ALLOCATE(sumrho(0:product(NVoxFine)-1),STAT=ErrCode2)
        IF (ErrCode2/=0) CALL StopProgram('Error allocating array sumrho! Too little memory?')
	sumrho=0.0E0
      ENDIF
      END SUBROUTINE SetupRhoAndFFT


!*****************************************
      SUBROUTINE ReadAndFillOtherDensities
!*****************************************
      USE SF_Module
      USE SF_Composition
      USE FFT_Interfaces
      USE SF_Interfaces
      USE SF_RhoIOInterfaces
      IMPLICIT NONE
      integer :: ErrCode2, i
      complex :: StrFac, StructureFactor

      IF (terminal.and.(RefRhoType/='none'.or.ModelRhoType/='none'.or.EnvRhoType/='none')) write(*,'('''')')

      IF ((RefRhoType/='none').and.(RefRhoType/='internal')) THEN
        ALLOCATE(NVoxRef(1:SDim))
        CALL NVoxFromFile(refrhofile, refrhotype,NVoxRef)
        ALLOCATE(refrho(0:product(NVoxRef)-1),STAT=ErrCode2)
        IF (ErrCode2/=0) CALL StopProgram('Error allocating array refrho! Too little memory?')
        IF (Terminal) write(*,'(a)') 'Reading the reference density from file '//trim(RefRhoFile)//'...'
        CALL ReadRho(refrho,refrhofile,refrhotype)
        IF (Terminal) write(*,'(a)') '...done.'
!        IF (any(TmpNVox/=NVox)) CALL ResampleDensity(TmpNVox,NVox,refrho)
      ENDIF
      IF ((ModelRhoType=='patterson').or.(ModelRhoType=='supminfunction').or.(ModelRhoType=='random')) THEN
        ALLOCATE(NVoxModel(1:SDim))
        NVoxModel=NVox
        ALLOCATE(modelrho(0:NVoxTot-1),STAT=ErrCode2)
	SELECT CASE (ModelRhoType)
	  CASE('patterson')
            CALL SupPatterson(pattersonmodel)
          CASE('supminfunction')
            CALL SupPatterson(supminfunction)
          CASE('random')
	    ! Do nothing here: a new random model is generated for every repeated run -> generated at the beginning of the subroutine chargeflipping
!            CALL GenerateRandomModel
	END SELECT
      ELSEIF (ModelRhoType/='none') THEN
        ALLOCATE(NVoxModel(1:SDim))
        CALL NVoxFromFile(modelrhofile, modelrhotype,NVoxModel)        
        ALLOCATE(modelrho(0:product(NVoxModel)-1),STAT=ErrCode2)
        IF (ErrCode2/=0) CALL StopProgram('Error allocating array modelrho! Too little memory?')
        IF (Terminal) write(*,'(a)') 'Reading the model density from file '//trim(ModelRhoFile)//'...'
        CALL ReadRho(modelrho,modelrhofile,modelrhotype)
        IF (Terminal) write(*,'(a)') '...done.'
      ENDIF
      IF (EnvRhoType/='none') THEN
        ALLOCATE(NVoxEnv(1:SDim))
        CALL NVoxFromFile(envrhofile, envrhotype,NVoxEnv)        
        ALLOCATE(envrho(0:product(NVoxEnv)-1),STAT=ErrCode2)
        IF (ErrCode2/=0) CALL StopProgram('Error allocating array envrho! Too little memory?')
        IF (Terminal) write(*,'(a)') 'Reading the envelope from file '//trim(EnvRhoFile)//'...'
        CALL ReadRho(envrho,envrhofile,envrhotype)
        IF (Terminal) write(*,'(a)') '...done.'
!        IF (any(TmpNVox/=NVox)) CALL ResampleDensity(TmpNVox,NVox,envrho)
      ENDIF
      END SUBROUTINE ReadAndFillOtherDensities
                  
! Sort the array using heap sort
!***************************
      subroutine heap(n,ra)
!***************************
      dimension ra(n)
      if(n.le.1) go to 9999
      l=n/2+1
      ir=n
10    continue
      if(l.gt.1) then
        l=l-1
        rra=ra(l)
      else
        rra=ra(ir)
        ra(ir)=ra(1)
        ir=ir-1
        if(ir.eq.1) then
          ra(1)=rra
          go to 9999
        endif
      endif
      i=l
      j=l+l
20    if(j.le.ir) then
        if(j.lt.ir) then
          if(ra(j).lt.ra(j+1)) j=j+1
        endif
        if(rra.lt.ra(j)) then
          ra(i)=ra(j)
          i=j
          j=j+j
        else
          j=ir+1
        endif
        go to 20
      endif
      ra(i)=rra
      go to 10
9999  return
      end

      REAL FUNCTION GaussDev()
! **************************************************
       IMPLICIT NONE
       REAL :: fac,rsq,v1,v2, MY_RAND
1      v1=MY_RAND()
       v1=2.*v1-1.
       v2=MY_RAND()
       v2=2.*v2-1.
       rsq=v1**2+v2**2
       IF (rsq>=1..or.rsq==0.) GOTO 1
       fac=SQRT(-2*log(rsq)/rsq)
       GaussDev=v1*fac
       END FUNCTION GaussDev

! Bring a number to interval 0-1
!*****************************************
      REAL FUNCTION From0to1_old(Num,eps)
!*****************************************
      REAL :: num,eps
        From0to1_old=Num
	DO
	  if (From0to1_old>-eps) EXIT
	  From0to1_old=From0to1_old+1.0E0
	ENDDO
        DO
	  if (From0to1_old<1.0E0-eps) EXIT
	  From0to1_old=From0to1_old-1.0E0
	ENDDO
	
      END FUNCTION From0to1_old


! Delete all spaces from a string
!*****************************************
      SUBROUTINE RemoveSpaces(ioText)
!*****************************************
      implicit none
      character*(*) :: ioText
      integer :: ii
        ii=0
        DO
          ii=ii+1
	  IF (ioText(ii:ii)==' ') THEN
	    ioText(ii:)=ioText(ii+1:)
	    ii=ii-1
          ENDIF
	  IF (ii>=len_trim(ioText)) EXIT
        ENDDO
      END SUBROUTINE RemoveSpaces

! A routine that finds a representaton of a number by a fraction
!*********************************************
      SUBROUTINE DecToFraction(InDec,Nom,Den)
!*********************************************
        REAL :: InDec,ModDec
	REAL, PARAMETER :: Toler=0.001
	INTEGER :: Nom,Den,LNom,LDen,RNom,RDen,IntDec
	
	ModDec=abs(InDec)-int(abs(InDec))
	IntDec=int(abs(indec))
	LNom=0
	LDen=1
	RNom=1
	RDen=1	
	DO
	  IF (abs(float(LNom)/float(LDen)-abs(ModDec))<Toler) THEN
	    Nom=LNom
	    Den=LDen
	    EXIT
	  ELSEIF (abs(float(RNom)/float(RDen)-abs(ModDec))<Toler) THEN
	    Nom=RNom
	    Den=RDen
	    EXIT
	  ELSE
            IF (abs(ModDec)<(float(LNom+RNom)/float(LDen+RDen))) THEN
	      RNom=LNom+RNom
	      RDen=LDen+RDen	      
	    ELSE
	      LNom=LNom+RNom
	      LDen=LDen+RDen	      
	    ENDIF
	  ENDIF
	ENDDO
	nom=nom+den*IntDec
	IF (InDec<0) Nom=-Nom
      END SUBROUTINE DecToFraction

! Initialize the random number generator
!*****************************************
      INTEGER FUNCTION GetSeedFromTime()
!*****************************************
      character*10 :: timestring
      real :: dummy
      ! Get a quaasirandom seed from the miliseconds of the current time
      call date_and_time(time=timestring)
      read(timestring,'(f10.3)') dummy
      GetSeedFromTime=nint(1000.*dummy)
      END FUNCTION GetSeedFromTime


! Working subroutine for analysis of the connectivity of the density
!*****************************************
      SUBROUTINE RestoreConnectivity(gamma)
!*****************************************
      USE SF_Module
      IMPLICIT NONE
      REAL, PARAMETER :: GoodConnectivity=0.9
      REAL, PARAMETER :: denstepfac=0.1
      REAL :: gamma !degree of overprojection
      REAL, DIMENSION(0:NVoxTot-1) :: exprho, origrho
      REAL :: Connect,denstep,stdev,denlimit,rhoave
      INTEGER :: l1, l2, l3, i, NPivots
      INTEGER, DIMENSION(SDim) :: hkl, hkltmp
      LOGICAL :: change
!      character*(filelength) :: rhofiletmp
!      character*(linelength) :: rhotypetmp
!	rhofiletmp='before_connect.xplor'
!        rhotypetmp='xplor'
!        CALL WriteRho(rhofiletmp,rhotypetmp)
      origrho=rho
      rhoave=sum(rho)/float(NVoxTot)
      rho=rho-rhoave
      stdev=sqrt(sum((rho)**2)/float(NVoxTot-1))
      denstep=denstepfac*stdev
      denlimit=stdev*connstdevfac
100   CONTINUE
      CALL CalculateConnectivity(Connect,stdev,GoodConnectivity)
!      write(*,*) 'Connectivity:',Connect
!        rho=float(skelet)
!	rhofiletmp='skelet.xplor'
!        rhotypetmp='xplor'
!        CALL WriteRho(rhofiletmp,rhotypetmp)
!        STOP
      IF (Connect>GoodConnectivity) THEN
        IF (abs(gamma)>eps) THEN
	  rho=(1.+gamma)*rho-gamma*origrho
	ENDIF       
        RETURN
      ENDIF
!      write(*,*) 'stdev:',stdev
      !Expand rho by incrementing every neighbor of a connected graph by denstep
      exprho=rho
      NPivots=0
200   DO i=0,NVoxTot-1
        IF (skelet(i)<0) THEN
          exprho(i)=min(denlimit+denstep/2.,exprho(i)+denstep)
	ELSEIF (skelet(i)==0) THEN !Search for neighbors - if there is a neighbor, increase by denstep, and assign the number of the corresponding graph
	  call coordinates(i,hkl)
	  change=.false.
	  DO l1=1,3
	    DO l2=-1,1,2
              hkltmp=hkl
	      hkltmp(l1)=mod(hkl(l1)+l2+NVox(l1),NVox(l1))
              call ArrayLocation(hkltmp,l3)
	      IF ((skelet(l3)/=0).and.(skelet(l3)<NVoxTot)) THEN
	        exprho(i)=min(denlimit+denstep/2.,exprho(i)+denstep)
		if (skelet(i)==0) THEN
		  skelet(i)=2*NVoxTot-abs(skelet(l3)) !Adding 2*NVoxTot to distinguish newly added points from previously added
!                  write(*,*) skelet(l3),skelet(i)
		  NPivots=NPivots+1
		endif
		change=.true.
		EXIT
	      ENDIF
            ENDDO
	    IF (change) EXIT
	  ENDDO
	ENDIF
      ENDDO      
      DO i=0,NVoxTot-1
        if (skelet(i)>NVoxTot) skelet(i)=skelet(i)-2*NVoxTot
      ENDDO
      !At this moment we have increased the density of every neighbor by denstep,
      !and assigned the point temporarily to its graph
      !Now search for pivot points, i.e. new connections between graphs
      NPivots=0
      DO i=0,NVoxTot-1
        IF ((exprho(i)>denlimit).and.(skelet(i)<0)) THEN !Potential candidate 
	  call coordinates(i,hkl)
	  change=.false.
	  DO l1=1,3
	    DO l2=-1,1,2
              hkltmp=hkl
	      hkltmp(l1)=mod(hkl(l1)+l2+NVox(l1),NVox(l1))
              call ArrayLocation(hkltmp,l3)
	      IF ((skelet(l3)/=skelet(i)).and.(skelet(l3)/=0).and.(exprho(l3)>denlimit)) THEN
                skelet(i)=abs(skelet(i)) !This point is touching another graph - make it a regular graph point
		Npivots=Npivots+1
                change=.true.
		EXIT
	      ENDIF
            ENDDO
	    IF (change) EXIT
	  ENDDO
        ENDIF	
      ENDDO
!      write(*,*) 'Pivots:',NPivots
      IF (NPivots>0) THEN !Some new connections - update rho
!        write(*,*) 'Pivots:',NPivots
	DO i=0,NVoxTot-1
	  if (skelet(i)>0) rho(i)=exprho(i)
	ENDDO
        goto 100 !Recalculate connectivity, update skeleton, and start again
      ELSE
        goto 200 !Continue expansion
      ENDIF      
      END SUBROUTINE RestoreConnectivity

! Working subroutine for analysis of the connectivity of the density
!*****************************************
      SUBROUTINE CalculateConnectivity(Connectivity,stdev,GoodConnectivity)
!*****************************************
      USE SF_Module
      IMPLICIT NONE
      TYPE Node_Type
        integer :: pos,grnr,nbnr !Position in the rho array, number of component the node belongs to, number of neighbors
	integer, dimension(1:6) :: nb !list of neighbors
      END TYPE Node_Type
      TYPE(Node_Type),ALLOCATABLE, DIMENSION(:) :: graph
      INTEGER, ALLOCATABLE, DIMENSION(:) :: tbv !list of "to-be-visited" nodes
      INTEGER :: Ntbv,curtbv,curpos !number of "to-be-visited" points, curent visited point
      INTEGER, ALLOCATABLE,DIMENSION(:) :: GraphList
      INTEGER, DIMENSION(0:NVoxTot-1) :: pos2node
      INTEGER, DIMENSION(SDim) :: hkl,hkltmp
      INTEGER :: l1,l2,l3,i,ii,NGraph,NGoodGraph,NTotalGrPoints,NConnectedPoints
      REAL :: stdev,zerolevel,Connectivity,GoodConnectivity
      
      IF (abs(stdev)<eps) THEN
        zerolevel=sum(rho)/float(NVoxTot)
        stdev=sqrt(sum((rho-zerolevel)**2)/float(NVoxTot-1))
!        write(*,*) 'Standard deviation:',stdev
      ENDIF
!      rho=rho-zerolevel
      NTotalGrPoints=0
      DO i=0,NVoxTot-1
        if (rho(i)>connstdevfac*stdev) NTotalGrPoints=NTotalGrPoints+1
      ENDDO
!      write(*,*) 'NTotalGrPoints:',NTotalGrPoints
!      write(*,*) 'Proportion above stdevfac*stdev:',float(NTotalGrPoints)/float(NVoxTot)
!      write(*,*) 'Stdevfac*stdev:',connstdevfac,stdev,connstdevfac*stdev
!      read(*,*)
!     Breadth-first algorithm (or something quite similar to it)
      ALLOCATE(graph(1:NTotalGrPoints))
      ALLOCATE(tbv(1:2*NTotalGrPoints))
      !Fill the positions of the graph nodes
      pos2node=0
      ii=0
      DO i=0,NVoxTot-1
        IF (rho(i)>connstdevfac*stdev) THEN
	  ii=ii+1
          graph(ii)%pos=i !Numerate the graph nodes
	  pos2node(i)=ii    !Allows for back-tracing the node number from its position in a constant time
        ENDIF
      ENDDO
      ii=0
      DO i=0,NVoxTot-1
        IF (rho(i)>connstdevfac*stdev) THEN
          ii=ii+1
          graph(ii)%grnr=0
	  graph(ii)%nbnr=0
	  !find the good neighbors
	  call coordinates(i,hkl)
	  DO l1=1,3
	    DO l2=-1,1,2
              hkltmp=hkl
	      hkltmp(l1)=mod(hkl(l1)+l2+NVox(l1),NVox(l1))
              call ArrayLocation(hkltmp,l3)
	      IF (rho(l3)>connstdevfac*stdev) THEN !Find that node in the list
	        graph(ii)%nbnr=graph(ii)%nbnr+1
                graph(ii)%nb(graph(ii)%nbnr)=pos2node(l3)
	      ENDIF
            ENDDO
	  ENDDO
	ENDIF
      ENDDO
      !Graph is filled, assign the component numbers
      tbv=0
      Ntbv=0
      i=0
      NGraph=0
      DO i=1,NTotalGrPoints !Cycle over all nodes of the graph
        IF (graph(i)%grnr==0) THEN !new graph
	  NGraph=NGraph+1
	  Ntbv=1
	  tbv(1)=i
	  curtbv=0
	  DO !Cycle over all "to-be-visited" points
            curtbv=curtbv+1
	    curpos=tbv(curtbv)
            if (graph(curpos)%grnr==0) THEN
	      graph(tbv(curtbv))%grnr=NGraph
	      DO l1=1,graph(curpos)%nbnr
	        IF (graph(graph(curpos)%nb(l1))%grnr==0) THEN
		  Ntbv=Ntbv+1
		  tbv(Ntbv)=graph(curpos)%nb(l1)
	        ENDIF
	      ENDDO
	    ENDIF
	    if (curtbv==Ntbv) EXIT
	  ENDDO
	ENDIF
      ENDDO
      ALLOCATE(GraphList(1:NGraph))
      GraphList=0
      DO i=1,NTotalGrPoints
        GraphList(graph(i)%grnr)=GraphList(graph(i)%grnr)+1
      ENDDO
!      CALL DATE_AND_TIME (Date, Time)
!      read(time,*) secend
!      write(*,*) 'Time:',secend-secstart
      NConnectedPoints=0
      NGoodGraph=0
      DO i=1,NGraph
!        write(*,*) i, Graphlist(i)
        IF (GraphList(i)>=LargeGraph) THEN
	  NConnectedPoints=NConnectedPoints+GraphList(i)
          NGoodGraph=NGoodGraph+1
	ENDIF
      ENDDO
!      write(*,*) 'NGoodGraph:',NGoodGraph, NGraph
!      write(*,*) 'Total connectivity:',NGraph,NConnectedPoints, NTotalGrPoints, float(NConnectedPoints)/float(NTotalGrPoints)
      Connectivity=float(NConnectedPoints)/float(NTotalGrPoints)
      !Fill the skelet aray to contain the connectivity information
      skelet=0
      DO i=1,NTotalGrPoints
        skelet(graph(i)%pos)=graph(i)%grnr
      ENDDO
!      IF (Connectivity>GoodConnectivity) THEN
!        write(*,*) 'Suppressing all small graphs'
!	DO i=1,NTotalGrPoints
!	  IF (GraphList(graph(i)%grnr)<Largegraph) THEN
!	    rho(graph(i)%pos)=0.0
!	  ENDIF
!	ENDDO
!      ENDIF
      DEALLOCATE(graph)
      DEALLOCATE(tbv)
      DEALLOCATE(GraphList)

      END SUBROUTINE CalculateConnectivity

! Check if given reflection occupies a Friedel independent position n the SF array
!*****************************************
      LOGICAL function IndepSFPos(hkl,SFpos)
!*****************************************
        use sf_module
	integer, dimension(:) :: hkl
	integer, dimension(SDim) ::sfcoo
	integer :: SFpos,i
	
        IndepSFPos=.true.
	IF (hkl(1)==0) THEN
	  call CoordinatesSF(SFpos,sfCoo)
          DO i=1,SDim
	    IF (SFCoo(i)==0) CYCLE !Skip zero indices
	    IF (SFCoo(i)>NVox(i)/2) THEN !The first non-zero index must be smaller than NVox(i)/2
	      IndepSFPos=.false.              
            ENDIF
            EXIT
          ENDDO
        ENDIF
      END FUNCTION IndepSFPos
      
! Sets the parameters of the iterative map
!*****************************************
      SUBROUTINE SetGdmPar(key)
!*****************************************
      USE SF_module
      USE Input_module      
      IMPLICIT NONE
      CHARACTER*(*) :: key
      CHARACTER*(linelength) :: copykey
      CHARACTER*(20) :: fwkey
      
      copykey=key
      fwkey=FirstWord(copykey,.true.)
      SELECT CASE (fwkey)
        CASE ('cf')
          gdmpar%beta1=1.
          gdmpar%gamma1D=1.
          IF (fdeltaf) THEN
            gdmpar%gamma1R=1.	      
          ELSE
            gdmpar%gamma1R=0.
          ENDIF
          gdmpar%beta2=0.
          gdmpar%gamma2R=0.
          gdmpar%gamma2D=0.
        CASE ('aar')
          gdmpar%beta1=0.5
          gdmpar%gamma1D=1.
          gdmpar%gamma1R=1.
          gdmpar%beta2=0.
          gdmpar%gamma2R=0.
          gdmpar%gamma2D=0.
        CASE ('dm')
          gdmpar%beta1=1.
          gdmpar%gamma1R=0.
          gdmpar%gamma1D=1.
          gdmpar%beta2=-1.
          gdmpar%gamma2D=0.
          gdmpar%gamma2R=-1.
        CASE ('lde')
          gdmpar%beta1=1.
          gdmpar%gamma1D=0.
          gdmpar%gamma1R=0.
          gdmpar%beta2=0.
          gdmpar%gamma2D=0.
          gdmpar%gamma2R=0.
        CASE ('workcode')
          gdmpar%beta1=1.0
          gdmpar%gamma1R=0.
          gdmpar%gamma1D=1.0
          gdmpar%beta2=-1.0
          gdmpar%gamma2D=0.
          gdmpar%gamma2R=-1.0
        CASE ('general')
	  read(copykey,*) gdmpar%beta1,gdmpar%gamma1R,gdmpar%gamma1D, & 
	                  gdmpar%beta2,gdmpar%gamma2D,gdmpar%gamma2R
      END SELECT
      IF ((abs(gdmpar%beta2)>eps).or.(abs(gdmpar%beta1+gdmpar%beta2-1.)>eps)) THEN
        gdmpar%needrhoold=.true.
        gdmpar%needrho1=.true.
      ELSE
        gdmpar%needrhoold=.false.
        gdmpar%needrho1=.false.
      ENDIF
      END SUBROUTINE SetGdmPar
      


! Check, if current vector represents a vector between two equivalent origins in the space group
!WARNING: This is a quick hack, that absolutely needs to be improved!
!*****************************************
      LOGICAL FUNCTION SGAllowedOrigin(Shift) result(GoodShift)
!*****************************************
      USE SF_module
      IMPLICIT NONE
      REAL, DIMENSION(1:SDim) :: Shift
      REAL, DIMENSION(1:SDim) :: Mask
      INTEGER, DIMENSION(1:SDim,1:SDim) :: Omega
      INTEGER :: sgaoi, sgaoii

      Omega=0
      DO sgaoi=1,NSymPrim
        Omega=Omega+SmOp(sgaoi)%R
      ENDDO
      mask=1.0E0
      DO sgaoi=1,SDim
        if (Omega(sgaoi,sgaoi)/=0) mask=0.0 !origin-free direction
      ENDDO
      Mask=mask*shift
      DO sgaoi=1,SDim
        GoodShift=.false.
	DO sgaoii=1,4
          IF (abs(sgaoii*mask(sgaoi)-anint(sgaoii*mask(sgaoi)))<eps) THEN
	    GoodShift=.true.
	    EXIT
	  ENDIF
	ENDDO
	IF (.not.GoodShift) EXIT
      ENDDO
      END FUNCTION SGAllowedOrigin
      
 ! Analysis of multiple runs - finding families of solutions
!*****************************************
      SUBROUTINE DensityCorrelationTable
!*****************************************
      USE SF_module
      USE SF_RhoIOInterfaces
      USE FFT_Interfaces
      IMPLICIT NONE
      REAL, PARAMETER :: rmsfrac=100.
      INTEGER :: i, ii, iii, NBestOrder
      Character*(6) :: BRForm='(i...)'
      INTEGER, DIMENSION(:,:), ALLOCATABLE :: Cluster
      REAL, DIMENSION(:,:), ALLOCATABLE :: MatchTable
      INTEGER :: NRho,NCluster, NBranch, mincl1, mincl2
      REAL :: mindist, rdummy, rmsmatch
      CHARACTER*256 :: runtstring
      LOGICAL :: HaveRunts, SGAllowedOrigin
      
        CALL CountDensities(NRho)
	IF (NBestOrder==0) THEN
	  write(lgu,'(a)') &
	    'No densities corrsponding to the input file settings have been found. No cluster analysis can be performed.'
	  RETURN
	ENDIF
        IF (.not.allocated(refrho)) ALLOCATE(RefRho(0:NVoxTot-1))
	ALLOCATE(MatchTable(1:NRho,1:NRho))
	MatchTable=0
        NBestOrder=int(log10(float(NRho))+eps)+1
        WRITE(BRForm(3:5),'(i1,''.'',i1)') NBestOrder,NBestOrder
        IF (Terminal) WRITE(*,*) 'Pair      Agreement     Shift'
        CALL FramedLabel(2,'Agreement between the saved densities:')
        WRITE(lgu,*) 'Pair     Agreement      Shift'
        rmsmatch=0.0
	DO i=1,NRho-1
	  CALL GetTheDensity(i)
	  refrho=rho
	  DO ii=i+1,NRho
	    CALL GetTheDensity(ii)
            CALL FFT(rho2sf)
            CALL SF2MEMRefl
            CALL MatchReference(1,1)
	    rmsmatch=rmsmatch+TwoDenMatch**2
	    IF (.not.SGAllowedOrigin(TwoDenShift)) TwoDenMatch=TwoDenMatch+200.
	    MatchTable(i,ii)=TwoDenMatch
	    MatchTable(ii,i)=TwoDenMatch
	    IF (Terminal) THEN
	      WRITE(*,'(2i4)',ADVANCE='NO') i,ii
              write(*,'(f7.2,8X)', ADVANCE='NO') TwoDenMatch
              DO iii=1,SDim
                write(*,'(f9.5)',ADVANCE='NO') TwoDenShift(iii)
              ENDDO
              WRITE(*,*)
	    ENDIF
	    WRITE(lgu,'(2i4)',ADVANCE='NO') i,ii
            write(lgu,'(f7.2,8X)', ADVANCE='NO') TwoDenMatch
            DO iii=1,SDim
              write(lgu,'(f9.5)',ADVANCE='NO') TwoDenShift(iii)
            ENDDO
            WRITE(lgu,*)
	  ENDDO
	ENDDO
        rmsmatch=sqrt(rmsmatch/(NRho*(NRho-1)/2))	
        CALL FramedLabel(2,'Agreement table:')
	write(lgu,'(8X)',ADVANCE='NO')
	DO i=1,NRho
	  write(lgu,'(i5)', ADVANCE='NO') i
	ENDDO
	write(lgu,*)
	DO i=1,NRho
	write(lgu,'(i5,3X)',ADVANCE='NO') i
	  DO ii=1,NRho
	    write(lgu,'(i5)',ADVANCE='NO') nint(MatchTable(i,ii))
	  ENDDO
	  write(lgu,*)
	ENDDO	
	write(lgu,*)
!	write(*,*) 'Root mean square match:', rmsmatch
	!And now do some cluser analysis
	ALLOCATE(Cluster(1:NRho, 1:NRho))
	Cluster=-1
        NCluster=NRho
	NBranch=1
	DO i=1,NRho
          Cluster(i,1)=i
	ENDDO
        DO	  
          mindist=1000000000
	  DO i=1,NCluster
	    DO ii=1,NCluster
	      if(i==ii) CYCLE
	      rdummy=InterClusterDist(i,ii,NBranch)
	      IF (rdummy<mindist) THEN
	        mindist=rdummy
                mincl1=i
	        mincl2=ii
              ENDIF
	    ENDDO
	  ENDDO
	  CALL CopyClusters(NBranch,NBranch+1)
	  NBranch=NBranch+1
	  CALL Mergeclusters(mincl1, mincl2, NBranch)
	  NCluster=NCluster-1
	  IF (NCluster==1) EXIT
	ENDDO
	! And now write out the statistics in a reasonable manner - that's the most difficult part
        WRITE(lgu,'(a)')  'Final statistics:'
        IF (Terminal) write(*,'(a)') 'Final statistics:'
	DO i=2,NBranch-1,1
	  IF (Terminal)  write(*,*) 'Step:',i
	  write(lgu,'(a,i3)') 'Step:',i
	  runtstring=' Outliers:'
	  HaveRunts=.false.
	  DO ii=1,NRho-i+1 !gives the number of clusters for each branch
	    !For each cluster write the list of members + the distance
	    mincl1=0
	    DO iii=1,NRho
	      IF (Cluster(iii,i)==ii) THEN
	        mincl1=mincl1+1
	        mincl2=iii
	      ENDIF
	    ENDDO
	    IF (mincl1>1) THEN
	      rdummy=IntraClusterDist(ii,i)
	      IF (rdummy<rmsmatch*rmsfrac) THEN
                IF (Terminal) write(*,'(i3,f7.2,'':  '')', ADVANCE='NO') ii, rdummy
                write(lgu,'(i3,f7.2,'':  '')', ADVANCE='NO') ii, rdummy
	        DO iii=1,NRho
	          IF (Cluster(iii,i)==ii) THEN
		    IF (Terminal) write(*,'(i4)', ADVANCE='NO') iii
		    write(lgu,'(i4)', ADVANCE='NO') iii
	          ENDIF
		ENDDO	    
                IF (Terminal) write(*,*)
		write(lgu,*)
              ENDIF
	    ELSE
	      write(runtstring,'(a,i3)') trim(runtstring),mincl2
	      HaveRunts=.true.
	    ENDIF           
	  ENDDO
	  IF (HaveRunts) THEN
	    IF (Terminal) write(*,*) trim(runtstring)
	    write(lgu,'(a)') trim(runtstring)
          ENDIF
          IF (Terminal) write(*,*)
          write(lgu,*)
	ENDDO
	DEALLOCATE(Cluster)
	DEALLOCATE(MatchTable)
        CONTAINS
	  SUBROUTINE CountDensities(cntd)
	    IMPLICIT NONE
	    integer :: cntd,cdi
            CHARACTER*(FileLength) :: densityname
	    LOGICAL :: exflag
            densityname=rhoprefix
            NBestOrder=0
	    cntd=0
	    
	    DO cdi=4,1,-1 !Hopefully nobody tests more than 9999 densities! :-) 
              densityname=rhoprefix
              WRITE(BRForm(3:5),'(i1,''.'',i1)') cdi,cdi
              WRITE(densityname(len_trim(rhoprefix)+1:len_trim(rhoprefix)+cdi),BRForm) 1
              densityname=trim(densityname)//'_'//OutRhoFile(1)
	      INQUIRE (FILE=densityname,EXIST=exflag)
	      IF (exflag) THEN
                NBestOrder=cdi
	        EXIT
	      ENDIF
            ENDDO
	    IF (NBestOrder==0) RETURN
            DO
              cntd=cntd+1
              densityname=rhoprefix
              WRITE(densityname(len_trim(rhoprefix)+1:len_trim(rhoprefix)+cdi),BRForm) cntd
              densityname=trim(densityname)//'_'//OutRhoFile(1)
	      INQUIRE (FILE=densityname,EXIST=exflag)
              IF (.not.exflag) THEN
                cntd=cntd-1
	        EXIT
	      ENDIF	      
	    ENDDO
	  END SUBROUTINE CountDensities

	  SUBROUTINE GetTheDensity(number)
            IMPLICIT NONE
	    integer :: number
            CHARACTER*(FileLength) :: densityname
            densityname=rhoprefix
            WRITE(densityname(len_trim(rhoprefix)+1:len_trim(rhoprefix)+NBestOrder),BRForm) number
            densityname=trim(densityname)//'_'//OutRhoFile(1)
            SELECT CASE(OutRhoType(1))
	      CASE('jana')
	        CALL ReadJana(rho, densityname)
	      CASE('xplor')
	        CALL ReadXplor(rho,densityname,.false.)
	      CASE('ccp4')
	        CALL ReadCCP4(rho, densityname)
	    END SELECT
	  END SUBROUTINE GetTheDensity

          REAL FUNCTION IntraClusterDist(clnr, Branch) result(intradist)
	    IMPLICIT NONE
	    INTEGER :: aci,acii,clnr, Branch,ndist
	    intradist=0.0
	    ndist=0
	    DO aci=1,NRho
	      DO acii=aci+1,NRho
	        if ((aci/=acii).and.(Cluster(aci,Branch)==clnr).and.(Cluster(acii,Branch)==clnr)) THEN
                  intradist=max(intradist,MatchTable(aci,acii))
!                  intradist=intradist+MatchTable(aci,acii)**2
!		  ndist=ndist+1
		ENDIF
	      ENDDO
	    ENDDO
!	    IF (ndist>0) intradist=sqrt(intradist/float(ndist))
          END FUNCTION IntraClusterDist

          REAL FUNCTION InterClusterDist(clnr1, clnr2, Branch) result(interdist)
	    IMPLICIT NONE
	    INTEGER :: ici,icii,clnr1, clnr2, Branch
	    interdist=1000000000
!	    interdist=0
	    DO ici=1,NRho
	      DO icii=ici+1,NRho
	        if (((Cluster(ici,Branch)==clnr1).and.(Cluster(icii,Branch)==clnr2)).or. &
		    ((Cluster(ici,Branch)==clnr2).and.(Cluster(icii,Branch)==clnr1))) THEN
                  interdist=min(interdist,MatchTable(ici,icii))
!                  interdist=max(interdist,MatchTable(ici,icii))
	        ENDIF
	      ENDDO
	    ENDDO
          END FUNCTION InterClusterDist
      
          SUBROUTINE CopyClusters(Branch1, Branch2)
	    IMPLICIT NONE
	    INTEGER :: Branch1, Branch2
	    Cluster(1:NRho,Branch2)=Cluster(1:NRho,Branch1)
	  END SUBROUTINE CopyClusters

          SUBROUTINE MergeClusters(incl1, incl2, Branch)
	    IMPLICIT NONE
	    INTEGER :: mci,incl1,incl2,cl1, cl2, Branch
	    IF (incl2<incl1) THEN
	      cl1=incl2
	      cl2=incl1
	    ELSE
	      cl1=incl1
	      cl2=incl2
	    ENDIF
	    DO mci=1,NRho
	      IF (Cluster(mci,Branch)==cl2) Cluster(mci,Branch)=cl1
	    ENDDO
	    DO mci=1,NRho
	      IF (Cluster(mci,Branch)==NCluster) Cluster(mci,Branch)=cl2
	    ENDDO
	  END SUBROUTINE MergeClusters
      END SUBROUTINE DensityCorrelationTable

      SUBROUTINE WriteForViewer
        USE SF_Module
	USE SF_RhoIOInterfaces
	IMPLICIT NONE
	CHARACTER*20 :: rhofiletmp, rhotypetmp
	character*(LineLength) :: mvcommand
	SELECT CASE (SystemFlag)
	  CASE(UnknownSystem,Unix)
	    mvcommand='mv -f'
	  CASE(Windows)
	    mvcommand='move /Y'
	END SELECT
        IF (ViewProgress==chimera) THEN
          rhofiletmp='tmpsf4viewer2.ccp4'
          CALL WriteCCP4(rho,rhofiletmp)
	  call system(trim(mvcommand)//' tmpsf4viewer2.ccp4 tmpsf4viewer.ccp4')
        ELSEIF (viewprogress==vesta) THEN
          rhofiletmp='tmpsf4viewer2.xplor'
          CALL WriteXplor(rho,rhofiletmp)
	  call system(trim(mvcommand)//' tmpsf4viewer2.xplor tmpsf4viewer.xplor')
!	  call system('VESTA tmpsf4viewer.xplor &')
	ENDIF
      END SUBROUTINE WriteForViewer

      SUBROUTINE RemoveForViewer
        USE SF_Module
	IMPLICIT NONE
	CHARACTER*20 :: rhofiletmp, rhotypetmp
	character*(LineLength) :: rmcommand
	SELECT CASE (SystemFlag)
	  CASE(UnknownSystem,Unix)
	    rmcommand='rm -f'
	  CASE(Windows)
	    rmcommand='del'
	END SELECT
        call system(trim(rmcommand)//' sftmplock')
        IF (ViewProgress==chimera) THEN
	  call system(trim(rmcommand)//' watchsuperfliptmp.py')
	ENDIF
      END SUBROUTINE RemoveForViewer

      SUBROUTINE RunViewer
        USE SF_Module
	IMPLICIT NONE
	logical :: exflag
	character*(LineLength) :: cpcommand, runcommand, lockcommand
	SELECT CASE (SystemFlag)	
	  CASE(UnknownSystem,Unix)
	    cpcommand='cp watchsuperflip.py watchsuperfliptmp.py'
	    runcommand='chimera watchsuperfliptmp.py &'
	    lockcommand='touch sftmplock'
	  CASE(Windows)
	    cpcommand='copy watchsuperflip.py watchsuperfliptmp.py'
	    runcommand='start chimera watchsuperfliptmp.py'
	    lockcommand='echo " " > sftmplock'
	END SELECT
	IF (ViewProgress==chimera) THEN
	  call system(trim(lockcommand))
          INQUIRE (FILE='watchsuperflip.py',EXIST=exflag)
	  IF (exflag) THEN
	    call system(trim(cpcommand))
	  ELSE
            call WriteChimeraScript
	  ENDIF
          call system(trim(runcommand))
        ELSEIF (Viewprogress==vesta) THEN
	  call system(trim(lockcommand))
	ENDIF
	CONTAINS
	SUBROUTINE WriteChimeraScript
	  OPEN(77,file='watchsuperfliptmp.py')
	  write(77,'(a)') 'class monitor_volume:'
	  write(77,'(a)') '   def __init__(self, map_file, lock_file, delay = 100):'
	  write(77,'(a)') '     self.map_file = map_file'
	  write(77,'(a)') '     self.lock_file = lock_file'
	  write(77,'(a)') '     self.delay = delay'				 
	  write(77,'(a)') '     import chimera'
	  write(77,'(a)') '     self.m = chimera.openModels.open(map_file)[0]'
	  write(77,'(a)') '     self.m.show()'
	  write(77,'(a)') '     self.reopen()'
	  write(77,'(a)') '   def reopen(self):'
	  write(77,'(a)') '     from os.path import exists'
	  write(77,'(a)') '     self.m.data.values_changed()'
	  write(77,'(a)') '     self.m.show()'
	  write(77,'(a)') '     if not exists(self.lock_file):'
	  write(77,'(a)') '       self.m.data.values_changed()'
	  write(77,'(a)') '       self.m.show()'
	  write(77,'(a)') '       return'
	  write(77,'(a)') '     from chimera.tkgui import app'
	  write(77,'(a)') '     app.after(self.delay, self.reopen)'
	  write(77,'(a)') 'monitor_volume(''tmpsf4viewer.ccp4'', ''sftmplock'')'
	  CLOSE(77)
	END SUBROUTINE WriteChimeraScript					 
      END SUBROUTINE RunViewer

!Try to gues which systems are we using based on the line-ends
!*****************************************
      SUBROUTINE TestSystem
!*****************************************
      USE SF_Module
      IMPLICIT NONE
      INTEGER :: reclone, i, ii, icharlist(1:12)
      CHARACTER*1 :: chr

      open(77,file='sftestsystem.tmp')
      write(77,'(a1)') ' '
      close(77)
      inquire(iolength=reclone) chr
      open(77,file='sftestsystem.tmp',ACCESS='DIRECT',RECL=reclone)
      i=0
      DO
        i=i+1
        read(77,rec=i,err=100) chr
        icharlist(i)=ichar(chr)
      ENDDO
100   i=i-1
      close(77, STATUS='DELETE')
      IF ((i==2).and.(icharlist(2)==10)) THEN
        SystemFlag=Unix
      ELSEIF ((i==3).and.(icharlist(2)==13).and.(icharlist(3)==10)) THEN
        SystemFlag=Windows
      ELSE
        SystemFlag=UnknownSystem
      ENDIF
      END SUBROUTINE TestSystem

! Handling of various messages passed to program via the file ComFile (jobname.BMcom)
!*****************************************
      function UserCom(key,valflag)
!*****************************************
      USE SF_Module

      IMPLICIT NONE
      
      LOGICAL :: UserCom,exflag !the function itself,existence of the file
      CHARACTER*3 :: valflag 	!flag describing the type if the input variable
      character*(*) :: key	!keyword
      UserCom=.false.
      Com%int=0
      Com%rl=0.0E0
      Com%chr=''
      INQUIRE (FILE=ComFile,EXIST=exflag) !test, if the file exists, if not, nothing happens
      IF (exflag) THEN
	OPEN(99,file=ComFile,ERR=997) 			!open the file
        READ(99,'(a)',ERR=998,END=998) buffer   	!read a line from it
!       if the start of the text in the file equals to the keyword passed, do something
	if (buffer(1:len_trim(key))==trim(key)) then 
          IF (valflag=='INT') THEN 				!if value is asked, try to read it
            READ(buffer(len_trim(key)+1:),*,ERR=998,END=998) com%int !read with default format
          ELSEIF (valflag=='REA') THEN 				
            READ(buffer(len_trim(key)+1:),*,ERR=998,END=998) com%rl
          ELSEIF (valflag=='CHR') THEN 				
            READ(buffer(len_trim(key)+1:),*,ERR=998,END=998) com%chr
          END IF
          UserCom=.true.
          close(99,STATUS='DELETE',ERR=999)		!close and delete the file, the information was read
        ELSE
          close(99,ERR=999)				!close the file, don't delete
	end if
      END IF
      RETURN
      !file could not be open, write a message and continue
997   WRITE(lgu,'(3a)') 'Cannot open the file ',trim(ComFile), 'although it exists. Ignoring it.'
      if (terminal) WRITE(*,'(3a)') 'Cannot open the file ',trim(ComFile), 'although it exists. Ignoring it.'
      RETURN
      !file could be open, but an error occured while trying to read the content. Write message and continue.
998   WRITE(lgu,'(3a)') 'Reading error or invalid key/value in the ',trim(ComFile),'. Ignoring it and deleting.'
      if (terminal) WRITE(lgu,'(3a)') 'Reading error or invalid key/value in the ',trim(ComFile), &
                                      '. Ignoring it and deleting.'
      CLOSE(99,STATUS='DELETE',ERR=999)
      RETURN
      !can't close and delete the file, but the contents might have been read properly
999   WRITE(lgu,'(3a)') 'Cannot delete file',trim(ComFile), &
                       ', problem with permission rights? Delete the file manually or correct the problem.'
      IF (Terminal) &
         WRITE(lgu,'(3a)') 'Cannot delete file',trim(ComFile), &
                           ', problem with permission rights? Delete the file manually or correct the problem.'
      
      END FUNCTION UserCom

      SUBROUTINE TimeTick(level, text, unit)
      integer :: level
      character*(*) :: text
      character*25 :: curtext
      integer, optional :: unit
      real, dimension(0:5), save :: oldtime
      integer :: curunit
      real :: curtime
      call cpu_time(curtime)
      if (len_trim(text)>0) THEN
        if (present(unit)) THEN
	  curunit=unit
	else
	  curunit=6
	endif
	curtext=text
        write(curunit,'(a,f15.3,50X)') curtext//':',1000*(curtime-oldtime(level))
      ENDIF
      oldtime(level)=curtime
      END SUBROUTINE TimeTick


      SUBROUTINE ErrorMsg(message)
        USE SF_Module
        IMPLICIT NONE
        CHARACTER*(*) :: message
        LOGICAL :: LogIsOpen
        INQUIRE(unit=lgu,opened=logisopen)
        INQUIRE(unit=lgu,opened=logisopen)
        IF (LogIsOpen) write(lgu,'(a)') trim(message)
        write(*,'(a)') trim(message)
      END SUBROUTINE ErrorMsg


!***********************************************************************
      subroutine StopProgram(message)
!***********************************************************************
      USE SF_Module
      IMPLICIT NONE
      CHARACTER*(*) :: message
      LOGICAL :: LogIsOpen
      
      INQUIRE(unit=lgu,opened=logisopen)
      IF (LogIsOpen) THEN
        IF (len_trim(message)>0) write(lgu,'(a)') trim(message)
	close(lgu)
      ENDIF
      IF (len_trim(message)>0) write(*,'(a)') trim(message)
      if (KeepWindow) THEN
        write(*,'(a)') 'Press Enter to exit'
	read(*,*)
      ENDIF     
      STOP
      END SUBROUTINE StopProgram

!***********************************************************************
      subroutine Wait(mseconds)
!***********************************************************************
      IMPLICIT NONE
      integer mseconds
      integer Time1,Time2,rate,maxcnt
      call system_clock(Time1,rate)
1000  CALL system_clock(Time2)
      IF ((1000.*float(Time2-Time1)/float(rate)<mseconds).or.(time2<time1)) goto 1000
      end subroutine Wait

! Deallocate all arrays, close files
!*****************************************************************************
      SUBROUTINE FinalCleanUp
!*****************************************************************************
      USE SF_Module
      USE SF_Composition
      IMPLICIT NONE

      CLOSE(lgu)
      IF (detailedphases>0) CLOSE(detphu)
      IF (UsePhases=='code') CLOSE(cwfu)
      DEALLOCATE(NVox)
      DEALLOCATE(NVoxIter)
      DEALLOCATE(NVoxFine)
      DEALLOCATE(NVoxSF)
      DEALLOCATE(DimStep)
      DEALLOCATE(DimStepSF)
      IF (ALLOCATED(NVoxRef)) DEALLOCATE(NVoxRef)
      IF (ALLOCATED(NVoxModel)) DEALLOCATE(NVoxModel)
      IF (ALLOCATED(NVoxEnv)) DEALLOCATE(NVoxEnv)
      IF (SDim>RDim) DEALLOCATE(QVec)
      DO cnt1=1,size(SmOp)
        DEALLOCATE(SmOp(cnt1)%R, SmOp(cnt1)%t)
      ENDDO
      DEALLOCATE(SmOp)
      DEALLOCATE(Center)
      IF (Derivesym=='use') THEN
        DO cnt1=1,size(SmOpOrig)
          DEALLOCATE(SmOpOrig(cnt1)%R, SmOpOrig(cnt1)%t)
        ENDDO
        DEALLOCATE(SmOpOrig)
        DEALLOCATE(CenterOrig)
      ENDIF
      IF (NTwin>0) THEN
        DO cnt1=1,NTwin
	  DEALLOCATE(TwinOp(cnt1)%R, twinOp(cnt1)%t)
	ENDDO
        DEALLOCATE(TwinOp)
      ENDIF
      DEALLOCATE(SurviveTwin)
      IF (allocated(Refl)) THEN
        DO cnt1=1,size(Refl)
          DEALLOCATE(Refl(cnt1)%hklfg)
        ENDDO
        DEALLOCATE(Refl)
      ENDIF
      IF (allocated(MEMRefl)) THEN
        DO cnt1=1,size(MEMRefl)
          DEALLOCATE(MEMRefl(cnt1)%hklfg)
        ENDDO
        DEALLOCATE(MEMRefl)
      ENDIF
      IF (hmclasses>0) DEALLOCATE(HMGramm)
      IF (allocated(MEMRefl2Refl)) DEALLOCATE(MEMRefl2Refl)
      DEALLOCATE(rho)      
      IF (allocated(rhoold)) DEALLOCATE(rhoold)
      IF (allocated(rho1)) DEALLOCATE(rho1)
      IF (allocated(refrho)) DEALLOCATE(refrho)
      IF (allocated(modelrho)) DEALLOCATE(modelrho)
      IF (allocated(envrho)) DEALLOCATE(envrho)
      IF (allocated(sumrho)) DEALLOCATE(sumrho)
      IF (allocated(rho50)) DEALLOCATE(rho50)
      IF (allocated(TwoDenShift)) deallocate(TwoDenShift)
      DEALLOCATE(sf)
      IF (NGroups>0) DEALLOCATE(GroupInt)
      DEALLOCATE(Cell)
      DEALLOCATE(RCell)
      DEALLOCATE(G)
      DEALLOCATE(GS)
      DEALLOCATE(Gi)
      DEALLOCATE(GSi)
      DEALLOCATE(BestRhoProps)
      deallocate(Generators)
      deallocate(Gencenters)
      IF (Allocated(PresModeLimits)) DEALLOCATE(PresModeLimits)
      IF (Allocated(PresModeTimes)) DEALLOCATE(PresModeTimes)
      IF (allocated(WeakRefl)) DEALLOCATE(WeakRefl, SaveWeakRefl)
      IF (ElNr>0) DEALLOCATE(Compos)
      IF (allocated(Atoms)) DEALLOCATE(Atoms)
      IF (allocated(sgelsy_zeroarr)) DEALLOCATE(sgelsy_zeroarr)
      IF (allocated(sgelsy_work)) deallocate(sgelsy_work)
      END SUBROUTINE FinalCleanUp

