      module symmetry_tools
        use matrices
	use solve_linear_module

!********************************************
      !parameters, data structures, variables
!********************************************
      IMPLICIT NONE
      integer, parameter :: BigBase=64
      real, parameter :: eps=0.00001, bigeps=0.001
      integer, parameter :: SGlistLength=509
      integer, parameter :: CrSysTriclinic  =1, &
                            CrSysMonoclinic =2, &
                            CrSysOrtorhombic=3, &
                            CrSysTetragonal =4, &
                            CrSysTrigonal   =5, &
                            CrSysHexagonal  =6, &
                            CrSysCubic      =7


      TYPE SymOp_type !Symmetry operation
        REAL, POINTER, DIMENSION(:) :: t !Translational vector
	INTEGER, POINTER, DIMENSION(:,:) :: R !Rotation matrix
      END TYPE SymOp_type

      type SymOpPropsType
        integer :: rank, det, order,IntrOrder,RotSense,trace, dirscore
        integer, pointer, dimension(:) :: dir
        real, pointer, dimension(:) :: IntrVec
      end type SymOpPropsType

      TYPE SGlist_type !Symmetry operation
        CHARACTER*132 :: SGFP
	CHARACTER*20  :: Orshift
	CHARACTER*20  :: HM
	CHARACTER*25  :: Hall
	CHARACTER     :: std
	INTEGER :: SGNr !Space group number in IT
      END TYPE SGlist_type

      type(SGlist_type), DIMENSION(1:SGlistLength) :: SGlist
      integer ::sgfp_verbosity=0, sgi
      integer ::sgfp_errorcode=0
      

!********************************
      !Data statements
!********************************
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=  1, 38)/ &
         '33001041a0'              , '(0,0,0)'       ,'P1'            , 'P 1'               ,    1, 'S' ,   &
         '331210341a0'             , '(0,0,0)'       ,'A1'            , 'A 1'               ,    1, ''  ,   &
         '331210541a0'             , '(0,0,0)'       ,'B1'            , 'B 1'               ,    1, ''  ,   &
         '331210641a0'             , '(0,0,0)'       ,'C1'            , 'C 1'               ,    1, ''  ,   &
         '331210741a0'             , '(0,0,0)'       ,'I1'            , 'I 1'               ,    1, ''  ,   &
         '3322103541a0'            , '(0,0,0)'       ,'F1'            , 'F 1'               ,    1, ''  ,   &
         '331310H41a0'             , '(0,0,0)'       ,'R1'            , 'R 1'               ,    1, ''  ,   &
         '3300100n{0'              , '(0,0,0)'       ,'P-1'           , '-P 1'              ,    2, 'S' ,   &
         '33121030n{0'             , '(0,0,0)'       ,'A-1'           , '-A 1'              ,    2, ''  ,   &
         '33121050n{0'             , '(0,0,0)'       ,'B-1'           , '-B 1'              ,    2, ''  ,   &
         '33121060n{0'             , '(0,0,0)'       ,'C-1'           , '-C 1'              ,    2, ''  ,   &
         '33121070n{0'             , '(0,0,0)'       ,'I-1'           , '-I 1'              ,    2, ''  ,   &
         '332210350n{0'            , '(0,0,0)'       ,'F-1'           , '-F 1'              ,    2, ''  ,   &
         '331310H0n{0'             , '(0,0,0)'       ,'R-1'           , '-R 1'              ,    2, ''  ,   &
         '3300100qW0'              , '(0,0,0)'       ,'P2'            , 'P 2y'              ,    3, 'S' ,   &
         '3300103}00'              , '(0,0,0)'       ,'P211'          , 'P 2x'              ,    3, ''  ,   &
         '3300100o00'              , '(0,0,0)'       ,'P112'          , 'P 2'               ,    3, ''  ,   &
         '3300120qW2'              , '(0,0,0)'       ,'P21'           , 'P 2yb'             ,    4, 'S' ,   &
         '3300120qW2'              , '(3/4,0,0)'     ,'P21:A'         , 'P 2yab'            ,    4, ''  ,   &
         '3300120qW2'              , '(0,0,3/4)'     ,'P21:C'         , 'P 2ybc'            ,    4, ''  ,   &
         '3300120qW2'              , '(3/4,0,3/4)'   ,'P21:N'         , 'P 2yn'             ,    4, ''  ,   &
         '3300123}04'              , '(0,0,0)'       ,'P2111'         , 'P 2xa'             ,    4, ''  ,   &
         '3300120o01'              , '(0,0,0)'       ,'P1121'         , 'P 2c'              ,    4, ''  ,   &
         '33121060qW0'             , '(0,0,0)'       ,'C2'            , 'C 2y'              ,    5, 'S' ,   &
         '33121060qW0'             , '(3/4,0,0)'     ,'C21'           , 'C 2yb'             ,    5, ''  ,   &
         '33121030qW0'             , '(0,0,0)'       ,'A2'            , 'A 2y'              ,    5, ''  ,   &
         '33121070qW0'             , '(0,0,0)'       ,'I2'            , 'I 2y'              ,    5, ''  ,   &
         '33121070qW0'             , '(3/4,0,3/4)'   ,'I21'           , 'I 2yb'             ,    5, ''  ,   &
         '332210350qW0'            , '(0,0,0)'       ,'F2'            , 'F 2y'              ,    5, ''  ,   &
         '33121030o00'             , '(0,0,0)'       ,'A112'          , 'A 2'               ,    5, ''  ,   &
         '33121050o00'             , '(0,0,0)'       ,'B112'          , 'B 2'               ,    5, ''  ,   &
         '33121070o00'             , '(0,0,0)'       ,'I112'          , 'I 2'               ,    5, ''  ,   &
         '3300103}20'              , '(0,0,0)'       ,'Pm'            , 'P -2y'             ,    6, 'S' ,   &
         '33001041Y0'              , '(0,0,0)'       ,'P11m'          , 'P -2'              ,    6, ''  ,   &
         '3300123}21'              , '(0,0,0)'       ,'Pc'            , 'P -2yc'            ,    7, 'S' ,   &
         '3300123}21'              , '(0,3/4,0)'     ,'Pcs'           , 'P -2ybc'           ,    7, ''  ,   &
         '3300123}24'              , '(0,0,0)'       ,'Pa'            , 'P -2ya'            ,    7, 'A' ,   &
         '3300123}25'              , '(0,0,0)'       ,'Pn'            , 'P -2yac'           ,    7, 'A' /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi= 39, 76)/ &
         '3300123}25'              , '(0,3/4,0)'     ,'Pns'           , 'P -2yabc'          ,    7, ''  ,   &
         '33001241Y4'              , '(0,0,0)'       ,'P11a'          , 'P -2a'             ,    7, ''  ,   &
         '33001241Y2'              , '(0,0,0)'       ,'P11b'          , 'P -2b'             ,    7, ''  ,   &
         '33001241Y6'              , '(0,0,0)'       ,'P11n'          , 'P -2ab'            ,    7, ''  ,   &
         '33121063}20'             , '(0,0,0)'       ,'Cm'            , 'C -2y'             ,    8, 'S' ,   &
         '33121033}20'             , '(0,0,0)'       ,'Am'            , 'A -2y'             ,    8, ''  ,   &
         '33121073}20'             , '(0,0,0)'       ,'Im'            , 'I -2y'             ,    8, ''  ,   &
         '331210541Y0'             , '(0,0,0)'       ,'B11m'          , 'B -2'              ,    8, ''  ,   &
         '33121263}21'             , '(0,0,0)'       ,'Cc'            , 'C -2yc'            ,    9, 'S' ,   &
         '33121273}21'             , '(0,0,0)'       ,'Ic'            , 'I -2yc'            ,    9, ''  ,   &
         '33121273}21'             , '(0,1/4,0)'     ,'Ia'            , 'I -2ya'            ,    9, ''  ,   &
         '33121233}24'             , '(0,0,0)'       ,'Aa'            , 'A -2ya'            ,    9, ''  ,   &
         '33121233}24'             , '(0,3/4,0)'     ,'An'            , 'A -2yac'           ,    9, ''  ,   &
         '331212541Y2'             , '(0,0,0)'       ,'B11b'          , 'B -2b'             ,    9, ''  ,   &
         '3300200n{03}20'          , '(0,0,0)'       ,'P2/m'          , '-P 2y'             ,   10, 'S' ,   &
         '3300200n{041Y0'          , '(0,0,0)'       ,'P112/m'        , '-P 2'              ,   10, ''  ,   &
         '3300200n{00qY0'          , '(0,0,0)'       ,'P2/m11'        , '-P 2x'             ,   10, ''  ,   &
         '3300220n{03}22'          , '(0,0,0)'       ,'P21/m'         , '-P 2yb'            ,   11, 'S' ,   &
         '3300220n{041Y1'          , '(0,0,0)'       ,'P1121/m'       , '-P 2c'             ,   11, ''  ,   &
         '3300220n{00qY4'          , '(0,0,0)'       ,'P21/m11'       , '-P 2xa'            ,   11, ''  ,   &
         '33122060n{03}20'         , '(0,0,0)'       ,'C2/m'          , '-C 2y'             ,   12, 'S' ,   &
         '33122030n{03}20'         , '(0,0,0)'       ,'A2/m'          , '-A 2y'             ,   12, ''  ,   &
         '33122030n{041Y0'         , '(0,0,0)'       ,'A112/m'        , '-A 2'              ,   12, ''  ,   &
         '33122050n{041Y0'         , '(0,0,0)'       ,'B112/m'        , '-B 2'              ,   12, ''  ,   &
         '33122070n{041Y0'         , '(0,0,0)'       ,'I112/m'        , '-I 2'              ,   12, ''  ,   &
         '33122070n{03}20'         , '(0,0,0)'       ,'I2/m'          , '-I 2y'             ,   12, ''  ,   &
         '33122270n{041Y2'         , '(0,0,0)'       ,'I112/b'        , '-I 2b'             ,   12, ''  ,   &
         '33122270n{041Y2'         , '(1/4,1/4,1/4)' ,'I112/a'        , '-I 2a'             ,   12, ''  ,   &
         '3300220n{03}21'          , '(0,0,0)'       ,'P2/c'          , '-P 2yc'            ,   13, 'S' ,   &
         '3300220n{03}24'          , '(0,0,0)'       ,'P2/a'          , '-P 2ya'            ,   13, 'A' ,   &
         '3300220n{03}25'          , '(0,0,0)'       ,'P2/n'          , '-P 2yac'           ,   13, 'A' ,   &
         '3300220n{041Y2'          , '(0,0,0)'       ,'P112/b'        , '-P 2b'             ,   13, ''  ,   &
         '3300220n{03}23'          , '(0,0,0)'       ,'P21/c'         , '-P 2ybc'           ,   14, 'S' ,   &
         '3300220n{03}26'          , '(0,0,0)'       ,'P21/a'         , '-P 2yab'           ,   14, 'A' ,   &
         '3300220n{041Y5'          , '(0,0,0)'       ,'P1121/a'       , '-P 2ac'            ,   14, ''  ,   &
         '3300220n{041Y3'          , '(0,0,0)'       ,'P1121/b'       , '-P 2bc'            ,   14, ''  ,   &
         '3300220n{03}27'          , '(0,0,0)'       ,'P21/n'         , '-P 2yn'            ,   14, 'A' ,   &
         '3300220n{041Y7'          , '(0,0,0)'       ,'P1121/n'       , '-P 2n'             ,   14, ''  /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi= 77,114)/ &
         '3300220n{00qY7'          , '(0,0,0)'       ,'P21/n11'       , '-P 2xn'            ,   14, ''  ,   &
         '33122250n{03}26'         , '(0,0,0)'       ,'B21/c'         , '-B 2ybc'           ,   14, ''  ,   &
         '33122260n{03}21'         , '(0,0,0)'       ,'C2/c'          , '-C 2yc'            ,   15, 'S' ,   &
         '33122230n{03}24'         , '(0,0,0)'       ,'A2/a'          , '-A 2ya'            ,   15, ''  ,   &
         '33122260n{03}21'         , '(1/4,3/4,0)'   ,'C2/n'          , '-C 2ybc'           ,   15, 'A' ,   &
         '33122270n{03}21'         , '(0,0,0)'       ,'I2/c'          , '-I 2yc'            ,   15, ''  ,   &
         '33122270n{03}21'         , '(1/4,1/4,1/4)' ,'I2/a'          , '-I 2ya'            ,   15, 'S'  ,   &
         '33122250n{041Y2'         , '(0,0,0)'       ,'B112/b'        , '-B 2b'             ,   15, ''  ,   &
         '3300203}000qW0'          , '(0,0,0)'       ,'P222'          , 'P 2 2'             ,   16, 'S' ,   &
         '3300223}000qW1'          , '(0,0,0)'       ,'P2221'         , 'P 2c 2'            ,   17, 'S' ,   &
         '3300223}060qW6'          , '(0,0,0)'       ,'P21212'        , 'P 2 2ab'           ,   18, 'S' ,   &
         '3300223}040qW3'          , '(0,3/4,1/2)'   ,'P212121'       , 'P 2ac 2ab'         ,   19, 'S' ,   &
         '33122263}000qW1'         , '(0,0,0)'       ,'C2221'         , 'C 2c 2'            ,   20, 'S' ,   &
         '33122063}000qW0'         , '(0,0,0)'       ,'C222'          , 'C 2 2'             ,   21, 'S' ,   &
         '332220353}000qW0'        , '(0,0,0)'       ,'F222'          , 'F 2 2'             ,   22, 'S' ,   &
         '33122073}000qW0'         , '(0,0,0)'       ,'I222'          , 'I 2 2'             ,   23, 'S' ,   &
         '33122273}000qW1'         , '(3/4,0,3/4)'   ,'I212121'       , 'I 2b 2c'           ,   24, 'S' ,   &
         '3300203}200qY0'          , '(0,0,0)'       ,'Pmm2'          , 'P 2 -2'            ,   25, 'S' ,   &
         '3300223}210qY0'          , '(0,0,0)'       ,'Pmc21'         , 'P 2c -2'           ,   26, 'S' ,   &
         '3300223}210qY1'          , '(0,0,0)'       ,'Pcc2'          , 'P 2 -2c'           ,   27, 'S' ,   &
         '3300223}240qY4'          , '(0,0,0)'       ,'Pma2'          , 'P 2 -2a'           ,   28, 'S' ,   &
         '3300223}240qY1'          , '(3/4,0,0)'     ,'Pca21'         , 'P 2c -2ac'         ,   29, 'S' ,   &
         '3300223}230qY3'          , '(0,0,0)'       ,'Pnc2'          , 'P 2 -2bc'          ,   30, 'S' ,   &
         '3300223}250qY0'          , '(0,0,0)'       ,'Pmn21'         , 'P 2ac -2'          ,   31, 'S' ,   &
         '3300223}260qY6'          , '(0,0,0)'       ,'Pba2'          , 'P 2 -2ab'          ,   32, 'S' ,   &
         '3300223}240qY3'          , '(3/4,3/4,0)'   ,'Pna21'         , 'P 2c -2n'          ,   33, 'S' ,   &
         '3300223}270qY7'          , '(0,0,0)'       ,'Pnn2'          , 'P 2 -2n'           ,   34, 'S' ,   &
         '33122063}200qY0'         , '(0,0,0)'       ,'Cmm2'          , 'C 2 -2'            ,   35, 'S' ,   &
         '33122263}210qY0'         , '(0,0,0)'       ,'Cmc21'         , 'C 2c -2'           ,   36, 'S' ,   &
         '33122263}210qY1'         , '(0,0,0)'       ,'Ccc2'          , 'C 2 -2c'           ,   37, 'S' ,   &
         '33122033}200qY0'         , '(0,0,0)'       ,'Amm2'          , 'A 2 -2'            ,   38, 'S' ,   &
         '33122233}220qY2'         , '(0,0,0)'       ,'Abm2'          , 'A 2 -2c'           ,   39, 'S' ,   &
         '33122233}240qY4'         , '(0,0,0)'       ,'Ama2'          , 'A 2 -2a'           ,   40, 'S' ,   &
         '33122233}260qY6'         , '(0,0,0)'       ,'Aba2'          , 'A 2 -2ac'          ,   41, 'S' ,   &
         '332220353}200qY0'        , '(0,0,0)'       ,'Fmm2'          , 'F 2 -2'            ,   42, 'S' ,   &
         '332224353}2L0qYL'        , '(0,0,0)'       ,'Fdd2'          , 'F 2 -2d'           ,   43, 'S' ,   &
         '33122073}200qY0'         , '(0,0,0)'       ,'Imm2'          , 'I 2 -2'            ,   44, 'S' ,   &
         '33122273}210qY1'         , '(0,0,0)'       ,'Iba2'          , 'I 2 -2c'           ,   45, 'S' /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=115,152)/ &
         '33122273}240qY4'         , '(0,0,0)'       ,'Ima2'          , 'I 2 -2a'           ,   46, 'S' ,   &
         '3300300n{041Y03}20'      , '(0,0,0)'       ,'Pmmm'          , '-P 2 2'            ,   47, 'S' ,   &
         '3300320n{741Y73}27'      , '(0,0,0)'       ,'Pnnn:1'        , 'P 2 2 -1n'         ,   48, ''  ,   &
         '3300320n{741Y73}27'      , '(1/4,1/4,1/4)' ,'Pnnn'          , '-P 2ab 2bc'        ,   48, 'S' ,   &
         '3300320n{041Y03}21'      , '(0,0,0)'       ,'Pccm'          , '-P 2 2c'           ,   49, 'S' ,   &
         '3300320n{641Y63}26'      , '(0,0,0)'       ,'Pban:1'        , 'P 2 2 -1ab'        ,   50, ''  ,   &
         '3300320n{341Y33}23'      , '(0,0,0)'       ,'Pncb:1'        , 'P 2 2 -1bc'        ,   50, ''  ,   &
         '3300320n{541Y53}25'      , '(0,0,0)'       ,'Pcna:1'        , 'P 2 2 -1ac'        ,   50, ''  ,   &
         '3300320n{641Y63}26'      , '(1/4,1/4,0)'   ,'Pban'          , '-P 2ab 2b'         ,   50, 'S' ,   &
         '3300320n{041Y43}20'      , '(0,0,0)'       ,'Pmma'          , '-P 2a 2a'          ,   51, 'S' ,   &
         '3300320n{041Y43}27'      , '(0,0,0)'       ,'Pnna'          , '-P 2a 2bc'         ,   52, 'S' ,   &
         '3300320n{041Y53}25'      , '(0,0,0)'       ,'Pmna'          , '-P 2ac 2'          ,   53, 'S' ,   &
         '3300320n{041Y43}21'      , '(0,0,0)'       ,'Pcca'          , '-P 2a 2ac'         ,   54, 'S' ,   &
         '3300320n{041Y03}26'      , '(0,0,0)'       ,'Pbam'          , '-P 2 2ab'          ,   55, 'S' ,   &
         '3300320n{041Y63}23'      , '(0,0,0)'       ,'Pccn'          , '-P 2ab 2ac'        ,   56, 'S' ,   &
         '3300320n{041Y13}23'      , '(0,0,0)'       ,'Pbcm'          , '-P 2c 2b'          ,   57, 'S' ,   &
         '3300320n{041Y03}27'      , '(0,0,0)'       ,'Pnnm'          , '-P 2 2n'           ,   58, 'S' ,   &
         '3300320n{641Y63}20'      , '(0,0,0)'       ,'Pmmn:1'        , 'P 2 2ab -1ab'      ,   59, ''  ,   &
         '3300320n{341Y03}20'      , '(0,0,0)'       ,'Pnmm:1'        , 'P 2bc 2 -1bc'      ,   59, ''  ,   &
         '3300320n{541Y03}25'      , '(0,0,0)'       ,'Pmnm:1'        , 'P 2ac 2ac -1ac'    ,   59, ''  ,   &
         '3300320n{641Y63}20'      , '(1/4,3/4,0)'   ,'Pmmn'          , '-P 2ab 2a'         ,   59, 'S' ,   &
         '3300320n{041Y73}21'      , '(0,0,0)'       ,'Pbcn'          , '-P 2n 2ab'         ,   60, 'S' ,   &
         '3300320n{041Y53}23'      , '(0,0,0)'       ,'Pbca'          , '-P 2ac 2ab'        ,   61, 'S' ,   &
         '3300320n{041Y53}22'      , '(0,0,0)'       ,'Pnma'          , '-P 2ac 2n'         ,   62, 'S' ,   &
         '33123260n{041Y13}21'     , '(0,0,0)'       ,'Cmcm'          , '-C 2c 2'           ,   63, 'S' ,   &
         '33123260n{041Y53}25'     , '(0,0,0)'       ,'Cmca'          , '-C 2bc 2'          ,   64, 'S' ,   &
         '33123060n{041Y03}20'     , '(0,0,0)'       ,'Cmmm'          , '-C 2 2'            ,   65, 'S' ,   &
         '33123260n{041Y03}21'     , '(0,0,0)'       ,'Cccm'          , '-C 2 2c'           ,   66, 'S' ,   &
         '33123260n{041Y43}20'     , '(1/4,3/4,0)'   ,'Cmma'          , '-C 2b 2'           ,   67, 'S' ,   &
         '33123260n{541Y53}25'     , '(0,0,0)'       ,'Ccca:1'        , 'C 2 2 -1bc'        ,   68, ''  ,   &
         '33123260n{541Y53}25'     , '(0,0,0)'       ,'Cccb:1'        , 'C 2 2 -1bc'        ,   68, ''  ,   &
         '33123230n{641Y63}26'     , '(0,0,0)'       ,'Abaa:1'        , 'A 2 2 -1ac'        ,   68, ''  ,   &
         '33123230n{641Y63}26'     , '(0,0,0)'       ,'Acaa:1'        , 'A 2 2 -1ac'        ,   68, ''  ,   &
         '33123250n{641Y63}26'     , '(0,0,0)'       ,'Bbcb:1'        , 'B 2 2 -1bc'        ,   68, ''  ,   &
         '33123250n{641Y63}26'     , '(0,0,0)'       ,'Bbab:1'        , 'B 2 2 -1bc'        ,   68, ''  ,   &
         '33123260n{541Y53}25'     , '(0,1/4,1/4)'   ,'Ccca'          , '-C 2b 2bc'         ,   68, 'S' ,   &
         '332230350n{041Y03}20'    , '(0,0,0)'       ,'Fmmm'          , '-F 2 2'            ,   69, 'S' ,   &
         '332234350n{L41YL3}2L'    , '(0,0,0)'       ,'Fddd:1'        , 'F 2 2 -1d'         ,   70, ''  /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=153,190)/ &
         '332234350n{L41YL3}2L'    , '(1/8,1/8,1/8)' ,'Fddd'          , '-F 2uv 2vw'        ,   70, 'S' ,   &
         '33123070n{041Y03}20'     , '(0,0,0)'       ,'Immm'          , '-I 2 2'            ,   71, 'S' ,   &
         '33123270n{041Y03}21'     , '(0,0,0)'       ,'Ibam'          , '-I 2 2c'           ,   72, 'S' ,   &
         '33123270n{041Y23}24'     , '(0,0,0)'       ,'Ibca'          , '-I 2b 2c'          ,   73, 'S' ,   &
         '33123270n{041Y23}22'     , '(0,0,0)'       ,'Imma'          , '-I 2b 2'           ,   74, 'S' ,   &
         '3300223}040qW0'          , '(0,0,0)'       ,'P2122'         , 'P 2a 2a'           ,    0, ''  ,   &
         '3300223}000qW2'          , '(0,3/4,0)'     ,'P2212'         , 'P 2 2b'            ,    0, ''  ,   &
         '3300223}000qW3'          , '(0,0,0)'       ,'P22121'        , 'P 2bc 2'           ,    0, ''  ,   &
         '3300223}050qW0'          , '(0,0,0)'       ,'P21221'        , 'P 2ac 2ac'         ,    0, ''  ,   &
         '33122233}040qW0'         , '(0,0,0)'       ,'A2122'         , 'A 2a 2a'           ,    0, ''  ,   &
         '33122253}000qW2'         , '(0,3/4,0)'     ,'B2212'         , 'B 2 2b'            ,    0, ''  ,   &
         '33122033}000qW0'         , '(0,0,0)'       ,'A222'          , 'A 2 2'             ,    0, ''  ,   &
         '33122053}000qW0'         , '(0,0,0)'       ,'B222'          , 'B 2 2'             ,    0, ''  ,   &
         '33002041Y03}20'          , '(0,0,0)'       ,'P2mm'          , 'P -2 2'            ,    0, ''  ,   &
         '33002041Y00qY0'          , '(0,0,0)'       ,'Pm2m'          , 'P -2 -2'           ,    0, ''  ,   &
         '3300223}200qY1'          , '(0,0,0)'       ,'Pcm21'         , 'P 2c -2c'          ,    0, ''  ,   &
         '33002241Y43}20'          , '(0,0,0)'       ,'P21ma'         , 'P -2a 2a'          ,    0, ''  ,   &
         '33002241Y03}24'          , '(0,0,0)'       ,'P21am'         , 'P -2 2a'           ,    0, ''  ,   &
         '33002241Y00qY2'          , '(0,0,0)'       ,'Pb21m'         , 'P -2 -2b'          ,    0, ''  ,   &
         '33002241Y20qY0'          , '(0,0,0)'       ,'Pm21b'         , 'P -2b -2'          ,    0, ''  ,   &
         '33002241Y43}24'          , '(0,0,0)'       ,'P2aa'          , 'P -2a 2'           ,    0, ''  ,   &
         '33002241Y20qY2'          , '(0,0,0)'       ,'Pb2b'          , 'P -2b -2b'         ,    0, ''  ,   &
         '3300223}220qY2'          , '(0,0,0)'       ,'Pbm2'          , 'P 2 -2b'           ,    0, ''  ,   &
         '33002241Y23}22'          , '(0,0,0)'       ,'P2mb'          , 'P -2b 2'           ,    0, ''  ,   &
         '33002241Y13}21'          , '(0,0,0)'       ,'P2cm'          , 'P -2c 2'           ,    0, ''  ,   &
         '33002241Y10qY1'          , '(0,0,0)'       ,'Pc2m'          , 'P -2c -2c'         ,    0, ''  ,   &
         '33002241Y40qY4'          , '(0,0,0)'       ,'Pm2a'          , 'P -2a -2a'         ,    0, ''  ,   &
         '3300223}210qY2'          , '(0,3/4,0)'     ,'Pbc21'         , 'P 2c -2b'          ,    0, ''  ,   &
         '33002241Y23}24'          , '(0,3/4,0)'     ,'P21ab'         , 'P -2b 2a'          ,    0, ''  ,   &
         '33002241Y43}21'          , '(0,0,3/4)'     ,'P21ca'         , 'P -2ac 2a'         ,    0, ''  ,   &
         '33002241Y20qY1'          , '(0,0,3/4)'     ,'Pc21b'         , 'P -2bc -2c'        ,    0, ''  ,   &
         '33002241Y40qY2'          , '(3/4,0,0)'     ,'Pb21a'         , 'P -2a -2ab'        ,    0, ''  ,   &
         '3300223}250qY5'          , '(0,0,0)'       ,'Pcn2'          , 'P 2 -2ac'          ,    0, ''  ,   &
         '33002241Y53}25'          , '(0,0,0)'       ,'P2na'          , 'P -2ac 2'          ,    0, ''  ,   &
         '33002241Y63}26'          , '(0,0,0)'       ,'P2an'          , 'P -2ab 2'          ,    0, ''  ,   &
         '33002241Y60qY6'          , '(0,0,0)'       ,'Pb2n'          , 'P -2ab -2ab'       ,    0, ''  ,   &
         '33002241Y30qY3'          , '(0,0,0)'       ,'Pn2b'          , 'P -2bc -2bc'       ,    0, ''  ,   &
         '3300223}200qY3'          , '(0,0,0)'       ,'Pnm21'         , 'P 2bc -2bc'        ,    0, ''  /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=191,228)/ &
         '33002241Y63}20'          , '(0,0,0)'       ,'P21mn'         , 'P -2ab 2ab'        ,    0, ''  ,   &
         '33002241Y03}25'          , '(0,0,0)'       ,'P21nm'         , 'P -2 2ac'          ,    0, ''  ,   &
         '33002241Y00qY3'          , '(0,0,0)'       ,'Pn21m'         , 'P -2 -2bc'         ,    0, ''  ,   &
         '33002241Y60qY0'          , '(0,0,0)'       ,'Pm21n'         , 'P -2ab -2'         ,    0, ''  ,   &
         '33002241Y33}23'          , '(0,0,0)'       ,'P2cb'          , 'P -2bc 2'          ,    0, ''  ,   &
         '33002241Y50qY5'          , '(0,0,0)'       ,'Pc2a'          , 'P -2ac -2ac'       ,    0, ''  ,   &
         '3300223}250qY2'          , '(3/4,3/4,0)'   ,'Pbn21'         , 'P 2c -2ab'         ,    0, ''  ,   &
         '33002241Y23}25'          , '(0,3/4,3/4)'   ,'P21nb'         , 'P -2bc 2a'         ,    0, ''  ,   &
         '33002241Y63}21'          , '(0,3/4,3/4)'   ,'P21cn'         , 'P -2n 2a'          ,    0, ''  ,   &
         '33002241Y60qY1'          , '(3/4,0,3/4)'   ,'Pc21n'         , 'P -2n -2ac'        ,    0, ''  ,   &
         '33002241Y40qY3'          , '(3/4,0,3/4)'   ,'Pn21a'         , 'P -2ac -2n'        ,    0, ''  ,   &
         '33002241Y73}27'          , '(0,0,0)'       ,'P2nn'          , 'P -2n 2'           ,    0, ''  ,   &
         '33002241Y70qY7'          , '(0,0,0)'       ,'Pn2n'          , 'P -2n -2n'         ,    0, ''  ,   &
         '331220341Y03}20'         , '(0,0,0)'       ,'A2mm'          , 'A -2 2'            ,    0, ''  ,   &
         '331220541Y00qY0'         , '(0,0,0)'       ,'Bm2m'          , 'B -2 -2'           ,    0, ''  ,   &
         '33122263}200qY1'         , '(0,0,0)'       ,'Ccm21'         , 'C 2c -2c'          ,    0, ''  ,   &
         '331222341Y43}20'         , '(0,0,0)'       ,'A21ma'         , 'A -2a 2a'          ,    0, ''  ,   &
         '331222341Y03}24'         , '(0,0,0)'       ,'A21am'         , 'A -2 2a'           ,    0, ''  ,   &
         '331222541Y00qY2'         , '(0,0,0)'       ,'Bb21m'         , 'B -2 -2b'          ,    0, ''  ,   &
         '331222541Y20qY0'         , '(0,0,0)'       ,'Bm21b'         , 'B -2b -2'          ,    0, ''  ,   &
         '331222341Y43}24'         , '(0,0,0)'       ,'A2aa'          , 'A -2a 2'           ,    0, ''  ,   &
         '331222541Y20qY2'         , '(0,0,0)'       ,'Bb2b'          , 'B -2b -2b'         ,    0, ''  ,   &
         '33122053}200qY0'         , '(0,0,0)'       ,'Bmm2'          , 'B 2 -2'            ,    0, ''  ,   &
         '331220541Y03}20'         , '(0,0,0)'       ,'B2mm'          , 'B -2 2'            ,    0, ''  ,   &
         '331220641Y03}20'         , '(0,0,0)'       ,'C2mm'          , 'C -2 2'            ,    0, ''  ,   &
         '331220641Y00qY0'         , '(0,0,0)'       ,'Cm2m'          , 'C -2 -2'           ,    0, ''  ,   &
         '331220341Y00qY0'         , '(0,0,0)'       ,'Am2m'          , 'A -2 -2'           ,    0, ''  ,   &
         '33122253}240qY4'         , '(0,0,0)'       ,'Bma2'          , 'B 2 -2c'           ,    0, ''  ,   &
         '331222541Y43}24'         , '(0,0,0)'       ,'B2cm'          , 'B -2c 2'           ,    0, ''  ,   &
         '331222641Y43}24'         , '(0,0,0)'       ,'C2mb'          , 'C -2b 2'           ,    0, ''  ,   &
         '331222641Y40qY4'         , '(0,0,0)'       ,'Cm2a'          , 'C -2b -2b'         ,    0, ''  ,   &
         '331222341Y20qY2'         , '(0,0,0)'       ,'Ac2m'          , 'A -2c -2c'         ,    0, ''  ,   &
         '33122253}220qY2'         , '(0,0,0)'       ,'Bbm2'          , 'B 2 -2b'           ,    0, ''  ,   &
         '331222541Y23}22'         , '(0,0,0)'       ,'B2mb'          , 'B -2b 2'           ,    0, ''  ,   &
         '331222641Y13}21'         , '(0,0,0)'       ,'C2cm'          , 'C -2c 2'           ,    0, ''  ,   &
         '331222641Y10qY1'         , '(0,0,0)'       ,'Cc2m'          , 'C -2c -2c'         ,    0, ''  ,   &
         '331222341Y40qY4'         , '(0,0,0)'       ,'Am2a'          , 'A -2a -2a'         ,    0, ''  ,   &
         '33122253}260qY6'         , '(0,0,0)'       ,'Bba2'          , 'B 2 -2bc'          ,    0, ''  /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=229,266)/ &
         '331222541Y63}26'         , '(0,0,0)'       ,'B2cb'          , 'B -2bc 2'          ,    0, ''  ,   &
         '331222641Y53}25'         , '(0,0,0)'       ,'C2cb'          , 'C -2bc 2'          ,    0, ''  ,   &
         '331222641Y50qY5'         , '(0,0,0)'       ,'Cc2a'          , 'C -2bc -2bc'       ,    0, ''  ,   &
         '331222341Y60qY6'         , '(0,0,0)'       ,'Ac2a'          , 'A -2ac -2ac'       ,    0, ''  ,   &
         '3322203541Y03}20'        , '(0,0,0)'       ,'F2mm'          , 'F -2 2'            ,    0, ''  ,   &
         '3322203541Y00qY0'        , '(0,0,0)'       ,'Fm2m'          , 'F -2 -2'           ,    0, ''  ,   &
         '3322243541YL3}2L'        , '(0,0,0)'       ,'F2dd'          , 'F -2d 2'           ,    0, ''  ,   &
         '3322243541YL0qYL'        , '(0,0,0)'       ,'Fd2d'          , 'F -2d -2d'         ,    0, ''  ,   &
         '331220741Y03}20'         , '(0,0,0)'       ,'I2mm'          , 'I -2 2'            ,    0, ''  ,   &
         '331220741Y00qY0'         , '(0,0,0)'       ,'Im2m'          , 'I -2 -2'           ,    0, ''  ,   &
         '331222741Y43}24'         , '(0,0,0)'       ,'I2cb'          , 'I -2a 2'           ,    0, ''  ,   &
         '331222741Y20qY2'         , '(0,0,0)'       ,'Ic2a'          , 'I -2b -2b'         ,    0, ''  ,   &
         '33122273}220qY2'         , '(0,0,0)'       ,'Ibm2'          , 'I 2 -2b'           ,    0, ''  ,   &
         '331222741Y23}22'         , '(0,0,0)'       ,'I2mb'          , 'I -2b 2'           ,    0, ''  ,   &
         '331222741Y13}21'         , '(0,0,0)'       ,'I2cm'          , 'I -2c 2'           ,    0, ''  ,   &
         '331222741Y10qY1'         , '(0,0,0)'       ,'Ic2m'          , 'I -2c -2c'         ,    0, ''  ,   &
         '331222741Y40qY4'         , '(0,0,0)'       ,'Im2a'          , 'I -2a -2a'         ,    0, ''  ,   &
         '3300320n{041Y43}24'      , '(0,0,0)'       ,'Pmaa'          , '-P 2a 2'           ,    0, ''  ,   &
         '3300320n{041Y23}20'      , '(0,0,0)'       ,'Pbmb'          , '-P 2b 2b'          ,    0, ''  ,   &
         '3300320n{341Y33}23'      , '(0,1/4,1/4)'   ,'Pncb'          , '-P 2b 2bc'         ,    0, ''  ,   &
         '3300320n{541Y53}25'      , '(1/4,0,1/4)'   ,'Pcna'          , '-P 2a 2c'          ,    0, ''  ,   &
         '3300320n{041Y23}22'      , '(0,0,0)'       ,'Pmmb'          , '-P 2b 2'           ,    0, ''  ,   &
         '3300320n{041Y03}22'      , '(0,0,0)'       ,'Pbmm'          , '-P 2 2b'           ,    0, ''  ,   &
         '3300320n{041Y13}20'      , '(0,0,0)'       ,'Pcmm'          , '-P 2c 2c'          ,    0, ''  ,   &
         '3300320n{041Y13}21'      , '(0,0,0)'       ,'Pmcm'          , '-P 2c 2'           ,    0, ''  ,   &
         '3300320n{041Y03}24'      , '(0,0,0)'       ,'Pmam'          , '-P 2 2a'           ,    0, ''  ,   &
         '3300320n{041Y23}25'      , '(0,0,0)'       ,'Pnnb'          , '-P 2b 2n'          ,    0, ''  ,   &
         '3300320n{041Y73}25'      , '(0,0,0)'       ,'Pbnn'          , '-P 2n 2b'          ,    0, ''  ,   &
         '3300320n{041Y63}27'      , '(0,0,0)'       ,'Pcnn'          , '-P 2ab 2c'         ,    0, ''  ,   &
         '3300320n{041Y63}21'      , '(0,0,0)'       ,'Pncn'          , '-P 2ab 2n'         ,    0, ''  ,   &
         '3300320n{041Y73}24'      , '(0,0,0)'       ,'Pnan'          , '-P 2n 2bc'         ,    0, ''  ,   &
         '3300320n{041Y33}20'      , '(0,0,0)'       ,'Pnmb'          , '-P 2bc 2bc'        ,    0, ''  ,   &
         '3300320n{041Y63}20'      , '(0,0,0)'       ,'Pbmn'          , '-P 2ab 2ab'        ,    0, ''  ,   &
         '3300320n{041Y03}25'      , '(0,0,0)'       ,'Pcnm'          , '-P 2 2ac'          ,    0, ''  ,   &
         '3300320n{041Y03}23'      , '(0,0,0)'       ,'Pncm'          , '-P 2 2bc'          ,    0, ''  ,   &
         '3300320n{041Y63}26'      , '(0,0,0)'       ,'Pman'          , '-P 2ab 2'          ,    0, ''  ,   &
         '3300320n{041Y23}23'      , '(0,0,0)'       ,'Pccb'          , '-P 2b 2c'          ,    0, ''  ,   &
         '3300320n{041Y43}26'      , '(0,0,0)'       ,'Pbaa'          , '-P 2a 2b'          ,    0, ''  /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=267,304)/ &
         '3300320n{041Y53}24'      , '(0,0,0)'       ,'Pcaa'          , '-P 2ac 2c'         ,    0, ''  ,   &
         '3300320n{041Y33}21'      , '(0,0,0)'       ,'Pbcb'          , '-P 2bc 2b'         ,    0, ''  ,   &
         '3300320n{041Y23}24'      , '(0,0,0)'       ,'Pbab'          , '-P 2b 2ab'         ,    0, ''  ,   &
         '3300320n{041Y33}23'      , '(0,0,0)'       ,'Pmcb'          , '-P 2bc 2'          ,    0, ''  ,   &
         '3300320n{041Y53}20'      , '(0,0,0)'       ,'Pcma'          , '-P 2ac 2ac'        ,    0, ''  ,   &
         '3300320n{041Y53}26'      , '(0,0,0)'       ,'Pnaa'          , '-P 2ac 2bc'        ,    0, ''  ,   &
         '3300320n{041Y33}25'      , '(0,0,0)'       ,'Pbnb'          , '-P 2bc 2ab'        ,    0, ''  ,   &
         '3300320n{041Y13}24'      , '(0,0,0)'       ,'Pcam'          , '-P 2c 2ac'         ,    0, ''  ,   &
         '3300320n{041Y53}21'      , '(0,0,0)'       ,'Pmca'          , '-P 2ac 2a'         ,    0, ''  ,   &
         '3300320n{041Y23}26'      , '(0,0,0)'       ,'Pmab'          , '-P 2b 2a'          ,    0, ''  ,   &
         '3300320n{041Y43}22'      , '(0,0,0)'       ,'Pbma'          , '-P 2a 2ab'         ,    0, ''  ,   &
         '3300320n{041Y33}22'      , '(0,0,0)'       ,'Pcmb'          , '-P 2bc 2c'         ,    0, ''  ,   &
         '3300320n{041Y73}27'      , '(0,0,0)'       ,'Pmnn'          , '-P 2n 2'           ,    0, ''  ,   &
         '3300320n{041Y73}20'      , '(0,0,0)'       ,'Pnmn'          , '-P 2n 2n'          ,    0, ''  ,   &
         '3300320n{341Y03}20'      , '(0,3/4,3/4)'   ,'Pnmm'          , '-P 2c 2bc'         ,    0, ''  ,   &
         '3300320n{541Y03}25'      , '(1/4,0,3/4)'   ,'Pmnm'          , '-P 2c 2a'          ,    0, ''  ,   &
         '3300320n{041Y73}26'      , '(0,0,0)'       ,'Pcan'          , '-P 2n 2c'          ,    0, ''  ,   &
         '3300320n{041Y43}23'      , '(0,0,0)'       ,'Pnca'          , '-P 2a 2n'          ,    0, ''  ,   &
         '3300320n{041Y33}24'      , '(0,0,0)'       ,'Pnab'          , '-P 2bc 2n'         ,    0, ''  ,   &
         '3300320n{041Y53}27'      , '(0,0,0)'       ,'Pbna'          , '-P 2ac 2b'         ,    0, ''  ,   &
         '3300320n{041Y23}27'      , '(0,0,0)'       ,'Pcnb'          , '-P 2b 2ac'         ,    0, ''  ,   &
         '3300320n{041Y33}26'      , '(0,0,0)'       ,'Pcab'          , '-P 2bc 2ac'        ,    0, ''  ,   &
         '3300320n{041Y33}27'      , '(0,0,0)'       ,'Pmnb'          , '-P 2bc 2a'         ,    0, ''  ,   &
         '3300320n{041Y13}27'      , '(0,0,0)'       ,'Pbnm'          , '-P 2c 2ab'         ,    0, ''  ,   &
         '3300320n{041Y73}22'      , '(0,0,0)'       ,'Pcmn'          , '-P 2n 2ac'         ,    0, ''  ,   &
         '3300320n{041Y73}23'      , '(0,0,0)'       ,'Pmcn'          , '-P 2n 2a'          ,    0, ''  ,   &
         '3300320n{041Y13}26'      , '(0,0,0)'       ,'Pnam'          , '-P 2c 2n'          ,    0, ''  ,   &
         '33123260n{041Y13}20'     , '(0,0,0)'       ,'Ccmm'          , '-C 2c 2c'          ,    0, ''  ,   &
         '33123230n{041Y43}20'     , '(0,0,0)'       ,'Amma'          , '-A 2a 2a'          ,    0, ''  ,   &
         '33123230n{041Y03}24'     , '(0,0,0)'       ,'Amam'          , '-A 2 2a'           ,    0, ''  ,   &
         '33123250n{041Y03}22'     , '(0,0,0)'       ,'Bbmm'          , '-B 2 2b'           ,    0, ''  ,   &
         '33123250n{041Y23}22'     , '(0,0,0)'       ,'Bmmb'          , '-B 2b 2'           ,    0, ''  ,   &
         '33123260n{041Y53}20'     , '(0,0,0)'       ,'Ccmb'          , '-C 2bc 2bc'        ,    0, ''  ,   &
         '33123230n{041Y63}20'     , '(0,0,0)'       ,'Abma'          , '-A 2ac 2ac'        ,    0, ''  ,   &
         '33123230n{041Y03}26'     , '(0,0,0)'       ,'Acam'          , '-A 2 2ac'          ,    0, ''  ,   &
         '33123250n{041Y03}26'     , '(0,0,0)'       ,'Bbcm'          , '-B 2 2bc'          ,    0, ''  ,   &
         '33123250n{041Y63}26'     , '(0,0,0)'       ,'Bmab'          , '-B 2bc 2'          ,    0, ''  ,   &
         '33123030n{041Y03}20'     , '(0,0,0)'       ,'Ammm'          , '-A 2 2'            ,    0, ''  /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=305,342)/ &
         '33123050n{041Y03}20'     , '(0,0,0)'       ,'Bmmm'          , '-B 2 2'            ,    0, ''  ,   &
         '33123230n{041Y43}24'     , '(0,0,0)'       ,'Amaa'          , '-A 2a 2'           ,    0, ''  ,   &
         '33123250n{041Y23}20'     , '(0,0,0)'       ,'Bbmb'          , '-B 2b 2b'          ,    0, ''  ,   &
         '33123260n{041Y43}20'     , '(0,0,0)'       ,'Cmmb'          , '-C 2b 2b'          ,    0, ''  ,   &
         '33123230n{041Y03}22'     , '(0,1/4,3/4)'   ,'Abmm'          , '-A 2c 2c'          ,    0, ''  ,   &
         '33123230n{041Y03}22'     , '(0,0,0)'       ,'Acmm'          , '-A 2 2c'           ,    0, ''  ,   &
         '33123250n{041Y03}24'     , '(0,0,0)'       ,'Bmcm'          , '-B 2 2c'           ,    0, ''  ,   &
         '33123250n{041Y03}24'     , '(1/4,0,3/4)'   ,'Bmam'          , '-B 2c 2'           ,    0, ''  ,   &
         '33123260n{541Y53}25'     , '(1/4,1/2,1/4)' ,'Cccb'          , '-C 2b 2c'          ,    0, ''  ,   &
         '33123230n{641Y63}26'     , '(1/4,1/2,1/4)' ,'Abaa'          , '-A 2a 2c'          ,    0, ''  ,   &
         '33123230n{641Y63}26'     , '(1/4,1/4,1/2)' ,'Acaa'          , '-A 2ac 2c'         ,    0, ''  ,   &
         '33123250n{641Y63}26'     , '(1/4,1/4,1/2)' ,'Bbcb'          , '-B 2bc 2b'         ,    0, ''  ,   &
         '33123250n{641Y63}26'     , '(0,1/4,1/4)'   ,'Bbab'          , '-B 2b 2bc'         ,    0, ''  ,   &
         '33123270n{041Y43}24'     , '(0,0,0)'       ,'Imcb'          , '-I 2a 2'           ,    0, ''  ,   &
         '33123270n{041Y23}20'     , '(0,0,0)'       ,'Icma'          , '-I 2b 2b'          ,    0, ''  ,   &
         '33123270n{041Y23}24'     , '(1/4,3/4,1/4)' ,'Icab'          , '-I 2a 2b'          ,    0, ''  ,   &
         '33123270n{041Y23}22'     , '(1/4,1/4,1/4)' ,'Immb'          , '-I 2a 2a'          ,    0, ''  ,   &
         '33123270n{041Y03}22'     , '(1/4,1/4,3/4)' ,'Ibmm'          , '-I 2c 2c'          ,    0, ''  ,   &
         '33123270n{041Y03}22'     , '(0,0,0)'       ,'Icmm'          , '-I 2 2b'           ,    0, ''  ,   &
         '33123270n{041Y03}24'     , '(0,0,0)'       ,'Imcm'          , '-I 2 2a'           ,    0, ''  ,   &
         '33123270n{041Y03}24'     , '(1/4,3/4,3/4)' ,'Imam'          , '-I 2c 2'           ,    0, ''  ,   &
         '3300101xQ0'              , '(0,0,0)'       ,'P4'            , 'P 4'               ,   75, 'S' ,   &
         '3300141xQ1'              , '(0,0,0)'       ,'P41'           , 'P 4w'              ,   76, 'S' ,   &
         '3300121xQ1'              , '(0,0,0)'       ,'P42'           , 'P 4c'              ,   77, 'S' ,   &
         '3300141xQ3'              , '(0,0,0)'       ,'P43'           , 'P 4cw'             ,   78, 'S' ,   &
         '33121071xQ0'             , '(0,0,0)'       ,'I4'            , 'I 4'               ,   79, 'S' ,   &
         '33121471xQ9'             , '(0,0,0)'       ,'I41'           , 'I 4bw'             ,   80, 'S' ,   &
         '3300101xO0'              , '(0,0,0)'       ,'P-4'           , 'P -4'              ,   81, 'S' ,   &
         '33121071xO0'             , '(0,0,0)'       ,'I-4'           , 'I -4'              ,   82, 'S' ,   &
         '3300200n{01xQ0'          , '(0,0,0)'       ,'P4/m'          , '-P 4'              ,   83, 'S' ,   &
         '3300220n{01xQ1'          , '(0,0,0)'       ,'P42/m'         , '-P 4c'             ,   84, 'S' ,   &
         '3300220n{61xQ6'          , '(0,0,0)'       ,'P4/n:1'        , 'P 4ab -1ab'        ,   85, ''  ,   &
         '3300220n{61xQ6'          , '(1/4,3/4,0)'   ,'P4/n'          , '-P 4a'             ,   85, 'S' ,   &
         '3300220n{71xQ7'          , '(0,0,0)'       ,'P42/n:1'       , 'P 4n -1n'          ,   86, ''  ,   &
         '3300220n{71xQ7'          , '(3/4,3/4,1/4)' ,'P42/n'         , '-P 4bc'            ,   86, 'S' ,   &
         '33122070n{01xQ0'         , '(0,0,0)'       ,'I4/m'          , '-I 4'              ,   87, 'S' ,   &
         '33122470n{91xQ9'         , '(0,0,0)'       ,'I41/a:1'       , 'I 4bw -1bw'        ,   88, ''  ,   &
         '33122470n{91xQ9'         , '(0,1/4,1/8)'   ,'I41/a'         , '-I 4ad'            ,   88, 'S' /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=343,380)/ &
         '3300201xQ03}00'          , '(0,0,0)'       ,'P422'          , 'P 4 2'             ,   89, 'S' ,   &
         '3300221xQ63}06'          , '(0,0,0)'       ,'P4212'         , 'P 4ab 2ab'         ,   90, 'S' ,   &
         '3300241xQ13}00'          , '(0,0,3/4)'     ,'P4122'         , 'P 4w 2c'           ,   91, 'S' ,   &
         '3300241xQ93}0Z'          , '(3/4,3/4,1/2)' ,'P41212'        , 'P 4abw 2nw'        ,   92, 'S' ,   &
         '3300221xQ13}00'          , '(0,0,0)'       ,'P4222'         , 'P 4c 2'            ,   93, 'S' ,   &
         '3300221xQ73}07'          , '(0,0,0)'       ,'P42212'        , 'P 4n 2n'           ,   94, 'S' ,   &
         '3300241xQ33}00'          , '(0,0,3/4)'     ,'P4322'         , 'P 4cw 2c'          ,   95, 'S' ,   &
         '3300241xQB3}0X'          , '(3/4,3/4,0)'   ,'P43212'        , 'P 4nw 2abw'        ,   96, 'S' ,   &
         '33122071xQ03}00'         , '(0,0,0)'       ,'I422'          , 'I 4 2'             ,   97, 'S' ,   &
         '33122471xQ93}09'         , '(0,0,0)'       ,'I4122'         , 'I 4bw 2bw'         ,   98, 'S' ,   &
         '3300201xQ03}20'          , '(0,0,0)'       ,'P4mm'          , 'P 4 -2'            ,   99, 'S' ,   &
         '3300221xQ03}26'          , '(0,0,0)'       ,'P4bm'          , 'P 4 -2ab'          ,  100, 'S' ,   &
         '3300221xQ13}21'          , '(0,0,0)'       ,'P42cm'         , 'P 4c -2c'          ,  101, 'S' ,   &
         '3300221xQ73}27'          , '(0,0,0)'       ,'P42nm'         , 'P 4n -2n'          ,  102, 'S' ,   &
         '3300221xQ03}21'          , '(0,0,0)'       ,'P4cc'          , 'P 4 -2c'           ,  103, 'S' ,   &
         '3300221xQ03}27'          , '(0,0,0)'       ,'P4nc'          , 'P 4 -2n'           ,  104, 'S' ,   &
         '3300221xQ13}20'          , '(0,0,0)'       ,'P42mc'         , 'P 4c -2'           ,  105, 'S' ,   &
         '3300221xQ13}26'          , '(0,0,0)'       ,'P42bc'         , 'P 4c -2ab'         ,  106, 'S' ,   &
         '33122071xQ03}20'         , '(0,0,0)'       ,'I4mm'          , 'I 4 -2'            ,  107, 'S' ,   &
         '33122271xQ03}21'         , '(0,0,0)'       ,'I4cm'          , 'I 4 -2c'           ,  108, 'S' ,   &
         '33122471xQ93}20'         , '(0,0,0)'       ,'I41md'         , 'I 4bw -2'          ,  109, 'S' ,   &
         '33122471xQ93}22'         , '(0,0,0)'       ,'I41cd'         , 'I 4bw -2c'         ,  110, 'S' ,   &
         '3300201xO02}m0'          , '(0,0,0)'       ,'P-42m'         , 'P -4 2'            ,  111, 'S' ,   &
         '3300221xO02}m1'          , '(0,0,0)'       ,'P-42c'         , 'P -4 2c'           ,  112, 'S' ,   &
         '3300221xO02}m6'          , '(0,0,0)'       ,'P-421m'        , 'P -4 2ab'          ,  113, 'S' ,   &
         '3300221xO02}m7'          , '(0,0,0)'       ,'P-421c'        , 'P -4 2n'           ,  114, 'S' ,   &
         '3300201xO03}20'          , '(0,0,0)'       ,'P-4m2'         , 'P -4 -2'           ,  115, 'S' ,   &
         '3300221xO03}21'          , '(0,0,0)'       ,'P-4c2'         , 'P -4 -2c'          ,  116, 'S' ,   &
         '3300221xO03}26'          , '(0,0,0)'       ,'P-4b2'         , 'P -4 -2ab'         ,  117, 'S' ,   &
         '3300221xO03}27'          , '(0,0,0)'       ,'P-4n2'         , 'P -4 -2n'          ,  118, 'S' ,   &
         '33122071xO03}20'         , '(0,0,0)'       ,'I-4m2'         , 'I -4 -2'           ,  119, 'S' ,   &
         '33122271xO03}21'         , '(0,0,0)'       ,'I-4c2'         , 'I -4 -2c'          ,  120, 'S' ,   &
         '33122071xO02}m0'         , '(0,0,0)'       ,'I-42m'         , 'I -4 2'            ,  121, 'S' ,   &
         '33122471xO02}m9'         , '(0,0,0)'       ,'I-42d'         , 'I -4 2bw'          ,  122, 'S' ,   &
         '3300300n{01xQ03}20'      , '(0,0,0)'       ,'P4/mmm'        , '-P 4 2'            ,  123, 'S' ,   &
         '3300320n{01xQ03}21'      , '(0,0,0)'       ,'P4/mcc'        , '-P 4 2c'           ,  124, 'S' ,   &
         '3300320n{61xQ03}26'      , '(0,0,0)'       ,'P4/nbm:1'      , 'P 4 2 -1ab'        ,  125, ''  ,   &
         '3300320n{61xQ03}26'      , '(3/4,3/4,0)'   ,'P4/nbm'        , '-P 4a 2b'          ,  125, 'S' /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=381,418)/ &
         '3300320n{71xQ03}27'      , '(0,0,0)'       ,'P4/nnc:1'      , 'P 4 2 -1n'         ,  126, ''  ,   &
         '3300320n{71xQ03}27'      , '(3/4,3/4,1/4)' ,'P4/nnc'        , '-P 4a 2bc'         ,  126, 'S' ,   &
         '3300320n{01xQ03}26'      , '(0,0,0)'       ,'P4/mbm'        , '-P 4 2ab'          ,  127, 'S' ,   &
         '3300320n{01xQ03}27'      , '(0,0,0)'       ,'P4/mnc'        , '-P 4 2n'           ,  128, 'S' ,   &
         '3300320n{61xQ63}20'      , '(0,0,0)'       ,'P4/nmm:1'      , 'P 4ab 2ab -1ab'    ,  129, ''  ,   &
         '3300320n{61xQ63}20'      , '(3/4,1/4,0)'   ,'P4/nmm'        , '-P 4a 2a'          ,  129, 'S' ,   &
         '3300320n{61xQ63}21'      , '(0,0,0)'       ,'P4/ncc:1'      , 'P 4ab 2n -1ab'     ,  130, ''  ,   &
         '3300320n{61xQ63}21'      , '(3/4,1/4,0)'   ,'P4/ncc'        , '-P 4a 2ac'         ,  130, 'S' ,   &
         '3300320n{01xQ13}20'      , '(0,0,0)'       ,'P42/mmc'       , '-P 4c 2'           ,  131, 'S' ,   &
         '3300320n{01xQ13}21'      , '(0,0,0)'       ,'P42/mcm'       , '-P 4c 2c'          ,  132, 'S' ,   &
         '3300320n{61xQ13}26'      , '(0,1/2,3/4)'   ,'P42/nbc:1'     , 'P 4n 2c -1n'       ,  133, ''  ,   &
         '3300320n{61xQ13}26'      , '(3/4,3/4,0)'   ,'P42/nbc'       , '-P 4ac 2b'         ,  133, 'S' ,   &
         '3300320n{71xQ73}27'      , '(0,0,0)'       ,'P42/nnm:1'     , 'P 4n 2 -1n'        ,  134, ''  ,   &
         '3300320n{71xQ73}27'      , '(1/4,3/4,1/4)' ,'P42/nnm'       , '-P 4ac 2bc'        ,  134, 'S' ,   &
         '3300320n{01xQ13}26'      , '(0,0,0)'       ,'P42/mbc'       , '-P 4c 2ab'         ,  135, 'S' ,   &
         '3300320n{01xQ73}27'      , '(0,0,0)'       ,'P42/mnm'       , '-P 4n 2n'          ,  136, 'S' ,   &
         '3300320n{71xQ73}20'      , '(0,0,0)'       ,'P42/nmc:1'     , 'P 4n 2n -1n'       ,  137, ''  ,   &
         '3300320n{71xQ73}20'      , '(3/4,1/4,1/4)' ,'P42/nmc'       , '-P 4ac 2a'         ,  137, 'S' ,   &
         '3300320n{01xQ33}23'      , '(1/4,1/4,3/4)' ,'P42/ncm:1'     , 'P 4n 2ab -1n'      ,  138, ''  ,   &
         '3300320n{01xQ33}23'      , '(1/2,0,0)'     ,'P42/ncm'       , '-P 4ac 2ac'        ,  138, 'S' ,   &
         '33123070n{01xQ03}20'     , '(0,0,0)'       ,'I4/mmm'        , '-I 4 2'            ,  139, ''  ,   &
         '33123270n{01xQ03}21'     , '(0,0,0)'       ,'I4/mcm'        , '-I 4 2c'           ,  140, 'S' ,   &
         '33123470n{91xQ93}20'     , '(0,0,0)'       ,'I41/amd:1'     , 'I 4bw 2bw -1bw'    ,  141, ''  ,   &
         '33123470n{91xQ93}20'     , '(0,3/4,1/8)'   ,'I41/amd'       , '-I 4bd 2'          ,  141, 'S' ,   &
         '33123470n{91xQ93}22'     , '(0,0,0)'       ,'I41/acd:1'     , 'I 4bw 2aw -1bw'    ,  142, ''  ,   &
         '33123470n{91xQ93}22'     , '(1/2,3/4,1/8)' ,'I41/acd:3'     , '-I 4ad 2c'         ,  142, ''  ,   &
         '33123470n{91xQ93}22'     , '(0,3/4,1/8)'   ,'I41/acd'       , '-I 4bd 2c'         ,  142, 'S' ,   &
         '3300101w90'              , '(0,0,0)'       ,'P3'            , 'P 3'               ,  143, 'S' ,   &
         '3300131w91'              , '(0,0,0)'       ,'P31'           , 'P 31'              ,  144, 'S' ,   &
         '3300131w92'              , '(0,0,0)'       ,'P32'           , 'P 32'              ,  145, 'S' ,   &
         '331310H1w90'             , '(0,0,0)'       ,'R3'            , 'R 3'               ,  146, 'S' ,   &
         '3300102f00'              , '(0,0,0)'       ,'R3r'           , 'P 3*'              ,  146, ''  ,   &
         '3300103Xv0'              , '(0,0,0)'       ,'P-3'           , '-P 3'              ,  147, 'S' ,   &
         '331310H3Xv0'             , '(0,0,0)'       ,'R-3'           , '-R 3'              ,  148, 'S' ,   &
         '3300101t20'              , '(0,0,0)'       ,'R-3r'          , '-P 3*'             ,  148, ''  ,   &
         '3300201w9042p0'          , '(0,0,0)'       ,'P312'          , 'P 3 2'             ,  149, 'S' ,   &
         '3300201w903Sr0'          , '(0,0,0)'       ,'P321'          , 'P 3 2"'            ,  150, 'S' ,   &
         '3300231w9142p0'          , '(0,0,0)'       ,'P3112'         , 'P 31 2c (0 0 1)'   ,  151, 'S' /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=419,456)/ &
         '3300231w913Sr0'          , '(0,0,1/6)'     ,'P3121'         , 'P 31 2"'           ,  152, 'S' ,   &
         '3300231w9242p0'          , '(0,0,0)'       ,'P3212'         , 'P 32 2c (0 0 -1)'  ,  153, 'S' ,   &
         '3300231w923Sr0'          , '(0,0,5/6)'     ,'P3221'         , 'P 32 2"'           ,  154, 'S' ,   &
         '331320H1w903Sr0'         , '(0,0,0)'       ,'R32'           , 'R 3 2"'            ,  155, 'S' ,   &
         '3300202f001po0'          , '(0,0,0)'       ,'R32r'          , 'P 3* 2'            ,  155, ''  ,   &
         '3300201w9042r0'          , '(0,0,0)'       ,'P3m1'          , 'P 3 -2"'           ,  156, 'S' ,   &
         '3300201w903St0'          , '(0,0,0)'       ,'P31m'          , 'P 3 -2'            ,  157, 'S' ,   &
         '3300221w9042r1'          , '(0,0,0)'       ,'P3c1'          , 'P 3 -2"c'          ,  158, 'S' ,   &
         '3300221w903St1'          , '(0,0,0)'       ,'P31c'          , 'P 3 -2c'           ,  159, 'S' ,   &
         '331320H1w9042r0'         , '(0,0,0)'       ,'R3m'           , 'R 3 -2"'           ,  160, 'S' ,   &
         '3300202f0040m0'          , '(0,0,0)'       ,'R3mr'          , 'P 3* -2'           ,  160, ''  ,   &
         '331322H1w9042r1'         , '(0,0,0)'       ,'R3c'           , 'R 3 -2"c'          ,  161, 'S' ,   &
         '3300222f0040m7'          , '(0,0,0)'       ,'R3cr'          , 'P 3* -2n'          ,  161, ''  ,   &
         '3300203Xv03St0'          , '(0,0,0)'       ,'P-31m'         , '-P 3 2'            ,  162, 'S' ,   &
         '3300223Xv03St1'          , '(0,0,0)'       ,'P-31c'         , '-P 3 2c'           ,  163, 'S' ,   &
         '3300203Xv042r0'          , '(0,0,0)'       ,'P-3m1'         , '-P 3 2"'           ,  164, 'S' ,   &
         '3300223Xv042r1'          , '(0,0,0)'       ,'P-3c1'         , '-P 3 2"c'          ,  165, 'S' ,   &
         '331320H3Xv042r0'         , '(0,0,0)'       ,'R-3m'          , '-R 3 2"'           ,  166, 'S' ,   &
         '3300201t2040m0'          , '(0,0,0)'       ,'R-3mr'         , '-P 3* 2'           ,  166, ''  ,   &
         '331322H3Xv042r1'         , '(0,0,0)'       ,'R-3c'          , '-R 3 2"c'          ,  167, 'S' ,   &
         '3300221t2040m7'          , '(0,0,0)'       ,'R-3cr'         , '-P 3* 2n'          ,  167, ''  ,   &
         '3300103Xx0'              , '(0,0,0)'       ,'P6'            , 'P 6'               ,  168, 'S' ,   &
         '3300163Xx01'             , '(0,0,0)'       ,'P61'           , 'P 61'              ,  169, 'S' ,   &
         '3300163Xx05'             , '(0,0,0)'       ,'P65'           , 'P 65'              ,  170, 'S' ,   &
         '3300133Xx1'              , '(0,0,0)'       ,'P62'           , 'P 62'              ,  171, 'S' ,   &
         '3300133Xx2'              , '(0,0,0)'       ,'P64'           , 'P 64'              ,  172, 'S' ,   &
         '3300123Xx1'              , '(0,0,0)'       ,'P63'           , 'P 6c'              ,  173, 'S' ,   &
         '3300101w70'              , '(0,0,0)'       ,'P-6'           , 'P -6'              ,  174, 'S' ,   &
         '3300200n{03Xx0'          , '(0,0,0)'       ,'P6/m'          , '-P 6'              ,  175, 'S' ,   &
         '3300220n{03Xx1'          , '(0,0,0)'       ,'P63/m'         , '-P 6c'             ,  176, 'S' ,   &
         '3300203Xx042p0'          , '(0,0,0)'       ,'P622'          , 'P 6 2'             ,  177, 'S' ,   &
         '3300263Xx0142p00'        , '(0,0,11/12)'   ,'P6122'         , 'P 61 2 (0 0 -1)'   ,  178, 'S' ,   &
         '3300263Xx0542p00'        , '(0,0,1/12)'    ,'P6522'         , 'P 65 2 ( 0 0 1)'   ,  179, 'S' ,   &
         '3300233Xx142p0'          , '(0,0,5/6)'     ,'P6222'         , 'P 62 2c (0 0 1)'   ,  180, 'S' ,   &
         '3300233Xx242p0'          , '(0,0,1/6)'     ,'P6422'         , 'P 64 2c (0 0 -1)'  ,  181, 'S' ,   &
         '3300223Xx142p0'          , '(0,0,3/4)'     ,'P6322'         , 'P 6c 2c'           ,  182, 'S' ,   &
         '3300203Xx042r0'          , '(0,0,0)'       ,'P6mm'          , 'P 6 -2'            ,  183, 'S' ,   &
         '3300223Xx042r1'          , '(0,0,0)'       ,'P6cc'          , 'P 6 -2c'           ,  184, 'S' /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=457,494)/ &
         '3300223Xx142r1'          , '(0,0,0)'       ,'P63cm'         , 'P 6c -2'           ,  185, 'S' ,   &
         '3300223Xx142r0'          , '(0,0,0)'       ,'P63mc'         , 'P 6c -2c'          ,  186, 'S' ,   &
         '3300201w7042r0'          , '(0,0,0)'       ,'P-6m2'         , 'P -6 2'            ,  187, 'S' ,   &
         '3300221w7042r1'          , '(0,0,3/4)'     ,'P-6c2'         , 'P -6c 2'           ,  188, 'S' ,   &
         '3300201w703St0'          , '(0,0,0)'       ,'P-62m'         , 'P -6 -2'           ,  189, 'S' ,   &
         '3300221w703St1'          , '(0,0,3/4)'     ,'P-62c'         , 'P -6c -2c'         ,  190, 'S' ,   &
         '3300300n{03Xx042r0'      , '(0,0,0)'       ,'P6/mmm'        , '-P 6 2'            ,  191, 'S' ,   &
         '3300320n{03Xx042r1'      , '(0,0,0)'       ,'P6/mcc'        , '-P 6 2c'           ,  192, 'S' ,   &
         '3300320n{03Xx142r1'      , '(0,0,0)'       ,'P63/mcm'       , '-P 6c 2'           ,  193, 'S' ,   &
         '3300320n{03Xx142r0'      , '(0,0,0)'       ,'P63/mmc'       , '-P 6c 2c'          ,  194, 'S' ,   &
         '3300202f002I80'          , '(0,0,0)'       ,'P23'           , 'P 2 2 3'           ,  195, 'S' ,   &
         '332220352f002I80'        , '(0,0,0)'       ,'F23'           , 'F 2 2 3'           ,  196, 'S' ,   &
         '33122072f002I80'         , '(0,0,0)'       ,'I23'           , 'I 2 2 3'           ,  197, 'S' ,   &
         '3300222f002I83'          , '(0,0,0)'       ,'P213'          , 'P 2ac 2ab 3'       ,  198, 'S' ,   &
         '33122272f002I81'         , '(3/4,3/4,3/4)' ,'I213'          , 'I 2b 2c 3'         ,  199, 'S' ,   &
         '3300202IE02ew0'          , '(0,0,0)'       ,'Pm-3'          , '-P 2 2 3'          ,  200, 'S' ,   &
         '3300222IE72ew7'          , '(0,0,0)'       ,'Pn-3:1'        , 'P 2 2 3 -1n'       ,  201, ''  ,   &
         '3300222IE72ew7'          , '(1/4,1/4,1/4)' ,'Pn-3'          , '-P 2ab 2bc 3'      ,  201, 'S' ,   &
         '332220352IE02ew0'        , '(0,0,0)'       ,'Fm-3'          , '-F 2 2 3'          ,  202, 'S' ,   &
         '332224352IEL2ewL'        , '(0,0,0)'       ,'Fd-3:1'        , 'F 2 2 3 -1d'       ,  203, 'S' ,   &
         '332224352IEL2ewL'        , '(5/8,1/8,1/8)' ,'Fd-3'          , '-F 2uv 2vw 3'      ,  203, 'S' ,   &
         '33122072IE02ew0'         , '(0,0,0)'       ,'Im-3'          , '-I 2 2 3'          ,  204, 'S' ,   &
         '3300222IE02ew3'          , '(0,1/2,1/2)'   ,'Pa-3'          , '-P 2ac 2ab 3'      ,  205, 'S' ,   &
         '3300222IE02ew6'          , '(0,0,1/2)'     ,'Pb-3'          , '-P 2bc 2ac 3'      ,  205, 'A' ,   &
         '33122272IE02ew1'         , '(1/4,1/4,3/4)' ,'Ia-3'          , '-I 2b 2c 3'        ,  206, 'S' ,   &
         '3300203}w02Fo0'          , '(0,0,0)'       ,'P432'          , 'P 4 2 3'           ,  207, 'S' ,   &
         '3300223}w72Fo7'          , '(0,0,0)'       ,'P4232'         , 'P 4n 2 3'          ,  208, 'S' ,   &
         '332220353}w02Fo0'        , '(0,0,0)'       ,'F432'          , 'F 4 2 3'           ,  209, 'S' ,   &
         '332224353}wL2FoL'        , '(0,0,0)'       ,'F4132'         , 'F 4d 2 3'          ,  210, 'S' ,   &
         '33122073}w02Fo0'         , '(0,0,0)'       ,'I432'          , 'I 4 2 3'           ,  211, 'S' ,   &
         '3300243}ws2Fos'          , '(5/8,1/8,3/8)' ,'P4332'         , 'P 4acd 2ab 3'      ,  212, 'S' ,   &
         '3300243}wM2Fol'          , '(3/8,1/8,7/8)' ,'P4132'         , 'P 4bd 2ab 3'       ,  213, 'S' ,   &
         '33122473}wH2Foy'         , '(3/8,5/8,1/8)' ,'I4132'         , 'I 4bd 2c 3'        ,  214, 'S' ,   &
         '3300201xO02DG0'          , '(0,0,0)'       ,'P-43m'         , 'P -4 2 3'          ,  215, 'S' ,   &
         '332220351xO02DG0'        , '(0,0,0)'       ,'F-43m'         , 'F -4 2 3'          ,  216, 'S' ,   &
         '33122071xO02DG0'         , '(0,0,0)'       ,'I-43m'         , 'I -4 2 3'          ,  217, 'S' ,   &
         '3300221xO72DG7'          , '(0,0,0)'       ,'P-43n'         , 'P -4n 2 3'         ,  218, 'S' ,   &
         '332222351xO42DG4'        , '(0,0,0)'       ,'F-43c'         , 'F -4c 2 3'         ,  219, 'S' /
       DATA (SGlist(sgi)%SGFP,SGlist(sgi)%Orshift, SGlist(sgi)%HM,SGlist(sgi)%Hall,SGlist(sgi)%SGNr,SGlist(sgi)%std,sgi=495,509)/ &
         '33122471xO62DGO'         , '(3/8,7/8,5/8)' ,'I-43d'         , 'I -4bd 2c 3'       ,  220, 'S' ,   &
         '3300202IE03}w0'          , '(0,0,0)'       ,'Pm-3m'         , '-P 4 2 3'          ,  221, 'S' ,   &
         '3300222IE73}w0'          , '(0,0,0)'       ,'Pn-3n:1'       , 'P 4 2 3 -1n'       ,  222, ''  ,   &
         '3300222IE73}w0'          , '(1/4,1/4,1/4)' ,'Pn-3n'         , '-P 4a 2bc 3'       ,  222, 'S' ,   &
         '3300222IE03}w7'          , '(0,0,0)'       ,'Pm-3n'         , '-P 4n 2 3'         ,  223, 'S' ,   &
         '3300222IE73}w7'          , '(0,0,0)'       ,'Pn-3m:1'       , 'P 4n 2 3 -1n'      ,  224, ''  ,   &
         '3300222IE73}w7'          , '(1/4,1/4,1/4)' ,'Pn-3m'         , '-P 4bc 2bc 3'      ,  224, 'S' ,   &
         '332220352IE03}w0'        , '(0,0,0)'       ,'Fm-3m'         , '-F 4 2 3'          ,  225, 'S' ,   &
         '332222352IE03}w4'        , '(0,0,0)'       ,'Fm-3c'         , '-F 4c 2 3'         ,  226, 'S' ,   &
         '332224352IEL3}wL'        , '(0,0,0)'       ,'Fd-3m:1'       , 'F 4d 2 3 -1d'      ,  227, ''  ,   &
         '332224352IEL3}wL'        , '(1/8,5/8,5/8)' ,'Fd-3m'         , '-F 4vw 2vw 3'      ,  227, 'S' ,   &
         '332224352IEL3}wr'        , '(3/4,1/4,3/4)' ,'Fd-3c:1'       , 'F 4d 2 3 -1cd'     ,  228, ''  ,   &
         '332224352IEL3}wr'        , '(5/8,1/8,1/8)' ,'Fd-3c'         , '-F 4cvw 2vw 3'     ,  228, 'S' ,   &
         '33122072IE03}w0'         , '(0,0,0)'       ,'Im-3m'         , '-I 4 2 3'          ,  229, 'S' ,   &
         '33122472IE03}wL'         , '(3/4,3/4,1/4)' ,'Ia-3d'         , '-I 4bd 2c 3'       ,  230, 'S' /

!********************************
      !Interfaces
!********************************
      INTERFACE

      FUNCTION NextCoo(Dim,StartC,EndC,Coo,AxOrd)
        IMPLICIT NONE
        LOGICAL :: NextCoo
        INTEGER :: Dim,i
        INTEGER, DIMENSION(Dim) :: StartC,EndC,Coo,AxOrd
      END FUNCTION NextCoo
      
      SUBROUTINE RemoveSpaces(ioText)
        implicit none
        character*(*) :: ioText
      END SUBROUTINE RemoveSpaces

      SUBROUTINE DecToFraction(InDec,Nom,Den)
        REAL :: InDec
	INTEGER :: Nom,Den
      END SUBROUTINE DecToFraction
      
      END INTERFACE

      contains

!*********************************************************************
      !General tools for handling symmetry operations and space groups
!*********************************************************************

!*****************************************
        SUBROUTINE ReallocateSG(SG,SDim,NSym,NSymNew)
	  integer :: SDim
	  type(SymOp_Type), dimension(:), allocatable :: SG
	  integer :: NSym, NSymNew
	  type(SymOp_Type), dimension(:), allocatable :: TmpSG
	  integer :: l1

	  IF (NSymNew==0) THEN !Only deallocate SG
  	    CALL DeallocateSG(SG)
	  ELSEIF (NSym==0) THEN !Only allocate new SG
	    CALL AllocateSG(SG, SDim, NSymNew)
	  ELSE
  	    call AllocateSG(TmpSG,SDim,NSym)
  	    call CopySG(SG,TmpSG)
  	    CALL DeallocateSG(SG)
  	    CALL AllocateSG(SG,SDim, NSymNew)
	    CALL CopySG(TmpSG, SG)
            CALL DeallocateSG(TmpSG)
	  ENDIF
	END SUBROUTINE ReallocateSG

!*****************************************
        SUBROUTINE AllocateSG(SG, SDim, NSym)
	  type(SymOp_type), allocatable, dimension(:) :: SG
	  integer :: SDim, NSym, i
	  IF (Allocated(SG)) CALL DeallocateSG(SG)
	  ALLOCATE(SG(1:NSym))
	  DO i=1,NSym
	    ALLOCATE(SG(i)%R(1:SDim,1:SDim))
	    ALLOCATE(SG(i)%t(1:SDim))
	  ENDDO
	END SUBROUTINE AllocateSG

!*****************************************
        SUBROUTINE DeallocateSG(SG)
	  type(SymOp_type), allocatable, dimension(:) :: SG
	  integer :: i
	  IF (.not.Allocated(SG)) RETURN
	  DO i=1,size(SG)
	    DEALLOCATE(SG(i)%R)
	    DEALLOCATE(SG(i)%t)
	  ENDDO
	  DEALLOCATE(SG)
	END SUBROUTINE DeallocateSG

!*****************************************
        SUBROUTINE CopySG(SG1,SG2)
	  type(SymOp_type), dimension(:) :: SG1,SG2
	  integer :: NSym1,NSym2,i
	  NSym1=size(SG1)
	  NSym2=size(SG2)
	  DO i=1,min(NSym1,NSym2)
            CALL CopySymOp(SG1(i),SG2(i))
	  ENDDO
	END SUBROUTINE CopySG

!*****************************************
        SUBROUTINE AllocateSymProps(SymProps, SDim)
	  type(SymOpPropsType) :: SymProps
	  integer :: SDim
	  ALLOCATE(SymProps%dir(SDim),SymProps%IntrVec(SDim))
	END SUBROUTINE AllocateSymProps
	
!*****************************************
        SUBROUTINE DeallocateSymProps(SymProps)
	  type(SymOpPropsType) :: SymProps
          DEALLOCATE(SymProps%dir)
          DEALLOCATE(SymProps%IntrVec)
	END SUBROUTINE DeallocateSymProps

!*****************************************
       subroutine printsymop(symop,form)
         integer :: i,ii,dim
         type(SymOp_type) :: symop
         CHARACTER(len = *), optional :: form       
         CHARACTER(len = 15) :: prnform
         dim=size(SymOp%R,1)
         write(prnform,'(i2)') dim
         IF (present(form)) THEN
	   prnform='('//trim(prnform)//trim(form)//')'
         ELSE
           prnform='('//trim(prnform)//'i3,3X,f6.3)'
         ENDIF
         DO i=1,dim
           write(*,prnform) (SymOp%R(i,ii),ii=1,dim), SymOp%t(i)
         ENDDO
         write(*,*) 
       end subroutine printsymop

!*****************************************
       CHARACTER*256 FUNCTION SymOpIntTab(InSmOp,axcode)
  	 IMPLICIT NONE
    	 TYPE(SymOp_type) :: InSmop
    	 INTEGER :: soiti,soitii,tnom,tden,SDim
  	 CHARACTER*132 :: sympart
  	 CHARACTER*(*) :: axcode
  	 CHARACTER*9 :: axnames='xyztuvqrs'
	 SDim=size(InSmOp%t)
  	 SymOpIntTab=''
  	 DO soiti=1,SDim
    	   sympart=''
  	   IF (abs(InSmOp%t(soiti))>eps) THEN
  	     CALL DecToFraction(InSmOp%t(soiti),tnom,tden)
  	     IF (tden<10) THEN
  	 	write(sympart,'(i1,a,i1)') tnom,'/',tden	
  	     ELSE	      
  	 	write(sympart,'(f6.3)')InSmOp%t(soiti)
    	     ENDIF
  	   ENDIF
  	   DO soitii=1,SDim
  	     IF (InSmOp%R(soiti,soitii)/=0) THEN
    	       IF (InSmOp%R(soiti,soitii)<0) THEN
    	         WRITE(sympart(len_trim(sympart)+1:),'(a1)') '-'
  	       ELSEIF(len_trim(sympart)>0) THEN
  	 	 WRITE(sympart(len_trim(sympart)+1:),'(a1)') '+'
  	       ENDIF
    	       IF (axcode=='x123') THEN
  	 	 write(sympart(len_trim(sympart)+1:),'(a1,i1)') 'x',soitii
  	       ELSEIF (axcode(1:3)=='xyz') THEN
  	 	 write(sympart(len_trim(sympart)+1:),'(a1)') axnames(soitii:soitii)
  	       ENDIF
  	     ENDIF
  	   ENDDO
    	   IF (len_trim(sympart)<13) THEN
    	     write(SymOpIntTab(len_trim(SymOpIntTab)+14-len_trim(sympart):),'(a)') trim(sympart)
    	   ELSE
    	     SymOpIntTab=trim(SymOpIntTab)//' '//sympart
  	   ENDIF
  	   IF ((axcode(1:3)=='xyz').and.(soiti<SDim)) SymOpIntTab(len_trim(SymOpIntTab)+1:)=','
  	 ENDDO
       END FUNCTION SymOpIntTab

! Multiplication of two symmetry operations S3=R2*R1
!*****************************************
      SUBROUTINE MultSymOp(SymOp2,SymOp1,SymOp3)
        TYPE(SymOp_type) :: SymOp2, SymOp1, SymOp3
        SymOp3%R=multmat(SymOp2%R,SymOp1%R)
        SymOp3%T=SymOp2%T+multmat(SymOp2%R,SymOp1%T)
        SymOp3%T=From0to1(SymOp3%T)
      END SUBROUTINE MultSymOp
!      END FUNCTION MultSymOp

! Copy SymOp1 to SymOp2
!*****************************************
      SUBROUTINE CopySymOp(SymOp1,SymOp2)
        TYPE(SymOp_type) :: SymOp2, SymOp1
        SymOp2%R=SymOp1%R
	SymOp2%t=SymOp1%t
      END SUBROUTINE CopySymOp

!Function EqualOp is true if the two input operations are equivalent
!*******************************************************************
        LOGICAL FUNCTION EqualOp(R1,T1,R2,T2,n)
        real, parameter :: verybigeps=0.05
        integer :: i,n
        integer, dimension(1:n,1:n) :: R1,R2
        real, dimension(1:n) :: T1,T2
        logical :: equal
      
        equal=.true.
        IF (any(R1/=R2)) equal=.false.
        DO i=1,n
          IF (abs((abs(T1(i)-T2(i))-abs(float(nint(T1(i)-T2(i))))))>verybigeps) THEN
	    equal=.false.
	    EXIT
          ENDIF
	  IF (.not.equal) EXIT
        ENDDO
        EqualOp=equal
        END FUNCTION EqualOp

        LOGICAL FUNCTION SymOpsJustShifted(RMat1,TVec1,RMat2,TVec2,Center)
	IMPLICIT NONE
        INTEGER, DIMENSION(:,:) :: RMat1,RMat2
        REAL, DIMENSION(:) :: TVec1,TVec2
	REAL, DIMENSION(:,:) :: Center
	REAL, DIMENSION(1:size(TVec1)) :: sjsdvec,sjsdvec2, tmptvec
        INTEGER, DIMENSION(1:size(TVec1)) :: sjsstartc,sjsendc,sjscoo, axord
        INTEGER :: l2,ROrder,SDim,NCenter
        REAL, DIMENSION(1:size(TVec1),1:size(TVec1)) :: TestR
	NCenter=size(Center,2)
	SDim=size(TVec1)
      
        axord=(/(l2,l2=1,SDim)/)
        SymOpsJustShifted=.false.
        IF (any(RMat1/=RMat2)) RETURN
        !We get here only if the two matrices are identical
        !Check if also the TVec is identical or centering-related
	symopsjustshifted=.false.
        DO l2=1,NCenter
          tmptvec=TVec1+Center(1:SDim,l2)
	  sjsdvec=tmptvec-TVec2
	  if (maxval(from0to1(sjsdvec))<bigeps) THEN
	    SymOpsJustShifted=.true.
	    EXIT
	  ENDIF
          CALL ProjectionOperator(float(Rmat1),TestR,ROrder,SDim)
          sjsstartc=0          !\
	  sjsendc=1	       !- This loop is needed for diagonal elements, where lattice translations can transform intrinsic vector to an origin-dependent one
	  sjscoo=sjsstartc     !/ Assumption here is that al centering vectors are in interval <0;1)
	  DO
            sjsdvec2=multmat(TestR,sjsdvec-sjscoo)
	    if (maxval(from0to1(sjsdvec2))<bigeps) THEN
	      SymOpsJustShifted=.true.
	      EXIT
	    ENDIF
	    IF (.not.NextCoo(SDim,sjsstartc,sjsendc,sjscoo,axord)) EXIT
	  ENDDO
	  IF (SymOpsJustShifted) EXIT
	ENDDO
	END FUNCTION SymOpsJustShifted

!****************************************
      CHARACTER*48 FUNCTION SymOpSymbol(InSmOp,SymProps,RDim, Form)
	IMPLICIT NONE
	TYPE(SymOp_type) :: InSmOp
	TYPE (SymOpPropsType) :: SymProps
	REAL :: rdummy
        INTEGER :: sosi,sosii,sosiii,Form !FORM=1 - LONG SYMBOL, FORM=2 - SHORT SYMBOL
	INTEGER :: SDim, RDim
	CHARACTER*10 :: direction,internpart
	SDim=size(InSmOp%t)
	direction=' '
        IF (RDim/=3) THEN
	  SymOpSymbol=' '
        ELSE
	  IF ((SymProps%rank==3).and.(SymProps%det==-1).and.(SymProps%Order==2)) THEN
	    SymOpSymbol=' -1'
	  ELSE
	    write(SymOpSymbol,'(i3)') SymProps%det*SymProps%Order
          ENDIF
	  IF ((SymOpSymbol==' -1').or.(SymOpSymbol=='  1')) RETURN
	  
          if (SymOpSymbol==' -2') SymOpSymbol='  m'
          if ((SymOpSymbol==' -6').and.(SymProps%trace==0)) SymOpSymbol=' -3'

          IF (Form==2) THEN
            write(direction,'(''('',i2,'','',i2,'','',i2,'')'')') (SymProps%dir(sosi),sosi=1,RDim)
            !remove the spaces
            DO sosi=1,len(direction)-1
              if (direction(sosi:sosi)==' ') direction(sosi:)=direction(sosi+1:)
            ENDDO
	  ENDIF
	  !now change the symbol, if intrinsic shift is non-zero
	  IF (SymProps%IntrOrder>1) THEN
	    IF (SymProps%rank==2) THEN !This is a rotation axis
	      !Usually only one intrinsic component, but for diagonal screws in hexagonal axes more than one different component possible 
	      rdummy=1000000.
	      DO sosi=1,RDim
                if (abs(SymProps%IntrVec(sosi))>bigeps) THEN
                  IF (((SymProps%IntrVec(sosi)>0.).and.(SymProps%IntrVec(sosi) <abs(rdummy))).or. &
                     ((SymProps%IntrVec(sosi)<0.).and.(abs(SymProps%IntrVec(sosi))<abs(rdummy)))) rdummy=abs(SymProps%IntrVec(sosi))
                ENDIF
	      ENDDO
	      sosii=nint(float(SymProps%order)*rdummy)
              IF (Form==2) THEN
                IF (SymProps%RotSense==1) THEN
                  write(SymOpSymbol(4:),'(''+_'',i1)') sosii
	        ELSEIF(SymProps%RotSense==-1) THEN
                  write(SymOpSymbol(4:),'(''-_'',i1)') sosii
	        ELSE
	          write(SymOpSymbol(4:),'(''_'',i1)') sosii
	        ENDIF
	      ELSE
                write(SymOpSymbol(4:),'(i1)') sosii
	      ENDIF
	    ELSEIF (SymProps%rank==1) THEN !This is a mirror plane
	      !Count the number of non-zero components
              sosii=0
	      DO sosi=1,RDim
	        if (abs(SymProps%IntrVec(sosi))>bigeps) sosii=sosii+1
	      ENDDO
	      SELECT CASE (sosii)
	      CASE (1)
	        IF (SymProps%IntrOrder==2) THEN
		  IF (abs(SymProps%IntrVec(1))>bigeps) THEN
		    SymOpSymbol='a'
	          ELSEIF (abs(SymProps%IntrVec(2))>bigeps) THEN
		    SymOpSymbol='b'
	          ELSEIF (abs(SymProps%IntrVec(3))>bigeps) THEN
		    SymOpSymbol='c'
	          ENDIF
	        ELSE
                  SymOpSymbol='?'
		ENDIF
	      CASE DEFAULT !Warning - this inorrectly (but more logically) assigns "n" to the c-glide in rhombohedral axes
	        IF (SymProps%IntrOrder==2) THEN
                  SymOpSymbol='n'
	        ELSEIF (SymProps%IntrOrder==6) THEN
                  SymOpSymbol='g'
		ELSEIF (SymProps%IntrOrder==4) THEN
                  sosiii=0
		  rdummy=0.0
		  DO sosi=1,RDim
                    sosiii=sosiii+abs(SymProps%dir(sosi))
		     rdummy=rdummy+abs(SymProps%IntrVec(sosi))
                  ENDDO
!		  write(*,*) sosiii,SymProps%dir
!		  write(*,*) rdummy,SymProps%IntrVec
		  IF ((sosiii==1).or.(((sosii)==3).and.(sosiii==2).and.(abs(rdummy-nint(rdummy))>bigeps))) THEN
                    SymOpSymbol='d'
		  ELSEIF (sosiii==2) THEN
                    SymOpSymbol='g'
		  ENDIF
		ELSE
                  SymOpSymbol='?'		
		ENDIF
              END SELECT
	    ENDIF
	  ENDIF
	  IF ((RDim==3).and.(SDim==4))THEN !Additional symbol for the internal dimension
	    IF ((InSmOp%R(SDim,SDim)==1).and.(InSmOp%t(SDim)>bigeps)) THEN !Non-zero intrinsic shift in internal dimension
	      sosii=nint(1./InSmOp%t(SDim))
	      SELECT CASE (sosii)
	        CASE (2)
                  internpart='s'
	        CASE (3)
                  internpart='t'
	        CASE (4)
                  internpart='q'
	        CASE (6)
                  internpart='h'
	        CASE DEFAULT
                  internpart='?'
	      END SELECT
	    ELSE
              internpart='0'
	    ENDIF
            SymOpSymbol=trim(SymOpSymbol)//'|'//internpart
	  ENDIF
          IF (Form==2) SymOpSymbol=trim(SymOpSymbol)//direction
	ENDIF
	END FUNCTION SymOpSymbol

!****************************************
      SUBROUTINE FillSymProps(InSmOp,SymProps,RDim)
	IMPLICIT NONE
	REAL, PARAMETER :: verybigeps=0.03
	TYPE(SymOp_type) :: InSmOp
	INTEGER :: RDim, SDim
	REAL, DIMENSION(1:RDim,1:RDim) :: Matrix,DMatrix,DMatrix2
	REAL, DIMENSION(1:size(InSmOp%t),1:size(InSmOp%t)) :: SDMatrix
	INTEGER, DIMENSION(1:size(InSmOp%t)) :: StartC,EndC,Coo,AxOrd
	REAL, DIMENSION(size(InSmOp%t)) :: dvec,dvec2,dvec3,randvec1
	REAL :: rdummy
        INTEGER :: sosi,sosii,tmprank
	LOGICAL :: different, NextCoo
	TYPE (SymOpPropsType) :: SymProps
        SDim=size(InSmOp%t)
        DO sosi=1,RDim
!	  randvec1(sosi)=(My_Rand()+My_Rand()+My_Rand())/3. !To avoid zero components with extremely high probablity
          randvec1(sosi)=0.1*sqrt(float(sosi)+0.5) !Totally random vectors are not good, since sometimes they produces accidental coincidences
	ENDDO

	
        IF (SDim>RDim) THEN
	  DO sosi=1,RDim
	    DO sosii=1,RDim
	      Matrix(sosi,sosii)=InSmOp%R(sosi,sosii)
	    ENDDO
	  ENDDO
	ELSE
	  Matrix=InSmOp%R
	ENDIF
	!Get the dimension of the invariant subspace of the matrix
        DMatrix=Matrix
	SymProps%trace=0
	DO sosi=1,RDim
	  DMatrix(sosi,sosi)=DMatrix(sosi,sosi)-1
          SymProps%trace=SymProps%trace+nint(Matrix(sosi,sosi))
	ENDDO
        CALL GetRank(DMatrix,RDim,RDim,SymProps%rank)
	!Get the order of the symop
        DMatrix=0.
        DO sosi=1,RDim
          DMatrix(sosi,sosi)=1.
        ENDDO
	SymProps%order=0
	DO
	  DMatrix2(1:RDim,1:RDim)=multmat(DMatrix(1:RDim,1:RDim),Matrix(1:RDim,1:RDim))
	  SymProps%order=SymProps%order+1            
	  IF (IsUnitMatrix(DMatrix2)) EXIT
	  DMatrix=DMatrix2
	ENDDO
        
	!Get the determinant of the matrix
	SymProps%det=nint(Determinant(Matrix,eps))
	  
        !Get the projection operator (attention, here and only here we use SDim instead of RDim)
        CALL ProjectionOperator(float(InSmOp%R),SDMatrix,sosi,SDim)
        !Extract the intrinsic part of the translation vector
        SymProps%IntrVec=multmat(SDMatrix,InSmOp%t)
        IF ((ANY(abs(SDMatrix-nint(SDMatrix))>eps)).and.(count(abs(SymProps%IntrVec(1:SDim))>eps)>0)) THEN !Enter this loop only if the projection operator contains fractional numbers    
          tmprank=count(abs(SymProps%IntrVec(1:SDim))>eps)
!          do sosi=1,SDim
!	    write(*,*) abs(SymProps%IntrVec(sosi)),eps,abs(SymProps%IntrVec(sosi))>eps
!	  enddo
          dvec3=SymProps%IntrVec
	  StartC=-SymProps%order+1
	  EndC=SymProps%Order-1
	  Axord=(/(sosi,sosi=1,SDim)/)
	  Coo=StartC
!	  write(*,*) StartC
!	  write(*,*) SDMatrix
!  	  write(*,*) 'pred:',(SymProps%IntrVec(sosi),sosi=1,SDim)         
	  DO
!            write(*,*) Coo
            dvec=multmat(SDMatrix,float(Coo))
            dvec2=0.0
	    DO sosi=1,SDim
	      IF (abs(dvec(sosi))>eps) THEN
	        dvec2(sosi)=SymProps%IntrVec(sosi)-dvec(sosi)
	      ELSE
                dvec2(sosi)=SymProps%IntrVec(sosi)
	      ENDIF
	    ENDDO
!            write(*,*) (dvec(sosi),sosi=1,sdim)
!            write(*,*) (dvec2(sosi),sosi=1,sdim)
	    sosi=count(abs(dvec2)>eps)
!	    write(*,*) 'pocet nenulovych:',sosi,tmprank,all(abs(dvec2)<=abs(dvec3)+eps)
	    IF ((sosi<tmprank).and.(all(abs(dvec2)<=abs(dvec3)+eps))) THEN
              dvec3=dvec2
              tmprank=sosi
!              write(*,*) 'beru!'
	    ENDIF
	    IF (tmprank==0) EXIT
	    IF (.not.NextCoo(SDim,StartC,EndC,Coo,AxOrd)) EXIT            
	  ENDDO
          SymProps%IntrVec(1:SDim)=dvec3(1:SDim)
!	  write(*,*) 'po:',(SymProps%IntrVec(sosi),sosi=1,SDim)         
        ENDIF
        dvec=0.;SymProps%IntrOrder=0
	DO
	  dvec=dvec+SymProps%IntrVec
	  SymProps%IntrOrder=SymProps%IntrOrder+1
	  different=.false.
	  DO sosi=1,RDim
	    IF (abs(dvec(sosi)-float(nint(dvec(sosi))))>verybigeps) THEN
	      different=.true.
	      EXIT
	    ENDIF	      
	  ENDDO
	  IF (.not.different) EXIT
	ENDDO

        !Get the direction of the symop
        IF (SymProps%det==-1) THEN !Check the rank of the inversed matrix
          DMatrix=-Matrix
          DO sosi=1,RDim
	    DMatrix(sosi,sosi)=DMatrix(sosi,sosi)-1
	  ENDDO
          CALL GetRank(DMatrix,RDim,RDim,tmprank)
        ELSE
	  tmprank=SymProps%rank
	ENDIF
!        IF ((SymProps%rank==RDim-1).or.(SymProps%rank==1)) THEN !Only in these cases the direction makes sense
        IF ((tmprank==RDim-1).or.(tmprank==1)) THEN !Only in these cases the direction makes sense
  	  IF (SymProps%det==-1) THEN !Get the projection operator of the inversed matrix -> get the axis instead of rotoinverison
            CALL ProjectionOperator(-Matrix,DMatrix,sosi,RDim)
	  ELSE
            CALL ProjectionOperator(Matrix,DMatrix,sosi,RDim)
          ENDIF
          dvec(1:RDim)=multmat(DMatrix,randvec1(1:RDim))
	  !Normalize the vector by the component with the smallest absolute value
          rdummy=10000.
          DO sosi=1,RDim
            if (abs(dvec(sosi))>bigeps) THEN
              IF (((dvec(sosi)>0.).and.(    dvec(sosi) <abs(rdummy))).or. &
                 ((dvec(sosi)<0.).and.(abs(dvec(sosi))<abs(rdummy)))) rdummy=dvec(sosi)
            ENDIF
          ENDDO
          SymProps%dir=nint(dvec/rdummy)
	  sosii=1
	  SymProps%dirscore=0
	  DO sosi=1,RDim
            SymProps%dirscore=SymProps%dirscore+sosii*SymProps%dir(sosi)
	    sosii=sosii*10
	  ENDDO
        ELSE
	  SymProps%dir=0
	ENDIF

        IF ((SymProps%Order>2).and.(SymProps%IntrOrder>2).and.(SymProps%det==1).and.(RDim==3)) THEN !Screw that needs determination of sign, only for 3D
          dvec(1:RDim)=multmat(Matrix,randvec1(1:RDim))
          DO sosi=1,RDim
	    DMatrix2(1,sosi)=randvec1(sosi)
	    DMatrix2(2,sosi)=dvec(sosi)
	    DMatrix2(3,sosi)=SymProps%IntrVec(sosi)
	  ENDDO
	  rdummy=Determinant(DMatrix2,eps)
	  IF (rdummy>bigeps) THEN
	    SymProps%rotsense=1
	  ELSEIF (rdummy<bigeps) THEN
	    SymProps%rotsense=-1
	  ELSE
	    SymProps%rotsense=0
	  ENDIF
	ELSE
	  SymProps%rotsense=0
	ENDIF

	END SUBROUTINE FillSymProps

!*********************************************************************
        SUBROUTINE TentativeHMSymbol(SG,Center,QVec,RDim,SGsymbol)
  	  type(SymOp_type), dimension(:) :: SG
          real, dimension(:,:) :: Center, QVec
	  integer :: SDim, RDim, NSym, NCenter
	  character*(*) :: SGSymbol

    	  integer :: i,ii,iii,iv
    	  INTEGER :: CrSystem
    	  INTEGER, DIMENSION(1:6) :: OrderSum
    	  type(SymOpPropsType),allocatable, dimension(:) ::SymProps
    	  REAL, DIMENSION(1:RDim) :: qr, qi
    	  CHARACTER*48 :: SGPart1,SGPart2
          CHARACTER*1  :: CenSymbol
    	  CHARACTER*3 :: AxStr
	  CHARACTER*132 :: Buffer
    	  INTEGER, DIMENSION(1:4,1:6) :: & !Array holding the special directions, the fourth element is a unique number of that direction calculated as a+10*b+100*c
    	    SpDir = RESHAPE ( &
    	     (/ 1, 0, 0, 1,   &
    		0, 1, 0, 10,  & 
    		0, 0, 1, 100, & 
    		1, 1, 0, 11,  & 
    		2, 1, 0, 12,  & 
    		1, 1, 1, 111  & 
    	      /),(/4,6/))
	  SDim=size(Center,1)
	  NSym=size(SG)
	  NCenter=size(center,2)

          ALLOCATE(SymProps(1:NSym))
	  DO i=1,NSym
            CALL AllocateSymProps(SymProps(i),SDim)
            CALL FillSymProps(SG(i),SymProps(i),RDim)
          ENDDO

	  SGSymbol=''
     	  IF (((SDim-RDim)>1).or.(RDim/=3)) RETURN !No space group symbol for exotic dimensions
     	  CALL FindCenteringSymbol(Center,CenSymbol)
     	  OrderSum=0
     	  AxStr='000'
     	  DO i=1,NSym
     	    OrderSum(SymProps(i)%order)=OrderSum(SymProps(i)%order)+1
     	    ii=SymProps(i)%Dir(1)+10*SymProps(i)%Dir(2)+100*SymProps(i)%Dir(3)
     	    DO iii=1,3
     	      if (ii==SpDir(4,iii)) AxStr(iii:iii)='1'
     	    ENDDO
     	  ENDDO
     	  IF (NSym==1) THEN
     	    CrSystem=CrSysTriclinic
     	  ELSEIF ((NSym==2).and.((SymProps(1)%rank==3).or.(SymProps(2)%rank==3))) THEN
     	    CrSystem=CrSysTriclinic
     	  ELSEIF (OrderSum(3)>=4) THEN
     	    CrSystem=CrSysCubic
     	  ELSEIF ((OrderSum(6)>0).or.(OrderSum(3)>0)) THEN
     	    CrSystem=CrSysHexagonal  !Includes hexagonal and trigonal
!    	   ELSEIF (OrderSum(3)>0) THEN
!    	     CrSystem=CrSysTrigonal
     	  ELSEIF (OrderSum(4)>0) THEN
     	    CrSystem=CrSysTetragonal
     	  ELSEIF (AxStr=='111') THEN
     	    CrSystem=CrSysOrtorhombic
     	  ELSE
     	    CrSystem=CrSysMonoclinic 
     	  ENDIF
     	  SGSymbol=''
     	  SELECT CASE(CrSystem)
     	    CASE(CrSysTriclinic)
     	      IF (NSym==2) THEN
     		SGSymbol='-1'
     	      ELSE
     		SGSymbol='1'		
     	      ENDIF
     	    CASE(CrSysMonoclinic)
     	      DO ii=1,3
     		CALL AxOperators(SGPart1,SGPArt2,ii)
     		IF ((len_trim(SGPart1)>0).and.(len_trim(SGPart2)>0)) THEN
     		  SGSymbol=trim(SGSymbol)//trim(SGPart1)//'/'//trim(SGPart2)
     		ELSEIF ((len_trim(SGPart1)>0).or.(len_trim(SGPart2)>0)) THEN
     		  SGSymbol=trim(SGSymbol)//trim(SGPart1)//trim(SGPart2)
     		ELSE
     		  SGSymbol=trim(SGSymbol)//'1'  	
     		ENDIF
     	      ENDDO
     	    CASE(CrSysOrtorhombic)
     	      DO ii=1,3
     		CALL AxOperators(SGPart1,SGPArt2,ii)
     		IF ((len_trim(SGPart1)>0).and.(len_trim(SGPart2)>0)) THEN
     		  SGSymbol=trim(SGSymbol)//trim(SGPart2)
     		ELSE
     		  SGSymbol=trim(SGSymbol)//trim(SGPart1)//trim(SGPart2) 	
     		ENDIF
     	      ENDDO
     	    CASE(CrSysTetragonal)
     	      !First find the appropriate symbol for the special direction
     	      CALL AxOperators(SGPart1,SGPart2,3)
     	      IF (len_trim(SGPart2)/=0) THEN
     		SGSymbol=trim(SGPart1)//'/'//trim(SGPart2)
     	      ELSE
     		SGSymbol=trim(SGPart1)
     	      ENDIF
     	      !Now the axial operators
     	      DO i=1,4,3
     		CALL AxOperators(SGPart1,SGPart2,i)
     		IF ((len_trim(SGPart1)>0).and.(len_trim(SGPart2)>0)) THEN
     		  SGSymbol=trim(SGSymbol)//trim(SGPart2)
     		ELSEIF ((len_trim(SGPart1)>0).or.(len_trim(SGPart2)>0)) THEN
     		  SGSymbol=trim(SGSymbol)//trim(SGPart1)//trim(SGPart2) 	
     		ENDIF
     	      ENDDO	     
     	    CASE(CrSysHexagonal)
     	      !First find the appropriate symbol for the special direction
     	      CALL AxOperators(SGPart1,SGPart2,3)
     	      IF ((len_trim(SGPart2)/=0).and.(index(SGPart1,'-')==0)) THEN
     		SGSymbol=trim(SGPart1)//'/'//trim(SGPart2)
     	      ELSE
     		SGSymbol=trim(SGPart1)
     	      ENDIF
     	      !Now the axial operators of 100 and 210
     	      DO i=1,5,4
     		CALL AxOperators(SGPart1,SGPart2,i)
     		Buffer=adjustl(SGSymbol)
     		IF ((len_trim(SGPart1)>0).and.(len_trim(SGPart2)>0)) THEN
     		  SGSymbol=trim(SGSymbol)//trim(SGPart2)
     		ELSEIF ((len_trim(SGPart1)>0).or.(len_trim(SGPart2)>0)) THEN
     		  SGSymbol=trim(SGSymbol)//trim(SGPart1)//trim(SGPart2) 	
     		ELSEIF((CenSymbol/='R').and.((Buffer(1:1)=='3').or.(Buffer(1:2)=='-3'))) THEN
     		  SGSymbol=trim(SGSymbol)//'1'
     		ENDIF
     	      ENDDO	     
     	    CASE(CrSysCubic)
     	      !First the 100 direction
     	      call AxOperators(SGPart1,SGPart2,3)
     	      IF (len_trim(SGPart2)/=0) THEN
     		SGSymbol=SGPart2
     	      ELSE
     		SGSymbol=SGPart1
     	      ENDIF
     	      !then the body diagonal direction - quite simple
     	      IF (OrderSum(6)/=0) THEN
     		SGSymbol=trim(SGSymbol)//'-3'
     	      ELSE
     		SGSymbol=trim(SGSymbol)//'3'
     	      ENDIF
     	      !And now the face diagonal
     	      call AxOperators(SGPart1,SGPart2,4)
     	      IF (len_trim(SGPart2)/=0) THEN
     		SGSymbol=trim(SGSymbol)//SGPart2
     	      ELSE
     		SGSymbol=trim(SGSymbol)//SGPart1
     	      ENDIF
     	  END SELECT
     	  SGSymbol=CenSymbol//SGSymbol
     	  !remove the spaces
     	  i=0
     	  DO
     	    i=i+1
     	    if (SGSymbol(i:i)==' ') THEN
     	      SGSymbol(i:)=SGSymbol(i+1:)
     	      i=i-1
     	    endif
     	    if (i>=len_trim(SGSymbol)) EXIT
     	  ENDDO
     	  IF ((SDim==4).and.(RDim==3)) THEN !Modify the symbol to "normal" form for (3+1)D group
     	    i=0
     	    SGPart1=''
     	    SGPart2=''
     	    DO
     	      i=i+1
     	      IF (SGSymbol(i:i)=='|') THEN
     		SGPart1=trim(SGPart1)//SGSymbol(i+1:i+1)
     		SGSymbol(i:)=SGSymbol(i+2:)
     		i=i-1
     	      ENDIF
     	      IF (i==len_trim(SGSymbol)) EXIT
     	    ENDDO
     	    !Create the symbol of the q-vector
     	    !first extract qi:
     	    qi=0.0
     	    DO i=1,NSym
     	      qr=multmat(SG(i)%R(1:RDim,1:RDim),QVec(1:RDim,1))
     	      qi=qi+SG(i)%R(SDim,SDim)*qr
     	    ENDDO
     	    qi=qi/NSym
     	    qr=QVec(1:RDim,1)-qi
     	    !dvector holds q^i, svector q^r
     	    DO i=1,RDim
     	      IF (abs(qi(i))>bigeps) THEN !non-zero incommensurate part
     		SELECT CASE (i)
     		  CASE (1)
     		    SGPart2=trim(SGPart2)//'a'
     		  CASE (2)
     		    SGPart2=trim(SGPart2)//'b'
     		  CASE (3)
     		    SGPart2=trim(SGPart2)//'g'
     		END SELECT
     	      ELSEIF (abs(qr(i))>bigeps) THEN !non-zero rational part
     		SGPart2=trim(SGPart2)//'1/'
     		write(SGPart2(len_trim(SGPart2)+1:),'(i1)') abs(nint(1./qr(i)))
     	      ELSE
     	       SGPart2=trim(SGPart2)//'0'
     	      ENDIF
     	    ENDDO
     	    SGSymbol=trim(SGSymbol)//'('//trim(SGPart2)//')'//SGPart1
     	  ENDIF
     	  CALL StandardizeSGSymbol(SGSymbol) !Change nonstandard symbols to standard

	  
	  DO i=1,NSym
	    CALL DeallocateSymProps(SymProps(i))
          ENDDO
	  DEALLOCATE(SymProps)

	CONTAINS
          SUBROUTINE AxOperators(Part1,Part2,dir)
     	  IMPLICIT NONE
          INTEGER :: aoi,dir,maxorder
     	  CHARACTER*48 :: Part1,Part2
     	  Part1=''
     	  Part2=''
     	  maxorder=0
          DO aoi=1,NSym
            IF (SymProps(aoi)%dirscore/=SpDir(4,dir)) CYCLE !Take only symops along the current axis
            IF ((((SymProps(aoi)%det==1).and.(SymProps(aoi)%rank==2)).or.& !This selects rotation axes
     	  	(SymProps(aoi)%order>2)).and.(SymProps(aoi)%order>=maxorder)) THEN !This takes into account also rotoinversion axes -3, -4, -6
     	      maxorder=SymProps(aoi)%order
     	      IF ((Part1(2:3)/=' 6').and.(Part1(2:3)/=' 4').and.(SymProps(aoi)%rotsense/=-1)) THEN !Protection against replacement of 6 by -6 and 4 by -4, and against taking the negative screw axis
     	  	Part1=SymOpSymbol(SG(aoi),SymProps(aoi),RDim,1)
              ENDIF
     	    ELSEIF (SymProps(aoi)%rank==1) THEN
              Part2=SymOpSymbol(SG(aoi),SymProps(aoi),RDim,1)
            ENDIF
          ENDDO
          END SUBROUTINE AxOperators

          SUBROUTINE StandardizeSGSymbol(Symbol)
          CHARACTER*(*) Symbol
          IF(Symbol=='Fd-3g') Symbol='Fd-3m'
          IF(Symbol=='R-3g') Symbol='R-3c'
          IF(Symbol=='Abm2') Symbol='Aem2'
          IF(Symbol=='Acm2') Symbol='Aem2'
          IF(Symbol=='Aba2') Symbol='Aea2'
          IF(Symbol=='Aca2') Symbol='Aea2'
          IF(Symbol=='Cmca') Symbol='Cmce'
          IF(Symbol=='Cmcb') Symbol='Cmce'
          IF(Symbol=='Cmma') Symbol='Cmme'
          IF(Symbol=='Cmmb') Symbol='Cmme'
          IF(Symbol=='Ccca') Symbol='Ccce'
          IF(Symbol=='Cccb') Symbol='Ccce'
          IF(Symbol=='Ibaa') Symbol='Ibca'
          IF(Symbol=='I4bm') Symbol='I4cm'
          IF(Symbol=='I41bd') Symbol='I41cd'
          IF(Symbol=='I-4b2') Symbol='I-4c2'
          IF(Symbol=='I4/mbm') Symbol='I4/mcm'
          IF(Symbol=='I41/abd') Symbol='I41/acd'
          IF(Symbol=='P3111') Symbol='P31'
          IF(Symbol=='P3211') Symbol='P32'
          IF(Symbol=='P311') Symbol='P3'
          IF(Symbol=='P-311') Symbol='P-3'
          IF(Symbol=='61221') Symbol='6122'
          IF(Symbol=='612121') Symbol='6122'
          IF(Symbol=='61212') Symbol='6122'
          IF(Symbol=='65221') Symbol='6522'
          IF(Symbol=='652121') Symbol='6522'
          IF(Symbol=='65212') Symbol='6522'
          IF(Symbol=='R-3g') Symbol='R-3c'
          IF(Symbol=='Pn-3c') Symbol='Pn-3n'
          IF(Symbol=='Pm-3c') Symbol='Pm-3n'
          IF(Symbol=='Fd-3g') Symbol='Fd-3m'

          END SUBROUTINE StandardizeSGSymbol
	
	  
!**********************************************
          SUBROUTINE FindCenteringSymbol(Cvec,CSymbol)
          IMPLICIT NONE
          INTEGER :: NCvec,fcsi
          REAL, DIMENSION(:,:) :: CVec
          REAL :: FootPrint
          CHARACTER*1 :: CSymbol
          NCVec=size(CVec,2)
          FootPrint=0.
          DO fcsi=1,NCVec
            FootPrint=FootPrint+CVec(1,fcsi)+10.*CVec(2,fcsi)+100.*CVec(3,fcsi)
          ENDDO
          SELECT CASE (NCVec)
            CASE(1)
              CSymbol='P'
            CASE(2)
              IF (abs(FootPrint-5.5)<bigeps) THEN
        	CSymbol='C'
              ELSEIF(abs(FootPrint-50.5)<bigeps) THEN
        	CSymbol='B'
              ELSEIF(abs(FootPrint-55.0)<bigeps) THEN
        	CSymbol='A'
              ELSEIF(abs(FootPrint-55.5)<bigeps) THEN
        	CSymbol='I'
              ELSE
               CSymbol='X'
              ENDIF
            CASE(3)
              IF (abs(FootPrint-111.0)<bigeps) THEN
        	CSymbol='R'
              ELSE
               CSymbol='X'
              ENDIF
            CASE(4)
              IF (abs(FootPrint-111.0)<bigeps) THEN
        	IF ((sum(CVec(1:3,2))==sum(CVec(1:3,2))).and.(sum(CVec(1:3,3))==sum(CVec(1:3,4)))) THEN
        	  CSymbol='F'
        	ELSE
        	  CSymbol='X'
        	ENDIF
              ELSE
               CSymbol='X'
              ENDIF
            CASE DEFAULT
              CSymbol='X'
          END SELECT
          IF ((SDim==4).and.(RDim==3)) THEN !Change centering to 'X' if centering in the fourth dimension is found
            DO fcsi=1,NCVec
              if (abs(CVec(4,fcsi))>bigeps) THEN
        	CSymbol='X'
        	EXIT		 
              ENDIF
            ENDDO
          ENDIF
          END SUBROUTINE FindCenteringSymbol

	END SUBROUTINE TentativeHMSymbol

!*********************************************************************
	SUBROUTINE ProjectionOperator(RmatIn,RmatOut,Order,Dim)
  	IMPLICIT NONE
  	INTEGER :: Dim
  	REAL, DIMENSION(1:Dim,1:Dim) :: RMatIn,RMatOut,dmat,dmat2
  	INTEGER :: Order
  	!Construct the projection oeprator
  	Order=0
  	dmat=UnitMat(Dim)
  	RMatOut=0.0
  	DO
  	  dmat2=multmat(dmat,RMatIn)
  	  RMatOut=RMatOut+dmat2
  	  dmat=dmat2
  	  Order=Order+1
  	  IF (IsUnitMatrix(dmat2)) EXIT
  	ENDDO
  	RMatOut=RMatOut/float(Order)	    
	END SUBROUTINE ProjectionOperator

!*********************************************************************
        subroutine MapToMinimalPositiveT(SG, Center)
          IMPLICIT NONE
	  type(SymOp_type), dimension(:) :: SG
	  real, dimension(:,:) :: Center
          integer, dimension(1:size(Center,1)) :: StartC, EndC, RunC, AxOrd, mind, maxd
          real, dimension(1:size(Center,1)) :: TVec, BestTVec, CVec,BestCVec, stepd
          real, allocatable, dimension(:,:) :: P, HP, ProjOp
	  integer :: SDim,NSym,NCenter,i,ii,iii,n, SmOpCnt, Order
	  NSym=size(SG)
	  NCenter=size(Center,2)
	  SDim=size(Center,1)
          AxOrd=(/(i,i=1,SDim)/)
          DO SmOpCnt=1,NSym
	    SG(SmOpCnt)%t=from0to1(SG(SmOpCnt)%t)
	    BestTVec=SG(SmOpCnt)%t
!	    write(*,*) 'Start:',BestTVec
	    IF (NCenter==1) CYCLE
	    IF (all(abs(BestTVec)<eps)) CYCLE
            mind=-1
            maxd=1
	    stepd=1.0
            DO i=1,SDim
              DO ii=1,NCenter
                if ((Center(i,ii)>eps).and.(Center(i,ii)<stepd(i))) stepd(i)=Center(i,ii)
              ENDDO
            ENDDO
	    StartC=int(float(mind)/stepd)
	    EndC  =int(float(maxd)/stepd)
            RunC  =StartC
	    DO
              CVec=float(RunC)*stepd
	      IF (IsCenter(CVec)) THEN
                TVec=SG(SmOpCnt)%t+CVec
!		write(*,*) 'Candidate:',TVec, BestTVec
                IF (FirstTVecBetter(TVec,BestTVec)) THEN
		  BestTVec=TVec
!		  write(*,*) 'accepted'
                  IF (all(abs(BestTVec)<eps)) EXIT
		ELSE
!		  write(*,*) 'Not accepted'
		ENDIF
	      ENDIF
	      IF (.not.NextCoo(SDim,StartC,EndC,RunC,AxOrd)) EXIT
            ENDDO
            SG(SmOpCnt)%t=BestTVec
	  ENDDO

	  CONTAINS
	    LOGICAL FUNCTION IsCenter(InVec)
	      IMPLICIT NONE
	      real, dimension(1:SDim) :: InVec, TmpVec
              integer :: ici
	      TmpVec=From0to1(InVec)
	      IF (NCenter==1) THEN
	        IF (any(abs(TmpVec)>eps)) THEN
		  IsCenter=.false.		
		ELSE
		  IsCenter=.true.
		ENDIF
                RETURN              
	      ENDIF
              IsCenter=.false.
	      DO ici=1,NCenter
                IF (maxval(abs(Tmpvec-Center(1:SDim,ici)))<eps) THEN
	          IsCenter=.true.
                  EXIT
	        ENDIF
	      ENDDO
	    END FUNCTION IsCenter
	
	  LOGICAL FUNCTION FirstTVecBetter(V1, V2)
	    REAL, DIMENSION(1:SDim) :: V1, V2
	    INTEGER :: vibl, zerocount1, zerocount2, posicount1, posicount2
	    REAL :: length1,length2
	    FirstTVecBetter=.false.
	    length1=sum(V1**2)
            IF (any((V1<-eps)).and.( length1>eps)) RETURN !All negative, don't like that
	    length2=sum(V2**2)
	    zerocount1=count(abs(v1)<eps)
	    zerocount2=count(abs(v2)<eps)
!	    write(*,*) 'lengths:',length1, length2
!	    write(*,*) 'zerocounts:',zerocount1, zerocount2
	    IF (zerocount1>zerocount2) THEN
              FirstTVecBetter=.true.	    
	      RETURN
	    ELSEIF (zerocount1<zerocount2) THEN
              FirstTVecBetter=.false.	    
	      RETURN
	    ELSEIF (length1>length2+eps) THEN
              FirstTVecBetter=.false.	    
	      RETURN
	    ELSEIF (length1<length2-eps) THEN
              FirstTVecBetter=.true.	    
	      RETURN
	    ELSE
	      DO vibl=1,SDim
	        IF (v1(vibl)-v2(vibl)>eps) THEN
                  FirstTVecBetter=.true.	    
	          RETURN
	        ELSEIF (v1(vibl)-v2(vibl)<-eps) THEN
                  FirstTVecBetter=.false.
		  RETURN
	        ENDIF
	      ENDDO
	    ENDIF
	  END FUNCTION FirstTVecBetter

        end subroutine MapToMinimalPositiveT

!*********************************************************************
	SUBROUTINE StationaryPoint(SymOp,StatP)
	  IMPLICIT NONE	  
	  type(SymOp_type) :: SymOp
	  real, dimension(:) :: StatP
	  real, dimension(1:size(StatP),1:Size(StatP)) :: IR
	  !Find some solution of the equation (I-SymOp%R)x=SymOp%t
	  IR=UnitMat(size(StatP))-SymOp%R
	  CALL SolveLinear(IR,SymOp%t,StatP)
	end subroutine StationaryPoint
	
!*********************************************************************
        INTEGER FUNCTION SiteSymmetry(point,SG,Center)
          IMPLICIT NONE
	  type(SymOp_type), dimension(:) :: SG
	  real, dimension(:,:) :: Center
	  real, dimension(:) :: Point
	  real, dimension(1:size(point)) :: dummypt
	  integer :: ssi,ssii,NSym,NCenter,SDim
	  NSym=size(SG)
	  NCenter=size(Center,2)
	  SDim=size(Center,1)
          SiteSymmetry=1
          DO ssi=1,NCenter
            DO ssii=2,NSym
	      dummypt=multmat(SG(ssii)%R,point)+Center(1:SDim,ssi)+SG(ssii)%t
	      dummypt=from0to1(dummypt-point,bigeps)
	      if (maxval(abs(dummypt))<bigeps) SiteSymmetry=SiteSymmetry+1
            ENDDO
          ENDDO
	END FUNCTION SiteSymmetry

!*********************************************************************
        SUBROUTINE IntrinsicVector(SymOp,IntrVec)
	  IMPLICIT NONE	  
	  type(SymOp_type) :: SymOp
	  real, dimension(:) :: IntrVec
	  real, dimension(1:size(IntrVec),1:Size(IntrVec)) :: ProjOp
	  integer :: sdim, rank
          SDim=size(IntrVec)
	  CALL ProjectionOperator(float(SymOp%R),ProjOp,rank,SDim)
	  IntrVec=multmat(ProjOp,SymOp%t)	
        END SUBROUTINE IntrinsicVector

!*********************************************************************
        LOGICAL FUNCTION IsCenterExt(InVec,Center)
	  IMPLICIT NONE
          real, dimension(:) :: InVec
          real, dimension(:,:) :: Center
	  real, dimension(1:size(Center,1)) :: TmpVec
          integer :: ici,NCenter, SDim
          NCenter=size(Center,2)
          SDim=size(Center,1)

	  TmpVec=From0to1(InVec)
	  IF (NCenter==1) THEN
	    IF (any(abs(TmpVec)>eps)) THEN
	      IsCenterExt=.false.		
	    ELSE
	      IsCenterExt=.true.
	    ENDIF
            RETURN              
	  ENDIF
          IsCenterExt=.false.
	  DO ici=1,NCenter
            IF (maxval(abs(Tmpvec-Center(1:SDim,ici)))<eps) THEN
	      IsCenterExt=.true.
              EXIT
	    ENDIF
	  ENDDO
	END FUNCTION IsCenterExt


!*********************************************************************
	subroutine GenerateSGFromGen(SGGen,Center,SG)
	  IMPLICIT NONE
  	  type(SymOp_type), dimension(:) :: SGGen
  	  real, dimension(:,:) :: Center
  	  type(SymOp_type), allocatable, dimension(:) :: SG
	  INTEGER :: SDim, NGen, NCenter, NSym, AllocatedNSym=48 !48 is a good initial guess; is enough for all 3D space groups in the standard setting
  	  type(SymOp_type) :: SymOp
          logical, dimension(1:1000,1:1000) :: CheckedPairs !I am lazy and convinced that nobody will ever test a space group with more that 1000 symmetry operations in the primitive cell
	  logical :: different
	  integer :: i,ii,iii,iv, NSymNew
	  SDim=size(Center,1)
	  NGen=size(SGGen)	  
	  NCenter=size(Center,2)
	  NSym=size(SG)

	  ALLOCATE(SymOp%R(1:SDim,1:SDim))
	  ALLOCATE(SymOp%t(1:SDim))
	  !Prepate the SG array
	  CALL ReallocateSG(SG, SDim, 0, AllocatedNSym)
	  NSym=NGen
	  DO i=1,NGen
	    CALL CopySymOp(SGGen(i),SG(i))
	  ENDDO
	  IF ((NGen==1).and.(all(SGGen(1)%R==UnitMat(SDim)))) GOTO 9999
	  NSymNew=NSym
	  CheckedPairs=.false.
1000      CONTINUE
          DO i=1,NSym
	    DO ii=1,NSym
	      IF (CheckedPairs(i,ii)) CYCLE
	      different=.true.
	      CALL MultSymOp(SG(ii),SG(i),SymOp)
	      DO iii=1,NSymNew
	        DO iv=1,NCenter
		  IF (EqualOp(SymOp%R,SymOp%t+Center(1:SDim,iv),SG(iii)%R,SG(iii)%t,SDim)) THEN
		    different=.false.
		    exit
		  ENDIF
		ENDDO
	      ENDDO
	      IF (different) THEN
	        IF (NSymNew>=AllocatedNSym) CALL ReallocateSG(SG,SDim, NSymNew,NSymNew+1)
                NSymNew=NSymNew+1
		CALL CopySymOp(SymOp,SG(NSymNew))
	      ENDIF
	      CheckedPairs(i,ii)=.true.
	      different=.true.
	      CALL MultSymOp(SG(i),SG(ii),SymOp)
	      DO iii=1,NSymNew
	        DO iv=1,NCenter
		  IF (EqualOp(SymOp%R,SymOp%t+Center(1:SDim,iv),SG(iii)%R,SG(iii)%t,SDim)) THEN
		    different=.false.
		    exit
		  ENDIF
		ENDDO
	      ENDDO
	      IF (different) THEN
	        IF (NSymNew>=AllocatedNSym) CALL ReallocateSG(SG,SDim, NSymNew,NSymNew+1)
                NSymNew=NSymNew+1
		CALL CopySymOp(SymOp,SG(NSymNew))
	      ENDIF
	      CheckedPairs(ii,i)=.true.
	    ENDDO
	  ENDDO
          IF (NSymNew/=NSym) THEN
	    NSym=NSymNew
	    GOTO 1000
	  ENDIF
9999	  CALL ReallocateSG(SG,SDim,AllocatedNSym,NSym) !Remove redundant elements from the array
	   DEALLOCATE(SymOp%R)
	   DEALLOCATE(SymOp%t)
	end subroutine GenerateSGFromGen
	
!*********************************************************************
	subroutine ShiftSGOrigin(SG,OrigShift,remap)
	  IMPLICIT NONE
  	  type(SymOp_type), dimension(:) :: SG
  	  real, dimension(:) :: OrigShift
	  logical :: remap
	  INTEGER :: SDim, NSym
	  real, allocatable, dimension(:) :: tnew
          integer :: i	  
	  SDim=size(SG(1)%t)
	  NSym=size(SG)
	  allocate(tnew(1:SDim))
	  DO i=1,NSym
	    tnew=SG(i)%t-multmat(SG(i)%R-UnitMat(SDim),OrigShift)
            IF (remap) THEN
	      SG(i)%t=From0to1(tnew)
	    ELSE
	      SG(i)%t=tnew
	    ENDIF
	  ENDDO
	  DEALLOCATE(tnew)
	end subroutine ShiftSGOrigin

!*********************************************************************
        LOGICAL FUNCTION EquivalentOrigin(Orig,SG,Center)
          IMPLICIT NONE
          real, dimension(:) :: Orig
          type(SymOp_type), dimension(:) :: SG
          real, dimension(:,:) :: Center
          real, dimension(1:size(Center,1)) :: ImRtO
          integer :: i, NSym, NCenter, SDim
          NSym=size(SG)
          NCenter=size(Center,2)
          SDim=size(Center,1)
          EquivalentOrigin=.true.
          DO i=1,NSym
            ImRtO=multmat(SG(i)%R-UnitMat(SDim),Orig)
            IF (.not.IsCenterExt(ImRtO,Center)) THEN
              EquivalentOrigin=.false.
              EXIT
            ENDIF
          ENDDO
        END FUNCTION EquivalentOrigin


!*********************************************************************
      !Routines related to operations needed for fingerprint
!*********************************************************************

  	SUBROUTINE SGFP(InSG, Center, OrigShift,FP)
	  IMPLICIT NONE
  	  type(SymOp_type), dimension(:) :: InSG
          real, dimension(:,:) :: Center
          real, dimension(:) :: OrigShift
	  character*(*) :: FP
  	  type(SymOp_type), allocatable, dimension(:) :: SG
	  integer, dimension(1:size(InSG)) :: GenList
	  integer, dimension(1:size(Center,2)) :: CenGenList
	  character*(132) :: Header,ShiftString
	  integer :: SDim, NSym, NCenter, GenNr,CenGenNr,Mbase, MatBase,CenBase,TBase
	  INTEGER :: i

          SDim=size(Center,1)
          NSym=size(InSG)
	  NCenter=size(Center,2)
          CALL AllocateSG(SG,SDim,NSym)
          CALL CopySG(InSG,SG)
          
          IF (sgfp_verbosity==3) THEN
            write(*,*) 'Space group entering the procedure:'
            DO i=1,NSym
	      CALL PrintSymOp(SG(i))
            ENDDO
	    write(*,*)
	  ENDIF
          !Prepare the information about centering. It does not depend on pos
          call SortCenteringVectors(Center)
     	  call SearchCenterGenerators(Center,CenGenNr,CenGenList)
          IF (sgfp_verbosity==3) THEN
            IF (CenGenNr>0) THEN
              write(*,*) 'Centering generators:',CenGenList(1:CenGenNr)
            ELSE
              write(*,*) 'Primitive cell - no centering generators'
     	    ENDIF
          ENDIF

	  !Set the translation vectors to a standard form
	  CALL StandardIntrVec(SG, Center)
          CALL MapToMinimalPositiveT(SG,Center)
          IF (sgfp_verbosity==3) THEN
            write(*,*) 'Afted intrinsic vectors have been standardized:'
            DO i=1,NSym
              CALL PrintSymOp(SG(i))
            ENDDO
	    write(*,*)
          ENDIF

          CALL SortSymOpsForGen(SG, Center) !Sort for search for generators
          CALL SearchGenerators(SG, Center, GenNr, GenList)
	  IF (sgfp_verbosity>1) THEN
            write(*,'(a,i3,a)',ADVANCE='no') 'Generators: '
            DO i=1,GenNr
	      FP=SymOpIntTab(SG(GenList(i)),'xyz ')
	      CALL RemoveSpaces(FP)
	      write(*,'(a25)',ADVANCE='NO') FP
	    ENDDO
  	    write(*,*)
          ENDIF
          IF (sgfp_verbosity==3) THEN
            write(*,*) 'Standardized SG representation:'
            DO i=1,NSym
              FP=SymOpIntTab(SG(i),'xyz ')
              CALL RemoveSpaces(FP)
              write(*,'(i3,2X,a20,5X,a8)') i,trim(Fp),SymOpFP(SG(i),0,0)
	    ENDDO
          ENDIF
          CALL SGFPHeader(SG, Center,GenNr,GenList,CenGenNr,CenGenList,MatBase,CenBase,TBase,Header)
          IF (sgfp_errorcode/=0) GOTO 9999
          FP=Header     
	  IF (CenGenNr>0) THEN
	    DO i=1,CenGenNr
              FP=trim(FP)//CenterFP(Center(1:SDim,CenGenList(i)),CenBase)
              IF (sgfp_errorcode/=0) GOTO 9999
	    ENDDO
	  ENDIF
	  DO i=1,GenNr
	    FP=trim(FP)//SymOpFP(SG(GenList(i)),MatBase,TBase)
            IF (sgfp_errorcode/=0) GOTO 9999
	  ENDDO
	  CALL VectorToString(OrigShift,ShiftString)
	  FP=trim(FP)//' '//trim(ShiftString)            
          IF (sgfp_verbosity>0) write(*,'(2a)') 'Fingerprint: ', trim(FP)//' '//trim(ShiftString)
9999      CONTINUE
	  CALL DeallocateSG(SG)
        end subroutine SGFP

      SUBROUTINE PrivilegedOrigin(OrigSG, Center,OrigShiftToStd)
        IMPLICIT NONE
        INTERFACE
          INTEGER FUNCTION LCM(inarr)
            IMPLICIT NONE
	    INTEGER, DIMENSION(:) :: inarr
	    INTEGER, DIMENSION(1:size(inarr)) :: cumarr
            INTEGER :: minpos
          END FUNCTION LCM
        END INTERFACE

        type OrigProps_type
          integer :: SiteMult,FreeDim
          logical :: Accepted
        end type OrigProps_type
        type(OrigProps_type), ALLOCATABLE, DIMENSION(:) :: OrigProps
        integer, parameter :: MaxOrShift=100000
        type(SymOp_type), dimension(:) :: OrigSG
        real, dimension(:,:) :: Center
        real, dimension(:) :: OrigShiftToStd
        type(SymOp_type), allocatable, dimension(:) :: TmpSG, SG
        integer, allocatable, dimension(:,:) :: Omega
        integer, dimension(1:size(Center,1)) :: StartC, EndC, RunC, AxOrd, TSymMinMultiple, &
                                                mintden,curtden,SearchStep
        real, dimension(1:size(Center,1)) :: CVec,DVec, OrigShift,SumTVec
        real, allocatable, dimension(:,:) :: goodorshift
        real, allocatable, dimension(:,:,:) :: AlltList
        integer, allocatable, dimension(:) :: GoodShiftPos
        integer, dimension(1:size(OrigSG)) :: GenList
        integer :: SDim,NSym,NCenter,i,ii,iii,n, tnom,tden, GenNr
        integer :: NOrShift,NGoodOrShift,Maxsitesym,MinFreeDim
        real :: sumt
        logical :: Equivalent
        
        NSym=size(OrigSG)
        NCenter=size(Center,2)
        SDim=size(Center,1)
        AxOrd=(/(i,i=1,SDim)/)
        ALLOCATE(Omega(1:SDim,1:SDim))
        ALLOCATE(GoodOrShift(1:SDim,1:MaxOrShift))
        CALL AllocateSG(TmpSG,SDim,NSym)
        CALL AllocateSG(SG,SDim,NSym)
        CALL CopySG(OrigSG,SG)
!       Shift the space group to a reasonable starting origin. Not strictly necessary, but speeds up the search.
        OrigShiftToStd=0.0
        CALL CopySG(SG,TmpSG)
        CALL StandardIntrVec(TmpSG, Center)
        CALL MapToMinimalPositiveT(SG,Center)
        CALL SortSymOpsForGen(TmpSG, Center) !Sort for search for generators
        CALL SearchGenerators(TmpSG, Center, GenNr, GenList)
        SumTVec=0.0
        DO i=1,NSym
          SumTVec=SumTVec+TmpSG(i)%t
        ENDDO
        SumTVec=From0to1(SumTVec)
        DO
          DO i=GenNr,1,-1
            CALL StationaryPoint(TmpSG(GenList(i)),CVec)
            CALL ShiftSGOrigin(TmpSG,-CVec,.true.)
            OrigShiftToStd=OrigShiftToStd-CVec
          ENDDO
          DVec=0.0
          DO i=1,NSym
            DO ii=1,NCenter
              DVec=DVec+TmpSG(i)%t+Center(1:SDim,ii)
            ENDDO
          ENDDO
          DVec=From0to1(DVec)
!          write(*,*) DVec, SumTVec
          IF (all(abs(DVec-SumTVec)<eps)) EXIT
          SumTVec=DVec
        ENDDO
        CALL ShiftSGOrigin(SG,OrigShiftToStd,.false.)
        call MapToMinimalPositiveT(SG,Center)

!       Fill TSymMinMultiple: The ratio between the curent minimum t-denominator, and the smallest allowed step in the origin shift:
        TSymMinMultiple=1
        DO i=1,NSym
          Omega=SG(i)%R-UnitMat(SDim)
          DO ii=1,SDim
            n=max(sum(abs(omega(1:SDim,ii))),1)
            TSymMinMultiple(ii)=lcm((/n,TSymMinMultiple(ii)/))
          ENDDO
        ENDDO
        IF (NCenter>1) THEN
          n=1
          DO ii=1,SDim
            DO i=2,NCenter          
              CALL DecToFraction(Center(ii,i),tnom,tden)
              n=lcm((/n,tden/))
            ENDDO
            TSymMinMultiple(ii)=lcm((/n,TSymMinMultiple(ii)/))
          ENDDO
        ENDIF
        
        !Determine the starting mintden - denominators of the translational parts
        mintden=1
        DO i=2,NSym
          DO iii=1,NCenter
            CVec=From0to1(SG(i)%t +Center(1:SDim,iii))
            DO ii=1,SDim
              CALL DecToFraction(CVec(ii),tnom,tden)
              mintden(ii)=max(mintden(ii),tden)
            ENDDO
          ENDDO
        ENDDO
        DO ii=1,SDim
          SearchStep(ii)=max(mintden(ii),1)*TSymMinMultiple(ii)
        ENDDO
!        write(*,*) mintden, SearchStep
        StartC=0
        ENdC=SearchStep-1
        RunC=StartC
        NOrShift=0
        DO
          OrigShift=float(RunC)/float(SearchStep)
          curtden=1
          DO i=2,NSym
	    DVec=SG(i)%t-multmat(SG(i)%R-UnitMat(SDim),OrigShift)            
            DO iii=1,NCenter
              CVec=From0to1(DVec+Center(1:SDim,iii))
              DO ii=1,SDim
                CALL DecToFraction(CVec(ii),tnom,tden)
                curtden(ii)=max(curtden(ii),tden)
              ENDDO
            ENDDO
            IF (any(curtden>mintden)) EXIT
          ENDDO
!          write(*,*) curtden, mintden, searchstep, TSymMinMultiple
!          DO ii=1,NSym
!            call printsymop(TmpSG(ii))
!          ENDDO
          IF ((any(curtden<mintden)).and.(all(curtden<=mintden)).and.(all(mod(mintden,curtden)==0))) THEN !Found a "nicer" origin -> restart the whole process with a new, coarser search step
            mintden=curtden
            DO ii=1,SDim
              SearchStep(ii)=max(curtden(ii),1)*TSymMinMultiple(ii)
            ENDDO
            EndC=SearchStep-1
            RunC=0            
            NOrShift=0
            CALL ShiftSGOrigin(SG,origshift,.true.)
            OrigShiftToStd=OrigShiftToStd+origshift
            CYCLE
          ELSEIF (all(curtden==mintden)) THEN !Another acceptable origin
            CALL CopySG(SG,TmpSG)
            call ShiftSGOrigin(TmpSG,origshift,.true.)
!            call MapToMinimalPositiveT(TmpSG,Center)
!            write(*,*) 'Mintden, curtden:',mintden, curtden
            Equivalent=.false.
            DO i=1,NOrShift
              CVec=OrigShift-GoodOrShift(1:SDim,i)
              IF (EquivalentOrigin(CVec,TmpSG,Center)) THEN
!                write(*,*) OrigShift, 'is equivalent to ',GoodOrShift(1:SDim,i)
                Equivalent=.true.
                EXIT
              ENDIF
            ENDDO
            IF (.not.Equivalent) THEN
              NOrShift=NOrShift+1
              IF (NOrShift>MaxOrShift) THEN
              write(*,*) 'This group gives a crazy number of origins! Something is wrong. Please contact the author of the program.'
              stop
              ENDIF
              GoodOrShift(1:SDim,NOrShift)=OrigShift
!              write(*,'(a,3f9.4,a,3i2)') 'OrigShift:',OrigShift, '  Denom:', curtden
            ENDIF
          ENDIF  
          IF (.not.NextCoo(SDim,StartC,EndC,RunC,AxOrd)) EXIT            
        ENDDO
        ! Fill the properties of the origin:
        ALLOCATE(OrigProps(1:NOrShift))
        DO i=1,NOrShift
          OrigProps(i)%Accepted=.true.
          CALL CopySG(SG,TmpSG)
          call ShiftSGOrigin(TmpSG,GoodOrShift(1:SDim,i),.true.)
          ! Site multiplicity
          CVec=0.0
          OrigProps(i)%SiteMult=SiteSymmetry(CVec,TmpSG,Center)
          !Number of fixed dimensions
          Omega=0
          DO ii=1,NSym
            DO iii=1,NCenter
              IF (sum(abs(from0to1(TmpSG(ii)%t+Center(1:SDim,iii))))<eps) Omega=Omega+TmpSG(ii)%R
            ENDDO
          ENDDO
          CALL GetRank(float(Omega),SDim,SDim,OrigProps(i)%FreeDim)
        ENDDO
!       And now filter the origins to get only the acceptable ones:
        !The site multiplicity
        MaxSiteSym=0
        DO i=1,NOrShift
          IF (OrigProps(i)%Accepted) MaxSiteSym=max(MaxSiteSym,OrigProps(i)%SiteMult)
        ENDDO
        DO i=1,NOrShift
          IF(OrigProps(i)%SiteMult<MaxSiteSym) OrigProps(i)%Accepted=.false.
        ENDDO
        !The number of fixed dimensions
        MinFreeDim=100
        DO i=1,NOrShift
          IF (.not.OrigProps(i)%Accepted) cycle
          MinFreeDim=min(MinFreeDim,OrigProps(i)%FreeDim)
        ENDDO
        DO i=1,NOrShift
          IF(OrigProps(i)%FreeDim>MinFreeDim) OrigProps(i)%Accepted=.false.
        ENDDO

        !Count the number of origins:
        NGoodOrShift=0
        DO i=1,NOrShift
          IF (OrigProps(i)%Accepted) NGoodOrShift=NGoodOrShift+1
        ENDDO
        IF (NGoodOrShift>1) THEN
          !Even after filtering more possible origins remain. Final choice must be made
          ALLOCATE(AlltList(1:SDim,1:NSym,1:NGoodOrShift))
          ALLOCATE(GoodShiftPos(1:NGoodOrShift))
          CALL SortSymOpsForGen(SG, Center) !Sort for search for generators
          n=0
          DO i=1,NOrShift
            IF (.not.OrigProps(i)%Accepted) cycle
            n=n+1
            GoodShiftPos(n)=i
            CALL CopySG(SG,TmpSG)
            call ShiftSGOrigin(TmpSG,GoodOrShift(1:SDim,i),.true.)
            CALL StandardIntrVec(TmpSG, Center)
            call MapToMinimalPositiveT(TmpSG,Center)
            DO ii=1,NSym
              AlltList(1:SDim,ii,n)=TmpSG(ii)%t
            ENDDO
          ENDDO
          n=NGoodOrShift
          DO i=1,NSym
            DO ii=1,SDim
              sumt=10000000.
              DO iii=1,NGoodOrShift
                IF(.not.OrigProps(GoodShiftPos(iii))%Accepted) CYCLE
                sumt=min(sumt,AllTList(ii,i,iii))
              ENDDO
              DO iii=1,NGoodOrShift
                IF(.not.OrigProps(GoodShiftPos(iii))%Accepted) CYCLE
                IF (AllTList(ii,i,iii)>sumt+eps) THEN
                  OrigProps(GoodShiftPos(iii))%Accepted=.false.
                  n=n-1
                ENDIF
              ENDDO
              IF (n==1) EXIT
            ENDDO
            IF (n==1) EXIT
          ENDDO
          DEALLOCATE(AlltList)
          DEALLOCATE(GoodShiftPos)
        ENDIF
        DO i=1,NOrShift
          IF (OrigProps(i)%Accepted) THEN
            OrigShiftToStd=from0to1(OrigShiftToStd+GoodOrShift(1:SDim,i))
            call ShiftSGOrigin(OrigSG,OrigShiftToStd,.true.)
            IF (EquivalentOrigin(OrigShiftToStd,OrigSG,Center)) OrigShiftToStd=0.0
!            write(*,'(3f9.4)') OrigShiftToStd(1:SDim)
!            DO ii=1,NSym
!              call printsymop(TmpSG(ii))
!            ENDDO
            EXIT
          ENDIF
        ENDDO

        DEALLOCATE(Omega)
        DEALLOCATE(GoodOrShift)
        DEALLOCATE(OrigProps)
        CALL DeallocateSG(TmpSG)
        CALL DeallocateSG(SG)
      END SUBROUTINE PrivilegedOrigin

        subroutine StandardIntrVec(SG, Center)
          IMPLICIT NONE
	  type(SymOp_type), dimension(:) :: SG
	  real, dimension(:,:) :: Center
          integer, dimension(1:size(Center,1)) :: StartC, EndC, RunC, AxOrd, mind, maxd
          real, dimension(1:size(Center,1)) :: IntrVec, BestIntrVec, CVec,BestCVec, stepd
          real, allocatable, dimension(:,:) :: P, HP, ProjOp
	  integer :: SDim,NSym,NCenter,i,ii,iii,n, SmOpCnt, Order
	  NSym=size(SG)
	  NCenter=size(Center,2)
	  SDim=size(Center,1)
          ALLOCATE(P(1:SDim,1:2**SDim))
          ALLOCATE(HP(1:SDim,1:2**SDim))
          ALLOCATE(ProjOp(1:SDim,1:SDim))
          AxOrd=(/(i,i=1,SDim)/)
          DO SmOpCnt=1,NSym
            mind=-1
            maxd=1
	    stepd=1.0
            DO i=1,SDim
              DO ii=1,NCenter
                if ((Center(i,ii)>eps).and.(Center(i,ii)<stepd(i))) stepd(i)=Center(i,ii)
              ENDDO
            ENDDO
            CALL ProjectionOperator(float(SG(SmOpCnt)%R),ProjOp,Order,SDim)
	    BestIntrVec=multmat(ProjOp,SG(SmOpCnt)%t)
	    BestCVec=0.0E0
!	    CALL PrintSymOp(SG(SmOpCnt))
!            call printmat(BestIntrVec)
	    StartC=int(float(mind)/stepd)
	    EndC  =int(float(maxd)/stepd)
            RunC  =StartC
	    DO
              CVec=float(RunC)*stepd
!	      IF ((IsCenter(CVec)).and.(all((SG(SmOpCnt)%t+CVec)>-eps))) THEN
	      IF ((IsCenter(CVec))) THEN
                IntrVec=multmat(ProjOp,SG(SmOpCnt)%t+CVec)
!	        write(*,*) 'Original t:',SG(SmOpCnt)%t
!	        write(*,*) 'Centering:', CVec
!	        write(*,*) 'Original IntrVec:',BestIntrVec
!	        write(*,*) 'New IntrVec     :',IntrVec
                IF (FirstIntrVecBetter(IntrVec,BestIntrVec)) THEN
		  BestIntrVec=IntrVec
		  BestCVec=CVec		  
		ELSEIF(all(abs(IntrVec-BestIntrVec)<eps)) THEN !Equal intrinsic vectors, let's pick up the nicest t
		  IF (FirstTVecBetter(SG(SmOpCnt)%t+CVec,SG(SmOpCnt)%t+BestCVec)) THEN
                    BestCVec=CVec		  
		  ENDIF
		ENDIF
!		write(*,*)
	      ENDIF
	      IF (.not.NextCoo(SDim,StartC,EndC,RunC,AxOrd)) EXIT
            ENDDO
!            write(*,*) 'And the winner is:',BestIntrVec
!            write(*,*) 'Corresponding translation:',BestCVec
!	    read(*,*)
            SG(SmOpCnt)%t=SG(SmOpCnt)%t+BestCVec
	  ENDDO

          DEALLOCATE(P)
          DEALLOCATE(HP)
          DEALLOCATE(ProjOp)

	  CONTAINS
	    LOGICAL FUNCTION IsCenter(InVec)
	      IMPLICIT NONE
	      real, dimension(1:SDim) :: InVec, TmpVec
              integer :: ici
	      TmpVec=From0to1(InVec)
	      IF (NCenter==1) THEN
	        IF (any(abs(TmpVec)>eps)) THEN
		  IsCenter=.false.		
		ELSE
		  IsCenter=.true.
		ENDIF
                RETURN              
	      ENDIF
              IsCenter=.false.
	      DO ici=1,NCenter
                IF (maxval(abs(Tmpvec-Center(1:SDim,ici)))<eps) THEN
	          IsCenter=.true.
                  EXIT
	        ENDIF
	      ENDDO
	    END FUNCTION IsCenter
	
	  LOGICAL FUNCTION FirstIntrVecBetter(V1, V2)
	    REAL, DIMENSION(1:SDim) :: V1, V2
	    INTEGER :: vibl, zerocount1, zerocount2, posicount1, posicount2
	    REAL :: length1,length2
	    FirstIntrVecBetter=.false.
	    length1=sum(V1**2)
            IF (all((V1<eps)).and.( length1>eps)) RETURN !All negative, don't like that
	    length2=sum(V2**2)
	    zerocount1=count(abs(v1)<eps)
	    zerocount2=count(abs(v2)<eps)
	    posicount1=count(v1>eps)
	    posicount2=count(v2>eps)
	    IF (zerocount1>zerocount2) THEN
              FirstIntrVecBetter=.true.	    
	      RETURN
	    ELSEIF (zerocount1<zerocount2) THEN
              FirstIntrVecBetter=.false.	    
	      RETURN
	    ELSEIF (length1>length2+eps) THEN
              FirstIntrVecBetter=.false.	    
	      RETURN
	    ELSEIF (length1<length2-eps) THEN
              FirstIntrVecBetter=.true.	    
	      RETURN
	    ELSEIF (posicount1>posicount2) THEN
              FirstIntrVecBetter=.true.
	      RETURN
	    ELSEIF (posicount1<posicount2) THEN
              FirstIntrVecBetter=.false.
	      RETURN
	    ELSE
	      DO vibl=1,SDim
	        IF (v1(vibl)-v2(vibl)>eps) THEN
                  FirstIntrVecBetter=.true.	    
	          RETURN
	        ELSEIF (v1(vibl)-v2(vibl)<-eps) THEN
                  FirstIntrVecBetter=.false.
		  RETURN
	        ENDIF
	      ENDDO
	    ENDIF
	  END FUNCTION FirstIntrVecBetter

	  LOGICAL FUNCTION FirstTVecBetter(V1, V2)
	    REAL, DIMENSION(1:SDim) :: V1, V2
	    INTEGER :: vibl, zerocount1, zerocount2, posicount1, posicount2
	    REAL :: length1,length2
	    FirstTVecBetter=.false.
	    length1=sum(V1**2)
	    length2=sum(V2**2)
	    zerocount1=count(abs(v1)<eps)
	    zerocount2=count(abs(v2)<eps)
!	    write(*,*) 'lengths:',length1, length2
!	    write(*,*) 'zerocounts:',zerocount1, zerocount2
	    IF (zerocount1>zerocount2) THEN
              FirstTVecBetter=.true.	    
	      RETURN
	    ELSEIF (zerocount1<zerocount2) THEN
              FirstTVecBetter=.false.	    
	      RETURN
	    ELSEIF (length1>length2+eps) THEN
              FirstTVecBetter=.false.	    
	      RETURN
	    ELSEIF (length1<length2-eps) THEN
              FirstTVecBetter=.true.	    
	      RETURN
	    ELSE
	      DO vibl=1,SDim
	        IF (v1(vibl)-v2(vibl)>eps) THEN
                  FirstTVecBetter=.true.	    
	          RETURN
	        ELSEIF (v1(vibl)-v2(vibl)<-eps) THEN
                  FirstTVecBetter=.false.
		  RETURN
	        ENDIF
	      ENDDO
	    ENDIF
	  END FUNCTION FirstTVecBetter

        end subroutine StandardIntrVec

      subroutine SearchCenterGenerators(Center, GenNr, CenGenList)
      IMPLICIT NONE
	  real, dimension(:,:) :: Center
	  integer :: GenNr
	  integer, dimension(:) :: CenGenList
	  real, allocatable, dimension(:,:) :: CurGen, CurCenter
          integer, dimension(1:size(Center,1)) :: StartC, EndC, RunC, AxOrd
	  integer :: SDim,NCenter,CurNCen,i,ii,iii
	  logical :: Finished
	  Finished=.false.
	  GenNr=0
	  NCenter=size(Center,2)
	  SDim=size(Center,1)
	  IF (NCenter==1) THEN
	    CenGenList(1)=1
	    GenNr=0
	    RETURN
	  ENDIF
	  ALLOCATE(CurCenter(1:SDim,1:NCenter))
	  DO
	    GenNr=GenNr+1
	    StartC=1
	    EndC=NCenter
	    RunC=StartC
        AxOrd(1:GenNr)=(/(i,i=GenNr,1,-1)/)
	    IF (ALLOCATED(CurGen)) DEALLOCATE(CurGen)
	    ALLOCATE(CurGen(1:SDim,1:GenNr))
	    DO
          DO i=1,GenNr
	        CurGen(1:SDim,i)=Center(1:SDim,RunC(i))
	      ENDDO
	      CALL GenerateCenterFromGen(CurNCen)
	      IF (CurNCen==NCenter) THEN
	        DO i=1,GenNr
		      CenGenList(i)=RunC(i)
            ENDDO
            Finished=.true.
		    EXIT
	      ENDIF
	      IF (.not.NextCoo(GenNr,StartC(1:GenNr),EndC(1:GenNr),RunC(1:GenNr),AxOrd(1:GenNr))) EXIT	      
	    ENDDO
	    IF (Finished) EXIT
	  ENDDO
	  DEALLOCATE(CurGen)
	  DEALLOCATE(CurCenter)
	  
	  CONTAINS
	  subroutine GenerateCenterFromGen(CurNCen)
	    IMPLICIT NONE
	    integer :: CurNCen, NCenNew
	    real, dimension(1:SDim,1:NCenter) :: CurCenter
	    real, dimension(1:SDim) :: TmpCen
	    logical, dimension(1:NCenter,1:NCenter) :: CheckedPairs
	    logical :: different
	    
	    CurNCen=GenNr
	    DO i=1,GenNr
	      CurCenter(1:SDim,i)=CurGen(1:SDim,i)
	    ENDDO
	    NCenNew=CurNCen
	    CheckedPairs=.false.
1000    CONTINUE
        DO i=1,CurNCen
	      DO ii=i,CurNCen
	        IF (CheckedPairs(i,ii)) CYCLE
	        different=.true.
	        TmpCen=from0to1(CurCenter(1:SDim,i)+CurCenter(1:SDim,ii))
	        DO iii=1,NCenNew
	          IF (all(abs(TmpCen-CurCenter(1:SDim,iii))<eps)) THEN
                different=.false.
                exit
	          ENDIF
	        ENDDO
	        IF (different) THEN
              NCenNew=NCenNew+1
              CurCenter(1:SDim,NCenNew)=TmpCen
	        ENDIF
	        CheckedPairs(i,ii)=.true.
	        CheckedPairs(ii,i)=.true.
	      ENDDO
	    ENDDO
        IF (NCenNew/=CurNCen) THEN
	      CurNCen=NCenNew
	      GOTO 1000
	    ENDIF

      end subroutine GenerateCenterFromGen

	end subroutine SearchCenterGenerators
	
        subroutine SearchGenerators(SG, Center, GenNr,GenList)
          IMPLICIT NONE
	  type(SymOp_type), dimension(:) :: SG
	  type(SymOp_type), allocatable, dimension(:) :: TmpSG,Gen
	  real, dimension(:,:) :: Center
	  integer, dimension(1:size(SG),1:size(SG)) :: GrpTable
	  logical, dimension(1:size(SG)) :: Present
	  integer :: GenNr
	  integer, dimension(:) :: GenList
          integer, dimension(1:size(Center,1)) :: StartC, EndC, RunC, AxOrd
          integer, dimension(1:size(Center,1),1:size(Center,1)) :: TmpMat
	  integer :: SDim,NSym,TmpNSym,NCenter,i,ii,iii, NFound,NFoundOld
	  logical :: Finished
	  Finished=.false.
	  GenNr=0
	  NSym=size(SG)
	  NCenter=size(Center,2)
	  SDim=size(Center,1)
          !Fill the Group table
          DO i=1,NSym
	    DO ii=1,NSym
	      TmpMat=multmat(SG(i)%R,SG(ii)%R)
	      DO iii=1,NSym
	        IF (all(TmpMat==SG(iii)%R)) THEN
                  GrpTable(i,ii)=iii
	          EXIT
	        ENDIF
              ENDDO
	    ENDDO
	  ENDDO
	  DO
	    GenNr=GenNr+1
	    CALL AllocateSG(Gen,SDim,GenNr)
	    StartC=1
	    EndC=NSym
	    RunC=StartC
            AxOrd(1:GenNr)=(/(i,i=GenNr,1,-1)/)
	    DO
              Present=.false.	      
	      DO i=1,GenNr
	        Present(RunC(i))=.true.
	      ENDDO
              NFound=count(Present)
              NFoundOld=NFound
              DO
	        DO i=1,NSym
		  IF (Present(i)) THEN
                    DO ii=1,NSym
		      IF (Present(ii)) THEN
		        IF (.not.Present(GrpTable(i,ii))) NFound=NFound+1
		        Present(GrpTable(i,ii))=.true.
		        IF (.not.Present(GrpTable(ii,i))) NFound=NFound+1
		        Present(GrpTable(ii,i))=.true.
		      ENDIF
		    ENDDO
		  ENDIF
	        ENDDO
                IF ((NFound==NFoundOld).or.(NFound==NSym)) EXIT
                NFoundOld=NFound
	      ENDDO
	      IF (NFound==NSym) EXIT
	      IF (.not.NextCoo(GenNr,StartC(1:GenNr),EndC(1:GenNr),RunC(1:GenNr),AxOrd(1:GenNr))) EXIT	      
	    ENDDO
	    IF (NFound==NSym) THEN
	      DO i=1,GenNr
	        GenList(i)=RunC(i)
	      ENDDO
	      EXIT
	    ENDIF
	  ENDDO
	  CALL DeallocateSG(TmpSG)
	  CALL DeallocateSG(Gen)
	end subroutine SearchGenerators

	SUBROUTINE SortSymOpsForGen(SG, Center)
	  IMPLICIT NONE
  	  type(SymOp_type), dimension(:) :: SG
  	  type(SymOp_type) :: TmpSymOp
	  real, dimension(:,:) :: Center
	  REAL, ALLOCATABLE, DIMENSION(:,:) :: PosList, ProjOp
          INTEGER, ALLOCATABLE, DIMENSION(:) :: NrPosList, OrderList, FreeDimList
	  INTEGER :: NSym, SDim, i,ii, iii
	  SDim=size(Center,1)
	  NSym=size(SG)
	  ALLOCATE(NrPosList(1:NSym))
	  ALLOCATE(OrderList(1:NSym))
	  ALLOCATE(FreeDimList(1:NSym))
	  ALLOCATE(ProjOp(1:SDim, 1:SDim))
	  ALLOCATE(TmpSymOp%R(1:SDim, 1:SDim))
	  ALLOCATE(TmpSymOp%t(1:SDim))
	  FreeDimList=0
	  NrPosList=0
	  DO i=1,NSym
	    call ProjectionOperator(float(SG(i)%R), ProjOp, OrderList(i), SDim)
            call GetRank(ProjOp,SDim, SDim, FreeDimList(i))
	  ENDDO
	  DO i=NSym,2,-1
	    DO ii=2,i
	      IF(FirstSymOpBigger(SG(ii-1),NrPosList(ii-1),FreeDimList(ii-1),OrderList(ii-1),&
	         SG(ii),NrPosList(ii), FreeDimList(ii), OrderList(ii))) THEN
	        CALL CopySymOp(SG(ii), TmpSymOp)
		CALL CopySymOp(SG(ii-1),SG(ii))
		CALL CopySymOp(TmpSymOp,SG(ii-1))
		iii=NrPosList(ii)
		NrPosList(ii)=NrPosList(ii-1)
		NrPosList(ii-1)=iii
		iii=OrderList(ii)
		OrderList(ii)=OrderList(ii-1)
		OrderList(ii-1)=iii
		iii=FreeDimList(ii)
		FreeDimList(ii)=FreeDimList(ii-1)
		FreeDimList(ii-1)=iii
              ENDIF
	    ENDDO
	  ENDDO
!          Write(*,*) 'Sorted operations:'
!	  DO i=1, NSym
!	    write(*,*) NrPosList(i), FreeDimList(i),NDiagList(i),OrderList(i)
!	    call printsymop(SG(i))
!	  ENDDO
          DEALLOCATE(NrPosList)
	  DEALLOCATE(OrderList)
	  DEALLOCATE(FreeDimList)
	  DEALLOCATE(ProjOp)
	  DEALLOCATE(TmpSymOp%R)
	  DEALLOCATE(TmpSymOp%t)
	  IF (ALLOCATED(PosList)) DEALLOCATE(PosList)
	  
	  CONTAINS
	  LOGICAL FUNCTION FirstSymOpBigger(SymOp1,NrPos1,FreeDim1,Order1,SymOp2,NrPos2,FreeDim2,Order2)
	    IMPLICIT NONE
	    Type(SymOp_Type) :: SymOp1, SymOp2
	    INTEGER :: NrPos1, NrPos2, FreeDim1, FreeDim2, Order1, Order2, vibl1, vibl2, crit1, crit2, crit3
	    IF (all(SymOp2%R+UnitMat(SDim)==0)) THEN
	      FirstSymOpBigger=.true.
              RETURN	      
	    ELSEIF (all(SymOp1%R+UnitMat(SDim)==0)) THEN
	      FirstSymOpBigger=.false.
	      RETURN	
            ENDIF
	    !SetUp the criteria based on key. If Crit>0, then FirstSymOpBigger=.true.
	      Crit2=Order2-Order1
	      Crit3=FreeDim2-FreeDim1
	    IF(Crit2>0) THEN
	      FirstSymOpBigger=.true.
	      RETURN
	    ELSEIF(Crit2<0) THEN
	      FirstSymOpBigger=.false.
	      RETURN
	    ELSEIF(Crit3>0) THEN
	      FirstSymOpBigger=.true.
	      RETURN
	    ELSEIF(Crit3<0) THEN
	      FirstSymOpBigger=.false.
	      RETURN
	    ELSE !Prefer matrix with smaller numbers at the beginning, but the best is inversion center
	      DO vibl1=1,SDim	        
		DO vibl2=vibl1,1,-1
		  IF (SymOP1%R(vibl1,vibl2)>SymOP2%R(vibl1,vibl2)) THEN
	            FirstSymOpBigger=.false.
                    RETURN
	          ELSEIF (SymOP1%R(vibl1,vibl2)<SymOP2%R(vibl1,vibl2)) THEN
	            FirstSymOpBigger=.true.
                    RETURN
		  ELSEIF (SymOP1%R(vibl2,vibl1)>SymOP2%R(vibl2,vibl1)) THEN
	            FirstSymOpBigger=.false.
                    RETURN
	          ELSEIF (SymOP1%R(vibl2,vibl1)<SymOP2%R(vibl2,vibl1)) THEN
	            FirstSymOpBigger=.true.
                    RETURN
	          ENDIF
	        ENDDO
	      ENDDO
	    ENDIF
	  END FUNCTION FirstSymOpBigger

	END SUBROUTINE SortSymOpsForGen

	SUBROUTINE SortPosVectors(VecList, TList)
	  IMPLICIT NONE
          REAL, DIMENSION(:,:) :: VecList, TList
	  REAL, DIMENSION(1:size(VecList,1)) :: TmpVec
	  INTEGER :: NVec, SDim, i,ii
	  SDim=size(VecList,1)
	  NVec=size(VecList,2)
	  DO i=NVec,2,-1
	    DO ii=2,i
	      IF(FirstVecBigger(VecList(1:SDim,ii-1),VecList(1:SDim,ii))) THEN
	        TmpVec=VecList(1:SDim,ii)
		VecList(1:SDim,ii)=VecList(1:SDim,ii-1)
		VecList(1:SDim,ii-1)=TmpVec
	        TmpVec=TList(1:SDim,ii)
		TList(1:SDim,ii)=TList(1:SDim,ii-1)
		TList(1:SDim,ii-1)=TmpVec
              ENDIF
	    ENDDO
	  ENDDO

	  CONTAINS
	  LOGICAL FUNCTION FirstVecBigger(V1, V2)
	    REAL, DIMENSION(1:SDim) :: V1, V2
	    INTEGER :: vibl
	    FirstVecBigger=.false.
	    DO vibl=1,SDim
	      IF (v1(vibl)-v2(vibl)>eps) THEN
	        FirstVecBigger=.true.
		EXIT
	      ELSEIF (v1(vibl)-v2(vibl)<-eps) THEN
	        FirstVecBigger=.false.
		EXIT
	      ENDIF
	    ENDDO
	  END FUNCTION FirstVecBigger

	END SUBROUTINE SortPosVectors

	SUBROUTINE SortCenteringVectors(Center)
	  IMPLICIT NONE
          REAL, DIMENSION(:,:) :: Center
	  REAL, DIMENSION(1:size(Center,1)) :: TmpVec
	  INTEGER :: NVec, SDim, i,ii
	  SDim=size(Center,1)
	  NVec=size(Center,2)
	  DO i=NVec,2,-1
	    DO ii=2,i
	      IF(FirstVecBigger(Center(1:SDim,ii-1),Center(1:SDim,ii))) THEN
	        TmpVec=Center(1:SDim,ii)
		Center(1:SDim,ii)=Center(1:SDim,ii-1)
		Center(1:SDim,ii-1)=TmpVec
              ENDIF
	    ENDDO
	  ENDDO

	  CONTAINS
	  LOGICAL FUNCTION FirstVecBigger(V1, V2)
	    REAL, DIMENSION(1:SDim) :: V1, V2
	    INTEGER :: vibl
	    FirstVecBigger=.false.
	    DO vibl=1,SDim
	      IF (v1(vibl)-v2(vibl)>eps) THEN
	        FirstVecBigger=.true.
		EXIT
	      ELSEIF (v1(vibl)-v2(vibl)<-eps) THEN
	        FirstVecBigger=.false.
		EXIT
	      ENDIF
	    ENDDO
	  END FUNCTION FirstVecBigger

	END SUBROUTINE SortCenteringVectors

!*********************************************************************
      !Routines used directly for producing the finger print
!*********************************************************************

!*********************************************************************
        subroutine VectorToString(InVec,String)
          IMPLICIT NONE
	  real, dimension(:) :: InVec
	  character*(*) :: String
	  character*132 :: OneComp
	  integer :: i, SDim,Nom,Den
	  SDim=size(InVec)
	  
	  String='('
  	  DO i=1,SDim
    	    onecomp=''
  	    IF (abs(InVec(i))>eps) THEN
  	      CALL DecToFraction(InVec(i),Nom,Den)
  	      IF (Den<100) THEN
  	  	write(onecomp,'(i3,a,i2)') Nom,'/',Den
	      ELSE
  	  	write(onecomp,'(f9.5)') InVec(i)
    	      ENDIF
  	    ELSE
	      onecomp='0'
	    ENDIF
	    
	    IF (i<SDim) THEN
              String=trim(String)//trim(onecomp)//','
	    ELSE
              String=trim(String)//trim(onecomp)//')'
	    ENDIF
	  ENDDO
	  CALL RemoveSpaces(String)
	end subroutine VectorToString
	
!*********************************************************************
	subroutine SGFPHeader(SG, Center,GenNr,GenList,CenGenNr,CenGenList,MatBase,CenBase,TBase,Header)
          IMPLICIT NONE
	  type(SymOp_type), dimension(:) :: SG
	  real, dimension(:,:) :: Center
	  real, allocatable,dimension(:) :: TmpList
	  integer, dimension(:) :: GenList, CenGenList
	  integer :: SDim,NSym,NCenter,GenNr,CenGenNr,MatBase,CenBase,TBase,CurMatBase
	  integer :: i
	  character*(*) :: Header
	  NSym=size(SG)
	  NCenter=size(Center,2)
	  SDim=size(Center,1)
	  MatBase=0
	  ALLOCATE(TmpList(1:GenNr*SDim))	  
	  DO i=1,GenNr
	    CurMatBase=maxval(abs(SG(GenList(i))%R))
	    MatBase=max(MatBase,CurMatBase)
	    TmpList((i-1)*SDim+1:i*SDim)=SG(GenList(i))%t
	  ENDDO
          MatBase=MatBase*2+1
          i=0
	  DO
	    i=i+1
	    IF (all(abs(abs(TmpList*i)-abs(nint(TmpList*i)))<bigeps)) THEN
	      TBase=i
	      EXIT
	    ENDIF
	  ENDDO
	  IF (TBase==1) TBase=0
          CenBase=0
	  DEALLOCATE(TmpList)
	  ALLOCATE(TmpList(1:CenGenNr*SDim))	  
	  DO i=1,CenGenNr
	    TmpList((i-1)*SDim+1:i*SDim)=Center(1:SDim,CenGenList(i))
	  ENDDO
          i=0
	  DO
	    i=i+1
	    IF (all(abs(abs(TmpList*i)-abs(nint(TmpList*i)))<bigeps)) THEN
	      CenBase=i
	      EXIT
	    ENDIF
	  ENDDO
	  IF (CenBase==1) CenBase=0
!	  write(*,*) 'MatBase, TBase, CenBase:',MatBase, TBase,CenBase
	  DEALLOCATE(TmpList)
	  Header=''
	  Header(1:1)=Nr2Char(SDim)
	  Header(2:2)=Nr2Char(MatBase)
	  Header(3:3)=Nr2Char(CenGenNr)
	  Header(4:4)=Nr2Char(CenBase)
	  Header(5:5)=Nr2Char(GenNr)
	  Header(6:6)=Nr2Char(TBase)
	  
        end subroutine SGFPHeader

!*********************************************************************
	function CenterFP(Center,Base)
	  use matrices
	  IMPLICIT NONE
	  character*(132) :: CenterFP
	  REAL, DIMENSION(:) :: Center
	  INTEGER :: Base
	  integer :: SDim,  MatCharNr, VecCharNr
	  INTEGER :: i,ii, iii
	  CHARACTER*(size(Center)) :: VecRep
	  SDim=size(Center)
	  IF (Base==0) THEN
	    i=0
	    DO
	      i=i+1
	      IF (all((abs(abs(Center*i)-nint(abs(Center*i))))<eps)) THEN
	        Base=i
		EXIT
              ENDIF
	    ENDDO
	  ENDIF
	  VecCharNr=1+int(log(float(Base**(SDim)))/log(float(BigBase))-eps)
	  VecRep=''
	  DO i=1,SDim
	    VecRep(i:i)=Nr2Char(nint(Center(i)*Base))
            IF (sgfp_errorcode/=0) RETURN
	  ENDDO
	  VecRep=Convertbase(VecRep,Base,BigBase)
          IF (sgfp_errorcode/=0) RETURN
	  CALL PadZeroes(VecRep,VecCharNr)
	  CenterFP=trim(VecRep)
	end function CenterFP
	
!*********************************************************************
	function SymOpFP(SymOp,Mbase,TBase)
	  use matrices
	  IMPLICIT NONE
	  character*(132) :: SymOpFP
	  TYPE(SymOp_type) :: SymOp
	  integer :: Mbase, TBase
	  integer :: SDim, MatCharNr, VecCharNr
	  INTEGER :: i,ii, iii
	  CHARACTER*(size(SymOp%t)*size(symop%t)) :: MatRep
	  CHARACTER*(size(SymOp%t)) :: VecRep
	  SDim=size(SymOp%t)
	  IF (MBase==0) THEN
	    MBase=maxval(abs(SymOp%R))*2+1
	  ENDIF
	  IF (TBase==0) THEN
	    i=0
	    DO
	      i=i+1
	      IF (all((abs(abs(SymOp%t*i)-nint(abs(SymOp%t*i))))<eps)) THEN
	        TBase=i
		EXIT
              ENDIF
	    ENDDO
	  ENDIF
	  MatCharNr=1+int(log(float(MBase**(SDim**2)))/log(float(BigBase))-eps)
	  VecCharNr=1+int(log(float(TBase**(SDim)))/log(float(BigBase))-eps)
	  !Encode matrix
	  iii=0
	  DO i=1,SDim
	    DO ii=1,SDim
	      iii=iii+1
	      MatRep(iii:iii)=Nr2Char(SymOp%R(i,ii)+Mbase/2)
	    ENDDO
            IF (sgfp_errorcode/=0) RETURN
	  ENDDO
	  MatRep=Convertbase(MatRep,MBase,BigBase)
	  CALL PadZeroes(MatRep,MatCharNr)
	  VecRep=''
	  DO i=1,SDim
	    VecRep(i:i)=Nr2Char(nint(SymOp%t(i)*TBase))
	  ENDDO
          IF (sgfp_errorcode/=0) RETURN
	  VecRep=Convertbase(VecRep,TBase,BigBase)
	  CALL PadZeroes(VecRep,VecCharNr)
	  SymOpFP=trim(MatRep)//trim(VecRep)
	end function SymOpFP

!*********************************************************************
	function ConvertBase(InString,InBase,OutBase)
	  character*(*) :: InString
	  integer :: InBase, OutBase
	  character*132 :: ConvertBase
	  integer(KIND=8) :: Number
	  Number=String2Nr(InString,InBase)
          IF (sgfp_errorcode/=0) RETURN
	  ConvertBase=Nr2String(Number,OutBase)
	end function ConvertBase
	
!*********************************************************************
	subroutine PadZeroes(InString,CharNr)
	  character*(*) :: InString
	  integer :: CharNr, LenDif,i
	  LenDif=CharNr-len_trim(InString)
	  IF (LenDif>0) THEN
            DO i=CharNr,1,-1
	      IF (i<=LenDif) THEN
                InString(i:i)='0'
	      ELSE
	        InString(i:i)=Instring(i-LenDif:i-LenDif)
	      ENDIF
	    ENDDO
	  ENDIF  	  
	end subroutine PadZeroes
	
!*********************************************************************
	function String2Nr(InString, Base)
	  character*(*) :: InString
	  integer(kind=8) :: String2Nr
	  integer :: Base, BasePower, s2ni, Digit
	  String2Nr=0
	  BasePower=1
	  DO s2ni=len_trim(InString),1,-1	    
	    Digit=Char2Nr(InString(s2ni:s2ni))
	    String2Nr=String2Nr+Digit*BasePower
	    BasePower=BasePower*Base
	  ENDDO
	end function String2Nr
	
!*********************************************************************
	function Char2Nr(InChar)
	  character :: InChar
	  integer :: Char2Nr
	  integer :: AsciiCode
	  AsciiCode=iachar(InChar)
	  SELECT CASE(AsciiCode)
	    CASE(48:57)
	      Char2Nr=AsciiCode-48
            CASE(65:90)
              Char2Nr=AsciiCode-55
	    CASE(97:122)
              Char2Nr=AsciiCode-61
	    CASE(123)
              Char2Nr=62
	    CASE(125)
	      Char2Nr=63
	    CASE DEFAULT
              sgfp_errorcode=-1
              Char2Nr=100
!              STOP 'Function Char2Nr: Cannot convert character to number'
	  END SELECT	  
	end function Char2Nr
	
!*********************************************************************
	function Nr2String(InNr, Base)
	  character*(132) :: Nr2String
	  integer(kind=8) :: InNr
	  integer :: Base, NrDigits, Remainder,n2si
	  Nr2String=''
	  NrDigits=1+int(log(dble(InNr))/log(dble(Base)))
          Remainder=InNr
	  n2si=0
	  DO n2si=NrDigits,1,-1
	    Nr2String(n2si:n2si)=Nr2Char(mod(Remainder,base))
	    Remainder=Remainder/base
	  ENDDO
	end function Nr2String

!*********************************************************************
	function Nr2Char(InNr)
	  character :: Nr2Char
	  integer :: InNr
	  SELECT CASE(InNr)
	    CASE(0:9)
	      Nr2Char=achar(InNr+48)
            CASE(10:35)
              Nr2Char=achar(InNr+55)
	    CASE(36:61)
              Nr2Char=achar(InNr+61)
	    CASE(62)
              Nr2Char='{'
	    CASE(63)
	      Nr2Char='}'
	    CASE DEFAULT
              sgfp_errorcode=-2
              Nr2Char='!'
!              STOP 'Function Nr2Char: Cannot convert number to character'
	  END SELECT
	end function Nr2Char
	
!*********************************************************************
      !Hall symbol interpretation
!*********************************************************************


!*********************************************************************
    	subroutine SGfromHall(Hall,NCenter,Center, NSymPrim,SG)
  	  IMPLICIT NONE
	  character*(*) :: Hall
  	  type(SymOp_type), allocatable, dimension(:) :: SG, SGGen
  	  real, allocatable, dimension(:,:) :: Center
	  real, dimension(1:3) :: Origshift
  	  integer :: i, NCenter,NSymPrim, NGen, rot
	  character :: dir
	  logical :: IsInv

    	  call StandardizeHall(Hall,OrigShift)
	  !Set up centering vectors
	  CALL HallCenSmb2Centers(HallCenterSmb(Hall),NCenter,Center)
  	  IsInv=HallIsInv(Hall)
	  NGen=HallNGen(Hall)
          IF (NGen==0) THEN
            NSymPrim=0
            RETURN
          ENDIF
	  IF (HallGen(Hall,1)/='1') NGen=NGen+1 !Add identity
	  IF (HallIsInv(Hall)) NGen=NGen+1
  	  CALL AllocateSG(SGGen,3,NGen)
  	  !Define the identity
	  SGGen(1)%R=UnitMat(3); SGGen(1)%t=0.0
	  IF (HallGen(Hall,1)/='1') THEN
            do i=1,HallNGen(Hall)
              call HallGenSmb2SymOp(trim(HallGen(Hall,i)),i,Rot, dir, SGGen(i+1)) !Rot and dir should contain the rotation and direction of the preceding symop on entry, on leave the current rotation and direction
	    enddo
	  ENDIF
	  IF (IsInv) THEN
	    SGGen(NGen)%R=-1*UnitMat(3); SGGen(NGen)%t=0.0
	  ENDIF
	  CALL GenerateSGFromGen(SGGen,Center,SG)
	  NSymPrim=size(SG)
	  CALL ShiftSGOrigin(SG,OrigShift,.true.)
	  CALL DeallocateSG(SGGen)
	end subroutine SGfromHall
    	
!*********************************************************************
	subroutine HallGenSmb2SymOp(GenSmb,GenNr,Rot, Dir, SymOp)
	  IMPLICIT NONE
	  character*(*) :: GenSmb !Symbol of the generator
          integer :: GenNr, Rot, CurRot !Position in the Hall symbol, preceding rotation order and current rotation order
          real, dimension(1:3) :: CurTrVec
	  type (SymOp_type) :: SymOp !Resulting symmetry operation
	  character*(len(GenSmb)+1) ::GenSmbTmp !Temporary symbol
	  character :: dir !direction of the preceding generator in the hall symbol
          integer :: i, det, FracTrans
	  ! Internal convention in this routine: The six possibilities for off-axis 2-fold rotations 
	  ! (see Table A1.4.2.5 in IT B, 2001, p 113), are here denoted with capital A through F after the rotation-order symbol
	  IF (GenSmb(1:1)=='-') THEN
	    det=-1; GenSmbTmp=GenSmb(2:)
	  ELSE
	    det=1; GenSmbTmp=GenSmb(1:)
	  ENDIF
	  read(GenSmbTmp(1:1),'(i1)') CurRot !Fill the current rotation order in CurRot
	  IF (CurRot==1) THEN
	      GenSmbTmp(2:)=' '//GenSmbTmp(2:) !This is just to make all symbols occupy two characters
	  ELSEIF ((gensmbtmp(2:2)/='x').and.(gensmbtmp(2:2)/='y').and.(gensmbtmp(2:2)/='z').and. &
	     (gensmbtmp(2:2)/='''').and.(gensmbtmp(2:2)/='"').and.(gensmbtmp(2:2)/='*')) THEN !Add the default directions of axes must be added
            IF (GenNr==1) THEN !default direction c
	      GenSmbTmp(2:)='z'//GenSmbTmp(2:)
	    ELSEIF((GenNr==2).and.(CurRot==2)) THEN
	      if ((rot==2).or.(rot==4)) THEN
	        GenSmbTmp(2:)='x'//GenSmbTmp(2:)
	      elseif((rot==3).or.(rot==6)) THEN
	        GenSmbTmp(2:)='E'//GenSmbTmp(2:)	        
	      endif	      
	    ELSEIF ((GenNr==3).and.(CurRot==3)) THEN
	      GenSmbTmp(2:)='*'//GenSmbTmp(2:)
	    ENDIF
	  ENDIF
	  IF (gensmbtmp(2:2)=='''') THEN
	    SELECT CASE (dir)
	      CASE('x')
	        gensmbtmp(2:2)='A'
	      CASE('y')
	        gensmbtmp(2:2)='C'
	      CASE('z')
	        gensmbtmp(2:2)='E'
	    END SELECT
	  ELSEIF (gensmbtmp(2:2)=='"') THEN
	    SELECT CASE (dir)
	      CASE('x')
	        gensmbtmp(2:2)='B'
	      CASE('y')
	        gensmbtmp(2:2)='D'
	      CASE('z')
	        gensmbtmp(2:2)='F'
	    END SELECT
	  ENDIF
          !Fill the matrices according to IT B 2001, p. 113
	  !Attention, due to the fortran array conventions the matrices are listed columnwise (first column, then second column, then third column)
	  SELECT CASE (GenSmbTmp(1:2))
	    CASE('1 ')
	      SymOp%R=RESHAPE((/1,0,0,0,1,0,0,0,1/),(/3,3/))
	    CASE('2x')
	      SymOp%R=RESHAPE((/1,0,0,0,-1,0,0,0,-1/),(/3,3/))
	    CASE('2y')
	      SymOp%R=RESHAPE((/-1,0,0,0,1,0,0,0,-1/),(/3,3/))
	    CASE('2z')
	      SymOp%R=RESHAPE((/-1,0,0,0,-1,0,0,0,1/),(/3,3/))
	    CASE('3x')
	      SymOp%R=RESHAPE((/1,0,0,0,0,1,0,-1,-1/),(/3,3/))
	    CASE('3y')
	      SymOp%R=RESHAPE((/-1,0,-1,0,1,0,1,0,0/),(/3,3/))
	    CASE('3z')
	      SymOp%R=RESHAPE((/0,1,0,-1,-1,0,0,0,1/),(/3,3/))
	    CASE('4x')
	      SymOp%R=RESHAPE((/1,0,0,0,0,1,0,-1,0/),(/3,3/))
	    CASE('4y')
	      SymOp%R=RESHAPE((/0,0,-1,0,1,0,1,0,0/),(/3,3/))
	    CASE('4z')
	      SymOp%R=RESHAPE((/0,1,0,-1,0,0,0,0,1/),(/3,3/))
	    CASE('6x')
	      SymOp%R=RESHAPE((/1,0,0,0,1,1,0,-1,0/),(/3,3/))
	    CASE('6y')
	      SymOp%R=RESHAPE((/0,0,-1,0,1,0,1,0,1/),(/3,3/))
	    CASE('6z')
	      SymOp%R=RESHAPE((/1,1,0,-1,0,0,0,0,1/),(/3,3/))
	    CASE('2A')
	      SymOp%R=RESHAPE((/-1,0,0,0,0,-1,0,-1,0/),(/3,3/))
	    CASE('2B')
	      SymOp%R=RESHAPE((/-1,0,0,0,0,1,0,1,0/),(/3,3/))
	    CASE('2C')
	      SymOp%R=RESHAPE((/0,0,-1,0,-1,0,-1,0,0/),(/3,3/))
	    CASE('2D')
	      SymOp%R=RESHAPE((/0,0,1,0,-1,0,1,0,0/),(/3,3/))
	    CASE('2E')
	      SymOp%R=RESHAPE((/0,-1,0,-1,0,0,0,0,-1/),(/3,3/))
	    CASE('2F')
	      SymOp%R=RESHAPE((/0,1,0,1,0,0,0,0,-1/),(/3,3/))
	    CASE('3*')
	      SymOp%R=RESHAPE((/0,1,0,0,0,1,1,0,0/),(/3,3/))
	    CASE DEFAULT
	      write(*,*) 'Invalid rotation-matrix identificator in the Hall symbol:'//trim(GenSmbTmp)
	      STOP 
	  END SELECT
	  SymOp%R=det*SymOp%R
	  Rot=CurRot
	  Dir=GenSmbTmp(2:2)
	  !And now determine the translation vector
	  SymOp%t=0.0
	  DO i=3,len_trim(GenSmbTmp)
	    IF (scan(GenSmbTmp(i:i),'123456')==0) THEN
	      CALL HallTransSmb2Vec(GenSmbTmp(i:i),CurTrVec)
	    ELSE
	      read(GenSmbTmp(i:i),'(i1)') FracTrans
              CurTrVec=0.0
	      SELECT CASE (Dir)
	        CASE ('x')
		  CurTrVec(1)=float(FracTrans)/Rot
	        CASE ('y')
		  CurTrVec(2)=float(FracTrans)/Rot
	        CASE ('z')
		  CurTrVec(3)=float(FracTrans)/Rot
		CASE DEFAULT
                  write(*,*) 'Invalid translation-part identificator in the Hall symbol:'//trim(GenSmbTmp)
                  STOP
	      END SELECT	      
	    ENDIF
            SymOp%t=SymOp%t+CurTrVec
	  ENDDO
	end subroutine HallGenSmb2SymOp
	
!*********************************************************************
	subroutine HallTransSmb2Vec(transsmb,Vec)
	  IMPLICIT NONE
	  character :: TransSmb
	  real, dimension(1:3) :: Vec
	  SELECT CASE(TransSmb)
	    CASE('a')
	      Vec=(/0.5,0.0,0.0/)
	    CASE('b')
	      Vec=(/0.0,0.5,0.0/)
	    CASE('c')
	      Vec=(/0.0,0.0,0.5/)
	    CASE('n')
	      Vec=(/0.5,0.5,0.5/)
	    CASE('u')
	      Vec=(/0.25,0.0,0.0/)
	    CASE('v')
	      Vec=(/0.0,0.25,0.0/)
	    CASE('w')
	      Vec=(/0.0,0.0,0.25/)
	    CASE('d')
	      Vec=(/0.25,0.25,0.25/)
	  END SELECT
	 
	end subroutine HallTransSmb2Vec
	
!*********************************************************************
	subroutine HallCenSmb2Centers(CenSym,NCenter,Center)
	  IMPLICIT NONE
	  character :: CenSym
	  integer :: NCenter
	  real, dimension(:,:), allocatable :: Center
          IF (allocated(center)) deallocate(Center)
	  select case(Censym)
	    case ('p')
	      allocate(center(1:3,1:1))
	      center(1:3,1)=(/0.0,0.0,0.0/)
	      NCenter=1
	    case ('a')
	      allocate(center(1:3,1:2))
	      center(1:3,1)=(/0.0,0.0,0.0/)
	      center(1:3,2)=(/0.0,0.5,0.5/)
	      NCenter=2
	    case ('b')
	      allocate(center(1:3,1:2))
	      center(1:3,1)=(/0.0,0.0,0.0/)
	      center(1:3,2)=(/0.5,0.0,0.5/)
	      NCenter=2
	    case ('c')
	      allocate(center(1:3,1:2))
	      center(1:3,1)=(/0.0,0.0,0.0/)
	      center(1:3,2)=(/0.5,0.5,0.0/)
	      NCenter=2
	    case ('i')
	      allocate(center(1:3,1:2))
	      center(1:3,1)=(/0.0,0.0,0.0/)
	      center(1:3,2)=(/0.5,0.5,0.5/)
	      NCenter=2
	    case ('r')
	      allocate(center(1:3,1:3))
	      center(1:3,1)=(/0.0,0.0,0.0/)
	      center(1:3,2)=(/0.666666667,0.333333333,0.333333333/)
	      center(1:3,3)=(/0.333333333,0.666666667,0.666666667/)
	      NCenter=3
	    case ('h')
	      allocate(center(1:3,1:3))
	      center(1:3,1)=(/0.0,0.0,0.0/)
	      center(1:3,2)=(/0.666666667,0.333333333,0.0/)
	      center(1:3,3)=(/0.333333333,0.666666667,0.0/)
	      NCenter=3
	    case ('f')
	      allocate(center(1:3,1:4))
	      center(1:3,1)=(/0.0,0.0,0.0/)
	      center(1:3,2)=(/0.0,0.5,0.5/)
	      center(1:3,3)=(/0.5,0.0,0.5/)
	      center(1:3,4)=(/0.5,0.5,0.0/)
	      NCenter=4
	  end select
	end subroutine HallCenSmb2Centers
	
!*********************************************************************
    	subroutine StandardizeHall(Hall,OrigShift)
  	  IMPLICIT NONE
    	  character*(*) :: Hall
	  real, dimension(3) :: OrigShift
	  integer, dimension(3) :: IOrigShift
  	  integer :: i, ii
  	  do
  	    IF (Hall(1:1)==' ') THEN
  	      Hall=Hall(2:)
  	    ELSE
  	      EXIT
  	    ENDIF
  	  enddo
  	  i=0
	  DO
  	    i=i+1
  	    IF (Hall(i:i)==char(9)) Hall(i:i)=' '
	    IF ((i==len_trim(Hall)).or.(i==len(Hall)-1)) EXIT
  	  ENDDO
  	  i=0
	  DO
  	    i=i+1
  	    IF (Hall(i:i+1)=='  ') THEN
	      Hall(i:)=Hall(i+1:)
  	      i=i-1
	    ENDIF
	    IF ((i==len_trim(Hall)).or.(i==len(Hall)-1)) EXIT
  	  ENDDO
    	  !Set capital leters to lower case
  	  DO i=1,len_trim(Hall)
  	    IF ((iachar(Hall(i:i))>=65).and.(iachar(Hall(i:i))<=90)) Hall(i:i)=achar(iachar(Hall(i:i))+32)
  	  ENDDO
          if (index(Hall,'(') /=0) THEN
	    i=index(Hall,'(')
	    ii=index(Hall,')')
	    IF (ii<i+6) STOP 'Invalid origin-shift part of the Hall Symbol'
	    READ(Hall(i+1:ii-1),*) IOrigShift
	    OrigShift=float(IOrigShift)/12
	  ELSE
	    OrigShift=0.0
	  ENDIF
	  Hall=Hall(1:i-1)
    	end subroutine StandardizeHall
    	
!*********************************************************************
    	function HallCenterSmb(Hall)
  	  IMPLICIT NONE
    	  character*(*) :: Hall
  	  character :: HallCenterSmb
  	  IF (Hall(1:1)=='-') THEN
  	    HallCenterSmb=Hall(2:2)
  	  ELSE
  	    HallCenterSmb=Hall(1:1)      
  	  ENDIF
    	end function HallCenterSmb

!*********************************************************************
    	logical function HallIsInv(Hall)
  	  IMPLICIT NONE
    	  character*(*) :: Hall
  	  IF (Hall(1:1)=='-') THEN
  	    HallIsInv=.true.
  	  ELSE
  	    HallIsInv=.false.
  	  ENDIF   
    	end function HallIsInv

!*********************************************************************
    	integer function HallNGen(Hall)
  	  IMPLICIT NONE
    	  character*(*) :: Hall
  	  integer :: i
  	  !Important: standardized Hall symbol is assumed!
  	  HallNGen=0
  	  DO i=1,len_trim(Hall)
  	    if (Hall(i:i)==' ') HallNGen=HallNGen+1
  	  ENDDO
    	end function HallNGen

!*********************************************************************
    	function HallGen(Hall,GenNr)
  	  IMPLICIT NONE
    	  character*(*) :: Hall
  	  character*(len(Hall)) :: HallGen
  	  integer :: GenNr
  	  integer :: i, pos
  	  !Important: standardized Hall symbol is assumed!  	  
	  HallGen=Hall
	  DO i=1,GenNr
            pos=index(HallGen,' ')
  	    HallGen=HallGen(pos+1:)
  	  ENDDO
	  pos=index(trim(HallGen),' ')
  	  if (pos>0) HallGen=HallGen(:pos-1)
    	end function HallGen

      end module symmetry_tools
!*********************************************************************
      !End of module symmetry_tools
!*********************************************************************
